/*
**	FILE:			LgiRes.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			3/8/99
**	DESCRIPTION:	Lgi Resource Editor
**
**	Copyright (C) 1999, Matthew Allen
**		fret@ozemail.com.au
*/

#include "LgiRes.h"
#include "LgiRes_Dialog.h"
#include "LgiRes_Menu.h"

char AppName[]				= "Lgi Resource Editor";
char HelpFile[]				= "Help.html";
char OptionsFileName[]		= "Options.r";
#define VIEW_PULSE_RATE		100

char *TypeNames[] = {
	"Folder",
	"Bitmap",
	"Icon",
	"Cursor",
	"Dialog",
	"String",
	"Menu",
	0};

//////////////////////////////////////////////////////////////////////////////
GImageList *OpenImageList(char *FileName)
{
	GImageList *ImageList = 0;
	if (FileName)
	{
		GSurface *pDC = LoadDC(FileName);
		if (pDC)
		{
			ImageList = NEW(GImageList(16, 16, pDC));
			if (ImageList)
			{
				ImageList->Create(pDC->X(), pDC->Y(), GdcD->GetBits());
			}
		}

		DeleteObj(pDC);
	}
	return ImageList;
}

//////////////////////////////////////////////////////////////////////////////
Resource::Resource(AppWnd *w, int t, bool enabled)
{
	AppWindow = w;
	ResType = t;
	Item = 0;
	SysObject = false;
}

Resource::~Resource()
{
	DeleteObj(Item);
}

bool Resource::IsSelected()
{
	return Item?Item->Selected():false;
}

bool Resource::Attach(GView *Parent)
{
	GView *w = Wnd();
	if (w)
	{
		return w->Attach(Parent);
	}
	return false;
}

//////////////////////////////////////////////////////////////////////////////
ResFolder::ResFolder(AppWnd *w, int t, bool enabled) : Resource(w, t, enabled)
{
	Wnd()->Name("<untitled>");
	Wnd()->Enabled(enabled);
}

//////////////////////////////////////////////////////////////////////////////
ObjTreeItem::ObjTreeItem(Resource *Object)
{
	Obj = Object;
}

ObjTreeItem::~ObjTreeItem()
{
}

char *ObjTreeItem::GetText(int i)
{
	if (Obj)
	{
		int Type = Obj->Type();
		if (Type > 0)
		{
			return Obj->Wnd()->Name();
		}
		else
		{
			return TypeNames[-Type];
		}
	}

	return "#NO_OBJ";
}

int ObjTreeItem::GetImage(int Flags)
{
	if (Obj)
	{
		int Type = Obj->Type();
		if (Type > 0)
		{
			return Obj->Type();
		}
		else
		{
			if (Obj->Wnd()->Enabled())
			{
				return 0;
			}
			else
			{
				return TYPE_DIR_DISABLED;
			}
		}
	}

	return 0;
}

void ObjTreeItem::OnSelect()
{
	if (Obj)
	{
		Obj->App()->OnResourceSelect(Obj);
	}
}

void ObjTreeItem::OnMouseClick(GMouse &m)
{
	if (NOT Obj) return;
	if (m.Right() AND m.Down())
	{
		Tree->Select(this);
		GSubMenu *RClick = NEW(GSubMenu);
		if (RClick)
		{
			if (Obj->Wnd()->Enabled())
			{
				if (Obj->Type() > 0)
				{
					// Resource
					RClick->AppendItem("Delete", IDM_DELETE, NOT Obj->SystemObject());
					RClick->AppendItem("Rename", IDM_RENAME, NOT Obj->SystemObject());
				}
				else
				{
					// Folder
					RClick->AppendItem("New", IDM_NEW, true);
					RClick->AppendSeparator();
					GSubMenu *Insert = RClick->AppendSub("Import from...");
					if (Insert)
					{
						Insert->AppendItem("Lgi File", IDM_IMPORT, true);
						Insert->AppendItem("Win32 Resource Script", IDM_IMPORT_WIN32, false);
					}
				}

				// Custom entries
				if (NOT Obj->SystemObject())
				{
					Obj->OnRightClick(RClick);
				}
			}
			else
			{
				RClick->AppendItem("Not implemented", 0, false);
			}

			if (Tree->GetMouse(m, true))
			{
				int Cmd = 0;
				switch (Cmd = RClick->Float(Tree, m.x, m.y))
				{
					case IDM_NEW:
					{
						Obj->App()->NewObject(-Obj->Type());
						break;
					}
					case IDM_DELETE:
					{
						Obj->App()->SetDirty(true);
						Obj->App()->DelObject(Obj);
						break;
					}
					case IDM_RENAME:
					{
						GInput Dlg(Tree, GetText(), "Enter the name for the object", "Object Name");
						if (Dlg.DoModal())
						{
							Obj->Wnd()->Name(Dlg.Str);
							Update();
							Obj->App()->SetDirty(true);
						}
						break;
					}
					case IDM_IMPORT:
					{
						GFileSelect Select;
						Select.Parent(Obj->App());
						Select.Type("Text", "*.txt");
						if (Select.Open())
						{
							GFile F;
							if (F.Open(Select.Name(), O_READ))
							{
								Resource *Res = Obj->App()->NewObject(-Obj->Type());
								if (Res)
								{
									// TODO
									// Res->Read();
								}
							}
							else
							{
								LgiMsg(Obj->App(), "Couldn't open file for reading.");
							}
						}
						break;
					}
					case IDM_IMPORT_WIN32:
					{
						/*
						List<ResDialog> l;
						if (ImportWin32Dialogs(l, MainWnd))
						{
							for (ResDialog *r = l.First(); r; r = l.Next())
							{
								Obj->App()->InsertObject(TYPE_DIALOG, r);
							}
						}
						*/
						break;
					}
					default:
					{
						Obj->OnCommand(Cmd);
						break;
					}
				}
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
FieldView::FieldView()
{
	Dlg = this;
	Source = 0;
	Ignore = true;

	#ifdef WIN32
	WndStyle |= WS_TABSTOP;
	WndExStyle |= WS_EX_CONTROLPARENT;
	#endif
}

FieldView::~FieldView()
{
}

void FieldView::Serialize(bool Write)
{
	Ignore = NOT Write;
	ProcessFields(Fields, Write);
	Ignore = false;
}

void FieldView::OnSelect(FieldSource *s)
{
	Ignore = true;
	if (Source)
	{
		Options = 0;

		// remove current fields
		for (DataDlgField *f = Fields.First(); f; f = Fields.Next())
		{
			DeleteObj(f);
		}
		
		Fields.Empty();

		// remove all children
		for (GView *c = Children.First(); c; c = Children.First())
		{
			c->Detach();
			DeleteObj(c);
		}
	}

	if (s)
	{
		// Add new fields
		Source = s;
		Options = &Source->Values;

		if (Source->GetFields(Fields))
		{
			int y = 10;
			for (DataDlgField *f = Fields.First(); f; f = Fields.Next())
			{
				switch (f->GetType())
				{
					case DATA_STR:
					case DATA_FLOAT:
					case DATA_INT:
					{
						Children.Insert(NEW(GText(-1, 10, y, 50, 20, f->GetDesc())));
						Children.Insert(NEW(GEdit(f->GetCtrl(), 70, y, 100, 18, "")));
						break;
					}
					case DATA_BOOL:
					{
						Children.Insert(NEW(GCheckBox(f->GetCtrl(), 70, y, 100, 20, f->GetDesc())));
						break;
					}
				}

				y += 25;
			}

			AttachChildren();
			OnPosChange();
		}

		Serialize(false);
		Ignore = false;
	}
}

void FieldView::OnPosChange()
{
	for (GView *w = Children.First(); w; w = Children.Next())
	{
		GEdit *e = dynamic_cast<GEdit*>(w);
		if (e)
		{
			GRegion r = e->GetPos();
			r.x2 = X()-10;
			e->SetPos(r);
		}
	}
}

int FieldView::OnEvent(GMessage *m)
{
	return GLayout::OnEvent(m);
}

int FieldView::OnNotify(GView *Ctrl, int Flags)
{
	if (NOT Ignore)
	{
		for (DataDlgField *f = Fields.First(); f; f = Fields.Next())
		{
			if (Ctrl->GetId() == f->GetCtrl())
			{
				// write the values back
				Serialize(true);
				Source->OnFieldChange();

				break;
			}
		}
	}

	return 0;
}

void FieldView::OnPaint(GSurface *pDC)
{
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle();
}

//////////////////////////////////////////////////////////////////////////////
ObjContainer::ObjContainer(AppWnd *w) :
	GTree(100, 0, 0, 100, 100, "LgiResObjTree")
{
	Window = w;
	Sunken(false);

	Insert(Bitmaps = NEW(ObjTreeItem( NEW(ResFolder(Window, -TYPE_BITMAP, false)) ))); 
	Insert(Icons =   NEW(ObjTreeItem( NEW(ResFolder(Window, -TYPE_ICON, false)) )));
	Insert(Cursors = NEW(ObjTreeItem( NEW(ResFolder(Window, -TYPE_CURSOR, false)) )));
	Insert(Dialogs = NEW(ObjTreeItem( NEW(ResFolder(Window, -TYPE_DIALOG)) )));
	Insert(Strings = NEW(ObjTreeItem( NEW(ResFolder(Window, -TYPE_STRING)) )));
	Insert(Menus =   NEW(ObjTreeItem( NEW(ResFolder(Window, -TYPE_MENU)) )));

	char *f = LgiFindFile("icons.bmp");
	if (f)
	{
		Images = OpenImageList(f);
		DeleteArray(f);
		SetImageList(Images, false);
	}
}

ObjContainer::~ObjContainer()
{
	DeleteObj(Images);
}

bool ObjContainer::AppendChildren(ObjTreeItem *Res, List<Resource> &Lst)
{
	bool Status = true;
	if (Res)
	{
		GTreeItem *Item = Res->GetChild();
		while (Item)
		{
			ObjTreeItem *i = dynamic_cast<ObjTreeItem*>(Item);
			if (i) Lst.Insert(i->GetObj());
			else Status = false;
			Item = Item->GetNext();
		}
	}
	return Status;
}

bool ObjContainer::ListObjects(List<Resource> &Lst)
{
	bool Status = AppendChildren(Bitmaps, Lst);
	Status &= AppendChildren(Icons, Lst);
	Status &= AppendChildren(Cursors, Lst);
	Status &= AppendChildren(Dialogs, Lst);
	Status &= AppendChildren(Strings, Lst);
	Status &= AppendChildren(Menus, Lst);
	return Status;
}

//////////////////////////////////////////////////////////////////////////////

#define IDM_WEB				100
#define IDM_BMP				101

class AboutDlg : public GDialog {

	char Text[512];

public:
	AboutDlg(GView *parent)
	{
		Parent = parent;
		Pos.ZOff(380, 195);
		MoveToCenter();
		
		char Str[256];
		sprintf(Str, "About %s", AppName);
		Name(Str);

		char *Os = "Default";
		#ifdef WIN32
		Os = "Win32";
		#endif
		#ifdef BEOS
		Os = "BeOS";
		#endif
		#ifdef LINUX
		Os = "Linux";
		#endif

		char *Build = "Release";
		#ifdef _DEBUG
		Build = "Debug";
		#endif

		sprintf(Text,
			"%s v%02.2f (%s %s)\n"
			"Build: %s, %s\n"
			"\n"
			"http://www.ozemail.com.au/~fret\n"
			"fret@ozemail.com.au\n"
			"\n"
			" 1999 Matthew Allen",
			AppName,
			VER,
			Os,
			Build,
			__DATE__,
			__TIME__);

		char *FileName = LgiFindFile("about2.bmp");
		if (FileName)
		{
			Children.Insert(NEW(GBitmap(IDM_BMP, 10, 10, FileName)));
		}

		Children.Insert(NEW(GText(IDM_STATIC, 180, 10, 190, 145, Text)));
		Children.Insert(NEW(GButton(IDM_WEB, 200, 148, 90, 20, "Webpage")));
		Children.Insert(NEW(GButton(IDOK, 300, 148, 50, 20, "Ok")));

		DoModal();
	}

	int OnNotify(GView *Ctrl, int Flags)
	{
		switch (Ctrl->GetId())
		{
			case IDM_WEB:
			{
				LgiExecute("http://www.ozemail.com.au/~fret", NULL, NULL);
				break;
			}
			case IDOK:
			{
				EndModal(0);
				break;
			}
		}
		return 0;
	}
};

//////////////////////////////////////////////////////////////////////////////
AppWnd::AppWnd() :
	GDocApp(AppName, MAKEINTRESOURCE(IDI_ICON1))
{
	LastRes = 0;
	Fields = 0;

	if (_Create())
	{
		SetupUi();
	}
}

AppWnd::~AppWnd()
{
	OnResourceSelect(NULL);
	_Destroy();
}

void AppWnd::SetupUi()
{
	if (_LoadMenu())
	{
		if (_FileMenu)
		{
			_FileMenu->AppendItem("Import Win32 Script", IDM_IMPORT_WIN32, true, 5);
			_FileMenu->AppendItem("Import LgiRes Language", IDM_IMPORT_LANG, true, 6);
			_FileMenu->AppendSeparator(7);
		}

		GSubMenu *NewObj = Menu->AppendSub("&New");
		if (NewObj)
		{
			NewObj->AppendItem("&Dialog", IDM_NEW_DLG, true);
			NewObj->AppendItem("&String Group", IDM_NEW_STRGRP, true);
			NewObj->AppendItem("&Menu", IDM_NEW_MENU, true);
		}

		/*
		Edit = Menu->AppendSub("&Edit");
		if (Edit)
		{
			Edit->AppendItem("&Undo", IDM_UNDO, TRUE);
			Edit->AppendItem("&Redo", IDM_REDO, TRUE);
			Edit->AppendSeparator();
			Edit->AppendItem("&Copy", IDM_COPY, TRUE);
			Edit->AppendItem("&Paste", IDM_PASTE, TRUE);
		}
		*/

		Help = Menu->AppendSub("&Help");
		if (Help)
		{
			Help->AppendItem("&Help",		IDM_HELP, TRUE);
			Help->AppendSeparator();
			Help->AppendItem("&About",		IDM_ABOUT, TRUE);
		}

		#ifdef WIN32
		Menu->Attach(this);
		#endif
	}

	Status = 0;
	StatusInfo[0] = StatusInfo[1] = 0;
	/*
	Status = NEW(GStatusBar);
	if (Status)
	{
		Status->Attach(this);

		StatusInfo[0] = Status->AppendPane("", 2);
		StatusInfo[1] = Status->AppendPane("", 160);
		if (StatusInfo[1]) StatusInfo[1]->Sunken(TRUE);
	}
	*/

	MainSplit = NEW(GSplitter);
	if (MainSplit)
	{
		MainSplit->Attach(this);
		MainSplit->Split(240);

		SubSplit = NEW(GSplitter);
		if (SubSplit)
		{
			MainSplit->SetViewA(SubSplit, false);
			MainSplit->SetViewB(NULL);
			SubSplit->Border(false);
			SubSplit->IsVertical(false);
			SubSplit->Split(200);

			SubSplit->SetViewA(Objs = NEW(ObjContainer(this)));
			if (Objs)
			{
				Objs->AskImage(true);
				Objs->AskText(true);
			}
			SubSplit->SetViewB(Fields = NEW(FieldView));
		}
	}

	Visible(TRUE);

	char Opt[256];
	if (LgiApp->GetOption(Opt, 'o'))
	{
		LoadLgi(Opt);
	}
}

int AppWnd::OnEvent(GMessage *m)
{
	GMru::OnEvent(m);

	switch (MsgCode(m))
	{
		#ifdef WIN32
		case WM_NCACTIVATE:
		{	
			GApp::ActiveDialog = (m->a) ? Handle() : 0;
			break;
		}
		#endif
		case WM_CHANGE:
		{
			return OnNotify((GView*) m->a, m->b);
		}
		case WM_DESCRIBE:
		{
			char *Text = (char*) m->b;
			if (Text)
			{
				SetStatusText(Text, STATUS_NORMAL);
			}
			break;
		}
	}
	return GWindow::OnEvent(m);
}

int AppWnd::OnCommand(int Cmd, int Event, HWND Handle)
{
	switch (Cmd)
	{
		case IDM_NEW_DLG:
		{
			NewObject(TYPE_DIALOG);
			break;
		}
		case IDM_NEW_STRGRP:
		{
			NewObject(TYPE_STRING);
			break;
		}
		case IDM_NEW_MENU:
		{
			NewObject(TYPE_MENU);
			break;
		}
		case IDM_CLOSE:
		{
			Empty();
			break;
		}
		case IDM_IMPORT_WIN32:
		{
			LoadWin32();
			break;
		}
		case IDM_IMPORT_LANG:
		{
			ImportLang();
			break;
		}
		case IDM_EXIT:
		{
			PostQuitMessage(0);
			break;
		}
		case IDM_HELP:
		{
			char ExeName[256];
			BOOL Done = FALSE;

			GetModuleFileName(NULL, ExeName, sizeof(ExeName));

			while (NOT Done)
			{
				if (LgiTrimDir(ExeName))
				{
					if (LgiExecute(HelpFile, NULL, ExeName))
					{
						strcat(ExeName, "\\help");
						if (LgiExecute(HelpFile, NULL, ExeName))
						{
							Done = TRUE;
						}
						
						LgiTrimDir(ExeName);
					}
					else
					{
						Done = TRUE;
					}
				}
				else
				{
					Done = TRUE;
				}
			}
			break;
		}
		case IDM_ABOUT:
		{
			AboutDlg Dlg(this);
			break;
		}
	}

	return GDocApp::OnCommand(Cmd, Event, Handle);
}

int AppWnd::OnNotify(GView *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
		default:
		{
			break;
		}
	}
	return 0;
}

int AppWnd::GetUniqueStrRef(int	Start)
{
	int Ref = Start;

	if (Objs)
	{
		List<Resource> l;
		if (Objs->ListObjects(l))
		{
			while (true)
			{
				bool Found = false;
				for (Resource *r = l.First(); r AND NOT Found; r = l.Next())
				{
					ResStringGroup *Grp = dynamic_cast<ResStringGroup*>(r);
					if (Grp)
					{
						Found = Grp->FindRef(Ref);
					}
				}

				if (Found)
				{
					Ref++;
				}
				else
				{
					break;
				}
			}
		}
	}

	return Ref;
}

ResString *AppWnd::GetStrFromRef(int Ref)
{
	ResString *Str = 0;

	if (Objs)
	{
		List<Resource> l;
		if (Objs->ListObjects(l))
		{
			for (Resource *r = l.First(); r AND NOT Str; r = l.Next())
			{
				ResStringGroup *Grp = dynamic_cast<ResStringGroup*>(r);
				if (Grp)
				{
					Str = Grp->FindRef(Ref);
				}
			}
		}
	}

	return Str;
}

ResStringGroup *AppWnd::GetDialogSymbols()
{
	if (Objs)
	{
		List<Resource> l;
		if (Objs->ListObjects(l))
		{
			for (Resource *r = l.First(); r; r = l.Next())
			{
				ResStringGroup *Grp = dynamic_cast<ResStringGroup*>(r);
				if (Grp)
				{
					char *ObjName = Grp->Wnd()->Name();
					if (stricmp(ObjName, StrDialogSymbols) == 0)
					{
						return Grp;
					}
				}
			}
		}
	}

	return NULL;
}

void AppWnd::OnReceiveFiles(List<char> &Files)
{
	char *f = Files.First();
	if (f)
	{
		LoadLgi(f);
	}
}

void AppWnd::SetStatusText(char *Text, int Pane)
{
	if (Pane >= 0 AND Pane < STATUS_MAX AND StatusInfo[Pane])
	{
		StatusInfo[Pane]->Name(Text);
	}
}

Resource *AppWnd::NewObject(int Type, bool Select)
{
	Resource *r = 0;
	ObjTreeItem *Dir = 0;

	switch (Type)
	{
		case TYPE_BITMAP:
		{
			break;
		}
		case TYPE_ICON:
		{
			break;
		}
		case TYPE_CURSOR:
		{
			break;
		}
		case TYPE_DIALOG:
		{
			r = NEW(ResDialog(this));
			Dir = Objs->Dialogs;
			break;
		}
		case TYPE_STRING:
		{
			r = NEW(ResStringGroup(this));
			Dir = Objs->Strings;
			break;
		}
		case TYPE_MENU:
		{
			r = NEW(ResMenu(this));
			Dir = Objs->Menus;
  			break;
		}
	}

	if (r)
	{
		ObjTreeItem *Item = NEW(ObjTreeItem(r));
		if (Item)
		{
			r->Item = Item;
			Dir->Insert(Item);
			Dir->Update();
			Dir->Expanded(true);
		
			if (Select)
			{
				Objs->Select(Item);
			}
		}

		r->Create();

		if (Item)
		{
			Item->Update();
		}

		SetDirty(true);
	}

	return r;
}

bool AppWnd::InsertObject(int Type, Resource *r, bool Select)
{
	bool Status = false;
	if (r)
	{
		ObjTreeItem *Dir = 0;

		switch (Type)
		{
			case TYPE_BITMAP:
			{
				Dir = Objs->Bitmaps;
				break;
			}
			case TYPE_ICON:
			{
				Dir = Objs->Icons;
				break;
			}
			case TYPE_CURSOR:
			{
				Dir = Objs->Cursors;
				break;
			}
			case TYPE_DIALOG:
			{
				Dir = Objs->Dialogs;
				break;
			}
			case TYPE_STRING:
			{
				Dir = Objs->Strings;
				break;
			}
			case TYPE_MENU:
			{
				Dir = Objs->Menus;
				break;
			}
		}

		if (Dir)
		{
			ObjTreeItem *Item = NEW(ObjTreeItem(r));
			if (Item)
			{
				char *Name = Item->GetText();

				r->Item = Item;
				Dir->Insert(Item, (Name AND Name[0] == '_') ? 0 : -1);
				Dir->Update();
				Dir->Expanded(true);
				
				if (Select)
				{
					Objs->Select(Item);
				}

				Status = true;
			}
		}
	}

	return Status;
}

void AppWnd::DelObject(Resource *r)
{
	OnResourceSelect(0);
	DeleteObj(r);
}

bool AppWnd::ListObjects(List<Resource> &Lst)
{
	if (Objs)
	{
		return Objs->ListObjects(Lst);
	}

	return false;
}

void AppWnd::OnObjChange(FieldSource *r)
{
	if (Fields)
	{
		Fields->Serialize(false);
		SetDirty(true);
	}
}

void AppWnd::OnObjSelect(FieldSource *r)
{
	if (Fields)
	{
		Fields->OnSelect(r);
	}
}

void AppWnd::OnResourceSelect(Resource *r)
{
	if (LastRes)
	{
		OnObjSelect(NULL);
		MainSplit->SetViewB(0);
		LastRes->DestroyUI();
		LastRes = 0;
	}
	
	if (r)
	{
		GView *Wnd = r->CreateUI();
		if (Wnd)
		{
			if (SubSplit)
			{
				MainSplit->SetViewB(Wnd, false);
			}

			Wnd->Invalidate();
			LastRes = r;
		}
	}
}

void AppWnd::ImportLang()
{
	// open dialog
	GFileSelect Select;

	Select.Parent(this);
	Select.Type("Lgi Resource", "*.lr");
	if (Select.Open())
	{
		GFile F;
		if (F.Open(Select.Name(), O_READ))
		{
			// read the file in
			int Len = F.GetSize();
			char *Data = NEW(char[Len+1]);
			if (Data)
			{
				F.Read(Data, Len);
				Data[Len] = 0;

				// convert file to Xml objects
				Xml Xml;
				if (Xml.ParseXml(Data))
				{
					ResStringGroup Grp(this);
					List<ResMenu> Menus;

					List<ResString> Strings;
					for (XmlTag *t = Xml.Values.First(); t; )
					{
						if (stricmp(t->Name, "menu") == 0)
						{
							ResMenu *Menu = NEW(ResMenu(this));
							if (Menu AND Menu->Read(Xml.Values))
							{
								Menus.Insert(Menu);
							}
						}
						else if (stricmp(t->Name, "string") == 0)
						{
							ResString *s = NEW(ResString(&Grp));
							if (s AND s->Read(Xml.Values))
							{
								Strings.Insert(s);
							}
						}

						t = Xml.Values.Next();
					}

					Grp.SetLanguages();

					if (Strings.GetItems() > 0 AND
						Grp.GetLanguages() > 0)
					{
						List<LangInfo> Langs;
						for (int i=0; i<Grp.GetLanguages(); i++)
						{
							Langs.Insert(Grp.GetLanguage(i));
						}

						LangDlg Dlg(this, Langs);
						if (Dlg.DoModal() == IDOK AND
							Dlg.Lang)
						{
							int Matches = 0;
							int NotFound = 0;
							int Imported = 0;
							for (ResString *s=Strings.First(); s; s=Strings.Next())
							{
								ResString *d=GetStrFromRef(s->Ref);
								if (d)
								{
									Matches++;

									char *Str = s->Get(Dlg.Lang->Id);
									if (Str)
									{
										d->Set(Str, Dlg.Lang->Id);
										Imported++;
									}
								}
								else
								{
									NotFound++;
								}
							}

							List<Resource> Lst;
							if (ListObjects(Lst))
							{
								for (ResMenu *m=Menus.First(); m; m=Menus.Next())
								{
									// find matching menu in our list
									ResMenu *Match = 0;
									for (Resource *r=Lst.First(); r; r=Lst.Next())
									{
										ResMenu *n = dynamic_cast<ResMenu*>(r);
										if (n AND stricmp(n->Name(), m->Name()) == 0)
										{
											Match = n;
											break;
										}
									}

									if (Match)
									{
										// match strings
										List<ResString> *Src = m->GetStrs();
										List<ResString> *Dst = Match->GetStrs();
										
										for (ResString *s=Src->First(); s; s = Src->Next())
										{
											for (ResString *d=Dst->First(); d; d=Dst->Next())
											{
												if (s->Ref == d->Ref)
												{
													char *Str = s->Get(Dlg.Lang->Id);
													if (Str)
													{
														d->Set(Str, Dlg.Lang->Id);
													}
													break;
												}
											}
										}

										Match->SetLanguages();
									}
								}

								for (Resource *r=Lst.First(); r; r=Lst.Next())
								{
									ResStringGroup *StrRes = dynamic_cast<ResStringGroup*>(r);
									if (StrRes)
									{
										StrRes->SetLanguages();
									}
								}
							}

							LgiMsg(this, "Imported: %i\nMatched: %i\nNot matched: %i\nTotal: %i\n\nImport complete.", AppName, MB_OK, Imported, Matches, NotFound, Matches+NotFound);
						}
					}
					else
					{
						LgiMsg(this, "No language information to import", AppName, MB_OK);
					}

					for (ResString *s=Strings.First(); s; s=Strings.Next())
					{
						Grp.DeleteStr(s);
					}
				}

				DeleteArray(Data);
			}
		}
	}
}

void AppWnd::Empty()
{
	// Delete any existing objects
	List<Resource> l;
	if (ListObjects(l))
	{
		for (Resource *r = l.First(); r; )
		{
			if (r->SystemObject())
			{
				l.Delete(r);
				r = l.Current();
			}
			else
			{
				r=l.Next();
			}
		}

		for (r = l.First(); r; r=l.Next())
		{
			DelObject(r);
		}
	}
}

bool AppWnd::OnRequestClose()
{
	return SetDirty(false);
}

bool AppWnd::OpenFile(char *FileName)
{
	if (stristr(FileName, ".lr"))
	{
		return LoadLgi(FileName);
	}
	else if (stristr(FileName, ".rc"))
	{
		return LoadWin32(FileName);
	}

	return false;
}

bool AppWnd::SaveFile(char *FileName)
{
	if (stristr(FileName, ".lr"))
	{
		return SaveLgi(FileName);
	}
	else if (stristr(FileName, ".rc"))
	{
	}

	return false;
}

void AppWnd::GetFileTypes(GFileSelect *Dlg, bool Write)
{
	Dlg->Type("Lgi Resource", "*.lr");
	if (NOT Write)
	{
		Dlg->Type("All Files", "*.*");
	}
}

// Lgi load/save
bool AppWnd::LoadLgi(char *FileName)
{
	bool Status = false;

	Empty();
	if (FileName)
	{
		GFile f;
		if (f.Open(FileName, O_READ))
		{
			GProgressDlg Progress(this);

			Progress.SetDescription("Initializing...");
			Progress.SetType("Tags");

			int Len = f.GetSize();
			char *Data = NEW(char[Len+1]);
			if (Data)
			{
				f.Read(Data, Len);
				Data[Len] = 0;

				// convert file to Xml objects
				Xml Xml;
				Progress.SetDescription("Lexing...");
				if (Xml.ParseXml(Data))
				{
					Progress.SetLimits(0, Xml.Values.GetItems()-1);

					// convert Xml list into objects
					Xml.Values.First();
					XmlTag *This = 0;
					int i=0;
					DoEvery Timer(500);
					for (XmlTag *t = Xml.Values.Current(); t AND t != This; t = Xml.Values.Current(), i++)
					{
						if (Timer.DoNow())
						{
							Progress.Value(Xml.Values.IndexOf(t));
							#ifdef WIN32
							GApp::ObjInstance()->Run(false);
							#endif
						}

						This = t;
						int RType = 0;
						if (stricmp(t->Name, "dialog") == 0)
						{
							RType = TYPE_DIALOG;
						}
						else if (stricmp(t->Name, "string-group") == 0)
						{
							RType = TYPE_STRING;
						}
						else if (stricmp(t->Name, "menu") == 0)
						{
							RType = TYPE_MENU;
						}

						if (RType > 0)
						{
							Resource *r = NewObject(RType, false);
							if (r)
							{
								r->Read(Xml.Values);
							}
						}
					}

					SortDialogs();
					Status = true;
				}

				// clean up
				DeleteArray(Data);
			}
		}
	}

	return Status;
}

int DialogNameCompare(ResDialog *a, ResDialog *b, int Data)
{
	char *A = (a)?a->Name():0;
	char *B = (b)?b->Name():0;
	if (A AND B) return stricmp(A, B);
	return -1;
}

void AppWnd::SortDialogs()
{
	List<Resource> Lst;
	if (ListObjects(Lst))
	{
		List<ResDialog> Dlgs;
		for (Resource *r = Lst.First(); r; r = Lst.Next())
		{
			ResDialog *Dlg = dynamic_cast<ResDialog*>(r);
			if (Dlg)
			{
				Dlgs.Insert(Dlg);
				Dlg->Item->Remove();
			}
		}

		Dlgs.Sort(DialogNameCompare, 0);

		for (ResDialog *d = Dlgs.First(); d; d = Dlgs.Next())
		{
			Objs->Dialogs->Insert(d->Item);
		}
	}
}

class ResTreeNode
{
public:
	char *Str;
	ResTreeNode *a, *b;
	ResTreeNode(char *s)
	{
		a = b = 0;
		Str = s;
	}

	~ResTreeNode()
	{
		DeleteArray(Str);
		DeleteObj(a);
		DeleteObj(b);
	}

	void List(List<char> &l)
	{
		if (a)
		{
			a->List(l);
		}

		if (Str)
		{
			l.Insert(Str);
		}

		if (b)
		{
			b->List(l);
		}
	}

	bool Add(char *s)
	{
		int Comp = (Str AND s) ? stricmp(Str, s) : -1;
		if (Comp == 0)
		{
			return false;
		}

		if (Comp < 0)
		{
			if (a)
			{
				return a->Add(s);
			}
			else
			{
				a = NEW(ResTreeNode(s));
			}
		}

		if (Comp > 0)
		{
			if (b)
			{
				return b->Add(s);
			}
			else
			{
				b = NEW(ResTreeNode(s));
			}
		}
		return true;
	}
};

class ResTree {

	ResTreeNode *Root;

public:
	ResTree()
	{
		Root = 0;
	}

	~ResTree()
	{
		DeleteObj(Root);
	}

	bool Add(char *s)
	{
		if (s)
		{
			if (NOT Root)
			{
				Root = NEW(ResTreeNode(NewStr(s)));
				return true;
			}
			else
			{
				return Root->Add(NewStr(s));
			}
		}

		return false;
	}

	void List(List<char> &l)
	{
		if (Root)
		{
			Root->List(l);
		}
	}
};

char *HeaderStr = "// This file generated by LgiRes\r\n\r\n";
	
bool AppWnd::WriteDefines(List<Resource> &StringGrps, GFile &Defs)
{
	bool Status = false;
	ResTree Tree;

	// Empty file
	Defs.Seek(0, SEEK_SET);
	Defs.SetSize(0);
	Defs.Write(HeaderStr, strlen(HeaderStr));

	// make a unique list of #define's
	List<Resource> Lst;
	if (ListObjects(Lst))
	{
		List<ResString> Final;
		for (Resource *r = Lst.First(); r; r = Lst.Next())
		{
			List<ResString> *StrList = r->GetStrs();
			if (StrList)
			{
				Status = true;
				for (ResString *s = StrList->First(); s; s = StrList->Next())
				{
					if (Tree.Add(s->Define))
					{
						Final.Insert(s);
					}
				}
			}
		}

		// write the list out
		for (ResString *s = Final.First(); s; s = Final.Next())
		{
			if (strlen(s->Define) > 0 AND
				stricmp(s->Define, "IDOK") != 0 AND
				stricmp(s->Define, "IDCANCEL") != 0 AND
				stricmp(s->Define, "IDC_STATIC") != 0 AND
				stricmp(s->Define, "-1") != 0)
			{
				int Tabs = (43 - strlen(s->Define)) / 4;
				char Tab[32];
				ZeroObj(Tab);
				for (int i=0; i<Tabs; i++) Tab[i] = '\t';
				Defs.Print("#define %s%s%i\r\n", s->Define, Tab, s->Id);
			}
		}
	}

	return Status;
}

bool AppWnd::SaveLgi(char *FileName)
{
	bool Status = false;
	if (FileName)
	{
		GFile f;
		GFile Defs;
		
		char DefsName[256];
		strcpy(DefsName, FileName);
		LgiTrimDir(DefsName);
		strcat(DefsName, DIR_STR);
		strcat(DefsName, "resdefs.h");

		if (f.Open(FileName, O_WRITE) AND
			Defs.Open(DefsName, O_WRITE))
		{
			f.SetSize(0);
			Defs.SetSize(0);
			Defs.Print("// Generated by LgiRes\r\n\r\n");

			List<Resource> l;
			if (ListObjects(l))
			{
				// Remove all duplicate symbol Id's from the dialogs
				for (Resource *r = l.First(); r; r = l.Next())
				{
					ResDialog *Dlg = dynamic_cast<ResDialog*>(r);
					if (Dlg)
					{
						Dlg->CleanSymbols();
					}
				}

				// Write all string lists out first so that when we load objects
				// back in again the strings will already be loaded and can
				// be referenced
				List<Resource> StringGrps;
				for (r = l.First(); r; r = l.Next())
				{
					if (r->Type() == TYPE_STRING)
					{
						r->Write(f);
						f.Write("\n", 1);
						StringGrps.Insert(r);
					}
				}

				// write defines
				WriteDefines(StringGrps, Defs);

				// now write the rest of the objects out
				for (r = l.First(); r; r = l.Next())
				{
					if (r->Type() != TYPE_STRING)
					{
						r->Write(f);
						f.Write("\n", 1);
					}
				}

				Status = true;
			}
		}
		else
		{
			LgiMsg(this,
					"Couldn't open these files for output:\n"
					"\t%s\n"
					"\t%s\n"
					"\n"
					"Maybe they are read only or locked by another application.",
					AppName,
					MB_OK,
					FileName,
					DefsName);
		}
	}

	return Status;
}

// Win32 load/save
#define ADJUST_CTRLS_X				2
#define ADJUST_CTRLS_Y				12

#define IMP_MODE_SEARCH				0
#define IMP_MODE_DIALOG				1
#define IMP_MODE_DLG_CTRLS			2
#define IMP_MODE_STRINGS			3
#define IMP_MODE_MENU				4

#include "GToken.h"

class ImportDefine
{
public:
	char *Name;
	char *Value;

	ImportDefine()
	{
		Name = Value = 0;
	}

	ImportDefine(char *Line)
	{
		Name = Value = 0;
		if (Line AND *Line == '#')
		{
			Line++;
			if (strnicmp(Line, "define", 6) == 0)
			{
				Line += 6;
				Line = Skip(Line);
				char *Start = Line;
				char *WhiteSpace = " \r\n\t";
				while (*Line AND NOT strchr(WhiteSpace, *Line))
				{
					Line++;
				}

				Name = NewStr(Start, (int)Line-(int)Start);
				Line = Skip(Line);
				Start = Line;
				while (*Line AND NOT strchr(WhiteSpace, *Line))
				{
					Line++;
				}
				if (Start != Line)
				{
					Value = NewStr(Start, (int)Line-(int)Start);
				}
			}
		}
	}

	~ImportDefine()
	{
		DeleteArray(Name);
		DeleteArray(Value);
	}
};

class DefineList : public List<ImportDefine>
{
	int NestLevel;

public:
	bool Defined;
	List<char> IncludeDirs;

	DefineList()
	{
		Defined = true;
		NestLevel = 0;
	}

	~DefineList()
	{
		for (ImportDefine *i = First(); i; i = Next())
		{
			DeleteObj(i);
		}

		for (char *c = IncludeDirs.First(); c; c = IncludeDirs.Next())
		{
			DeleteArray(c);
		}
	}
	
	void DefineSymbol(char *Name, char *Value = 0)
	{
		ImportDefine *Def = NEW(ImportDefine);
		if (Def)
		{
			Def->Name = NewStr(Name);
			if (Value) Def->Value = NewStr(Value);
			Insert(Def);
		}
	}

	ImportDefine *GetDefine(char *Name)
	{
		if (Name)
		{
			for (ImportDefine *i = First(); i; i = Next())
			{
				if (i->Name AND stricmp(i->Name, Name) == 0)
				{
					return i;
				}
			}
		}
		return NULL;
	}

	void ProcessLine(char *Line)
	{
		if (NestLevel > 16)
		{
			return;
		}

		if (Line AND *Line == '#')
		{
			Line++;
			TokenParser T(Line);

			if (T.Tokens > 0)
			{
				if (stricmp(T.Token[0], "define") == 0) // #define
				{
					ImportDefine *Def = NEW(ImportDefine(Line-1));
					if (Def)
					{
						if (Def->Name)
						{
							Insert(Def);
						}
						else
						{
							DeleteObj(Def);
						}
					}
				}
				else if (stricmp(T.Token[0], "include") == 0) // #include
				{
					NestLevel++;

					GFile F;
					if (T.Tokens > 1)
					{
						for (char *IncPath = IncludeDirs.First(); IncPath; IncPath = IncludeDirs.Next())
						{
							char FullPath[256];

							strcpy(FullPath, IncPath);
							if (FullPath[strlen(FullPath)-1] != DIR_CHAR)
							{
								strcat(FullPath, DIR_STR);
							}
							strcat(FullPath, T.Token[1]);

							if (F.Open(FullPath, O_READ))
							{
								char Line[1024];
								while (NOT F.Eof())
								{
									F.ReadStr(Line, sizeof(Line));
									char *p = Skip(Line);
									if (*p == '#')
									{
										ProcessLine(p);
									}
								}

								break;
							}
						}
					}

					NestLevel--;
				}
				else if (stricmp(T.Token[0], "if") == 0) // #if
				{
				}
				else if (stricmp(T.Token[0], "ifdef") == 0) // #if
				{
					if (T.Tokens > 1)
					{
						Defined = GetDefine(T.Token[1]) != 0;
					}
				}
				else if (stricmp(T.Token[0], "endif") == 0) // #endif
				{
					Defined = true;
				}
				else if (stricmp(T.Token[0], "pragma") == 0)
				{
					ImportDefine *Def = NEW(ImportDefine);
					if (Def)
					{
						char *Str = Line + 7;
						char *First = strchr(Str, '(');
						char *Second = (First) ? strchr(First+1, ')') : 0;
						if (First AND Second)
						{
							Insert(Def);
							Def->Name = NewStr(Str, (int)First-(int)Str);
							First++;
							Def->Value = NewStr(First, (int)Second-(int)First);
						}
						else
						{
							DeleteObj(Def);
						}
					}
				}
				else if (stricmp(T.Token[0], "undef") == 0)
				{
				}
			}
		}
		// else it's not for us anyway
	}
};

bool AppWnd::LoadWin32(char *FileName)
{
	bool Status = false;
	GFileSelect Select;

	Empty();

	if (NOT FileName)
	{
		Select.Parent(this);
		Select.Type("Win32 Resource Script", "*.rc");
		if (Select.Open())
		{
			FileName = Select.Name();
		}
	}

	if (FileName)
	{
		GProgressDlg Progress(this);

		Progress.SetDescription("Initializing...");
		Progress.SetType("K");
		Progress.SetScale(1.0/1024.0);

		GFile F;
		if (F.Open(Select.Name(), O_READ))
		{
			DefineList Defines;
			ResStringGroup *String = NEW(ResStringGroup(this));
			TokenParser T;
			int Mode = IMP_MODE_SEARCH;
			char Line[1024];

			// Language
			char *Language = 0;
			int LanguageId = 0;

			// Dialogs
			List<ResDialog> DlgList;
			CtrlDlg *Dlg = 0;

			// Menus
			ResDialog *Dialog = 0;
			ResMenu *Menu = 0;
			List<ResMenu> Menus;
			ResMenuItem *MenuItem[32];
			int MenuLevel = 0;
			bool MenuNewLang = false;
			int MenuNextItem = 0;

			// Include defines
			char IncPath[256];
			strcpy(IncPath, Select.Name());
			LgiTrimDir(IncPath);
			Defines.IncludeDirs.Insert(NewStr(IncPath));

			Defines.DefineSymbol("_WIN32");
			Defines.DefineSymbol("IDC_STATIC", "-1");

			DoEvery Ticker(200);
			Progress.SetDescription("Reading resources...");
			Progress.SetLimits(0, F.GetSize()-1);

			if (String)
			{
				InsertObject(TYPE_STRING, String, false);
			}

			while (NOT F.Eof())
			{
				F.ReadStr(Line, sizeof(Line));

				if (Ticker.DoNow())
				{
					Progress.Value(F.GetPosition());
					#ifdef WIN32
					UpdateWindow(Progress.Handle());
					#endif
				}
				
				// Exclude comments
				for (int k=0; Line[k]; k++)
				{
					if (Line[k] == '/' AND Line[k+1] == '/')
					{
						Line[k] = 0;
						break;
					}
				}
				
				// Skip white space
				char *p = Skip(Line);
				Defines.ProcessLine(Line);

				// Process line
				if (Defines.Defined)
				{
					switch (Mode)
					{
						case IMP_MODE_SEARCH:
						{
							DeleteObj(Dialog);
							Dlg = 0;
							if (*p != '#')
							{
								TokenParser T(Line);
								if (T.Tokens > 1 AND
									stricmp(T.Token[1], "DIALOG") == 0)
								{
									Mode = IMP_MODE_DIALOG;
									Dialog = NEW(ResDialog(this));
									if (Dialog)
									{
										Dialog->Create();
										Dialog->Name(T.Token[0]);

										Dlg = dynamic_cast<CtrlDlg*>(Dialog->Children.First());
										if (Dlg)
										{
											int Pos[4] = {0, 0, 0, 0};
											int i = 0;
											for (; i<T.Tokens AND NOT IsNumber(T.Token[i]); i++);
											for (int n=0; n<4; n++)
											{
												if (i+n<T.Tokens AND IsNumber(T.Token[i+n]))
												{
													Pos[n] = atoi(T.Token[i+n]);
												}
												else
												{
													break;
												}
											}

											GRegion r(Pos[0], Pos[1], Pos[0]+(Pos[2]*DIALOG_X), Pos[1]+(Pos[3]*DIALOG_Y));
											r.Offset(7, 7);
											Dlg->SetPos(r);
											Dlg->Str->Define = NewStr(T.Token[0]);
											ImportDefine *Def = Defines.GetDefine(T.Token[0]);
											if (Def)
											{
												Dlg->Str->Id = atoi(Def->Value);
											}
										}
									}

									T.Empty();
									break;
								}

								if (T.Tokens > 0 AND
									stricmp(T.Token[1], "MENU") == 0)
								{
									ZeroObj(MenuItem);
									Mode = IMP_MODE_MENU;
									Menu = 0;

									// Check for preexisting menu in another language
									MenuNewLang = false;
									for (ResMenu *m = Menus.First(); m; m = Menus.Next())
									{
										if (stricmp(m->Name(), T.Token[0]) == 0)
										{
											MenuNewLang = true;
											Menu = m;
											break;
										}
									}

									// If it doesn't preexist then create it
									if (NOT Menu)
									{
										Menu = NEW(ResMenu(this));
										if (Menu)
										{
											Menus.Insert(Menu);
											Menu->Create();
											Menu->Name(T.Token[0]);
										}
									}

									T.Empty();
									break;
								}

								if (T.Tokens > 0 AND
									stricmp(T.Token[0], "STRINGTABLE") == 0)
								{
									Mode = IMP_MODE_STRINGS;
									if (String)
									{
										String->Name("_Win32 Imports_");
									}

									T.Empty();
									break;
								}

								if (T.Tokens > 2 AND
									stricmp(T.Token[0], "LANGUAGE") == 0)
								{
									LanguageId = 0;
									DeleteArray(Language);
									Language = NewStr(T.Token[1]);
									if (Language)
									{
										LangInfo *Info = LangFind(Language);
										if (Info)
										{
											LanguageId = Info->Id;
											ResDialog::AddLanguage(Info->Id);
										}
									}

									T.Empty();
									break;
								}
							}
							break;
						}
						case IMP_MODE_DIALOG:
						{
							T.Parse(Line);
							if (T.Tokens > 0 AND Dlg)
							{
								if (stricmp(T.Token[0], "CAPTION") == 0)
								{
									char *Caption = T.Get(1);
									if (Caption)
									{
										Dlg->Str->Set(Caption, LanguageId);
									}
								}
								else if (stricmp(T.Token[0], "BEGIN") == 0)
								{
									Mode = IMP_MODE_DLG_CTRLS;
								}
							}
							T.Empty();
							break;
						}
						case IMP_MODE_DLG_CTRLS:
						{
							TokenParser CtrlTokens(Line);
							T.AppendTokens(&CtrlTokens);

							if (stricmp(T.Token[0], "LTEXT") == 0)
							{
								if (T.Tokens >= 7)
								{
									CtrlText *Ctrl = NEW(CtrlText(Dialog));
									if (Ctrl)
									{
										Ctrl->Str->Set(T.Token[1], LanguageId);
										Ctrl->Str->Define = NewStr(T.Token[2]);
										ImportDefine *Def = Defines.GetDefine(T.Token[2]);
										if (Def)
										{
											Ctrl->Str->Id = atoi(Def->Value);
										}

										GRegion r;
										r.ZOff(atoi(T.Token[5])*CTRL_X, atoi(T.Token[6])*CTRL_Y);
										r.Offset(atoi(T.Token[3])*CTRL_X+ADJUST_CTRLS_X, atoi(T.Token[4])*CTRL_Y+ADJUST_CTRLS_Y);
										Ctrl->SetPos(r);

										Dlg->Children.Insert(Ctrl);
										Ctrl->SetParent(Dlg);
									}
									T.Empty();
								}
							}
							else if (stricmp(T.Token[0], "EDITTEXT") == 0)
							{
								if (T.Tokens >= 7)
								{
									CtrlEditbox *Ctrl = NEW(CtrlEditbox(Dialog));
									if (Ctrl)
									{
										Ctrl->Str->Define = NewStr(T.Token[1]);
										ImportDefine *Def = Defines.GetDefine(T.Token[1]);
										if (Def)
										{
											Ctrl->Str->Id = atoi(Def->Value);
										}

										GRegion r;
										r.ZOff(atoi(T.Token[4])*CTRL_X, atoi(T.Token[5])*CTRL_Y);
										r.Offset(atoi(T.Token[2])*CTRL_X+ADJUST_CTRLS_X, atoi(T.Token[3])*CTRL_Y+ADJUST_CTRLS_Y);
										Ctrl->SetPos(r);

										Dlg->Children.Insert(Ctrl);
										Ctrl->SetParent(Dlg);
									}
									T.Empty();
								}
							}
							else if (stricmp(T.Token[0], "COMBOBOX") == 0)
							{
								if (T.Tokens >= 6)
								{
									CtrlComboBox *Ctrl = NEW(CtrlComboBox(Dialog));
									if (Ctrl)
									{
										Ctrl->Str->Define = NewStr(T.Token[1]);
										ImportDefine *Def = Defines.GetDefine(T.Token[1]);
										if (Def)
										{
											Ctrl->Str->Id = atoi(Def->Value);
										}

										GRegion r;
										r.ZOff(atoi(T.Token[4])*CTRL_X, atoi(T.Token[5])*CTRL_Y);
										r.Offset(atoi(T.Token[2])*CTRL_X+ADJUST_CTRLS_X, atoi(T.Token[3])*CTRL_Y+ADJUST_CTRLS_Y);
										Ctrl->SetPos(r);

										Dlg->Children.Insert(Ctrl);
										Ctrl->SetParent(Dlg);
									}
									T.Empty();
								}
							}
							else if (stricmp(T.Token[0], "SCROLLBAR") == 0)
							{
								if (T.Tokens == 6)
								{
									CtrlScrollBar *Ctrl = NEW(CtrlScrollBar(Dialog));
									if (Ctrl)
									{
										Ctrl->Str->Define = NewStr(T.Token[1]);
										ImportDefine *Def = Defines.GetDefine(T.Token[1]);
										if (Def)
										{
											Ctrl->Str->Id = atoi(Def->Value);
										}

										GRegion r;
										r.ZOff(atoi(T.Token[4])*CTRL_X, atoi(T.Token[5])*CTRL_Y);
										r.Offset(atoi(T.Token[2])*CTRL_X+ADJUST_CTRLS_X, atoi(T.Token[3])*CTRL_Y+ADJUST_CTRLS_Y);
										Ctrl->SetPos(r);

										Dlg->Children.Insert(Ctrl);
										Ctrl->SetParent(Dlg);
									}
									T.Empty();
								}
							}
							else if (stricmp(T.Token[0], "GROUPBOX") == 0)
							{
								if (T.Tokens >= 7)
								{
									CtrlGroup *Ctrl = NEW(CtrlGroup(Dialog));
									if (Ctrl)
									{
										Ctrl->Str->Set(T.Token[1], LanguageId);
										Ctrl->Str->Define = NewStr(T.Token[2]);
										ImportDefine *Def = Defines.GetDefine(T.Token[2]);
										if (Def)
										{
											Ctrl->Str->Id = atoi(Def->Value);
										}

										GRegion r;
										r.ZOff(atoi(T.Token[5])*CTRL_X, atoi(T.Token[6])*CTRL_Y);
										r.Offset(atoi(T.Token[3])*CTRL_X+ADJUST_CTRLS_X, atoi(T.Token[4])*CTRL_Y+ADJUST_CTRLS_Y);
										Ctrl->SetPos(r);

										Dlg->Children.Insert(Ctrl);
										Ctrl->SetParent(Dlg);
									}
									T.Empty();
								}
							}
							else if (stricmp(T.Token[0], "PUSHBUTTON") == 0 OR
									 stricmp(T.Token[0], "DEFPUSHBUTTON") == 0)
							{
								if (T.Tokens >= 7)
								{
									CtrlButton *Ctrl = NEW(CtrlButton(Dialog));
									if (Ctrl)
									{
										Ctrl->Str->Set(T.Token[1], LanguageId);
										Ctrl->Str->Define = NewStr(T.Token[2]);
										ImportDefine *Def = Defines.GetDefine(T.Token[2]);
										if (Def)
										{
											Ctrl->Str->Id = atoi(Def->Value);
										}

										GRegion r;
										r.ZOff(atoi(T.Token[5])*CTRL_X, atoi(T.Token[6])*CTRL_Y);
										r.Offset(atoi(T.Token[3])*CTRL_X+ADJUST_CTRLS_X, atoi(T.Token[4])*CTRL_Y+ADJUST_CTRLS_Y);
										Ctrl->SetPos(r);

										Dlg->Children.Insert(Ctrl);
										Ctrl->SetParent(Dlg);
									}
									T.Empty();
								}
							}
							else if (stricmp(T.Token[0], "CONTROL") == 0)
							{
								if (T.Tokens >= 7)
								{
									char *Caption = T.Token[1];
									char *Id = T.Token[2];
									char *Type = T.Token[3];
									bool Checkbox = false;
									bool Radio = false;
									bool Done = false;

									// loop through styles
									int i;
									for (i=4; NOT Done AND i<T.Tokens; i++)
									{
										Done = true;
										if (stricmp(T.Token[i], "BS_AUTOCHECKBOX") == 0)
										{
											Checkbox = true;
										}
										else if (stricmp(T.Token[i], "BS_AUTORADIOBUTTON") == 0)
										{
											Radio = true;
										}

										if (i<T.Tokens-1 AND stricmp(T.Token[i+1], "|") == 0)
										{
											// next token is an OR operator
											i++;
											Done = false; // read through more styles
										}
									}

									GRegion r(0, 0, 0, 0);
									if (i + 3 < T.Tokens AND Type)
									{
										ResDialogCtrl *Ctrl = 0;
										if (stricmp(Type, "Button") == 0)
										{
											if (Radio)
											{
												Ctrl = NEW(CtrlRadio(Dialog));
											}
											else if (Checkbox)
											{
												Ctrl = NEW(CtrlCheckbox(Dialog));
											}
										}
										else if (stricmp(Type, "SysListView32") == 0)
										{
											Ctrl = NEW(CtrlList(Dialog));
										}

										if (Ctrl)
										{
											r.ZOff(atoi(T.Token[i+2])*CTRL_X, atoi(T.Token[i+3])*CTRL_Y);
											r.Offset(atoi(T.Token[i])*CTRL_X+ADJUST_CTRLS_X, atoi(T.Token[i+1])*CTRL_Y+ADJUST_CTRLS_Y);
											Ctrl->SetPos(r);

											if (Caption) Ctrl->Str->Set(Caption, LanguageId);
											if (Id) Ctrl->Str->Define = NewStr(Id);
											ImportDefine *Def = Defines.GetDefine(Id);
											if (Def)
											{
												Ctrl->Str->Id = atoi(Def->Value);
											}

											Dlg->Children.Insert(Ctrl);
											Ctrl->SetParent(Dlg);
										}

										T.Empty();
									}
								}
							}
							else if (stricmp(T.Token[0], "END") == 0)
							{
								// search for an existing dialog resource in
								// another language
								ResDialog *Match = 0;
								CtrlDlg *MatchObj = 0;
								for (ResDialog *d = DlgList.First(); d; d = DlgList.Next())
								{
									GView *Wnd = d->Children.First();
									if (Wnd)
									{
										CtrlDlg *Obj = dynamic_cast<CtrlDlg*>(Wnd);
										if (Obj)
										{
											if (Obj->Str->Id == Dlg->Str->Id)
											{
												MatchObj = Obj;
												Match = d;
												break;
											}
										}
									}
								}

								if (Match)
								{
									// Merge the controls from "Dlg" to "MatchObj"
									List<ResDialogCtrl> Old;
									List<ResDialogCtrl> New;
									
									Dlg->ListChildren(New);
									MatchObj->ListChildren(Old);

									// add the language strings for the caption
									// without clobbering the languages already
									// present
									for (StrLang *s = Dlg->Str->Items.First();
										s; s = Dlg->Str->Items.Next())
									{
										if (NOT MatchObj->Str->Get(s->Lang))
										{
											MatchObj->Str->Set(s->Str, s->Lang);
										}
									}

									for (ResDialogCtrl *c = New.First(); c; c = New.Next())
									{
										ResDialogCtrl *MatchCtrl = 0;

										// try matching by Id
										{
											for (ResDialogCtrl *Mc = Old.First(); Mc; Mc = Old.Next())
											{
												if (Mc->Str->Id == c->Str->Id AND
													Mc->Str->Id > 0)
												{
													MatchCtrl = Mc;
													break;
												}
											}
										}

										// ok no Id match, match by location and type
										if (NOT MatchCtrl)
										{
											List<ResDialogCtrl> Overlapping;
											for (ResDialogCtrl *Mc = Old.First(); Mc; Mc = Old.Next())
											{
												GRegion a = Mc->GetPos();
												GRegion b = c->GetPos();
												GRegion c = a;
												c.Bound(&b);
												if (c.Valid())
												{
													int Sa = a.X() * a.Y();
													int Sb = b.X() * b.Y();
													int Sc = c.X() * c.Y();
													int Total = Sa + Sb - Sc;
													double Amount = (double) Sc / (double) Total;
													if (Amount > 0.5)
													{
														// mostly similar in size
														Overlapping.Insert(Mc);
													}
												}
											}

											if (Overlapping.GetItems() == 1)
											{
												MatchCtrl = Overlapping.First();
											}
										}

										if (MatchCtrl)
										{
											// woohoo we are cool
											for (StrLang *s = c->Str->Items.First();
												s; s = c->Str->Items.Next())
											{
												MatchCtrl->Str->Set(s->Str, s->Lang);
											}
										}
									}

									// Delete the duplicate
									OnObjSelect(0);
									DeleteObj(Dialog);
								}
								else
								{
									// Insert the dialog
									InsertObject(TYPE_DIALOG, Dialog, false);
									DlgList.Insert(Dialog);
								}

								Dialog = 0;
								Dlg = 0;
								Status = true;
								T.Empty();
								Mode = IMP_MODE_SEARCH;
							}
							else
							{
								T.Empty();
							}

							break;
						}
						case IMP_MODE_STRINGS:
						{
							TokenParser CtrlTokens(Line);
							T.AppendTokens(&CtrlTokens);

							if (stricmp(T.Token[0], "BEGIN") == 0)
							{
								T.Empty();
							}
							else if (stricmp(T.Token[0], "END") == 0)
							{
								Status = true;
								T.Empty();
								Mode = IMP_MODE_SEARCH;
							}
							else
							{
								if (T.Tokens > 1)
								{
									ResString *Str = String->FindName(T.Token[0]);
									if (NOT Str)
									{
										Str = String->CreateStr();
										if (Str)
										{
											ImportDefine *Def = Defines.GetDefine(T.Token[0]);
											if (Def)
											{
												Str->Id = atoi(Def->Value);
											}

											DeleteArray(Str->Define);
											Str->Define = NewStr(T.Token[0]);
											Str->UnDupelicate();
										}
									}

									if (Str)
									{
										// get the language
										LangInfo *Lang = LangFind(Language);
										int SLang = (Lang) ? Lang->Id : LangIdEnglish;
										StrLang *s = 0;

										// look for language present in string object
										for (s = Str->Items.First(); s AND s->Lang != SLang; s = Str->Items.Next());
										
										// if not present then add it
										if (NOT s)
										{
											s = NEW(StrLang);
											if (s)
											{
												Str->Items.Insert(s);
												s->Lang = SLang;
											}
										}

										// set the value
										if (s)
										{
											DeleteArray(s->Str);
											s->Str = NewStr(T.Token[1]);
										}
									}

									T.Empty();
								}
							}

							break;
						}
						case IMP_MODE_MENU:
						{
							TokenParser CtrlTokens(Line);
							T.AppendTokens(&CtrlTokens);

							if (T.Tokens >= 1)
							{
								if (stricmp(T.Token[0], "BEGIN") == 0)
								{
									MenuLevel++;
									T.Empty();
								}
								else if (stricmp(T.Token[0], "END") == 0)
								{
									MenuLevel--;
									T.Empty();
									if (MenuLevel == 0)
									{
										Status = true;
										Mode = IMP_MODE_SEARCH;
										Menu->SetLanguages();

										if (NOT MenuNewLang)
										{
											InsertObject(TYPE_MENU, Menu, false);
										}

										Menu = 0;
									}
								}
								else
								{
									ResMenuItem *i = 0;
									char *Text = T.Token[1];
									if (Text)
									{
										if (stricmp(T.Token[0], "POPUP") == 0)
										{
											if (MenuNewLang)
											{
												GTreeItem *Ri = 0;
												if (MenuItem[MenuLevel])
												{
													Ri = MenuItem[MenuLevel]->GetNext();
												}
												else
												{
													if (MenuLevel == 1)
													{
														Ri = Menu->ItemAt(0);
													}
													else if (MenuItem[MenuLevel-1])
													{
														Ri = MenuItem[MenuLevel-1]->GetChild();
													}
												}

												if (Ri)
												{
													// Seek up to the next submenu
													while (NOT Ri->GetChild())
													{
														Ri = Ri->GetNext();
													}

													i = dynamic_cast<ResMenuItem*>(Ri);
													char *si = i->Str.Get(LangIdEnglish);
													if (i)
													{
														MenuItem[MenuLevel] = i;
													}
												}
											}
											else
											{
												MenuItem[MenuLevel] = i = NEW(ResMenuItem(Menu));
											}

											if (i)
											{
												LangInfo *Lang = LangFind(Language);
												i->Str.Set(Text, (Lang) ? Lang->Id : LangIdEnglish);
											}

											MenuNextItem = 0;
										}
										else if (stricmp(T.Token[0], "MENUITEM") == 0)
										{
											if (MenuNewLang)
											{
												if (MenuItem[MenuLevel-1] AND T.Tokens > 2)
												{
													ImportDefine *id = Defines.GetDefine(T.Token[2]);
													if (id)
													{
														int Id = atoi(id->Value);

														int n = 0;
														for (GTreeItem *o = MenuItem[MenuLevel-1]->GetChild(); o; o = o->GetNext(), n++)
														{
															ResMenuItem *Res = dynamic_cast<ResMenuItem*>(o);
															if (Res AND Res->Str.Id == Id)
															{
																i = Res;
																break;
															}
														}
													}
												}

												MenuNextItem++;
											}
											else
											{
												i = NEW(ResMenuItem(Menu));
											}

											if (i)
											{
												if (stricmp(Text, "SEPARATOR") == 0)
												{
													// Set separator
													i->Separator(true);
												}
												else
												{
													if (NOT MenuNewLang)
													{
														// Set Id
														i->Str.Define = NewStr(T.Token[2]);
														if (i->Str.Define)
														{
															ImportDefine *id = Defines.GetDefine(i->Str.Define);
															if (id)
															{
																i->Str.Id = atoi(id->Value);
																i->Str.UnDupelicate();
															}
														}
													}

													// Set Text
													LangInfo *Lang = LangFind(Language);
													i->Str.Set(Text, (Lang) ? Lang->Id : LangIdEnglish);
												}
											}
										}
									}

									if (i AND NOT MenuNewLang)
									{
										if (MenuLevel == 1)
										{
											Menu->Insert(i);
										}
										else if (MenuItem[MenuLevel-1])
										{
											MenuItem[MenuLevel-1]->Insert(i);
										}
									}

									T.Empty();
								}
							}

							break;
						}
					}
				}
			}

			DeleteObj(Dialog);

			if (String->GetItems() > 0)
			{
				String->SetLanguages();
			}
		}

		Invalidate();
	}

	return Status;
}

bool AppWnd::SaveWin32()
{
	return false;
}

/////////////////////////////////////////////////////////////////////////
ResFrame::ResFrame(Resource *child)
{
	Child = child;
}

ResFrame::~ResFrame()
{
	if (Child)
	{
		Child->App()->OnObjSelect(NULL);

		Children.Delete(Child->Wnd());
		Child->Wnd()->SetParent(NULL);
	}
}

bool ResFrame::Attach(GView *p)
{
	bool Status = GLayout::Attach(p);
	if (Status AND Child)
	{
		Child->Attach(this);
	}
	return Status;
}

bool ResFrame::Pour(GComplexRegion &r)
{
	GRegion *Best = FindLargest(r);
	if (Best)
	{
		SetPos(*Best);
		return TRUE;
	}
	return FALSE;
}

void ResFrame::OnKey(GKey &k)
{
	if (k.Down AND Child)
	{
		switch (k.c)
		{
			case VK_DELETE:
			{
				if (k.Shift())
				{
					Child->Copy(true);
				}
				else
				{
					Child->Delete();
				}
				break;
			}
			case 'x':
			case 'X':
			{
				if (k.Ctrl())
				{
					Child->Copy(true);
				}
				break;
			}
			case 'c':
			case 'C':
			{
				if (k.Ctrl())
				{
					Child->Copy();
				}
				break;
			}
			case VK_INSERT:
			{
				if (k.Ctrl())
				{
					Child->Copy();
				}
				else if (k.Shift())
				{
					Child->Paste();
				}
				break;
			}
			case 'v':
			case 'V':
			{
				if (k.Ctrl())
				{
					Child->Paste();
				}
				break;
			}
		}
	}

	Child->Wnd()->OnKey(k);
}

void ResFrame::OnPaint(GSurface *pDC)
{
	// Draw nice frame
	GRegion r(0, 0, X()-1, Y()-1);
	LgiThinBorder(pDC, r, RAISED);
	pDC->Colour(LC_MED, 24);
	LgiFlatBorder(pDC, r, 4);
	LgiWideBorder(pDC, r, SUNKEN);

	// Set the child to the client area
	Child->Wnd()->SetPos(r);

	// Draw the dialog & controls
	GView::OnPaint(pDC);
}

