/*
**	FILE:			LgiRes_Dialog.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			5/8/1999
**	DESCRIPTION:	Dialog Resource Editor
**
**
**	Copyright (C) 1999, Matthew Allen
**		fret@ozemail.com.au
*/

////////////////////////////////////////////////////////////////////
#include "LgiRes.h"
#include "LgiRes_Dialog.h"

////////////////////////////////////////////////////////////////////
#define IDM_MOVE_LEFT				702
#define IDM_MOVE_RIGHT				703
#define IDM_SET_LANG				704
#define IDM_TAB_ORDER				705
#define	IDM_DUMP					706

////////////////////////////////////////////////////////////////////
#define IDC_LIST					100
#define IDC_UP						101
#define IDC_DOWN					102

// Name mapping table
class LgiObjectName
{
public:
	int Type;
	char *ObjectName;
	char *ResourceName;
}
NameMap[] =
{
	// ID			Lgi's name		Resource editor name
	{UI_DIALOG,		"GDialog",		Res_Dialog},
	{UI_TEXT,		"GText",		Res_StaticText},
	{UI_EDITBOX,	"GEdit",		Res_EditBox},
	{UI_CHECKBOX,	"GCheckBox",	Res_CheckBox},
	{UI_BUTTON,		"GButton",		Res_Button},
	{UI_GROUP,		"GRadioGroup",	Res_Group},
	{UI_RADIO,		"GRadioButton",	Res_RadioBox},
	{UI_TABS,		"GTabView",	Res_TabView},
	{UI_TAB,		"GTabPage",		Res_Tab},
	{UI_LIST,		"GList",		Res_ListView},
	{UI_COLUMN,		"GListColumn",	Res_Column},
	{UI_COMBO,		"GCombo",		Res_ComboBox},
	{UI_TREE,		"GTree",		Res_TreeView},
	{UI_BITMAP,		"GBitmap",		Res_Bitmap},
	{UI_PROGRESS,	"GProgress",	Res_Progress},
	{UI_SCROLL_BAR,	"GScrollBar",	Res_ScrollBar},
	{0, 0}
};

class CtrlItem : public GListItem
{
	friend class TabOrder;

	ResDialogCtrl *Ctrl;

public:
	CtrlItem(ResDialogCtrl *ctrl)
	{
		Ctrl = ctrl;
	}

	char *GetText(int Col)
	{
		switch (Col)
		{
			case 0:
			{
				if (Ctrl AND
					Ctrl->Str AND
					Ctrl->Str->Define)
				{
					return Ctrl->Str->Define;
				}
				break;
			}
			case 1:
			{
				if (Ctrl AND
					Ctrl->Str AND
					Ctrl->Str->Get())
				{
					return Ctrl->Str->Get();
				}
				break;
			}
		}
		return "";
	}
};

class TabOrder : public GDialog
{
	ResDialog *Dlg;
	GList *Lst;
	GButton *Ok;
	GButton *Cancel;
	GButton *Up;
	GButton *Down;
	ResDialogCtrl *Parent;

public:
	TabOrder(ResDialog *dlg)
	{
		Dlg = dlg;
		SetParent(Dlg);

		Children.Insert(Lst = NEW(GList(IDC_LIST, 10, 10, 350, 300)));
		Children.Insert(Ok = NEW(GButton(IDOK, Lst->GetPos().x2 + 10, 10, 60, 20, "Ok")));
		Children.Insert(Cancel = NEW(GButton(IDCANCEL, Lst->GetPos().x2 + 10, Ok->GetPos().y2 + 5, 60, 20, "Cancel")));
		Children.Insert(Up = NEW(GButton(IDC_UP, Lst->GetPos().x2 + 10, Cancel->GetPos().y2 + 15, 60, 20, "Up")));
		Children.Insert(Down = NEW(GButton(IDC_DOWN, Lst->GetPos().x2 + 10, Up->GetPos().y2 + 5, 60, 20, "Down")));

		GRegion r(0, 0, Ok->GetPos().x2 + 17, Lst->GetPos().y2 + 40);
		SetPos(r);
		MoveToCenter();

		Lst->AddColumn("#define", 150);
		Lst->AddColumn("Text", 150);
		Lst->MultiSelect(false);

		Parent = dynamic_cast<ResDialogCtrl*>(Dlg->Children.First());
		if (Parent)
		{
			for (ResDialogCtrl *Ctrl=dynamic_cast<ResDialogCtrl*>(Parent->Children.First());
				Ctrl; Ctrl=dynamic_cast<ResDialogCtrl*>(Parent->Children.Next()))
			{
				Lst->Insert(NEW(CtrlItem(Ctrl)));
			}
		}

		DoModal();
	}

	int OnNotify(GView *Ctrl, int Flags)
	{
		int MoveDir = 1;
		switch (Ctrl->GetId())
		{
			case IDC_UP:
			{
				MoveDir = -1;
				// fall through
			}
			case IDC_DOWN:
			{
				if (Lst AND Parent)
				{
					CtrlItem *Sel = dynamic_cast<CtrlItem*>(Lst->GetSelection());
					if (Sel)
					{
						int Index = Lst->IndexOf(Sel);
						CtrlItem *Up = dynamic_cast<CtrlItem*>(Lst->ItemAt(Index+MoveDir));
						if (Up)
						{
							Lst->Remove(Sel);
							Lst->Insert(Sel, Index+MoveDir);
							Sel->Select(true);
							Sel->ScrollTo();
						}
					}
				}
				break;
			}
			case IDOK:
			{
				if (Lst AND Parent)
				{
					int i=0;
					Parent->Children.Empty();
					for (CtrlItem *n=dynamic_cast<CtrlItem*>(Lst->First());
						n; n=dynamic_cast<CtrlItem*>(Lst->Next()))
					{
						Parent->Children.Insert(n->Ctrl, i++);
					}
				}
				// fall through
			}
			case IDCANCEL:
			{
				EndModal(0);
				break;
			}
		}

		return 0;
	}
};

////////////////////////////////////////////////////////////////////
void DrawGoobers(GSurface *pDC, GRegion &r, GRegion *Goobers)
{
	int Mx = (r.x2 + r.x1) / 2 - (GOOBER_SIZE / 2);
	int My = (r.y2 + r.y1) / 2 - (GOOBER_SIZE / 2);

	pDC->Colour(Rgb24(255, 0, 0), 24);

	Goobers[0].x1 = r.x1 - GOOBER_BORDER;
	Goobers[0].y1 = r.y1 - GOOBER_BORDER;
	Goobers[0].x2 = r.x1 - (GOOBER_BORDER - GOOBER_SIZE);
	Goobers[0].y2 = r.y1 - (GOOBER_BORDER - GOOBER_SIZE);

	Goobers[1].x1 = Mx;
	Goobers[1].y1 = r.y1 - GOOBER_BORDER;
	Goobers[1].x2 = Mx + GOOBER_SIZE;
	Goobers[1].y2 = r.y1 - (GOOBER_BORDER - GOOBER_SIZE);

	Goobers[2].x1 = r.x2 + (GOOBER_BORDER - GOOBER_SIZE);
	Goobers[2].y1 = r.y1 - GOOBER_BORDER;
	Goobers[2].x2 = r.x2 + GOOBER_BORDER;
	Goobers[2].y2 = r.y1 - (GOOBER_BORDER - GOOBER_SIZE);

	Goobers[3].x1 = r.x2 + (GOOBER_BORDER - GOOBER_SIZE);
	Goobers[3].y1 = My;
	Goobers[3].x2 = r.x2 + GOOBER_BORDER;
	Goobers[3].y2 = My + GOOBER_SIZE;

	Goobers[4].x1 = r.x2 + (GOOBER_BORDER - GOOBER_SIZE);
	Goobers[4].y1 = r.y2 + (GOOBER_BORDER - GOOBER_SIZE);
	Goobers[4].x2 = r.x2 + GOOBER_BORDER;
	Goobers[4].y2 = r.y2 + GOOBER_BORDER;

	Goobers[5].x1 = Mx;
	Goobers[5].y1 = r.y2 + (GOOBER_BORDER - GOOBER_SIZE);
	Goobers[5].x2 = Mx + GOOBER_SIZE;
	Goobers[5].y2 = r.y2 + GOOBER_BORDER;

	Goobers[6].x1 = r.x1 - GOOBER_BORDER;
	Goobers[6].y1 = r.y2 + (GOOBER_BORDER - GOOBER_SIZE);
	Goobers[6].x2 = r.x1 - (GOOBER_BORDER - GOOBER_SIZE);
	Goobers[6].y2 = r.y2 + GOOBER_BORDER;

	Goobers[7].x1 = r.x1 - GOOBER_BORDER;
	Goobers[7].y1 = My;
	Goobers[7].x2 = r.x1 - (GOOBER_BORDER - GOOBER_SIZE);
	Goobers[7].y2 = My + GOOBER_SIZE;

	for (int i=0; i<8; i++)
	{
		pDC->Box(Goobers+i);
	}
}

////////////////////////////////////////////////////////////////////
int ResDialogCtrl::TabDepth = 0;

ResDialogCtrl::ResDialogCtrl(ResDialog *dlg, char *CtrlTypeName) :
	ResObject(CtrlTypeName)
{
	Dlg = dlg;
	DragCtrl = -1;
	AcceptChildren = false;
	Movable = true;
	MoveCtrl = false;
	Client.ZOff(-1, -1);
	SelectMode = 0;
	SelectStart.ZOff(-1, -1);

	if (Dlg->CreateSymbols)
	{
		// We create a symbol for this resource
		Str = (Dlg AND Dlg->Symbols) ? Dlg->Symbols->CreateStr() : 0;
		if (Str)
		{
			Str->UpdateWnd = this;
		}
	}
	else
	{
		// Someone else is going to create the symbol
		Str = 0;
	}
}

ResDialogCtrl::~ResDialogCtrl()
{
	if (ResDialog::Symbols)
	{
		ResDialog::Symbols->DeleteStr(Str);
	}

	if (Dlg)
	{
		Dlg->OnDeselect(this);
	}
}

char *ResDialogCtrl::GetRefText()
{
	static char Buf[64];
	if (Str)
	{
		sprintf(Buf, "Ref=%i", Str->Ref);
	}
	else
	{
		Buf[0] = 0;
	}
	return Buf;
}

void ResDialogCtrl::ListChildren(List<ResDialogCtrl> &l, bool Deep)
{
	for (GView *w = Children.First(); w; w = Children.Next())
	{
		ResDialogCtrl *c = dynamic_cast<ResDialogCtrl*>(w);
		if (c)
		{
			l.Insert(c);
			if (Deep)
			{
				c->ListChildren(l);
			}
		}
	}
}

GRegion ResDialogCtrl::GetMinSize()
{
	GRegion m(0, 0, GRID_X-1, GRID_Y-1);

	if (IsContainer())
	{
		for (GView *c=Children.First(); c; c=Children.Next())
		{
			m.Union(&c->GetPos());
		}
	}

	return m;
}

bool ResDialogCtrl::SetPos(GRegion &p, bool Repaint)
{
	ResDialogCtrl *ParentCtrl = dynamic_cast<ResDialogCtrl*>(GetParent());
	GRegion m = GetMinSize();
	if (m.X() > p.X()) p.x2 = p.x1 + m.X() - 1;
	if (m.Y() > p.Y()) p.y2 = p.y1 + m.Y() - 1;

	if (p != Pos)
	{
		/*
		if (ParentCtrl)
		{
			if (p.x1 < ParentCtrl->Client.x1)
			{
				p.Offset(ParentCtrl->Client.x1 - p.x1, 0);
			}

			if (p.y1 < ParentCtrl->Client.y1)
			{
				p.Offset(0, ParentCtrl->Client.y1 - p.y1);
			}
		}
		*/

		// set our size
		bool Status = GView::SetPos(p, Repaint);
		
		// tell everyone else about the change
		Serialize(true);
		if (Dlg->Resource::IsSelected())
		{
			Dlg->App()->OnObjChange(this);
		}

		// check our parents are big enough to show us...
		if (ParentCtrl)
		{
			GRegion t = ParentCtrl->GetPos();
			ParentCtrl->SetPos(t, true);
		}
		
		return Status;
	}

	return true;
}

void ResDialogCtrl::TabString(char *Str)
{
	if (Str)
	{
		for (int i=0; i<TabDepth; i++)
		{
			*Str++ = 9;
		}
		*Str = 0;
	}
}

void ResDialogCtrl::StrFromRef(int Ref)
{
	// get the string object
	Str = Dlg->App()->GetStrFromRef(Ref);
	
	// if this assert fails then the Ref doesn't exist
	// and the string can't be found
	ASSERT(Str);

	// if this assert fails then the Str is already
	// associated with a control, and thus we would
	// duplicate the pointer to the string if we let
	// it go by
	ASSERT(Str->UpdateWnd == 0);

	// set the string's control to us
	Str->UpdateWnd = this;

	// make the strings refid unique
	Str->UnDupelicate();
}

#define VAL_Text					"Text"
#define VAL_x1						"x1"
#define VAL_y1						"y1"
#define VAL_x2						"x2"
#define VAL_y2						"y2"
#define VAL_Visible					"Visible"
#define VAL_Enabled					"Enabled"

bool ResDialogCtrl::GetFields(DataDlgFieldList &Fields, ObjProperties *Props)
{
	// Set Fields
	// Fields.Insert(NEW(DataDlgField(DATA_STR, 100, VAL_Text, "Text")));

	if (Str)
	{
		Str->GetFields(Fields, &Values);
	}

	Fields.Insert(NEW(DataDlgField(DATA_INT, 101, VAL_x1, "x1")));
	Fields.Insert(NEW(DataDlgField(DATA_INT, 102, VAL_y1, "y1")));
	Fields.Insert(NEW(DataDlgField(DATA_INT, 103, VAL_x2, "x2")));
	Fields.Insert(NEW(DataDlgField(DATA_INT, 104, VAL_y2, "y2")));
	Fields.Insert(NEW(DataDlgField(DATA_BOOL, 105, VAL_Visible, "Visible")));
	Fields.Insert(NEW(DataDlgField(DATA_BOOL, 106, VAL_Enabled, "Enabled")));

	// Copy data from object in Values
	Serialize(true);

	return true;
}

void ResDialogCtrl::OnFieldChange()
{
	Serialize(false);
}

bool ResDialogCtrl::Serialize(bool Write)
{
	if (Str)
	{
		Str->Serialize(Write, &Values);
	}

	if (Write)
	{
		GRegion r = GetPos();

		Values.Set(VAL_x1, r.x1);
		Values.Set(VAL_y1, r.y1);
		Values.Set(VAL_x2, r.x2);
		Values.Set(VAL_y2, r.y2);
		Values.Set(VAL_Visible, 1);
		Values.Set(VAL_Enabled, Enabled());
	}
	else
	{
		GRegion r = GetPos();
		GRegion Old = r;
		char *n;

		Values.Get(VAL_x1, r.x1);
		Values.Get(VAL_y1, r.y1);
		Values.Get(VAL_x2, r.x2);
		Values.Get(VAL_y2, r.y2);
		SetPos(r);
		r.Union(&Old);
		r.Size(-GOOBER_BORDER, -GOOBER_BORDER);
		if (Parent)
		{
			Parent->Invalidate(&r);
		}

		int Enable = true;
		if (Values.Get(VAL_Enabled, Enable))
		{
			Enabled(Enable);
		}
	}

	return true;
}

bool ResDialogCtrl::AttachCtrl(ResDialogCtrl *Ctrl, GRegion *r)
{
	bool Status = false;
	if (Ctrl)
	{
		Children.Insert(Ctrl);
		Ctrl->SetParent(this);
		if (r)
		{
			Dlg->SnapRect(r, this);
			Ctrl->SetPos(*r);
		}

		if (Dlg->Resource::IsSelected())
		{
			Dlg->OnSelect(Ctrl);
		}

		Status = true;
	}
	return Status;
}

void ResDialogCtrl::OnPaint(GSurface *pDC)
{
	if (DragCtrl >= 0)
	{
		int Op = pDC->Op();
		pDC->Op(GDC_XOR);

		GRegion r = DragRgn;
		r.Normal();
		pDC->Colour(0xFFFFFF, 24);
		pDC->Box(&r);

		pDC->Op(Op);
	}
}

void ResDialogCtrl::OnMouseClick(GMouse &m)
{
	if (m.Down())
	{
		if (m.Left())
		{
			if (Dlg)
			{
				bool Processed = false;
				bool ClickedThis = Client.Overlap(m.x, m.y);

				GMouse Ms = m;
				for (GView *p = this; p AND p != Dlg; p = p->GetParent())
				{
					GRegion r = p->GetPos();
					Ms.x += r.x1;
					Ms.y += r.y1;
				}

				Dlg->OnMouseClick(Ms);

				if (ClickedThis AND
					NOT Dlg->IsDraging())
				{
					DragCtrl = Dlg->CurrentTool();
					if ((DragCtrl > 0 AND AcceptChildren) OR
						((DragCtrl == 0) AND NOT Movable))
					{
						GdcPt2 p(m.x, m.y);
						Dlg->SnapPoint(&p, ParentCtrl());

						DragStart.x = DragRgn.x1 = DragRgn.x2 = p.x;
						DragStart.y = DragRgn.y1 = DragRgn.y2 = p.y;
						Capture(true);
						Processed = true;
					}
					else
					{
						DragCtrl = -1;
						if (Movable)
						{
							DragRgn.x1 = m.x;
							DragRgn.y1 = m.y;
							MoveCtrl = true;
							Processed = true;
							Capture(true);
						}
					}

					SelectMode = (m.Shift()) ? 2 : 1;
					SelectStart = GetPos();
				}
			}
		}
		else
		{
			GSubMenu *RClick = NEW(GSubMenu);
			if (RClick)
			{
				bool PasteData = false;
				{
					GClipBoard c(Dlg);
					char *Clip = c.Text();
					if (Clip)
					{
						char *p = Clip;
						while (*p AND strchr(" \r\n\t", *p)) p++;
						PasteData = *p == '<';
						DeleteArray(Clip);
					}
				}

				RClick->AppendItem("Cut", IDM_CUT, Dlg->Selection.GetItems()>0);
				RClick->AppendItem("Copy", IDM_COPY, Dlg->Selection.GetItems()>0);
				RClick->AppendItem("Paste", IDM_PASTE, PasteData);
				RClick->AppendSeparator();
				RClick->AppendItem("Delete", IDM_DELETE, Dlg->Selection.GetItems()>0);

				if (Dlg->GetMouse(m, true))
				{
					int Cmd = 0;
					switch (Cmd = RClick->Float(Dlg, m.x, m.y))
					{
						case IDM_DELETE:
						{
							Dlg->Delete();
							break;
						}
						case IDM_CUT:
						{
							Dlg->Copy(true);
							break;
						}
						case IDM_COPY:
						{
							Dlg->Copy();
							break;
						}
						case IDM_PASTE:
						{
							Dlg->Paste();
							break;
						}
					}
				}

				DeleteObject(RClick);
			}
		}
	}
	else
	{
		if (DragCtrl >= 0)
		{
			bool Exit = false;

			if (Dlg AND
				(DragRgn.X() > 1 OR
				DragRgn.Y() > 1))
			{
				if (DragCtrl == 0)
				{
					Dlg->SelectRect(this, &DragRgn, SelectMode != 2);
				}
				else
				{
					ResDialogCtrl *Ctrl = Dlg->CreateCtrl(DragCtrl);
					if (Ctrl)
					{
						AttachCtrl(Ctrl, &DragRgn);
					}
				}

				Exit = true;
			}

			DragCtrl = -1;
			Invalidate();
			Capture(false);

			if (Exit)
			{
				return;
			}
		}
		
		if (MoveCtrl)
		{
			Capture(false);
			MoveCtrl = false;
		}
		
		if (SelectMode > 0)
		{
			GRegion r = GetPos();
			if (SelectStart == r)
			{
				Dlg->OnSelect(this, SelectMode != 2);
			}
			SelectMode = 0;
		}
	}
}

void ResDialogCtrl::OnMouseMove(GMouse &m)
{
	// Drag a rubber band...
	if (DragCtrl >= 0)
	{
		GRegion Old = DragRgn;

		DragRgn.x1 = DragStart.x;
		DragRgn.y1 = DragStart.y;
		DragRgn.x2 = m.x;
		DragRgn.y2 = m.y;
		DragRgn.Normal();

		Dlg->SnapRect(&DragRgn, this);

		Old.Union(&DragRgn);
		Old.Size(-1, -1);

		Invalidate(&Old);
	}
	
	// Move some ctrl(s)
	if (MoveCtrl)
	{
		if (NOT Dlg->IsSelected(this))
		{
			Dlg->OnSelect(this);
		}

		GRegion Old = GetPos();
		GRegion New = Old;
		New.Offset(	m.x-DragRgn.x1,
					m.y-DragRgn.y1);

		GdcPt2 p(New.x1, New.y1);
		Dlg->SnapPoint(&p, ParentCtrl());
		New.Set(p.x, p.y, p.x+New.X()-1, p.y+New.Y()-1);
		Dlg->MoveSelection(New.x1 - Old.x1, New.y1 - Old.y1);
	}
}

char *ReadInt(char *s, int &Value)
{
	char *c = strchr(s, ',');
	if (c)
	{
		*c = 0;
		Value = atoi(s);
		return c+1;
	}
	Value = atoi(s);
	return 0;
}

void ResDialogCtrl::ReadPos(char *Str)
{
	if (Str)
	{
		char *s = NewStr(Str);
		if (s)
		{
			char *p = ReadInt(s, Pos.x1);
			if (p) p = ReadInt(p, Pos.y1);
			if (p) p = ReadInt(p, Pos.x2);
			if (p) p = ReadInt(p, Pos.y2);

			DeleteArray(s);
		}
	}
}

////////////////////////////////////////////////////////////////////
CtrlDlg::CtrlDlg(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_Dialog)
{
	Movable = false;
	AcceptChildren = true;
}

void CtrlDlg::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);
	
	// Draw the border
	LgiWideBorder(pDC, r, RAISED);
	pDC->Colour(LC_MED, 24);
	LgiFlatBorder(pDC, r, 1);

	// Draw the title bar
	Title = r;
	Title.y2 = min(Title.y2, Title.y1 + 13);
	pDC->Colour(GetSysColor(COLOR_ACTIVECAPTION), 24);
	pDC->Rectangle(&Title);
	
	if (Str)
	{
		char *Text = Str->Get();
		if (Text)
		{
			SysFont->Fore(GetSysColor(COLOR_CAPTIONTEXT));
			SysFont->Transparent(true);
			SysFont->Text(pDC, Title.x1 + 10, Title.y1, Text);
		}
	}

	// Draw the client area
	Client = r;
	Client.y1 = Title.y2 + 1;

	// Draw the grid
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&Client);
	pDC->Colour(Rgb24(0x80, 0x80, 0x80), 24);
	for (int y=Client.y1; y<Client.y2; y+=GRID_Y)
	{
		for (int x=Client.x1; x<Client.x2; x+=GRID_X)
		{
			pDC->Set(x, y);
		}
	}

	// Draw the children
	// GView::OnPaint(pDC);

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

// Text box

CtrlText::CtrlText(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_StaticText)
{
}

void CtrlText::OnPaint(GSurface *pDC)
{
	Client.ZOff(X()-1, Y()-1);
	char *Text = Str->Get();
	SysFont->Fore(0);
	SysFont->Transparent(true);

	if (Text)
	{
		int y = 0;
		char *Start = Text;
		for (char *s = Text; 1; s++)
		{
			if ((*s == '\\' AND *(s+1) == 'n') OR (*s == 0))
			{
				SysFont->Text(pDC, 0, y, Start, (int)s-(int)Start);
				y += 15;
				Start = s + 2;
				if (*s == '\\') s++;
			}

			if (*s == 0) break;
		}
	}
}

// Editbox

CtrlEditbox::CtrlEditbox(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_EditBox)
{
	Password = false;
	MultiLine = false;
}

void CtrlEditbox::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);
	Client = r;
	
	// Draw the ctrl
	LgiWideBorder(pDC, r, SUNKEN);
	pDC->Colour(Enabled() ? LC_WORKSPACE : LC_MED, 24);
	pDC->Rectangle(&r);

	char *Text = Str->Get();
	SysFont->Fore(Enabled() ? 0 : LC_LOW);
	SysFont->Transparent(true);
	if (Text)
	{
		if (Password)
		{
			char *t = NewStr(Text);
			if (t)
			{
				for (char *p = t; *p; p++) *p = '*';
				SysFont->Text(pDC, 4, 4, t);
				DeleteArray(t);
			}
		}
		else
		{
			SysFont->Text(pDC, 4, 4, Text);
		}
	}

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

#define VAL_Password		"Password"
#define VAL_MultiLine		"MultiLine"

bool CtrlEditbox::GetFields(DataDlgFieldList &Fields, ObjProperties *Props)
{
	bool Status = ResDialogCtrl::GetFields(Fields, Props);
	if (Status)
	{
		Fields.Insert(NEW(DataDlgField(DATA_BOOL, 300, VAL_Password, "Password")));
		Fields.Insert(NEW(DataDlgField(DATA_BOOL, 301, VAL_MultiLine, "MultiLine")));
	}
	return Status;
}

bool CtrlEditbox::Serialize(bool Write)
{
	bool Status = ResDialogCtrl::Serialize(Write);
	if (Status)
	{
		if (Write)
		{
			Values.Set(VAL_Password, Password);
			Values.Set(VAL_MultiLine, MultiLine);
		}
		else
		{
			int i = false;
			if (Values.Get(VAL_Password, i))
			{
				Password = i;
			}
			if (Values.Get(VAL_MultiLine, i))
			{
				MultiLine = i;
			}
		}
	}
	return Status;
}

// Check box
CtrlCheckbox::CtrlCheckbox(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_CheckBox)
{
}

void CtrlCheckbox::OnPaint(GSurface *pDC)
{
	Client.ZOff(X()-1, Y()-1);
	GRegion r(0, 0, 12, 12);
	
	// Draw the ctrl
	LgiWideBorder(pDC, r, SUNKEN);
	pDC->Colour(LC_WORKSPACE, 24);
	pDC->Rectangle(&r);

	GdcPt2 Pt[6] = {
		GdcPt2(3, 4),
		GdcPt2(3, 7),
		GdcPt2(5, 10),
		GdcPt2(10, 5),
		GdcPt2(10, 2),
		GdcPt2(5, 7)};
	pDC->Colour(0);
	pDC->Polygon(6, Pt);
	pDC->Set(3, 5);

	char *Text = Str->Get();
	if (Text)
	{
		SysFont->Fore(GetSysColor(COLOR_BTNTEXT));
		SysFont->Transparent(true);
		SysFont->Text(pDC, r.x2 + 10, r.y1-2, Text);
	}

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

// Button

CtrlButton::CtrlButton(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_Button)
{
}

void CtrlButton::OnPaint(GSurface *pDC)
{
	Client.ZOff(X()-1, Y()-1);
	GRegion r = Client;
	char *Text = Str->Get();
	int x, y;
	
	// Draw the ctrl
	LgiWideBorder(pDC, r, RAISED);
	SysFont->Fore(GetSysColor(COLOR_BTNTEXT));
	SysFont->Back(LC_MED);
	SysFont->Transparent(false);
	SysFont->Size(&x, &y, Text);
	SysFont->Text(pDC, r.x1 + ((r.X()-x)/2), r.y1 + ((r.Y()-y)/2), Text, -1, &r);

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

// Group

CtrlGroup::CtrlGroup(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_Group)
{
	AcceptChildren = true;
}

void CtrlGroup::OnPaint(GSurface *pDC)
{
	Client.ZOff(X()-1, Y()-1);
	GRegion r = Client;
	
	// Draw the ctrl
	r.y1 += 5;
	LgiWideBorder(pDC, r, CHISEL);
	r.y1 -= 5;
	SysFont->Fore(GetSysColor(COLOR_BTNTEXT));
	SysFont->Back(LC_MED);
	SysFont->Transparent(false);
	char *Text = Str->Get();
	SysFont->Text(pDC, r.x1 + 8, r.y1 - 2, Text);

	// Draw children
	//GWindow::OnPaint(pDC);

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

//Radio button

uint32 RadioBmp[] = {
0xC0C0C0C0, 0xC0C0C0C0, 0xC0C0C0C0, 0x80808080, 0x80808080, 0x80808080, 0xC0C0C0C0, 0xC0C0C0C0, 
0xC0C0C0C0, 0xC0C0C0C0, 0x8080C0C0, 0x80808080, 0x00000000, 0x00000000, 0x00000000, 0x80808080, 
0xC0C08080, 0xC0C0C0C0, 0x80C0C0C0, 0x00008080, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 
0x00000000, 0xFFFF0000, 0xC0C0C0FF, 0x80C0C0C0, 0x00008080, 0xFFFFFF00, 0xFFFFFFFF, 0xFFFFFFFF, 
0xFFFFFFFF, 0xDFFFFFFF, 0xFFFFDFDF, 0xC0C0C0FF, 0x00808080, 0xFFFF0000, 0xFFFFFFFF, 0x00FFFFFF, 
0x00000000, 0xFFFFFF00, 0xFFFFFFFF, 0xDFDFFFFF, 0xFFFFFFDF, 0x00808080, 0xFFFF0000, 0xFFFFFFFF, 
0x00000000, 0x00000000, 0x00000000, 0xFFFFFFFF, 0xDFDFFFFF, 0xFFFFFFDF, 0x00808080, 0xFFFF0000, 
0xFFFFFFFF, 0x00000000, 0x00000000, 0x00000000, 0xFFFFFFFF, 0xDFDFFFFF, 0xFFFFFFDF, 0x00808080, 
0xFFFF0000, 0xFFFFFFFF, 0x00FFFFFF, 0x00000000, 0xFFFFFF00, 0xFFFFFFFF, 0xDFDFFFFF, 0xFFFFFFDF, 
0x80C0C0C0, 0x00008080, 0xFFFFFF00, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xDFFFFFFF, 0xFFFFDFDF, 
0xC0C0C0FF, 0x80C0C0C0, 0xDFDF8080, 0xDFDFDFDF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xDFDFDFDF, 
0xFFFFDFDF, 0xC0C0C0FF, 0xC0C0C0C0, 0xFFFFC0C0, 0xFFFFFFFF, 0xDFDFDFDF, 0xDFDFDFDF, 0xDFDFDFDF, 
0xFFFFFFFF, 0xC0C0FFFF, 0xC0C0C0C0, 0xC0C0C0C0, 0xC0C0C0C0, 0xC0C0C0C0, 0xFFFFFFFF, 0xFFFFFFFF, 
0xFFFFFFFF, 0xC0C0C0C0, 0xC0C0C0C0, 0xC0C0C0C0};

CtrlRadio::CtrlRadio(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_RadioBox)
{
	Bmp = NEW(GMemDC);
	if (Bmp AND Bmp->Create(12, 12, 24))
	{
		int Len = ((Bmp->X()*Bmp->GetBits())+31)/32*4;
		for (int y=0; y<Bmp->Y(); y++)
		{
			for (int x=0; x<Bmp->X(); x++)
			{
				uchar *s = ((uchar*) RadioBmp) + (y * Len) + (x * 3);
				Bmp->Colour(Rgb24(s[0], s[1], s[2]), 24);
				Bmp->Set(x, y);
			}
		}
	}
}

CtrlRadio::~CtrlRadio()
{
	DeleteObj(Bmp);
}

void CtrlRadio::OnPaint(GSurface *pDC)
{
	Client.ZOff(X()-1, Y()-1);
	GRegion r(0, 0, 12, 12);
	
	// Draw the ctrl
	if (RadioBmp)
	{
		pDC->Blt(r.x1, r.y1, Bmp);
	}

	char *Text = Str->Get();
	if (Text)
	{
		SysFont->Fore(GetSysColor(COLOR_BTNTEXT));
		SysFont->Transparent(true);
		SysFont->Text(pDC, r.x2 + 10, r.y1-2, Text);
	}

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

// Tab

CtrlTab::CtrlTab(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_Tab)
{
}

void CtrlTab::ListChildren(List<ResDialogCtrl> &l, bool Deep)
{
	ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(GetParent());
	CtrlTabs *Par = dynamic_cast<CtrlTabs*>(Ctrl);
	ASSERT(Par);

	int MyIndex = Par->Tabs.IndexOf(this);
	ASSERT(MyIndex >= 0);

	List<GView> *CList = (Par->Current == MyIndex) ? &Par->Children : &Children;
	for (GView *w = CList->First(); w; w = CList->Next())
	{
		ResDialogCtrl *c = dynamic_cast<ResDialogCtrl*>(w);
		if (c)
		{
			l.Insert(c);
			if (Deep)
			{
				c->ListChildren(l, Deep);
			}
		}
	}
}

// Tab control

CtrlTabs::CtrlTabs(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_TabView)
{
	AcceptChildren = true;
	Current = 0;

	for (int i=0; i<3; i++)
	{
		CtrlTab *t = NEW(CtrlTab(dlg));
		if (t)
		{
			char Text[256];
			sprintf(Text, "Tab %i", i+1);
			if (t->Str)
			{
				t->Str->Set(Text);
				t->SetParent(this);
				Tabs.Insert(t);
			}
			else
			{
				DeleteObj(t);
			}
		}
	}
}

CtrlTabs::~CtrlTabs()
{
	Empty();
}

GRegion CtrlTabs::GetMinSize()
{
	List<ResDialogCtrl> l;
	ListChildren(l, false);

	GRegion r(0, 0, GRID_X-1, GRID_Y-1);

	// don't resize smaller than the tabs
	for (CtrlTab *t = Tabs.First(); t; t = Tabs.Next())
	{
		r.x2 = max(r.x2, t->r.x2 + 10);
		r.y2 = max(r.y2, t->r.y2 + 10);
	}

	// don't resize smaller than any of the children
	// on any of the tabs
	for (ResDialogCtrl *c=l.First(); c; c=l.Next())
	{
		r.Union(&c->GetPos());
	}

	return r;
}

void CtrlTabs::ListChildren(List<ResDialogCtrl> &l, bool Deep)
{
	int n=0;
	for (CtrlTab *t = Tabs.First(); t; t = Tabs.Next(), n++)
	{
		List<GView> *CList = (Current == n) ? &Children : &t->Children;
		for (GView *w = CList->First(); w; w = CList->Next())
		{
			ResDialogCtrl *c = dynamic_cast<ResDialogCtrl*>(w);
			if (c)
			{
				l.Insert(c);
				c->ListChildren(l, Deep);
			}
		}
	}
}

void CtrlTabs::Empty()
{
	ToTab();
	for (CtrlTab *t = Tabs.First(); t; t = Tabs.Next())
	{
		DeleteObj(t);
	}
	Tabs.Empty();
}

void CtrlTabs::OnPaint(GSurface *pDC)
{
	// Draw the ctrl
	Title.ZOff(X()-1, 17);
	Client.ZOff(X()-1, Y()-1);
	Client.y1 = Title.y2;
	GRegion r = Client;
	LgiWideBorder(pDC, r, RAISED);

	// Draw the tabs
	int i = 0;
	int x = 2;
	for (CtrlTab *Tab = Tabs.First(); Tab; Tab = Tabs.Next(), i++)
	{
		int Sx, Sy;
		char *Str = Tab->Str->Get();
		SysFont->Size(&Sx, &Sy, Str);

		int Width = 12 + Sx;
		GRegion t(x, Title.y1 + 2, x + Width - 1, Title.y2 - 1);

		if (Current == i)
		{
			t.Size(-2, -2);
			if (Tab->Children.GetItems() > 0)
			{
				FromTab();
			}
		}

		Tab->r = t;

		pDC->Colour(LC_LIGHT, 24);
		pDC->Line(t.x1, t.y1+2, t.x1, t.y2);
		pDC->Set(t.x1+1, t.y1+1);
		pDC->Line(t.x1+2, t.y1, t.x2-2, t.y1);

		pDC->Colour(LC_MED, 24);
		pDC->Line(t.x1+1, t.y1+2, t.x1+1, t.y2);
		pDC->Line(t.x1+2, t.y1+1, t.x2-2, t.y1+1);

		pDC->Colour(LC_LOW, 24);
		pDC->Line(t.x2-1, t.y1, t.x2-1, t.y2);
		pDC->Colour(0, 24);
		pDC->Line(t.x2, t.y1+2, t.x2, t.y2);

		t.Size(2, 2);
		t.y2 += 2;

		SysFont->Fore(GetSysColor(COLOR_BTNTEXT));
		SysFont->Back(LC_MED);
		SysFont->Transparent(false);
		SysFont->Text(pDC, t.x1 + ((t.X()-Sx)/2), t.y1 + ((t.Y()-Sy)/2), Str, -1, &t);

		x += Width + ((Current == i) ? 2 : 1);
	}

	// Draw children
	//GWindow::OnPaint(pDC);

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

void CtrlTabs::ToTab()
{
	CtrlTab *Cur = Tabs.ItemAt(Current);
	if (Cur)
	{
		// move all our children into the tab losing focus
		Cur->Children.Empty();
		for (GView *c = Children.First(); c; c = Children.Next())
		{
			Cur->Children.Insert(c);
			c->SetParent(Cur);
		}
		Children.Empty();
	}
}

void CtrlTabs::FromTab()
{
	CtrlTab *Cur = Tabs.ItemAt(Current);
	if (Cur)
	{
		// load all our children from the new tab
		Children.Empty();
		for (GView *c = Cur->Children.First(); c; c = Cur->Children.Next())
		{
			Children.Insert(c);
			c->SetParent(this);
		}
		Cur->Children.Empty();
	}
}

void CtrlTabs::OnMouseClick(GMouse &m)
{
	if (m.Down())
	{
		if (Title.Overlap(m.x, m.y))
		{
			// select current tab
			int i = 0;
			for (CtrlTab *Tab = Tabs.First(); Tab; Tab = Tabs.Next(), i++)
			{
				if (Tab->r.Overlap(m.x, m.y) AND
					i != Current)
				{
					ToTab();
					Current = i;
					FromTab();

					Dlg->OnSelect(this);
					Invalidate();
				}
			}
		}

		if (m.Right() AND Title.Overlap(m.x, m.y))
		{
			GSubMenu *RClick = NEW(GSubMenu);
			if (RClick)
			{
				bool HasTab = Tabs.ItemAt(Current);

				RClick->AppendItem("New tab", IDM_NEW, TRUE);
				RClick->AppendItem("Delete tab", IDM_DELETE, HasTab);
				RClick->AppendItem("Rename tab", IDM_RENAME, HasTab);
				RClick->AppendItem("Move tab left", IDM_MOVE_LEFT, HasTab);
				RClick->AppendItem("Move tab right", IDM_MOVE_RIGHT, HasTab);

				if (GetMouse(m, true))
				{
					switch (RClick->Float(this, m.x, m.y, false))
					{
						case IDM_NEW:
						{
							CtrlTab *t = NEW(CtrlTab(Dlg));
							if (t)
							{
								char Text[256];
								sprintf(Text, "Tab %i", Tabs.GetItems()+1);
								t->Str->Set(Text);
								t->SetParent(this);

								Tabs.Insert(t);
							}
							break;
						}
						case IDM_DELETE:
						{
							CtrlTab *t = Tabs.ItemAt(Current);
							if (t)
							{
								ToTab();
								Tabs.Delete(t);
								DeleteObj(t);
								FromTab();
							}
							break;
						}
						case IDM_RENAME:
						{
							CtrlTab *t = Tabs.ItemAt(Current);
							if (t)
							{
								GInput Input(this, t->Str->Get(), "Enter tab name:", "Rename");
								Input.SetParent(Dlg);
								if (Input.DoModal() AND Input.Str)
								{
									t->Str->Set(Input.Str);
								}
							}
							break;
						}
						case IDM_MOVE_LEFT:
						{
							CtrlTab *t = Tabs.ItemAt(Current);
							if (t AND Current > 0)
							{
								Tabs.Delete(t);
								Tabs.Insert(t, --Current);
							}
							break;
						}
						case IDM_MOVE_RIGHT:
						{
							CtrlTab *t = Tabs.ItemAt(Current);
							if (t AND Current < Tabs.GetItems()-1)
							{
								Tabs.Delete(t);
								Tabs.Insert(t, ++Current);
							}
							break;
						}
					}

					Invalidate();
				}
			}

			return;
		}
	}

	ResDialogCtrl::OnMouseClick(m);
}

// List column

ListCol::ListCol(ResDialog *dlg, char *s, int Width) :
	ResDialogCtrl(dlg, Res_Column)
{
	if (s AND Str)
	{
		Str->Set(s);
	}
	
	Pos.ZOff(Width-1, 18);
}

// List control

CtrlList::CtrlList(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_ListView)
{
	DragCol = -1;
}

CtrlList::~CtrlList()
{
	Empty();
}

void CtrlList::ListChildren(List<ResDialogCtrl> &l, bool Deep)
{
	if (Deep)
	{
		for (GView *w = Cols.First(); w; w = Cols.Next())
		{
			ResDialogCtrl *c = dynamic_cast<ResDialogCtrl*>(w);
			if (c)
			{
				l.Insert(c);
				c->ListChildren(l);
			}
		}
	}
}

void CtrlList::Empty()
{
	for (ListCol *c = Cols.First(); c; c = Cols.Next())
	{
		DeleteObj(c);
	}
	Cols.Empty();
}

void CtrlList::OnMouseClick(GMouse &m)
{
	if (m.Down())
	{
		if (Title.Overlap(m.x, m.y))
		{
			int x=0;
			ListCol *c = 0;
			int DragOver = -1;
			for (ListCol *Col = Cols.First(); Col; Col = Cols.Next())
			{
				if (m.x >= Col->r().x1 AND m.x <= Col->r().x2)
				{
					if (m.x > Col->r().x2 - 6)
					{
						DragOver = Cols.IndexOf(Col);
					}
					if (m.x < Col->r().x1 + 6)
					{
						DragOver = Cols.IndexOf(Col) - 1;
					}
					c = Col;
					break;
				}
				x += Col->r().X();
			}

			if (m.Left())
			{
				if (DragOver >= 0)
				{
					DragCol = DragOver;
					Capture(true);
				}
			}
			else if (m.Right())
			{
				GSubMenu *RClick = NEW(GSubMenu);
				if (RClick)
				{
					bool HasCol = c != 0;
					RClick->AppendItem("New column", IDM_NEW, TRUE);
					RClick->AppendItem("Delete column", IDM_DELETE, HasCol);
					RClick->AppendItem("Rename column", IDM_RENAME, HasCol);
					RClick->AppendItem("Move column left", IDM_MOVE_LEFT, HasCol);
					RClick->AppendItem("Move column right", IDM_MOVE_RIGHT, HasCol);

					if (GetMouse(m, true))
					{
						switch (RClick->Float(this, m.x, m.y, false))
						{
							case IDM_NEW:
							{
								ListCol *c = NEW(ListCol(Dlg));
								if (c)
								{
									char Text[256];
									sprintf(Text, "Col %i", Cols.GetItems()+1);
									c->Str->Set(Text);
									Cols.Insert(c);
								}
								break;
							}
							case IDM_DELETE:
							{
								Cols.Delete(c);
								DeleteObj(c);
								break;
							}
							case IDM_RENAME:
							{
								if (c)
								{
									GInput Input(this, c->Str->Get(), "Enter column name:", "Rename");
									Input.SetParent(Dlg);
									if (Input.DoModal() AND Input.Str)
									{
										c->Str->Set(Input.Str);
									}
								}
								break;
							}
							case IDM_MOVE_LEFT:
							{
								int Current = Cols.IndexOf(c);
								if (c AND Current > 0)
								{
									Cols.Delete(c);
									Cols.Insert(c, --Current);
								}
								break;
							}
							case IDM_MOVE_RIGHT:
							{
								int Current = Cols.IndexOf(c);
								if (c AND Current < Cols.GetItems()-1)
								{
									Cols.Delete(c);
									Cols.Insert(c, ++Current);
								}
								break;
							}
						}

						Invalidate();
					}

					DeleteObj(RClick);
					return;
				}
			}
		}
	}
	else
	{
		if (DragCol >= 0)
		{
			Capture(false);
			DragCol = -1;
		}
	}

	ResDialogCtrl::OnMouseClick(m);
}

void CtrlList::OnMouseMove(GMouse &m)
{
	if (DragCol >= 0)
	{
		int i=0, x=0;;
		for (ListCol *Col = Cols.First(); Col; Col = Cols.Next(), i++)
		{
			if (i == DragCol)
			{
				int Dx = (m.x - x - Title.x1);
				
				GRegion r = Col->GetPos();
				r.x2 = r.x1 + Dx;
				Col->SetPos(r);
				break;
			}

			x += Col->r().X();
		}

		Invalidate();
	}

	ResDialogCtrl::OnMouseMove(m);
}

void CtrlList::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);

	// Draw the ctrl
	LgiWideBorder(pDC, r, SUNKEN);
	Title = r;
	Title.y2 = Title.y1 + 15;
	Client = r;
	Client.y1 = Title.y2 + 1;

	pDC->Colour(LC_WORKSPACE, 24);
	pDC->Rectangle(&Client);

	int x = Title.x1;
	for (ListCol *c = Cols.First(); c; c = Cols.Next())
	{
		int Width = c->r().X();
		c->r().Set(x, Title.y1, x + Width - 1, Title.y2);
		GRegion r = c->r();
		r.x2 = min(r.x2, Title.x2);
		x = r.x2 + 1;
		if (r.Valid())
		{
			LgiWideBorder(pDC, r, RAISED);

			SysFont->Fore(GetSysColor(COLOR_BTNTEXT));
			SysFont->Back(LC_MED);
			SysFont->Transparent(false);

			int Sx, Sy;
			char *Str = c->Str->Get();
			if (NOT Str) Str = "";
			SysFont->Size(&Sx, &Sy, Str);
			SysFont->Text(pDC, r.x1 + 2, r.y1 + ((r.Y()-Sy)/2) - 1, Str, -1, &r);
		}
	}

	GRegion Client(x, Title.y1, Title.x2, Title.y2);
	if (Client.Valid())
	{
		LgiWideBorder(pDC, Client, RAISED);
		pDC->Colour(LC_MED, 24);;
		pDC->Rectangle(&Client);
	}

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

// Combo box
CtrlComboBox::CtrlComboBox(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_ComboBox)
{
}

void CtrlComboBox::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);
	Client = r;
	
	// Draw the ctrl
	LgiWideBorder(pDC, r, SUNKEN);
	
	// Allocate space
	GRegion e = r;
	e.x2 -= 15;
	GRegion d = r;
	d.x1 = e.x2 + 1;

	// Draw edit
	pDC->Colour(LC_WORKSPACE, 24);
	pDC->Rectangle(&e);

	// Draw drap down
	LgiWideBorder(pDC, d, RAISED);
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&d);

	pDC->Colour(0, 24);
	int Size = 4;
	int cx = (d.X()/2) + d.x1 - 4;
	int cy = (d.Y()/2) + d.y1 - 3;
	for (int i=1; i<=Size; i++)
	{
		pDC->Line(cx+i, cy+i, cx+(Size*2)-i, cy+i);
	}

	// Text
	char *Text = Str->Get();
	SysFont->Fore(0);
	SysFont->Transparent(true);
	if (Text) SysFont->Text(pDC, 4, 4, Text);

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

////////////////////////////////////////////////////////////////////
CtrlScrollBar::CtrlScrollBar(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_ScrollBar)
{
}

void CtrlScrollBar::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);
	Client = r;
	
	// Draw the ctrl
	bool Vertical = r.Y() > r.X();
	int ButSize = Vertical ? r.X() : r.Y();
	GRegion a, b, c;
	if (Vertical)
	{
		a.Set(r.x1, r.y1, r.x2, r.y1 + ButSize);
		c.Set(r.x1, r.y2 - ButSize, r.x2, r.y2);
		b.Set(r.x1, a.y2 + 1, r.x2, c.y1 - 1);
	}
	else
	{
		a.Set(r.x1, r.y1, r.x1 + ButSize, r.y2);
		c.Set(r.x2 - ButSize, r.y1, r.x2, r.y2);
		b.Set(a.x2 + 1, r.y1, c.x1 - 1, r.y2);
	}

	// Buttons
	LgiWideBorder(pDC, a, RAISED);
	LgiWideBorder(pDC, c, RAISED);
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&a);
	pDC->Rectangle(&c);

	// Arrows
	pDC->Colour(0);
	int x = a.x1 + (a.X()/2) - 2;
	int y = a.y1 + (a.Y()/2);
	for (int i=0; i<5; i++)
	{
		pDC->Line(x+i, y-i, x+i, y+i);
	}
	x = c.x1 + (c.X()/2) - 2;
	y = c.y1 + (c.Y()/2);
	for (i=0; i<5; i++)
	{
		pDC->Line(x+i, y-(4-i), x+i, y+(4-i));
	}

	// Slider
	pDC->Colour(0xd0d0d0, 24);
	pDC->Rectangle(&b);

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

////////////////////////////////////////////////////////////////////
CtrlTree::CtrlTree(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_TreeView)
{
}

void CtrlTree::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);
	Client = r;

	LgiWideBorder(pDC, r, SUNKEN);
	pDC->Colour(Rgb24(255, 255, 255), 24);
	pDC->Rectangle(&r);
	SysFont->Colour(0, 0xffffff);
	SysFont->Transparent(true);
	SysFont->Text(pDC, r.x1 + 3, r.y1 + 3, "Tree", -1, &r);

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

////////////////////////////////////////////////////////////////////
CtrlBitmap::CtrlBitmap(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_Bitmap)
{
}

void CtrlBitmap::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);
	Client = r;

	LgiWideBorder(pDC, r, SUNKEN);
	pDC->Colour(Rgb24(255, 255, 255), 24);
	pDC->Rectangle(&r);
	pDC->Colour(0, 24);
	pDC->Line(r.x1, r.y1, r.x2, r.y2);
	pDC->Line(r.x2, r.y1, r.x1, r.y2);

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

////////////////////////////////////////////////////////////////////
CtrlProgress::CtrlProgress(ResDialog *dlg) :
	ResDialogCtrl(dlg, Res_Progress)
{
}

void CtrlProgress::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);
	Client = r;

	LgiWideBorder(pDC, r, SUNKEN);

	COLOUR Flat = Rgb24(0x7f, 0x7f, 0x7f);
	COLOUR White = Rgb24(0xff, 0xff, 0xff);
	int Pos = 60;
	int x = Pos * r.X() / 100;

	pDC->Colour(Flat, 24);
	pDC->Rectangle(r.x1, r.y1, r.x1 + x, r.y2);
	
	pDC->Colour(White, 24);
	pDC->Rectangle(r.x1 + x + 1, r.y1, r.x2, r.y2);

	// Draw any rubber band
	ResDialogCtrl::OnPaint(pDC);
}

////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////
ResStringGroup *ResDialog::Symbols = 0;
int ResDialog::SymbolRefs = 0;
bool ResDialog::CreateSymbols = true;

void ResDialog::AddLanguage(int Id)
{
	if (Symbols)
	{
		Symbols->AppendLanguage(Id);
	}
}

int GooberCaps[] = {
	RESIZE_X1 | RESIZE_Y1,
	RESIZE_Y1,
	RESIZE_X2 | RESIZE_Y1,
	RESIZE_X2,
	RESIZE_X2 | RESIZE_Y2,
	RESIZE_Y2,
	RESIZE_X1 | RESIZE_Y2,
	RESIZE_X1};

ResDialog::ResDialog(AppWnd *w, int type) : 
	Resource(w, type)
{
	// Maintain a string group just for our dialog
	// defines 
	if (NOT Symbols AND w)
	{
		// First check to see of the symbols have been loaded from a file
		// and we don't have a pointer to it yet
		Symbols = App()->GetDialogSymbols();
		if (NOT Symbols)
		{
			// else we need to create the symbols object
			Symbols = NEW(ResStringGroup(w));
			if (Symbols)
			{
				Symbols->Name(StrDialogSymbols);
				w->InsertObject(TYPE_STRING, Symbols);
			}
		}

		if (Symbols)
		{
			Symbols->SystemObject(true);
		}
	}
	SymbolRefs++;

	// Init dialog resource
	Ui = 0;
	DlgPos.ZOff(299, 199);

	DragGoober = -1;
	DragX = 0;
	DragY = 0;
	DragCtrl = 0;
}

ResDialog::~ResDialog()
{
	// Decrement and delete the shared string group
	SymbolRefs--;
	if (SymbolRefs < 1)
	{
		App()->DelObject(Symbols);
		Symbols = 0;
	}

	// Delete our Ui
	DeleteObj(Ui);
}

char *ResDialog::Name()
{
	ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(Children.First());
	if (Ctrl AND Ctrl->Str AND Ctrl->Str->Define)
	{
		return Ctrl->Str->Define;
	}
	return "<no object>";
}

bool ResDialog::Name(char *n)
{
	ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(Children.First());
	if (Ctrl AND Ctrl->Str)
	{
		DeleteArray(Ctrl->Str->Define);
		Ctrl->Str->Define = NewStr((n)?n:"");
		return Ctrl->Str->Define != 0;
	}
	return false;
}

char *ResDialog::StringFromRef(int Ref)
{
	if (Symbols)
	{
		ResString *Str = Symbols->FindRef(Ref);
		if (Str)
		{
			return Str->Get();
		}
	}

	return 0;
}

bool ResDialog::Res_GetProperties(ResObject *Obj, ObjProperties *Props)
{
	ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(Obj);
	if (Ctrl AND Props)
	{
		Ctrl->Serialize(true);
		*Props = Ctrl->Values;
		return true;
	}
	return false;
}

bool ResDialog::Res_SetProperties(ResObject *Obj, ObjProperties *Props)
{
	ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(Obj);
	if (Ctrl AND Props)
	{
		Ctrl->Values = *Props;
		Ctrl->Serialize(false);
		return true;
	}
	return false;
}

ResObject *ResDialog::CreateObject(char *Name, ResObject *Parent)
{
	return dynamic_cast<ResObject*>(CreateCtrl(Name));
}

void ResDialog::Res_SetPos(ResObject *Obj, int x1, int y1, int x2, int y2)
{
	if (Obj)
	{
		ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>((ResDialogCtrl*)Obj);
		if (Ctrl)
		{
			Ctrl->SetPos(GRegion(x1, y1, x2, y2));
		}
	}
}

void ResDialog::Res_SetPos(ResObject *Obj, char *s)
{
	if (Obj)
	{
		ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>((ResDialogCtrl*)Obj);
		if (Ctrl)
		{
			Ctrl->ReadPos(s);
		}
	}
}

GRegion ResDialog::Res_GetPos(ResObject *Obj)
{
	if (Obj)
	{
		ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>((ResDialogCtrl*)Obj);
		if (Ctrl)
		{
			return Ctrl->GetPos();
		}
	}
	return GRegion(0, 0, 0, 0);
}

int ResDialog::Res_GetStrRef(ResObject *Obj)
{
	if (Obj)
	{
		ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>((ResDialogCtrl*)Obj);
		if (Ctrl)
		{
			return Ctrl->Str->Ref;
		}
	}
	return -1;
}

void ResDialog::Res_SetStrRef(ResObject *Obj, int Ref)
{
	ResDialogCtrl *Ctrl = 0;
	if (Obj AND Symbols)
	{
		Ctrl = dynamic_cast<ResDialogCtrl*>((ResDialogCtrl*)Obj);
		if (Ctrl)
		{
			Ctrl->StrFromRef(Ref);
		}
	}

	ASSERT(Ctrl AND Ctrl->Str);
}

void ResDialog::Res_Attach(ResObject *Obj, ResObject *Parent)
{
	if (Obj AND Parent)
	{
		ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>((ResDialogCtrl*)Obj);
		ResDialogCtrl *Par = dynamic_cast<ResDialogCtrl*>((ResDialogCtrl*)Parent);
		if (Ctrl AND Par)
		{
			Par->AttachCtrl(Ctrl);
		}
	}
}

bool ResDialog::Res_GetChildren(ResObject *Obj, List<ResObject> *l, bool Deep)
{
	bool Status = false;
	if (Obj)
	{
		ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>((ResDialogCtrl*)Obj);
		if (Ctrl)
		{
			Status = true;

			List<ResDialogCtrl> Child;
			Ctrl->ListChildren(Child, Deep);
			for (ResDialogCtrl *o = Child.First(); o; o = Child.Next())
			{
				l->Insert(o);
			}
		}
	}

	return Status;
}

void ResDialog::Res_Append(ResObject *Obj, ResObject *Parent)
{
	if (Obj AND Parent)
	{
		CtrlTabs *Tabs = dynamic_cast<CtrlTabs*>(Obj);
		CtrlTab *Tab = dynamic_cast<CtrlTab*>(Parent);
		if (Tabs AND Tab)
		{
			Tab->SetParent(Tabs);
			Tabs->Tabs.Insert(Tab);

			if (Tabs->Tabs.GetItems() == 1)
			{
				Tabs->FromTab();
			}
			return;
		}

		CtrlList *Lst = dynamic_cast<CtrlList*>(Obj);
		ListCol *Col = dynamic_cast<ListCol*>(Parent);
		if (Lst AND Col)
		{
			Lst->Cols.Insert(Col);
			return;
		}
	}
}

bool ResDialog::Res_GetItems(ResObject *Obj, List<ResObject> *l)
{
	if (Obj AND l)
	{
		CtrlTabs *Tabs = dynamic_cast<CtrlTabs*>(Obj);
		if (Tabs)
		{
			for (	CtrlTab *Tab = Tabs->Tabs.First();
					Tab;
					Tab = Tabs->Tabs.Next())
			{
				l->Insert(Tab);
			}

			return true;
		}

		CtrlList *Lst = dynamic_cast<CtrlList*>(Obj);
		if (Lst)
		{
			for (	ListCol *Col = Lst->Cols.First();
					Col;
					Col = Lst->Cols.Next())
			{
				l->Insert(Col);
			}

			return true;
		}
	}

	return false;
}

void ResDialog::Create()
{
	CtrlDlg *Dlg = NEW(CtrlDlg(this));
	if (Dlg)
	{
		GRegion r = DlgPos;
		r.Offset(GOOBER_BORDER, GOOBER_BORDER);

		Children.Insert(Dlg);
		Dlg->SetParent(this);
		Dlg->SetPos(r);
		if (Dlg->Str)
		{
			Dlg->Str->Set("Dialog");
		}
	}
}

void ResDialog::Delete()
{
	// Deselect the dialog ctrl
	OnDeselect(dynamic_cast<ResDialogCtrl*>(Children.First()));

	// Delete selected controls
	for (ResDialogCtrl *c = Selection.First(); c; c = Selection.First())
	{
		DeleteObj(c);
	}

	// Repaint
	GView::Invalidate();
}

bool IsChild(ResDialogCtrl *Parent, ResDialogCtrl *Child)
{
	if (Parent AND Child)
	{
		for (	ResDialogCtrl *c = dynamic_cast<ResDialogCtrl*>(Child->GetParent());
				c;
				c = dynamic_cast<ResDialogCtrl*>(c->GetParent()))
		{
			if (c == Parent)
			{
				return true;
			}
		}
	}

	return false;
}

void GetTopCtrls(List<ResDialogCtrl> &Top, List<ResDialogCtrl> &Selection)
{
	// all children will automatically be cut as well

	for (ResDialogCtrl *c = Selection.First(); c; c = Selection.Next())
	{
		// is c a child of an item already in Top?
		bool Ignore = false;
		for (ResDialogCtrl *p = Top.First(); p; p = Top.Next())
		{
			if (IsChild(p, c))
			{
				Ignore = true;
				break;
			}
		}

		if (NOT Ignore)
		{
			// is not a child
			Top.Insert(c);
		}
	}
}

void ResDialog::Copy(bool Delete)
{
	// Deselect the dialog... can't cut that
	OnDeselect(dynamic_cast<ResDialogCtrl*>(Children.First()));
	
	// Get top level list
	List<ResDialogCtrl> Top;
	GetTopCtrls(Top, Selection);

	GFile F;
	char *FileName = "c:\\temp\\~clip.txt";
	if (F.Open(FileName, O_WRITE))
	{
		// ok we have a top level list of ctrls
		// write them to the file
		List<ResDialogCtrl> All;
		ResDialogCtrl *c;

		// write the string resources first
		for (c = Top.First(); c; c = Top.Next())
		{
			All.Insert(c);
			c->ListChildren(All, true);
		}

		// write the strings out at the top of the block
		// so that we can reference them from the objects
		// below.
		for (c = All.First(); c; c = All.Next())
		{
			// Write the string out
			c->Str->Write(F);
		}

		// write the objects themselves
		for (c = Top.First(); c; c = Top.Next())
		{
			Res_Write(c, F);
		}

		if (Delete)
		{
			// delete them
			for (c = Selection.First(); c; c = Selection.First())
			{
				DeleteObj(c);
			}
		}
		
		// Repaint
		GView::Invalidate();

		// Read the file in and copy to the clipboard
		if (F.Open(FileName, O_READ))
		{
			int Len = F.GetSize();
			char *Data = NEW(char[Len+1]);
			if (Data)
			{
				F.Read(Data, Len);
				Data[Len] = 0;
				
				{
					GClipBoard Clip(Ui);
					Clip.Text(Data);
				}

				DeleteArray(Data);
			}
			
			F.Close();
		}

		// Delete temp file
		unlink(FileName);
	}
}

/*
		// This last ref is the last unique ref we got. when
		// we send it back to GetUniqueStrRef we increment it so
		// that we don't get the same unique ref for all these controls
		// the reason is that the GetUniqueStrRef doesn't know about our
		// controls are thus doesn't count them when looking for a unique
		// value
		int LastRef = 1;

		for (c = All.First(); c; c = All.Next())
		{
			StringId *Id = NEW(StringId);
			if (Id)
			{
				// Id object remembers the old and new ref values
				Id->Str = c->Str;
				Id->OldRef = c->Str->Ref;
				LastRef = Id->NewRef = AppWindow->GetUniqueStrRef(LastRef + 1);
				Strings.Insert(Id);
			}
		}

		for (Id = Strings.First(); Id; Id = Strings.Next())
		{
			// Id object remembers the old and new ref values
			Id->Str->Ref = Id->NewRef;
			
			// Write the string out with the new ref
			Id->Str->Write(F);
		}

		// revert string ref's back to their original value
		// this way when pasting the strings back in they
		// don't conflict ref wise.
		for (Id = Strings.First(); Id; Id = Strings.Next())
		{
			Id->Str->Ref = Id->OldRef;
			DeleteObj(Id);
		}
		

*/
class StringId
{
public:
	ResString *Str;
	int OldRef;
	int NewRef;
};

void ResDialog::Paste()
{
	// Get the clipboard data
	char *Data = 0;
	{
		GClipBoard Clip(Ui);
		Data = Clip.Text();
	}
	if (Data)
	{
		ResDialogCtrl *Container = 0;

		// Find a container to plonk the controls in
		ResDialogCtrl *c = Selection.First();
		if (c AND c->IsContainer())
		{
			Container = c;
		}

		if (NOT Container)
		{
			// Otherwise just use the dialog as the container
			Container = dynamic_cast<ResDialogCtrl*>(Children.First());
		}

		if (Container)
		{
			// remap list
			List<StringId> Strings;
			int NextRef = 0;

			// Deselect everything
			OnSelect(NULL);

			// Parse the data
			List<ResString> NewStrs;
			Xml Parser;
			Parser.ParseXml(Data);
			
			// Create the new controls, strings first
			// that way we can setup the remapping properly to avoid
			// string ref duplicates
			XmlTag *t;
			for (t = Parser.Values.First(); t; t = Parser.Values.Current())
			{
				if (stricmp(t->Name, "string") == 0)
				{
					// string tag
					ASSERT(Symbols);

					ResString *Str = Symbols->CreateStr();
					if (Str AND Str->Read(Parser.Values))
					{
						// setup remap object, so that we can make all the strings
						// unique
						StringId *Id = NEW(StringId);
						ASSERT(Id);
						Id->Str = Str;
						Id->OldRef = Str->Ref;
						NextRef = Str->Ref = Id->NewRef = AppWindow->GetUniqueStrRef(NextRef + 1);
						Strings.Insert(Id);

						// insert out new string
						NewStrs.Insert(Str);
						Parser.Values.Next();
					}
					else
					{
						break;
					}
				}
				else
				{
					// object tag

					// all strings should have been processed by the time 
					// we get in here. We remap the incomming string refs to 
					// unique values so that we don't get conflicts later.
					char *RefStr;
					if (t->Get("Ref", RefStr))
					{
						int r = atoi(RefStr);
						for (StringId *i=Strings.First(); i; i=Strings.Next())
						{
							if (i->OldRef == r)
							{
								// this string ref is to be remapped
								char Buf[32];
								sprintf(Buf, "%i", i->NewRef);
								t->Set("Ref", Buf);

								// delete the string ref map
								// it's not needed anymore
								Strings.Delete(i);
								DeleteObj(i);
								
								// leave the loop
								RefStr = 0;
								break;
							}
						}

						// if this assert failes then no map for this
						// string was found. every incomming string needs
						// to be remapped
						ASSERT(RefStr == 0);
					}

					Parser.Values.Next();
				}
			}

			// load all the objects now
			List<ResDialogCtrl> NewCtrls;
			for (t = Parser.Values.First(); t; t = Parser.Values.Current())
			{
				if (stricmp(t->Name, "string") != 0)
				{
					// object (not string)
					CreateSymbols = false;
					ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(CreateCtrl(t->Name));
					CreateSymbols = true;
					if (Ctrl AND Res_Read(Ctrl, Parser.Values))
					{
						NewCtrls.Insert(Ctrl);
					}
					else
					{
						break;
					}
				}
				else
				{
					Parser.Values.Next();
				}
			}

			// calculate the new control set's dimensions so we
			// can paste them in at the top of the container
			ResDialogCtrl *c = NewCtrls.First();
			if (c)
			{
				GRegion All = c->GetPos();
				while (c = NewCtrls.Next())
				{
					All.Union(&c->GetPos());
				}

				// now paste in the controls
				for (c = NewCtrls.First(); c; c = NewCtrls.Next())
				{
					GRegion p = c->GetPos();
					p.Offset(Container->Client.x1 + GRID_X - All.x1,
							 Container->Client.y1 + GRID_Y-All.y1);
					c->SetPos(p);

					Container->Children.Insert(c);
					c->SetParent(Container);
					OnSelect(c, false);
				}

				// reset parent size to fit
				GRegion cp = Container->GetPos();
				Container->SetPos(cp, true);
			}

			// Deduplicate all these new strings
			for (ResString *s = NewStrs.First(); s; s = NewStrs.Next())
			{
				s->UnDupelicate();
			}

			// Repaint
			GView::Invalidate();
		}
	}
}

void ResDialog::SnapPoint(GdcPt2 *p, ResDialogCtrl *From)
{
	ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(Children.First());
	if (p AND Ctrl)
	{
		int Ox = -Ctrl->Client.x1;
		int Oy = -Ctrl->Client.y1;
		GView *Parent = dynamic_cast<GView*>(Ctrl);
		for (GView *w = From; w AND w != Parent; w = w->GetParent())
		{
			Ox += w->GetPos().x1;
			Oy += w->GetPos().y1;
		}

		p->x += Ox;
		p->y += Oy;

		p->x = ((p->x + (GRID_X / 2)) / GRID_X * GRID_X);
		p->y = ((p->y + (GRID_X / 2)) / GRID_X * GRID_X);

		p->x -= Ox;
		p->y -= Oy;
	}
}

void ResDialog::SnapRect(GRegion *r, ResDialogCtrl *From)
{
	ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(Children.First());
	if (r AND Ctrl)
	{
		int Ox = -Ctrl->Client.x1;
		int Oy = -Ctrl->Client.y1;
		GView *Parent = dynamic_cast<GView*>(Ctrl);
		for (GView *w = From; w AND w != Parent; w = w->GetParent())
		{
			Ox += w->GetPos().x1;
			Oy += w->GetPos().y1;
		}

		r->Normal();
		r->Offset(Ox, Oy);

		r->x1 = ((r->x1 + (GRID_X / 2)) / GRID_X * GRID_X);
		r->y1 = ((r->y1 + (GRID_X / 2)) / GRID_X * GRID_X);
		r->x2 = ((r->x2 + (GRID_X / 2)) / GRID_X * GRID_X) - 1;
		r->y2 = ((r->y2 + (GRID_X / 2)) / GRID_X * GRID_X) - 1;

		r->Offset(-Ox, -Oy);
	}
}

bool ResDialog::IsDraging()
{
	return DragGoober >= 0;
}

int ResDialog::CurrentTool()
{
	return (Ui) ? Ui->CurrentTool() : 0;
}

void ResDialog::MoveSelection(int Dx, int Dy)
{
	ResDialogCtrl *w = Selection.First();
	if (w)
	{
		ResDialogCtrl *Parent = w->ParentCtrl();
		if (Parent)
		{
			// find dimensions of group
			GRegion All = w->GetPos();
			for (; w; w = Selection.Next())
			{
				All.Union(&w->GetPos());
			}

			// limit the move to the top-left corner of the parent's client
			if (All.x1 + Dx < Parent->Client.x1)
			{
				Dx = Parent->Client.x1 - All.x1;
			}
			if (All.y1 + Dy < Parent->Client.y1)
			{
				Dy = Parent->Client.y1 - All.y1;
			}

			// move the ctrls
			for (w = Selection.First(); w; w = Selection.Next())
			{
				GRegion Old = w->GetPos();
				GRegion New = Old;
				New.Offset(Dx, Dy);
				w->SetPos(New, false);

				Old.Size(-GOOBER_BORDER, -GOOBER_BORDER);
				New.Size(-GOOBER_BORDER, -GOOBER_BORDER);
				New.Union(&Old);
				
				w->Parent->Invalidate(&New);
			}
		}
	}
}

void ResDialog::SelectRect(ResDialogCtrl *Parent, GRegion *r, bool ClearPrev)
{
	if (ClearPrev)
	{
		Selection.Empty();
	}

	if (Parent AND r)
	{
		for (GView *c = Parent->Children.First(); c; c = Parent->Children.Next())
		{
			if (c->GetPos().Overlap(r))
			{
				ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(c);
				if (Ctrl)
				{
					if (Selection.IndexOf(Ctrl) >= 0)
					{
						Selection.Delete(Ctrl);
					}
					else
					{
						Selection.Insert(Ctrl);
					}
				}
			}
		}
	}

	Invalidate();

	if (AppWindow)
	{
		AppWindow->OnObjSelect((Selection.GetItems() == 1) ? Selection.First() : 0);
	}
}

bool ResDialog::IsSelected(ResDialogCtrl *Ctrl)
{
	return Selection.IndexOf(Ctrl) >= 0;
}

void ResDialog::OnSelect(ResDialogCtrl *Wnd, bool ClearPrev)
{
	if (ClearPrev)
	{
		Selection.Empty();
	}

	if (Wnd)
	{
		ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(Wnd);
		if (Ctrl)
		{
			Selection.Insert(Ctrl);
			if (Ui) Ui->SelectTool(UI_CURSOR);
			Invalidate();
		}

		if (AppWindow)
		{
			AppWindow->OnObjSelect((Selection.GetItems() == 1) ? Selection.First() : 0);
		}
	}
}

void ResDialog::OnDeselect(ResDialogCtrl *Wnd)
{
	if (Selection.HasItem(Wnd))
	{
		Selection.Delete(Wnd);
		AppWindow->OnObjSelect(0);
	}
}

ResDialogCtrl *ResDialog::CreateCtrl(char *Name)
{
	if (Name)
	{
		for (LgiObjectName *o = NameMap; o->Type; o++)
		{
			if (stricmp(Name, o->ResourceName) == 0)
			{
				return CreateCtrl(o->Type);
			}
		}
	}

	return 0;
}

ResDialogCtrl *ResDialog::CreateCtrl(int Tool)
{
	ResDialogCtrl *Ctrl = 0;

	switch (Tool)
	{
		case UI_DIALOG:
		{
			Ctrl = NEW(CtrlDlg(this));
			break;
		}
		case UI_TEXT:
		{
			Ctrl = NEW(CtrlText(this));
			if (Ctrl AND Ctrl->Str)
			{
				Ctrl->Str->Set("Some text");
			}
			break;
		}
		case UI_EDITBOX:
		{
			Ctrl = NEW(CtrlEditbox(this));
			break;
		}
		case UI_CHECKBOX:
		{
			Ctrl = NEW(CtrlCheckbox(this));
			if (Ctrl AND Ctrl->Str)
			{
				Ctrl->Str->Set("Checkbox");
			}
			break;
		}
		case UI_BUTTON:
		{
			Ctrl = NEW(CtrlButton(this));
			if (Ctrl AND Ctrl->Str)
			{
				static int i = 1;
				char Text[256];
				sprintf(Text, "Button %i", i++);
				Ctrl->Str->Set(Text);
			}
			break;
		}
		case UI_GROUP:
		{
			Ctrl = NEW(CtrlGroup(this));
			if (Ctrl AND Ctrl->Str)
			{
				Ctrl->Str->Set("Text");
			}
			break;
		}
		case UI_RADIO:
		{
			Ctrl = NEW(CtrlRadio(this));
			if (Ctrl AND Ctrl->Str)
			{
				Ctrl->Str->Set("Radio");
			}
			break;
		}
		case UI_TAB:
		{
			Ctrl = NEW(CtrlTab(this));
			break;
		}
		case UI_TABS:
		{
			Ctrl = NEW(CtrlTabs(this));
			break;
		}
		case UI_LIST:
		{
			Ctrl = NEW(CtrlList(this));
			break;
		}
		case UI_COLUMN:
		{
			Ctrl = NEW(ListCol(this));
			break;
		}
		case UI_COMBO:
		{
			Ctrl = NEW(CtrlComboBox(this));
			break;
		}
		case UI_TREE:
		{
			Ctrl = NEW(CtrlTree(this));
			break;
		}
		case UI_BITMAP:
		{
			Ctrl = NEW(CtrlBitmap(this));
			break;
		}
		case UI_SCROLL_BAR:
		{
			Ctrl = NEW(CtrlScrollBar(this));
			break;
		}
		case UI_PROGRESS:
		{
			Ctrl = NEW(CtrlProgress(this));
			break;
		}
	}

	return Ctrl;
}

GView *ResDialog::CreateUI()
{
	return Ui = NEW(ResDialogUi(this));
}

void ResDialog::DestroyUI()
{
	DeleteObj(Ui);
}

void ResDialog::DrawSelection(GSurface *pDC)
{
	// Draw selection
	for (ResDialogCtrl *Ctrl = Selection.First(); Ctrl; Ctrl = Selection.Next())
	{
		GRegion r = Ctrl->GetPos();
		for (GView *p = Ctrl->GetParent();
			p AND p != this; p = p->GetParent())
		{
			GRegion Wp = p->GetPos();
			r.Offset(Wp.x1, Wp.y1);
		}

		DrawGoobers(pDC, r, Ctrl->Goobers);
	}
}

void ResDialog::_Paint(GSurface *pDC)
{
	COLOUR Back = LC_MED;
	GRegion c(0, 0, X()-1, Y()-1);

	ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(Children.First());
	if (Ctrl)
	{
		c.ZOff(	Ctrl->X() + (2 * GOOBER_BORDER),
				Ctrl->Y() + (2 * GOOBER_BORDER));
	}

	GSurface *pMemDC = NEW(GMemDC);
	if (pMemDC AND pMemDC->Create(c.X(), c.Y(), GdcD->GetBits()))
	{
		// Draw client
		pMemDC->Colour(LC_WORKSPACE, 24);
		pMemDC->Rectangle();

		// Paint children
		GView::_Paint(pMemDC);

		if (Parent AND Parent->Focus())
		{
			// Draw selection
			DrawSelection(pMemDC);
		}

		// Put on screen
		pDC->Blt(0, 0, pMemDC);

		// Draw other non Mem-DC regions
		pDC->Colour(LC_WORKSPACE, 24);
		if (X() > c.X())
		{
			pDC->Rectangle(c.x2 + 1, 0, X()-1, c.y2);
		}
		if (Y() > c.Y())
		{
			pDC->Rectangle(0, c.y2 + 1, X()-1, Y()-1);
		}
	}
	else
	{
		// error
		pDC->Colour(Rgb24(255, 0, 255), 24);
		pDC->Rectangle();
	}

	DeleteObj(pMemDC);
}

void ResDialog::OnLanguageChange()
{
	if (Ui AND Ui->StatusInfo)
	{
		LangInfo *l = Symbols->GetLanguage(Symbols->GetCurrent());
		if (l)
		{
			char Str[256];
			sprintf(Str, "Current Language: %s (Id: %i)", l->Name, l->Id);
			Ui->StatusInfo->Name(Str);
		}
	}
}

void ResDialog::OnKey(GKey &k)
{
	if (k.Down AND k.Ctrl())
	{
		switch (k.c)
		{
			case VK_UP:
			{
				if (Symbols->GetCurrent() > 0)
				{
					LangInfo *l = Symbols->GetLanguage(Symbols->GetCurrent() - 1);
					if (l)
					{
						Symbols->SetCurrent(l->Id);
					}
				}
				Invalidate();
				OnLanguageChange();
				break;
			}
			case VK_DOWN:
			{
				if (Symbols->GetCurrent() < Symbols->GetLanguages() - 1)
				{
					LangInfo *l = Symbols->GetLanguage(Symbols->GetCurrent() + 1);
					if (l)
					{
						Symbols->SetCurrent(l->Id);
					}
				}
				Invalidate();
				OnLanguageChange();
				break;
			}
		}
	}
}

void ResDialog::OnMouseClick(GMouse &m)
{
	if (m.Down())
	{	
		if (Parent) Parent->Focus(true);

		if (m.Left())
		{
			DragGoober = -1;
			for (ResDialogCtrl *c = Selection.First(); c; c = Selection.Next())
			{
				for (int i=0; i<8; i++)
				{
					if (c->Goobers[i].Overlap(m.x, m.y))
					{
						DragGoober = i;
						DragCtrl = c;
						break;
					}
				}
			}

			if (DragGoober >= 0)
			{
				DragRgn = DragCtrl->GetPos();
				DragX = DragY = 0;
				int Cap = GooberCaps[DragGoober];
				
				// Lock the dialog to the top left corner
				if (dynamic_cast<CtrlDlg*>(DragCtrl))
				{
					Cap &= ~(RESIZE_X1|RESIZE_Y1);
				}
				
				if (NOT Cap)
				{
					// Can't move the object anyway so abort the drag
					DragGoober = -1;
				}
				else
				{
					// Allow movement along the appropriate edge
					if (Cap & RESIZE_X1)
					{
						DragX = &DragRgn.x1;
					}
					if (Cap & RESIZE_Y1)
					{
						DragY = &DragRgn.y1;
					}
					if (Cap & RESIZE_X2)
					{
						DragX = &DragRgn.x2;
					}
					if (Cap & RESIZE_Y2)
					{
						DragY = &DragRgn.y2;
					}

					// Remember the offset to the mouse from the egdes
					if (DragX) DragOx = m.x - *DragX;
					if (DragY) DragOy = m.y - *DragY;

					Capture(true);
				}
			}
		}
	}
	else
	{
		if (DragGoober >= 0)
		{
			Capture(false);
			DragGoober = -1;
			DragX = 0;
			DragY = 0;
		}
	}
}

void ResDialog::OnMouseMove(GMouse &m)
{
	if (DragGoober >= 0)
	{
		if (DragX)
		{
			*DragX = m.x - DragOx;
		}
		if (DragY)
		{
			*DragY = m.y - DragOy;
		}

		// DragRgn in real coords
		GRegion Old = DragCtrl->GetPos();
		GRegion New = DragRgn;
		
		if (DragCtrl != Children.First())
		{
			SnapRect(&New, DragCtrl->ParentCtrl());
		}

		// New now in snapped coords
		
		// If that means the dragging control changes then
		if (New != Old)
		{
			// change everyone else by the same amount
			for (ResDialogCtrl *c = Selection.First(); c; c = Selection.Next())
			{
				GRegion OldPos = c->GetPos();
				GRegion NewPos = OldPos;

				NewPos.x1 += New.x1 - Old.x1;
				NewPos.y1 += New.y1 - Old.y1;
				NewPos.x2 += New.x2 - Old.x2;
				NewPos.y2 += New.y2 - Old.y2;

				c->SetPos(NewPos);
				OldPos.Union(&NewPos);
				OldPos.Size(-GOOBER_BORDER, -GOOBER_BORDER);
				c->GetParent()->Invalidate(&OldPos);
			}
		}
	}
}

bool ResDialog::Read(List<XmlTag> &Tags)
{
	bool Status = false;

	// turn symbol creation off so that ctrls find their
	// symbol via Id matching instead of creating their own
	CreateSymbols = false;

	ResDialogCtrl *Dlg = dynamic_cast<ResDialogCtrl*>(Children.First());
	if (Dlg)
	{
		Symbols->DeleteStr(Dlg->Str);
		Dlg->Str = 0;

		// Load the resource
		Status = Res_Read(Dlg, Tags);
		Item->Update();
	}

	CreateSymbols = true;
	return Status;
}

void ResDialog::CleanSymbols()
{
	// removed unreferenced dialog strings
	if (Symbols)
	{
		Symbols->RemoveUnReferenced();
	}

	// re-ID duplicate entries
	ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(Children.First());
	if (Ctrl)
	{
		// list all the entries
		List<ResDialogCtrl> l;
		Ctrl->ListChildren(l);
		l.Insert(Ctrl); // insert the dialog too

		// remove duplicate string entries
		for (ResDialogCtrl *c = l.First(); c; c = l.Next())
		{
			ASSERT(c->Str);
			c->Str->UnDupelicate();
		}
	}

	// sort list (cause I need to read the file myself)
	if (Symbols)
	{
		Symbols->Sort();
	}
}

bool ResDialog::Write(GFile &f)
{
	bool Status = false;
	ResDialogCtrl *Ctrl = dynamic_cast<ResDialogCtrl*>(Children.First());
	if (Ctrl)
	{
		// duplicates symbols should have been removed before the 
		// strings were written out
		// so we don't have to do it here.

		// write the resources out
		if (Ctrl)
		{
			Status = Res_Write(Ctrl, f);
		}
	}

	return Status;
}

void ResDialog::OnRightClick(GSubMenu *RClick)
{
	if (RClick)
	{
		if (Enabled())
		{
			RClick->AppendSeparator();
			if (Type() > 0)
			{
				RClick->AppendItem("Dump to C++", IDM_DUMP, true);

				GSubMenu *Export = RClick->AppendSub("Export to...");
				if (Export)
				{
					Export->AppendItem("Lgi File", IDM_EXPORT, true);
					Export->AppendItem("Win32 Resource Script", IDM_EXPORT_WIN32, false);
				}
			}
		}
	}
}

char *TypeOfRes(ResDialogCtrl *Ctrl)
{

	// the default
	return "GWindow";
}

class ControlVar
{
public:
	char Type[32];
	char VarName[32];
	ResDialogCtrl *Ctrl;
};

char *TextOfCtrl(ResDialogCtrl *Ctrl)
{
	static char Buf[256];

	switch (Ctrl->GetType())
	{
		// Has text
		case UI_TEXT:
		case UI_EDITBOX:
		case UI_CHECKBOX:
		case UI_BUTTON:
		case UI_GROUP:
		case UI_RADIO:
		{
			char *s = Ctrl->Str->Get();
			sprintf(Buf, ", \"%s\"", s?s:"");
			return Buf;
		}

		// Not processed
		case UI_COLUMN:
		case UI_TAB:
		{
			LgiAssert(0);
			break;
		}

		// No text...
		case UI_BITMAP:
		case UI_PROGRESS:
		case UI_TABS:
		case UI_LIST:
		case UI_COMBO:
		case UI_TREE:
		case UI_SCROLL_BAR:
		{
			break;
		}
	}

	return "";
}

void ResDialog::OnCommand(int Cmd)
{
	switch (Cmd)
	{
		case IDM_DUMP:
		{
			StringPipe Buf;
			char Str[256];

			int i=0;
			List<ControlVar> Ctrls;
			ResDialogCtrl *Dlg = dynamic_cast<ResDialogCtrl*>(Children.First());
			if (Dlg)
			{
				for (ResDialogCtrl *c=dynamic_cast<ResDialogCtrl*>(Dlg->Children.First()); c; c=dynamic_cast<ResDialogCtrl*>(Dlg->Children.Next()))
				{
					ControlVar *cd = NEW(ControlVar);
					if (cd)
					{
						cd->Ctrl = c;
						sprintf(cd->VarName, "Ctrl%i", i++);
						strcpy(cd->Type, "GView"); // default type

						for (LgiObjectName *on=NameMap; on->Type; on++)
						{
							if (c->GetType() == on->Type)
							{
								strcpy(cd->Type, on->ObjectName);
								break;
							}
						}

						Ctrls.Insert(cd);
					}
				}

				// class def
				Buf.Push(	"class Dlg : public GDialog\n"
							"{\n");

				ControlVar *cd;
				for (cd=Ctrls.First(); cd; cd=Ctrls.Next())
				{
					sprintf(Str, "\t%s *%s;\n", cd->Type, cd->VarName);
					Buf.Push(Str);
				}

				Buf.Push(	"\n"
							"public:\n"
							"\tDlg(GWindow *Parent);\n"
							"\t~Dlg();\n"
							"\n"
							"\tint OnNotify(GView *Ctrl, int Flags);\n"
							"};\n"
							"\n");

				// class impl
				Buf.Push(	"Dlg::Dlg(GWindow *Parent)\n"
							"{\n"
							"\tSetParent(Parent);\n");

				sprintf(Str, "\tName(\"%s\");\n\tPos.ZOff(%i, %i);\n", Dlg->Str->Get(), Dlg->X(), Dlg->Y());
				Buf.Push(Str);
				Buf.Push("\tMoveToCenter();\n\n");

				for (cd=Ctrls.First(); cd; cd=Ctrls.Next())
				{
					sprintf(Str,
							"\tChildren.Insert(%s = NEW(%s(%s, %i, %i, %i, %i%s)));\n",
							cd->VarName,
							cd->Type,
							cd->Ctrl->Str->Define,
							cd->Ctrl->GetPos().x1,
							cd->Ctrl->GetPos().y1,
							cd->Ctrl->X(),
							cd->Ctrl->Y(),
							TextOfCtrl(cd->Ctrl));
					Buf.Push(Str);
				}

				Buf.Push(	"}\n"
							"\n"
							"int Dlg::OnNotify(GView *Ctrl, int Flags)\n"
							"{\n"
							"\tswitch (Ctrl->GetId())\n"
							"\t{\n"
							"\t\tcase IDOK:\n"
							"\t\t{\n"
							"\t\t\t// Do something here\n"
							"\t\t\t// fall thru\n"
							"\t\t}\n"
							"\t\tcase IDCANCEL:\n"
							"\t\t{\n"
							"\t\t\tEndModal(Ctrl->GetId());\n"
							"\t\t\tbreak;\n"
							"\t\t}\n"
							"\t}\n"
							"\n"
							"\treturn 0;\n"
							"}\n");

				// Output to clipboard
				char *Text = Buf.NewStr();
				if (Text)
				{
					GClipBoard Clip(Ui);
					Clip.Text(Text);
					DeleteArray(Text);
				}
			}
			break;
		}
		case IDM_EXPORT:
		{
			GFileSelect Select;
			Select.Parent(AppWindow);
			Select.Type("Text", "*.txt");
			if (Select.Save())
			{
				GFile F;
				if (F.Open(Select.Name(), O_WRITE))
				{
					F.SetSize(0);
					// Serialize(F, true);
				}
				else
				{
					LgiMsg(AppWindow, "Couldn't open file for writing.");
				}
			}
			break;
		}
		case IDM_EXPORT_WIN32:
		{
			break;
		}
	}
}

int ResDialog::OnCommand(int Cmd, int Event, OsView hWnd)
{
	switch (Cmd)
	{
		case IDM_SET_LANG:
		{
			Symbols->SetCurrent();
			OnSelect(Selection.First());

			Invalidate();
			OnLanguageChange();
			break;
		}
		case IDM_TAB_ORDER:
		{
			TabOrder Dlg(this);
			break;
		}
	}

	return 0;
}

ResString *ResDialog::CreateSymbol()
{
	return (Symbols) ? Symbols->CreateStr() : 0;
}

////////////////////////////////////////////////////////////////////
ResDialogUi::ResDialogUi(ResDialog *Res)
{
	Dialog = Res;
	Tools = 0;
	Status = 0;
	StatusInfo = 0;

	if (Res)
	{
		Res->OnSelect(Res->Selection.First());
	}
}

ResDialogUi::~ResDialogUi()
{
}

void ResDialogUi::OnPaint(GSurface *pDC)
{
	GComplexRegion Client(0, 0, X()-1, Y()-1);
	for (GView *w = Children.First(); w; w = Children.Next())
	{
		GRegion r = w->GetPos();
		Client.Subtract(&r);
	}

	pDC->Colour(LC_MED, 24);
	for (GRegion *r = Client.First(); r; r = Client.Next())
	{
		pDC->Rectangle(r);
	}
}

void ResDialogUi::Pour()
{
	RECT rcClient;

	if (NOT GetClientRect(Handle(), &rcClient))
	{
		return;
	}

	GComplexRegion Client(rcClient);
	GComplexRegion Update;

	for (GView *v = Children.First(); v; v = Children.Next())
	{
		GRegion OldPos = v->GetPos();
		Update.Union(&OldPos);

		if (v->Pour(Client))
		{
			if (NOT v->Visible())
			{
				v->Visible(TRUE);
			}

			if (OldPos != v->GetPos())
			{
				// position has changed update...
				v->Invalidate();
			}

			Client.Subtract(&v->GetPos());
			Update.Subtract(&v->GetPos());
		}
		else
		{
			// make the view not visible
			v->Visible(FALSE);
		}
	}

	for (int i=0; i<Update.GetSize(); i++)
	{
		Invalidate(Update[i]);
	}
}

void ResDialogUi::OnPosChange()
{
	Pour();
}

void ResDialogUi::OnCreate()
{
	Tools = NEW(GToolBar);
	if (Tools)
	{
		char *FileName = LgiFindFile("DialogIcons.bmp");
		if (FileName AND Tools->SetBitmap(FileName, 16, 16))
		{
			Tools->Attach(this);

			Tools->AppendButton("Cursor", 0, TBT_RADIO);
			Tools->AppendButton("Text", 0, TBT_RADIO);
			Tools->AppendButton("Edit", 0, TBT_RADIO);
			Tools->AppendButton("Checkbox", 0, TBT_RADIO);
			Tools->AppendButton("Button", 0, TBT_RADIO);
			Tools->AppendButton("Group", 0, TBT_RADIO);
			Tools->AppendButton("Radio Button", 0, TBT_RADIO);
			Tools->AppendButton("Tabs", 0, TBT_RADIO);
			Tools->AppendButton("ListView", 0, TBT_RADIO);
			Tools->AppendButton("Combobox", 0, TBT_RADIO);
			Tools->AppendButton("Tree", 0, TBT_RADIO);
			Tools->AppendButton("Bitmap", 0, TBT_RADIO);
			Tools->AppendButton("Progress", 0, TBT_RADIO);
			Tools->AppendButton("Scroll Bar", 0, TBT_RADIO);
			Tools->AppendSeparator();

			Tools->AppendButton("Change language", IDM_SET_LANG, TBT_PUSH);
			Tools->AppendButton("Tab Order", IDM_TAB_ORDER, TBT_PUSH);
		}
		else
		{
			DeleteObj(Tools);
		}
		DeleteArray(FileName);
	}

	Status = NEW(GStatusBar);
	if (Status)
	{
		Status->Attach(this);
		StatusInfo = Status->AppendPane("", 2);
	}

	ResFrame *Frame = NEW(ResFrame(Dialog));
	if (Frame)
	{
		Frame->Attach(this);
	}
}

int ResDialogUi::CurrentTool()
{
	if (Tools)
	{
		int i=0;
		for (GView *w = Tools->Children.First(); w; w = Tools->Children.Next(), i++)
		{
			GToolButton *But = dynamic_cast<GToolButton*>(w);
			if (But)
			{
				if (But->Value())
				{
					return i;
				}
			}
		}
	}

	return -1;
}

void ResDialogUi::SelectTool(int i)
{
	if (Tools)
	{
		GView *w = Tools->Children.ItemAt(i);
		if (w)
		{
			GToolButton *But = dynamic_cast<GToolButton*>(w);
			if (But)
			{
				But->Value(true);
			}
		}
	}
}

int ResDialogUi::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case WM_COMMAND:
		{
			Dialog->OnCommand(LOWORD(Msg->a), HIWORD(Msg->a), (HWND) Msg->b);
			break;
		}
		case WM_DESCRIBE:
		{
			char *Text = (char*) MsgB(Msg);
			if (Text)
			{
				StatusInfo->Name(Text);
			}
			break;
		}
	}

	return GView::OnEvent(Msg);
}
