/*
**	FILE:			LgiRes_Dialog.h
**	AUTHOR:			Matthew Allen
**	DATE:			5/8/1999
**	DESCRIPTION:	Dialog Resource Editor
**
**
**	Copyright (C) 1999, Matthew Allen
**		fret@ozemail.com.au
*/

#ifndef __LGIRES_DIALOG_H
#define __LGIRES_DIALOG_H

#include "GXml.h"
#include "Res.h"
#include "LgiRes_String.h"

class ResDialog;
class ResDialogUi;

#define UI_CURSOR				0
#define UI_TEXT					1
#define UI_EDITBOX				2
#define UI_CHECKBOX				3
#define UI_BUTTON				4
#define UI_GROUP				5
#define UI_RADIO				6
#define UI_TABS					7
#define	UI_LIST					8
#define UI_COMBO				9
#define UI_TREE					10
#define UI_BITMAP				11
#define UI_PROGRESS				12
#define UI_SCROLL_BAR			13
#define UI_TAB					14
#define UI_COLUMN				15
#define UI_DIALOG				100

#define GOOBER_SIZE				5
#define GOOBER_BORDER			7

#define GRID_X					7
#define GRID_Y					7

////////////////////////////////////////////////////////////////
class ResDialogCtrl : public GView, public FieldSource, public ResObject
{
	friend class ResDialog;

protected:
	static int TabDepth;
	void TabString(char *Str);
	void ReadPos(char *Str);
	char *GetRefText();

	GRegion Goobers[8];

	ResDialog *Dlg;
	GRegion Title;
	GRegion Client;
	int DragCtrl;
	GRegion DragRgn;
	GdcPt2 DragStart;
	bool MoveCtrl;
	int SelectMode;
	GRegion SelectStart;

	bool AcceptChildren;
	bool Movable;

public:
	ResString *Str;
	
	ResDialogCtrl(ResDialog *dlg, char *CtrlTypeName);
	~ResDialogCtrl();

	ResDialogCtrl *ParentCtrl() { return dynamic_cast<ResDialogCtrl*>(GetParent()); }

	bool IsContainer() { return AcceptChildren; }
	void OnPaint(GSurface *pDC);
	void OnMouseClick(GMouse &m);
	void OnMouseMove(GMouse &m);
	bool SetPos(GRegion &p, bool Repaint = FALSE);
	void StrFromRef(int Id);

	bool GetFields(DataDlgFieldList &Fields, ObjProperties *Props = 0);
	void OnFieldChange();

	virtual int GetType() = 0;
	virtual GRegion GetMinSize();
	virtual bool Serialize(bool Write);
	virtual void ListChildren(List<ResDialogCtrl> &l, bool Deep = true);
	virtual bool AttachCtrl(ResDialogCtrl *Ctrl, GRegion *r = 0);
};

#define RESIZE_X1			0x0001
#define RESIZE_Y1			0x0002
#define RESIZE_X2			0x0004
#define RESIZE_Y2			0x0008

class ResDialog : public Resource, public GView, public ResFactory
{
	friend class ResDialogCtrl;
	friend class ResDialogUi;
	friend class CtrlDlg;

protected:
	static int SymbolRefs;
	static ResStringGroup *Symbols;
	static bool CreateSymbols;

	ResDialogUi *Ui;
	List<ResDialogCtrl> Selection;
	GRegion DlgPos;
	
	int DragGoober;
	int *DragX, *DragY;
	int DragOx, DragOy;
	GRegion DragRgn;
	ResDialogCtrl *DragCtrl;

	void DrawSelection(GSurface *pDC);

public:
	ResDialog(AppWnd *w, int type = TYPE_DIALOG);
	~ResDialog();
	void Create();

	GView *Wnd() { return dynamic_cast<GView*>(this); }
	static void AddLanguage(int Id);

	// GObject overrides
	char *Name();
	bool Name(char *n);

	// Factory
	char *StringFromRef(int Ref);
	ResObject *CreateObject(char *Name, ResObject *Parent);

	int Res_GetStrRef(ResObject *Obj);
	void Res_SetStrRef(ResObject *Obj, int Id);
	void Res_SetPos(ResObject *Obj, int x1, int y1, int x2, int y2);
	void Res_SetPos(ResObject *Obj, char *s);
	GRegion Res_GetPos(ResObject *Obj);
	void Res_Attach(ResObject *Obj, ResObject *Parent);
	bool Res_GetChildren(ResObject *Obj, List<ResObject> *l, bool Deep);
	void Res_Append(ResObject *Obj, ResObject *Parent);
	bool Res_GetItems(ResObject *Obj, List<ResObject> *l);
	bool Res_GetProperties(ResObject *Obj, ObjProperties *Props);
	bool Res_SetProperties(ResObject *Obj, ObjProperties *Props);

	// Implementation
	int CurrentTool();
	ResDialogCtrl *CreateCtrl(char *Tool);
	ResDialogCtrl *CreateCtrl(int Tool);
	bool IsSelected(ResDialogCtrl *Ctrl);
	bool IsDraging();
	void SnapPoint(GdcPt2 *p, ResDialogCtrl *From);
	void SnapRect(GRegion *r, ResDialogCtrl *From);
	void MoveSelection(int Dx, int Dy);
	void SelectRect(ResDialogCtrl *Parent, GRegion *r, bool ClearPrev = true);
	void CleanSymbols();
	ResString *CreateSymbol();

	void Delete();
	void Copy(bool Delete = false);
	void Paste();

	GView *CreateUI();
	void DestroyUI();
	void OnMouseClick(GMouse &m);
	void OnMouseMove(GMouse &m);
	void OnKey(GKey &k);
	void OnSelect(ResDialogCtrl *Wnd, bool ClearPrev = true);
	void OnDeselect(ResDialogCtrl *Wnd);
	void OnRightClick(GSubMenu *RClick);
	void OnCommand(int Cmd);
	int OnCommand(int Cmd, int Event, OsView hWnd);
	void OnLanguageChange();

	void _Paint(GSurface *pDC);

	bool Read(List<XmlTag> &Tags);
	bool Write(GFile &f);
};

class ResDialogUi : public GLayout
{
	friend class ResDialog;

	GToolBar *Tools;
	ResDialog *Dialog;
	GStatusBar *Status;
	GStatusPane *StatusInfo;

public:
	ResDialogUi(ResDialog *Res);
	~ResDialogUi();

	void OnPaint(GSurface *pDC);
	void Pour();
	void OnPosChange();
	void OnCreate();
	int OnEvent(GMessage *Msg);

	int CurrentTool();
	void SelectTool(int i);
};

class ResDialogFrame : public GLayout
{
	ResDialog *Dialog;

public:
	ResDialogFrame(ResDialog *d);
	~ResDialogFrame();

	bool Pour(GComplexRegion &r);
	void OnKey(GKey &k);
	void OnPaint(GSurface *pDC);
};

///////////////////////////////////////////////////////////////////////
class CtrlDlg : public ResDialogCtrl
{
public:
	CtrlDlg(ResDialog *dlg);

	int GetType() { return UI_DIALOG; }
	void OnPaint(GSurface *pDC);
};

class CtrlText : public ResDialogCtrl
{
public:
	CtrlText(ResDialog *dlg);

	int GetType() { return UI_TEXT; }
	void OnPaint(GSurface *pDC);
};

class CtrlEditbox : public ResDialogCtrl
{
	bool Password;
	bool MultiLine;

public:
	CtrlEditbox(ResDialog *dlg);

	int GetType() { return UI_EDITBOX; }
	void OnPaint(GSurface *pDC);

	bool GetFields(DataDlgFieldList &Fields, ObjProperties *Props = 0);
	bool Serialize(bool Write);
};

class CtrlCheckbox : public ResDialogCtrl
{
public:
	CtrlCheckbox(ResDialog *dlg);

	int GetType() { return UI_CHECKBOX; }
	void OnPaint(GSurface *pDC);
};

class CtrlButton : public ResDialogCtrl
{
public:
	CtrlButton(ResDialog *dlg);

	int GetType() { return UI_BUTTON; }
	void OnPaint(GSurface *pDC);
};

class CtrlGroup : public ResDialogCtrl
{
public:
	CtrlGroup(ResDialog *dlg);

	int GetType() { return UI_GROUP; }
	void OnPaint(GSurface *pDC);
};

class CtrlRadio : public ResDialogCtrl
{
	GSurface *Bmp;

public:
	CtrlRadio(ResDialog *dlg);
	~CtrlRadio();

	int GetType() { return UI_RADIO; }
	void OnPaint(GSurface *pDC);
};

class CtrlTab : public ResDialogCtrl
{
public:
	GRegion r;

	CtrlTab(ResDialog *dlg);

	int GetType() { return UI_TAB; }
	void ListChildren(List<ResDialogCtrl> &l, bool Deep);
};

class CtrlTabs : public ResDialogCtrl
{
	friend class CtrlTab;

	int Current;

public:
	List<CtrlTab> Tabs;

	CtrlTabs(ResDialog *dlg);
	~CtrlTabs();

	int GetType() { return UI_TABS; }
	void ListChildren(List<ResDialogCtrl> &l, bool Deep);
	void Empty();
	void OnPaint(GSurface *pDC);
	void ToTab();
	void FromTab();
	void OnMouseClick(GMouse &m);

	GRegion GetMinSize();
};

class ListCol : public ResDialogCtrl
{
public:
	GRegion &r() { return Pos; }

	int GetType() { return UI_COLUMN; }
	ListCol(ResDialog *dlg, char *Str = 0, int Width = 50);
};

class CtrlList : public ResDialogCtrl
{
	int DragCol;

public:
	List<ListCol> Cols;

	CtrlList(ResDialog *dlg);
	~CtrlList();

	int GetType() { return UI_LIST; }
	void ListChildren(List<ResDialogCtrl> &l, bool Deep);
	void Empty();
	void OnMouseClick(GMouse &m);
	void OnMouseMove(GMouse &m);
	void OnPaint(GSurface *pDC);
};

class CtrlComboBox : public ResDialogCtrl
{
public:
	CtrlComboBox(ResDialog *dlg);

	int GetType() { return UI_COMBO; }
	void OnPaint(GSurface *pDC);
};

class CtrlScrollBar : public ResDialogCtrl
{
public:
	CtrlScrollBar(ResDialog *dlg);

	int GetType() { return UI_SCROLL_BAR; }
	void OnPaint(GSurface *pDC);
};

class CtrlTree : public ResDialogCtrl
{
public:
	CtrlTree(ResDialog *dlg);

	int GetType() { return UI_TREE; }
	void OnPaint(GSurface *pDC);
};

class CtrlBitmap : public ResDialogCtrl
{
public:
	CtrlBitmap(ResDialog *dlg);

	int GetType() { return UI_BITMAP; }
	void OnPaint(GSurface *pDC);
};

class CtrlProgress : public ResDialogCtrl
{
public:
	CtrlProgress(ResDialog *dlg);

	int GetType() { return UI_PROGRESS; }
	void OnPaint(GSurface *pDC);
};

////////////////////////////////////////////////////////////////
#endif
