/*
**	FILE:			LgiRes_Menu.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			6/3/2000
**	DESCRIPTION:	Menu Resource Editor
**
**
**	Copyright (C) 1999, Matthew Allen
**		fret@ozemail.com.au
*/

////////////////////////////////////////////////////////////////////
#include "LgiRes.h"
#include "LgiRes_Menu.h"

#define IDC_TREE				100

#define IDM_NEW_SUB				200
#define IDM_NEW_ITEM			201
#define IDM_DELETE_ITEM			202
#define IDM_SET_LANG			300

//////////////////////////////////////////////////////////////////
#define VAL_Separator			"Sep"
#define VAL_Enabled				"Enabled"

ResMenuItem::ResMenuItem(ResMenu *menu) :
	Str((menu)?menu->Group:0)
{
	Sep = false;
	Enabled = true;
	Menu = menu;

	if (Menu AND Menu->Group)
	{
		Str.Ref = Menu->Group->UniqueId();
	}
}

char *ResMenuItem::GetText(int i)
{
	if (Sep)
	{
		return "--------------";
	}

	char *s = Str.Get();
	return (s) ? s : "<null>";
}

bool ResMenuItem::GetFields(DataDlgFieldList &Fields, ObjProperties *Props)
{
	Str.GetFields(Fields);
	Fields.Insert(NEW(DataDlgField(DATA_BOOL, 300, VAL_Separator, "Separator")));
	Fields.Insert(NEW(DataDlgField(DATA_BOOL, 301, VAL_Enabled, "Enabled")));

	Serialize(true);
	return true;
}

void ResMenuItem::OnFieldChange()
{
	Serialize(false);
	Update();
}

bool ResMenuItem::Serialize(bool Write)
{
	Str.Serialize(Write, &Values);

	if (Write)
	{
		Values.Set(VAL_Separator, Sep);
		Values.Set(VAL_Enabled, Enabled);
	}
	else
	{
		int n = false;
		if (Values.Get(VAL_Separator, n)) Sep = n;
		if (Values.Get(VAL_Enabled, n)) Enabled = n;
	}

	return true;
}

bool ResMenuItem::Read(List<XmlTag> &Tags, ResMenuItem *Parent)
{
	bool Status = false;
	XmlTag *t = Tags.Current();
	if (t)
	{
		bool SubMenu = stricmp(t->Name, "submenu") == 0;
		bool MenuItem = stricmp(t->Name, "menuitem") == 0;

		if (SubMenu OR MenuItem)
		{
			// Read item
			char *n = 0;
			if (t->Get("sep", n))
			{
				Sep = atoi(n);
			}
			if (t->Get("enabled", n))
			{
				Enabled = atoi(n);
			}
			if (t->Get("ref", n) AND
				Menu AND
				Menu->Group)
			{
				ResString *s = Menu->Group->FindRef(atoi(n));
				if (s)
				{
					// Copy the data from the string resource
					// allocated by the string group into
					// our object, allocated as part of this
					// item
					Str = *s;

					// Delete the string group allocated version
					// leaving the group containing just our
					// copy of the string
					DeleteObj(s);
				}
			}

			// Attach item
			Status = MenuItem;
			if (Parent)
			{
				Parent->Insert(this);
			}
			else if (Menu)
			{
				Menu->Insert(this);
			}

			// Read sub items
			if (SubMenu)
			{
				for (t = Tags.Next(); t; t = Tags.Next())
				{
					if (stricmp(t->Name, "/submenu") == 0)
					{
						Status = true;
						break;
					}
					else
					{
						ResMenuItem *i = NEW(ResMenuItem(Menu));
						if (i AND NOT i->Read(Tags, this))
						{
							DeleteObj(i);
						}
					}
				}
			}
		}
	}

	return Status;
}

bool ResMenuItem::Write(GFile &f, int Tabs)
{
	char *s = NEW(char[256]);
	if (s)
	{
		memset(s, 0, 256);
		memset(s, 9, Tabs);

		bool SubMenu = GetChild() != 0;
		char *PostTabs = s+strlen(s);
		if (Sep)
		{
			sprintf(PostTabs, "<menuitem sep=1>\n");
		}
		else
		{
			sprintf(PostTabs, "<%s sep=%i enabled=%i ref=%i>\n", (SubMenu)?"submenu":"menuitem", Sep, Enabled, Str.Ref);
		}

		f.Write(s, strlen(s));
		if (SubMenu)
		{
			for (ResMenuItem *i = dynamic_cast<ResMenuItem*>(GetChild()); i; i = dynamic_cast<ResMenuItem*>(i->GetNext()))
			{
				i->Write(f, Tabs+1);
			}

			strcpy(PostTabs, "</submenu>\n");
			f.Write(s, strlen(s));
		}

		DeleteArray(s);

		return true;
	}

	return false;
}

//////////////////////////////////////////////////////////////////
ResMenu::ResMenu(AppWnd *w, int type) :
	Resource(w, type),
	GTree(IDC_TREE, 0, 0, 200, 200, "Menu resource")
{
	Group = NEW(ResStringGroup(w));
}

ResMenu::~ResMenu()
{
}

void ResMenu::Create()
{
	Name("IDM_MENU");
	if (Resource::Item) Resource::Item->Update();
	Visible(true);
}

GView *ResMenu::CreateUI()
{
	return Ui = NEW(ResMenuUi(this));
}

void ResMenu::DestroyUI()
{
	DeleteObj(Ui);
}

void ResMenu::OnItemClick(GTreeItem *Item, GMouse &m)
{
	GTree::OnItemClick(Item, m);
}

void ResMenu::OnItemBeginDrag(GTreeItem *Item, int Flags)
{
	GTree::OnItemBeginDrag(Item, Flags);
}

void ResMenu::OnItemExpand(GTreeItem *Item, bool Expand)
{
	GTree::OnItemExpand(Item, Expand);
}

void ResMenu::OnItemSelect(GTreeItem *Item)
{
	GTree::OnItemSelect(Item);

	if (AppWindow)
	{
		ResMenuItem *Item = dynamic_cast<ResMenuItem*>(Selection());
		AppWindow->OnObjSelect(Item);
	}
}

void ResMenu::OnRightClick(GSubMenu *RClick)
{
}

void ResMenu::OnCommand(int Cmd)
{
}

int ResMenu::OnCommand(int Cmd, int Event, OsView hWnd)
{
	switch (Cmd)
	{
		case IDM_NEW_SUB:
		{
			ResMenuItem *Item = dynamic_cast<ResMenuItem*>(Selection());
			GTreeItem *New = 0;
			if (Item)
			{
				New = Item->Insert(NEW(ResMenuItem(this)));
			}
			else
			{
				New = Insert(NEW(ResMenuItem(this)));
			}

			if (New)
			{
				Select(New);
			}
			break;
		}
		case IDM_NEW_ITEM:
		{
			ResMenuItem *Item = dynamic_cast<ResMenuItem*>(Selection());
			if (Item)
			{
				int n = Item->IndexOf();
				if (n >= 0) n++;

				GTreeItem *Parent = Item->GetParent();
				GTreeItem *New = 0;
				if (Parent)
				{
					New = Parent->Insert(NEW(ResMenuItem(this)), n);
				}
				else
				{
					New = Insert(NEW(ResMenuItem(this)), n);
				}

				if (New)
				{
					Select(New);
				}
			}
			break;
		}
		case IDM_DELETE_ITEM:
		{
			ResMenuItem *Item = dynamic_cast<ResMenuItem*>(Selection());
			if (Item)
			{
				Remove(Item);
				DeleteObj(Item);
			}			
			break;
		}
		case IDM_SET_LANG:
		{
			if (Group)
			{
				Group->SetCurrent();
				Invalidate();
			}
			break;
		}
		case IDM_NEW_LANG:
		{
			if (Group)
			{
				Group->OnCommand(IDM_NEW_LANG, 0, 0);
				Invalidate();
			}
			break;
		}
		case IDM_DELETE_LANG:
		{
			if (Group)
			{
				Group->OnCommand(IDM_DELETE_LANG, 0, 0);
				Invalidate();
			}
			break;
		}
	}

	return 0;
}

bool ResMenu::Read(List<XmlTag> &Tags)
{
	bool Status = false;

	XmlTag *t = 0;
	for (t = Tags.Current(); t; t = Tags.Next())
	{
		if (stricmp(t->Name, "menu") == 0)
		{
			char *p = 0;
			if (t->Get("Name", p))
			{
				Name(p);
				t = Tags.Next();
				break;
			}
		}
	}

	for (; t; t = Tags.Next())
	{
		if (stricmp(t->Name, "/menu") == 0)
		{
			Tags.Next();
			Status = true;
			break;
		}
		else
		{
			if (stricmp(t->Name, "string-group") == 0)
			{
				if (Group)
				{
					Group->Read(Tags);
				}
				t = Tags.Current();
			}

			ResMenuItem *i = NEW(ResMenuItem(this));
			if (i AND i->Read(Tags))
			{
			}
			else
			{
				DeleteObj(i);
			}
		}
	}

	return Status;
}

bool ResMenu::Write(GFile &f)
{
	bool Status = false;

	char Str[256];
	
	// Write start tag
	sprintf(Str, "<menu name=\"%s\">\n", (Name())?Name():"");
	f.Write(Str, strlen(Str));

	// Write group
	if (Group)
	{
		Group->Write(f);
	}

	// Write submenus/items
	for (GTreeItem *i = Items.First(); i; i = Items.Next())
	{
		ResMenuItem *m = dynamic_cast<ResMenuItem*>(i);
		if (m)
		{
			m->Write(f, 1);
		}
	}

	// Write end tag
	strcpy(Str, "</menu>\n");
	f.Write(Str, strlen(Str));

	return Status;
}

//////////////////////////////////////////////////////////////////////////
ResMenuUi::ResMenuUi(ResMenu *Res)
{
	Menu = Res;
	Tools = 0;
	Status = 0;
	StatusInfo = 0;
}

ResMenuUi::~ResMenuUi()
{
}

void ResMenuUi::OnPaint(GSurface *pDC)
{
	GComplexRegion Client(0, 0, X()-1, Y()-1);
	for (GView *w = Children.First(); w; w = Children.Next())
	{
		GRegion r = w->GetPos();
		Client.Subtract(&r);
	}

	pDC->Colour(LC_MED, 24);
	for (GRegion *r = Client.First(); r; r = Client.Next())
	{
		pDC->Rectangle(r);
	}
}

void ResMenuUi::Pour()
{
	RECT rcClient;
	if (NOT GetClientRect(Handle(), &rcClient)) return;

	GComplexRegion Client(rcClient);
	GComplexRegion Update;

	for (GView *v = Children.First(); v; v = Children.Next())
	{
		GRegion OldPos = v->GetPos();
		Update.Union(&OldPos);

		if (v->Pour(Client))
		{
			if (NOT v->Visible())
			{
				v->Visible(TRUE);
			}

			if (OldPos != v->GetPos())
			{
				// position has changed update...
				v->Invalidate();
			}

			Client.Subtract(&v->GetPos());
			Update.Subtract(&v->GetPos());
		}
		else
		{
			// make the view not visible
			v->Visible(FALSE);
		}
	}

	for (int i=0; i<Update.GetSize(); i++)
	{
		Invalidate(Update[i]);
	}
}

void ResMenuUi::OnPosChange()
{
	Pour();
}

void ResMenuUi::OnCreate()
{
	Tools = NEW(GToolBar);
	if (Tools)
	{
		char *FileName = LgiFindFile("MenuIcons.bmp");
		if (FileName AND Tools->SetBitmap(FileName, 16, 16))
		{
			Tools->Attach(this);

			Tools->AppendButton("New child menu", IDM_NEW_SUB);
			Tools->AppendButton("New menu item", IDM_NEW_ITEM);
			Tools->AppendButton("Delete menu item", IDM_DELETE_ITEM);
			Tools->AppendSeparator();

			Tools->AppendButton("Change language", IDM_SET_LANG, TBT_PUSH);
			Tools->AppendButton("New language", IDM_NEW_LANG, TBT_PUSH);
			Tools->AppendButton("Delete language", IDM_DELETE_LANG, TBT_PUSH);
		}
		else
		{
			DeleteObj(Tools);
		}
		DeleteArray(FileName);
	}

	Status = NEW(GStatusBar);
	if (Status)
	{
		Status->Attach(this);
		StatusInfo = Status->AppendPane("", 2);
	}

	ResFrame *Frame = NEW(ResFrame(Menu));
	if (Frame)
	{
		Frame->Attach(this);
	}
}

int ResMenuUi::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case WM_COMMAND:
		{
			Menu->OnCommand(LOWORD(Msg->a), HIWORD(Msg->a), (HWND) Msg->b);
			break;
		}
		case WM_DESCRIBE:
		{
			char *Text = (char*) MsgB(Msg);
			if (Text)
			{
				StatusInfo->Name(Text);
			}
			break;
		}
	}

	return GView::OnEvent(Msg);
}

int ResMenuUi::CurrentTool()
{
	return false;
}

void ResMenuUi::SelectTool(int i)
{
}


