/*
**	FILE:			LgiRes_Menu.h
**	AUTHOR:			Matthew Allen
**	DATE:			6/3/2000
**	DESCRIPTION:	Menu Resource Editor
**
**
**	Copyright (C) 1999, Matthew Allen
**		fret@ozemail.com.au
*/

#ifndef __LGIRES_MENU_H
#define __LGIRES_MENU_H

#include "GXml.h"
#include "Res.h"
#include "LgiRes_String.h"

class ResMenu;
class ResMenuUi;

class ResMenuItem : public GTreeItem, public FieldSource
{
	ResMenu *Menu;
	bool Sep;
	bool Enabled;

public:
	ResString Str;

	ResMenuItem(ResMenu *menu);

	bool Separator() { return Sep; }
	void Separator(bool i) { Sep = i; }
	bool Enable() { return Enabled; }
	void Enable(bool i) { Enabled = i; }

	char *GetText(int i=0);
	bool GetFields(DataDlgFieldList &Fields, ObjProperties *Props = 0);
	void OnFieldChange();
	bool Serialize(bool Write);

	bool Read(List<XmlTag> &Tags, ResMenuItem *Parent = 0);
	bool Write(GFile &f, int Tabs);
};

class ResMenu : public Resource, public GTree
{
	friend class ResMenuItem;

protected:
	ResMenuUi *Ui;
	ResStringGroup *Group;

public:
	ResMenu(AppWnd *w, int type = TYPE_MENU);
	~ResMenu();

	void Create();
	GView *Wnd() { return dynamic_cast<GView*>(this); }
	void SetLanguages() { if (Group) Group->SetLanguages(); }
	List<ResString> *GetStrs() { return (Group)?Group->GetStrs():0; }

	// Tree
	void OnItemClick(GTreeItem *Item, GMouse &m);
	void OnItemBeginDrag(GTreeItem *Item, int Flags);
	void OnItemExpand(GTreeItem *Item, bool Expand);
	void OnItemSelect(GTreeItem *Item);

	// Resource
	GView *CreateUI();
	void DestroyUI();
	void OnRightClick(GSubMenu *RClick);
	void OnCommand(int Cmd);
	int OnCommand(int Cmd, int Event, OsView hWnd);

	// Serialize
	bool Read(List<XmlTag> &Tags);
	bool Write(GFile &f);
};

class ResMenuUi : public GLayout
{
	GToolBar *Tools;
	ResMenu *Menu;
	GStatusBar *Status;
	GStatusPane *StatusInfo;

public:
	ResMenuUi(ResMenu *Res);
	~ResMenuUi();

	void OnPaint(GSurface *pDC);
	void Pour();
	void OnPosChange();
	void OnCreate();
	int OnEvent(GMessage *Msg);

	int CurrentTool();
	void SelectTool(int i);
};

#endif