/*
**	FILE:			LgiRes_String.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			10/9/1999
**	DESCRIPTION:	String Resource Editor
**
**
**	Copyright (C) 1999, Matthew Allen
**		fret@ozemail.com.au
*/

////////////////////////////////////////////////////////////////////
#include "LgiRes.h"
#include "LgiRes_String.h"

////////////////////////////////////////////////////////////////////////////
LangDlg::LangDlg(GView *parent, List<LangInfo> &l, int Init)
{
	Lang = 0;
	Parent = (parent) ? parent : MainWnd;
	Pos.ZOff(260, 90);
	MoveToCenter();
	Name("Language");

	Children.Insert(NEW(GText(-1, 10, 10, -1, -1, "Select language:")));
	Children.Insert(Sel = NEW(GCombo(-1, 20, 30, 150, 200, "")));
	if (Sel)
	{
		for (LangInfo *li = l.First(); li; li = l.Next())
		{
			Langs.Insert(li);
			Sel->Insert(li->Name);
		}

		if (Init >= 0)
		{
			Sel->Index(Init);
		}
	}

	Children.Insert(NEW(GButton(IDOK, 180, 5, 60, 20, "Ok")));
	Children.Insert(NEW(GButton(IDCANCEL, 180, 30, 60, 20, "Cancel")));
}

int LangDlg::OnNotify(GView *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
		case IDOK:
		{
			Lang = Langs.ItemAt(Sel->Index());
			EndModal(1);
			break;
		}
		case IDCANCEL:
		{
			EndModal(0);
			break;
		}
	}

	return 0;
}	

////////////////////////////////////////////////////////////////////
StrLang::StrLang()
{
	Lang = LANG_ENGLISH;
	Str = 0;
}

StrLang::~StrLang()
{
	DeleteArray(Str);
}

////////////////////////////////////////////////////////////////////
ResString::ResString(ResStringGroup *grp)
{
	Ref = 0;
	Group = grp;
	Id = 0;
	Define = 0;
	UpdateWnd = 0;

	if (Group)
	{
		ASSERT(NOT Group->Strs.HasItem(this));

		Group->Strs.Insert(this);
		Group->GList::Insert(this);
	}
}

ResString::~ResString()
{
	if (Group)
	{
		if (NOT Group->Strs.Delete(this))
		{
			LgiAssert(0);
		}

		Group->GList::Remove(this);
	}

	for (StrLang *s = Items.First(); s; s = Items.Next())
	{
		DeleteObj(s);
	}

	DeleteArray(Define);
}

ResString::operator =(ResString &s)
{
	memcpy(RefStr, s.RefStr, sizeof(RefStr));
	memcpy(IdStr, s.IdStr, sizeof(IdStr));
	Ref = s.Ref;
	Id = s.Id;
	DeleteObj(Define);
	Define = NewStr(s.Define);
	
	for (StrLang *l = s.Items.First(); l; l = s.Items.Next())
	{
		Set(l->Str, l->Lang);
	}

	return 0;
}

char *ResString::Get(int Lang)
{
	if (Lang < 0) Lang = Group->Lang[Group->CurrentLang]->Id;
	StrLang *s = GetLang(Lang);
	if (s)
	{
		return s->Str;
	}

	return 0;
}

void ResString::Set(char *p, int Lang)
{
	if (Lang < 0) Lang = Group->Lang[Group->CurrentLang]->Id;
	StrLang *s = GetLang(Lang);
	if (NOT s)
	{
		Items.Insert(s = NEW(StrLang));
		if (s)
		{
			s->Lang = Lang;
		}
	}
	if (s)
	{
		DeleteArray(s->Str);
		if (p) s->Str = NewStr(p);

		Update();
		if (UpdateWnd)
		{
			UpdateWnd->Invalidate();
		}
	}
}

void ResString::UnDupelicate()
{
	// reset Ref if it's a duplicate
	int OldStrRef = Ref;
	Ref = 0x80000000;
	ResString *Dupe = Group->App()->GetStrFromRef(OldStrRef);
	// now we return the Id to the string
	Ref = OldStrRef;

	if (Dupe OR	Ref < 0)
	{
		// there is a duplicate string Id
		// so assign a new Id
		Ref = Group->App()->GetUniqueStrRef();
	}				
}

char **ResString::GetIndex(int i)
{
	StrLang *s = Items.ItemAt(i);
	return (s) ? &s->Str : 0;
}

bool ResString::GetFields(DataDlgFieldList &Fields, ObjProperties *Props)
{
	ObjProperties *p = (Props) ? Props : &Values;

	if (NOT Props)
	{
		// string view [literal]
		Fields.Insert(NEW(DataDlgField(DATA_INT, 199, VAL_Ref, "Ref")));
		Fields.Insert(NEW(DataDlgField(DATA_INT, 200, VAL_Id, "Id")));
	}
	// else dialog view [embeded]
	Fields.Insert(NEW(DataDlgField(DATA_STR, 201, VAL_Define, "#define")));

	if (Group)
	{
		if (Props)
		{
			Fields.Insert(NEW(DataDlgField(DATA_STR, 202, Group->Lang[Group->CurrentLang]->Name, Group->Lang[Group->CurrentLang]->Name)));
		}
		else
		{
			for (int i=0; i<Group->Langs; i++)
			{
				Fields.Insert(NEW(DataDlgField(DATA_STR, 202+i, Group->Lang[i]->Name, Group->Lang[i]->Name)));
			}
		}
	}

	// Copy data from object in Values
	Serialize(true, p);

	return true;
}

void ResString::OnFieldChange()
{
	Serialize(false);
}

bool ResString::Read(List<XmlTag> &Tags)
{
	XmlTag *t = Tags.Current();
	if (t AND stricmp(t->Name, "string") == 0)
	{
		char *n = 0;
		if (t->Get("Ref", n))
		{
			Ref = atoi(n);

			if (Ref == 199)
			{
				int h=0;
			}
		}
		if (t->Get("Id", n))
		{
			Id = atoi(n);
		}
		if (t->Get("Define", n))
		{
			DeleteArray(Define);
			Define = NewStr(n);
		}

		for (XmlValue *v = t->Values.First(); v; v = t->Values.Next())
		{
			if (v->Name AND strnicmp(v->Name, "Text(", 5) == 0)
			{
				char *Start = strchr(v->Name, '(');
				char *End = (Start) ? strchr(Start, ')') : 0;
				if (Start AND End)
				{
					*End = 0;
					Set(v->Value, atoi(Start+1));
				}
			}
		}

		return true;
	}
	return false;
}

bool ResString::Write(GFile &f)
{
	char Str[256];

	sprintf(Str, "<string Ref=%i Id=%i Define=\"%s\"", Ref, Id, (Define)?Define:"");
	f.Write(Str, strlen(Str));
	for (StrLang *s = Items.First(); s; s = Items.Next())
	{
		if (ValidStr(s->Str))
		{
			sprintf(Str, "\n\tText(%i)=\"%s\"", s->Lang, (s->Str)?s->Str:"");
			f.Write(Str, strlen(Str));
		}
	}
	f.Write(">\n", 2);

	return true;
}

StrLang *ResString::GetLang(int i)
{
	for (StrLang *s = Items.First(); s; s = Items.Next())
	{
		if (s->Lang == i)
		{
			return s;
		}
	}

	return 0;
}

bool ResString::Serialize(bool Write, ObjProperties *Props)
{
	ObjProperties *p = (Props) ? Props : &Values;

	if (Write)
	{
		p->Set(VAL_Ref, Ref);
		p->Set(VAL_Id, Id);
		p->Set(VAL_Define, Define);

		if (Group)
		{
			for (int i=0; i<Group->Langs; i++)
			{
				StrLang *s = GetLang(Group->Lang[i]->Id);
				if (s)
				{
					p->Set(Group->Lang[i]->Name, s->Str);
				}
			}
		}
	}
	else
	{
		p->Get(VAL_Ref, Ref);
		p->Get(VAL_Id, Id);
		char *n = 0;
		if (p->Get(VAL_Define, n))
		{
			DeleteArray(Define);
			Define = NewStr(n);
		}

		if (Group)
		{
			for (int i=0; i<Group->Langs; i++)
			{
				StrLang *s = GetLang(Group->Lang[i]->Id);
				if (NOT s)
				{
					s = NEW(StrLang);
					if (s)
					{
						Items.Insert(s);
						s->Lang = Group->Lang[i]->Id;
					}
				}
				if (s)
				{
					char *n = 0;
					if (p->Get(Group->Lang[i]->Name, n))
					{
						DeleteArray(s->Str);
						s->Str = NewStr(n);
					}
				}
			}
		}

		Update();
	}

	return true;
}

int ResString::GetCols()
{
	return 3 + Group->Langs;
}

char *ResString::GetText(int i)
{
	switch (i)
	{
		case 0:
		{
			return (Define)?Define:"";
		}
		case 1:
		{
			sprintf(RefStr, "%i", Ref);
			return RefStr;
		}
		case 2:
		{
			sprintf(IdStr, "%i", Id);
			return IdStr;
		}
		default:
		{
			if (i-3 < Group->Langs)
			{
				LangInfo *Info = (Group->Lang) ? Group->Lang[i-3] : 0;
				if (Info)
				{
					for (StrLang *s = Items.First(); s; s = Items.Next())
					{
						if (s->Lang == Info->Id)
						{
							return (s->Str) ? s->Str : "";
						}
					}
				}
			}
			break;
		}
	}

	return "";
}

////////////////////////////////////////////////////////////////////
ResStringGroup::ResStringGroup(AppWnd *w, int type) : 
	Resource(w, type),
	GList(90, 0, 0, 200, 200)
{
	Lang = 0;
	Ui = 0;
	Wnd()->Name("<untitled>");

	Sunken(false);
	Langs = 0;
	Lang = 0;
	CurrentLang = 0;
	SortAscend = true;
	SortCol = 0;
	AppendLanguage(LangIdEnglish);
}

ResStringGroup::~ResStringGroup()
{
	ResString *s;
	while (s=Strs.First())
	{
		DeleteObj(s);
	}

	DeleteObj(Ui);
	DeleteArray(Lang);
}

int ResString::Compare(ResString *r, int Column)
{
	bool Mod = (Column < 0) ? -1 : 1;
	Column = abs(Column) - 1;

	if (r)
	{
		switch (Column)
		{
			case 0:
			{
				return stricmp(Define, r->Define) * Mod;
				break;
			}
			case 1:
			{
				return (Ref - r->Ref) * Mod;
				break;
			}
			case 2:
			{
				return (Id - r->Id) * Mod;
				break;
			}
			default:
			{
				int Col = Column - 3;
				if (Group AND Col >= 0 AND Col < Group->Langs)
				{
					int Lang = Group->Lang[Column - 3]->Id;
					char *a = Get(Lang);
					char *b = r->Get(Lang);
					if (a AND b)
					{
						return stricmp(a, b) * Mod;
					}
				}
				break;
			}
		}
	}

	return -1;
}

int ResStringCompareFunc(GListItem *a, GListItem *b, int Data)
{
	ResString *A = dynamic_cast<ResString*>(a);
	if (A)
	{
		return A->Compare(dynamic_cast<ResString*>(b), Data);
	}
	return -1;
}

void ResStringGroup::OnColumnClick(int Col, GMouse &m)
{
	if (SortCol == Col)
	{
		SortAscend = NOT SortAscend;
	}
	else
	{
		SortCol = Col;
		SortAscend = true;
	}

	GList::Sort(ResStringCompareFunc, (SortCol + 1) * ((SortAscend) ? 1 : -1));
}

void ResStringGroup::OnItemClick(GListItem *Item, GMouse &m)
{
}

void ResStringGroup::OnItemSelect(GListItem *Item)
{
	ResString *s = dynamic_cast<ResString*>(Item);
	if (s AND AppWindow)
	{
		AppWindow->OnObjSelect(s);
	}
}

void ResStringGroup::UpdateColumns()
{
	EmptyColumns();
	AddColumn("#define", 150);
	AddColumn("Ref", 70);
	AddColumn("Id", 70);

	if (Lang)
	{
		for (int i=0; i<Langs; i++)
		{
			AddColumn(Lang[i]->Name, 100);
		}
	}
}

void ResStringGroup::DeleteLanguage(int Id)
{
	// Check for presence
	bool Present = false;
	for (int i=0; i<Langs AND NOT Present; i++)
	{
		Present |= Lang[i]->Id == Id;
	}

	if (Present)
	{
		// Remove lang
		int NewLangs = Langs - 1;
		LangInfo **NewLang = NEW(LangInfo*[NewLangs]);
		if (NewLang)
		{
			// Add the new language to this string group
			int n=0;
			for (int i=0; i<Langs; i++)
			{
				if (Lang[i]->Id != Id)
				{
					NewLang[n++] = Lang[i];
				}
			}

			DeleteArray(Lang);
			Lang = NewLang;
			Langs = NewLangs;
			UpdateColumns();
		}
	}
}

void ResStringGroup::AppendLanguage(int Id)
{
	// Check for presence
	bool Present = false;
	for (int i=0; i<Langs AND NOT Present; i++)
	{
		Present |= Lang[i]->Id == Id;
	}

	if (NOT Present)
	{
		LangInfo *n = LangFind(0, Id);
		if (n)
		{
			int NewLangs = Langs + 1;
			LangInfo **NewLang = NEW(LangInfo*[NewLangs]);
			if (NewLang)
			{
				// Add the new language to this string group
				memcpy(NewLang, Lang, sizeof(*NewLang)*Langs);
				NewLang[Langs] = n;
				DeleteArray(Lang);
				Lang = NewLang;
				Langs = NewLangs;
				UpdateColumns();
			}
		}
	}
}

void ResStringGroup::Sort()
{
	List<ResString> Temp;
	for (ResString *s = Strs.First(); s; s = Strs.Next())
	{
		Temp.Insert(s);
	}

	Strs.Empty();
	
	for (s = Temp.First(); s; s = Temp.Next())
	{
		int n=0;
		ResString *d = 0;
		for (d = Strs.First(); d AND s->Ref >= d->Ref; d = Strs.Next(), n++)
		{
		}

		Strs.Insert(s, n);
	}
}

void ResStringGroup::RemoveUnReferenced()
{
	for (ResString *s = Strs.First(); s; s = Strs.Next())
	{
		if (NOT s->UpdateWnd)
		{
			DeleteStr(s);
		}
	}
}

void ResStringGroup::SetCurrent(int Id)
{
	if (Id < 0)
	{
		// List all the languages being edited
		List<LangInfo> l;
		for (int i=0; i<Langs; i++)
		{
			l.Insert(Lang[i]);
		}

		// Display the list
		LangDlg Dlg(Ui, l, CurrentLang);
		if (Dlg.DoModal() AND Dlg.Lang)
		{
			if (Dlg.Lang)
			{
				Id = Dlg.Lang->Id;
			}
		}
	}

	for (int i=0; i<Langs; i++)
	{
		if (Id == Lang[i]->Id)
		{
			CurrentLang = i;
			break;
		}
	}
}

int ResStringGroup::OnCommand(int Cmd, int Event, OsView hWnd)
{
	switch (Cmd)
	{
		case IDM_NEW:
		{
			CreateStr(true);
			break;
		}
		case IDM_DELETE:
		{
			List<GListItem> l;
			if (GetSelection(l))
			{
				for (	ResString *s = dynamic_cast<ResString*>(l.First());
						s;
						s = dynamic_cast<ResString*>(l.Next()))
				{
					DeleteObj(s);
				}
			}
			break;
		}
		case IDM_NEW_LANG:
		{
			// List all the languages minus the ones already
			// being edited
			List<LangInfo> l;
			for (LangInfo *li = AllLanguages; li->Id; li++)
			{
				bool Present = false;
				for (int i=0; i<Langs; i++)
				{
					if (Lang[i]->Id == li->Id)
					{
						Present = true;
						break;
					}
				}

				if (NOT Present)
				{
					l.Insert(li);
				}
			}

			// Display the list
			LangDlg Dlg(this, l);
			if (Dlg.DoModal())
			{
				if (Dlg.Lang)
				{
					AppendLanguage(Dlg.Lang->Id);
				}
			}
			break;
		}
		case IDM_DELETE_LANG:
		{
			// List all the languages being edited
			List<LangInfo> l;
			for (int i=0; i<Langs; i++)
			{
				l.Insert(Lang[i]);
			}

			// Display the list
			LangDlg Dlg(this, l);
			if (Dlg.DoModal() AND Dlg.Lang)
			{
				if (Dlg.Lang)
				{
					DeleteLanguage(Dlg.Lang->Id);
					CurrentLang = limit(CurrentLang, 0, Langs-1);
				}
			}
			break;
		}
	}
	return 0;
}

ResString *ResStringGroup::CreateStr(bool User)
{
	ResString *s = NEW(ResString(this));
	if (s)
	{
		// Strs.Insert(s);
		// Insert(s);
		s->Id = s->Ref = App()->GetUniqueStrRef();

		char Str[256];
		sprintf(Str, "IDS_%i", s->Ref);
		s->Define = NewStr(Str);

		if (User)
		{
			s->GetList()->Select(NULL);
			s->ScrollTo();
			s->Select(true);
		}
	}
	return s;
}

void ResStringGroup::DeleteStr(ResString *Str)
{
	/*
	if (Str)
	{
		Strs.Delete(Str);
		GList::Delete(Str);
	}
	*/
	DeleteObj(Str);
}

ResString *ResStringGroup::FindName(char *Name)
{
	if (Name)
	{
		for (	ResString *s = Strs.First();
				s;
				s = Strs.Next())
		{
			if (stricmp(s->Define, Name) == 0)
			{
				return s;
			}
		}
	}
	return NULL;
}

ResString *ResStringGroup::FindRef(int Ref)
{
	for (	ResString *s = Strs.First();
			s;
			s = Strs.Next())
	{
		if (s->Ref == Ref)
		{
			return s;
		}
	}

	return NULL;
}

int ResStringGroup::UniqueId()
{
	int n = 1;
	for (ResString *i = dynamic_cast<ResString*>(Items.First()); i; i = dynamic_cast<ResString*>(Items.Next()))
	{
		n = max(n, i->Ref);
	}

	return n + 1;
}

int ResStringGroup::FindId(int Id, List<ResString*> &Strs)
{
	return 0;
}

void ResStringGroup::SetLanguages()
{
	List<LangInfo> l;
	bool EnglishFound = false;

	for (	ResString *s = Strs.First();
			s;
			s = Strs.Next())
	{
		for (StrLang *sl = s->Items.First(); sl; sl = s->Items.Next())
		{
			LangInfo *li = 0;
			for (li = l.First(); li AND li->Id != sl->Lang; li = l.Next());
			if (NOT li)
			{
				LangInfo *NewLang = LangFind(0, sl->Lang);
				if (NewLang) l.Insert(NewLang);
				if (sl->Lang == LangIdEnglish) EnglishFound = true;
			}
		}
	}

	CurrentLang = ((EnglishFound) ? 0 : 1);
	Langs = l.GetItems() + ((EnglishFound) ? 0 : 1);
	DeleteArray(Lang);
	Lang = NEW(LangInfo*[Langs]);
	if (Lang)
	{
		int n = 0;
		Lang[n++] = LangFind(0, LangIdEnglish);
		for (LangInfo *li = l.First(); li; li = l.Next())
		{
			if (li->Id != LangIdEnglish) // English
			{
				Lang[n++] = LangFind(0, li->Id);
			}
		}
	}

	UpdateColumns();
}

void ResStringGroup::DeleteSel()
{
}

void ResStringGroup::Copy(bool Delete)
{
}

void ResStringGroup::Paste()
{
}

GView *ResStringGroup::CreateUI()
{
	return Ui = NEW(ResStringUi(this));
}

void ResStringGroup::DestroyUI()
{
	DeleteObj(Ui);
}

bool ResStringGroup::Read(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *t = 0;
	for (t = Tags.Current(); t; t = Tags.Next())
	{
		if (stricmp(t->Name, "string-group") == 0)
		{
			char *p = 0;
			if (t->Get("Name", p))
			{
				Name(p);
				t = Tags.Next();
				Empty();
				break;
			}
		}
	}

	for (; t; t = Tags.Next())
	{
		if (stricmp(t->Name, "/string-group") == 0)
		{
			SetLanguages();
			Tags.Next();
			Status = true;
			break;
		}
		else
		{
			ResString *s = NEW(ResString(this));
			if (s)
			{
				s->Read(Tags);
			}
		}
	}

	if (Item)
	{
		Item->Update();
	}

	return Status;
}

bool ResStringGroup::Write(GFile &f)
{
	bool Status = false;

	char Str[256];
	sprintf(Str, "<string-group name=\"%s\">\n", (Name())?Name():"");
	f.Write(Str, strlen(Str));

	for (GListItem *i = Strs.First(); i; i = Strs.Next())
	{
		ResString *s = dynamic_cast<ResString*>(i);
		if (s AND (s->Define OR s->Items.GetItems() > 0))
		{
			Status |= s->Write(f);
		}
	}

	strcpy(Str, "</string-group>\n");
	f.Write(Str, strlen(Str));

	return Status;
}

void ResStringGroup::OnRightClick(GSubMenu *RClick)
{
	if (RClick)
	{
		if (Enabled())
		{
			RClick->AppendSeparator();
			if (Type() > 0)
			{
				GSubMenu *Export = RClick->AppendSub("Export to...");
				if (Export)
				{
					Export->AppendItem("Lgi File", IDM_EXPORT, true);
					Export->AppendItem("Win32 Resource Script", IDM_EXPORT_WIN32, false);
				}
			}
		}
	}
}

void ResStringGroup::OnCommand(int Cmd)
{
	switch (Cmd)
	{
		case IDM_IMPORT:
		{
			GFileSelect Select;
			Select.Parent(AppWindow);
			Select.Type("Text", "*.txt");
			if (Select.Open())
			{
				GFile F;
				if (F.Open(Select.Name(), O_READ))
				{
					Resource *Res = AppWindow->NewObject(-Type());
					if (Res)
					{
						// TODO
						// Res->Read();
					}
				}
				else
				{
					LgiMsg(AppWindow, "Couldn't open file for reading.");
				}
			}
			break;
		}
		case IDM_EXPORT:
		{
			GFileSelect Select;
			Select.Parent(AppWindow);
			Select.Type("Text", "*.txt");
			if (Select.Save())
			{
				GFile F;
				if (F.Open(Select.Name(), O_WRITE))
				{
					F.SetSize(0);
					// Serialize(F, true);
				}
				else
				{
					LgiMsg(AppWindow, "Couldn't open file for writing.");
				}
			}
			break;
		}
	}
}

////////////////////////////////////////////////////////////////////
ResStringUi::ResStringUi(ResStringGroup *Res)
{
	StringGrp = Res;
	Tools = 0;
	Status = 0;
	StatusInfo = 0;
}

ResStringUi::~ResStringUi()
{
}

void ResStringUi::OnPaint(GSurface *pDC)
{
	GComplexRegion Client(0, 0, X()-1, Y()-1);
	for (GView *w = Children.First(); w; w = Children.Next())
	{
		GRegion r = w->GetPos();
		Client.Subtract(&r);
	}

	pDC->Colour(LC_MED, 24);
	for (GRegion *r = Client.First(); r; r = Client.Next())
	{
		pDC->Rectangle(r);
	}
}

void ResStringUi::Pour()
{
	RECT rcClient;

	if (NOT GetClientRect(Handle(), &rcClient))
	{
		return;
	}

	GComplexRegion Client(rcClient);
	GComplexRegion Update;

	for (GView *v = Children.First(); v; v = Children.Next())
	{
		GRegion OldPos = v->GetPos();
		Update.Union(&OldPos);

		if (v->Pour(Client))
		{
			if (NOT v->Visible())
			{
				v->Visible(TRUE);
			}

			if (OldPos != v->GetPos())
			{
				// position has changed update...
				v->Invalidate();
			}

			Client.Subtract(&v->GetPos());
			Update.Subtract(&v->GetPos());
		}
		else
		{
			// make the view not visible
			v->Visible(FALSE);
		}
	}

	for (int i=0; i<Update.GetSize(); i++)
	{
		Invalidate(Update[i]);
	}
}

void ResStringUi::OnPosChange()
{
	Pour();
}

void ResStringUi::OnCreate()
{
	Tools = NEW(GToolBar);
	if (Tools)
	{
		char *FileName = LgiFindFile("StringIcons.bmp");
		if (FileName AND Tools->SetBitmap(FileName, 16, 16))
		{
			Tools->Attach(this);

			Tools->AppendButton("New String", IDM_NEW, TBT_PUSH, NOT StringGrp->SystemObject());
			Tools->AppendButton("Delete String", IDM_DELETE, TBT_PUSH, NOT StringGrp->SystemObject());
			Tools->AppendSeparator();

			Tools->AppendButton("New Language", IDM_NEW_LANG, TBT_PUSH);
			Tools->AppendButton("Delete Language", IDM_DELETE_LANG, TBT_PUSH);
		}
		else
		{
			DeleteObj(Tools);
		}
		DeleteArray(FileName);
	}

	Status = NEW(GStatusBar);
	if (Status)
	{
		Status->Attach(this);
		StatusInfo = Status->AppendPane("", 2);
	}

	ResFrame *Frame = NEW(ResFrame(StringGrp));
	if (Frame)
	{
		Frame->Attach(this);
	}
}

int ResStringUi::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case WM_COMMAND:
		{
			StringGrp->OnCommand(LOWORD(Msg->a), HIWORD(Msg->a), (HWND) Msg->b);
			break;
		}
		case WM_DESCRIBE:
		{
			char *Text = (char*) MsgB(Msg);
			if (Text)
			{
				StatusInfo->Name(Text);
			}
			break;
		}
	}

	return GView::OnEvent(Msg);
}
