/*
**	FILE:			LgiRes_String.h
**	AUTHOR:			Matthew Allen
**	DATE:			10/9/1999
**	DESCRIPTION:	Dialog Resource Editor
**
**
**	Copyright (C) 1999, Matthew Allen
**		fret@ozemail.com.au
*/

#ifndef __LGIRES_STRING_H
#define __LGIRES_STRING_H

#include "GXml.h"
#include "Res.h"

////////////////////////////////////////////////////////////////
class ResStringGroup;
class ResStringUi;

#define VAL_Ref						"Ref"
#define VAL_Id						"Id"
#define VAL_Define					"Define"
#define VAL_Text					"Text"

////////////////////////////////////////////////////////////////
class StrLang
{
public:
	int Lang;
	char *Str;

	StrLang();
	~StrLang();
};

class ResString : public GListItem, public FieldSource
{
	friend class ResStringGroup;
	friend class ResStringView;

protected:
	ResStringGroup *Group;

	char RefStr[16];
	char IdStr[16];
	char **GetIndex(int i);
	StrLang *GetLang(int i);

public:
	int Ref;			// globally unique
	int Id;				// numerical value used in code to reference
	char *Define;		// #define used in code to reference
	List<StrLang> Items;
	GView *UpdateWnd;

	ResString(ResStringGroup *group);
	~ResString();
	operator =(ResString &s);

	// Data
	char *Get(int Lang = -1);
	void Set(char *s, int Lang = -1);
	void UnDupelicate();
	int Compare(ResString *r, int Column);

	// Item
	int GetCols();
	char *GetText(int i);

	// Fields
	bool GetFields(DataDlgFieldList &Fields, ObjProperties *Props = 0);
	bool Serialize(bool Write, ObjProperties *Props = 0);
	void OnFieldChange();

	// Persistence
	bool Read(List<XmlTag> &Tags);
	bool Write(GFile &f);
};

#define RESIZE_X1			0x0001
#define RESIZE_Y1			0x0002
#define RESIZE_X2			0x0004
#define RESIZE_Y2			0x0008

class ResStringGroup : public Resource, public GList
{
	friend class ResString;
	friend class ResStringView;

protected:
	ResStringUi *Ui;
	List<ResString> Strs;
	int CurrentLang;
	int Langs;
	LangInfo **Lang;
	int SortCol;
	int SortAscend;

	void UpdateColumns();
	void OnColumnClick(int Col, GMouse &m);

public:
	ResStringGroup(AppWnd *w, int type = TYPE_STRING);
	~ResStringGroup();

	GView *Wnd() { return dynamic_cast<GView*>(this); }
	bool Attach(GView *Parent) { return GList::Attach(Parent); }

	// Methods
	ResString *CreateStr(bool User = false);
	void DeleteStr(ResString *Str);
	ResString *FindName(char *Name);
	ResString *FindRef(int Ref);
	int FindId(int Id, List<ResString*> &Strs);
	int UniqueId();
	int OnCommand(int Cmd, int Event, OsView hWnd);
	void SetLanguages();
	int GetLanguages() { return Langs; }
	LangInfo *GetLanguage(int i) { return (i>=0 AND i<Langs) ? Lang[i] : 0; }
	void AppendLanguage(int Id);
	void DeleteLanguage(int Id);
	void SetCurrent(int Id = -1);
	int GetCurrent() { return CurrentLang; }
	void Sort();
	void RemoveUnReferenced();
	List<ResString> *GetStrs() { return &Strs; }

	// Clipboard access
	void DeleteSel();
	void Copy(bool Delete = false);
	void Paste();

	// Resource
	GView *CreateUI();
	void DestroyUI();
	void OnRightClick(GSubMenu *RClick);
	void OnCommand(int Cmd);
	bool Read(List<XmlTag> &Tags);
	bool Write(GFile &f);

	// GList
	void OnItemClick(GListItem *Item, GMouse &m);
	void OnItemSelect(GListItem *Item);
	// void OnItemBeginDrag(GListItem *Item, int Flags);
	// void OnColumnClick(int Col, GMouse &m) {}
};

class ResStringUi : public GLayout
{
	GToolBar *Tools;
	ResStringGroup *StringGrp;
	GStatusBar *Status;
	GStatusPane *StatusInfo;

public:
	ResStringUi(ResStringGroup *Res);
	~ResStringUi();

	void OnPaint(GSurface *pDC);
	void Pour();
	void OnPosChange();
	void OnCreate();
	int OnEvent(GMessage *Msg);
};

class LangDlg : public GDialog
{
	GCombo *Sel;
	List<LangInfo> Langs;

public:
	LangInfo *Lang;

	LangDlg(GView *parent, List<LangInfo> &l, int Init = -1);
	int OnNotify(GView *Ctrl, int Flags);
};

////////////////////////////////////////////////////////////////
#endif
