/*hdr
**	FILE:		Gdc2.h
**	AUTHOR:		Matthew Allen
**	DATE:		20/2/97
**	DESCRIPTION:	GDC v2.xx header
**
**	Copyright (C) 1997, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include <math.h>
#include "Gdc2.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////
GdcViewDC::GdcViewDC(OsView view)
{
	View = view;
}

GdcViewDC::~GdcViewDC()
{
}

void GdcViewDC::GetOrigin(int &x, int &y)
{
	GDeviceContext::GetOrigin(x, y);
}

void GdcViewDC::SetOrigin(int x, int y)
{
	GDeviceContext::SetOrigin(x, y);
}

GRegion GdcViewDC::ClipRgn()
{
	return _Client;
}

void GdcViewDC::_SetClient(GRegion *c)
{
	_ClientClip = c != 0;
	if (c)
	{
		_Client = *c;

		os::IRect ir;
		ir.left = _Client.x1;
		ir.top = _Client.y1;
		ir.right = _Client.x2;
		ir.bottom = _Client.y2;

		os::Region br;
		br.Set(ir);

		View->SetDrawingRegion(br);
	}
	else
	{
		View->ClearDrawingRegion();
	}
}

GRegion GdcViewDC::ClipRgn(GRegion *Rgn)
{
	GRegion r(0, 0, 0, 0);

	if (Rgn)
	{
		r = *Rgn;
		if (_ClientClip)
		{
			r.Bound(&_Client);
		}
	}
	else if (_ClientClip)
	{
		r = _Client;
		Rgn = &r;
	}

	if (Rgn)
	{
		os::IRect ir;
		ir.left = r.x1;
		ir.top = r.y1;
		ir.right = r.x2;
		ir.bottom = r.y2;

		os::Region br;
		br.Set(ir);
		View->SetDrawingRegion(br);
	}
	else
	{
		View->ClearDrawingRegion();
	}
}

COLOUR GdcViewDC::Colour()
{
	os::Color32_s c = View->GetFgColor();
	return CBit(GetBits(), Rgb24(c.red, c.green, c.blue), 24);
}

COLOUR GdcViewDC::Colour(COLOUR c, int Bits)
{
	COLOUR Old = Colour();

	COLOUR Col = CBit(GetBits(), c, Bits>0 ? Bits : GetBits());
	View->SetFgColor(R24(Col), G24(Col), B24(Col));

	return Old;
}

int GdcViewDC::Op()
{
	switch (View->GetDrawingMode())
	{
		case os::DM_COPY:
			return GDC_SET;
		default:
			return GDC_SET;
	}
}

int GdcViewDC::Op(int NewOp)
{
	int Old = Op();

	switch (NewOp)
	{
		default:
			View->SetDrawingMode(os::DM_COPY);
			break;
		/* FIXME Whats the equivilent codes for those operators?
		case GDC_OR:
			View->SetDrawingMode();
			break;
		case GDC_AND:
			View->SetDrawingMode();
			break;
		case GDC_XOR:
			View->SetDrawingMode();
			break;
		*/
	}

	return Old;
}

int GdcViewDC::X()
{
	OsRect r = View->GetFrame();
	return (int) r.Width();
}

int GdcViewDC::Y()
{
	OsRect r = View->GetFrame();
	return (int) r.Height();
}

int GdcViewDC::GetBits()
{
	return GdcD->GetBits();
}

int GdcViewDC::GetFlags()
{
	return 0;
}

void GdcViewDC::Set(int x, int y)
{
	os::Rect r(x, y, x, y);
	View->FillRect(r);
}

COLOUR GdcViewDC::Get(int x, int y)
{
	// can't implement
	return 0;
}

void GdcViewDC::HLine(int x1, int x2, int y)
{
	os::Point From(x1, y);
	os::Point To(x2, y);
	View->DrawLine(From, To);
}

void GdcViewDC::VLine(int x, int y1, int y2)
{
	os::Point From(x, y1);
	os::Point To(x, y2);
	View->DrawLine(From, To);
}

void GdcViewDC::Line(int x1, int y1, int x2, int y2)
{
	os::Point From(x1, y1);
	os::Point To(x2, y2);
	View->DrawLine(From, To);
}

void GdcViewDC::Circle(double cx, double cy, double radius)
{
	// FIXME
}

void GdcViewDC::FilledCircle(double cx, double cy, double radius)
{
	// FIXME
}

void GdcViewDC::Arc(double cx, double cy, double radius, double start, double end)
{
	// FIXME
}

void GdcViewDC::FilledArc(double cx, double cy, double radius, double start, double end)
{
	// FIXME
}

void GdcViewDC::Ellipse(double cx, double cy, double x, double y)
{
	// FIXME
}

void GdcViewDC::FilledEllipse(double cx, double cy, double x, double y)
{
	// FIXME
}

void GdcViewDC::Box(int x1, int y1, int x2, int y2)
{
	os::Point From(x1, y1);
	os::Point To(x2, y1);
	View->DrawLine(From, To);
	To.x = x1;
	To.y = y2;
	View->DrawLine(From, To);
	From.x = x2;
	From.y = y2;
	View->DrawLine(From, To);
	To.x = x2;
	To.y = y1;
	View->DrawLine(From, To);
}

void GdcViewDC::Box(GRegion *a)
{
	Box(a->x1, a->y1, a->x2, a->y2);
}

void GdcViewDC::Rectangle(int x1, int y1, int x2, int y2)
{
	OsRect r(x1, y1, x2, y2);
	View->FillRect(r);
}

void GdcViewDC::Rectangle(GRegion *a)
{
	View->FillRect(*a);
}

void GdcViewDC::Blt(int x, int y, GSurface *Src, GRegion *a)
{
	if (Src)
	{
		if (Src->IsScreen())
		{
			// can't do
		}
		else if (Src->GetBitmap())
		{
			os::Rect Def(0, 0, Src->X()-1, Src->Y()-1);
			os::Rect s;

			if (a)
			{
				s = *a;
			}
			else
			{
				s = Def;
			}

			os::Rect d(x, y, x+s.Width()-1, y + s.Height()-1);

			View->DrawBitmap(Src->GetBitmap(), s, d);
		}
	}
}

void GdcViewDC::StretchBlt(GRegion *d, GSurface *Src, GRegion *s)
{
	// FIXME
}

void GdcViewDC::Polygon(int Points, GdcPt2 *Data)
{
	// FIXME
}

void GdcViewDC::Text(GdcFont *pFont, int x, int y, char *Str, int Len)
{
	if (pFont)
	{
	}
}

void GdcViewDC::Bezier(int Threshold, GdcPt2 *Pt)
{
	// FIXME
}

void GdcViewDC::FloodFill(int x, int y, int Mode, COLOUR Border, GRegion *Bounds)
{
	// FIXME
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
GMemDC::GMemDC()
{
	Bmp = 0;
}

GMemDC::GMemDC(GSurface *pDC)
{
	Bmp = 0;

	if (pDC AND
		Create(pDC->X(), pDC->Y(), pDC->GetBits()))
	{
		Blt(0, 0, pDC);
	}
}

GMemDC::~GMemDC()
{
	DeleteObj(Bmp);
}

bool GMemDC::Create(int x, int y, int Bits, int LineLen, bool KeepData)
{
	bool Status = false;
	
	// set minimum size
	if (x < 1) x = 1;
	if (y < 1) y = 1;

	// allocate bitmap
	os::color_space Cs;
	switch (Bits)
	{
		case 8:
		{
			Cs = os::CS_CMAP8;
			break;
		}
		case 15:
		{
			Cs = os::CS_RGB15;
			break;
		}
		default:
		case 16:
		{
			Cs = os::CS_RGB16;
			break;
		}
		case 24:
		{
			Cs = os::CS_RGB24;
			break;
		}
		case 32:
		{
			Cs = os::CS_RGB32;
			break;
		}
	}

	Bmp = NEW(os::Bitmap(x, y, Cs));
	if (Bmp)
	{
		pMem = NEW(GdcSurface);
		if (pMem)
		{
			pMem->Base = (uchar*) Bmp->LockRaster();
			pMem->x = x;
			pMem->y = y;
			pMem->Bits = Bits;
			pMem->Line = Bmp->GetBytesPerRow();
			pMem->Flags = 0;
			
			Status = TRUE;

			int NewOp = (pApp) ? Op() : GDC_SET;

			if (	(Flags & GDC_OWN_APPLICATOR) AND
				NOT (Flags & GDC_CACHED_APPLICATOR))
			{
				DeleteObj(pApp);
			}

			for (int i=0; i<GDC_CACHE_SIZE; i++)
			{
				DeleteObj(pAppCache[i]);
			}

			if (NewOp < GDC_CACHE_SIZE AND NOT DrawOnAlpha())
			{
				pApp = (pAppCache[NewOp]) ? pAppCache[NewOp] : pAppCache[NewOp] = CreateApplicator(NewOp);
				Flags &= ~GDC_OWN_APPLICATOR;
				Flags |= GDC_CACHED_APPLICATOR;
			}
			else
			{
				pApp = CreateApplicator(NewOp);
				Flags &= ~GDC_CACHED_APPLICATOR;
				Flags |= GDC_OWN_APPLICATOR;
			}

			Clip.ZOff(X()-1, Y()-1);
		}
	}

	return Status;
}

void GMemDC::Blt(int x, int y, GSurface *Src, GRegion *a)
{
	if (Src)
	{
		if (Src->IsScreen())
		{
			// screen to memory Blt
		}
		else
		{
			// memory to memory Blt
			GSurface::Blt(x, y, Src, a);
		}
	}
}

void GMemDC::StretchBlt(GRegion *d, GSurface *Src, GRegion *s)
{
}

void GMemDC::HLine(int x1, int x2, int y, COLOUR a, COLOUR b)
{
	if (x1 > x2) LgiSwap(x1, x2);

	if (x1 < Clip.x1) x1 = Clip.x1;
	if (x2 > Clip.x2) x2 = Clip.x2;
	if (	x1 <= x2 AND
		y >= Clip.y1 AND
		y <= Clip.y2)
	{
		COLOUR Prev = pApp->c;

		pApp->SetPtr(x1, y);
		for (; x1 <= x2; x1++)
		{
			if (x1 & 1)
			{
				pApp->c = a;
			}
			else
			{
				pApp->c = b;
			}

			pApp->Set();
			pApp->IncX();
		}

		pApp->c = Prev;
	}
}

void GMemDC::VLine(int x, int y1, int y2, COLOUR a, COLOUR b)
{
	if (y1 > y2) LgiSwap(y1, y2);
	
	if (y1 < Clip.y1) y1 = Clip.y1;
	if (y2 > Clip.y2) y2 = Clip.y2;
	if (	y1 <= y2 AND
		x >= Clip.x1 AND
		x <= Clip.x2)
	{
		COLOUR Prev = pApp->c;

		pApp->SetPtr(x, y1);
		for (; y1 <= y2; y1++)
		{
			if (y1 & 1)
			{
				pApp->c = a;
			}
			else
			{
				pApp->c = b;
			}

			pApp->Set();
			pApp->IncY();
		}

		pApp->c = Prev;
	}
}


