
#ifndef __LGI_OS_DEFS_H
#define __LGI_OS_DEFS_H

#include <util/application.h>
#include <util/message.h>
#include <gui/window.h>
#include <gui/view.h>
#include <gui/requesters.h>
#include <gui/bitmap.h>
#include <storage/directory.h>

#include "assert.h"

//////////////////////////////////////////////////////////////////
// Typedefs
typedef			long long int		quad;
typedef	signed long long int		int64;
typedef unsigned long long int		uint64;

typedef os::Window					*OsWindow;
typedef os::View						*OsView;
typedef os::Bitmap					*OsBitmap;
typedef int							OsThread;
typedef os::Message					GMessage;

/////////////////////////////////////////////////////////////////
// Defines
// #define _DEBUG 1

#define MAKEINTRESOURCE(i)		((char*)(i))

// assert
#ifdef _DEBUG
LgiFunc void _lgi_assert(bool b, char *test, char *file, int line);
#define LgiAssert(b)				_lgi_assert(b, #b, __FILE__, __LINE__)
#else
#define LgiAssert(b)			
#endif

#define LgiSleep(i)				snooze(i*1000)
LgiFunc void _lgi_yield();
#define LgiYield()				_lgi_yield()

#define min(a, b)					((a) < (b)) ? (a) : (b)
#define max(a, b)					((a) > (b)) ? (a) : (b)

#define K_CHAR					0x0

#define SND_ASYNC 				1

#define DOUBLE_CLICK_THRESHOLD	5

#define MB_OK						1
#define MB_YESNO					2
#define MB_YESNOCANCEL			3
#define IDYES						4
#define IDNO						5

#define MK_LBUTTON				0x01
#define MK_RBUTTON				0x02
#define MK_MBUTTON				0x04
#define MK_CONTROL				0x08
#define MK_ALT					0x10
#define MK_SHIFT					0x20

// Window flags
#define GWF_VISIBLE				0x00000001
#define GWF_ENABLED				0x00000002
#define GWF_FOCUS					0x00000004
#define GWF_OVER					0x00000008
#define GWF_DROP_TARGET			0x00000010
#define GWF_SUNKEN				0x00000020
#define GWF_FLAT					0x00000040
#define GWF_RAISED				0x00000080
#define GWF_BORDER				0x00000100
#define GWF_DIALOG				0x00000200
#define GWF_DESTRUCTOR			0x00000400
#define GWF_QUIT_WND				0x00000800

// Edge types
#define SUNKEN					1
#define RAISED					2
#define CHISEL					3
#define FLAT						4

// Widgets
#define CTRL_SUNKEN				0x80000000
#define CTRL_FLAT					0x40000000
#define CTRL_RAISED				0x20000000

// Directories
#define DIR_CHAR					'/'
#define DIR_STR					"/"

#define IsSlash(c)				(((c)=='/')OR((c)=='\\'))
#define IsQuote(c)				(((c)=='\"')OR((c)=='\''))

// Messages
#define WM_USER					(0xF0000000) // how strange, this isn't defined in BeOS?? :P

// wParam = bool Inside;			// is the mouse inside the client area?
// lParam = MAKELONG(x, y);		// mouse location
#define WM_MOUSEENTER			(WM_USER+0x100)
#define WM_MOUSEEXIT				(WM_USER+0x101)

// wParam = (GView*) Wnd;
// lParam = (int) Flags;
#define WM_CHANGE					(WM_USER+0x102)

// wParam = (GView*) Wnd;
// lParam = (char*) Text;		// description from window
#define WM_DESCRIBE				(WM_USER+0x103)

// return (bool)
#define WM_WANT_DIALOG_PROC		(WM_USER+0x104)

#define WM_MENU					(WM_USER+0x105)
#define WM_COMMAND				(WM_USER+0x106)
#define WM_DRAG_DROP				(WM_USER+0x107)
#define WM_VSCROLL				(WM_USER+0x108)
#define WM_HSCROLL				(WM_USER+0x109)
#define WM_PULSE					(WM_USER+0x10a)

// Dialog stuff
#define IDOK						1
#define IDCANCEL					2

// Keys
#define VK_SHIFT					1
#define VK_DELETE					os::VK_DELETE
#define VK_BACK					os::VK_BACKSPACE
#define VK_RIGHT					os::VK_RIGHT_ARROW
#define VK_LEFT					os::VK_LEFT_ARROW
#define VK_UP						os::VK_UP_ARROW
#define VK_DOWN					os::VK_DOWN_ARROW
#define VK_PRIOR					os::VK_PAGE_UP
#define VK_NEXT					os::VK_PAGE_DOWN
#define VK_RETURN					os::VK_RETURN
#define VK_HOME					os::VK_HOME
#define VK_END					os::VK_END
#define VK_ESCAPE					os::VK_ESCAPE
#define VK_INSERT					os::VK_INSERT

// The BeOS defines actually overlap some other keys
// so I don't use them, instead I just defined my own block
// of codes
#define VK_F1						15
#define VK_F2						16
#define VK_F3						17
#define VK_F4						18
#define VK_F5						19
#define VK_F6						20
#define VK_F7						21
#define VK_F8						22
#define VK_F9						23
#define VK_F10					24
#define VK_F11					25
#define VK_F12					26

/////////////////////////////////////////////////////////////////////////////////////
#define MsgCode(m)				(m->GetCode())
LgiFunc int MsgA(GMessage *m);
LgiFunc int MsgB(GMessage *m);
LgiFunc GMessage &CreateMsg(int m, int a, int b);

/////////////////////////////////////////////////////////////////////////////////////
// Externs
extern int stricmp(char *a, char *b);
extern char *strnistr(char *a, char *b, int n);
extern int strnicmp(char *a, char *b, int i);
extern void FormatSize(char *Str, quad Size);
extern void strupr(char *a);
extern void strlwr(char *a);
extern int stricmp(char *a, char *b);

#endif















