#include <atheos/time.h>

#include "Lgi.h"

/////////////////////////////////////////////////////////////////////////////
GApp *GApp::TheApp = 0;

GApp::GApp(char *AppMime, int Args, char **Arg) :
	os::Application(AppMime)
{
	TheApp = this;

	_AppFile = NewStr(Arg[0]);

	int CmdLen = 0;
	for (int i=1; i<Args; i++)
	{
		CmdLen += strlen(Arg[i]) + ((i > 1) ? 1 : 0);
	}
	CmdLine = NEW(char[CmdLen+1]);
	if (CmdLine)
	{
		CmdLine[0] = 0;
		for (int i=1; i<Args; i++)
		{
			if (i > 1)
			{
				strcat(CmdLine, " ");
			}
			strcat(CmdLine, Arg[i]);
		}
	}

	srand(get_system_time());
	LgiInitColours();
	AppWnd = 0;

	// System font
	SystemFont = 0;
	GdcFontType SysFontType;
	if (SysFontType.GetSystemFont("System"))
	{
		SystemFont = SysFontType.CreateFont();
		if (SystemFont)
		{
			SystemFont->Transparent(true);
		}
	}

	if (NOT SystemFont)
	{
		LgiMsg(0, "Error: Couldn't create system font.", "Lgi Error: GApp::GApp", MB_OK);
		LgiExitApp();
	}
}

GApp::~GApp()
{
	TheApp = 0;
}

bool GApp::Run(bool Loop)
{
	printf("Enter app loop...\n");
	os::Application::Run();
	return 0;
}

void GApp::OnReceiveFiles(List<char> &Files)
{
	if (AppWnd)
	{
		AppWnd->OnReceiveFiles(Files);
	}
}

XmlTag *GApp::GetConfig(char *Tag)
{
	if (NOT Config)
	{
		char Path[256];
		if (LgiGetExePath(Path))
		{
			if (Path[strlen(Path)-1] != DIR_CHAR) strcat(Path, DIR_STR);
			strcat(Path, "lgi.conf");
			if (FileExists(Path))
			{
				Config = NEW(Xml);
				if (Config)
				{
					if (Config->ParseXmlFile(Path) <= 0)
					{
						DeleteObj(Config);
					}
				}
			}
		}
	}

	if (Tag AND Config)
	{
		for (XmlTag *t=Config->Values.First(); t; t=Config->Values.Next())
		{
			if (t->Name AND
				stricmp(t->Name, Tag) == 0)
			{
				return t;
			}
		}
	}

	return 0;
}

bool GApp::GetOption(char *Dest, char Option)
{
	/*
	printf("GApp::GetOption(%c)\n", Option);
	printf("CmdLine: '%s'\n", CmdLine);
	*/

	char *c = CmdLine;
	while (c AND *c)
	{
		if (*c == '/' OR *c == '\\' OR *c == '-')
		{
			c++;
			if (*c == Option)
			{
				c++;
				if (*c)
				{
					while (*c AND strchr(" \t\r\n", *c))
					{
						c++;
					}

					for (char *i = c; *i AND *i != ' ';)
					{
						*Dest++ = *i++;
					}
					*Dest++ = 0;
					return TRUE;
				}
			}
		}

		c = strchr(c, ' ');
		if (c) c++;
	}

	return FALSE;
}

void GApp::OnCmdLine(char *Cmd)
{
	char WhiteSpace[] = " \r\n\t";

	DeleteArray(CmdLine);
	CmdLine = NewStr(Cmd);

	if (CmdLine)
	{
		List<char> Files;
		
		char *Delim = "\'\"";
		char *s;
		for (s = CmdLine; *s; )
		{
			// skip ws
			while (strchr(WhiteSpace, *s)) s++;

			// read to end of token
			char *e = s;
			if (strchr(Delim, *s))
			{
				for (e = ++s; *e AND NOT strchr(Delim, *e); e++);
			}
			else
			{
				for (; *e AND NOT strchr(WhiteSpace, *e); e++);
			}

			char *Arg = NewStr(s, (int)e-(int)s);
			if (Arg)
			{
				if (FileExists(Arg))
				{
					Files.Insert(Arg);
				}
				else
				{
					DeleteArray(Arg);
				}
			}

			// next
			s = (*e) ? e + 1 : e;
		}

		// call app
		if (Files.GetItems() > 0)
		{
			OnReceiveFiles(Files);
		}

		// clear up
		DeleteArrayList(&Files, char);
	}
}

void GApp::Exit(int Code)
{
	#ifdef _DEBUG
	printf("\nGApp::Exit(%i) called.\n", Code);
	#endif

	if (Code)
	{
		exit(Code);
	}
	else
	{
		os::Application::GetInstance()->PostMessage(os::M_QUIT);
	}
}












