/*hdr
**      FILE:           GFileSelect.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           18/7/98
**      DESCRIPTION:    Common file/directory selection dialog
**
**      Copyright (C) 1998-2001, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdlib.h>
#include "Lgi.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
GFileSelect::GFileSelect()
{
	Size = 4096;
	FileStr = NEW(char[Size]);
	if (FileStr)
	{
		*FileStr = 0;
	}
	else
	{
		Size = 0;
	}
	
	MultiSel = FALSE;
	InitDir = 0;
	DefExt = 0;
	TitleStr = 0;
}

GFileSelect::~GFileSelect()
{
	DeleteArray(FileStr);
	DeleteArray(InitDir);
	DeleteArray(DefExt);
	DeleteArray(TitleStr);

	for (GFileType *Type = TypeList.First(); Type; Type = TypeList.Next())
	{
		DeleteObj(Type);
	}
}

char *GFileSelect::Name()
{
	return FileStr;
}

bool GFileSelect::Name(char *n)
{
	bool Status = FALSE;

	if (n AND FileStr)
	{
		strncpy(FileStr, n, Size);
	}

	return Status;
}

char *GFileSelect::operator [](int i)
{
	return NULL;
}

int GFileSelect::Names()
{
	return 1;
}

char *GFileSelect::File()
{
	char *Ret = 0;
	char *Last = FileStr;
	char *Dir = strchr(Last, '/');
	while (Dir)
	{
		Last = Dir + 1;
		Dir = strchr(Last, '/');
	}
	return Last;
}

char *GFileSelect::Extension()
{
	char *Ret = File();
	char *Dot = strchr(Ret, '.');
	if (Dot)
	{
		return Dot + 1;
	}
	return NULL;
}

int GFileSelect::Types()
{
	return TypeList.GetItems();
}

void GFileSelect::ClearTypes()
{
	DeleteObjList(&TypeList, GFileType);
}

GFileType *GFileSelect::TypeAt(int n)
{
	return TypeList[n];
}

bool GFileSelect::Type(char *Description, char *Extension, int Data)
{
	GFileType *Type = NEW(GFileType);
	if (Type)
	{
		Type->Description(Description);
		Type->Extension(Extension);
		TypeList.Insert(Type);
	}

	return Type != 0;
}

int GFileSelect::SelectedType()
{
	return -1;
}

GView *GFileSelect::Parent()
{
	return ParentWnd;
}

void GFileSelect::Parent(GView *Window)
{
	ParentWnd = Window;
}

bool GFileSelect::MultiSelect()
{
	return MultiSel;
}

void GFileSelect::MultiSelect(bool Multi)
{
	MultiSel = Multi;
}

char *GFileSelect::InitialDir()
{
	return InitDir;
}

void GFileSelect::InitialDir(char *Dir)
{
	DeleteArray(InitDir);
	InitDir = NewStr(Dir);
}

char *GFileSelect::Title()
{
	return TitleStr;
}

void GFileSelect::Title(char *title)
{
	DeleteArray(TitleStr);
	TitleStr = NewStr(title);
}

char *GFileSelect::DefaultExtension()
{
	return DefExt;
}

void GFileSelect::DefaultExtension(char *DefExt)
{
	DeleteArray(DefExt);
	DefExt = NewStr(DefExt);
}

bool GFileSelect::Open()
{
	return false;
}

bool GFileSelect::OpenFolder()
{
	return false;
}

bool GFileSelect::Save()
{
	return false;
}

