/*
**	FILE:			GLayout.cpp
**	AUTHOR:		Matthew Allen
**	DATE:			19/10/2001
**	DESCRIPTION:	Standard Views
**
**	Copyright (C) 2001, Matthew Allen
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
GLayout::GLayout() :
	GView( NEW(AView<os::View>(this)) )
{
	_PourLargest = false;
	VScroll = 0;
	HScroll = 0;
	XScroll = YScroll = false;
}

GLayout::~GLayout()
{
	DeleteObj(HScroll);
	DeleteObj(VScroll);
}

bool GLayout::GetPourLargest()
{
	return _PourLargest;
}

void GLayout::SetPourLargest(bool i)
{
	_PourLargest = i;
}

bool GLayout::Pour(GComplexRegion &r)
{
	if (_PourLargest)
	{
		GRegion *Best = FindLargest(r);
		if (Best)
		{
			SetPos(*Best, true);
			return true;
		}
	}
	
	return false;
}

void GLayout::GetScrollPos(int &x, int &y)
{
	if (HScroll)
	{
		x = HScroll->Pos();
	}
	else
	{
		x = 0;
	}

	if (VScroll)
	{
		y = VScroll->Pos();
	}
	else
	{
		y = 0;
	}
}

void GLayout::SetScrollPos(int x, int y)
{
	if (HScroll)
	{
		HScroll->SetPos(x);
	}

	if (VScroll)
	{
		VScroll->SetPos(y);
	}
}

bool GLayout::SetScrollBars(bool x, bool y)
{
	bool Has = (HScroll != 0) OR (VScroll != 0);
	bool Changing = Has ^ (x OR y);
	bool Attached = IsAttached();

	if (Changing)
	{
		if (Attached)
		{
			// remove scroll bars
		}
	
		if (x)
		{
			if (NOT HScroll)
			{
				HScroll = NEW(GScrollBar);
				if (HScroll)
				{
					HScroll->SetId(IDC_HSCROLL);
				}
			}
		}
		else
		{
			DeleteObj(HScroll);
		}
		if (y)
		{
			if (NOT VScroll)
			{
				VScroll = NEW(GScrollBar);
				if (VScroll)
				{
					VScroll->SetId(IDC_VSCROLL);
				}
			}
		}
		else
		{
			DeleteObj(VScroll);
		}
	
		if ((x OR y) AND
			Attached)
		{
			Sunken(false);
		}
	}
	
	return true;
}

int GLayout::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case WM_CHANGE:
		{
			if (GetParent())
			{
				GetParent()->OnEvent(Msg);
			}
			break;
		}
		case WM_VSCROLL:
		{
			if (VScroll)
			{
				OnNotify(VScroll, MsgA(Msg));
			}
			break;
		}
		case WM_HSCROLL:
		{
			if (HScroll)
			{
				OnNotify(HScroll, MsgA(Msg));
			}
			break;
		}
	}
	
	if (VScroll) VScroll->OnEvent(Msg);
	if (HScroll) HScroll->OnEvent(Msg);
	return GView::OnEvent(Msg);
}



