/*hdr
**      FILE:           GuiMenu.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           18/7/98
**      DESCRIPTION:    Gui menu system
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

///////////////////////////////////////////////////////////////////////////////////////////////
extern char *NewStrLessAnd(char *s);
char *NewStrLessAnd(char *s)
{
	char *ns = NewStr(s);
	if (ns)
	{
		char *In = ns;
		char *Out = ns;
		while (*In)
		{
			if (*In != '&')
			{
				*Out++ = *In;
			}
			In++;
		}
		*Out = 0;
	}
	return ns;
}

///////////////////////////////////////////////////////////////////////////////////////////////
GSubMenu::GSubMenu(char *name)
{
	Info = 0;
	if (name)
	{
		char *Temp = NewStrLessAnd(name);
		if (Temp)
		{
			Info = NEW(os::Menu(os::Rect(0,0,1,1), Temp, os::ITEMS_IN_COLUMN));
			DeleteArray(Temp);
		}
	}

	Parent = 0;
	Menu = 0;
	Window = 0;
}

GSubMenu::~GSubMenu()
{
	DeleteObj(Info);
}

GMenuItem *GSubMenu::AppendItem(char *Str, int Id, bool Enabled, int Where)
{
	GMenuItem *i = NEW(GMenuItem);
	if (i)
	{
		Items.Insert(i, Where);

		i->Menu = Menu;
		i->Parent = this;
		i->Position = Items.IndexOf(i);
		i->Name(Str);
		i->Id(Id);
		i->Enabled(Enabled);

		// Info->AddItem(i->Info, i->Position);
	}

	return i;
}

GMenuItem *GSubMenu::AppendSeparator(int Where)
{
	GMenuItem *i = NEW(GMenuItem(0, 0, -1));
	if (i)
	{
		Items.Insert(i, Where);

		i->Menu = Menu;
		i->Parent = this;
		i->Position = Items.IndexOf(i);

		// Info->AddItem(i->Info, i->Position);
	}
	
	return i;
}

GSubMenu *GSubMenu::AppendSub(char *Str, int Where)
{
	GSubMenu *Sub = Sub = NEW(GSubMenu(Str));
	GMenuItem *i = NEW(GMenuItem(0, Sub, -1));
	if (i)
	{
		Items.Insert(i, Where);

		i->Name(Str);
		i->Menu = Menu;
		i->Parent = this;
		i->Position = Items.IndexOf(i);

		Sub->Menu = Menu;
		Sub->Parent = i;
		Sub->Window = Window;

		// Info->AddItem(i->Info, i->Position);
	}

	return Sub;
}

void GSubMenu::Empty()
{
	GMenuItem *i = 0;
	while (i = Items.First())
	{
		DeleteObj(i);
	}
}

GMenuItem *GSubMenu::ItemAt(int i)
{
	return Items.ItemAt(i);
}

bool GSubMenu::RemoveItem(int i)
{
	bool Status = false;
	GMenuItem *Item = Items.ItemAt(i);
	if (Item)
	{
		Items.Delete(Item);
		Status = true;
	}
	return Status;
}

bool GSubMenu::RemoveItem(GMenuItem *Item)
{
	bool Status = false;
	if (Item AND Items.HasItem(Item))
	{
		Items.Delete(Item);
		Status = true;
	}
	return Status;
}

int GSubMenu::Float(GView *Parent, int x, int y, bool Left)
{
	return -1;
}

GMenuItem *GSubMenu::FindItem(int Id)
{
	for (int n=0; n<Items.GetItems(); n++)
	{
		GMenuItem *i = Items.ItemAt(n);
		if (i)
		{
			GSubMenu *Sub = i->Sub();
	
			if (i->Id() == Id)
			{
				return i;
			}
			else if (Sub)
			{
				i = Sub->FindItem(Id);
				if (i)
				{
					return i;
				}
			}
		}
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////
GMenuItem::GMenuItem()
{
	Menu = 0;
	Parent = 0;
	Child = 0;
	Position = -1;
	_Icon = -1;
}

GMenuItem::GMenuItem(GMenu *m, GSubMenu *p, int Pos)
{
	Menu = m;
	Parent = p;
	Child = 0;
	Position = Pos;
	_Icon = -1;
}

GMenuItem::~GMenuItem()
{
	if (Parent)
	{
		Parent->Items.Delete(this);
	}
}

void GMenuItem::_Measure(GdcPt2 &Size)
{
}

void GMenuItem::_Paint(GSurface *pDC, int Flags)
{
	if (_Icon >= 0)
	{
		if (Parent AND
			Parent->GetImageList())
		{
			Parent->GetImageList()->Draw(pDC, -16, 0, _Icon);
		}
	}
}

void GMenuItem::_PaintText(GSurface *pDC, int x, int y, int Width)
{
}

void GMenuItem::Icon(int i)
{
	_Icon = i;
}

int GMenuItem::Icon()
{
	return _Icon;
}

GSubMenu *GMenuItem::Create()
{
	return 0;
}

bool GMenuItem::Remove()
{
	if (Parent)
	{
		Parent->Items.Delete(this);
	}
	
	return true;
}

GSubMenu *GMenuItem::GetParent()
{
	return Parent;
}

void GMenuItem::Id(int i)
{
}

int GMenuItem::Id()
{
	return 0;
}

void GMenuItem::Separator(bool s)
{
}

bool GMenuItem::Separator()
{
	return 0;
}

void GMenuItem::Checked(bool c)
{
}

bool GMenuItem::Checked()
{
	return false;
}

bool GMenuItem::Name(char *n)
{
	bool Status = false;
	char *p = (n) ? NewStrLessAnd(n) : NewStr("");
	if (p)
	{
		char *Tab = strchr(p, '\t');
		if (Tab)
		{
			*Tab = 0;
			Tab++;
			
			if (strnicmp(Tab, "Ctrl+", 5) == 0)
			{
			}
			else if (stricmp(Tab, "Del") == 0)
			{
			}
		}

		Status = GObject::Name(p);
		DeleteArray(p);
	}

	return Status;
}

char *GMenuItem::Name()
{
	return (char*)0;
}

void GMenuItem::Enabled(bool e)
{
}

bool GMenuItem::Enabled()
{
	return false;
}

void GMenuItem::Focus(bool f)
{
}

bool GMenuItem::Focus()
{
	return false;
}

void GMenuItem::Visible(bool v)
{
}

bool GMenuItem::Visible()
{
	return true;
}

void GMenuItem::Sub(GSubMenu *s)
{
	DeleteObj(Child);
	Child = s;
	if (Child)
	{
		// fixme.... reallocate the item to have a submenu
	}
}

GSubMenu *GMenuItem::Sub()
{
	return Child;
}

///////////////////////////////////////////////////////////////////////////////////////////////
GFont *GMenu::_Font = 0;

GMenu::GMenu()
{
	Window = NULL;
	Menu = this;
}

GMenu::~GMenu()
{
	Detach();
}

GFont *GMenu::GetFont()
{
	if (NOT _Font)
	{
		GdcFontType Type;
		if (Type.GetSystemFont("Menu"))
		{
			_Font = Type.CreateFont();
		}
	}

	return _Font ? _Font : SysFont;
}

bool GMenu::Attach(GView *p)
{
printf("GMenu::Attach(GView*p) start\n");

	bool Status = false;
	Window = p;
	if (Window)
	{
		os::Window *w = Window->GetWindow();
		if (w)
		{
			bool Lock = w->Lock() == 0;
printf("GMenu::Attach(GView*p) lock=%i\n", Lock);

			os::Rect r = w->GetBounds();
			r.bottom = r.top + 20;

printf("GMenu::Attach(GView*p) bounds=%.1f,%.1f,%.1f,%.1f\n", r.left, r.top, r.right, r.bottom);

			Info = NEW( os::Menu(r,
									"main_menu",
									os::ITEMS_IN_ROW,
									os::CF_FOLLOW_LEFT | os::CF_FOLLOW_RIGHT,
									os::WID_FULL_UPDATE_ON_H_RESIZE) );

printf("GMenu::Attach(GView*p) Info=%p\n", Info);

			w->AddChild(Info);

			if (Lock) w->Unlock();
		}

		Status = true;
	}

printf("GMenu::Attach(GView*p) end\n");

	return Status;
}

bool GMenu::Detach()
{
	bool Status = FALSE;

	if (Window)
	{
		Window = 0;
	}

	return Status;
}

///////////////////////////////////////////////////////////////////////////////////////////
GCommand::GCommand()
{
	Flags = GWF_ENABLED | GWF_VISIBLE;
	Id = 0;
	ToolButton = 0;
	MenuItem = 0;
	Accelerator = 0;
	TipHelp = 0;
}

GCommand::~GCommand()
{
	DeleteArray(Accelerator);
	DeleteArray(TipHelp);
}

bool GCommand::Enabled()
{
	if (ToolButton)
	{
		return ToolButton->Enabled();
	}
	else if (MenuItem)
	{
		return MenuItem->Enabled();
	}
	
	return false;
}

void GCommand::Enabled(bool e)
{
	if (Enabled() != e)
	{
		if (ToolButton)
		{
			ToolButton->Enabled(e);
		}
		if (MenuItem)
		{
			MenuItem->Enabled(e);
		}
	}
}

bool GCommand::Value()
{
	bool v = FALSE;
	if (ToolButton)
	{
		v |= ToolButton->Value();
	}
	if (MenuItem)
	{
		v |= MenuItem->Checked();
	}
	return v;
}

void GCommand::Value(bool v)
{
	if (ToolButton)
	{
		ToolButton->Value(v);
	}
	if (MenuItem)
	{
		MenuItem->Checked(v);
	}
}



