#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
#define SPLITER_BORDER			4
#define SPLITER_BAR				5
#define SPLITER_INSET			2
#define SPLITER_MIN_X			(2*SPLITER_INSET)
#define SPLITER_MIN_Y			(2*SPLITER_INSET)

#define SPLITER_FULL_DRAG

GSplitter::GSplitter()
{
	Name("LGI_Spliter");
	SplitSet = SplitPos = -1;
	Moving = FALSE;
	_BorderSize = 1;

	BorderA = BorderB = true;
	PosA.ZOff(0, 0);
	PosB.ZOff(0, 0);
	ViewA = 0;
	ViewB = 0;
	SplitFollow = false;

	IsVertical(true);
	Border(true);
	Raised(true);
	Split(0);

	SetPourLargest(true);

	GetBView()->SetViewColor(B_TRANSPARENT_COLOR);
}

GSplitter::~GSplitter()
{
	DeleteObj(ViewA);
	DeleteObj(ViewB);
}

void GSplitter::IsVertical(bool v)
{
	Vertical = v;
}

void GSplitter::SetViewA(GView *a, bool Border)
{
	BorderA = Border;
	ViewA = a;
	if (a AND GetParent())
	{
		a->Attach(this);
		a->Visible(true);

		CalcRegions(Pos);
		Invalidate();
	}
}

void GSplitter::SetViewB(GView *b, bool Border)
{
	BorderB = Border;
	ViewB = b;
	if (b AND GetParent())
	{
		b->Attach(this);
		b->Visible(true);

		CalcRegions(Pos);
		Invalidate();
	}
}

void GSplitter::Split(int s)
{
	SplitSet = s;
	int Limit = ((Vertical) ? Pos.X() : Pos.Y()) - 18;
	int NewPos = (s > 0) ? max(s, 4) : max(4, Limit+s);
	if (NewPos != SplitPos)
	{
		SplitPos = NewPos;
		CalcRegions(Pos);
		if (Visible())
		{
			GView::Invalidate();
		}
	}
}

void GSplitter::CalcRegions(GRegion Rect, bool Follow)
{
	PosA = Rect;
	PosA.Offset(-PosA.x1, -PosA.y1);
	if (Border())
	{
		PosA.Size(SPLITER_BORDER, SPLITER_BORDER);
	}
	
	//printf("this=%p, gsplitter::calc(%i, %i, %i, %i)\n", this, PosA.x1, PosA.y1, PosA.x2, PosA.y2);

	PosB = PosA;

	if (SplitFollow AND Follow)
	{
		SplitPos = max(1, ((Vertical) ? Pos.X()-10 : Pos.Y()-10) + SplitSet);
	}

	if (Vertical)
	{
		PosA.x2 = PosA.x1 + SplitPos;
		PosB.x1 = PosA.x2 + SPLITER_BAR;
	}
	else
	{
		PosA.y2 = PosA.y1 + SplitPos;
		PosB.y1 = PosA.y2 + SPLITER_BAR;
	}

	if (ViewA)
	{
		GRegion r = PosA;
		if (BorderA)
		{
			r.Size(SPLITER_INSET, SPLITER_INSET);
		}

		r.Offset(Rect.x1, Rect.y1);
		ViewA->SetPos(r, true);

		GComplexRegion Rgn(r);
		ViewA->Pour(Rgn);
	}
	if (ViewB)
	{
		GRegion r = PosB;
		if (BorderB)
		{
			r.Size(SPLITER_INSET, SPLITER_INSET);
		}

		r.Offset(Rect.x1, Rect.y1);
		ViewB->SetPos(r, true);

		GComplexRegion Rgn(r);
		ViewB->Pour(Rgn);
	}
}

void GSplitter::OnPosChange()
{
	CalcRegions(GetClient(), true);
}

void GSplitter::OnPaint(GSurface *pDC)
{
	GRegion r = GetClient();
	r.Offset(-r.x1, -r.y1);

	// printf("gsplitter::onpaint(%i, %i, %i, %i)\n", r.x1, r.y1, r.x2, r.y2);

	if (Border())
	{
		// edge
		LgiFlatBorder(pDC, r, SPLITER_BORDER);
	}

	// bar
	pDC->Colour(LC_MED, 24);
	if (Vertical)
	{
		pDC->Rectangle(PosA.x2+1, r.y1, PosB.x1-1, r.y2);
	}
	else
	{
		pDC->Rectangle(r.x1, PosA.y2+1, r.x2, PosB.y1-1);
	}

	// PosA region
	r = PosA;
	if (r.X() > SPLITER_MIN_X AND r.Y() > SPLITER_MIN_Y)
	{
		if (BorderA)
		{
			LgiWideBorder(pDC, r, SUNKEN);
		}

		if (ViewA)
		{
		}
		else
		{
			pDC->Colour(0xFFFFFF);
			pDC->Rectangle(&r);
		}
	}
	else
	{
		// too small
	}

	r = PosB;
	if (r.X() > SPLITER_MIN_X AND r.Y() > SPLITER_MIN_Y)
	{
		if (BorderB)
		{
			LgiWideBorder(pDC, r, SUNKEN);
		}

		if (ViewB)
		{
		}
		else
		{
			pDC->Colour(0xFFFFFF);
			pDC->Rectangle(&r);
		}
	}
	else
	{
		// too small
	}
}

void GSplitter::OnMouseClick(GMouse &m)
{
	if (m.Down())
	{
		if (Vertical)
		{
			if (m.x > PosA.x2 AND m.x < PosB.x1 AND m.y > PosA.y1 AND m.y < PosA.y2)
			{
				Capture(TRUE);
				Moving = TRUE;

				Bar.x1 = PosA.x2;
				Bar.x2 = PosB.x1;
				Bar.y1 = PosA.y1;
				Bar.y2 = PosA.y2;
				Offset = m.x - Bar.x1;

				#ifndef SPLITER_FULL_DRAG
				HashRectangle(this, &Bar);
				#endif
			}
		}
		else
		{
			if (m.x > PosA.x1 AND m.x < PosA.x2 AND m.y > PosA.y2 AND m.y < PosB.y1)
			{
				Capture(TRUE);
				Moving = TRUE;

				Bar.x1 = PosA.x1;
				Bar.x2 = PosA.x2;
				Bar.y1 = PosA.y2;
				Bar.y2 = PosB.y1;
				Offset = m.y - Bar.y1;

				#ifndef SPLITER_FULL_DRAG
				HashRectangle(this, &Bar);
				#endif
			}
		}
	}
	else
	{
		#ifdef SPLITER_FULL_DRAG
		Capture(FALSE);
		Moving = FALSE;
		#else
		if (Moving)
		{
			HashRectangle(this, &Bar);
			Capture(FALSE);
			Moving = FALSE;

			int NewPos;
			if (Vertical)
			{
				NewPos = limit(Bar.x1-4, 4, Pos.X()-18);
			}
			else
			{
				NewPos = limit(Bar.y1-4, 4, Pos.Y()-18);
			}

			if (NewPos != SplitPos)
			{
				SplitPos = NewPos;
				CalcRegions(Pos);
				GView::Invalidate();
			}
		}
		#endif
	}
}

void GSplitter::OnMouseMove(GMouse &m)
{
	if (Moving)
	{
		#ifdef SPLITER_FULL_DRAG
		if (Vertical)
		{
			Bar.Offset((m.x - Offset) - Bar.x1, 0);
		}
		else
		{
			Bar.Offset(0, (m.y - Offset) - Bar.y1);
		}

		int NewPos;
		if (Vertical)
		{
			NewPos = limit(Bar.x1-4, 4, Pos.X()-18);
		}
		else
		{
			NewPos = limit(Bar.y1-4, 4, Pos.Y()-18);
		}

		if (NewPos != SplitPos)
		{
			SplitPos = NewPos;
			CalcRegions(GetClient());
			Invalidate();
		}
		#else
		HashRectangle(this, &Bar);
		if (Vertical)
		{
			Bar.Offset((m.x - Offset) - Bar.x1, 0);
		}
		else
		{
			Bar.Offset(0, (m.y - Offset) - Bar.y1);
		}
		HashRectangle(this, &Bar);
		#endif
	}
}

bool GSplitter::Attach(GView *p)
{
	bool Status = GLayout::Attach(p);
	if (Status)
	{
		if (ViewA)
		{
			ViewA->Attach(this);
		}

		if (ViewB)
		{
			ViewB->Attach(this);
		}
	}

	return Status;
}

GView *GSplitter::FindControl(OsView hCtrl)
{
	GView *c = 0;
	if (ViewA) c = ViewA->FindControl(hCtrl);
	if (c) return c;

	if (ViewB) c = ViewB->FindControl(hCtrl);
	return c;
}

