/*hdr
**      FILE:           Gui.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           23/4/98
**      DESCRIPTION:    Memory subsystem
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdlib.h>

#include "Lgi.h"
#include "DragAndDrop.h"

////////////////////////////////////////////////////////////////////////////
class ALock
{
	os::Handler *o;
	os::Looper *l;
	bool Locked;

	void _New()
	{
	}

public:
	ALock()
	{
		o = 0;
		l = 0;
		Locked = false;
	}

	~ALock()
	{
		if (l AND Locked)
		{
			l->Unlock();
		}
	}

	bool Lock(os::Handler *Obj)
	{
		o = Obj;
		if (o)
		{
			l = o->GetLooper();
			if (l)
			{
				Locked = l->Lock() == 0;
			}

			return true;
		}

		return false;
	}

	bool Lock(GView *v)
	{
		return v ? Lock(v->Handle()) : false;
	}
};

void _lgi_yield()
{
}

void _lgi_assert(bool b, char *test, char *file, int line)
{
	if (NOT b)
	{
		LgiMsg(0,
				"Assert failed, file: %s, line: %i\n"
				"%s\n",
				"Debug",
				MB_OK,
				file, line, test);

		printf("Assert failed, file: %s, line: %i\n%s\n", file, line, test);
		exit(-1);					
	}
}

#define ADJ_LEFT			1
#define ADJ_RIGHT			2
#define ADJ_UP			3
#define ADJ_DOWN			4

extern int IsAdjacent(GRegion &a, GRegion &b);
int IsAdjacent(GRegion &a, GRegion &b)
{
	if (	(a.x1 == b.x2 + 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_LEFT;
	}

	if (	(a.x2 == b.x1 - 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_RIGHT;
	}

	if (	(a.y1 == b.y2 + 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_UP;
	}

	if (	(a.y2 == b.y1 - 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_DOWN;
	}

	return 0;
}

extern GRegion JoinAdjacent(GRegion &a, GRegion &b, int Adj);
GRegion JoinAdjacent(GRegion &a, GRegion &b, int Adj)
{
	GRegion t;

	switch (Adj)
	{
		case ADJ_LEFT:
		case ADJ_RIGHT:
		{
			t.y1 = max(a.y1, b.y1);
			t.y2 = min(a.y2, b.y2);
			t.x1 = min(a.x1, b.x1);
			t.x2 = max(a.x2, b.x2);
			break;
		}
		case ADJ_UP:
		case ADJ_DOWN:
		{
			t.y1 = min(a.y1, b.y1);
			t.y2 = max(a.y2, b.y2);
			t.x1 = max(a.x1, b.x1);
			t.x2 = min(a.x2, b.x2);
			break;
		}
	}

	return t;
}

////////////////////////////////////////////////////////////////////////////
bool LgiIsKeyDown(int Key)
{
	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
GFlags::GFlags()
{
	WndFlags = GWF_VISIBLE | GWF_ENABLED;
}

bool GFlags::Enabled()
{
	return TestFlag(WndFlags, GWF_ENABLED);
}

void GFlags::Enabled(bool i)
{
	if (i) SetFlag(WndFlags, GWF_ENABLED);
	else ClearFlag(WndFlags, GWF_ENABLED);
}

bool GFlags::Visible()
{
	return TestFlag(WndFlags, GWF_VISIBLE);
}

void GFlags::Visible(bool i)
{
	if (i) SetFlag(WndFlags, GWF_VISIBLE);
	else ClearFlag(WndFlags, GWF_VISIBLE);
}

bool GFlags::Focus()
{
	return TestFlag(WndFlags, GWF_FOCUS);
}

void GFlags::Focus(bool i)
{
	if (i) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);
}

bool GFlags::DropTarget()
{
	return TestFlag(WndFlags, GWF_DROP_TARGET);
}

void GFlags::DropTarget(bool i)
{
	if (i) SetFlag(WndFlags, GWF_DROP_TARGET);
	else ClearFlag(WndFlags, GWF_DROP_TARGET);
}

#define GWF_ALL (GWF_SUNKEN | GWF_FLAT | GWF_RAISED)

bool GFlags::Sunken()
{
	return TestFlag(WndFlags, GWF_SUNKEN);
}

void GFlags::Sunken(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_SUNKEN);
}

bool GFlags::Flat()
{
	return TestFlag(WndFlags, GWF_FLAT);
}

void GFlags::Flat(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_FLAT);
}

bool GFlags::Raised()
{
	return TestFlag(WndFlags, GWF_RAISED);
}

void GFlags::Raised(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_RAISED);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
GView::GView(OsView view)
{
	_View = view;
	_BorderSize = 2;
	Parent = 0;
	Notify = 0;
	Pos.ZOff(0, 0);
	WndFlags = GWF_ENABLED | GWF_VISIBLE;
}

GView::~GView()
{
	DeleteObj(_View);
	WndFlags = 0;
}

void GView::_Paint(GSurface *pDC)
{
	GRegion r(0, 0, Pos.X()-1, Pos.Y()-1);
	bool HasEdge = (Sunken() OR Raised()) AND _BorderSize;
	if (HasEdge)
	{
		if (_BorderSize == 2)
		{
			LgiWideBorder(pDC, r, Sunken() ? SUNKEN : RAISED);
		}
		else if (_BorderSize == 1)
		{
			LgiThinBorder(pDC, r, Sunken() ? SUNKEN : RAISED);
		}

		pDC->SetOrigin(-r.x1, -r.y1);
		r.Offset(-r.x1, -r.y1);
		pDC->_SetClient(&r);
	}
	
	OnPaint(pDC);
	
	if (HasEdge)
	{
		pDC->ClipRgn(0);
	}
}

GRegion *GView::FindLargest(GComplexRegion &r)
{
	int Pixels = 0;
	GRegion *Best = 0;
	static GRegion Final;

	// do initial run through the list to find largest single region
	for (GRegion *i = r.First(); i; i = r.Next())
	{
		int Pix = i->X() * i->Y();
		if (Pix > Pixels)
		{
			Pixels = Pix;
			Best = i;
		}
	}

	if (Best)
	{
		Final = *Best;
		Pixels = Final.X() * Final.Y();

		int LastPixels = Pixels;
		GRegion LastRgn = Final;
		int ThisPixels = Pixels;
		GRegion ThisRgn = Final;

		GComplexRegion TempList;
		for (GRegion *i = r.First(); i; i = r.Next())
		{
			TempList.Union(i);
		}
		TempList.Subtract(Best);

		do
		{
			LastPixels = ThisPixels;
			LastRgn = ThisRgn;

			// search for adjoining rectangles that maybe we can add
			for (GRegion *i = TempList.First(); i; i = TempList.Next())
			{
				int Adj = IsAdjacent(ThisRgn, *i);
				if (Adj)
				{
					GRegion t = JoinAdjacent(ThisRgn, *i, Adj);
					int Pix = t.X() * t.Y();
					if (Pix > ThisPixels)
					{
						ThisPixels = Pix;
						ThisRgn = t;

						TempList.Subtract(i);
					}
				}
			}
		} while (LastPixels < ThisPixels);

		Final = ThisRgn;
	}

	return &Final;
}

GRegion *GView::FindSmallestFit(GComplexRegion &r, int Sx, int Sy)
{
	int X = 1000000;
	int Y = 1000000;
	GRegion *Best = 0;

	for (GRegion *i = r.First(); i; i = r.Next())
	{
		if ((i->X() >= Sx AND i->Y() >= Sy) AND
			(i->X() < X OR i->Y() < Y))
		{
			X = i->X();
			Y = i->Y();
			Best = i;
		}
	}

	return Best;
}

GRegion *GView::FindLargestEdge(GComplexRegion &r, int Edge)
{
	GRegion *Best = 0;

	for (GRegion *i = r.First(); i; i = r.Next())
	{
		if (NOT Best)
		{
			Best = i;
		}

		if ((Edge & GV_EDGE_TOP) AND (i->y1 < Best->y1) OR
			(Edge & GV_EDGE_RIGHT) AND (i->x2 > Best->x2) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 > Best->y2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 < Best->x1))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_TOP) AND (i->y1 == Best->y1) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 == Best->y2)) AND
			(i->X() > Best->X()))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_RIGHT) AND (i->x2 == Best->x2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 == Best->x1)) AND
			(i->Y() > Best->Y()))
		{
			Best = i;
		}
	}

	return Best;
}

int GView::GetCtrlValue(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Value() : -1;
}

void GView::SetCtrlValue(int Id, int i)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Value(i);
	}
}

char *GView::GetCtrlName(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Name() : 0;
}

void GView::SetCtrlName(int Id, char *s)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Name(s);
	}
}

bool GView::GetCtrlEnabled(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Enabled() : 0;
}

void GView::SetCtrlEnabled(int Id, bool Enabled)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Enabled(Enabled);
	}
}

void GView::SetId(int i)
{
	CtrlId = i;
}

bool GView::Sunken()
{
	return GFlags::Sunken();
}

void GView::Sunken(bool i)
{
	GFlags::Sunken(i);
}

bool GView::Flat()
{
	return GFlags::Flat();
}

void GView::Flat(bool i)
{
	GFlags::Flat(i);
}

bool GView::Raised()
{
	return GFlags::Raised();
}

void GView::Raised(bool i)
{
	GFlags::Raised(i);
}

void GView::Quit(bool DontDelete)
{
	
}

bool GView::AttachChildren()
{
	bool Status = true;

	for (GView *c=Children.First(); c; c=Children.Next())
	{
		Status |= c->Attach(this);
	}

	return Status;
}

GRegion &GView::GetClient()
{
	static GRegion Client;
	Client.ZOff(Pos.X()-1, Pos.Y()-1);
	if (Sunken() OR Raised()) Client.Size(_BorderSize, _BorderSize);
	
	return Client;
}

GFont *GView::GetFont()
{
	return 0;
}

void GView::SetFont(GFont *Fnt)
{
}

bool GView::IsOver(GMouse &m)
{
	return	(m.x >= 0) AND
			(m.y >= 0) AND
			(m.x < Pos.X()) AND
			(m.y < Pos.Y());
}

GView *GView::WindowFromPoint(int x, int y)
{
	for (GView *c = Children.First(); c; c = Children.Next())
	{
		if (c->Pos.Overlap(x, y))
		{
			return c;
		}
	}

	if (x >= 0 AND y >= 0 AND x < Pos.X() AND y < Pos.Y())
	{
		return this;
	}

	return NULL;
}

void GView::OnPosChange()
{
}

bool GView::SetPos(GRegion &p, bool Repaint)
{
	Pos = p;

	{
		ALock l;
		if (l.Lock(this))
		{
			_View->SetFrame(Pos);
		}
	}
	
	return true;
}

bool GView::Invalidate(GRegion *r, bool Repaint)
{
	ALock l;
	if (l.Lock(this))
	{
		if (r)
		{
			_View->Invalidate(*r, true);
		}
		else
		{
			_View->Invalidate();
		}

		if (Repaint)
		{
			_View->Flush();
		}
	}

	return false;
}

bool GView::InvalidateFrame(GRegion *r, bool Repaint)
{
	return Invalidate(r, Repaint);
}

void GView::SetPulse(int Length)
{
}

bool GView::InThread()
{
	return true;
}

bool GView::PostEvent(int Cmd, int a, int b)
{
	bool Status = false;

	if (GetWindow())
	{
		GMessage m = CreateMsg(Cmd, a, b);
		GetWindow()->PostMessage(&m);
		Status = true;
	}

	return Status;
}

int GView::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case WM_CHANGE:
		{
			GView *Ctrl = dynamic_cast<GView*>((GView*) MsgA(Msg));
			if (Ctrl)
			{
				OnNotify(Ctrl, MsgB(Msg));
			}
			break;
		}
		case WM_COMMAND:
		{
			int32 a=MsgA(Msg);
			return OnCommand(a&0xFFFF, a>>16, (OsView) MsgB(Msg));
		}
		default:
		{
			break;
		}
	}

	return 0;
}

int GView::OnNotify(GView *Ctrl, int Flags)
{
	if (Ctrl AND Parent)
	{
		// default behaviour is just to pass the 
		// notification up to the parent
		return Parent->OnNotify(Ctrl, Flags);
	}

	return 0;
}

bool GView::Name(char *n)
{
	bool Status = GObject::Name(n);

	ALock l;
	if (l.Lock(this))
	{
		os::Control *c = dynamic_cast<os::Control*>(_View);
		if (c)
		{
			std::string s = n;
			c->SetLabel(s);
		}
	}
	else
	{
		Status = false;
	}

	return Status;
}

char *GView::Name()
{
	return GObject::Name();
}

bool GView::Enabled()
{
	return TestFlag(WndFlags, GWF_ENABLED);
}

void GView::Enabled(bool e)
{
	GFlags::Enabled(e);
	Invalidate();

	ALock l;
	if (l.Lock(this))
	{
		os::Control *c = dynamic_cast<os::Control*>(_View);
		if (c)
		{
			c->SetEnable(e);
		}
	}
}

bool GView::Visible()
{
	return GFlags::Visible();
}

void GView::Visible(bool v)
{
	bool Change = GFlags::Visible() ^ v;

	GFlags::Visible(v);

	if (Change)
	{
		ALock l;
		if (l.Lock(this))
		{
			_View->Show(v);
		}
	}
}

bool GView::Focus()
{
	bool f = TestFlag(WndFlags, GWF_FOCUS);

	ALock l;
	if (l.Lock(this))
	{
		f = _View->HasFocus();
	}

	return f;
}

void GView::Focus(bool f)
{
	GFlags::Focus(f);

	ALock l;
	if (l.Lock(this))
	{
		_View->MakeFocus(f);
	}
}

bool GView::IsCapturing()
{
	return false;
}

bool GView::Capture(bool c)
{
	return false;
}

bool GView::GetMouse(GMouse &m, bool ScreenCoords)
{
	ALock l;
	if (l.Lock(this))
	{
		os::Point Pos;
		uint32 Buttons;

		_View->GetMouse(&Pos, &Buttons);
		
		m.x = (int) Pos.x;
		m.y = (int) Pos.y;
		// FIXME buttons...
	}

	return false;
}

bool GView::IsAttached()
{
	bool Status = false;

	ALock l;
	if (l.Lock(this))
	{
		Status = _View->GetParent() != 0;
	}

	return Status;
}

bool GView::Attach(GView *parent)
{
	bool Status = false;

printf("\tGView::Attach(...) start\n");

	Parent = parent;
	if (Parent)
	{
printf("\tGView::Attach(...) _View=%p\n", _View);
		if (_View)
		{
			// Detach from whatever..
printf("\tGView::Attach(...) _View->Parent=%p\n", _View->GetParent());
			if (_View->GetParent())
			{
				_View->RemoveThis();
			}

			// Attech to new parent..
			GWindow *w = dynamic_cast<GWindow*>(Parent);

printf("\tGView::Attach(...) w=%p\n", w);

			if (w)
			{
printf("\tGView::Attach(...) w->Wnd=%p\n", w->Wnd);
				if (w->Wnd)
				{
					// Add view to os::Window object
					w->Wnd->AddChild(_View);
					Status = true;
				}
			}
			else
			{
printf("\tGView::Attach(...) Parent->_View=%p\n", Parent->_View);
				if (Parent->_View)
				{
					// Add view to another os::View object
					Parent->_View->AddChild(_View);
printf("\tGView::Attach(...) _View->Parent=%p\n", _View->GetParent());
					Status = true;
				}
			}

			ALock l;
			if (l.Lock(this))
			{
printf("\tGView::Attach(...) locked for set pos/name\n");
				SetPos(Pos);
				Name(Name());
			}
		}

		if (NOT Parent->Children.HasItem(this))
		{
printf("\tGView::Attach(...) into parent list\n");
			Parent->Children.Insert(this);
			Parent->OnChildrenChanged(this, true);
		}
	}

printf("\tGView::Attach(...) end\n");

	return Status;
}

bool GView::Detach()
{
	bool Status = false;
	if (Parent)
	{
		if (_View)
		{
			// Detach from whatever..
			_View->RemoveThis();
		}

		Parent->Children.Delete(this);
		Parent->OnChildrenChanged(this, false);

		Status = true;
	}
	return Status;
}

GView *GView::FindControl(OsView hCtrl)
{
	if (0 == hCtrl)
	{
		return this;
	}

	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(hCtrl);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

GView *GView::FindControl(int Id)
{
	if (GetId() == Id)
	{
		return this;
	}

	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(Id);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

#define MIN_WINDOW_POS	24

void GView::MoveOnScreen()
{
	GRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);

	if (Pos.x2 >= Screen.X())
		Pos.Offset(Screen.X() - Pos.x2, 0);
	if (Pos.y2 >= Screen.Y())
		Pos.Offset(0, Screen.Y() - Pos.y2);
	if (Pos.x1 < 0)
		Pos.Offset(-Pos.x1, 0);
	if (Pos.y1 < MIN_WINDOW_POS)
		Pos.Offset(0, MIN_WINDOW_POS-Pos.y1);
}

void GView::MoveToCenter()
{
	GRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	Pos.Offset(-Pos.x1, -Pos.y1);
	Pos.Offset((Screen.X() - Pos.X()) / 2, (Screen.Y() - Pos.Y()) / 2);
}

void GView::MoveToMouse()
{
	GMouse m;
	if (GetMouse(m, true))
	{
		Pos.Offset(-Pos.x1, -Pos.y1);
		Pos.Offset(m.x-(Pos.X()/2), m.y-(Pos.Y()/2));
	}
}

///////////////////////////////////////////////////////////////////
bool LgiIsMounted(char *Name)
{
	return false;
}

bool LgiMountVolume(char *Name)
{
	return false;
}











