/*hdr
**      FILE:           GWidgets.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           20/10/2001
**      DESCRIPTION:    Dialog components
**
**      Copyright (C) 2001 Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "Lgi.h"
#include <gui/textview.h>
#include <gui/checkbox.h>
#include <gui/stringview.h>
#include <gui/frameview.h>

#define TICKS_PER_SECOND						1000000
#define DOUBLE_CLICK_TIMEOUT				(TICKS_PER_SECOND/3)

#define DefaultViewColour()					

///////////////////////////////////////////////////////////////////////////////////////////
GDialog::GDialog() :
	GView( NEW(AView<os::View>(this)) )
#ifdef LGI_RES
	, ResObject(Res_Dialog)
#endif
{
}

GDialog::~GDialog()
{
}

bool GDialog::LoadFromResource(char *Resource)
{
	char n[256];
	bool Status = GLgiRes::LoadFromResource(Resource, Children, &Pos, n);
	if (Status)
	{
		Name(n);
	}
	return Status;
}

int GDialog::DoModal()
{
	return 0;
}

int GDialog::DoModeless()
{
	return 0;
}

int GDialog::OnEvent(GMessage *Msg)
{
	int Status = 0;

	return Status;
}

void GDialog::EndModal(int Code)
{
}

void GDialog::EndModeless(int Code)
{
	OnDestroy();
}

///////////////////////////////////////////////////////////////////////////////////////////
GControl::GControl(OsView view) : GView(view)
{
	Pos.ZOff(10, 10);
	SetId(0);
	Sunken(true);
}

GControl::~GControl()
{
}

int GControl::OnEvent(GMessage *Msg)
{
	return 0;
}

GdcPt2 GControl::SizeOfStr(char *Str)
{
	GdcPt2 Pt(0, 0);
	if (Str)
	{
	}

	return Pt;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Button
GButton::GButton(int id, int x, int y, int cx, int cy, char *name) :
	GControl( NEW(AView<os::Button>(this, "GButton", name, new os::Message(WM_CHANGE) )) )
#ifdef LGI_RES
	, ResObject(Res_Button)
#endif
{
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
}

GButton::~GButton()
{
}

bool GButton::Default()
{
	return false;
}

void GButton::Default(bool b)
{
}

int GButton::OnEvent(GMessage *Msg)
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Edit
GEdit::GEdit(int id, int x, int y, int cx, int cy, char *name) :
	GControl( NEW(AView<os::TextView>(this, "GEdit", name)) )
#ifdef LGI_RES
	, ResObject(Res_EditBox)
#endif
{
	GdcPt2 Size = SizeOfStr((name)?name:(char*)"A");
	if (cx < 0) cx = Size.x + 6;
	if (cy < 0) cy = Size.y + 4;

	MultiLineEdit = false;
	PasswordEdit = false;
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	SetPos(Pos);
	Name(name);
}

GEdit::~GEdit()
{
}

void GEdit::Password(bool i)
{
	PasswordEdit = i;
}

void GEdit::MultiLine(bool m)
{
	MultiLineEdit = m;
}

void GEdit::Value(int i)
{
	char Str[32];
	sprintf(Str, "%i", i);
	Name(Str);
}

int GEdit::Value()
{
	char *n = Name();
	return (n) ? atoi(n) : 0;
}

int GEdit::OnEvent(GMessage *Msg)
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Check box
GCheckBox::GCheckBox(int id, int x, int y, int cx, int cy, char *name, int InitState) :
	GControl( NEW(AView<os::CheckBox>(this, "GCheckBox", name, new os::Message(WM_CHANGE) )) )
#ifdef LGI_RES
	, ResObject(Res_CheckBox)
#endif
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	Val = InitState;
	Three = FALSE;
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	if (name) Name(name);
}

GCheckBox::~GCheckBox()
{
}

int GCheckBox::Value()
{
	return Val;
}

void GCheckBox::Value(int b)
{
	Val = b;
}

int GCheckBox::OnEvent(GMessage *Msg)
{
	return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
// Text
GText::GText(int id, int x, int y, int cx, int cy, char *name) :
	GControl( NEW(AView<os::StringView>(this, "GText", name)) )
#ifdef LGI_RES
	, ResObject(Res_StaticText)
#endif
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x;
	if (cy < 0) cy = Size.y;

	GText::Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	DefaultViewColour();
}

GText::~GText()
{
}

int GText::OnEvent(GMessage *Msg)
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio group
int GRadioGroup::NextId = 20000;

GRadioGroup::GRadioGroup(int id, int x, int y, int cx, int cy, char *name, int Init) :
	GControl( NEW(AView<os::FrameView>(this, std::string("GRadioGroup"), std::string(name))) )
#ifdef LGI_RES
	, ResObject(Res_Group)
#endif
{
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	InitialValue = Init;
	if (name) Name(name);
}

GRadioGroup::~GRadioGroup()
{
}

void GRadioGroup::OnCreate()
{
	AttachChildren();
	Value(InitialValue);
}

int GRadioGroup::Value()
{
	return InitialValue;
}

void GRadioGroup::Value(int Which)
{
	InitialValue = Which;
}

int GRadioGroup::OnNotify(GView *Ctrl, int Flags)
{
	return 0;
}

int GRadioGroup::OnEvent(GMessage *Msg)
{
	return 0;
}

GRadioButton *GRadioGroup::Append(int x, int y, char *name)
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio button
#include <gui/radiobutton.h>

GRadioButton::GRadioButton(int id, int x, int y, int cx, int cy, char *name, bool first) :
	GControl( NEW(AView<os::RadioButton>(this, "GRadioButton", name, new os::Message(WM_CHANGE))) )
#ifdef LGI_RES
	, ResObject(Res_RadioBox)
#endif
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	if (name) Name(name);
}

GRadioButton::~GRadioButton()
{
}

int GRadioButton::Value()
{
	int Status = 0;
	return Status;
}

void GRadioButton::Value(int i)
{
}

int GRadioButton::OnEvent(GMessage *Msg)
{
	return 0;
}

//////////////////////////////////////////////////////////////////////////////////
// Slider control
GSlider::GSlider(int id, int x, int y, int cx, int cy, char *name, bool vert) :
	GControl( NEW(AView<os::Control>(this, std::string("GSlider"), std::string(name), new os::Message(WM_CHANGE))) )
#ifdef LGI_RES
	, ResObject(Res_Slider)
#endif
{
	SetId(id);
	if (vert)
	{
		Pos.Set(x, y, x+13, y+cy);
	}
	else
	{
		Pos.Set(x, y, x+cx, y+13);
	}
	
	Name(name);
	Vertical = vert;
	Min = 0;
	Max = 1;
	Val = 0;
	
	Thumb.ZOff(-1, -1);
	Track = false;
}

GSlider::~GSlider()
{
}

void GSlider::Value(int i)
{
	Val = i;
	Invalidate();
}

int GSlider::Value()
{
	return Val;
}

void GSlider::GetLimits(int &min, int &max)
{
	min = Min;
	max = Max;
}

void GSlider::SetLimits(int min, int max)
{
	Min = min;
	Max = max;
	if (Val < Min) Val = Min;
	if (Val > Max) Val = Max;
}

int GSlider::OnEvent(GMessage *Msg)
{
	return 0;
}

/*
#define SLIDER_BAR_SIZE				2
#define SLIDER_THUMB_WIDTH			10
#define SLIDER_THUMB_HEIGHT			6

void GSlider::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);

	// fill area
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&r);

	// draw tray
	int Mid = (Vertical) ? r.X()/2 : r.Y()/2;
	GRegion Bar;
	if (Vertical)
	{
		Bar.ZOff(SLIDER_BAR_SIZE, Y()-1);
		Bar.Offset(Mid-(Bar.X()/2), 0);
	}
	else
	{
		Bar.ZOff(X()-1, SLIDER_BAR_SIZE);
		Bar.Offset(0, Mid-(Bar.Y()/2));
	}
	
	LgiThinBorder(pDC, Bar, SUNKEN);
	pDC->Colour(0);
	pDC->Line(Bar.x1, Bar.y1, Bar.x2, Bar.y2);
	
	// draw thumb
	int Length = ((Vertical) ? Y() : X()) - SLIDER_THUMB_HEIGHT - 1;
	int Pos = (Max != Min) ? Val * Length / (Max - Min) : 0;
	if (Vertical)
	{
		Thumb.ZOff(SLIDER_THUMB_WIDTH, SLIDER_THUMB_HEIGHT);
		Thumb.Offset(Mid - (Thumb.X()/2), Pos);
	}
	else
	{
		Thumb.ZOff(SLIDER_THUMB_HEIGHT, SLIDER_THUMB_WIDTH);
		Thumb.Offset(Pos, Mid - (Thumb.Y()/2));
	}
	
	GRegion t = Thumb;
	LgiWideBorder(pDC, t, RAISED);
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&t);
}

void GSlider::OnMouseClick(GMouse &m)
{
	if (m.Down())
	{
		if (Thumb.Overlap(m.x, m.y))
		{
			Tx = m.x - Thumb.x1;
			Ty = m.y - Thumb.y1;
			Track = true;
		}
	}
	else
	{
		Track = false;
	}
}

void GSlider::OnMouseMove(GMouse &m)
{
	if (Track)
	{
		int Length = ((Vertical) ? Y() : X()) - SLIDER_THUMB_HEIGHT - 1;
		int NewPos = (Vertical) ? m.y - Ty : m.x - Tx;
		if (NewPos < 0) NewPos = 0;
		if (NewPos > Length) NewPos = Length;
		int NewVal = (Min != Max) ? (NewPos*(Max-Min)) / Length : 0;
		if (NewVal != Val)
		{
			Val = NewVal;
			Invalidate();
			
			if (GetParent())
			{
				GetParent()->OnNotify(this, Val);
			}
		}
	}
}
*/

//////////////////////////////////////////////////////////////////////////////////
GBitmap::GBitmap(int id, int x, int y, char *FileName, bool Async) :
	GControl( NEW(AView<os::Control>(	this,
											std::string("GBitmap"),
											std::string(""),
											new os::Message(WM_CHANGE))) )
#ifdef LGI_RES
	, ResObject(Res_Bitmap)
#endif
{
	SetId(id);
	Pos.ZOff(1, 1);
	Pos.Offset(x, y);
	pDC = 0;
	DefaultViewColour();

	int Opt = GdcD->SetOption(GDC_IMPORT_8_AS_16, true);
	pDC = LoadDC(FileName);
	GdcD->SetOption(GDC_IMPORT_8_AS_16, Opt);

	if (pDC)
	{
		Pos.Dimension(pDC->X()+4, pDC->Y()+4);
	}
}

GBitmap::~GBitmap()
{
	DeleteObj(pDC);
}

int GBitmap::OnEvent(GMessage *Msg)
{
	return 0;
}

void GBitmap::OnPaint(GSurface *pScreen)
{
	GRegion a(0, 0, Pos.X()-1, Pos.Y()-1);
	LgiWideBorder(pScreen, a, SUNKEN);
	if (pDC)
	{
		pScreen->Blt(2, 2, pDC);
	}
	else
	{
		pScreen->Colour(CBit(pScreen->GetBits(), Rgb24(255, 255, 255)));
		pScreen->Rectangle(&a);
	}
}

void GBitmap::OnMouseClick(GMouse &m)
{
	if (NOT m.Down() AND GetParent())
	{
		GDialog *p = (GDialog*) GetParent();
		p->OnNotify(this, 0);
	}
}

void GBitmap::SetDC(GSurface *pNewDC)
{
	DeleteObj(pDC);
	pDC = pNewDC;
}

GSurface *GBitmap::GetDC()
{
	return pDC;
}

///////////////////////////////////////////////////////////////////////////////////////////
GItemContainer::GItemContainer()
{
	Flags = 0;
	ImageList = 0;
}

GItemContainer::~GItemContainer()
{
	if (OwnList())
	{
		DeleteObj(ImageList);
	}
	else
	{
		ImageList = 0;
	}
}

bool GItemContainer::SetImageList(GImageList *list, bool Own)
{
	ImageList = list;
	OwnList(Own);
	AskImage(ImageList != NULL);
	return ImageList != NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////
GProgress::GProgress(int id, int x, int y, int cx, int cy, char *name) :
	GControl( NEW(AView<os::Control>(	this,
											std::string("GProgress"),
											std::string(""),
											new os::Message(WM_CHANGE))) )
#ifdef LGI_RES
	, ResObject(Res_Progress)
#endif
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	c = Rgb24(50, 150, 255);
	DefaultViewColour();
}

GProgress::~GProgress()
{
}

void GProgress::Colour(COLOUR Col)
{
	c = Col;
	GView::Invalidate();
}

COLOUR GProgress::Colour()
{
	return c;
}

void GProgress::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);
	LgiThinBorder(pDC, r, SUNKEN);
	
	if (High != 0)
	{
		int Pos = (int) ((double) r.X() * (((double) Val - Low) / High));
		if (Pos > 0)
		{
			COLOUR High = Rgb24( (R24(c)+255)/2, (G24(c)+255)/2, (B24(c)+255)/2 );
			COLOUR Low = Rgb24( (R24(c)+0)/2, (G24(c)+0)/2, (B24(c)+0)/2 );
			GRegion p = r;

			p.x2 = p.x1 + Pos;
			r.x1 = p.x2 + 1;
			
			pDC->Colour(Low, 24);
			pDC->Line(p.x2, p.y2, p.x2, p.y1);
			pDC->Line(p.x2, p.y2, p.x1, p.y2);

			pDC->Colour(High, 24);
			pDC->Line(p.x1, p.y1, p.x2, p.y1);
			pDC->Line(p.x1, p.y1, p.x1, p.y2);

			p.Size(1, 1);
			pDC->Colour(c, 24);
			pDC->Rectangle(&p);
		}
	}

	if (r.Valid())
	{
		pDC->Colour(Rgb24(255, 255, 255), 24);
		pDC->Rectangle(&r);
	}
}

void GProgress::SetLimits(int l, int h)
{
	Low = l;
	High = h;
	GView::Invalidate();
}

void GProgress::GetLimits(int &l, int &h)
{
	l = Low;
	h = High;
}

void GProgress::Value(int v)
{
	Val = limit(v, Low, High);
	GView::Invalidate();
}

int GProgress::Value()
{
	return Val;
}

int GProgress::OnEvent(GMessage *Msg)
{
	return 0;
}


//////////////////////////////////////////////////////////////////////////////////////
#include <gui/scrollbar.h>

GScrollBar::GScrollBar() :
	GControl( NEW(AView<os::ScrollBar>(this, "GScrollBar", new os::Message(WM_CHANGE))) )
#ifdef LGI_RES
	, ResObject(Res_ScrollBar)
#endif
{
}

GScrollBar::GScrollBar(int id, int x, int y, int cx, int cy, char *name) :
	GControl( NEW(AView<os::ScrollBar>(this, "GScrollBar", new os::Message(WM_CHANGE))) )
#ifdef LGI_RES
	, ResObject(Res_ScrollBar)
#endif
{
	SetId(id);
	GRegion n(	x,
				y,
				x + ((cx > cy) ? cx : 16),
				y + ((cx <= cy) ? cy : 16));
	GView::SetPos(n);
	if (name) GView::Name(name);
	
	High = Low = PageSize = 0;
}

GScrollBar::~GScrollBar()
{
}

bool GScrollBar::Vertical()
{
	bool Status = false;
	return Status;
}

int GScrollBar::Pos()
{
	int i = 0;
	return i;
}

void GScrollBar::SetPos(int p)
{
}

void GScrollBar::Limits(int &Low, int &High)
{
}

void GScrollBar::SetLimits(int low, int high)
{
}

int GScrollBar::Page()
{
	return 0;
}

void GScrollBar::SetPage(int p)
{
}

bool GScrollBar::Attach(GView *parent)
{
	return GControl::Attach(parent);
}

int GScrollBar::OnEvent(GMessage *Msg)
{
	return FALSE;
}





