#include "Lgi.h"

#include <util/message.h>

class AWindow : public os::Window
{
	GWindow *Wnd;

public:
	AWindow(GWindow *wnd) :
		os::Window(os::Rect(100, 100, 300, 200), "Lgi GWindow", "")
	{
		Wnd = wnd;
	}

	bool OkToQuit();
	void HandleMessage(os::Message *m);
	void FrameSized(const os::Point &cDelta);
};

bool AWindow::OkToQuit(void)
{
	LgiCloseApp();
	return Wnd->OnRequestClose();
}

void AWindow::HandleMessage(os::Message *m)
{
	Wnd->OnEvent(m);
	os::Window::HandleMessage(m);
}

void AWindow::FrameSized(const os::Point &cDelta)
{
	Wnd->OnPosChange();
}

///////////////////////////////////////////////////////////////////////
GWindow::GWindow() :
	GView(0)
{
	Menu = 0;
	Wnd = NEW(AWindow(this));
	if (Wnd)
	{
printf("::GWindow(0) locked=%i\n", Wnd->IsLocked());

		GFlags::Visible(false);
		Wnd->Start();

printf("::GWindow(1) locked=%i\n", Wnd->IsLocked());
	}
}

GWindow::~GWindow()
{
	DeleteObj(Wnd);
	DeleteObj(Menu);
}

bool GWindow::Name(char *n)
{
printf("GWindow::Name(char*) before, locked=%i\n", Wnd->IsLocked());

	if (Wnd)
	{
		bool Lock = Wnd->Lock() == 0;
		Wnd->SetTitle(n?n:(char*)"");
		if (Lock) Wnd->Unlock();
	}

printf("GWindow::Name(char*) after, locked=%i\n", Wnd->IsLocked());

	return true;
}

char *GWindow::Name()
{
	char *Status = 0;

printf("GWindow::Name() before, locked=%i\n", Wnd->IsLocked());

	if (Wnd)
	{
		bool Lock = Wnd->Lock() == 0;
		Status = (char*) Wnd->GetTitle().c_str();
		if (Lock) Wnd->Unlock();
	}

printf("GWindow::Name() after, locked=%i\n", Wnd->IsLocked());

	return Status;
}

GRegion &GWindow::GetPos()
{
	static GRegion r(0, 0, 1, 1);

printf("GWindow::GetPos() before, locked=%i\n", Wnd->IsLocked());

	if (Wnd)
	{
		bool Lock = Wnd->Lock() == 0;
		r = Wnd->GetFrame();
		if (Lock) Wnd->Unlock();
	}

printf("GWindow::GetPos() after, locked=%i\n", Wnd->IsLocked());

	return r;
}

bool GWindow::SetPos(GRegion &p, bool Repaint)
{
printf("GWindow::SetPos() before, locked=%i\n", Wnd->IsLocked());

	if (Wnd)
	{
		bool Lock = Wnd->Lock() == 0;
		Wnd->SetFrame(p);
		if (Lock) Wnd->Unlock();
	}

printf("GWindow::SetPos() after, locked=%i\n", Wnd->IsLocked());

	return true;
}

void GWindow::OnChildrenChanged(GView *Wnd, bool Attaching)
{
}

void GWindow::OnCreate()
{
}

void GWindow::OnPaint(GSurface *pDC)
{
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle();
}

void GWindow::OnPosChange()
{
	GView::OnPosChange();
	Pour();
}

bool GWindow::Visible()
{
	return GFlags::Visible();
}

void GWindow::Visible(bool v)
{
printf("GWindow::Visible() before, locked=%i\n", Wnd->IsLocked());

	bool Change = GFlags::Visible() ^ v;
	GFlags::Visible(v);

	if (Change AND Wnd)
	{
		bool Lock = Wnd->Lock() == 0;
		if (v)
		{
			Wnd->Show();
			Pour();
		}
		else
		{
			Wnd->Hide();
		}	
		if (Lock) Wnd->Unlock();
	}

printf("GWindow::Visible() after, locked=%i\n", Wnd->IsLocked());
}

void GWindow::Pour()
{
printf("GWindow::Pour() before, locked=%i\n", Wnd->IsLocked());

	bool Lock = Wnd->Lock() == 0;

	GRegion r = GetClient();
	GComplexRegion Client(r);
	GComplexRegion Update;

	for (GView *w = Children.First(); w; w = Children.Next())
	{
		GRegion OldPos = w->Pos;
		Update.Union(&OldPos);

		if (w->Pour(Client))
		{
			if (NOT w->Visible())
			{
				w->Visible(true);
			}

			w->Invalidate();

			Client.Subtract(&w->Pos);
			Update.Subtract(&w->Pos);
		}
		else
		{
			// non-pourable
		}
	}

	if (Lock) Wnd->Unlock();

printf("GWindow::Pour() after, locked=%i\n", Wnd->IsLocked());
}

int GWindow::OnEvent(GMessage *m)
{
	switch (MsgCode(m))
	{
	}

	// printf("GWindow::OnEvent(%i)\n", MsgCode(m));

	return GView::OnEvent(m);
}




















