/*hdr
**	FILE:		Gdc2.h
**	AUTHOR:		Matthew Allen
**	DATE:		20/2/97
**	DESCRIPTION:	GDC v2.xx header
**
**	Copyright (C) 1997, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include <math.h>
#include "Gdc2.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////
GdcViewDC::GdcViewDC(BView *view)
{
	View = view;
	LgiAssert(View);
	
	_ClientClip = false;
	BitDepth = GdcD->GetBits();
}

GdcViewDC::~GdcViewDC()
{
	View = 0;
}

void GdcViewDC::GetOrigin(int &x, int &y)
{
	BPoint p = View->Origin();
	x = p.x;
	y = p.y;
}

void GdcViewDC::SetOrigin(int x, int y)
{
	View->SetOrigin(-x, -y);
}

GRegion GdcViewDC::ClipRgn()
{
	BRegion r;
	View->GetClippingRegion(&r);
	BRect rc = r.Frame();
	return GRegion(rc);
}

void GdcViewDC::_SetClient(GRegion *c)
{
	_ClientClip = c != 0;
	if (c)
	{
		_Client = *c;

		BRegion br;
		br.Set(_Client);
		View->ConstrainClippingRegion(&br);
	}
	else
	{
		View->ConstrainClippingRegion(0);
	}
}

GRegion GdcViewDC::ClipRgn(GRegion *Rgn)
{
	if (Rgn)
	{
		GRegion r = *Rgn;
		if (_ClientClip)
		{
			r.Bound(&_Client);
		}

		BRegion br;
		br.Set(r);
		View->ConstrainClippingRegion(&br);
	}
	else
	{
		if (_ClientClip)
		{
			BRegion br;
			br.Set(_Client);
			View->ConstrainClippingRegion(&br);
		}
		else
		{
			View->ConstrainClippingRegion(NULL);
		}
	}
}

COLOUR GdcViewDC::Colour()
{
	rgb_color Rgb = View->HighColor();
	return CBit(GetBits(), Rgb24(Rgb.red, Rgb.green, Rgb.blue), 24);
}

COLOUR GdcViewDC::Colour(COLOUR c, int Bits)
{
	COLOUR Prev = Colour();

	if (Bits < 1) Bits = GetBits();
	c = CBit(24, c, Bits);
	
	rgb_color Rgb;
	Rgb.red = R24(c);
	Rgb.green = G24(c);
	Rgb.blue = B24(c);
	View->SetHighColor(Rgb);
	
	return Prev;
}

int GdcViewDC::Op()
{
	int Mode = View->DrawingMode();
	switch (Mode)
	{
		case B_OP_ERASE:
		{
			return GDC_AND;
			break;
		}
		case B_OP_OVER:
		{
			return GDC_OR;
			break;
		}
		case B_OP_INVERT:
		{
			return GDC_XOR;
			break;
		}
	}
	return GDC_SET;
}

int GdcViewDC::Op(int NewOp)
{
	int Prev = Op();
	drawing_mode Mode = B_OP_COPY;
	switch (NewOp)
	{
		case GDC_AND:
		{
			Mode = B_OP_ERASE;
			break;
		}
		case GDC_OR:
		{
			Mode = B_OP_OVER;
			break;
		}
		case GDC_XOR:
		{
			Mode = B_OP_INVERT;
			break;
		}
	}
	View->SetDrawingMode(Mode);
	return Prev;
}

int GdcViewDC::X()
{
	BRect r = View->Bounds();
	return r.right - r.left + 1;
}

int GdcViewDC::Y()
{
	BRect r = View->Bounds();
	return r.bottom - r.top + 1;
}

int GdcViewDC::GetBits()
{
	return BitDepth;
}

int GdcViewDC::GetFlags()
{
	return Flags;
}

void GdcViewDC::Set(int x, int y)
{
	View->StrokeLine(BPoint(x, y), BPoint(x, y));
}

COLOUR GdcViewDC::Get(int x, int y)
{
	// you can't get the value of a pixel!
	// fix this Be!!!!!!!!!!
	return 0;
}

void GdcViewDC::HLine(int x1, int x2, int y)
{
	View->StrokeLine(BPoint(x1, y), BPoint(x2, y));
}

void GdcViewDC::VLine(int x, int y1, int y2)
{
	View->StrokeLine(BPoint(x, y1), BPoint(x, y2));
}

void GdcViewDC::Line(int x1, int y1, int x2, int y2)
{
	View->StrokeLine(BPoint(x1, y1), BPoint(x2, y2));
}

void GdcViewDC::Circle(double cx, double cy, double radius)
{
	View->StrokeEllipse(BPoint(cx, cy), radius, radius);
}

void GdcViewDC::FilledCircle(double cx, double cy, double radius)
{
	View->FillEllipse(BPoint(cx, cy), radius, radius);
}

void GdcViewDC::Arc(double cx, double cy, double radius, double start, double end)
{
	View->StrokeArc(BPoint(cx, cy), radius, radius, start, end - start);
}

void GdcViewDC::FilledArc(double cx, double cy, double radius, double start, double end)
{
	View->FillArc(BPoint(cx, cy), radius, radius, start, end - start);
}

void GdcViewDC::Ellipse(double cx, double cy, double x, double y)
{
	View->StrokeEllipse(BPoint(cx, cy), x, y);
}

void GdcViewDC::FilledEllipse(double cx, double cy, double x, double y)
{
	View->FillEllipse(BPoint(cx, cy), x, y);
}

void GdcViewDC::Box(int x1, int y1, int x2, int y2)
{
	View->StrokeRect(BRect(x1, y1, x2, y2));
}

void GdcViewDC::Box(GRegion *a)
{
	BRect r;
	if (a)
	{
		r = *a;
	}
	else
	{
		r.left = r.top = 0;
		r.right = X()-1;
		r.bottom = Y()-1;
	}
	
	View->StrokeRect(r);
}

void GdcViewDC::Rectangle(int x1, int y1, int x2, int y2)
{
	View->FillRect(BRect(x1, y1, x2, y2));
}

void GdcViewDC::Rectangle(GRegion *a)
{
	BRect r;
	if (a)
	{
		r = *a;
	}
	else
	{
		r.left = r.top = 0;
		r.right = X()-1;
		r.bottom = Y()-1;
	}
	
	View->FillRect(r);
}

void GdcViewDC::Blt(int x, int y, GSurface *Src, GRegion *a)
{
	if (Src)
	{
		if (Src->IsScreen())
		{
			// screen to screen Blt
		}
		else
		{
			// memory to screen Blt
			BBitmap *Bmp = Src->GetBitmap();
			if (Bmp)
			{
				BRect S;
				if (a)
				{
					S = *a;
				}
				else
				{
					S.left = S.top = 0;
					S.right = Src->X() - 1;
					S.bottom = Src->Y() - 1;
				}
				
				BRect D(x, y, x+S.Width(), y+S.Height());
				
				View->DrawBitmap(Bmp, S, D);
			}
		}
	}
}

void GdcViewDC::StretchBlt(GRegion *d, GSurface *Src, GRegion *s)
{
}

void GdcViewDC::Polygon(int Points, GdcPt2 *Data)
{
	if (Points > 0 AND Data)
	{
		BPoint *p = NEW(BPoint[Points]);
		if (p)
		{
			for (int i=0; i<Points; i++)
			{
				p[i].x = Data[i].x;
				p[i].y = Data[i].y;
			}
			
			View->StrokePolygon(p, Points);
			DeleteArray(p);
		}
	}
}

void GdcViewDC::Text(GdcFont *pFont, int x, int y, char *Str, int Len)
{
}

void GdcViewDC::Bezier(int Threshold, GdcPt2 *Pt)
{
}

void GdcViewDC::FloodFill(int x, int y, int Mode, COLOUR Border, GRegion *Bounds)
{
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
GMemDC::GMemDC()
{
	Bmp = 0;
}

GMemDC::GMemDC(GSurface *pDC)
{
	Bmp = 0;

	if (pDC AND Create(pDC->X(), pDC->Y(), pDC->GetBits()))
	{
		Blt(0, 0, pDC);
		if (pDC->Palette())
		{
			GdcPalette *Pal = NEW(GdcPalette(pDC->Palette()));
			if (Pal)
			{
				Palette(Pal, TRUE);
			}
		}
	}
}

GMemDC::~GMemDC()
{
	DeleteObj(Bmp);
}

bool GMemDC::Create(int x, int y, int Bits, int LineLen, bool KeepData)
{
	bool Status = FALSE;
	
	if (x < 1) x = 1;
	if (y < 1) y = 1;
	
	color_space Mode = B_RGB24;
	switch (Bits)
	{
		case 8:
		{
			Mode = B_CMAP8;
			break;
		}
		case 15:
		{
			Mode = B_RGB15;
			break;
		}
		case 16:
		{
			Mode = B_RGB16;
			break;
		}
		case 24:
		{
			Bits = 32;
			Mode = B_RGB32;
			break;
		}
		case 32:
		{
			Mode = B_RGB32;
			break;
		}
	}

	DeleteObj(Bmp);
	Bmp = NEW(BBitmap(BRect(0, 0, x-1, y-1), Mode, true, true));
	if (Bmp)
	{
		pMem = NEW(GdcSurface);
		if (pMem)
		{
			pMem->Base = (uchar*) Bmp->Bits();
			pMem->x = x;
			pMem->y = y;
			pMem->Bits = Bits;
			pMem->Line = Bmp->BytesPerRow();
			pMem->Flags = 0;
			
			// Msg("Base: %p x: %i y: %i Bits: %i Line: %i", pMem->Base, pMem->x, pMem->y, pMem->Bits, pMem->Line);
			Status = TRUE;

			int NewOp = (pApp) ? Op() : GDC_SET;

			if (	(Flags & GDC_OWN_APPLICATOR) AND
				NOT (Flags & GDC_CACHED_APPLICATOR))
			{
				DeleteObj(pApp);
			}

			for (int i=0; i<GDC_CACHE_SIZE; i++)
			{
				DeleteObj(pAppCache[i]);
			}

			if (NewOp < GDC_CACHE_SIZE AND NOT DrawOnAlpha())
			{
				pApp = (pAppCache[NewOp]) ? pAppCache[NewOp] : pAppCache[NewOp] = CreateApplicator(NewOp);
				Flags &= ~GDC_OWN_APPLICATOR;
				Flags |= GDC_CACHED_APPLICATOR;
			}
			else
			{
				pApp = CreateApplicator(NewOp);
				Flags &= ~GDC_CACHED_APPLICATOR;
				Flags |= GDC_OWN_APPLICATOR;
			}

			Clip.ZOff(X()-1, Y()-1);
		}
	}

	return Status;
}

void GMemDC::Blt(int x, int y, GSurface *Src, GRegion *a)
{
	if (Src)
	{
		if (Src->IsScreen())
		{
			// screen to memory Blt
		}
		else
		{
			// memory to memory Blt
			GSurface::Blt(x, y, Src, a);
		}
	}
}

void GMemDC::StretchBlt(GRegion *d, GSurface *Src, GRegion *s)
{
}

void GMemDC::HLine(int x1, int x2, int y, COLOUR a, COLOUR b)
{
	if (x1 > x2) LgiSwap(x1, x2);

	if (x1 < Clip.x1) x1 = Clip.x1;
	if (x2 > Clip.x2) x2 = Clip.x2;
	if (	x1 <= x2 AND
		y >= Clip.y1 AND
		y <= Clip.y2)
	{
		COLOUR Prev = pApp->c;

		pApp->SetPtr(x1, y);
		for (; x1 <= x2; x1++)
		{
			if (x1 & 1)
			{
				pApp->c = a;
			}
			else
			{
				pApp->c = b;
			}

			pApp->Set();
			pApp->IncX();
		}

		pApp->c = Prev;
	}
}

void GMemDC::VLine(int x, int y1, int y2, COLOUR a, COLOUR b)
{
	if (y1 > y2) LgiSwap(y1, y2);
	
	if (y1 < Clip.y1) y1 = Clip.y1;
	if (y2 > Clip.y2) y2 = Clip.y2;
	if (y1 <= y2 AND
		x >= Clip.x1 AND
		x <= Clip.x2)
	{
		COLOUR Prev = pApp->c;

		pApp->SetPtr(x, y1);
		for (; y1 <= y2; y1++)
		{
			if (y1 & 1)
			{
				pApp->c = a;
			}
			else
			{
				pApp->c = b;
			}

			pApp->Set();
			pApp->IncY();
		}

		pApp->c = Prev;
	}
}
