/*hdr
**	FILE:		File.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		11/7/95
**	DESCRIPTION:	The new file subsystem
**
**	Copyright (C) 1995, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

/****************************** Includes ************************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>
#include <unistd.h>

#include "GFile.h"

/****************************** Defines *************************************************************************************/

#define FILEDEBUG

#define FLOPPY_360K				0x0001
#define FLOPPY_720K				0x0002
#define FLOPPY_1_2M				0x0004
#define FLOPPY_1_4M				0x0008
#define FLOPPY_5_25				(FLOPPY_360K | FLOPPY_1_2M)
#define FLOPPY_3_5				(FLOPPY_720K | FLOPPY_1_4M)

/****************************** Globals *************************************************************************************/

FileDevice *FileDevice::Instance = 0;

/****************************** Helper Functions ****************************************************************************/

extern void GetFlagsStr(char *flagstr, short flags);
extern int LeapYear(int year);

int SizeofStr(char *s)
{
	return sizeof(ulong) + ((s) ? strlen(s) : 0);
}

void WriteStr(GFile &f, char *s)
{
	ulong Len = (s) ? strlen(s) : 0;
	f << Len;
	if (Len > 0)
	{
		f.Write(s, Len);
	}
}

char *ReadStr(GFile &f)
{
	ulong Len;
	f >> Len;
	char *s = 0;

	if (Len > 0)
	{
		s = NEW(char[Len+1]);
		if (s)
		{
			f.Read(s, Len);
			s[Len] = 0;
		}
		else
		{
			f.Seek(Len, SEEK_CUR);
		}
	}

	return s;
}

bool ResolveShortcut(char *LinkFile, char *Path, int Length) 
{
	bool Status = FALSE;

	if (LinkFile AND Path)
	{
		BSymLink Link(LinkFile);
		if (Link.InitCheck())
		{
			Link.ReadLink(Path, Length);
		}
		else
		{
			strcpy(Path, LinkFile);
		}
	}
	
	return Status;
}

int FileSize(char *FileName)
{
	GFile F;
	if (FileName AND F.Open(FileName, O_READ))
	{
		return F.GetSize();
	}
	return 0;
}

bool FileExists(char *Name)
{
	GFile f;
	return f.Open(Name, O_READ);
}

bool TrimDir(char *Path)
{
	if (Path)
	{
		char *p = Path + (strlen(Path) - 1);
		if (*p == DIR_CHAR)
		{
			*p = 0;
			p--;
		}
		
		for (; p > Path; p--)
		{
			if (*p == DIR_CHAR)
			{
				*p = 0;
				return true;
			}
		}

		p[1] = 0;
		return true;
	}
	return false;
}

/****************************** Classes *************************************************************************************/

FileDevice::FileDevice()
{
	Instance = this;
	FillDiskList();
	getcwd(CurDir, sizeof(CurDir));
}

FileDevice::~FileDevice()
{
	DeleteArray(DriveList);
}

int FileDevice::FillDiskList()
{
	NumDrive = 0;
	DriveList = NEW(DriveInfo[26]);
	if (DriveList)
	{
		memset(DriveList, 0, sizeof(DriveInfo[26]));

		BVolumeRoster Info;
		Info.Rewind();
		BVolume Volume;
		
		for (int i=0; i<26; i++)
		{
			if (Info.GetNextVolume(&Volume) == B_NO_ERROR)
			{
				DriveInfo *VInfo = DriveList + i;

				if (Volume.IsPersistent())
				{
					VInfo->Type = DIF_HARDDISK;
					
					if (Volume.IsReadOnly())
					{
						VInfo->Type = DIF_CDROM;
					}

					if (Volume.IsRemovable())
					{
						VInfo->Type = DIF_3_5FLOPPY;
					}
					
					if (Volume.IsShared())
					{
						VInfo->Type = DIF_NET;
					}
				}
				else
				{
					VInfo->Type = DIF_RAM;
				}

				Volume.GetName(VInfo->Name);
				VInfo->TotalSpace = Volume.Capacity();
				VInfo->FreeSpace = Volume.FreeBytes();
				NumDrive++;
			}
		}
	}

	return TRUE;
}

VDirectory *FileDevice::GetDir(char *DirName)
{
	VBeDir *d = NEW(VBeDir);
	if (d)
	{
		// d->First(DirName, 0, 0);
	}
	return d;
}

bool FileDevice::DeleteFile(char *FileName)
{
	if (FileName)
	{
		return unlink(FileName) == 0;
	}
	return false;
}

bool FileDevice::SetCurrentDirectory(char *PathName)
{
	if (PathName)
	{
		for (char *p = PathName; *p; )
		{
			int Dots = 0;
			if (*p == '.')
			{
				Dots++;
				p++;

				if (*p == '.')
				{
					Dots++;
					p++;
				}
				if (*p == DIR_CHAR) p++;
				if (Dots == 2)
				{
					// chop of directory off
					TrimDir(CurDir);
				}
			}
			else if (*p == DIR_CHAR)
			{
				CurDir[0] = 0;
			}
			else
			{
				if (CurDir[strlen(CurDir)-1] != DIR_CHAR)
				{
					strcat(CurDir, DIR_STR);
				}
			}
			
			strcat(CurDir, p);
			break;
		}
	}
	return false;
}

bool FileDevice::GetCurrentDirectory(char *PathName, int Length)
{
	if (PathName)
	{
		strncpy(PathName, CurDir, Length);
		return true;
	}
	
	return false;
}

bool FileDevice::MoveFile(char *OldName, char *NewName)
{
	return false;
}

bool FileDevice::GetVolumeInfomation(char Drive, VolumeInfo *pInfo)
{
	return false;
}

bool FileDevice::CreateDirectory(char *PathName)
{
	BDirectory Old;
	status_t Status = Old.CreateDirectory(PathName, &Old);
	return (Status == B_OK) OR (Status == B_FILE_EXISTS);
}

bool FileDevice::RemoveDirectory(char *PathName)
{
	bool Status = FALSE;
	BEntry Entry(PathName);
	if (Entry.Exists())
	{
		Status = (Entry.Remove() == B_NO_ERROR);
	}
	return Status;
}

void GetFlagsStr(char *flagstr, short flags)
{
	if (flagstr)
	{
		flagstr[0] = 0;

		if (flags & O_WRONLY)
		{
			strcat(flagstr,"w");
		}
		else if (flags & O_APPEND)
		{
			strcat(flagstr,"a");
		}
		else
		{
			strcat(flagstr,"r");
		}

		if ((flags & O_RDWR) == O_RDWR)
		{
			strcat(flagstr,"+");
		}

		strcat(flagstr,"b");
	}
}

bool Match(char *Name, char *Mask)
{
	strupr(Name);
	strupr(Mask);

	while (*Name AND *Mask)
	{
		if (*Mask == '*')
		{
			if (*Name == *(Mask+1))
			{
				Mask++;
			}
			else
			{
				Name++;
			}
		}
		else if (*Mask == '?' OR *Mask == *Name)
		{
			Mask++;
			Name++;
		}
		else
		{
			return FALSE;
		}
	}

	while (*Mask AND ((*Mask == '*') OR (*Mask == '.'))) Mask++;

	return (*Name == 0 AND *Mask == 0);
}

short DaysInMonth[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
int LeapYear(int year)
{
	if (year & 3)
	{
		return 0;
	}
	if ((year % 100 == 0) && !(year % 400 == 0))
	{
		return 0;
	}
	
	return 1;
}

/////////////////////////////////////////////////////////////////////////////////
bool VDirectory::ConvertToTime(char *Str, quad Time)
{
	if (Str)
	{
		time_t time = Time;
		struct tm *t = localtime(&time);
		if (t)
		{
			strftime(Str, 256, "%H:%I:%S %p", t);
			return true;
		}
	}
	return false;
}

bool VDirectory::ConvertToDate(char *Str, quad Time)
{
	if (Str)
	{
		time_t time = Time;
		struct tm *t = localtime(&time);
		if (t)
		{
			strftime(Str, 256, "%d/%m/%Y", t);
			return true;
		}
	}
	return false;
}

/////////////////////////////////////////////////////////////////////////////////
//////////////////////////// Directory //////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

VBeDir::VBeDir()
{
	BDir = 0;
}

VBeDir::~VBeDir()
{
	DeleteObj(BDir);
}

bool VBeDir::First(char *Name, char Must, char Search)
{
	bool Status = FALSE;
	if (Name)
	{
		char *e = Name + (strlen(Name)-1);
		for (; e>Name AND *e != DIR_CHAR; e--);
		if (e == Name) return false;
		*e = 0;
		
		if (NOT BDir)
		{
			BDir = NEW(BDirectory(Name));
		}
		else
		{
			BDir->Rewind();
		}
		
		*e = DIR_CHAR;
		
		if (BDir)
		{
			Status = (BDir->GetNextEntry(&Entry) == B_OK);
		}
	}
	return Status;
}

bool VBeDir::Next()
{
	bool Status = FALSE;
	if (BDir)
	{
		Status = (BDir->GetNextEntry(&Entry) == B_OK);
	}
	return Status;
}

bool VBeDir::Close()
{
	DeleteObj(BDir);
	return TRUE;
}

bool VBeDir::IsDir()
{
	return Entry.IsDirectory();
}

long VBeDir::GetAttributes()
{
	if (BDir)
	{
		struct stat Stat;
		if (Entry.GetStat(&Stat) == B_NO_ERROR)
		{
			return Stat.st_mode;
			// & S_IWUSR))
			// Attrib |= FA_READONLY;
		}
	}
	return 0;
}

char *VBeDir::GetName()
{
	static char Name[MAX_PATH];
	Name[0] = 0;
	
	if (BDir)
	{
		Entry.GetName(Name);
	}
	
	return Name;
}

void VBeDir::GetName(char *Buf, int Len)
{
	if (BDir AND Buf AND Len > 0)
	{
		Entry.GetName(Buf);
	}
}

const quad VBeDir::GetCreationTime()
{
	time_t t;
	if (Entry.GetCreationTime(&t) == B_OK)
	{
		return t;
	}
	
	return 0;
}

const quad VBeDir::GetLastAccessTime()
{
	time_t t;
	if (Entry.GetAccessTime(&t) == B_OK)
	{
		return t;
	}
	
	return 0;
}

const quad VBeDir::GetLastWriteTime()
{
	time_t t;
	if (Entry.GetModificationTime(&t) == B_OK)
	{
		return t;
	}
	
	return 0;
}

const quad VBeDir::GetSize()
{
	if (BDir)
	{
		struct stat Stat;
		if (Entry.GetStat(&Stat) == B_NO_ERROR)
		{
			return Stat.st_size;
		}
	}
	return 0;
}

/////////////////////////////////////////////////////////////////////////////////
//////////////////////////// File ///////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

GFile::GFile()
{
	Name = 0;
	Swap = FALSE;
	Status = FALSE;

	Pos = 0;
	BufferSize = 512;
	Buffer = 0;
	Start = End = 0;
}

GFile::~GFile()
{
	Close();
}

int GFile::BufferedRead(uchar *Buf, int Size)
{
	int Bytes = 0;
	int OldSize = Size;

	if (Buf AND Buffer)
	{
		while (Size > 0)
		{
			if (Pos < Start OR Pos >= End - Size)
			{
				Start = Pos;
				End = Start + BFile::ReadAt(Start, Buffer, BufferSize);
			}
	
			if (Pos >= Start)
			{
				int Len = min(End - Pos, Size);
				if (Len > 0)
				{
					memcpy(Buf, Buffer + (Pos-Start), Len);
					Pos += Len;
					Buf += Len;
					Size -= Len;
				}
				else
				{
					// read error
					break;
				}
			}
		}
		
		Bytes = OldSize - Size;
	}
	
	return Bytes;
}

int GFile::BufferedWrite(uchar *Buf, int Size)
{
	int Bytes = 0;
	int OldSize = Size;
	
	if (Buf)
	{
		End = Start; // buffer invalid
		Bytes = BFile::WriteAt(Pos, Buf, Size); // read bytes
		Pos += Bytes;
	}
	
	return Bytes;
}

bool GFile::Open(char *File, int Mode)
{
	int Status = FALSE;

	if (File)
	{
		int Other = 0;
		
		if (Mode == O_WRITE OR Mode == O_READWRITE)
		{
			Other |= B_CREATE_FILE;
		}
		
		Close();
		if (SetTo(File, Mode | Other) == B_NO_ERROR)
		{
			Name = NewStr(File);
			Pos = 0;
			Start = 0;
			End = 0;
			Buffer = NEW(uchar[BufferSize]);
			if (Buffer)
			{
				BFile::ReadAt(Start, Buffer, BufferSize);
				End = Start + BufferSize;				
			}
			Status = Buffer != 0;
		}
	}

	return Status;
}

bool GFile::Open()
{
	return InitCheck() == B_NO_ERROR;
}

bool GFile::Close()
{
	int Status = FALSE;

	if (Open())
	{
		Unset();
		Status = TRUE;
	}
	DeleteArray(Name);
	DeleteArray(Buffer);

	return Status;
}

int GFile::Print(char *Format, ...)
{
	int Chars = 0;

	if (Format)
	{
		char Buffer[1024];
		va_list Arg;

		va_start(Arg, Format);
		Chars = vsprintf(Buffer, Format, Arg);
		va_end(Arg);

		if (Chars > 0)
		{
			Write(Buffer, Chars);
		}
	}

	return Chars;
}

#define CHUNK		0xFFF0

int GFile::Read(void *Buffer, int Size)
{
	return BufferedRead((uchar*) Buffer, Size);
}

int GFile::Write(void *Buffer, int Size)
{
	return BufferedWrite((uchar*) Buffer, Size);
}

int GFile::Seek(int To, int Whence)
{
	int CurPos = Pos;
	int NewPos = CurPos;
	switch (Whence)
	{
		case SEEK_SET:
		{
			NewPos = To;
			break;
		}
		case SEEK_CUR:
		{
			NewPos = CurPos + To;
			break;
		}
		case SEEK_END:
		{
			NewPos = GetSize() + To;
			break;
		}
		default:
		{
			return false;
		}
	}
	
	BFile::Seek(NewPos, SEEK_SET);
	Pos = Position();
	End = Start;
	return (Pos == NewPos);
}

int GFile::GetPosition()
{
	// return BFile::Position();
	return Pos;
}

int GFile::GetSize()
{
	off_t Size = 0;
	BFile::GetSize(&Size);
	return Size;
}

bool GFile::SetSize(int Size)
{
	return BFile::SetSize(Size) == B_NO_ERROR;
}

bool GFile::Eof()
{
	return GetPosition() >= GetSize();
}

int GFile::SwapRead(uchar *Buf, int Size)
{
	int i = 0;
	Buf = &Buf[Size-1];
	while (Size--)
	{
		i += BufferedRead(Buf--, 1);
	}
	return i;
}

int GFile::SwapWrite(uchar *Buf, int Size)
{
	int i = 0;
	Buf = &Buf[Size-1];
	while (Size--)
	{
		i += BufferedWrite(Buf--, 1);
	}
	return i;
}

int GFile::ReadStr(char *Buf, int Size)
{
	int i = 0;
	if (Buf AND Size > 0)
	{
		char c;

		Size--;

		do
		{
			i += BufferedRead((uchar*) &c, 1);
			if (Eof())
			{
				break;
			}
			
			*Buf++ = c;
		
		} while (i < Size - 1 AND c != '\n');

		*Buf = 0;
	}

	return i;
}

int GFile::WriteStr(char *Buf, int Size)
{
	int i = 0;

	while (i <= Size)
	{
		i += BufferedWrite((uchar*) Buf, 1);
		Buf++;
		if (NOT *Buf OR *Buf == '\n') break;
	}

	return i;
}

void GFile::SetStatus(bool s)
{
	Status = s;
}

bool GFile::GetStatus()
{
	return Status;
}

#define RdIO { Status |= ((Swap) ? SwapRead((uchar*) &i, sizeof(i)) : BufferedRead((uchar*) &i, sizeof(i))) != sizeof(i); return *this; }
#define WrIO { Status |= ((Swap) ? SwapWrite((uchar*) &i, sizeof(i)) : BufferedWrite((uchar*) &i, sizeof(i))) != sizeof(i); return *this; }

GFile &GFile::operator >> (char &i) RdIO
GFile &GFile::operator >> (signed char &i) RdIO
GFile &GFile::operator >> (unsigned char &i) RdIO
GFile &GFile::operator >> (signed short &i) RdIO
GFile &GFile::operator >> (unsigned short &i) RdIO
GFile &GFile::operator >> (signed int &i) RdIO
GFile &GFile::operator >> (unsigned int &i) RdIO
GFile &GFile::operator >> (signed long &i) RdIO
GFile &GFile::operator >> (unsigned long &i) RdIO
GFile &GFile::operator >> (float &i) RdIO
GFile &GFile::operator >> (double &i) RdIO
GFile &GFile::operator >> (long double &i) RdIO
GFile &GFile::operator >> (quad &i) RdIO

GFile &GFile::operator << (char i) WrIO
GFile &GFile::operator << (signed char i) WrIO
GFile &GFile::operator << (unsigned char i) WrIO
GFile &GFile::operator << (signed short i) WrIO
GFile &GFile::operator << (unsigned short i) WrIO
GFile &GFile::operator << (signed int i) WrIO
GFile &GFile::operator << (unsigned int i) WrIO
GFile &GFile::operator << (signed long i) WrIO
GFile &GFile::operator << (unsigned long i) WrIO
GFile &GFile::operator << (float i) WrIO
GFile &GFile::operator << (double i) WrIO
GFile &GFile::operator << (long double i) WrIO
GFile &GFile::operator << (quad i) WrIO

VFileBuf::VFileBuf()
{
	BufSize = 0;
	BufPos = 0;
	Data = 0;
}

VFileBuf::~VFileBuf()
{
	Close();
}

bool VFileBuf::Open(char *Name, int Attrib)
{
	bool Status = FALSE;

	if (GFile::Open(Name, Attrib))
	{
		BufPos = 0;
		BufSize = GFile::GetSize();
		Data = NEW(uchar[BufSize]);
		if (Data)
		{
			GFile::Read(Data, BufSize);
			Status = TRUE;
		}
	}

	return Status;
}

bool VFileBuf::Open()
{
	return (Data != 0) AND GFile::Open();
}

bool VFileBuf::Close()
{
	BufSize = 0;
	BufPos = 0;
	DeleteArray(Data);
	GFile::Close();
	return TRUE;
}

int VFileBuf::Read(void *Buffer, int Size)
{
	int Len = 0;

	if (Buffer AND Data AND BufPos >= 0 AND BufPos + 1 < BufSize)
	{
		Len = min(Size, BufSize - BufPos);
		memcpy(Buffer, Data + BufPos, Len);
		BufPos += Len;
	}

	return Len;
}

int VFileBuf::Write(void *Buffer, int Size)
{
	int Len = 0;

	if (Buffer AND BufPos >= 0 AND BufPos + 1 < BufSize)
	{
		Len = min(Size, BufSize - BufPos);
		memcpy(Data + BufPos, Buffer, Len);
		BufPos += Len;

		Seek(BufPos, SEEK_SET);
		Len = GFile::Write(Buffer, Size);
	}

	return Len;
}

int VFileBuf::Seek(int To, int Whence)
{
	int Start = BufPos;

	switch (Whence)
	{
		case SEEK_SET:
		{
			BufPos = To;
			break;
		}
		case SEEK_CUR:
		{
			BufPos += To;
			break;
		}
		case SEEK_END:
		{
			BufPos = BufSize + To;
			break;
		}
	}

	if (BufPos != Start)
	{
		GFile::Seek(To, Whence);
	}
	
	return BufPos;
}

int VFileBuf::GetPosition()
{
	return BufPos;
}

int VFileBuf::GetSize()
{
	return BufSize;
}

bool VFileBuf::SetSize(int Size)
{
	bool Status = FALSE;

	if (Size != BufSize)
	{
		int Common = min(Size, BufSize);
		uchar *Temp = NEW(uchar[Size]);
		if (Temp)
		{
			memcpy(Temp, Data, Common);
			if (Common < Size)
			{
				memset(Temp + Common, 0, Size - Common);
			}

			DeleteArray(Data);
			Data = Temp;
			BufSize = Size;
			Status = GFile::SetSize(Size);
		}
	}

	return Status;
}

bool VFileBuf::Eof()
{
	return (BufPos >= 0 AND BufPos < BufSize);
}

int VFileBuf::ReadStr(char *Buf, int Size)
{
	int i;

	for (i=0; i<Size-1 AND Data[BufPos++] != '\n'; i++)
	{
	}
	
	if (Buf)
	{
		memcpy(Buf, Data+BufPos, i);
		Buf[Size-1] = 0;
		return i;
	}
	return 0;
}

int VFileBuf::WriteStr(char *Buf, int Size)
{
	if (Buf)
	{
		int i;
		uchar *p = Data + BufPos;
		for (i=0; i<Size AND Buf[BufPos] != '\n'; i++)
			*p++ = Buf[BufPos++];

		Seek(BufPos, SEEK_SET);
		GFile::Write(Buf, i);

		return i;
	}
	return 0;
}

int VFileBuf::SwapRead(uchar *Buf, int Size)
{
	int i=0;
	Buf = &Buf[Size-1];
	while (Size-- AND BufPos < BufSize)
	{
		*Buf-- = Data[BufPos++];
		i++;
	}
	return i;
}

int VFileBuf::SwapWrite(uchar *Buf, int Size)
{
	int i=0;
	int Start = BufPos;

	Buf = &Buf[Size-1];
	while (Size-- AND BufPos < BufSize)
	{
		Data[BufPos++] = *Buf--;
		i++;
	}

	Seek(Start, SEEK_SET);
	GFile::Write(Data + Start, i);

	return i;
}

#define BufRdIO { Status |= ((Swap) ? SwapRead((uchar*) &i, sizeof(i)) : Read(&i, sizeof(i))) != sizeof(i); return *this; }
#define BufWrIO { Status |= ((Swap) ? SwapWrite((uchar*) &i, sizeof(i)) : Write(&i, sizeof(i))) != sizeof(i); return *this; }

GFile &VFileBuf::operator >> (char &i) BufRdIO
GFile &VFileBuf::operator >> (signed char &i) BufRdIO
GFile &VFileBuf::operator >> (unsigned char &i) BufRdIO
GFile &VFileBuf::operator >> (signed short &i) BufRdIO
GFile &VFileBuf::operator >> (unsigned short &i) BufRdIO
GFile &VFileBuf::operator >> (signed int &i) BufRdIO
GFile &VFileBuf::operator >> (unsigned int &i) BufRdIO
GFile &VFileBuf::operator >> (signed long &i) BufRdIO
GFile &VFileBuf::operator >> (unsigned long &i) BufRdIO
GFile &VFileBuf::operator >> (float &i) BufRdIO
GFile &VFileBuf::operator >> (double &i) BufRdIO
GFile &VFileBuf::operator >> (long double &i) BufRdIO

GFile &VFileBuf::operator << (char i) BufWrIO
GFile &VFileBuf::operator << (signed char i) BufWrIO
GFile &VFileBuf::operator << (unsigned char i) BufWrIO
GFile &VFileBuf::operator << (signed short i) BufWrIO
GFile &VFileBuf::operator << (unsigned short i) BufWrIO
GFile &VFileBuf::operator << (signed int i) BufWrIO
GFile &VFileBuf::operator << (unsigned int i) BufWrIO
GFile &VFileBuf::operator << (signed long i) BufWrIO
GFile &VFileBuf::operator << (unsigned long i) BufWrIO
GFile &VFileBuf::operator << (float i) BufWrIO
GFile &VFileBuf::operator << (double i) BufWrIO
GFile &VFileBuf::operator << (long double i) BufWrIO

/////////////////////////////////////////////////////////////////////////////////
//////////////////////////// VDirView ///////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

VDirView::VDirView()
{
	Items = 0;
	Files = 0;
	Dirs = 0;
	strcpy(CurrentDir, ".");


	Index = 0;
	Root = 0;
}

VDirView::~VDirView()
{
	DeleteObj(Root);
	DeleteArray(Index);
}

bool VDirView::Read(char *Dir)
{
	if (NOT Dir) Dir = ".";
	DeleteObj(Root);

	Items = 0;
	Files = 0;
	Dirs = 0;
	DeleteArray(Index);

	Node *Current;
	if (strlen(CurrentDir) > 3)
	{
		Current = NEW(Node("..", FA_DIRECTORY, 0));
		if (Current)
		{
			if (NOT Root)
			{
				Root = Current;
			}
			else
			{
				Root->Add(Current);
			}

			Dirs++;
			Items++;
		}
	}

	bool Done = First("*.*", 0x10, 0x1F);
	while (NOT Done)
	{
		if (strcmp(GetName(), ".") AND strcmp(GetName(), ".."))
		{
			Current = NEW(Node(GetName(), GetAttributes(), GetSize()));
			if (Current)
			{
				if (NOT Root)
				{
					Root = Current;
				}
				else
				{
					Root->Add(Current);
				}

				Dirs++;
				Items++;
			}
		}
		Done = Next();
	}

	Done = First("*.*", 0x0, 0x2F);
	while (NOT Done)
	{
		Current = NEW(Node(GetName(), GetAttributes(), GetSize()));
		if (Current)
		{
			if (NOT Root)
			{
				Root = Current;
			}
			else
			{
				Root->Add(Current);
			}

			Files++;
			Items++;
		}
		Done = Next();
	}

	if (Root AND Items)
	{
		Index = NEW(Node*[Items]);
		if (Index)
		{
			Node **Ptr = Root->Traverse(Index, TRUE);
			Root->Traverse(Ptr, FALSE);
		}
	}

	return TRUE;
}

VDirView::Node::Node()
{
	Left = Right = 0;
}

VDirView::Node::Node(char *n, int a, quad s) : NodeInfo(n, a, s)
{
	Left = Right = 0;
}

VDirView::Node::~Node()
{
	DeleteObj(Left);
	DeleteObj(Right);
}

VDirView::Node **VDirView::Node::Traverse(VDirView::Node **ppNode, int Dir)
{
	if (Left)
	{
		ppNode = Left->Traverse(ppNode, Dir);
	}

	if (Dir)
	{
		if (GetAttributes() & FA_DIRECTORY)
		{
			*ppNode++ = this;
		}
	}
	else
	{
		if (NOT (GetAttributes() & FA_DIRECTORY))
		{
			*ppNode++ = this;
		}
	}

	if (Right)
	{
		ppNode = Right->Traverse(ppNode, Dir);
	}

	return ppNode;
}

void VDirView::Node::Add(Node *Info)
{
	int Result = *this ^ *Info;

	if (Result < 0)
	{
		if (Right)
		{
			Right->Add(Info);
		}
		else
		{
			Right = Info;
		}
	}
	else if (Result > 0)
	{
		if (Left)
		{
			Left->Add(Info);
		}
		else
		{
			Left = Info;
		}
	}
}

NodeInfo::NodeInfo()
{
	Name = 0;
	Attributes = 0;
	Size = 0;
}

NodeInfo::NodeInfo(char *n, int a, quad s)
{
	if (n)
	{
		Name = NEW(char[strlen(n)+1]);
		if (Name)
		{
			strcpy(Name, n);
		}
	}
	else
	{
		Name = 0;
	}
	Attributes = a;
	Size = s;
}

NodeInfo::~NodeInfo()
{
	DeleteArray(Name);
}

int NodeInfo::operator ^(NodeInfo &N)
{
	return stricmp(Name, N.Name);
}
