#include "Lgi.h"

/////////////////////////////////////////////////////////////////////////////
GApp *GApp::TheApp = 0;

GApp::GApp(char *Mime, int Args, char **Arg) : BApplication(Mime)
{
	LgiAssert(0);
	TheApp = this;

	srand(system_time());
	// CmdLine = NewStr(lpCmdLine);
	LgiInitColours();
	AppWnd = 0;

	// System font
	SystemFont = 0;
	GdcFontType SysFontType;
	if (SysFontType.GetSystemFont("System"))
	{
		SystemFont = SysFontType.CreateFont();
		if (SystemFont)
		{
			SystemFont->Transparent(true);
		}
	}

	if (NOT SystemFont)
	{
		LgiMsg(0, "Error: Couldn't create system font.", "Lgi Error: GApp::GApp", MB_OK);
		LgiExitApp();
	}
}

GApp::~GApp()
{
	// DeleteObj(AppWnd);	
	// DeleteObj(SystemFont);
	TheApp = 0;
}

bool GApp::Run(bool Loop)
{
	BApplication::Run();
}

void GApp::RefsReceived(BMessage *Msg)
{
	List<char> Files;
	int32 Count = 0;
	type_code Type = 0;
	
	if (Msg->GetInfo("refs", &Type, &Count) == B_OK)
	{
		for (int i=0; i<Count; i++)
		{
			entry_ref Ref;
			if (Msg->FindRef("refs", i, &Ref) == B_OK)
			{
				BPath Path;
				BEntry Entry(&Ref, true);
				Entry.GetPath(&Path);
				char *p = (char*) Path.Path();
				if (p)
				{
					Files.Insert(NewStr(p));
				}
			}
		}
	}
	
	if (AppWnd AND Files.GetItems() > 0)
	{
		AppWnd->OnReceiveFiles(Files);
	}
	
	for (char *f=Files.First(); f; f=Files.Next())
	{
		DeleteArray(f);
	}
}

void GApp::OnReceiveFiles(List<char> &Files)
{
	if (AppWnd)
	{
		AppWnd->OnReceiveFiles(Files);
	}
}

XmlTag *GApp::GetConfig(char *Tag)
{
	if (NOT Config)
	{
		char Path[256];
		if (LgiGetExePath(Path))
		{
			if (Path[strlen(Path)-1] != DIR_CHAR) strcat(Path, DIR_STR);
			strcat(Path, "lgi.conf");
			if (FileExists(Path))
			{
				Config = NEW(Xml);
				if (Config)
				{
					if (Config->ParseXmlFile(Path) <= 0)
					{
						DeleteObj(Config);
					}
				}
			}
		}
	}

	if (Tag AND Config)
	{
		for (XmlTag *t=Config->Values.First(); t; t=Config->Values.Next())
		{
			if (t->Name AND
				stricmp(t->Name, Tag) == 0)
			{
				return t;
			}
		}
	}

	return 0;
}

bool GApp::GetOption(char *Dest, char Option)
{
	/*
	printf("GApp::GetOption(%c)\n", Option);
	printf("CmdLine: '%s'\n", CmdLine);
	*/

	char *c = CmdLine;
	while (c AND *c)
	{
		if (*c == '/' OR *c == '\\' OR *c == '-')
		{
			c++;
			if (*c == Option)
			{
				c++;
				if (*c)
				{
					while (*c AND strchr(" \t\r\n", *c))
					{
						c++;
					}

					for (char *i = c; *i AND *i != ' ';)
					{
						*Dest++ = *i++;
					}
					*Dest++ = 0;
					return TRUE;
				}
			}
		}

		c = strchr(c, ' ');
		if (c) c++;
	}

	return FALSE;
}

void GApp::OnCmdLine(char *Cmd)
{
	char WhiteSpace[] = " \r\n\t";

	DeleteArray(CmdLine);
	CmdLine = NewStr(Cmd);

	if (CmdLine)
	{
		List<char> Files;
		
		char *Delim = "\'\"";
		char *s;
		for (s = CmdLine; *s; )
		{
			// skip ws
			while (strchr(WhiteSpace, *s)) s++;

			// read to end of token
			char *e = s;
			if (strchr(Delim, *s))
			{
				for (e = ++s; *e AND NOT strchr(Delim, *e); e++);
			}
			else
			{
				for (; *e AND NOT strchr(WhiteSpace, *e); e++);
			}

			char *Arg = NewStr(s, (int)e-(int)s);
			if (Arg)
			{
				if (FileExists(Arg))
				{
					Files.Insert(Arg);
				}
				else
				{
					DeleteArray(Arg);
				}
			}

			// next
			s = (*e) ? e + 1 : e;
		}

		// call app
		if (Files.GetItems() > 0)
		{
			OnReceiveFiles(Files);
		}

		// clear up
		DeleteArrayList(&Files, char);
	}
}

void GApp::Exit(int Code)
{
	// DeleteObj(AppWnd);
	be_app->PostMessage(B_QUIT_REQUESTED);
}

