// BeOS Implementation of General LGI functions
#include "Lgi.h"

////////////////////////////////////////////////////////////////
// Local helper functions
bool _lgi_check_file(char *Path)
{
	if (Path)
	{
		if (FileExists(Path))
		{
			// file is there
			return true;
		}
		else
		{
			// shortcut?
			char *e = Path + strlen(Path);
			strcpy(e, ".lnk");
			if (FileExists(Path))
			{
				// resolve shortcut
				char Link[256];
				if (ResolveShortcut(Path, Link, sizeof(Link)))
				{
					// check destination of link
					if (FileExists(Link))
					{
						strcpy(Path, Link);
						return true;
					}
				}
			}
			*e = 0;
		}
	}

	return false;
}

long _lgi_play_sound(void *Ptr)
{
	BEntry Entry((char*)Ptr);
	entry_ref Ref;
	if (Entry.GetRef(&Ref) == B_OK)
	{
		BSimpleGameSound Sound(&Ref);
		if (Sound.InitCheck() == B_OK)
		{
			Sound.StartPlaying();
			while (Sound.IsPlaying())
			{
			}
			return true;
		}
	}
	
	return false;
}

////////////////////////////////////////////////////////////////////////
// Implementations
bool LgiGetFileMimeType(char *File, char *Mime)
{
	bool Status = false;
	if (File AND Mime)
	{
		BFile f(File, B_READ_ONLY);
		BNodeInfo i(&f);
		Status = i.GetType(Mime) == B_OK;		
	}
	return Status;
}

bool LgiGetAppForMimeType(char *Mime, char *AppPath)
{
	bool Status = false;
	if (AppPath AND Mime)
	{
		BMimeType m(Mime);
		char DefApp[B_MIME_TYPE_LENGTH];
		if (m.GetPreferredApp(DefApp) == B_OK)
		{
			entry_ref ap;
			BMimeType a(DefApp);
			if (a.GetAppHint(&ap) == B_OK)
			{
				BPath p;
				BEntry e(&ap);
				if (e.GetPath(&p) == B_OK)
				{
					strcpy(AppPath, p.Path());
					Status = true;
				}
			}
		}
	}
	return Status;
}

int LgiRand(int Limit)
{
	return rand() % Limit;
}

int LgiMsg(GView *Parent, char *Str, char *Title, int Type, ...)
{
	int Res = 0;
	if (Str)
	{
		char Buffer[1024];
		va_list Arg;

		va_start(Arg, Type);
		vsprintf(Buffer, Str, Arg);
		va_end(Arg);

		BAlert *Dlg = 0;
		
		switch (Type)
		{
			case MB_OK:
			{
				Dlg = NEW(BAlert((Title) ? Title : "Message", Buffer, "Ok"));
				break;
			}
			case MB_YESNO:
			{
				Dlg = NEW(BAlert((Title) ? Title : "Message", Buffer, "Yes", "No", NULL, B_WIDTH_AS_USUAL, B_IDEA_ALERT));
				break;
			}
			case MB_YESNOCANCEL:
			{
				Dlg = NEW(BAlert((Title) ? Title : "Message", Buffer, "Yes", "No", "Cancel", B_WIDTH_AS_USUAL, B_IDEA_ALERT));
				break;
			}
			default:
			{
				ASSERT(0);
				break;
			}
		}
		
		if (Dlg)
		{
			Res = Dlg->Go();
		}
		
		if (Res >= 0 AND
			(Type == MB_YESNO OR Type == MB_YESNOCANCEL))
		{
			if (Res == 0)
				return IDYES;
			if (Res == 1)
				return IDNO;
			if (Res == 2)
				return IDCANCEL;
		}
	}
	return Res;
}

bool LgiPlaySound(char *FileName, int ASync)
{
	bool Status = false;

	if (ASync)
	{
		thread_id Thread = spawn_thread(_lgi_play_sound, "_lgi_play_sound", B_LOW_PRIORITY, FileName);
		if (Thread > 0)
		{
			resume_thread(Thread);
			Status = true;
		}
	}
	else
	{
		Status = _lgi_play_sound(FileName);
	}

	return Status;
}

bool LgiExecute(char *File, char *Args, char *Dir)
{
	printf("LgiExecute(%s, %s, %s)\n", File, Args, Dir);
	if (File)
	{
		char f[256];
		
		if (Dir)
		{
			sprintf(f, "%s/%s", Dir, File);
		}
		else
		{
			strcpy(f, File);
		}
	
		BRoster Roster;
		BEntry Entry(f);
		entry_ref Ref;
		if (Entry.GetRef(&Ref) == B_OK)
		{
			status_t s = Roster.Launch(&Ref);
			return s == B_OK OR s == B_ALREADY_RUNNING;
		}
		else
		{
			printf("LgiExecute(...), not a file\n");
			if (strnicmp(File, "http://", 7) == 0)
			{
				if (Roster.FindApp("text/html", &Ref) == B_OK)
				{
					/*
					char *Arg[2] = {Ref.name, File};
					return Roster.Launch(&Ref, 2, Arg) == B_OK;
					*/
					char *Arg[1] = {File};
					status_t s = Roster.Launch(&Ref, 1, Arg);
					return s == B_OK OR s == B_ALREADY_RUNNING;
				}
			}
		}
	}
	return false;
}

