/*
**	FILE:		GuiViews.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		18/7/98
**	DESCRIPTION:	Standard Views
**
**	Copyright (C) 1998, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
GLayout::GLayout() :
	GView(Redir = NEW(BViewRedir(this)))
{
	_PourLargest = false;
	VScroll = 0;
	HScroll = 0;
	XScroll = YScroll = false;
	Redir->SetScrollView(0);

	// GetBView()->SetResizingMode(B_FOLLOW_NONE);
	// GetBView()->SetViewColor(216, 216, 216);
	// GetBView()->SetViewColor(B_TRANSPARENT_COLOR);
}

GLayout::~GLayout()
{
	DeleteObj(HScroll);
	DeleteObj(VScroll);
}

bool GLayout::GetPourLargest()
{
	return _PourLargest;
}

void GLayout::SetPourLargest(bool i)
{
	_PourLargest = i;
}

bool GLayout::Pour(GComplexRegion &r)
{
	if (_PourLargest)
	{
		GRegion *Best = FindLargest(r);
		if (Best)
		{
			SetPos(*Best, true);
			return true;
		}
	}
	
	return false;
}

void GLayout::GetScrollPos(int &x, int &y)
{
	if (HScroll)
	{
		x = HScroll->Pos();
	}
	else
	{
		x = 0;
	}

	if (VScroll)
	{
		y = VScroll->Pos();
	}
	else
	{
		y = 0;
	}
}

void GLayout::SetScrollPos(int x, int y)
{
	if (HScroll)
	{
		HScroll->SetPos(x);
	}

	if (VScroll)
	{
		VScroll->SetPos(y);
	}
}

bool GLayout::Attach(GView *p)
{
	bool Status = GView::Attach(p);
	
	if (HScroll OR VScroll)
	{
		SetScrollBars(HScroll!=0, VScroll!=0);
		
		if (Redir->GetScrollView())
		{
			if (HScroll)
			{
				HScroll->Bar = Redir->GetScrollView()->ScrollBar(B_HORIZONTAL);
			}
			if (VScroll)
			{
				VScroll->Bar = Redir->GetScrollView()->ScrollBar(B_VERTICAL);
			}
		}
	}
	
	Visible(GFlags::Visible());

	return Status;
}

bool GLayout::SetScrollBars(bool x, bool y)
{
	bool Has = (HScroll != 0) OR (VScroll != 0);
	bool Changing = Has ^ (x OR y);
	bool Attached = (Handle() != 0) AND (Handle()->Parent() != 0);

	if (Changing)
	{
		if (Attached AND
			Redir->GetScrollView())
		{
			delete Redir->GetScrollView();
			Redir->SetScrollView(0);
		}
	
		if (x)
		{
			if (NOT HScroll)
			{
				HScroll = NEW(GScrollBar(0));
				if (HScroll)
				{
					HScroll->SetId(IDC_HSCROLL);
				}
			}
		}
		else
		{
			DeleteObj(HScroll);
		}
		if (y)
		{
			if (NOT VScroll)
			{
				VScroll = NEW(GScrollBar(0));
				if (VScroll)
				{
					VScroll->SetId(IDC_VSCROLL);
				}
			}
		}
		else
		{
			DeleteObj(VScroll);
		}
	
		if ((x OR y) AND
			Attached)
		{
			GRegion r = GetPos();
			
			Redir->SetScrollView(NEW(GScrollView("GViewScroller",
												this,
												B_FOLLOW_LEFT | B_FOLLOW_TOP,
												0,
												x,
												y,
												Sunken())));
			if (Redir->GetScrollView())
			{
				Sunken(false);
				
				SetPos(r);
				if (HScroll)
				{
					HScroll->Bar = Redir->GetScrollView()->ScrollBar(B_HORIZONTAL);
				}
				if (VScroll)
				{
					VScroll->Bar = Redir->GetScrollView()->ScrollBar(B_VERTICAL);
				}
			}
		}
	}
	
	return true;
}

GRegion &GLayout::GetPos()
{
	if (Redir->GetScrollView())
	{
		return Redir->GetScrollView()->GetPos();
	}

	return GView::GetPos();
}

bool GLayout::SetPos(GRegion &p, bool Repaint)
{
	if (Redir->GetScrollView())
	{
		Redir->GetScrollView()->SetPos(p);
		if (Repaint AND
			Redir->GetScrollView()->LockLooper())
		{
			Redir->GetScrollView()->Invalidate();
			Redir->GetScrollView()->UnlockLooper();
		}
	}
	else
	{
		GView::SetPos(p, Repaint);
	}
	
	return TRUE;
}

bool GLayout::Visible()
{
	return GView::Visible();
}

void GLayout::Visible(bool i)
{
	if (Redir->GetScrollView())
	{
		// Set view visible
		bool Lock = Redir->GetScrollView()->LockLooper();
		if (i)
		{
			if (NOT GFlags::Visible())
			{
				Redir->GetScrollView()->Show();
			}
		}
		else
		{
			if (GFlags::Visible())
			{
				Redir->GetScrollView()->Hide();
			}
		}
		if (Lock) Redir->GetScrollView()->UnlockLooper();

		// keep track of visibility (used when we're not attached)
		GFlags::Visible(i);
	}
	else
	{
		GView::Visible(i);
	}
}

int GLayout::OnEvent(BMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case WM_CHANGE:
		{
			if (GetParent())
			{
				GetParent()->OnEvent(Msg);
			}
			break;
		}
		case WM_VSCROLL:
		{
			if (VScroll)
			{
				OnNotify(VScroll, MsgA(Msg));
			}
			break;
		}
		case WM_HSCROLL:
		{
			if (HScroll)
			{
				OnNotify(HScroll, MsgA(Msg));
			}
			break;
		}
	}
	
	if (VScroll) VScroll->OnEvent(Msg);
	if (HScroll) HScroll->OnEvent(Msg);
	return GView::OnEvent(Msg);
}

