#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////////////
GStatusBar::GStatusBar()
{
	GView::Name("LGI_StatusBar");

	Font = NEW(GFont);
	if (Font)
	{
		Font->Transparent(TRUE);
		Font->Back(LC_MED);
		Font->Create("Ms Sans Serif", 10);
	}
}

GStatusBar::~GStatusBar()
{
	for (int i=0; i<GetSize(); i++)
	{
		DeleteObj((*this)[0]);
	}

	DeleteObj(Font);
}

bool GStatusBar::Pour(GComplexRegion &r)
{
	GRegion *Best = FindLargestEdge(r, GV_EDGE_BOTTOM);
	if (Best)
	{
		GRegion Take = *Best;
		
		int y = 21;
		// int y = 50;
		
		if (Take.Y() > y)
		{
			Take.y1 = max((Take.y2 - y), Take.y1);
			SetPos(Take);
			RePour();
			return TRUE;
		}
	}
	return FALSE;
}

void GStatusBar::OnPaint(GSurface *pDC)
{
	int Xm1 = pDC->X()-1;
	int Ym1 = pDC->Y()-1;

	pDC->Colour(LC_LOW, 24);
	pDC->Line(Xm1, Ym1, 0, Ym1);
	pDC->Line(Xm1, Ym1, Xm1, 0);

	pDC->Colour(LC_HIGH, 24);
	pDC->Line(0, 0, Xm1, 0);
	pDC->Line(0, 0, 0, Ym1);

	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(1, 1, Xm1-1, Ym1-1);

	#ifdef WIN32
	for (	GStatusPane *Pane = (GStatusPane*) Children.First();
			Pane;
			Pane = (GStatusPane*) Children.Next())
	{
		pDC->SetOrigin(-Pane->Pos.x1, -Pane->Pos.y1);
		Pane->OnPaint(pDC);
	}
	#endif
}

void GStatusBar::RePour()
{
	int X = Pos.X() - 5;
	int i = Children.GetItems()-1;

	for (	GStatusPane *Pane = (GStatusPane*) Children.Last();
			Pane;
			Pane = (GStatusPane*) Children.Prev(), i--)
	{
		if (i)
		{
			// not first
			X -= Pane->Width; 
			Pane->Pos.ZOff(Pane->Width, Pos.Y()-7);
			Pane->Pos.Offset(X, 3);
			X -= STATUSBAR_SEPARATOR;
		}
		else
		{
			// first
			Pane->Pos.ZOff(X-2, Pos.Y()-4);
			Pane->Pos.Offset(2, 2);
			X = 0;
		}

		Pane->SetPos(Pane->Pos);
	}
}

GStatusPane *GStatusBar::AppendPane(char *Text, int Width)
{
	GStatusPane *Pane = 0;
	if (Text)
	{
		Pane = NEW(GStatusPane);
		if (Pane)
		{
			Pane->SetParent(this);
			Pane->Name(Text);
			Pane->Width = Width;
			Children.Insert(Pane);

			#ifdef BEOS
			Pane->Attach(this);
			#endif
		}
	}
	return Pane;
}

bool GStatusBar::AppendPane(GStatusPane *Pane)
{
	if (Pane)
	{
		Pane->SetParent(this);
		Children.Insert(Pane);

		#ifdef BEOS
		Pane->Attach(this);
		#endif
	}
	return TRUE;
}

// Pane
GStatusPane::GStatusPane() :
	/*
	BView(	BRect(0, 0, 1, 1),
			NULL,
			B_FOLLOW_NONE,
			B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE),
	*/
	GView(NEW(BViewRedir(this)))
{
	Parent = 0;
	Flags = 0;
	Width = 32;
	Pos.ZOff(Width-1, 20);
	pDC = 0;
	_BorderSize = 1;
	
	Handle()->SetViewColor(B_TRANSPARENT_COLOR);
}

GStatusPane::~GStatusPane()
{
	DeleteObj(pDC)
}

bool GStatusPane::Name(char *n)
{
	bool Status = GObject::Name(n);
	GRegion p(0, 0, Pos.X()-1, Pos.Y()-1);
	p.Size(1, 1);
	Invalidate();
	return Status;
}

void GStatusPane::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, Pos.X()-1, Pos.Y()-1);
	int Off = 1;

	if (Sunken())
	{
		Off = 0;
		LgiThinBorder(pDC, r, SUNKEN);
	}
	
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&r);

	GStatusBar *Par = (GStatusBar*) GView::Parent;
	if (Par AND Par->Font)
	{
		Par->Font->Text(pDC, 2+Off, 1+Off, Name(), -1, &r);
	}
}

int GStatusPane::GetWidth()
{
	return Width;
}

void GStatusPane::SetWidth(int x)
{
	Width = x;
	if (GView::Parent)
	{
		((GStatusBar*) GView::Parent)->RePour();
	}
}

bool GStatusPane::Sunken()
{
	return GView::Sunken();
}

void GStatusPane::Sunken(bool s)
{
	bool Update = s ^ GView::Sunken();
	GView::Sunken(s);

	if (GetParent() AND Update)
	{
		GetParent()->Invalidate(&Pos);
	}
}

GSurface *GStatusPane::Bitmap()
{
	return pDC;
}

void GStatusPane::Bitmap(GSurface *pNewDC)
{
	DeleteObj(pDC);
	if (pNewDC)
	{
		pDC = NEW(GMemDC);
		if (pDC)
		{
			if (pDC->Create(pNewDC->X(), pNewDC->Y(), pNewDC->GetBits()))
			{
				pDC->Blt(0, 0, pNewDC);
			}
			else
			{
				DeleteObj(pDC);
			}
		}
	}

	if (GView::Parent)
	{
		GView::Parent->Invalidate(&Pos);
	}
}
