/*
**	FILE:		GuiToolbar.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		18/7/98
**	DESCRIPTION:	Toolbar classes
**
**	Copyright (C) 1998, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

////////////////////////////////////////////////////////////////////////
GToolBar *LgiLoadToolbar(GView *Parent, char *File, int x, int y)
{
	GToolBar *Toolbar = NEW(GToolBar);
	if (Toolbar)
	{
		char *FileName = LgiFindFile(File);
		bool Success = FileName AND Toolbar->SetBitmap(FileName, x, y);
		if (NOT Success)
		{
			LgiMsg(	Parent,
					"Can't find the graphic '%s' for the toolbar.\n"
					"You can find it in this program's archive.",
					"LgiLoadToolbar",
					MB_OK,
					File);
			DeleteObj(Toolbar);
		}
		DeleteArray(FileName);
	}
	return Toolbar;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GImageList::GImageList(int x, int y, GSurface *pDC)
{
	Sx = x;
	Sy = y;

	if (pDC)
	{
		int Bits = (GdcD->GetBits() == 32) ? 32 : 15;
		if (Create(pDC->X(), pDC->Y(), Bits))
		{
			GdcPalette *Pal = pDC->Palette();
			if (Pal)
			{
				Palette(NEW(GdcPalette(Pal)));
			}

			Blt(0, 0, pDC);
			
			COLOUR InitKey = Get(0, 0);

			for (int cy=0; cy<Y(); cy++)
			{
				switch (GetBits())
				{
					case 8:
					{
						uchar Key = InitKey;
						uchar *p = (*this)[cy];
						for (int cx=0; cx<X(); cx++)
						{
							if (p[cx] == Key)
							{
								p[cx] = B_TRANSPARENT_MAGIC_CMAP8;
							}
						}
						break;
					}
					case 15:
					{
						ushort Key = InitKey;
						ushort *p = (ushort*) (*this)[cy];
						ushort Trans = B_TRANSPARENT_MAGIC_RGBA15;
						
						for (int cx=0; cx<X(); cx++)
						{
							if (p[cx] == Trans)
							{
								p[cx] = Trans+1;
							}
							else if (p[cx] == Key)
							{
								p[cx] = Trans;
							}
						}
						break;
					}
					case 32:
					{
						ulong Key = InitKey;
						ulong *p = (ulong*) ((*this)[cy]);
						for (int cx=0; cx<X(); cx++)
						{
							if (	R32(p[cx]) == R32(Key) AND
									G32(p[cx]) == G32(Key) AND
									B32(p[cx]) == B32(Key))
							{
								p[cx] = B_TRANSPARENT_MAGIC_RGBA32;
							}
						}
						break;
					}
				}

			}
		}
	}
}

GImageList::~GImageList()
{
}

bool GImageList::Create(int x, int y, int Bits, int LineLen, bool KeepData)
{
	return GMemDC::Create(x, y, Bits, LineLen, KeepData);
}

void GImageList::Update(int Flags)
{
	if (Mask)
	{
	}
}

void GImageList::Draw(GSurface *pDest, int Dx, int Dy, int Image, int Flags)
{
	if (pDest)
	{
		COLOUR Key = Get(0, 0);
		GRegion r;
		r.ZOff(Sx-1, Sy-1);
		r.Offset(Image * Sx, 0);

		if (Flags & IMGLST_DISABLED)
		{
			COLOUR Low = LC_LOW;
			COLOUR High = LC_LIGHT;
			
			for (int y=0; y<r.Y(); y++)
			{
				for (int x=0; x<r.X(); x++)
				{
					if (Get(r.x1 + x, r.y1 + y) != Key)
					{
						pDest->Colour(Low, 24);
						pDest->Set(Dx + x, Dy + y);
						pDest->Colour(High, 24);
						pDest->Set(Dx + x + 1, Dy + y + 1);
					}
				}
			}
		}
		else
		{
			GdcViewDC *pViewDC = dynamic_cast<GdcViewDC*>(pDest);
			if (pDest->IsScreen() AND pViewDC)
			{
				BView *owner = pViewDC->Handle();
				BBitmap *Bmp = Handle();

				if (owner AND Bmp)
				{
					int x = Image * Sx;
					BRect S(x, 0, x+Sx-1, Sy-1);
					BRect D(Dx, Dy, Dx+(Sx-1), Dy+(Sx-1));

					drawing_mode Mode = owner->DrawingMode();
					owner->SetDrawingMode(B_OP_OVER);
					owner->DrawBitmap(Bmp, S, D);
					owner->SetDrawingMode(Mode);
				}
			}
			else
			{
				/*
				if (pDest->GetBits() == 16 AND
					GetBits() == 15)
				{
					// 15 -> 16 converter
					for (int y=0; y<r.Y(); y++)
					{
						for (int x=0; x<r.X(); x++)
						{
							COLOUR c = Get(r.x1 + x, r.y1 + y);
							if (c != Key)
							{
								pDest->Colour(Rgb15To16(c));
								pDest->Set(Dx + x, Dy + y);
							}
						}
					}
				}
				else if (pDest->GetBits() == 8)
				{
					color_map *CMap = (color_map*) system_colors();
					if (NOT CMap) return;

					switch (GetBits())
					{
						case 15:
						{
							for (int y=0; y<r.Y(); y++)
							{
								ushort *s = ((ushort*) (*this)[r.y1 + y]) + r.x1;
								for (int x=0; x<r.X(); x++, s++)
								{
									// COLOUR c = Get(r.x1 + x, r.y1 + y);
									if (*s != Key)
									{
										// CMap->index_map[Rgb24To15(c)];
										
										pDest->Colour(CMap->index_map[*s]);
										pDest->Set(Dx + x, Dy + y);
									}
								}
							}
							break;
						}
						case 16:
						{
							for (int y=0; y<r.Y(); y++)
							{
								ushort *s = ((ushort*) (*this)[r.y1 + y]) + r.x1;
								for (int x=0; x<r.X(); x++, s++)
								{
									// COLOUR c = Get(r.x1 + x, r.y1 + y);
									if (*s != Key)
									{
										// CMap->index_map[Rgb24To15(c)];
										
										pDest->Colour(CMap->index_map[Rgb16To15(*s)]);
										pDest->Set(Dx + x, Dy + y);
									}
								}
							}
							break;
						}
						case 32:
						{
							for (int y=0; y<r.Y(); y++)
							{
								ulong *s = ((ulong*) (*this)[r.y1 + y]) + r.x1;
								for (int x=0; x<r.X(); x++, s++)
								{
									if (*s != Key)
									{
										pDest->Colour(CMap->index_map[Rgb32To15(*s)]);
										pDest->Set(Dx + x, Dy + y);
									}
								}
							}
							break;
						}
					}
				}
				else
				*/
				{
					for (int y=0; y<r.Y(); y++)
					{
						int b = GetBits();
						
						for (int x=0; x<r.X(); x++)
						{
							COLOUR c = Get(r.x1 + x, r.y1 + y);
							if (c != Key)
							{
								pDest->Colour(c, b);
								pDest->Set(Dx + x, Dy + y);
							}
						}
					}
				}
				
				// end normal draw
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GToolButton::GToolButton(int Bx, int By) :
	/*
	BView(	BRect(0, 0, 1, 1),
			NULL,
			B_FOLLOW_NONE,
			B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE),
	GControl(this)
	*/
	GControl(NEW(BViewRedir(this)))
{
	Name("");
	Type = TBT_PUSH;
	SetId(IDM_NONE);
	SetDown(FALSE);
	Clicked = FALSE;
	ImgIndex = 0;

	Pos.ZOff(Bx+1, By+1);
	SetParent(0);
	_BorderSize = 0;

	GetBView()->SetViewColor(B_TRANSPARENT_COLOR);
}

GToolButton::~GToolButton()
{
}

void GToolButton::OnPaint(GSurface *pDC)
{
	GToolBar *Par = (GToolBar*) GView::Parent;
	if (Par AND Par->ImgList)
	{
		int MemBits = Par->ImgList->GetBits(); // GdcD->GetBits()
		
		GMemDC *pMemDC = NEW(GMemDC);
		if (pMemDC AND pMemDC->Create(Pos.X(), Pos.Y(), MemBits))
		{
			int Px = Pos.X()-1;
			int Py = Pos.Y()-1;
			COLOUR Grey = LC_MED;
			if (GdcD->GetBits() == 16)
			{
				Grey = Rgb24(R24(Grey), G24(Grey)-1, B24(Grey));
			}

			pMemDC->Colour(Grey, 24);
			pMemDC->Rectangle();
			if (((Type == TBT_RADIO) OR (Type == TBT_TOGGLE)) AND Down)
			{
				// draw highlight
				pMemDC->Colour(LC_LIGHT, 24);

				for (int y=0; y<pMemDC->Y(); y++)
				{
					for (int x=y&1; x<pMemDC->X(); x+=2)
					{
						pMemDC->Set(x, y);
					}
				}
			}

			int Off = BORDER_BUTTON + ((Down) ? 1 : 0);
			
			if (GetId() >= 0)
			{
				// button
				int Flags = 0;

				if (Enabled())
				{
					// draw normal
					Par->ImgList->Draw(pMemDC, Off, Off, ImgIndex, Flags);
				}
				else
				{
					// draw disabled
					Par->ImgList->Draw(pMemDC, Off, Off, ImgIndex, IMGLST_DISABLED | Flags);
				}

				// draw border
				if (Down)
				{
					pMemDC->Colour((Down) ? LC_LIGHT : LC_LOW, 24);
					pMemDC->Line(Px, Py, Px, 0);
					pMemDC->Line(Px, Py, 0, Py);

					pMemDC->Colour((Down) ? LC_LOW : LC_LIGHT, 24);
					pMemDC->Line(0, 0, 0, Py);
					pMemDC->Line(0, 0, Px, 0);
				}
			}
			else
			{
				// separator
				if (X() > Y())
				{
					int c = Y()/2-1;
					pMemDC->Colour(LC_LOW, 24);
					pMemDC->Line(2, c, Px-2, c);
					pMemDC->Colour(LC_LIGHT, 24);
					pMemDC->Line(2, c+1, Px-2, c+1);
				}
				else
				{
					int c = X()/2-1;
					pMemDC->Colour(LC_LOW, 24);
					pMemDC->Line(c, 2, c, Py-2);
					pMemDC->Colour(LC_LIGHT, 24);
					pMemDC->Line(c+1, 2, c+1, Py-2);
				}
			}

			pDC->Blt(0, 0, pMemDC);
		}

		DeleteObj(pMemDC);
	}
}

void GToolButton::Value(int b)
{
	switch (Type)
	{
		case TBT_PUSH:
		{
			// do nothing... can't set value
			break;
		}

		case TBT_TOGGLE:
		{
			SetDown(b);
			GView::Invalidate();
			break;
		}

		case TBT_RADIO:
		{
			if (GView::Parent AND b)
			{
				// Clear any other radio buttons that are down
				for (	GToolButton *But = (GToolButton*) GView::Parent->Children.Find(this);
						But AND But->GetId() >= 0;
						But = (GToolButton*) GView::Parent->Children.Next())
				{
					if (	But->Type == TBT_RADIO AND
						But != this AND
						But->Down)
					{
						But->SetDown(FALSE);
						But->GView::Invalidate();
					}
				}

				for (	GToolButton *But = (GToolButton*) GView::Parent->Children.Find(this);
						But AND But->GetId() >= 0;
						But = (GToolButton*) GView::Parent->Children.Prev())
				{
					if (	But->Type == TBT_RADIO AND
						But != this AND
						But->Down)
					{
						But->SetDown(FALSE);
						But->GView::Invalidate();
					}
				}
			}

			SetDown(b);
			GView::Invalidate();
			break;
		}
	}
}

void GToolButton::OnCommand(GMouse &m)
{
	if (GView::Parent)
	{
		((GToolBar*) GView::Parent)->PostCommand(GetId());
	}
}

void GToolButton::SendNotify(int Flags)
{
	GView *To = (Notify) ? Notify : GView::Parent;
	if (To)
	{
		BMessage Msg = CreateMsg(WM_CHANGE, (int)((GView*)this), Flags);
		To->OnEvent(&Msg);
	}
}

void GToolButton::OnMouseClick(GMouse &m)
{
	if (GetId() >= 0 AND Enabled())
	{
		switch (Type)
		{
			case TBT_PUSH:
			{
				bool Old = Down;

				Clicked = m.Down();
				Capture(m.Down());

				if (Old AND IsOver(m))
				{
					if (m.Left())
					{
						OnCommand(m);
					}
					SendNotify(m.Flags);
				}

				SetDown(m.Down());
				if (Old != Down)
				{
					GView::Invalidate();
				}
				break;
			}

			case TBT_TOGGLE:
			{
				if (m.Down())
				{
					if (m.Left())
					{
						Value(NOT Down);
						OnCommand(m);
					}
					SendNotify(m.Flags);
				}
				break;
			}

			case TBT_RADIO:
			{
				if (m.Down())
				{
					if (NOT Down AND m.Left())
					{
						Value(TRUE);
						OnCommand(m);
					}
					SendNotify(m.Flags);
				}
				break;
			}
		}
	}
}

void GToolButton::OnMouseMove(GMouse &m)
{
	if (GetParent())
	{
		((GToolBar*) GetParent())->PostDescription(this, GView::Name());
	}
}

void GToolButton::OnMouseEnter(GMouse &m)
{
	if (Clicked)
	{
		SetDown(TRUE);
		GView::Invalidate();
	}
	else
	{
		((GToolBar*) GView::Parent)->PostDescription(this, GView::Name());
	}
}

void GToolButton::OnMouseExit(GMouse &m)
{
	if (Clicked)
	{
		SetDown(FALSE);
		GView::Invalidate();
	}
	else
	{
		((GToolBar*) GView::Parent)->PostDescription(this, "");
	}
}

void GToolButton::SetDown(bool d)
{
	Down = d;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GToolBar::GToolBar()
{
	Name("LGI_Toolbar");
	Bx = By = 0;
	Sx = Sy = 10;
	Vertical = FALSE;

	LastIndex = 0;
	ImgList = 0;
	
	GetBView()->SetViewColor(B_TRANSPARENT_COLOR);
}

GToolBar::~GToolBar()
{
	DeleteObj(ImgList);
}

bool GToolBar::Pour(GComplexRegion &r)
{
	if (ImgList)
	{
		int PosX = BORDER_SHADE + BORDER_SPACER;
		int PosY = BORDER_SHADE + BORDER_SPACER;
		int SrcX = 0;
		int SrcY = 0;
		int EndX = 0;
		int EndY = 0;

		GRegion ButPos;
		GControl *But = (GControl*) Children.First();
		while (But)
		{
			ButPos = But->GetPos();

			if (But->GetId() == IDM_SEPARATOR)
			{
				// separator
				if (Vertical)
				{
					ButPos.ZOff(Bx+(2*BORDER_BUTTON)-1, BORDER_SEPARATOR+1);
				}
				else
				{
					ButPos.ZOff(BORDER_SEPARATOR+1, By+(2*BORDER_BUTTON)-1);
				}
			}
			else if (But->GetId() >= 0)
			{
				// other control
			}

			ButPos.Offset(PosX - ButPos.x1, PosY - ButPos.y1);

			if (But->GetId() == IDM_BREAK)
			{
				ButPos.ZOff(0, 0);
				if (Vertical)
				{
					PosX = EndX;
					PosY = BORDER_SHADE + BORDER_SPACER;
				}
				else
				{
					PosX = BORDER_SHADE + BORDER_SPACER;
					PosY = EndY;
				}
			}
			else
			{
				if (Vertical)
				{
					EndX = max(EndX, ButPos.x2 + 1);
					PosY = ButPos.y2 + 1;
				}
				else
				{
					PosX = ButPos.x2 + 1;
					EndY = max(EndY, ButPos.y2 + 1);
				}
			}

			But->SetPos(ButPos);
			But = (GControl*) Children.Next();
		}

		if (Vertical)
		{
			PosX = EndX;
		}
		else
		{
			PosY = EndY;
		}

		Sx = PosX + BORDER_SHADE + BORDER_SPACER;
		Sy = PosY + BORDER_SHADE + BORDER_SPACER;
		Pos.ZOff(Sx-1, Sy-1);

		GRegion *Best = FindLargest(r);
		if (Best)
		{
			Pos.Offset(Best->x1, Best->y1);
			Pos.Bound(Best);
			SetPos(Pos);

			return TRUE;
		}
	}

	return FALSE;
}

int GToolBar::PostCommand(int Cmd)
{
	if (GetParent())
	{
		// this needs to be an actual post message not a direct call to "OnEvent"
		// because if the GView wants to put up a modal dialog it will need to
		// be running from it's own loopers thread. And this is called from the
		// Mouse handler thread, the the windows thread.	
		
		// BMessage Msg = CreateMsg(WM_COMMAND, Cmd, 0);
		GMessage Msg(WM_COMMAND);
		Msg.AddInt32("Cmd", Cmd);
		return GetBWindow()->PostMessage(&Msg, GetParent()->GetBView());
	}
	return 0;
}

int GToolBar::PostDescription(GView *Ctrl, char *Text)
{
	if (GetParent())
	{
		BMessage Msg = CreateMsg(WM_DESCRIBE, (int) Ctrl, (int) Text);
		return GetParent()->OnEvent(&Msg);
	}
	return 0;
}

void GToolBar::OnMouseClick(GMouse &m)
{
}

void GToolBar::OnMouseEnter(GMouse &m)
{
}

void GToolBar::OnMouseExit(GMouse &m)
{
}

void GToolBar::OnMouseMove(GMouse &m)
{
}

void GToolBar::OnButtonClick(GToolButton *Btn)
{
}

int GToolBar::OnEvent(BMessage *Msg)
{
	switch (Msg->what)
	{
		case WM_CHANGE:
		{
			return GView::Parent->OnEvent(Msg);
		}
	}

	return GView::OnEvent(Msg);
}

void GToolBar::OnPaint(GSurface *pScr)
{
	GRegion r(0, 0, X()-1, Y()-1);

	LgiThinBorder(pScr, r, RAISED);
	pScr->Colour(LC_MED, 24);
	pScr->Rectangle(&r);
}

bool GToolBar::SetBitmap(char *File, int bx, int by)
{
	bool Status = FALSE;

	int Opt = GdcD->SetOption(GDC_IMPORT_8_AS_16, true);
	GSurface *pDC = LoadDC(File);
	GdcD->SetOption(GDC_IMPORT_8_AS_16, Opt);

	if (pDC)
	{
		Status = SetDC(pDC, bx, by);
		DeleteObj(pDC);
	}
	
	return Status;
}

bool GToolBar::SetDC(GSurface *pNewDC, int bx, int by)
{
	if (OwnImgList)
	{
		DeleteObj(ImgList);
	}

	if (pNewDC)
	{
		ImgList = NEW(GImageList(bx, by, pNewDC));
		if (ImgList)
		{
			Bx = bx;
			By = by;
			
			return TRUE;
		}
	}
	return FALSE;
}

bool GToolBar::SetImageList(GImageList *l, int bx, int by, bool Own)
{
	if (OwnImgList)
	{
		DeleteObj(ImgList);
	}
	
	ImgList = l;
	Bx = bx;
	By = by;
	OwnImgList = Own;
}

GToolButton *GToolBar::AppendButton(char *Tip, int Id, int Type, int Enabled, int IconId)
{
	GToolButton *But = NEW(GToolButton(Bx, By));
	if (But)
	{
		But->GView::Name(Tip);
		But->SetId(Id);
		But->Type = Type;
		But->Enabled(Enabled);
		
		if (IconId >= 0)
		{
			But->ImgIndex = IconId;
		}
		else
		{
			But->ImgIndex = LastIndex++;
		}
		
		Children.Insert(But);
		But->SetParent(this);
		GetBView()->AddChild(But->GetBView());
	}
	return But;
}

bool GToolBar::AppendSeparator()
{
	GToolButton *But = NEW(GToolButton(Bx, By));
	if (But)
	{
		But->SetId(IDM_SEPARATOR);
		Children.Insert(But);
		But->SetParent(this);
		GetBView()->AddChild(But->GetBView());
		
		return TRUE;
	}
	return FALSE;
}

bool GToolBar::AppendBreak()
{
	GToolButton *But = NEW(GToolButton(Bx, By));
	if (But)
	{
		But->SetId(IDM_BREAK);
		But->SetParent(this);
		Children.Insert(But);

		return TRUE;
	}
	return FALSE;
}

bool GToolBar::AppendControl(GView *Ctrl)
{
	bool Status = FALSE;
	if (Ctrl)
	{
		Ctrl->SetParent(this);
		Children.Insert(Ctrl);
		GetBView()->AddChild(Ctrl->GetBView());

		Status = TRUE;
	}
	return Status;
}

void GToolBar::Empty()
{
	for (GView *But = Children.First(); But; But = Children.Next())
	{
		DeleteObj(But);
	}
}

