/*hdr
**      FILE:           Gui.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           23/4/98
**      DESCRIPTION:    Memory subsystem
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdlib.h>

#include "SimpleGameSound.h"
#include "Roster.h"

#include "Lgi.h"
#include "DragAndDrop.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
struct GView::OsMouseInfo
{
	// View needing mouse info
	GView *Wnd;

	// Data
	int32 Clicks;
	int Sx, Sy;
	uint32 Buttons;
	BPoint Point;

	// Reason for thread
	bool Capture;
	bool MouseClick;
};

////////////////////////////////////////////////////////////////////////////////////////////////////
extern uint32 _lgi_mouse_keys();

class GMouseDoEnterExit : public GThread, public GSemaphore
{
	GView *Over;

public:
	GMouseDoEnterExit()
	{
		Over = 0;
		Run(true);
	}
	
	int Main()
	{
		// Run forever
		while (true)
		{
			if (Lock())
			{
				uint32 But;
				BPoint Pos;
				
				if (Over)
				{
					if (Over->Handle()->LockLooperWithTimeout(50 * 1000) == B_OK)
					{					
						Over->Handle()->GetMouse(&Pos, &But, false);
						Over->Handle()->UnlockLooper();
					}
					
					if (Pos.x < 0 OR
						Pos.y < 0 OR
						Pos.x >= Over->X() OR
						Pos.y >= Over->Y())
					{
						SendEvent(LGI_MOUSE_EXIT);
						Over = 0;
					}
				}
				
				Unlock();
			}
			
			// Go easy on the CPU
			LgiSleep(50);
		}
	}
	
	void SendEvent(uint32 Event)
	{
		if (Over)
		{
			uint32 But;
			BPoint Pos;
				
			if (Over->Handle()->LockLooperWithTimeout(300 * 1000) == B_OK)
			{					
				Over->Handle()->GetMouse(&Pos, &But, false);

				BMessage *Msg = new BMessage(Event);
				if (Msg)
				{
					Msg->AddPoint("pos", Pos);
					Msg->AddInt32("flags", But | _lgi_mouse_keys());
					Over->GetWindow()->PostMessage(Msg, Over->Handle());
				}

				Over->Handle()->UnlockLooper();
			}
		}
	}
	
	void SetMouseOver(GView *over)
	{
		if (Lock())
		{
			if (Over != over)
			{
				SendEvent(LGI_MOUSE_EXIT);
				Over = over;
				SendEvent(LGI_MOUSE_ENTER);
			}
			
			Unlock();
		}
	}
	
	void OnViewDelete(GView *View)
	{
		if (View == Over)
		{
			if (Lock())
			{
				Over = 0;
				Unlock();
			}
		}
	}

} EnterExitThread;

////////////////////////////////////////////////////////////////////////////////////////////////////
BViewRedir::BViewRedir(GView *wnd) :
	BView(	BRect(0, 0, 100, 100),
			"BViewRedir",
			B_FOLLOW_LEFT | B_FOLLOW_TOP,
			LGI_DRAW_VIEW_FLAGS)
{
	Wnd = wnd;
	ScrollView = 0;
	SetViewColor(B_TRANSPARENT_COLOR);
}

void BViewRedir::AttachedToWindow()
{
	Wnd->OnCreate();
}

void BViewRedir::DetachedFromWindow()
{
	Wnd->OnDestroy();
}

void BViewRedir::Draw(BRect UpdateRect)
{
	GdcViewDC DC(this);
	Wnd->_Paint(&DC);
	Window()->Sync();
}

void BViewRedir::FrameMoved(BPoint Point)
{
	if (NOT ScrollView)
	{
		Wnd->Sys_FrameMoved(Point);
	}
}

void BViewRedir::FrameResized(float width, float height)
{
	if (NOT ScrollView)
	{
		// Wnd->Pos.Dimension(width+1, height+1);
		Wnd->OnPosChange();
	}
}

void BViewRedir::Pulse()
{
	Wnd->OnPulse();
}

void BViewRedir::MessageReceived(BMessage *message)
{
	Wnd->OnEvent(message);
}

void BViewRedir::MakeFocus(bool f = true)
{
	BView::MakeFocus(f);
	Wnd->OnFocus(f);
}

void BViewRedir::KeyDown(const char *bytes, int32 numBytes)
{
	Wnd->Sys_KeyDown(bytes, numBytes);
}

void BViewRedir::KeyUp(const char *bytes, int32 numBytes)
{
	Wnd->Sys_KeyUp(bytes, numBytes);
}

void BViewRedir::MouseDown(BPoint point)
{
	Wnd->Sys_MouseDown(point);
}

void BViewRedir::MouseMoved(BPoint point, uint32 transit, const BMessage *message)
{
	EnterExitThread.SetMouseOver(Wnd);
	
	if (NOT Wnd->IsCapturing())
	{
		Wnd->Sys_MouseMoved(point, transit, message);
	}
}

bool BViewRedir::QuitRequested()
{
	GWindow *App = dynamic_cast<GWindow*>(Wnd);
	if (App)
	{
		return App->QuitRequested();
	}
	return true;
}

//////////////////////////////////////////////////////////////////////////////
GScroll::GScroll(GScrollView *notify, orientation posture) :
	BScrollBar(	(posture==B_VERTICAL)?BRect(0, 0, B_V_SCROLL_BAR_WIDTH, 100):BRect(0, 0, 100, B_H_SCROLL_BAR_HEIGHT),
				"Scrollbar",
				NULL,
				0, 99,
				posture)
{
	Notify = notify;
	Hide();
}

void GScroll::ValueChanged(float newValue)
{
	if (Notify)
	{
		Notify->OnNotify(this, newValue);
	}
}

///////////////////////////////////////////////////////////////////////////////////////
GScrollView::GScrollView(	char *name,
							GView *target,
							uint32 resizingMode,
							uint32 flags,
							bool horizontal,
							bool vertical,
							int border) :
				BView(		BRect(0, 0, 100, 100),
							name,
							resizingMode,
							flags | B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE)
{
	// Border
	Border = border;

	// Setup scrollbars
	X = Y = 0;
	if (horizontal)
	{
		X = NEW(GScroll(this, B_HORIZONTAL));
		if (X)
		{
			AddChild(X);
		}
	}

	if (vertical)
	{
		Y = NEW(GScroll(this, B_VERTICAL));
		if (Y)
		{
			AddChild(Y);
		}
	}

	// Setup target view
	Target = target;
	if (Target AND Target->_View)
	{
		bool Vis = NOT Target->_View->IsHidden();
		BView *Parent = Target->_View->Parent();
		BView *LockView = (Parent) ? Parent : Target->_View;
		bool Lock = LockView->LockLooper();

		// make our position the target's position
		Target->_View->SetResizingMode(B_FOLLOW_LEFT | B_FOLLOW_TOP);
		GRegion r(Target->_View->Frame());

		if (Parent)
		{
			// Replace the target view with ourselves
			Parent->RemoveChild(Target->_View);
			Parent->AddChild(this);
		}

		// attach the target to us
		AddChild(Target->_View);

		// set our position
		Target->_View->MoveTo(0, 0);
		SetPos(r);
		if (Vis)
		{
			if (IsHidden())
			{
				Show();
			}
		}
		else
		{
			if (NOT IsHidden())
			{
				Hide();
			}
		}

		// Other setup
		SetViewColor(B_TRANSPARENT_COLOR);

		if (Lock) LockView->UnlockLooper();
	}
}

GScrollView::~GScrollView()
{
	if (Parent())
	{
		bool Lock = LockLooper();
		if (Target AND Target->_View)
		{
			bool HasTarget = false;
			for (int i=0; i<CountChildren(); i++)
			{
				if (ChildAt(i) == Target->_View)
				{
					HasTarget = true;
					break;
				}
			}
		
			if (HasTarget)
			{
				GRegion r(Frame());
				Target->_View->RemoveSelf();
				Parent()->AddChild(Target->_View);
				Target->GView::SetPos(r);

				switch (Border)
				{
					case SUNKEN:
						Target->Sunken(true);
						break;
					case RAISED:
						Target->Raised(true);
						break;
				}
			}
			
			Target = 0;
		}
		
		if (Lock) UnlockLooper();
		RemoveSelf();
	}
}

BScrollBar *GScrollView::ScrollBar(orientation posture)
{
	if (posture == B_VERTICAL)
	{
		return Y;
	}
	else if (posture == B_HORIZONTAL)
	{
		return X;
	}

	return 0;
}

void GScrollView::Draw(BRect update)
{
	GdcViewDC DC(this);
	GRegion r(Frame());
	r.Offset(-r.x1, -r.y1);
	
	switch (Border)
	{
		case SUNKEN:
		case RAISED:
		{
			LgiWideBorder(&DC, r, Border);
			break;
		}
	}

	// printf("\tGScrollView::Draw(%p, %i)\n", this, Border==SUNKEN);
}

void GScrollView::FrameResized(float width, float height)
{
	BView::FrameResized(width, height);
	GRegion r(Frame());
	GRegion tr(Target->_View->Frame());

	/*
	printf("\tGScrollView::FrameResized pos(%i, %i, %i, %i) targetpos(%i, %i, %i, %i) targetvis(%i)\n", 
			r.x1, r.y1, r.x2, r.y2,
			tr.x1, tr.y1, tr.x2, tr.y2,
			NOT Target->IsHidden());
	*/

	r.Offset(-r.x1, -r.y1);

	
	// Border
	switch (Border)
	{
		case SUNKEN:
		case RAISED:
		{
			r.Size(2, 2);
			break;
		}
	}
	
	// Scrollbars
	if (X)
	{
		bool Lock = X->LockLooper();
		if (r.Y() > B_H_SCROLL_BAR_HEIGHT)
		{
			X->MoveTo(r.x1 - 1, r.y2 - B_H_SCROLL_BAR_HEIGHT);
			X->ResizeTo(r.X() - ((Y) ? B_V_SCROLL_BAR_WIDTH : 0), B_H_SCROLL_BAR_HEIGHT);
			r.y2 -= B_H_SCROLL_BAR_HEIGHT + 1;
			X->Show();
		}
		else if (NOT X->IsHidden())
		{
			X->Hide();
		}
		if (Lock) X->UnlockLooper();
	}
	if (Y)
	{
		bool Lock = Y->LockLooper();
		if (r.X() > B_V_SCROLL_BAR_WIDTH)
		{
			Y->MoveTo(r.x2 - B_V_SCROLL_BAR_WIDTH, r.y1 - 1);
			Y->ResizeTo(B_V_SCROLL_BAR_WIDTH, r.Y());
			r.x2 -= B_V_SCROLL_BAR_WIDTH + 1;
			Y->Show();
		}
		else if (NOT Y->IsHidden())
		{
			Y->Hide();
		}

		if (Lock) Y->UnlockLooper();
	}
	
	// View
	if (Target AND
		r.Valid())
	{
		Target->GView::SetPos(r);
		Target->OnPosChange();
	}
}

void GScrollView::OnNotify(GScroll *Bar, float Value)
{
	if (Target)
	{
		bool Hor = (int)X == (int)Bar;
		BMessage *m = NEW(BMessage(Hor ? WM_HSCROLL : WM_VSCROLL));
		m->AddInt32("a", Value);
		Target->GetWindow()->PostMessage(m, Target->_View);
	}
}

GRegion &GScrollView::GetPos()
{
	bool Lock = LockLooper();
	GRegion n(Frame());
	if (Lock) UnlockLooper();
	return n;
}

void GScrollView::SetPos(GRegion &r)
{
	Pos = r;

	bool Lock = LockLooper();
	MoveTo(Pos.x1, Pos.y1);
	ResizeTo(Pos.X()-1, Pos.Y()-1);
	if (Lock) UnlockLooper();
}

void GScrollView::AttachedToWindow()
{
	/*
	bool Lock = LockLooper();
	MoveTo(Pos.x1, Pos.y1);
	ResizeTo(Pos.X()-1, Pos.Y()-1);
	FrameResized(Pos.X()-1, Pos.Y()-1);
	if (Lock) UnlockLooper();
	*/
}

////////////////////////////////////////////////////////////////////////////////////////////////////
void _lgi_yield()
{
	BLooper *Looper = BLooper::LooperForThread(find_thread(0));
	if (Looper)
	{
		BWindow *Wnd = dynamic_cast<BWindow*>(Looper);
		if (Wnd)
		{
			if (Wnd->Lock())
			{
				Wnd->UpdateIfNeeded();
				Wnd->Unlock();
			}
		}
	}
}

uint32 _lgi_mouse_keys()
{
	int Keys = modifiers();
	
	return	(TestFlag(Keys, B_SHIFT_KEY) ? LGI_VMOUSE_SHIFT : 0) |
			(TestFlag(Keys, B_CONTROL_KEY) ? LGI_VMOUSE_ALT : 0) |
			(TestFlag(Keys, B_COMMAND_KEY) ? LGI_VMOUSE_SHIFT : 0);
}

void _lgi_assert(bool b, char *test, char *file, int line)
{
	if (NOT b)
	{
		LgiMsg(	0,
				"Assert failed, file: %s, line: %i\n"
				"%s\n",
				"Debug",
				MB_OK,
				file, line, test);
		exit(-1);					
	}
}

BMessage NewMsg(int m, int a, int b)
{
	BMessage Msg(m);
	Msg.AddInt32("a", a);
	Msg.AddInt32("b", b);
	return Msg;
}

////////////////////////////////////////////////////////////////////////////
GMessage CreateMsg(int m, int a, int b)
{
	GMessage Msg(m);
	Msg.AddInt32("a", a);
	Msg.AddInt32("b", b);
	return Msg;
}

int MsgA(GMessage *m)
{
	int32 i=0;
	if (m) m->FindInt32("a", &i);
	return i;
}

int MsgB(GMessage *m)
{
	int32 i=0;
	if (m) m->FindInt32("b", &i);
	return i;
}

////////////////////////////////////////////////////////////////////////////
bool LgiIsKeyDown(int Key)
{
	key_info KeyInfo;
	if (get_key_info(&KeyInfo) == B_OK)
	{
		switch (Key)
		{
		}
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
GFlags::GFlags()
{
	WndFlags = GWF_VISIBLE | GWF_ENABLED;
}

bool GFlags::Enabled()
{
	return TestFlag(WndFlags, GWF_ENABLED);
}

void GFlags::Enabled(bool i)
{
	if (i) SetFlag(WndFlags, GWF_ENABLED);
	else ClearFlag(WndFlags, GWF_ENABLED);
}

bool GFlags::Visible()
{
	return TestFlag(WndFlags, GWF_VISIBLE);
}

void GFlags::Visible(bool i)
{
	if (i) SetFlag(WndFlags, GWF_VISIBLE);
	else ClearFlag(WndFlags, GWF_VISIBLE);
}

bool GFlags::Focus()
{
	return TestFlag(WndFlags, GWF_FOCUS);
}

void GFlags::Focus(bool i)
{
	if (i) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);
}

bool GFlags::DropTarget()
{
	return TestFlag(WndFlags, GWF_DROP_TARGET);
}

void GFlags::DropTarget(bool i)
{
	if (i) SetFlag(WndFlags, GWF_DROP_TARGET);
	else ClearFlag(WndFlags, GWF_DROP_TARGET);
}

#define GWF_ALL (GWF_SUNKEN | GWF_FLAT | GWF_RAISED)

bool GFlags::Sunken()
{
	return TestFlag(WndFlags, GWF_SUNKEN);
}

void GFlags::Sunken(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_SUNKEN);
}

bool GFlags::Flat()
{
	return TestFlag(WndFlags, GWF_FLAT);
}

void GFlags::Flat(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_FLAT);
}

bool GFlags::Raised()
{
	return TestFlag(WndFlags, GWF_RAISED);
}

void GFlags::Raised(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_RAISED);
}

/*
bool GFlags::Border()
{
	return TestFlag(WndFlags, GWF_BORDER);
}

void GFlags::Border(bool i)
{
	if (i) SetFlag(WndFlags, GWF_BORDER);
	else ClearFlag(WndFlags, GWF_BORDER);
}
*/


////////////////////////////////////////////////////////////////////////////////////////////////////
GView::GView(BView *view)
{
	_View = view;
	// _View->SetViewColor(B_TRANSPARENT_32_BIT);
	WndFlags = GWF_ENABLED | GWF_VISIBLE;
	Pos.ZOff(0, 0);
	Parent = 0;
	Notify = 0;

	_MouseInfo = 0;
	_CaptureThread = 0;
	_PulseThread = 0;
	_PulseRate = -1;
	_BorderSize = 2;
}

GView::~GView()
{
	EnterExitThread.OnViewDelete(this);
	
	if (_CaptureThread)
	{
		kill_thread(_CaptureThread);
	}
	if (_PulseThread)
	{
		kill_thread(_PulseThread);
	}
	
	DeleteObj(_MouseInfo);

	BViewRedir *Redir = dynamic_cast<BViewRedir*>(_View);
	DeleteObj(Redir);
	_View = 0;
	WndFlags = 0;
}

#define ADJ_LEFT			1
#define ADJ_RIGHT			2
#define ADJ_UP				3
#define ADJ_DOWN			4

extern int IsAdjacent(GRegion &a, GRegion &b);
int IsAdjacent(GRegion &a, GRegion &b)
{
	if (	(a.x1 == b.x2 + 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_LEFT;
	}

	if (	(a.x2 == b.x1 - 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_RIGHT;
	}

	if (	(a.y1 == b.y2 + 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_UP;
	}

	if (	(a.y2 == b.y1 - 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_DOWN;
	}

	return 0;
}

extern GRegion JoinAdjacent(GRegion &a, GRegion &b, int Adj);
GRegion JoinAdjacent(GRegion &a, GRegion &b, int Adj)
{
	GRegion t;

	switch (Adj)
	{
		case ADJ_LEFT:
		case ADJ_RIGHT:
		{
			t.y1 = max(a.y1, b.y1);
			t.y2 = min(a.y2, b.y2);
			t.x1 = min(a.x1, b.x1);
			t.x2 = max(a.x2, b.x2);
			break;
		}
		case ADJ_UP:
		case ADJ_DOWN:
		{
			t.y1 = min(a.y1, b.y1);
			t.y2 = max(a.y2, b.y2);
			t.x1 = max(a.x1, b.x1);
			t.x2 = min(a.x2, b.x2);
			break;
		}
	}

	return t;
}

void GView::_Paint(GSurface *pDC)
{
	GRegion r(0, 0, Pos.X()-1, Pos.Y()-1);
	bool HasEdge = (Sunken() OR Raised()) AND _BorderSize;
	if (HasEdge)
	{
		if (_BorderSize == 2)
		{
			LgiWideBorder(pDC, r, Sunken() ? SUNKEN : RAISED);
		}
		else if (_BorderSize == 1)
		{
			LgiThinBorder(pDC, r, Sunken() ? SUNKEN : RAISED);
		}

		pDC->SetOrigin(-r.x1, -r.y1);
		r.Offset(-r.x1, -r.y1);
		pDC->_SetClient(&r);
	}
	
	OnPaint(pDC);
	
	if (HasEdge)
	{
		pDC->ClipRgn(0);
	}
}

GRegion *GView::FindLargest(GComplexRegion &r)
{
	int Pixels = 0;
	GRegion *Best = 0;
	static GRegion Final;

	// do initial run through the list to find largest single region
	for (GRegion *i = r.First(); i; i = r.Next())
	{
		int Pix = i->X() * i->Y();
		if (Pix > Pixels)
		{
			Pixels = Pix;
			Best = i;
		}
	}

	if (Best)
	{
		Final = *Best;
		Pixels = Final.X() * Final.Y();

		int LastPixels = Pixels;
		GRegion LastRgn = Final;
		int ThisPixels = Pixels;
		GRegion ThisRgn = Final;

		GComplexRegion TempList;
		for (GRegion *i = r.First(); i; i = r.Next())
		{
			TempList.Union(i);
		}
		TempList.Subtract(Best);

		do
		{
			LastPixels = ThisPixels;
			LastRgn = ThisRgn;

			// search for adjoining rectangles that maybe we can add
			for (GRegion *i = TempList.First(); i; i = TempList.Next())
			{
				int Adj = IsAdjacent(ThisRgn, *i);
				if (Adj)
				{
					GRegion t = JoinAdjacent(ThisRgn, *i, Adj);
					int Pix = t.X() * t.Y();
					if (Pix > ThisPixels)
					{
						ThisPixels = Pix;
						ThisRgn = t;

						TempList.Subtract(i);
					}
				}
			}
		} while (LastPixels < ThisPixels);

		Final = ThisRgn;
	}

	return &Final;
}

GRegion *GView::FindSmallestFit(GComplexRegion &r, int Sx, int Sy)
{
	int X = 1000000;
	int Y = 1000000;
	GRegion *Best = 0;

	for (GRegion *i = r.First(); i; i = r.Next())
	{
		if ((i->X() >= Sx AND i->Y() >= Sy) AND
			(i->X() < X OR i->Y() < Y))
		{
			X = i->X();
			Y = i->Y();
			Best = i;
		}
	}

	return Best;
}

GRegion *GView::FindLargestEdge(GComplexRegion &r, int Edge)
{
	GRegion *Best = 0;

	for (GRegion *i = r.First(); i; i = r.Next())
	{
		if (NOT Best)
		{
			Best = i;
		}

		if ((Edge & GV_EDGE_TOP) AND (i->y1 < Best->y1) OR
			(Edge & GV_EDGE_RIGHT) AND (i->x2 > Best->x2) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 > Best->y2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 < Best->x1))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_TOP) AND (i->y1 == Best->y1) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 == Best->y2)) AND
			(i->X() > Best->X()))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_RIGHT) AND (i->x2 == Best->x2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 == Best->x1)) AND
			(i->Y() > Best->Y()))
		{
			Best = i;
		}
	}

	return Best;
}

int GView::GetCtrlValue(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Value() : -1;
}

void GView::SetCtrlValue(int Id, int i)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Value(i);
	}
}

char *GView::GetCtrlName(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Name() : 0;
}

void GView::SetCtrlName(int Id, char *s)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Name(s);
	}
}

bool GView::GetCtrlEnabled(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Enabled() : 0;
}

void GView::SetCtrlEnabled(int Id, bool Enabled)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Enabled(Enabled);
	}
}

void GView::SetId(int i)
{
	CtrlId = i;
}

bool GView::Sunken()
{
	return GFlags::Sunken();
}

void GView::Sunken(bool i)
{
	GFlags::Sunken(i);
}

bool GView::Flat()
{
	return GFlags::Flat();
}

void GView::Flat(bool i)
{
	GFlags::Flat(i);
}

bool GView::Raised()
{
	return GFlags::Raised();
}

void GView::Raised(bool i)
{
	GFlags::Raised(i);
}

void GView::Quit(bool DontDelete)
{
	BWindow *Wnd = GetWindow();
	if (Wnd)
	{
		Wnd->PostMessage(NEW(BMessage(B_QUIT_REQUESTED)));
	}
}

bool GView::AttachChildren()
{
	bool Status = true;

	for (GView *c=Children.First(); c; c=Children.Next())
	{
		Status |= c->Attach(this);
	}

	return Status;
}

GRegion &GView::GetClient()
{
	/*
	bool Lock = GetBView()->LockLooper();
	BRect rc = GetBView()->Frame();
	if (Lock) GetBView()->UnlockLooper();

	static GRegion Client;
	Client.Set(0, 0, rc.right-rc.left, rc.bottom-rc.top);
	if (Sunken() OR Raised()) Client.Size(_BorderSize, _BorderSize);
	*/

	static GRegion Client;
	Client.ZOff(Pos.X()-1, Pos.Y()-1);
	if (Sunken() OR Raised()) Client.Size(_BorderSize, _BorderSize);
	
	return Client;
}

GFont *GView::GetFont()
{
	return SysFont;
}

void GView::SetFont(GFont *Fnt)
{
}

bool GView::IsOver(GMouse &m)
{
	return	(m.x >= 0) AND
			(m.y >= 0) AND
			(m.x < Pos.X()) AND
			(m.y < Pos.Y());
}

GView *GView::WindowFromPoint(int x, int y)
{
	for (GView *c = Children.First(); c; c = Children.Next())
	{
		if (c->Pos.Overlap(x, y))
		{
			return c;
		}
	}

	if (x >= 0 AND y >= 0 AND x < Pos.X() AND y < Pos.Y())
	{
		return this;
	}

	return NULL;
}

void GView::OnPosChange()
{
}

bool GView::SetPos(GRegion &p, bool Repaint)
{
	if (_View)
	{
		bool Lock = _View->LockLooper();
		_View->MoveTo(p.x1, p.y1);
		_View->ResizeTo(p.X()-1, p.Y()-1);
		
		if (Repaint)
		{
			_View->Invalidate();
		}
		
		if (Lock ) _View->UnlockLooper();
	}
	Pos = p;
	
	return TRUE;
}

class GuiLocker
{
	BHandler *Handler;
	bool Lock;

public:
	GuiLocker(BHandler *handler)
	{
		Lock = 0;
		Handler = handler;
		if (Handler)
		{
			Lock = Handler->LockLooperWithTimeout(20000) == B_OK;
		}
	}
	
	~GuiLocker()
	{
		if (Handler AND Lock)
		{
			Handler->UnlockLooper();
			Lock = 0;
		}
	}
	
	bool Locked()
	{
		return Lock;
	}
};

bool GView::Invalidate(GRegion *r, bool Repaint)
{
	// GuiLocker Lock(_View);
	if (_View->LockLooperWithTimeout(1000 * 1000) == B_OK) // 1 second
	{
		if (r)
		{
			GRegion c = GetClient();
			BRect Rc = *r;
			Rc.OffsetBy(c.x1, c.y1);
			_View->Invalidate(Rc);
		}
		else
		{
			_View->Invalidate();
		}
		
		if (Repaint)
		{
			BWindow *Wnd = _View->Window();
			if (Wnd)
			{
				thread_id MyThread = find_thread(0);
				thread_id WndThread = Wnd->Thread();
				
				if (MyThread == WndThread)
				{
					// in thread
					Wnd->UpdateIfNeeded();
				}
				else
				{
					// out of thread
					if (r)
					{
						BRect Rc = *r;
						BRegion *Rgn = NEW(BRegion(Rc));
						_View->ConstrainClippingRegion(Rgn);
						_View->Draw(Rc);
						_View->ConstrainClippingRegion(NULL);
					}
					else
					{
						BRect Rc = _View->Frame();
						_View->Draw(Rc);
					}
				}
			}
		}
		
		_View->UnlockLooper();
	}
	else
	{
		// printf("******** Invalidate couldn't lock ********\n");
	}
	
	return FALSE;
}

bool GView::InvalidateFrame(GRegion *r, bool Repaint)
{
	return FALSE;
}

long _lgi_pulse_thread(void *ptr)
{
	GView *Wnd = (GView*) ptr;

	if (Wnd)
	{
		while (Wnd->_PulseRate > 0)
		{
			LgiSleep(Wnd->_PulseRate);
			
			if (Wnd->Handle()->LockLooperWithTimeout(1000) == B_OK)
			{
				Wnd->GetWindow()->PostMessage(WM_PULSE, Wnd->Handle());
				Wnd->Handle()->UnlockLooper();
			}
		}
		
		Wnd->_PulseRate = -1;
		Wnd->_PulseThread = 0;
	}
	
	return 0;
}

void GView::SetPulse(int Length)
{
	if (_View->LockLooper())
	{
		if (_PulseThread)
		{
			kill_thread(_PulseThread);
			_PulseThread = 0;
		}
	
		_PulseRate = Length;
		
		if (Length > 0)
		{
			_PulseThread = spawn_thread(_lgi_pulse_thread, "_lgi_pulse_thread", B_LOW_PRIORITY, this);
			if (_PulseThread > 0)
			{
				resume_thread(_PulseThread);
			}
		}
		
		_View->UnlockLooper();
	}
}

BView *SearchChildren(BView *Parent, BPoint &Cursor)
{
	BView *Ret = 0;
	int32 Children = Parent->CountChildren();
	for (int i=0; i<Children; i++)
	{
		BView *View = Parent->ChildAt(i);
		if (View)
		{
			BRect r = View->Frame();
			if (Cursor.x >= r.left AND
				Cursor.y >= r.top AND
				Cursor.x < r.right AND
				Cursor.y < r.bottom)
			{
				BPoint Pt(Cursor.x-r.left, Cursor.y-r.top);
				BView *Child = SearchChildren(View, Pt);

				Cursor = Pt;
				if (Child)
				{
					Ret = Child;
				}
				else
				{
					Ret = View;
				}
				break;
			}
		}
	}

	return Ret;
}

bool GView::InThread()
{
	return find_thread(0) == GetWindow()->Thread();
}

bool GView::PostEvent(int Cmd, int a, int b)
{
	bool Status = false;

	if (GetWindow() AND
		GetWindow()->Lock())
	{
		BMessage *Msg = NEW(BMessage(Cmd));
		if (Msg)
		{
			Msg->AddInt32("a", a);
			Msg->AddInt32("b", b);
			Status = GetWindow()->PostMessage(Msg, Handle());
			DeleteObj(Msg);
		}
		GetWindow()->Unlock();
	}
	
	return Status;
}

int GView::OnEvent(BMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		// These are internal messages sent to a View to get it to
		// handle it's mouse click in thread.
		case LGI_MOUSE_CLICK:
		case LGI_MOUSE_MOVE:
		case LGI_MOUSE_ENTER:
		case LGI_MOUSE_EXIT:
		{
			BPoint p;
			GMouse m;
			if (Msg->FindPoint("pos", &p) == B_OK AND
				Msg->FindInt32("flags", (int32*)&m.Flags) == B_OK)
			{
				m.x = p.x;
				m.y = p.y;
				
				switch (MsgCode(Msg))
				{
					case LGI_MOUSE_CLICK:
						//printf("LGI_MOUSE_CLICK\n");
						OnMouseClick(m); break;
					case LGI_MOUSE_MOVE:
						//printf("LGI_MOUSE_MOVE\n");
						OnMouseMove(m); break;
					case LGI_MOUSE_ENTER:
						//printf("LGI_MOUSE_ENTER\n");
						OnMouseEnter(m); break;
					case LGI_MOUSE_EXIT:
						//printf("LGI_MOUSE_EXIT\n");
						OnMouseExit(m); break;
				}
			}
			break;
		}
		case B_SIMPLE_DATA:
		{
			if (GetParent())
			{
				GetParent()->OnEvent(Msg);
			}
			break;
		}
		case B_MOUSE_WHEEL_CHANGED:
		{
			float Wheel = 0.0;
			if (Msg->FindFloat("be:wheel_delta_y", &Wheel) == B_OK)
			{
				OnMouseWheel(Wheel * 3.0);
			}
			break;
		}
		case WM_PULSE:
		{
			OnPulse();
			break;
		}
		case WM_DRAG_DROP:
		{
			GDragDropSource *Source = 0;
			if (Msg->FindPointer("GDragDropSource", (void**) &Source) == B_OK)
			{
				BPoint Cursor;
				uint32 Buttons;
				Handle()->GetMouse(&Cursor, &Buttons);
			
				GDragDropTarget *Target = 0;
				BView *TargetView = SearchChildren(_View, Cursor);

				GdcPt2 MousePt(Cursor.x, Cursor.y);
				Target = dynamic_cast<GDragDropTarget*>(TargetView);
				if (Target)
				{
					if (Target->WillAccept(Source, MousePt, 0) != DROPEFFECT_NONE)
					{
						uchar *Data = 0;
						int Size = 0;
						if (Source->GetData(&Data, &Size))
						{
							Target->OnDrop(Data, Size, MousePt, 0);
						}
					}
				}
			}
			break;
		}
		case WM_CHANGE:
		{
			GView *Ctrl = dynamic_cast<GView*>((GView*) MsgA(Msg));
			if (Ctrl)
			{
				OnNotify(Ctrl, MsgB(Msg));
			}
			break;
		}
		case WM_COMMAND:
		{
			int32 a=MsgA(Msg);
			return OnCommand(a&0xFFFF, a>>16, (OsView) MsgB(Msg));
		}
		default:
		{
			BMessage *Msg = GetWindow()->CurrentMessage();
			if (Msg AND Msg->WasDropped())
			{
				Msg->PrintToStream();
			}
			break;
		}
	}

	return 0;
}

int GView::OnNotify(GView *Ctrl, int Flags)
{
	if (Ctrl AND Parent)
	{
		// default behaviour is just to pass the 
		// notification up to the parent
		return Parent->OnNotify(Ctrl, Flags);
	}

	return 0;
}

bool GView::Name(char *n)
{
	bool Status = GObject::Name(n);
	_View->SetName(n);
	return TRUE;
}

char *GView::Name()
{
	return GObject::Name();
}

bool GView::Enabled()
{
	return TestFlag(WndFlags, GWF_ENABLED);
}

void GView::Enabled(bool e)
{
	if (e) SetFlag(WndFlags, GWF_ENABLED);
	else ClearFlag(WndFlags, GWF_ENABLED);
	
	BControl *Ctrl = dynamic_cast<BControl*>(_View);
	if (Ctrl)
	{
		bool Lock = Ctrl->LockLooper();
		Ctrl->SetEnabled(e);
		if (Lock) Ctrl->UnlockLooper();
	}

	Invalidate();
}

bool GView::Visible()
{
	return GFlags::Visible();
}

void GView::Visible(bool v)
{
	if (_View)
	{
		bool Lock = _View->LockLooper();
		if (v)
		{
			if (NOT GFlags::Visible())
			{
				_View->Show();
			}
		}
		else
		{
			if (GFlags::Visible())
			{
				_View->Hide();
			}
		}
		if (Lock) _View->UnlockLooper();
	}

	GFlags::Visible(v);
}

bool GView::Focus()
{
	bool f = false;
	BWindow *Wnd = GetWindow();

	if (Wnd AND _View AND _View->LockLooper())
	{
		BView *Cur = Wnd->CurrentFocus();
		f = (_View == Cur);
		if (f) SetFlag(WndFlags, GWF_FOCUS);
		else ClearFlag(WndFlags, GWF_FOCUS);

		_View->UnlockLooper();
	}

	return TestFlag(WndFlags, GWF_FOCUS);
}

void GView::Focus(bool f)
{
	if (f) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);

	if (_View AND _View->LockLooper())
	{
		_View->MakeFocus(f);
		_View->UnlockLooper();
	}
}

bool GView::GetMouse(GMouse &m, bool ScreenCoords)
{
	if (Handle()->LockLooper())
	{
		// get mouse state
		BPoint Cursor;
		uint32 Buttons;
		
		Handle()->GetMouse(&Cursor, &Buttons);
		if (ScreenCoords)
		{
			Handle()->ConvertToScreen(&Cursor);

			// fudge???
			Cursor.x += 20;
			Cursor.y += 20;
		}
		
		// position
		m.x = Cursor.x;
		m.y = Cursor.y;

		// buttons
		m.Flags =	(TestFlag(Buttons, B_PRIMARY_MOUSE_BUTTON) ? MK_LEFT : 0) |
					(TestFlag(Buttons, B_TERTIARY_MOUSE_BUTTON) ? MK_MIDDLE : 0) |
					(TestFlag(Buttons, B_SECONDARY_MOUSE_BUTTON) ? MK_RIGHT : 0);

		// key states
		key_info KeyInfo;
		if (get_key_info(&KeyInfo) == B_OK)
		{
			m.Flags |=	(TestFlag(KeyInfo.modifiers, B_CONTROL_KEY) ? MK_CTRL : 0) |
						(TestFlag(KeyInfo.modifiers, B_MENU_KEY) ? MK_ALT : 0) |
						(TestFlag(KeyInfo.modifiers, B_SHIFT_KEY) ? MK_SHIFT : 0);
		}
		
		Handle()->UnlockLooper();
		
		return true;
	};

	return false;
}

int LastFunctionKey = 0;
void GView::Sys_KeyDown(const char *bytes, int32 numBytes)
{
	// printf("GView::Sys_KeyDown(%p, %i)\n", bytes, numBytes);

	key_info Info;
	get_key_info(&Info);

	for (int i=0; i<numBytes; i++)
	{
		GKey k;
		
		if (bytes[i] == B_FUNCTION_KEY)
		{
			k.c = 0;
			BMessage *msg = GetWindow()->CurrentMessage(); 
			if (msg)
			{ 
				int32 key; 
				msg->FindInt32("key", &key); 
				switch (key)
				{ 
					case B_F1_KEY: 
						k.c = VK_F1;
						break;
					case B_F2_KEY: 
						k.c = VK_F2;
						break;
					case B_F3_KEY: 
						k.c = VK_F3;
						break;
					case B_F4_KEY: 
						k.c = VK_F4;
						break;
					case B_F5_KEY: 
						k.c = VK_F5;
						break;
					case B_F6_KEY: 
						k.c = VK_F6;
						break;
					case B_F7_KEY: 
						k.c = VK_F7;
						break;
					case B_F8_KEY: 
						k.c = VK_F8;
						break;
					case B_F9_KEY: 
						k.c = VK_F9;
						break;
					case B_F10_KEY: 
						k.c = VK_F10;
						break;
					case B_F11_KEY: 
						k.c = VK_F11;
						break;
					case B_F12_KEY: 
						k.c = VK_F12;
						break;
					default:
						k.c = key;
						break;
				}
			}

			k.IsChar = false;
			LastFunctionKey = k.c;
		}
		else
		{
			k.c = bytes[i];
			k.IsChar = ((k.c < 127) AND (k.c >= ' ')) OR
						(k.c == 9) OR
						(k.c == 8) OR
						(k.c == 10);
		}
		
		k.Flags = modifiers();
		k.Data = 0;
		k.Down = TRUE;
		
		OnKey(k);
	}
}

void GView::Sys_KeyUp(const char *bytes, int32 numBytes)
{
	key_info Info;
	get_key_info(&Info);

	for (int i=0; i<numBytes; i++)
	{
		GKey k;
		
		k.c = bytes[i];
		k.Flags = modifiers();
		k.Data = 0;
		k.Down = FALSE;
		k.IsChar = 0;
		
		OnKey(k);
	}
}

int LastX = -DOUBLE_CLICK_THRESHOLD, LastY = -DOUBLE_CLICK_THRESHOLD;

long _lgi_mouse_thread(GView::OsMouseInfo *Info)
{
// printf("_lgi_mouse_thread: start\n");

	if (Info AND Info->Wnd)
	{
		bool Over = true;
		uint32 LastButton = Info->Buttons;

		// watch the mouse while the button is down
		while (Info->Buttons)
		{
			if (Info->Wnd->Handle()->LockLooper())
			{
				BPoint p;
				LastButton = Info->Buttons;
				Info->Wnd->Handle()->GetMouse(&p, &Info->Buttons, FALSE);
				if (p.x != Info->Point.x OR p.y != Info->Point.y)
				{
					// Position has changed...
					Info->Point = p;

					// Do captured mouse move...
					Over =	(Info->Point.x >= 0) AND
							(Info->Point.y >= 0) AND
							(Info->Point.x < Info->Sx) AND
							(Info->Point.y < Info->Sy);

					if (Info->Capture)
					{
						// Mouse is captured, send events mouse move events
						// While the capturing is on the BViewRedir doesn't
						// send events because we do it here.
						BMessage *Msg = new BMessage(LGI_MOUSE_MOVE);
						if (Msg)
						{
							Msg->AddPoint("pos", Info->Point);
							Msg->AddInt32("flags", Info->Buttons | _lgi_mouse_keys());
// printf("Captured move: %.0f,%.0f %08.8X\n", Info->Point.x, Info->Point.y, Info->Buttons);
							Info->Wnd->GetWindow()->PostMessage(Msg, Info->Wnd->Handle());
						}
					}
				}
				
				Info->Wnd->Handle()->UnlockLooper();
			}
			
			LgiSleep(50);
		}
		
		// button has been released
		if (Info->Wnd->Handle()->LockLooper())
		{
			if (Info->MouseClick)
			{
				// Do down click
				Info->Wnd->Handle()->GetMouse(&Info->Point, &Info->Buttons, false);
		
				BMessage *Msg = new BMessage(LGI_MOUSE_CLICK);
				if (Msg)
				{
					Msg->AddPoint("pos", Info->Point);
					Msg->AddInt32("flags", LastButton | _lgi_mouse_keys());
						
					Info->Wnd->GetWindow()->PostMessage(Msg, Info->Wnd->Handle());
// printf("_lgi_mouse_thread: up click to %p\n", Info->Wnd->Handle());
				}
			}
			// else is just to Capture the mouse
	
			OsView v = Info->Wnd->Handle();
			Info->Wnd->_CaptureThread = 0;
			Info->Wnd->_MouseInfo = 0;
			v->UnlockLooper();
		}
	}

// printf("_lgi_mouse_thread: end\n");

	return 0;
}

bool _setup_mouse_thread(GView::OsMouseInfo *&i, OsThread &t)
{
	if (NOT i)
	{
		i = NEW(GView::OsMouseInfo);
		if (i)
		{
			i->Wnd = 0;
			i->Clicks = 0;
			i->MouseClick = false;
			i->Capture = false;
		}
	}

	if (NOT t)
	{
		t = spawn_thread(_lgi_mouse_thread, "_lgi_mouse_thread", B_NORMAL_PRIORITY, i);
	}
	
	bool Status = i AND t > 0;

//printf("_setup_mouse_thread = %i\n", Status);

	return Status;
}

void GView::Sys_MouseDown(BPoint Point)
{
//printf("GView::Sys_MouseDown start, thread=%p\n", _CaptureThread);

	if (NOT _CaptureThread)
	{
		if (_setup_mouse_thread(_MouseInfo, _CaptureThread))
		{
			_MouseInfo->Wnd = this;
			_MouseInfo->Clicks = 1;
			_MouseInfo->Sx = Pos.X();
			_MouseInfo->Sy = Pos.Y();
			_MouseInfo->MouseClick = true;
		
			BMessage *Msg = (GetWindow())?GetWindow()->CurrentMessage():0;
			if (Msg)
			{
				Msg->FindInt32("clicks", &_MouseInfo->Clicks);
			}
	
			// Do down click
			GMouse m;
			Handle()->GetMouse(&_MouseInfo->Point, &_MouseInfo->Buttons, FALSE);
	
			m.x = _MouseInfo->Point.x;
			m.y = _MouseInfo->Point.y;
			m.Flags = _MouseInfo->Buttons | _lgi_mouse_keys();
			m.Down(true);
			m.Double(	(_MouseInfo->Clicks > 1) AND
						(abs(m.x-LastX) < DOUBLE_CLICK_THRESHOLD) AND
						(abs(m.y-LastY) < DOUBLE_CLICK_THRESHOLD));
			LastX = _MouseInfo->Point.x;
			LastY = _MouseInfo->Point.y;

// printf("GView::Sys_MouseDown before down click\n");
			OnMouseClick(m);
// printf("GView::Sys_MouseDown after down click\n");
	
			// Spawn thread to watch the mouse until the user
			// releases the button
			resume_thread(_CaptureThread);
		}
	}
}

void GView::Sys_MouseMoved(BPoint Point, uint32 transit, const BMessage *message)
{
	GMouse m;
	ulong Buttons;

	_View->GetMouse(&Point, &Buttons, FALSE);
	
	m.x = Point.x;
	m.y = Point.y;
	m.Flags = Buttons | _lgi_mouse_keys();
	m.Down(Buttons != 0);
	
	OnMouseMove(m);
//printf("GView::Sys_MouseMoved mouse move\n");
}

bool GView::IsCapturing()
{
	bool Status = false;

	if (Handle()->LockLooper())
	{
		Status =	_CaptureThread > 0 AND
					_MouseInfo AND
					_MouseInfo->Capture;

		Handle()->UnlockLooper();
	}

	return Status;
}

bool GView::Capture(bool c)
{
	bool Status = false;

// printf("GView::Capture(%i) thread=%p\n", c, _CaptureThread);

	if (Handle()->LockLooper())
	{
		bool HadThread = _CaptureThread > 0;
		
		if (c)
		{
			if (_setup_mouse_thread(_MouseInfo, _CaptureThread))
			{
				// setup new Info for capture
				_MouseInfo->Wnd = this;
				_MouseInfo->Clicks = 0;
				_MouseInfo->Sx = Pos.X();
				_MouseInfo->Sy = Pos.Y();

				Status = true;
			}
		}
		else
		{
			Status = true;
		}

		if (_MouseInfo)
		{
			// Has thread already
			_MouseInfo->Capture = c;

			if (NOT c AND
				NOT _MouseInfo->MouseClick)
			{
// printf("Capture(false) killing thread\n");
				kill_thread(_CaptureThread);
				DeleteObj(_MouseInfo);
				_CaptureThread = 0;
			}
			
		}

		if (NOT HadThread AND _CaptureThread)
		{
			resume_thread(_CaptureThread);
		}

		Handle()->UnlockLooper();
	}

	return Status;
}

void GView::Sys_FrameMoved(BPoint p)
{
	OnPosChange();
}

bool GView::IsAttached()
{
	return _View AND _View->Parent() != 0;
}

bool GView::Attach(GView *Wnd)
{
	bool Status = FALSE;
	Parent = Wnd;
	if (Parent)
	{
		BView *ParentSysView = _View->Parent();
		if (ParentSysView != Parent->_View)
		{
			if (ParentSysView)
			{
				// detach from the previous BView
				ParentSysView->RemoveChild(_View);
			}

			// bool Lock = _View->LockLooper();
			_View->MoveTo(Pos.x1, Pos.y1);
			_View->ResizeTo(Pos.X()-1, Pos.Y()-1);
			if (NOT GFlags::Visible() AND
				NOT _View->IsHidden())
			{
				_View->Hide();
			}
			if (NOT Enabled())
			{
				Enabled(false);
			}
			// if (Lock) _View->UnlockLooper();

			if (Parent->_View)
			{
				// attach to the new BView
				Parent->_View->AddChild(_View);
			}
		}

		if (NOT Parent->Children.HasItem(this))
		{
			// we aren't already a member of the parents children
			// list so add ourselves in.
			Parent->Children.Insert(this);
			Parent->OnChildrenChanged(this, true);
		}

		Status = true;
	}
	return Status;
}

bool GView::Detach()
{
	bool Status = FALSE;
	if (Parent)
	{
		Visible(FALSE);
		
		bool Lock = Parent->_View->LockLooper();
		Parent->_View->RemoveChild(_View);
		if (Lock) Parent->_View->UnlockLooper();
		
		Parent->Children.Delete(this);
		Parent->OnChildrenChanged(this, FALSE);

		Parent = 0;
		Status = TRUE;
	}
	return Status;
}

GView *GView::FindControl(OsView hCtrl)
{
	if (Handle() == hCtrl)
	{
		return this;
	}

	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(hCtrl);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

GView *GView::FindControl(int Id)
{
	if (GetId() == Id)
	{
		return this;
	}

	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(Id);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

#define MIN_WINDOW_X		4
#define MIN_WINDOW_Y		24

void GView::MoveOnScreen()
{
	GRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	GRegion p = GetPos();

	if (p.x2 >= Screen.X())
		p.Offset(Screen.X() - p.x2, 0);
	if (p.y2 >= Screen.Y())
		p.Offset(0, Screen.Y() - p.y2);
	if (p.x1 < MIN_WINDOW_X)
		p.Offset(MIN_WINDOW_X-p.x1, 0);
	if (p.y1 < MIN_WINDOW_Y)
		p.Offset(0, MIN_WINDOW_Y-p.y1);
		
	SetPos(p);
}

void GView::MoveToCenter()
{
	GRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	GRegion p = GetPos();

	p.Offset(-p.x1, -p.y1);
	p.Offset((Screen.X() - p.X()) / 2, (Screen.Y() - p.Y()) / 2);
	
	SetPos(p);
}

void GView::MoveToMouse()
{
	GMouse m;
	if (GetMouse(m, true))
	{
		GRegion p = GetPos();
		
		p.Offset(-p.x1, -p.y1);
		p.Offset(m.x-(p.X()/2), m.y-(p.Y()/2));
		
		SetPos(p);
	}
	else
	{
		// Next best thing
		MoveToCenter();
	}
}

///////////////////////////////////////////////////////////////////
bool LgiIsMounted(char *Name)
{
	return false;
}

bool LgiMountVolume(char *Name)
{
	return false;
}

