#include "Lgi.h"
#include "GWnd.h"

///////////////////////////////////////////////////////////////////////////////////////////////
GWnd::GWnd(GView *notify) :
	BWindow(BRect(10, 10, 100, 100), "(No Name)", B_TITLED_WINDOW, B_ASYNCHRONOUS_CONTROLS, B_CURRENT_WORKSPACE)
{
	Notify = notify;
}

GWnd::GWnd(	GView *notify,
			BRect frame,
			char *title,
			window_type type,
			uint32 flags,
			uint32 workspaces)
	: BWindow(frame, title, type, flags, workspaces)		
{
	Notify = notify;
}

bool GWnd::QuitRequested()
{
	if (Notify)
	{
		return Notify->QuitRequested();
	}
	
	return true;
}

void GWnd::MessageReceived(BMessage *Msg)
{
	BWindow::MessageReceived(Msg);
}

void GWnd::FrameMoved(BPoint origin)
{
	if (Notify)
	{
		Notify->Pos.Offset(	-Notify->Pos.x1 + origin.x,
							-Notify->Pos.y1 + origin.y);
	}
}

void GWnd::FrameResized(float width, float height)
{
}

////////////////////////////////////////////////////////////////////////////////////////////////
GWindow::GWindow() :
	GView(NEW(BViewRedir(this)))
{
	Menu = 0;
	Handle()->SetResizingMode(B_FOLLOW_ALL_SIDES);
	Handle()->SetViewColor(B_TRANSPARENT_COLOR);	// Rgb);

	Wnd = NEW(GWnd(this));
}

GWindow::~GWindow()
{
	DeleteObj(Menu);

	if (Handle()->LockLooper())
	{
		if (Handle()->Window())
		{
			Handle()->RemoveSelf();
		}

		Wnd->Hide();
		Wnd->Quit();
	}
}

bool GWindow::QuitRequested()
{
	bool Status = OnRequestClose();
	if (Status)
	{
		if (Handle()->LockLooper())
		{
			Handle()->RemoveSelf();
		}
		delete this;
	}
	return Status;
}

void GWindow::OnPaint(GSurface *pDC)
{
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle();
}

void GWindow::OnChildrenChanged(GView *Wnd, bool Attaching)
{
}

void GWindow::OnPosChange()
{
	GetPos();
	Pour();
}

void GWindow::OnCreate()
{
}

bool GWindow::Visible()
{
	bool Status = FALSE;
	if (Wnd)
	{
		if (Wnd->Lock())
		{
			Status = NOT Wnd->IsHidden();
			Wnd->Unlock();
		}
	}
	return Status;
}

void GWindow::Visible(bool v)
{
	if (Wnd)
	{
		if (Wnd->Lock())
		{
			if (v)
			{
				if (Wnd->IsHidden())
				{
					// Add BView here, to make sure "OnCreate" is called at the end
					// of a GWindow constructor not at the beginning, where it's not
					// useful.
					if (NOT Handle()->Parent())
					{
						// Set window pos
						Wnd->MoveTo(Pos.x1, Pos.y1);
						Wnd->ResizeTo(Pos.X(), Pos.Y());

						// Insert view
						BRect r = Wnd->Bounds();
						Handle()->ResizeTo(r.Width(), r.Height());
						Wnd->AddChild(Handle());
					}
					
					Wnd->Show();
					Invalidate(0, true);
				}
			}
			else
			{
				if (NOT Wnd->IsHidden())
				{
					Wnd->Hide();
				}
			}
			Wnd->Unlock();
		}
	}

	if (v)
	{
		Pour();
	}
}

bool GWindow::Name(char *n)
{
	bool Status = GObject::Name(n);
	if (Wnd)
	{
		bool Lock = Wnd->Lock();
		Wnd->SetTitle(n);
		if (Lock) Wnd->Unlock();
	}
	return Status;
}

char *GWindow::Name()
{
	return GObject::Name();
}

GRegion &GWindow::GetPos()
{
	GRegion r(0, 0, 1, 1);
	if (Wnd)
	{
		if (Wnd->Lock())
		{
			BRect frame = Wnd->Frame();
			r.x1 = frame.left;
			r.y1 = frame.top;
			r.x2 = frame.right;
			r.y2 = frame.bottom;
			Pos = frame;

			Wnd->Unlock();
		}
		else
		{
			r = Pos;
		}
	}
	return r;
}

bool GWindow::SetPos(GRegion &p, bool Repaint)
{
	if (Wnd)
	{
		if (Wnd->Lock())
		{
			Wnd->MoveTo(p.x1, p.y1);
			Wnd->ResizeTo(p.X(), p.Y());
			Wnd->Unlock();
		}
		Pos = p;
		
		return true;
	}

	return false;
}

void GWindow::Pour()
{
	bool SafeToLock = false;

	if (GetWindow()->IsLocked())
	{
		// I'm already locked... this could be bad if it's not me
		thread_id Thread = GetWindow()->LockingThread();
		if (Thread != -1)
		{
			if (Thread == find_thread(NULL))
			{
				// it's all ok, I'm locking me
				SafeToLock = true;
			}
			else
			{
				// someone else is locking us
				// ok who is locking me?
				thread_info Info;
				if (get_thread_info(Thread, &Info) == B_OK)
				{
					printf("Evil locking thread: %i (%s)\n", Thread, Info.name);
				}
				else
				{
					printf("Couldn't get evil thread info\n");
				}
			}
		}
	}
	else
	{
		SafeToLock = true;
	}
	
	if (NOT SafeToLock)
	{
		return;
	}
	
	bool Lock = Handle()->LockLooper();

	GetWindow()->BeginViewTransaction();
	GRegion r(Handle()->Frame());
	GComplexRegion Client(r);
	GComplexRegion Update;

	if (Menu AND Menu->Bar)
	{
		BRect Rc = Menu->Bar->Frame();
		GRegion Mp = Rc;
		Mp.x2 = 10000;
		Client.Subtract(&Mp);
	}

	for (GView *w = Children.First(); w; w = Children.Next())
	{
		GRegion OldPos = w->Pos;
		Update.Union(&OldPos);

		if (w->Pour(Client))
		{
			if (NOT w->Visible())
			{
				w->Visible(true);
			}

			w->Invalidate();

			Client.Subtract(&w->Pos);
			Update.Subtract(&w->Pos);
		}
		else
		{
			// non-pourable
		}
	}

	GetWindow()->EndViewTransaction();
	GetWindow()->Flush();

	Handle()->Invalidate();
	if (Lock) Handle()->UnlockLooper();
}

int GWindow::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case B_SIMPLE_DATA:
		{
			List<char> Files;
			int32 Count = 0;
			type_code Type = 0;
			
			if (Msg->GetInfo("refs", &Type, &Count) == B_OK)
			{
				for (int i=0; i<Count; i++)
				{
					entry_ref Ref;
					if (Msg->FindRef("refs", i, &Ref) == B_OK)
					{
						BPath Path;
						BEntry Entry(&Ref, true);
						Entry.GetPath(&Path);
						char *p = (char*) Path.Path();
						if (p)
						{
							Files.Insert(NewStr(p));
						}
					}
				}
			}
			
			if (Files.GetItems() > 0)
			{
				OnReceiveFiles(Files);
			}
			
			for (char *f=Files.First(); f; f=Files.Next())
			{
				DeleteArray(f);
			}
			break;
		}
		case WM_COMMAND:
		{
			int32 Cmd = 0;
			int32 Event = 0;
			Msg->FindInt32("Cmd", &Cmd);
			Msg->FindInt32("Event", &Event);
			OnCommand(Cmd, Event, 0);
			break;
		}
	}

	return GView::OnEvent(Msg);
}

