/*hdr
**	FILE:			Alpha.cpp
**	AUTHOR:		Matthew Allen
**	DATE:			4/3/98
**	DESCRIPTION:	Draw mode effects
**
**		Copyright (C) 1998, Matthew Allen
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "Gdc2.h"
#include "Alpha.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
#define Rc16(c)		( (((c) & 0xF800) >> 8) | (((c) & 0xF800) >> 13) )
#define Gc16(c)		( (((c) & 0x07E0) >> 3) | (((c) & 0x07E0) >> 9) )
#define Bc16(c)		( (((c) & 0x001F) << 3) | (((c) & 0x001F) >> 2) )

#define Rc15(c)		( (((c) & 0x7C00) >> 7) | (((c) & 0x7C00) >> 12) )
#define Gc15(c)		( (((c) & 0x03E0) >> 2) | (((c) & 0x03E0) >> 7) )
#define Bc15(c)		( (((c) & 0x001F) << 3) | (((c) & 0x001F) >> 2) )

///////////////////////////////////////////////////////////////////////////////////////////////////
GdcApp8Alpha::GdcApp8Alpha()
{
	Bits = 8;
	Bytes = 1;
	Op = GDC_ALPHA;
	ZeroObj(Remap);
}

void GdcApp8Alpha::SetAlpha(uchar A, GdcPalette *Pal)
{
	a = A;
	oma = 0xFF - A;

	if (Pal AND a < 255)
	{
		GdcRGB *p = (*Pal)[0];
		GdcRGB *Col = (*Pal)[c&0xFF];

		for (int i=0; i<Pal->GetSize(); i++)
		{
			COLOUR Rgb = Rgb24(	((oma * p[i].R) + (a * Col->R)) / 255,
						((oma * p[i].G) + (a * Col->G)) / 255,
						((oma * p[i].B) + (a * Col->B)) / 255);

			Remap[i] = Pal->MatchRgb(Rgb);
		}
	}
	else
	{
		for (int i=0; i<256; i++)
		{
			Remap[i] = c;
		}
	}
}

void GdcApp8Alpha::Set()
{
	*Ptr = Remap[*Ptr];
}

void GdcApp8Alpha::VLine(int y)
{
	while (y--)
	{
		*Ptr = Remap[*Ptr];
		Ptr += Dest->Line;
	}
}

void GdcApp8Alpha::Rectangle(int x, int y)
{
	while (y--)
	{
		uchar *p = Ptr;
		for (int X=0; X<x; X++)
		{
			*p++ = Remap[*p];
		}
		Ptr += Dest->Line;
	}
}

bool GdcApp8Alpha::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (NOT Src OR NOT Pal) return FALSE;
	if (NOT SPal) SPal = Pal;

	if (SrcAlpha)
	{
		GdcRGB *SRgb = (*SPal)[0];
		GdcRGB *DRgb = (*Pal)[0];
		if (NOT SRgb OR NOT DRgb) return FALSE;

		switch (Src->Bits)
		{
			case 8:
			{
				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					uchar *a = SrcAlpha->Base + (y * SrcAlpha->Line);
					uchar *d = Ptr;
					uchar oma;

					for (int x=0; x<Src->x; x++)
					{
						oma = 255 - *a;
						if (oma == 0)
						{
							*d = *s;
						}
						else if (oma < 0xFF)
						{
							uchar r = ((oma * SRgb[*s].R) + ((int)*a * DRgb[*d].R)) / 255;
							uchar g = ((oma * SRgb[*s].G) + ((int)*a * DRgb[*d].G)) / 255;
							uchar b = ((oma * SRgb[*s].B) + ((int)*a * DRgb[*d].B)) / 255;
							*d = Pal->MatchRgb(Rgb24(r, g, b));
						}

						d++;
						a++;
						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 16:
			{
				for (int y=0; y<Src->y; y++)
				{
					ushort *s = (ushort*) (Src->Base + (y * Src->Line));
					uchar *a = SrcAlpha->Base + (y * SrcAlpha->Line);
					uchar *d = Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar oma = 255 - *a;
						uchar r = ((oma * DRgb[*d].R) + ((int)*a * Rc16(*s))) / 255;
						uchar g = ((oma * DRgb[*d].G) + ((int)*a * Gc16(*s))) / 255;
						uchar b = ((oma * DRgb[*d].B) + ((int)*a * Bc16(*s))) / 255;
						*d++ = Pal->MatchRgb(Rgb24(r, g, b));
						a++;
						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 24:
			{
				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					uchar *a = SrcAlpha->Base + (y * SrcAlpha->Line);
					uchar *d = Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar oma = 255 - *a;
						uchar r = ((oma * DRgb[*d].R) + ((int)*a * s[0])) / 255;
						uchar g = ((oma * DRgb[*d].G) + ((int)*a * s[1])) / 255;
						uchar b = ((oma * DRgb[*d].B) + ((int)*a * s[2])) / 255;
						*d++ = Pal->MatchRgb(Rgb24(r, g, b));
						a++;
						s+=3;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 32:
			{
				for (int y=0; y<Src->y; y++)
				{
					ulong *s = (ulong*) (Src->Base + (y * Src->Line));
					uchar *a = SrcAlpha->Base + (y * SrcAlpha->Line);
					uchar *d = Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar oma = 255 - *a;
						uchar r = ((oma * DRgb[*d].R) + ((int)*a * R32(*s))) / 255;
						uchar g = ((oma * DRgb[*d].G) + ((int)*a * G32(*s))) / 255;
						uchar b = ((oma * DRgb[*d].B) + ((int)*a * B32(*s))) / 255;
						*d++ = Pal->MatchRgb(Rgb24(r, g, b));
						a++;
						s+=4;
					}

					Ptr += Dest->Line;
				}
				break;
			}
		}
	}
	else
	{
		// TODO
		// what do I do here????
		// I need an ORDER[1..8] MatchRgb func... other wise it'll suck speed wise

		GdcRGB *SRgb = (*SPal)[0];
		GdcRGB *DRgb = (*Pal)[0];
		if (NOT SRgb OR NOT DRgb) return FALSE;

		switch (Src->Bits)
		{
			case 8:
			{
				if (a == 255)
				{
					// do a straight blt
					for (int y=0; y<Src->y; y++)
					{
						uchar *s = Src->Base + (y * Src->Line);
						uchar *d = Ptr;

						memcpy(d, s, Src->x);
						Ptr += Dest->Line;
					}
				}
				else if (a == 0)
				{
					// do nothing baby
				}
				else
				{
				}
				break;
			}
			case 16:
			{
				break;
			}
			case 24:
			{
				break;
			}
			case 32:
			{
				break;
			}
		}
	}

	return FALSE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void GdcApp15Alpha::Set()
{
	ushort *p = (ushort*) Ptr;
	*p = Rgb15(	((oma * Rc15(*p)) + (a * Rc15(c))) / 255,
				((oma * Gc15(*p)) + (a * Gc15(c))) / 255,
				((oma * Bc15(*p)) + (a * Bc15(c))) / 255);
}

void GdcApp15Alpha::VLine(int height)
{
	ushort *p;
	int Cr = Rc15(c) * a;
	int Cg = Gc15(c) * a;
	int Cb = Bc15(c) * a;

	while (height--)
	{
		p = (ushort*) Ptr;
		*p = Rgb15(	((oma * Rc15(*p)) + Cr) / 255,
					((oma * Gc15(*p)) + Cg) / 255,
					((oma * Bc15(*p)) + Cb) / 255);

		Ptr += Dest->Line;
	}
}

void GdcApp15Alpha::Rectangle(int x, int y)
{
	ushort *p;
	int Cr = Rc15(c) * a;
	int Cg = Gc15(c) * a;
	int Cb = Bc15(c) * a;

	while (y--)
	{
		p = (ushort*) Ptr;
		for (int n=0; n<x; n++, p++)
		{
			*p = Rgb15(	((oma * Rc15(*p)) + Cr) / 255,
						((oma * Gc15(*p)) + Cg) / 255,
						((oma * Bc15(*p)) + Cb) / 255);
		}

		Ptr += Dest->Line;
	}
}

bool GdcApp15Alpha::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (NOT Src) return 0;

	if (SrcAlpha)
	{
		uchar lookup[256];
		for (int i=0; i<256; i++)
		{
			lookup[i] = (i * (int)a) / 255;
		}

		switch (Src->Bits)
		{
			case 8:
			{
				uchar r[256];
				uchar g[256];
				uchar b[256];

				for (int i=0; i<SPal->GetSize(); i++)
				{
					GdcRGB *p = (SPal) ? (*SPal)[i] : 0;
					if (p)
					{
						r[i] = p->R;
						g[i] = p->G;
						b[i] = p->B;
					}
					else
					{
						r[i] = i;
						g[i] = i;
						b[i] = i;
					}
				}

				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							*d = Rgb15(r[*s], g[*s], b[*s]);
						}
						else if (my_oma < 255)
						{
							*d = Rgb15(	((my_oma * Rc15(*d)) + (my_a * b[*s])) / 255,
										((my_oma * Gc15(*d)) + (my_a * g[*s])) / 255,
										((my_oma * Bc15(*d)) + (my_a * r[*s])) / 255);
						}

						d++;
						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 15:
			{
				for (int y=0; y<Src->y; y++)
				{
					ushort *s = (ushort*) (Src->Base + (y * Src->Line));
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							*d = *s;
						}
						else if (my_oma < 255)
						{
							*d = Rgb15(	((my_oma * Rc15(*d)) + (my_a * Rc15(*s))) / 255,
										((my_oma * Gc15(*d)) + (my_a * Gc15(*s))) / 255,
										((my_oma * Bc15(*d)) + (my_a * Bc15(*s))) / 255);
						}

						d++;
						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 16:
			{
				for (int y=0; y<Src->y; y++)
				{
					ushort *s = (ushort*) (Src->Base + (y * Src->Line));
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							*d = *s;
						}
						else if (my_oma < 255)
						{
							*d = Rgb15(	((my_oma * Rc15(*d)) + (my_a * Rc16(*s))) / 255,
										((my_oma * Gc15(*d)) + (my_a * Gc16(*s))) / 255,
										((my_oma * Bc15(*d)) + (my_a * Bc16(*s))) / 255);
						}

						d++;
						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 24:
			{
				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							*d++ = Rgb16(s[0], s[1], s[2]);
						}
						else if (my_oma < 255)
						{
							*d++ = Rgb15(	( ((int)my_a * Rc15(*d)) + (my_oma * s[0]) ) / 255,
											( ((int)my_a * Gc15(*d)) + (my_oma * s[1]) ) / 255,
											( ((int)my_a * Bc15(*d)) + (my_oma * s[2]) ) / 255);
						}

						s += 3;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 32:
			{
				for (int y=0; y<Src->y; y++)
				{
					ulong *s = (ulong*) (Src->Base + (y * Src->Line));
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							*d = *s;
						}
						else if (my_oma < 255)
						{
							*d++ = Rgb15(	( ((int)my_a * Rc15(*d)) + (my_oma * R32(*s)) ) / 255,
											( ((int)my_a * Gc15(*d)) + (my_oma * G32(*s)) ) / 255,
											( ((int)my_a * Bc15(*d)) + (my_oma * B32(*s)) ) / 255);
						}

						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
		}
	}
	else
	{
		switch (Src->Bits)
		{
			case 8:
			{
				// this code uses the input bitmap as a mask to say where
				// to draw the current colour at the given alpha value

				int Cr = Rc15(c) * a;
				int Cg = Gc15(c) * a;
				int Cb = Bc15(c) * a;

				for (int y=0; y<Src->y; y++)
				{
					ushort *d = (ushort*) Ptr;
					uchar *s = Src->Base + (Src->Line * y);
					ushort Dst;

					for (int x=0; x<Src->x; x++, d++, s++)
					{
						if (*s)
						{
							Dst = *d;
							*d = Rgb15(	((oma * Rc15(Dst)) + Cr) / 255,
										((oma * Gc15(Dst)) + Cg) / 255,
										((oma * Bc15(Dst)) + Cb) / 255);
						}
					}

					Ptr += Dest->Line;
				}

				break;
			}

			case 16:
			{
				// this code combines the colour bitmap with the destination
				// bitmap using the given alpha value
				for (int y=0; y<Src->y; y++)
				{
					ushort *s = (ushort*) (Src->Base + (y * Src->Line));
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++, d++, s++)
					{
						*d = Rgb15(	((oma * Rc15(*d)) + (a * Rc16(*s))) / 255,
									((oma * Gc15(*d)) + (a * Gc16(*s))) / 255,
									((oma * Bc15(*d)) + (a * Bc16(*s))) / 255);
					}

					Ptr += Dest->Line;
				}
				break;
			}
		}
	}

	return FALSE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
void GdcApp16Alpha::Set()
{
	ushort *p = (ushort*) Ptr;
	*p = Rgb16(	((oma * Rc16(*p)) + (a * Rc16(c))) / 255,
			((oma * Gc16(*p)) + (a * Gc16(c))) / 255,
			((oma * Bc16(*p)) + (a * Bc16(c))) / 255);
}

void GdcApp16Alpha::VLine(int height)
{
	ushort *p;
	int Cr = Rc16(c) * a;
	int Cg = Gc16(c) * a;
	int Cb = Bc16(c) * a;

	while (height--)
	{
		p = (ushort*) Ptr;
		*p = Rgb16(	((oma * Rc16(*p)) + Cr) / 255,
				((oma * Gc16(*p)) + Cg) / 255,
				((oma * Bc16(*p)) + Cb) / 255);

		Ptr += Dest->Line;
	}
}

void GdcApp16Alpha::Rectangle(int x, int y)
{
	ushort *p;
	int Cr = Rc16(c) * a;
	int Cg = Gc16(c) * a;
	int Cb = Bc16(c) * a;

	while (y--)
	{
		p = (ushort*) Ptr;
		for (int n=0; n<x; n++, p++)
		{
			*p = Rgb16(	((oma * Rc16(*p)) + Cr) / 255,
					((oma * Gc16(*p)) + Cg) / 255,
					((oma * Bc16(*p)) + Cb) / 255);
		}

		Ptr += Dest->Line;
	}
}

bool GdcApp16Alpha::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (NOT Src) return 0;

	if (SrcAlpha)
	{
		uchar lookup[256];
		for (int i=0; i<256; i++)
		{
			lookup[i] = (i * (int)a) / 255;
		}

		switch (Src->Bits)
		{
			case 8:
			{
				uchar r[256];
				uchar g[256];
				uchar b[256];

				for (int i=0; i<SPal->GetSize(); i++)
				{
					GdcRGB *p = (SPal) ? (*SPal)[i] : 0;
					if (p)
					{
						r[i] = p->R;
						g[i] = p->G;
						b[i] = p->B;
					}
					else
					{
						r[i] = i;
						g[i] = i;
						b[i] = i;
					}
				}

				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							*d = Rgb16(r[*s], g[*s], b[*s]);
						}
						else if (my_oma < 255)
						{
							*d = Rgb16(	((my_oma * Rc16(*d)) + (my_a * b[*s])) / 255,
											((my_oma * Gc16(*d)) + (my_a * g[*s])) / 255,
											((my_oma * Bc16(*d)) + (my_a * r[*s])) / 255);
						}

						d++;
						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 16:
			{
				for (int y=0; y<Src->y; y++)
				{
					ushort *s = (ushort*) (Src->Base + (y * Src->Line));
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							*d = *s;
						}
						else if (my_oma < 255)
						{
							*d = Rgb16(	((my_oma * Rc16(*d)) + (my_a * Rc16(*s))) / 255,
										((my_oma * Gc16(*d)) + (my_a * Gc16(*s))) / 255,
										((my_oma * Bc16(*d)) + (my_a * Bc16(*s))) / 255);
						}

						d++;
						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 24:
			{
				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							*d++ = Rgb16(s[0], s[1], s[2]);
						}
						else if (my_oma < 255)
						{
							*d++ = Rgb16(	( ((int)my_a * Rc16(*d)) + (my_oma * s[0]) ) / 255,
											( ((int)my_a * Gc16(*d)) + (my_oma * s[1]) ) / 255,
											( ((int)my_a * Bc16(*d)) + (my_oma * s[2]) ) / 255);
						}

						s += 3;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 32:
			{
				for (int y=0; y<Src->y; y++)
				{
					ulong *s = (ulong*) (Src->Base + (y * Src->Line));
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							*d = *s;
						}
						else if (my_oma < 255)
						{
							*d++ = Rgb16(	((my_oma * (*d)) + (my_a * B32(*s))) / 255,
											((my_oma * (*d)) + (my_a * G32(*s))) / 255,
											((my_oma * (*d)) + (my_a * R32(*s))) / 255);
						}

						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
		}
	}
	else
	{
		switch (Src->Bits)
		{
			case 8:
			{
				// this code uses the input bitmap as a mask to say where
				// to draw the current colour at the given alpha value

				int Cr = Rc16(c) * a;
				int Cg = Gc16(c) * a;
				int Cb = Bc16(c) * a;

				for (int y=0; y<Src->y; y++)
				{
					ushort *d = (ushort*) Ptr;
					uchar *s = Src->Base + (Src->Line * y);
					ushort Dst;

					for (int x=0; x<Src->x; x++, d++, s++)
					{
						if (*s)
						{
							Dst = *d;
							*d = Rgb16(	((oma * Rc16(Dst)) + Cr) / 255,
										((oma * Gc16(Dst)) + Cg) / 255,
										((oma * Bc16(Dst)) + Cb) / 255);
						}
					}

					Ptr += Dest->Line;
				}

				break;
			}

			case 16:
			{
				// this code combines the colour bitmap with the destination
				// bitmap using the given alpha value
				for (int y=0; y<Src->y; y++)
				{
					ushort *s = (ushort*) (Src->Base + (y * Src->Line));
					ushort *d = (ushort*) Ptr;

					for (int x=0; x<Src->x; x++, d++, s++)
					{
						*d = Rgb16(	((oma * Rc16(*d)) + (a * Rc16(*s))) / 255,
									((oma * Gc16(*d)) + (a * Gc16(*s))) / 255,
									((oma * Bc16(*d)) + (a * Bc16(*s))) / 255);
					}

					Ptr += Dest->Line;
				}
				break;
			}
		}
	}

	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
void GdcApp24Alpha::Set()
{
	COLOUR p = Get();
	Ptr[0] = ((oma * Ptr[0]) + (a * B24(c))) / 255;
	Ptr[1] = ((oma * Ptr[1]) + (a * G24(c))) / 255;
	Ptr[2] = ((oma * Ptr[2]) + (a * R24(c))) / 255;
}

void GdcApp24Alpha::VLine(int height)
{
	uchar *p = Ptr;
	int Cr = R24(c) * a;
	int Cg = G24(c) * a;
	int Cb = B24(c) * a;

	while (height--)
	{
		*p++ = ((oma * (*p)) + Cb) / 255;
		*p++ = ((oma * (*p)) + Cg) / 255;
		*p++ = ((oma * (*p)) + Cr) / 255;
		p = (Ptr += Dest->Line);
	}
}

void GdcApp24Alpha::Rectangle(int x, int y)
{
	uchar *p = Ptr;
	int Cr = R24(c) * a;
	int Cg = G24(c) * a;
	int Cb = B24(c) * a;

	while (y--)
	{
		for (int n=0; n<x; n++)
		{
			*p++ = ((oma * (*p)) + Cb) / 255;
			*p++ = ((oma * (*p)) + Cg) / 255;
			*p++ = ((oma * (*p)) + Cr) / 255;
		}

		p = (Ptr += Dest->Line);
	}
}

bool GdcApp24Alpha::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (NOT Src) return FALSE;

	if (SrcAlpha)
	{
		uchar lookup[256];
		for (int i=0; i<256; i++)
		{
			lookup[i] = (i * (int)a) / 255;
		}

		switch (Src->Bits)
		{
			case 8:
			{
				int Size = (SPal) ? SPal->GetSize() : 256;
				uchar r[256];
				uchar g[256];
				uchar b[256];

				for (int i=0; i<Size; i++)
				{
					GdcRGB *p = (SPal) ? (*SPal)[i] : 0;
					if (p)
					{
						r[i] = p->R;
						g[i] = p->G;
						b[i] = p->B;
					}
					else
					{
						r[i] = i;
						g[i] = i;
						b[i] = i;
					}
				}

				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					uchar *d = Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							*d++ = b[*s];
							*d++ = g[*s];
							*d++ = r[*s];
						}
						else if (my_oma < 255)
						{
							*d++ = ((my_oma * (*d)) + (my_a * b[*s])) / 255;
							*d++ = ((my_oma * (*d)) + (my_a * g[*s])) / 255;
							*d++ = ((my_oma * (*d)) + (my_a * r[*s])) / 255;
						}

						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 16:
			{
				for (int y=0; y<Src->y; y++)
				{
					ushort *s = (ushort*) (Src->Base + (y * Src->Line));
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					uchar *d = Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							*d++ = Bc16(*s);
							*d++ = Gc16(*s);
							*d++ = Rc16(*s);
						}
						else if (my_oma < 255)
						{
							*d++ = ((my_oma * (*d)) + (my_a * Bc16(*s))) / 255;
							*d++ = ((my_oma * (*d)) + (my_a * Gc16(*s))) / 255;
							*d++ = ((my_oma * (*d)) + (my_a * Rc16(*s))) / 255;
						}

						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 24:
			{
				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					uchar *d = Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							d[0] = s[0];
							d[1] = s[1];
							d[2] = s[2];
						}
						else if (my_oma < 255)
						{
							d[0] = ( (my_oma * d[0]) + (my_a * s[0]) ) / 255;
							d[1] = ( (my_oma * d[1]) + (my_a * s[1]) ) / 255;
							d[2] = ( (my_oma * d[2]) + (my_a * s[2]) ) / 255;
						}

						d += 3;
						s += 3;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 32:
			{
				for (int y=0; y<Src->y; y++)
				{
					ulong *s = (ulong*) (Src->Base + (y * Src->Line));
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					uchar *d = Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							*d++ = B32(*s);
							*d++ = G32(*s);
							*d++ = R32(*s);
						}
						else if (my_oma < 255)
						{
							*d++ = ((my_oma * (*d)) + (my_a * B32(*s))) / 255;
							*d++ = ((my_oma * (*d)) + (my_a * G32(*s))) / 255;
							*d++ = ((my_oma * (*d)) + (my_a * R32(*s))) / 255;
						}

						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
		}
	}
	else
	{
		switch (Src->Bits)
		{
			case 24:
			{
				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					uchar *d = Ptr;

					if (a == 0xFF)
					{
						memcpy(d, s, Src->x*3);
					}
					else if (a > 0)
					{
						for (int x=0; x<Src->x; x++)
						{
							*d++ = ((oma * (*d)) + (a * (*s++))) / 255;
							*d++ = ((oma * (*d)) + (a * (*s++))) / 255;
							*d++ = ((oma * (*d)) + (a * (*s++))) / 255;
						}
					}

					Ptr += Dest->Line;
				}
				break;
			}
		}
	}

	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
void GdcApp32Alpha::Set()
{
	ulong *p = (ulong*) Ptr;
	int R = ((oma * R32(*p)) + (a * R32(c))) / 255;
	int G = ((oma * G32(*p)) + (a * G32(c))) / 255;
	int B = ((oma * B32(*p)) + (a * B32(c))) / 255;

	*p = Rgb32(R, G, B);
}

void GdcApp32Alpha::VLine(int height)
{
	ulong *p;
	int Cr = R32(c) * a;
	int Cg = G32(c) * a;
	int Cb = B32(c) * a;

	while (height--)
	{
		p = (ulong*) Ptr;
		*p = Rgb32(	((oma * R32(*p)) + Cr) / 255,
				((oma * G32(*p)) + Cg) / 255,
				((oma * B32(*p)) + Cb) / 255);

		Ptr = (((uchar*) Ptr) + Dest->Line);
	}
}

void GdcApp32Alpha::Rectangle(int x, int y)
{
	ulong *p;
	int Cr = R32(c) * a;
	int Cg = G32(c) * a;
	int Cb = B32(c) * a;

	while (y--)
	{
		p = (ulong*) Ptr;
		for (int n=0; n<x; n++, p++)
		{
			*p = Rgb32(	((oma * R32(*p)) + Cr) / 255,
					((oma * G32(*p)) + Cg) / 255,
					((oma * B32(*p)) + Cb) / 255);
		}

		Ptr = (((uchar*) Ptr) + Dest->Line);
	}
}

bool GdcApp32Alpha::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (NOT Src) return 0;

	if (SrcAlpha)
	{
		uchar lookup[256];
		for (int i=0; i<256; i++)
		{
			lookup[i] = (i * (int)a) / 255;
		}

		switch (Src->Bits)
		{
			case 8:
			{
				uchar r[256];
				uchar g[256];
				uchar b[256];

				for (int i=0; i<SPal->GetSize(); i++)
				{
					GdcRGB *p = (SPal) ? (*SPal)[i] : 0;
					if (p)
					{
						r[i] = p->R;
						g[i] = p->G;
						b[i] = p->B;
					}
					else
					{
						r[i] = i;
						g[i] = i;
						b[i] = i;
					}
				}

				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					uchar *d = Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							d[0] = r[*s];
							d[1] = g[*s];
							d[2] = b[*s];
						}
						else if (my_oma < 255)
						{
							d[0] = ((my_oma * (*d)) + (my_a * r[*s])) / 255;
							d[1] = ((my_oma * (*d)) + (my_a * g[*s])) / 255;
							d[2] = ((my_oma * (*d)) + (my_a * b[*s])) / 255;
						}

						d += 4;
						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 16:
			{
				for (int y=0; y<Src->y; y++)
				{
					ushort *s = (ushort*) (Src->Base + (y * Src->Line));
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					uchar *d = Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							d[0] = Rc16(*s);
							d[1] = Gc16(*s);
							d[2] = Bc16(*s);
						}
						else if (my_oma < 255)
						{
							d[0] = ((my_oma * d[0]) + (my_a * Rc16(*s))) / 255;
							d[1] = ((my_oma * d[1]) + (my_a * Gc16(*s))) / 255;
							d[2] = ((my_oma * d[2]) + (my_a * Bc16(*s))) / 255;
						}

						s++;
						d+=4;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 24:
			{
				for (int y=0; y<Src->y; y++)
				{
					uchar *s = Src->Base + (y * Src->Line);
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					uchar *d = Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							d[0] = s[0];
							d[1] = s[1];
							d[2] = s[2];
						}
						else if (my_oma < 255)
						{
							d[0] = ( ((int)my_a * d[0]) + (my_oma * s[0]) ) / 255;
							d[1] = ( ((int)my_a * d[1]) + (my_oma * s[1]) ) / 255;
							d[2] = ( ((int)my_a * d[2]) + (my_oma * s[2]) ) / 255;
						}

						d += 4;
						s += 3;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 32:
			{
				for (int y=0; y<Src->y; y++)
				{
					ulong *s = (ulong*) (Src->Base + (y * Src->Line));
					uchar *src_a = SrcAlpha->Base + (y * SrcAlpha->Line);
					uchar *d = Ptr;

					for (int x=0; x<Src->x; x++)
					{
						uchar my_a = lookup[*src_a++];
						uchar my_oma = 255 - my_a;

						if (my_oma == 0)
						{
							d[0] = R32(*s);
							d[1] = G32(*s);
							d[2] = B32(*s);
						}
						else if (my_oma < 255)
						{
							d[0] = ((my_a * (*d)) + (my_oma * R32(*s))) / 255;
							d[1] = ((my_a * (*d)) + (my_oma * G32(*s))) / 255;
							d[2] = ((my_a * (*d)) + (my_oma * B32(*s))) / 255;
						}

						d += 4;
						s++;
					}

					Ptr += Dest->Line;
				}
				break;
			}
		}
	}
	else
	{
		switch (Src->Bits)
		{
			case 32:
			{
				for (int y=0; y<Src->y; y++)
				{
					ulong *s = (ulong*) (Src->Base + (y * Src->Line));
					ulong *d = (ulong*) Ptr;

					for (int x=0; x<Src->x; x++, d++, s++)
					{
						*d = Rgb32(	((oma * R32(*d)) + (a * R32(*s))) / 255,
								((oma * G32(*d)) + (a * G32(*s))) / 255,
								((oma * B32(*d)) + (a * B32(*s))) / 255);
					}

					Ptr = (((uchar*) Ptr) + Dest->Line);
				}
				break;
			}
		}
	}

	return FALSE;
}

GApplicator *GdcAlphaServer::Create(int Bits)
{
	switch (Bits)
	{
		case 8:
		{
			return NEW(GdcApp8Alpha);
			break;
		}
		case 15:
		{
			return NEW(GdcApp15Alpha);
			break;
		}
		case 16:
		{
			return NEW(GdcApp16Alpha);
			break;
		}
		case 24:
		{
			return NEW(GdcApp24Alpha);
			break;
		}
		case 32:
		{
			return NEW(GdcApp32Alpha);
			break;
		}
	}

	return NULL;
}
