/*hdr
**      FILE:           Filter2.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           25/3/97
**      DESCRIPTION:    Graphics file filters
**
**      Copyright (C) 1997, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "Gdc2.h"

#ifndef WIN32
#define BI_RLE8       1L
#define BI_RLE4       2L
#else
#include "Lgi.h"
#endif

/////////////////////////////////////////////////////////////////////////
int FindHeader(int Offset, char *Str, GFile &f)
{
	int i = 0;
	
	if (Offset >= 0)
	{
		f.Seek(Offset, SEEK_SET);
	}
	
	while (NOT f.Eof())
	{
		char c;
		
		f >> c;
		if (Str[i] == c || Str[i] == '?')
		{
			i++;
			if (!Str[i])
			{
				return TRUE;
			}
		}
		else
		{
			i = 0;
		}
	}

	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Windows BMP file format

class GdcBmp : public GFilter {

	int ActualBits;
	int ScanSize;
	bool SetSize(long *Info, GDeviceContext *pDC);

public:
	GdcBmp();

	int GetCapabilites() { return FILTER_CAP_READ | FILTER_CAP_WRITE; }
	bool ReadImage(GSurface *pDC);
	bool WriteImage(GSurface *pDC);
};

class GdcBmpFactory : public GFilterFactory
{
	bool CheckFile(char *File)
	{
		return (File) ? stristr(File, ".bmp") != 0 : false;
	}

	GFilter *NewObject()
	{
		return NEW(GdcBmp);
	}

} BmpFactory;

#define BMP_ID			0x4D42

#ifdef WIN32
#pragma pack(push, before_pack)
#pragma pack(1)
#endif

class BMP_FILE {
public:
	ushort		Type;
	ulong		Size;
	ushort		Reserved1;
	ushort		Reserved2;
	ulong		OffsetToData;
};

class BMP_WININFO {
public:
	long		Size;
	long		Sx;
	long		Sy;
	ushort		Planes;
	ushort		Bits;
	ulong		Compression;
	ulong		DataSize;
	long		XPels;
	long		YPels;
	ulong		ColoursUsed;
	ulong		ColourImportant;
};

class BMP_OS2INFO {
public:
	long		Size;
	ushort		Sx;
	ushort		Sy;
	ushort		Planes;
	ushort		Bits;
};

#ifdef WIN32
#pragma pack(pop, before_pack)
#endif

bool GdcBmp::SetSize(long *Info, GDeviceContext *pDC)
{
	bool Status = FALSE;

	switch (*Info)
	{
		case sizeof(BMP_WININFO):
		{
			BMP_WININFO *i = (BMP_WININFO*) Info;
			ActualBits = i->Bits;
			ScanSize = BMPWIDTH(i->Sx * i->Bits);
			Status = pDC->Create(i->Sx, i->Sy, max(i->Bits, 8), ScanSize);
			break;
		}

		case sizeof(BMP_OS2INFO):
		{
			BMP_OS2INFO *i = (BMP_OS2INFO*) Info;
			ActualBits = i->Bits;
			ScanSize = BMPWIDTH(i->Sx * i->Bits);
			Status = pDC->Create(i->Sx, i->Sy, max(i->Bits, 8), ScanSize);
			Seek(sizeof(BMP_OS2INFO)-sizeof(BMP_WININFO), SEEK_CUR);
			break;
		}
	}

	return Status;
}

bool GdcBmp::ReadImage(GSurface *pDC)
{
	bool Status = FALSE;
	ActualBits = 0;
	ScanSize = 0;

	if (pDC)
	{
		BMP_FILE File;
		BMP_WININFO Info;
		GRegion Cr;

		*this >> File.Type;
		*this >> File.Size;
		*this >> File.Reserved1;
		*this >> File.Reserved2;
		*this >> File.OffsetToData;

		*this >> Info.Size;
		*this >> Info.Sx;
		*this >> Info.Sy;
		*this >> Info.Planes;
		*this >> Info.Bits;
		*this >> Info.Compression;
		*this >> Info.DataSize;
		*this >> Info.XPels;
		*this >> Info.YPels;
		*this >> Info.ColoursUsed;
		*this >> Info.ColourImportant;

		if (File.Type == BMP_ID AND
			SetSize((long*) &Info, pDC))
		{
			if (pDC->GetBits() <= 8)
			{
				int Colours = 1 << ActualBits;
				GdcPalette *Palette = NEW(GdcPalette);
				if (Palette)
				{
					Palette->SetSize(Colours);
					GdcRGB *Start = (*Palette)[0];
					if (Start)
					{
						Read(Start, sizeof(GdcRGB)*Colours);
						Palette->SwapRAndB();
					}

					Palette->Update();
					pDC->Palette(Palette);
				}
			}

			GdcSurface *pMem = GetSurface(pDC);
			Seek(File.OffsetToData, SEEK_SET);

			Status = TRUE;

			if (Info.Compression == BI_RLE8)
			{
				// 8 bit RLE compressed image
				int Remaining = GetSize() - GetPosition();
				uchar *Data = NEW(uchar[Remaining]);
				if (Data)
				{
					if (Read(Data, Remaining) == Remaining)
					{
						int x=0, y=pDC->Y()-1;
						uchar *p = Data;
						bool Done = FALSE;

						while (NOT Done AND p < Data + Remaining)
						{
							uchar Length = *p++;
							uchar Colour = *p++;

							if (Length == 0)
							{
								switch (Colour)
								{
									case 0:
									{
										x = 0;
										y--;
										break;
									}
									case 1:
									{
										Done = TRUE;
										break;
									}
									case 2:
									{
										x += *p++;
										y -= *p++;
										break;
									}
									default:
									{
										// absolute mode
										uchar *Pixel = (*pDC)[y];
										if (Pixel AND y >= 0 AND y < pDC->Y())
										{
											int Len = min(Colour, pDC->X() - x);
											if (Len > 0)
											{
												memcpy(Pixel + x, p, Len);
												x += Colour;
												p += Colour;
											}
										}
										else
										{
											p += Colour;
										}
										if ((int) p & 1) p++;
										break;
									}
								}
							}
							else
							{
								// encoded mode
								uchar *Pixel = (*pDC)[y];
								if (Pixel AND y >= 0 AND y < pDC->Y())
								{
									int Len = min(Length, pDC->X() - x);
									if (Len > 0)
									{
										memset(Pixel + x, Colour, Len);
										x += Length;
									}
								}
							}
						}
					}
				}
			}
			else if (Info.Compression == BI_RLE4)
			{
				// 4 bit RLE compressed image

				// not implemented
			}
			else
			{
				// Progress
				if (Meter)
				{
					Meter->SetDescription("scanlines");
					Meter->SetLimits(0, pMem->y-1);
				}

				// straight
				switch (ActualBits)
				{
					case 1:
					{
						uchar *Buffer = NEW(uchar[ScanSize]);
						if (Buffer)
						{
							for (int y=pMem->y-1; y>=0; y--)
							{
								Status &= Read(Buffer, ScanSize) == ScanSize;
								if (Status)
								{
									uchar Mask = 0x80;
									uchar *d = Buffer;
									for (int x=0; x<pDC->X(); x++)
									{
										pDC->Colour((*d & Mask) ? 1 : 0);
										pDC->Set(x, y);

										Mask >>= 1;
										if (NOT Mask)
										{
											Mask = 0x80;
											d++;
										}
									}
								}
								else
								{
									break;
								}

								if (Meter) Meter->Value(pMem->y-1-y);
							}

							DeleteArray(Buffer);
						}
						break;
					}
					case 4:
					{
						uchar *Buffer = NEW(uchar[ScanSize]);
						if (Buffer)
						{
							for (int y=pMem->y-1; y>=0; y--)
							{
								Status &= Read(Buffer, ScanSize) == ScanSize;

								uchar *d = Buffer;
								for (int x=0; x<pDC->X(); x++)
								{
									if (x & 1)
									{
										pDC->Colour(*d & 0xf);
										d++;
									}
									else
									{
										pDC->Colour(*d >> 4);
									}
									pDC->Set(x, y);
								}

								if (Meter) Meter->Value(pMem->y-1-y);
							}

							DeleteArray(Buffer);
						}
						break;
					}
					case 24:
					#ifdef BEOS
					{
						for (int i=pMem->y-1; i>=0; i--)
						{
							uchar *Ptr = pMem->Base + (pMem->Line * i);
							Status &= Read(Ptr, ScanSize) == ScanSize;

							ulong *Out = ((ulong*)Ptr) + (pDC->X()-1);
							uchar *In = Ptr + ((pDC->X()-1) * 3);
							for (int x=0; x<pDC->X(); x++)
							{
								*Out-- = Rgb32(In[2], In[1], In[0]);
								In -= 3;
							}
							
							if (Meter) Meter->Value(pMem->y-1-i);
						}
						break;
					}
					#endif
					case 8:
					case 16:
					case 32:
					{
						for (int i=pMem->y-1; i>=0; i--)
						{
							Status &= Read(pMem->Base + (pMem->Line * i), ScanSize) == ScanSize;
							if (Meter) Meter->Value(pMem->y-1-i);
						}
						
						break;
					}
					default:
					{
						Status = FALSE;
						break;
					}
				}
			}

			Cr.ZOff(pDC->X()-1, pDC->Y()-1);
			pDC->Update(GDC_BITS_CHANGE|GDC_PAL_CHANGE);
		}
		else
		{
			Cr.ZOff(-1, -1);
		}

		pDC->ClipRgn(&Cr);
	}
	
	return Status;
}

bool GdcBmp::WriteImage(GSurface *pDC)
{
	bool Status = FALSE;
	
	if (pDC)
	{
		BMP_FILE File;
		BMP_WININFO Info;
		GdcSurface *pMem = GetSurface(pDC);
		int Colours = (pMem->Bits <= 8) ? 1 << pMem->Bits : 0;
		int UsedBits = pMem->Bits;

		GdcPalette *Palette = pDC->Palette();
		if (pMem->Bits <= 8 AND
			Palette)
		{
			int Size = Palette->GetSize();
			int Bits = 8;
			for (int c=256; c; c>>=1, Bits--)
			{
				if (c & Size)
				{
					Colours = c;
					UsedBits = Bits;
					break;
				}
			}
		}

		if (NOT pMem OR
			NOT pMem->x OR
			NOT pMem->y)
		{
			return FALSE;
		}

		GFile::SetSize(0);

		File.Type = BMP_ID;
		File.OffsetToData = sizeof(BMP_FILE) + sizeof(BMP_WININFO) + (Colours * 4);
		File.Size = File.OffsetToData + (pMem->Line * pMem->y);
		File.Reserved1 = 0;
		File.Reserved2 = 0;

		Info.Size = sizeof(BMP_WININFO);
		Info.Sx = pMem->x;
		Info.Sy = pMem->y;
		Info.Planes = 1;
		Info.Bits = UsedBits;
		Info.Compression = 0;
		Info.DataSize = 0;
		Info.XPels = 3000;
		Info.YPels = 3000;
		Info.ColoursUsed = Colours;
		Info.ColourImportant = Colours;

		if (Write(&File, sizeof(File)) AND
			Write(&Info, sizeof(Info)))
		{
			if (pMem->Bits <= 8)
			{
				int Done = 0;

				if (Palette)
				{
					GdcRGB *Start = (*Palette)[0];
					if (Start)
					{
						Palette->SwapRAndB();
						Write(Start, sizeof(GdcRGB)*Palette->GetSize());
						Palette->SwapRAndB();
						Done += Palette->GetSize();
					}
				}

				char Grey[4];
				Grey[3] = 0;

				while (Done < Colours)
				{
					int C = Done * 256 / Colours;

					Grey[0] = C;
					Grey[1] = C;
					Grey[2] = C;

					Write(Grey, 4);
					Done++;
				}
			}

			// Progress
			if (Meter)
			{
				Meter->SetDescription("scanlines");
				Meter->SetLimits(0, pMem->y-1);
			}

			int Bytes = BMPWIDTH(pMem->x * UsedBits);
			Status = TRUE;
			switch (UsedBits)
			{
				case 1:
				{
					uchar *Buffer = NEW(uchar[Bytes]);
					if (Buffer)
					{
						for (int i=pMem->y-1; i>=0; i--)
						{
							uchar *Src = pMem->Base + (i * pMem->Line);
							
							// assemble the bits
							memset(Buffer, 0, Bytes);
							for (int x=0; x<pMem->x; x++)
							{
								Buffer[x>>3] |= (Src[x] & 1) << (x & 3);
							}

							// write the bits
							Status &= Write(Buffer, Bytes) == Bytes;
							
							// update status
							if (Meter) Meter->Value(pMem->y-1-i);
						}

						DeleteArray(Buffer);
					}
					break;
				}
				case 4:
				{
					uchar *Buffer = NEW(uchar[Bytes]);
					if (Buffer)
					{
						for (int i=pMem->y-1; i>=0; i--)
						{
							uchar *Src = pMem->Base + (i * pMem->Line);

							// assemble the nibbles
							for (int x=0; x<pMem->x; x+=2)
							{
								Buffer[x>>1] = (Src[x]<<4) | (Src[x+1]&0x0f);
							}

							// write the line
							Status &= Write(Buffer, Bytes) == Bytes;
							
							// update status
							if (Meter) Meter->Value(pMem->y-1-i);
						}

						DeleteArray(Buffer);
					}
					break;
				}
				default:
				{
					for (int i=pMem->y-1; i>=0; i--)
					{
						Status &= Write(pMem->Base + (i * pMem->Line), Bytes) == Bytes;
						if (Meter) Meter->Value(pMem->y-1-i);
					}
					break;
				}
			}
		}
	}

	Close();
	
	return Status;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// ICO file format
class GdcIco : public GFilter
{
	int TruncSize(int i)
	{
		if (i >= 64) return 64;
		if (i >= 32) return 32;
		return 16;
	}

public:
	GdcIco();

	int GetCapabilites() { return FILTER_CAP_READ | FILTER_CAP_WRITE; }
	int GetImages() { return 1; }
	bool ReadImage(GSurface *pDC);
	bool WriteImage(GSurface *pDC);
};

class GdcIcoFactory : public GFilterFactory
{
	bool CheckFile(char *File)
	{
		return (File) ? stristr(File, ".ico") != 0 : false;
	}

	GFilter *NewObject()
	{
		return NEW(GdcIco);
	}

} IcoFactory;

GdcIco::GdcIco()
{
	Props.Set("Type", "Icon file");
	Props.Set("Extension", "ICO");
}

bool GdcIco::ReadImage(GSurface *pDC)
{
	bool Status = false;
	int MyBits = 0;

	int16 Reserved_1;
	int16 Type;
	int16 Count;

	*this >> Reserved_1;
	*this >> Type;
	*this >> Count;

	for (int Image = 0; Image < Count; Image++)
	{
		int8	Width;
		int8	Height;
		int8	ColorCount;
		int8	Reserved_2;
		int16	Planes;
		int16	BitCount;
		int32	BytesInRes;
		int32	ImageOffset;
		
		*this >> Width;
		*this >> Height;
		*this >> ColorCount;
		*this >> Reserved_2;
		*this >> Planes;
		*this >> BitCount;
		*this >> BytesInRes;
		*this >> ImageOffset;

		int CurrentPos = GetPosition();
		Seek(ImageOffset, SEEK_SET);

		BMP_WININFO	Header;
		GdcRGB *Colours;
		uchar *XorBytes;
		uchar *AndBytes;

		assert(sizeof(Header) == 40);
		Read(&Header, sizeof(Header));

		int Cols = 1 << (Header.Bits);
		Colours = NEW(GdcRGB[Cols]);
		if (Colours)
		{
			Read(Colours, sizeof(GdcRGB) * Cols);
		}

		int XorSize = BMPWIDTH(Header.Sx * Header.Bits) * (Header.Sy / 2);
		int AndSize = BMPWIDTH(Header.Sx) * (Header.Sy / 2);
		LgiAssert(Header.DataSize == XorSize + AndSize);

		XorBytes = NEW(uchar[XorSize]);
		if (XorBytes)
		{
			Read(XorBytes, XorSize);
		}

		AndBytes = NEW(uchar[AndSize]);
		if (AndBytes)
		{
			Read(AndBytes, AndSize);
		}

		if (Colours AND
			XorBytes AND
			AndBytes AND
			(Header.Bits > MyBits OR Width > pDC->X() OR Height > pDC->Y()) AND
			pDC->Create(Width, Height, max(8, Header.Bits) ))
		{
			MyBits = Header.Bits;
			pDC->Colour(0, 24);
			pDC->Rectangle();

			COLOUR Trans = 0xff;
			GdcPalette *Pal = NEW(GdcPalette);
			if (Pal)
			{
				Pal->SetSize(Cols);
				memcpy((*Pal)[0], Colours, sizeof(GdcRGB) * Cols);
				Pal->SwapRAndB();
				pDC->Palette(Pal, true);
				Trans = Pal->MatchRgb(Rgb24(255, 0, 255));
			}

			int XorLine = XorSize / Height;
			int AndLine = AndSize / Height;
			for (int y=0; y<Height; y++)
			{
				uchar *Dest = (*pDC)[y];
				uchar *Src = XorBytes + ((Height-y-1) * XorLine);
				uchar *Mask = AndBytes + ((Height-y-1) * AndLine);

				switch (Header.Bits)
				{
					case 1:
					{
						uchar Mask = 0x80;
						for (int x=0; x<Width; x++)
						{
							if (Src[x/8] & Mask)
							{
								Dest[x] = 1;
							}
							else
							{
								Dest[x] = 0;
							}
							Mask >>= 1;
							if (NOT Mask) Mask = 0x80;
						}
						Status = true;
						break;
					}
					case 4:
					{
						for (int x=0; x<Width; x++)
						{
							if (x & 1)
							{
								Dest[x] = Src[x>>1] & 0xf;
							}
							else
							{
								Dest[x] = Src[x>>1] >> 4;
							}
						}
						Status = true;
						break;
					}
					case 8:
					{
						// memcpy(Dest, Src, Width * (Header.Bits / 8));

						for (int x=0; x<Width; x++)
						{
							if (Mask[x>>3] & (0x80 >> (x&7)))
							{
								Dest[x] = Trans;
							}
							else
							{
								Dest[x] = Src[x];
							}
						}

						Status = true;
						break;
					}
				}
			}
		}

		DeleteArray(Colours);
		DeleteArray(XorBytes);
		DeleteArray(AndBytes);
		Seek(CurrentPos, SEEK_SET);
	}

	return Status;
}

bool GdcIco::WriteImage(GSurface *pDC)
{
	bool Status = false;

	if (pDC AND
		pDC->GetBits() <= 8)
	{
		SetSize(0);

		int ActualBits = pDC->GetBits();
		GdcPalette *Pal = pDC->Palette();
		if (Pal)
		{
			if (Pal->GetSize() <= 2)
			{
				ActualBits = 1;
			}
			else if (Pal->GetSize() <= 16)
			{
				ActualBits = 4;
			}
		}

		// write file header
		int Colours = 1 << ActualBits;
		int16 Reserved_1 = 0;
		int16 Type = 1;
		int16 Count = 1;

		*this << Reserved_1;
		*this << Type;
		*this << Count;

		// write directory list
		int8	Width = TruncSize(pDC->X());
		int8	Height = TruncSize(pDC->Y());
		int8	ColorCount = Colours;
		int8	Reserved_2 = 0;
		int16	Planes = 0;
		int16	BitCount = 0;

		int Line = (ActualBits * Width) / 8;
		int MaskLine = BMPWIDTH(Width/8);
		int32	BytesInRes = sizeof(BMP_WININFO) +
							(sizeof(GdcRGB) * Colours) +
							(Line * Height) +
							(MaskLine * Height);
		
		*this << Width;
		*this << Height;
		*this << ColorCount;
		*this << Reserved_2;
		*this << Planes;
		*this << BitCount;
		*this << BytesInRes;

		int32	ImageOffset = GetPosition() + sizeof(ImageOffset);
		*this << ImageOffset;

		// Write icon itself
		BMP_WININFO	Header;
		assert(sizeof(Header) == 40);
		Header.Size = sizeof(Header);
		Header.Sx = Width;
		Header.Sy = Height * 2;
		Header.Planes = 1;
		Header.Bits = ActualBits;
		Header.Compression = 0; // BI_RGB;
		Header.DataSize = Line * Height;
		Header.XPels = 0;
		Header.YPels = 0;
		Header.ColoursUsed = 0;
		Header.ColourImportant = 0;
		Write(&Header, sizeof(Header));

		// Write palette
		if (Pal)
		{
			Pal->SwapRAndB();
			Write((*Pal)[0], sizeof(GdcRGB) * (1 << ActualBits));
			Pal->SwapRAndB();
		}

		// Get background colour
		COLOUR Back = -1;
		Props.Get("back", (int&)Back);

		// Write "Colour" bits
		int y;
		for (y=Height-1; y>=0; y--)
		{
			uchar *Src = (*pDC)[y];

			// xor bits (colour)
			switch (ActualBits)
			{
				case 4:
				{
					uchar Dest;
					for (int x=0; x<Width; x++)
					{
						if (x&1)
						{
							Dest |=	(Back == Src[x]) ? 0 : (Src[x] & 0xf) << 4;
							*this << Dest;
							Dest = 0;
						}
						else
						{
							Dest |=	(Back == Src[x]) ? 0 : Src[x] & 0xf;
						}
					}

					Status = true;
					break;
				}
				case 8:
				{
					for (int x=0; x<Width; x++)
					{
						uchar Dest = (Back == Src[x]) ? 0 : Src[x];
						*this << Dest;
					}
					Status = true;
					break;
				}
			}
		}

		// Write "And" bits (mask)
		for (y=Height-1; y>=0; y--)
		{
			uchar *Src = (*pDC)[y];
			uchar Dest = 0;
			uchar Mask = 0x80;
			for (int x=0; x<Width; x++)
			{
				if (Src[x] == Back) Dest |= Mask;
				Mask >>= 1;
				if (NOT Mask)
				{
					*this << Dest;
					Dest = 0;
					Mask = 0x80;
				}
			}

			Status = true;
		}
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// Run length encoded DC
GdcRleDC::GdcRleDC()
{
	Flags = GDC_RLE_COLOUR | GDC_RLE_READONLY;
	Length = 0;
	Alloc = 0;
	Data = 0;
	ScanLine = 0;
	Key = 0;

	pMem = NEW(GdcSurface);
	if (pMem)
	{
		pMem->Base = 0;
		pMem->x = 0;
		pMem->y = 0;
		pMem->Bits = 0;
		pMem->Line = 0;
		pMem->Flags = 0;
	}
}

GdcRleDC::~GdcRleDC()
{
	Empty();
}

void GdcRleDC::Empty()
{
	Length = 0;
	Alloc = 0;
	DeleteArray(Data);
	DeleteArray(ScanLine);
}

bool GdcRleDC::Create(int x, int y, int bits, int LineLen)
{
	bool Status = GMemDC::Create(x, y, bits, LineLen);
	if (Status)
	{
		Flags &= ~GDC_RLE_READONLY;
	}
	return Status;
}

bool GdcRleDC::CreateInfo(int x, int y, int bits)
{
	bool Status = FALSE;

	DeleteObj(pMem);
	pMem = NEW(GdcSurface);
	if (pMem)
	{
		pMem->x = x;
		pMem->y = y;
		pMem->Bits = bits;
		pMem->Base = 0;
		pMem->Line = 0;
		pMem->Flags = 0;

		Flags |= GDC_RLE_READONLY;

		if (bits == 8)
		{
			Flags |= GDC_RLE_MONO;
		}
	}

	return Status;
}

void GdcRleDC::ReadOnly(bool Read)
{
	if (Read)
	{
		Flags |= GDC_RLE_READONLY;
	}
	else
	{
		if (pMem)
		{
			pMem->Base = NEW(uchar[pMem->Line * pMem->y]);
		}
		else
		{
			// just in case... shouldn't ever happen
			// am I paranoid?
			// 'course :)
			Create(1, 1, 24);
		}

		Flags &= ~GDC_RLE_READONLY;
	}
}

bool GdcRleDC::ReadOnly()
{
	return (Flags & GDC_RLE_READONLY) != 0;
}

void GdcRleDC::Mono(bool Mono)
{
	if (Mono)
	{
		Flags |= GDC_RLE_MONO;
		Flags &= ~GDC_RLE_COLOUR;
	}
	else
	{
		Flags &= ~GDC_RLE_MONO;
		Flags |= GDC_RLE_COLOUR;
	}
}

bool GdcRleDC::Mono()
{
	return (Flags & GDC_RLE_MONO) != 0;
}

bool GdcRleDC::SetLength(int Len)
{
	bool Status = TRUE;

	if (Len + 1 > Alloc)
	{
		int NewAlloc = (Len + 0x4000) & (~0x3FFF);
		uchar *Temp = NEW(uchar[NewAlloc]);
		if (Temp)
		{
			if (NewAlloc > Length)
			{
				memset(Temp + Length, 0, NewAlloc - Length);
			}

			memcpy(Temp, Data, Length);
			DeleteArray(Data);
			DeleteArray(ScanLine);
			Data = Temp;
			Length = Len;
			Alloc = NewAlloc;
		}
		else
		{
			Status = FALSE;
		}
	}
	else
	{
		// TODO: resize Data to be smaller if theres a
		// considerable size change
		Length = Len;
	}

	return Status;
}

void GdcRleDC::Update(int UpdateFlags)
{
	bool Error = FALSE;

	if (	(UpdateFlags & GDC_BITS_CHANGE) AND
		NOT (Flags & GDC_RLE_READONLY))
	{
		Empty();

		COLOUR Key = Get(0, 0);
		ulong Pos = 0;
		ulong PixelSize = GetBits() / 8;
		
		for (ulong y=0; NOT Error AND y<Y(); y++)
		{
			for (ulong x=0; x<X(); )
			{
				ulong Skip = 0;
				ulong Pixels = 0;
				uchar *Bits = 0;

				while (Get(x, y) == Key AND x < X())
				{
					Skip++;
					x++;
				}

				Bits = (*this)[y] + (x * PixelSize);
				while (Get(x, y) != Key AND x < X())
				{
					Pixels++;
					x++;
				}

				if (Flags & GDC_RLE_COLOUR)
				{
					if (SetLength(	Length +
							(2 * sizeof(ulong)) +
							(Pixels * PixelSize)))
					{
						*((ulong*)(Data+Pos)) = Skip;
						Pos += sizeof(Skip);
						*((ulong*)(Data+Pos)) = Pixels;
						Pos += sizeof(Pixels);
						
						int l = Pixels * PixelSize;
						memcpy(Data+Pos, Bits, l);
						Pos += l;
					}
					else
					{
						Error = TRUE;
						break;
					}
				}
				else if (Flags & GDC_RLE_MONO)
				{
					if (SetLength(	Length + (2 * sizeof(ulong))))
					{
						*((ulong*)(Data+Pos)) = Skip;
						Pos += sizeof(Skip);
						*((ulong*)(Data+Pos)) = Pixels;
						Pos += sizeof(Pixels);
					}
					else
					{
						Error = TRUE;
						break;
					}
				}
			}
		}
	}

	if (Error)
	{
		Empty();
	}
	else
	{
		FindScanLines();
	}
}

bool GdcRleDC::FindScanLines()
{
	bool Status = FALSE;
	bool Error = FALSE;
	DeleteArray(ScanLine);

	if (Data)
	{
		ScanLine = NEW(uchar*[Y()]);
		if (ScanLine)
		{
			ulong Pos = 0;
			ulong PixelSize = GetBits() / 8;
			
			for (ulong y=0; NOT Error AND y<Y(); y++)
			{
				ScanLine[y] = Data + Pos;
				ulong x;
				for (x=0; x<X();)
				{
					ulong Skip = 0;
					ulong Pixels = 0;

					Skip = *((ulong*)(Data + Pos));
					Pos += sizeof(Skip);
					Pixels = *((ulong*)(Data + Pos));
					Pos += sizeof(Pixels);

					if (Flags & GDC_RLE_COLOUR)
					{
						Pos += Pixels * PixelSize;
					}

					x += Pixels + Skip;
				}

				if (x != X() OR Pos > Length)
				{
					Error = TRUE;
				}
			}
		}

		if (Error)
		{
			DeleteArray(ScanLine);
		}
		else
		{
			Status = TRUE;
		}
	}

	return Status;
}

void GdcRleDC::Draw(GSurface *Dest, int Ox, int Oy)
{
	#ifdef ADJUST_ORIGIN
	int OriX, OriY;
	Dest->GetOrigin(OriX, OriY);
	Ox += OriX;
	Oy += OriY;
	#endif

	if (Dest AND Data)
	{
		bool ReBuildScanInfo = FALSE;
		
		if (NOT ScanLine) ReBuildScanInfo = TRUE;
		if (ScanLine AND Data AND ScanLine[0] != Data) ReBuildScanInfo = TRUE;

		if (ReBuildScanInfo)
		{
			// need to rebuild scan line info
			if (NOT FindScanLines())
			{
				// couldn't create scan line info so quit
				return;
			}
		}

		GRegion S;
		S.ZOff(X()-1, Y()-1);
		GRegion D;
		D = S;
		D.Offset(Ox, Oy);
		GRegion DClip = D;
		GRegion Temp = Dest->ClipRgn();
		D.Bound(&Temp);

		if (DClip == D AND (*Dest)[0])
		{
			// no clipping needed
			int PixLen = Dest->GetBits() / 8;

			if (Flags & GDC_RLE_COLOUR)
			{
				if (Dest->GetBits() == GetBits())
				{
					for (int y=0; y<Y(); y++)
					{
						uchar *s = ScanLine[y];
						uchar *d = (*Dest)[Oy+y];
						for (int x=0; x<X(); )
						{
							x += *((ulong*)s); s+=4;
							int Pixels = *((ulong*)s); s+=4;
							int Len = Pixels*PixLen;
							memcpy(d + ((Ox+x)*PixLen), s, Len);
							x += Pixels;
							s += Len;
						}
					}
				}
			}
			else if (Flags & GDC_RLE_MONO)
			{
				GApplicator *pDApp = Dest->Applicator();

				for (int y=0; y<Y(); y++)
				{
					uchar *s = ScanLine[y];
					for (int x=0; x<X(); )
					{
						x += *((ulong*)s); s+=4;
						int Pixels = *((ulong*)s); s+=4;
						pDApp->SetPtr(Ox+x, Oy+y);
						pDApp->Rectangle(Pixels, 1);
						x += Pixels;
					}
				}
			}
		}
		else if (DClip.Valid())
		{
			// clip
			int PixLen = Dest->GetBits() / 8;

			if (Flags & GDC_RLE_COLOUR)
			{
				if (Dest->GetBits() == GetBits())
				{
					for (int y=0; y<Y(); y++)
					{
						int Fy = Oy+y;
						if (Fy >= Temp.y1 AND Fy < Temp.y2) // clip y
						{
							uchar *s = ScanLine[y];
							uchar *d = (*Dest)[Fy];
							for (int x=0; x<X(); )
							{
								x += *((ulong*)s); s+=4;
								int Pixels = *((ulong*)s); s+=4;

								int Fx = Ox+x;
								int PreClipPixels = max(0, Temp.x1 - Fx);
								int PostClipPixels = max(0, (Pixels + Fx) - Temp.x2 - 1);
								int PixelsLeft = Pixels - (PreClipPixels + PostClipPixels);
								if (PixelsLeft > 0) // clip x
								{
									memcpy(	d + ((Fx - PreClipPixels) * PixLen),
										s + (PreClipPixels * PixLen),
										PixelsLeft * PixLen);
								}

								x += Pixels;
								s += Pixels * PixLen;
							}
						}
					}
				}
			}
			/*
			else if (Flags & GDC_RLE_MONO)
			{
				GApplicator *pDApp = Dest->Applicator();

				for (int y=0; y<Y(); y++)
				{
					uchar *s = ScanLine[y];
					for (int x=0; x<X(); )
					{
						x += *((ulong*)s); s+=4;
						int Pixels = *((ulong*)s); s+=4;
						pDApp->SetPtr(Ox+x, Oy+y);
						pDApp->Rectangle(Pixels, 1);
						x += Pixels;
					}
				}
			}
			*/
		}
	}
}

int GdcRleDC::SizeOf()
{
	return (sizeof(int) * 5) + Length;
}

bool GdcRleDC::Read(GFile &F)
{
	Empty();
	
	int Len = 0, x = 0, y = 0, bits = 0;
	char Monochrome;
	F >> Len;
	if (SetLength(Len))
	{
		F >> Key;
		F >> x;
		F >> y;
		F >> bits;
		F >> Monochrome;
		Mono(Monochrome);
		CreateInfo(x, y, bits);
		F.Read(Data, Len);
		return NOT F.GetStatus();
	}

	return FALSE;
}

bool GdcRleDC::Write(GFile &F)
{
	if (Data)
	{
		char Monochrome = Mono();
		
		F << Length;
		F << Key;
		F << X();
		F << Y();
		F << GetBits();
		F << Monochrome;

		F.Write(Data, Length);

		return NOT F.GetStatus();
	}

	return FALSE;
}

void GdcRleDC::Move(GdcRleDC *pDC)
{
	if (pDC)
	{
		Key = pDC->Key;
		Flags = pDC->Flags;
		Length = pDC->Length;
		Alloc = pDC->Alloc;
		Data = pDC->Data;
		ScanLine = pDC->ScanLine;
		pMem = pDC->pMem;
		pPalette = pDC->pPalette;

		pDC->pPalette = 0;
		pDC->pMem = 0;
		pDC->Length = 0;
		pDC->Alloc = 0;
		pDC->Data = 0;
		pDC->ScanLine = 0;
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////
// MBM
class GdcMbm : public GFilter {

	GDeviceContext	*pBmp;
	GdcRleDC	*pRle;

	#ifdef WIN32
	bool IsRead() { return Attributes & O_READ; }
	bool IsWrite() { return Attributes & O_WRITE; }
	#else
	bool IsRead() { return 0; }
	bool IsWrite() { return 0; }
	#endif

	bool ReadAny();
	bool WriteAny();

public:
	GdcMbm();
	~GdcMbm();

	bool Open(char *Name, int Attrib);
	bool Close();

	int GetCapabilites() { return FILTER_CAP_READ | FILTER_CAP_WRITE; }
	bool ReadImage(GSurface *pDC);
	bool WriteImage(GSurface *pDC);
	bool ReadRle(GdcRleDC *pDC);
	bool WriteRle(GdcRleDC *pDC);
};

#define MBM_HEADER				"MBMG"

// ver 1.0 defines
#define MBM_EOF					0
#define MBM_INFO				1
#define MBM_PALETTE				2
#define MBM_BITMAP				3

// ver 2.0 defines
#define MBM_INFO2				4
#define MBM_RLE					5
#define MBM_COMMENT				6
#define MBM_FREE_SPACE				7

class MbmObject {
public:
	MbmObject() {}
	virtual ~MbmObject() {}
	virtual int SizeOf() { return 0; }
};

class MbmString : public MbmObject {

	char *Str;

public:
	MbmString()
	{
		Str = 0;
	}

	~MbmString()
	{
		DeleteArray(Str);
	}

	operator char *() const { return Str; }
	char *operator =(char *c) { DeleteArray(Str); Str = NewStr(c); return Str; }
	int SizeOf() { return sizeof(ulong) + (Str) ? strlen(Str) : 0; }

	bool Read(GFile &F)
	{
		ulong Len = 0;
		F >> Len;
		DeleteArray(Str);
		Str = NEW(char[Len+1]);
		if (Str)
		{
			F.Read(Str, Len);
			Str[Len] = 0;
		}

		return NOT F.GetStatus();
	}

	bool Write(GFile &F)
	{
		ulong Len = 0;
		if (Str)
		{
			Len = strlen(Str);
			F << Len;
			F.Write(Str, Len);
		}
		else
		{
			F << Len;
		}

		return NOT F.GetStatus();
	}
};

// MBM_INFO
class MbmSimpleInfo : public MbmObject {
public:
	ushort Sx;
	ushort Sy;
	ushort Bits;
	ushort Gamma; // * 1000

	bool Read(GFile &F)
	{
		F >> Sx;
		F >> Sy;
		F >> Bits;
		F >> Gamma;

		return NOT F.GetStatus();
	}

	bool Write(GFile &F)
	{
		F << Sx;
		F << Sy;
		F << Bits;
		F << Gamma;

		return NOT F.GetStatus();
	}

	int SizeOf() { return sizeof(ushort) * 4; }
};

// MBM_INFO2
class MbmComplexInfo : public MbmObject {
public:
	MbmString	Name;
	ulong		Sx;
	ulong		Sy;
	ulong		Bits;
	ulong		LineLength;
	ushort		Channels[4];

	int SizeOf()
	{
		return	Name.SizeOf() + 
			sizeof(Sx) +
			sizeof(Sy) +
			sizeof(Bits) +
			sizeof(LineLength) +
			sizeof(Channels);
	}

	bool Read(GFile &F)
	{
		Name.Read(F);
		F >> Sx;
		F >> Sy;
		F >> Bits;
		F >> LineLength;
		F.Read(Channels, sizeof(Channels));

		return NOT F.GetStatus();
	}

	bool Write(GFile &F)
	{
		Name.Write(F);
		F << Sx;
		F << Sy;
		F << Bits;
		F << LineLength;
		F.Write(Channels, sizeof(Channels));

		return NOT F.GetStatus();
	}
};

GdcMbm::GdcMbm()
{
	pBmp = 0;
	pRle = 0;
}

GdcMbm::~GdcMbm()
{
}

bool GdcMbm::Open(char *Name, int Attrib)
{
	bool Status = GFile::Open(Name, Attrib);

	if (Status)
	{
		if (IsRead())
		{
			// check for MBM id code
			if (NOT FindHeader(0, MBM_HEADER, *this))
			{
				Status = FALSE;
				GFile::Close();
			}
		}
		else if (IsWrite())
		{
			// write MBM id code
			if (NOT Write((void*) MBM_HEADER, 4))
			{
				Status = FALSE;
				GFile::Close();
			}
		}
	}

	return Status;
}

bool GdcMbm::Close()
{
	#ifdef WIN32
	if (ValidHandle(hFile) AND IsWrite())
	{
		// write the EOF chunk
		int i = MBM_EOF;
		(*this) << i;
	}
	#endif

	bool Status = GFile::Close();
	return Status;
}

bool GdcMbm::ReadImage(GSurface *pDC)
{
	pBmp = pDC;
	pRle = NULL;
	return ReadAny();
}

bool GdcMbm::WriteImage(GSurface *pDC)
{
	pBmp = pDC;
	pRle = NULL;
	return WriteAny();
}

bool GdcMbm::ReadRle(GdcRleDC *pDC)
{
	pBmp = NULL;
	pRle = pDC;
	return ReadAny();
}

bool GdcMbm::WriteRle(GdcRleDC *pDC)
{
	pBmp = NULL;
	pRle = pDC;
	return WriteAny();
}

bool GdcMbm::ReadAny()
{
	bool Status = FALSE;
	bool Done = FALSE;

	static MbmSimpleInfo *SInfo = NULL;
	static MbmComplexInfo *CInfo = NULL;
	static GdcPalette *Pal = NULL;

	while (NOT Done)
	{
		ulong ChunkId = 0;
		ulong Size = 0;

		(*this) >> ChunkId;
		if (ChunkId != MBM_EOF)
		{
			(*this) >> Size;
		}

		switch (ChunkId)
		{
			case MBM_EOF:
			{
				Done = TRUE;
				break;
			}
			case MBM_INFO:
			{
				DeleteObj(CInfo);
				if (NOT SInfo) SInfo = NEW(MbmSimpleInfo);

				if (SInfo)
				{
					if (NOT SInfo->Read(*this))
					{
						Done = TRUE;
						Status = FALSE;
					}
				}
				break;
			}
			case MBM_INFO2:
			{
				DeleteObj(SInfo);
				if (NOT CInfo) CInfo = NEW(MbmComplexInfo);

				if (CInfo)
				{
					if (NOT CInfo->Read(*this))
					{
						Done = TRUE;
						Status = FALSE;
					}
					else
					{
						Props.Set("Name", (char*) CInfo->Name);
					}
				}
				break;
			}
			case MBM_PALETTE:
			{
				int Entries;
				(*this) >> Entries;

				uchar *PalData = NEW(uchar[Entries]);
				if (PalData)
				{
					Read(PalData, Entries);

					DeleteObj(Pal);
					Pal = NEW(GdcPalette(PalData));
				}
				break;
			}
			case MBM_BITMAP:
			{
				if (pBmp)
				{
					bool CreateStatus = FALSE;
					int LineLength = 0;

					if (CInfo)
					{
						// Prefer complex info
						LineLength = CInfo->LineLength;
						CreateStatus = pBmp->Create(CInfo->Sx, CInfo->Sy, CInfo->Bits, CInfo->LineLength);
					}
					else if (SInfo)
					{
						// Use simple info
						LineLength = ((SInfo->Sx * SInfo->Bits) + 7) / 8;
						CreateStatus = pBmp->Create(SInfo->Sx, SInfo->Sy, SInfo->Bits, LineLength);
					}

					if (CreateStatus AND (*pBmp)[0])
					{
						Status = Read((*pBmp)[0], LineLength * pBmp->Y());

						Props.Set("X", pBmp->X());
						Props.Set("Y", pBmp->Y());
						Props.Set("Bits", pBmp->GetBits());
					}
					else
					{
						Status = FALSE;
					}

					Done = TRUE;
				}
				break;
			}
			case MBM_RLE:
			{
				if (pRle)
				{
					Status = pRle->Read(*this);
					if (Status)
					{
						Props.Set("X", pRle->X());
						Props.Set("Y", pRle->Y());
						Props.Set("Bits", pRle->GetBits());
					}

					Done = TRUE;
				}
				break;
			}
			case MBM_COMMENT:
			{
				MbmString Str;
				Str.Read(*this);
				Props.Set("Comment", Str);
				break;
			}
			default:
			{
				// Skip unknown block
				Seek(Size, SEEK_CUR);
				break;
			}
		}
	}

	if (NOT Status)
	{
		DeleteObj(SInfo);
		DeleteObj(CInfo);
		DeleteObj(Pal);
	}

	return Status;
}

#define MakeChannel(Bits, Comp)			((Bits<<8)|(Comp))

bool GdcMbm::WriteAny()
{
	bool Status = FALSE;
	bool Error = FALSE;

	static MbmComplexInfo *CInfo = NULL;
	static GdcPalette *Pal = NULL;

	GDeviceContext *pCommon = 0;
	if (pBmp)
	{
		pCommon = pBmp;
	}
	else
	{
		pCommon = (GDeviceContext*) pRle;
	}

	if (pCommon)
	{
		// Info block
		DeleteObj(CInfo);
		CInfo = NEW(MbmComplexInfo);
		if (CInfo)
		{
			char *Name = 0;
			if (Props.Get("Name", Name))
			{
				CInfo->Name = Name;
			}
			else
			{
				CInfo->Name = "Untitled";
			}

			CInfo->Sx = pCommon->X();
			CInfo->Sy = pCommon->Y();
			CInfo->Bits = pCommon->GetBits();
			CInfo->LineLength = 0;
			
			ZeroObj(CInfo->Channels);
			switch (CInfo->Bits)
			{
				case 8:
				{
					CInfo->Channels[0] = MakeChannel(8, 'I');
					break;
				}
				case 15:
				{
					CInfo->Channels[0] = MakeChannel(5, 'B');
					CInfo->Channels[1] = MakeChannel(5, 'G');
					CInfo->Channels[2] = MakeChannel(5, 'R');
					CInfo->Channels[3] = MakeChannel(1, '?');
					break;
				}
				case 16:
				{
					CInfo->Channels[0] = MakeChannel(5, 'B');
					CInfo->Channels[1] = MakeChannel(6, 'G');
					CInfo->Channels[2] = MakeChannel(5, 'R');
					break;
				}
				case 24:
				{
					CInfo->Channels[0] = MakeChannel(8, 'R');
					CInfo->Channels[1] = MakeChannel(8, 'G');
					CInfo->Channels[2] = MakeChannel(8, 'B');
					break;
				}
				case 32:
				{
					CInfo->Channels[0] = MakeChannel(8, 'B');
					CInfo->Channels[1] = MakeChannel(8, 'G');
					CInfo->Channels[2] = MakeChannel(8, 'R');
					CInfo->Channels[3] = MakeChannel(8, 'A');
					break;
				}
			}

			int ChunkId = MBM_INFO2;
			int Size = CInfo->SizeOf();

			*this << ChunkId;
			*this << Size;
			if (NOT CInfo->Write(*this))
			{
				Error = TRUE;
			}
		}

		// Palette
		if (pCommon->GetBits() <= 8)
		{
			GdcPalette *Pal = pCommon->Palette();
			if (Pal)
			{
				int I = MBM_PALETTE;
				*this << I;
				I = sizeof(ulong) + (Pal->GetSize() * 3);
				*this << I;
				I = Pal->GetSize();
				*this << I;

				for (int i=0; i<I; i++)
				{
					GdcRGB *p = (*Pal)[i];
					*this << p->R;
					*this << p->G;
					*this << p->B;
				}
			}
			else
			{
				// Grayscale?
			}
		}

		if (pBmp)
		{
			if ((*pBmp)[0])
			{
				int ChunkId = MBM_PALETTE;
				int Line = GetLineLength(pBmp);
				int Size = Line * pBmp->Y();

				*this << ChunkId;
				*this << Size;
				if (Write((*pBmp)[0], Size) != Size)
				{
					Error = TRUE;
				}
			}
			else
			{
				Error = TRUE;
			}
		}
		else if (pRle)
		{
			int ChunkId = MBM_RLE;
			int Size = pRle->SizeOf();

			*this << ChunkId;
			*this << Size;

			if (NOT pRle->Write(*this))
			{
				Error = TRUE;
			}
		}
		else
		{
			Error = TRUE;
		}

		char *Comment = NULL;
		if (Props.Get("Comment", Comment))
		{
			MbmString Str;
			Str = Comment;

			int ChunkId = MBM_COMMENT;
			int Size = Str.SizeOf();

			*this << ChunkId;
			*this << Size;

			Str.Write(*this);
		}
	}

	if (NOT Error)
	{
		Status = TRUE;
	}

	return Status;
}

////////////////////////////////////////////////////////////////////
GFilterFactory *GFilterFactory::First = 0;
GFilterFactory::GFilterFactory()
{
	// add this filter to the global list
	Next = First;
	First = this;
}

GFilterFactory::~GFilterFactory()
{
	// delete from the global list
	if (First == this)
	{
		First = First->Next;
	}
	else
	{
		GFilterFactory *i = First;
		while (i->Next AND i->Next != this)
		{
			i = i->Next;
		}
		if (i->Next == this)
		{
			i->Next = Next;
		}
		else
		{
			// we aren't in the list
			// thats bad
			ASSERT(0);
		}
	}
}

GFilter *GFilterFactory::New(char *File)
{
	GFilterFactory *i = First;
	while (i)
	{
		if (i->CheckFile(File))
		{
			return i->NewObject();
		}
		i = i->Next;
	}
	return 0;
}

GFilter *GFilterFactory::NewAt(int n)
{
	int Status = 0;
	GFilterFactory *i = First;
	while (i)
	{
		if (Status++ == n)
		{
			return i->NewObject();
		}

		i = i->Next;
	}

	return 0;
}

int GFilterFactory::GetItems()
{
	int Status = 0;
	GFilterFactory *i = First;
	while (i)
	{
		Status++;
		i = i->Next;
	}
	return Status;
}

//////////////////////////////////////////////////////////////////////////
GSurface *LoadDC(char *Name)
{
	GSurface *pDC = 0;

	GFilter *F = GFilterFactory::New(Name);
	if (F AND F->Open(Name, O_READ))
	{
		// a file...
		pDC = NEW(GMemDC);
		if (pDC AND NOT F->ReadImage(pDC))
		{
			DeleteObj(pDC);
		}
	}
	DeleteObj(F);

	#if defined WIN32
	if (NOT pDC)
	{
		// a resource... lock and load gentlemen
		HRSRC hRsrc = FindResource(NULL, Name, RT_BITMAP);
		if (hRsrc)
		{
			int Size = SizeofResource(NULL, hRsrc);
			HGLOBAL hMem = LoadResource(NULL, hRsrc);
			if (hMem)
			{
				uchar *Ptr = (uchar*) LockResource(hMem);
				if (Ptr)
				{
					GClipBoard Clip(NULL);
					pDC = Clip.ConvertFromPtr(Ptr);
					GlobalUnlock(hMem);
				}
			}
		}
	}
	#elif defined BEOS
	if (pDC)
	{
		switch (pDC->GetBits())
		{
			case 8:
			{
				int Import8As16 = GdcD->GetOption(GDC_IMPORT_8_AS_16);
				if (Import8As16)
				{
					GSurface *pOld = pDC;
					GdcPalette *pPal = pDC->Palette();
					pDC = NEW(GMemDC);
					if (pOld AND
						pDC AND
						pDC->Create(pOld->X(), pOld->Y(), 16))
					{
						for (int y=0; y<pDC->Y(); y++)
						{
							uchar *s = (*pOld)[y];
							ushort *d = (ushort*) (*pDC)[y];
							
							for (int x=0; x<pDC->X(); x++)
							{
								GdcRGB *p = (*pPal)[s[x]];
								d[x] = (p) ? Rgb16(p->R, p->G, p->B) : 0;
							}
						}
						
						DeleteObj(pOld);
					}
				}
				else
				{
					// Create remap table
					color_map *Map = (color_map*) system_colors();
					GdcPalette *Palette = pDC->Palette();
					if (Map AND Palette)
					{
						char ReMap[256];
						for (int i=0; i<256; i++)
						{
							GdcRGB *p = (*Palette)[i];
							if (p)
							{
								COLOUR c = Rgb15(p->R, p->G, p->B);
								ReMap[i] = Map->index_map[c];
			
								p->R = Map->color_list[i].red;
								p->G = Map->color_list[i].green;
								p->B = Map->color_list[i].blue;
							}
							else
							{
								ReMap[i] = 0;
							}
						}
			
						// Remap colours to BeOS palette
						for (int y=0; y<pDC->Y(); y++)
						{
							uchar *d = (*pDC)[y];
							if (d)
							{
								for (int x=0; x<pDC->X(); x++)
								{
									d[x] = ReMap[d[x]];
								}
							}
						}
					}
				}
				break;
			}
		}
	}
	#endif

	return pDC;
}

bool WriteDC(char *Name, GSurface *pDC)
{
	bool Status = false;

	if (Name AND pDC)
	{
		GFilter *F = GFilterFactory::New(Name);
		if (F AND F->Open(Name, O_WRITE))
		{
			// a file...
			Status = F->WriteImage(pDC);
		}
		DeleteObj(F);
	}

	return Status;
}

GdcBmp::GdcBmp()
{
	Props.Set("Type", "Windows or OS/2 Bitmap");
	Props.Set("Extension", "BMP");
}
