/*hdr
**      FILE:           Jpeg.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           8/9/1998
**      DESCRIPTION:    Jpeg file filter
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "Gdc2.h"

#ifdef WIN32
#include "..\..\..\..\CodeLib\Jpeg\Code\jpeglib.h"
#else if BEOS
#include "../../../../CodeLib/jpeg/Code/jpeglib.h"
#endif
#include <setjmp.h>

// JPEG
class GdcJpeg : public GFilter {
public:
	GdcJpeg();

	int GetCapabilites() { return FILTER_CAP_READ | FILTER_CAP_WRITE; }
	bool ReadImage(GSurface *pDC);
	bool WriteImage(GSurface *pDC);
};

/////////////////////////////////////////////////////////////////////
class GdcJpegFactory : public GFilterFactory
{
	bool CheckFile(char *File)
	{
		return (File) ? stristr(File, ".jpeg") != 0 OR
						stristr(File, ".jpg") != 0 : false;
	}

	GFilter *NewObject()
	{
		return NEW(GdcJpeg);
	}

} JpegFactory;

/////////////////////////////////////////////////////////////////////
struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */

  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr *my_error_ptr;

METHODDEF(void)
my_error_exit (j_common_ptr cinfo)
{
	my_error_ptr myerr = (my_error_ptr) cinfo->err;
	(*cinfo->err->output_message) (cinfo);
	longjmp(myerr->setjmp_buffer, 1);
}

bool GdcJpeg::ReadImage(GSurface *pDC)
{
	bool Status = FALSE;
	char *filename = NewStr(Name);
	if (filename)
	{
		if (Open())
		{
			Close();
		}

		struct jpeg_decompress_struct cinfo;
		struct my_error_mgr jerr;

		FILE *infile;
		int row_stride;

		if (pDC AND (infile = fopen(filename, "rb")))
		{
			cinfo.err = jpeg_std_error(&jerr.pub);
			jerr.pub.error_exit = my_error_exit;

			if (setjmp(jerr.setjmp_buffer))
			{
				jpeg_destroy_decompress(&cinfo);
				fclose(infile);
				return 0;
			}

			jpeg_create_decompress(&cinfo);
			jpeg_stdio_src(&cinfo, infile);
			jpeg_read_header(&cinfo, TRUE);

			if (pDC->Create(cinfo.image_width, cinfo.image_height, cinfo.num_components*8))
			{
				// zero out bitmap
				pDC->Colour(0);
				pDC->Rectangle();

				// Progress
				if (Meter)
				{
					Meter->SetDescription("scanlines");
					Meter->SetLimits(0, cinfo.image_height-1);
				}

				// Read
				if (cinfo.num_components == 1)
				{
					// greyscale
					GdcPalette *Pal = NEW(GdcPalette(0, 256));
					if (Pal)
					{
						GdcRGB *p = (*Pal)[0];
						if (p)
						{
							for (int i=0; i<256; i++, p++)
							{
								p->R = i;
								p->G = i;
								p->B = i;
							}
						}
						pDC->Palette(Pal);
					}
				}

				// setup decompress
				jpeg_start_decompress(&cinfo);
				row_stride = cinfo.output_width * cinfo.output_components;

	#define Rc 0
	#define Gc (MaxRGB+1)
	#define Bc (MaxRGB+1)*2
	#define DownShift(x) ((int) ((x)+(1L << 15)) >> 16)
	#define UpShifted(x) ((int) ((x)*(1L << 16)+0.5))
	#define MaxRGB                          255

				long  *red   = NEW(long[3*(MaxRGB+1)]);
				long  *green = NEW(long[3*(MaxRGB+1)]);
				long  *blue  = NEW(long[3*(MaxRGB+1)]);
				uchar *range_table = NEW(uchar[3*(MaxRGB+1)]);

				if (red AND
					green AND
					blue AND
					range_table)
				{
					int i;
					for (i=0; i<=MaxRGB; i++)
					{
						range_table[i]=0;
						range_table[i+(MaxRGB+1)]=(uchar) i;
						range_table[i+(MaxRGB+1)*2]=MaxRGB;
					}
					uchar *range_limit=range_table+(MaxRGB+1);

					for (i=0; i<=MaxRGB; i++)
					{
						red[i+Rc]   = UpShifted(1.00000)*i;
						green[i+Rc] = 0;
						blue[i+Rc]  = UpShifted(1.40200*0.5) * ((i << 1)-MaxRGB);

						red[i+Gc]   = UpShifted(1.00000)*i;
						green[i+Gc] = (-UpShifted(0.34414*0.5)) * ((i << 1)-MaxRGB);
						blue[i+Gc]  = (-UpShifted(0.71414*0.5)) * ((i << 1)-MaxRGB);

						red[i+Bc]   = UpShifted(1.00000)*i;
						green[i+Bc] = UpShifted(1.77200*0.5) * ((i << 1)-MaxRGB);
						blue[i+Bc]  = 0;
					}

					// loop through scanlines
					while (cinfo.output_scanline < cinfo.output_height)
					{
						uchar *Ptr = (*pDC)[cinfo.output_scanline];
						if (Ptr)
						{
							if (jpeg_read_scanlines(&cinfo, &Ptr, 1))
							{
								switch (cinfo.jpeg_color_space)
								{
									case JCS_GRAYSCALE:
									{
										// do nothing
										break;
									}
									case JCS_CMYK:
									{
										/*
										RGB to CMY                         
										Cyan    = 1-Red                    
										Magenta = 1-Green                 
										Yellow  = 1-Blue                  

										CMY to RGB                   
										Red   = 1-Cyan
										Green = 1-Magenta 
										Blue  = 1-Yellow

										CMY to CMYK                             
										Black   = minimum (Cyan,Magenta,Yellow) 
										Cyan    = (Cyan-Black)/(1-Black)                
										Magenta = (Magenta-Black)/(1-Black)     
										Yellow  = (Yellow-Black)/(1-Black)      

										CMYK to CMY
										Cyan    = minimum(1,Cyan*(1-Black)+Black)
										Magenta = minimum(1,Magenta*(1-Black)+Black)
										Yellow  = minimum(1,Yellow*(1-Black)+Black)
										*/

										if (cinfo.num_components == 4)
										{
											Pixel32 *p = (Pixel32*) Ptr;
											for (int x=0; x<pDC->X(); x++, p++)
											{
												double C = (double) p->A / 255;
												double M = (double) p->R / 255;
												double Y = (double) p->G / 255;
												double K = (double) p->B / 255;

												p->R = (1.0 - min(1, C * (1.0 - K) + K)) * 255;
												p->G = (1.0 - min(1, M * (1.0 - K) + K)) * 255;
												p->B = (1.0 - min(1, Y * (1.0 - K) + K)) * 255;
											}
										}
										break;
									}
									case JCS_YCCK: // YCbCrK
									case 10000: // YCbCr
									{
										if (cinfo.num_components == 3)
										{
											uchar *p = (uchar*) Ptr;
											for (int x=0; x<pDC->X(); x++, p+=3)
											{
												/*
												Y =	0.299 * R	+ 0.587 * G	+ 0.114 * B 
												Cb =	-0.169 * R	- 0.331 * G	+ 0.500 * B 
												Cr =	0.500 * R	- 0.419 * G	- 0.081 * B
												*/

												int x=p[0];
												int y=p[1];
												int z=p[2];
												
												p[2]=range_limit[DownShift(red[x+Rc]+green[y+Rc]+blue[z+Rc])];
												p[1]=range_limit[DownShift(red[x+Gc]+green[y+Gc]+blue[z+Gc])];
												p[0]=range_limit[DownShift(red[x+Bc]+green[y+Bc]+blue[z+Bc])];
											}
										}
										break;
									}
									case JCS_YCbCr:
									case JCS_RGB:
									{
										if (cinfo.num_components == 3)
										{
											Pixel24 *p = (Pixel24*) Ptr;
											for (int x=0; x<pDC->X(); x++, p++)
											{
												uchar t = p->R;
												p->R = p->B;
												p->B = t;
											}
										}
										break;
									}
								}
							}
						}

						if (Meter) Meter->Value(cinfo.output_scanline);
					}
				}

				DeleteArray(red);
				DeleteArray(green);
				DeleteArray(blue);
				DeleteArray(range_table);

				jpeg_finish_decompress(&cinfo);
			}

			jpeg_destroy_decompress(&cinfo);
			fclose(infile);

			Status = TRUE;
		}

		DeleteArray(filename);
	}

	return Status;
}

bool GdcJpeg::WriteImage(GSurface *pDC)
{
	bool Status = FALSE;
	char *filename = NewStr(Name);

	if (Open())
	{
		SetSize(0);
		Close();
	}

	struct jpeg_compress_struct cinfo;
	struct my_error_mgr jerr;

	FILE *file;
	int row_stride;

	if (pDC AND (file = fopen(filename, "wb")))
	{
		cinfo.err = jpeg_std_error(&jerr.pub);
		jpeg_create_compress(&cinfo);
		jerr.pub.error_exit = my_error_exit;

		if (setjmp(jerr.setjmp_buffer))
		{
			jpeg_destroy_compress(&cinfo);
			fclose(file);
			return 0;
		}

		jpeg_stdio_dest(&cinfo, file);

		bool GreyScale = FALSE;
		GdcPalette *Pal = pDC->Palette();
		if (pDC->GetBits() == 8)
		{
			if (Pal)
			{
				GreyScale = TRUE;
				for (int i=0; i<Pal->GetSize(); i++)
				{
					GdcRGB *p = (*Pal)[i];
					if (	p AND
					       (p->R != i OR
						p->G != i OR
						p->B != i))
					{
						GreyScale = FALSE;
						break;
					}
				}
			}
			else if (NOT Pal)
			{
				GreyScale = TRUE;
			}
		}

		cinfo.image_width = pDC->X();
		cinfo.image_height = pDC->Y();
		if (GreyScale)
		{
			cinfo.input_components = 1;
			cinfo.in_color_space = JCS_GRAYSCALE;
		}
		else
		{
			cinfo.input_components = 3;
			cinfo.in_color_space = JCS_RGB;
		}

		jpeg_set_defaults(&cinfo);
		// jpeg_set_quality(&cinfo, quality, TRUE);
		jpeg_start_compress(&cinfo, TRUE);

		row_stride = pDC->X() * cinfo.input_components;
		uchar *Buffer = NEW(uchar[row_stride]);
		if (Buffer)
		{
			// Progress
			if (Meter)
			{
				Meter->SetDescription("scanlines");
				Meter->SetLimits(0, cinfo.image_height-1);
			}

			// Write
			while (cinfo.next_scanline < cinfo.image_height)
			{
				uchar *d = Buffer;

				switch (pDC->GetBits())
				{
					case 8:
					{
						if (GreyScale)
						{
							memcpy(Buffer, (*pDC)[cinfo.next_scanline], pDC->X());
						}
						else if (Pal)
						{
							GdcRGB *p = (*Pal)[0];
							if (p)
							{
								for (int x=0; x<pDC->X(); x++)
								{
									COLOUR c = pDC->Get(x, cinfo.next_scanline);
									*d++ = p[c].R;
									*d++ = p[c].G;
									*d++ = p[c].B;
								}
							}
						}
						break;
					}
					case 16:
					{
						for (int x=0; x<pDC->X(); x++)
						{
							COLOUR c = pDC->Get(x, cinfo.next_scanline);
							*d++ = (R16(c) << 3) | (R16(c) >> 5);
							*d++ = (G16(c) << 2) | (G16(c) >> 6);
							*d++ = (B16(c) << 3) | (B16(c) >> 5);
						}
						break;
					}
					case 24:
					{
						for (int x=0; x<pDC->X(); x++)
						{
							COLOUR c = pDC->Get(x, cinfo.next_scanline);
							*d++ = R24(c);
							*d++ = G24(c);
							*d++ = B24(c);
						}
						break;
					}
					case 32:
					{
						for (int x=0; x<pDC->X(); x++)
						{
							COLOUR c = pDC->Get(x, cinfo.next_scanline);
							*d++ = R32(c);
							*d++ = G32(c);
							*d++ = B32(c);
						}
						break;
					}
				}

				jpeg_write_scanlines(&cinfo, &Buffer, 1);
				if (Meter) Meter->Value(cinfo.next_scanline);
			}

			DeleteArray(Buffer);
		}

		jpeg_finish_compress(&cinfo);
		jpeg_destroy_compress(&cinfo);
		fclose(file);

		Status = TRUE;
	}

	DeleteArray(filename);

	return Status;
}

GdcJpeg::GdcJpeg()
{
	Props.Set("Type", "JPEG");
	Props.Set("Extension", "JPG,JPEG");
}
