/*hdr
**      FILE:           Tiff.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           9/10/98
**      DESCRIPTION:    Tiff filter
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>

#include "Lgi.h"
#include "Lzw.h"

// TIFF
#define TIFF_USE_LZW

class IFD {

	void *Data;

public:
	ushort	Tag;
	ushort	Type;
	ulong	Count;
	ulong	Offset;

	IFD();
	~IFD();
	void *GetData();
	char *Str();
	int ElementSize();
	int Value();
	int ArrayValue(int i);
	bool Read(GFile &f);
	bool Write(GFile &f);
};

class GdcTiff : public GFilter {

	List<IFD> Blocks;
	IFD *FindTag(ushort n);
	int ScanLength;

	#ifdef TIFF_USE_LZW
	Lzw Lib;
	#endif

	bool ProcessRead(GSurface *pDC);
	bool ProcessWrite(GSurface *pDC);

public:
	GView *Parent;
	OsView Handle();

	GdcTiff();
	~GdcTiff();

	int GetCapabilites() { return FILTER_CAP_READ; }
	bool ReadImage(GSurface *pDC);
	bool WriteImage(GSurface *pDC);
};

// Object Factory
class GdcTiffFactory : public GFilterFactory
{
	bool CheckFile(char *File)
	{
		return (File) ? stristr(File, ".tiff") != 0 OR
						stristr(File, ".tif") != 0 : false;
	}

	GFilter *NewObject()
	{
		return NEW(GdcTiff);
	}

} TiffFactory;

// Tiff tag numbers
#define TAG_ImageX						0x100
#define TAG_ImageY						0x101
#define TAG_Bits						0x102
#define TAG_Palette						0x140

#define TAG_Compression					0x103
#define TAG_RowsPerStrip				0x116
#define TAG_StripOffsets				0x111
#define TAG_StripByteOffsets			0x117
#define TAG_PhotometricInterpretation	0x106
#define TAG_Predictor					0x13D

// Tiff tag data types
#define TYPE_UBYTE						1
#define TYPE_USHORT						3
#define TYPE_ULONG						4

#define TYPE_SBYTE						6
#define TYPE_SSHORT						8
#define TYPE_SLONG						9

#define TYPE_ASCII						2
#define TYPE_URATIONAL					5
#define TYPE_UNDEFINED					7
#define TYPE_SRATIONAL					10
#define TYPE_FLOAT						11
#define TYPE_DOUBLE						12

////////////////////////////////////////////////////////////////////////////////////////////
/*
void Log(char *s, ...)
{
	GFile F;
	if (s AND F.Open("log.txt", O_WRITE))
	{
		char Buffer[512];

		va_list Arg;
		va_start(Arg, s);
		vsprintf(Buffer, s, Arg);
		va_end(Arg);

		F.Seek(0, SEEK_END);
		F.Write(Buffer, strlen(Buffer));
	}
}
*/

////////////////////////////////////////////////////////////////////////////////////////////
IFD::IFD()
{
	Data = 0;
}

IFD::~IFD()
{
	DeleteArray(Data);
}

void *IFD::GetData()
{
	return (Data) ? Data : &Offset;
}

int IFD::ElementSize()
{
	switch (Type)
	{
		case TYPE_UBYTE:
		case TYPE_SBYTE:
		case TYPE_ASCII:
		case TYPE_UNDEFINED:
		{
			return 1;
		}
		case TYPE_USHORT:
		case TYPE_SSHORT:
		{
			return 2;
		}
		case TYPE_ULONG:
		case TYPE_SLONG:
		case TYPE_FLOAT:
		{
			return 4;
		}
		case TYPE_URATIONAL:
		case TYPE_SRATIONAL:
		case TYPE_DOUBLE:
		{
			return 8;
		}
	}

	return 0;
}

char *IFD::Str()
{
	if (Data)
	{
		return (char*) Data;
	}
	return (char*) &Offset;
}

int IFD::Value()
{
	if (Count == 1)
	{
		switch (Type)
		{
			case TYPE_UBYTE:
			{
				return (int) ((uchar) (Offset & 0xFF));
			}
			case TYPE_USHORT:
			{
				return (int) ((ushort) (Offset & 0xFFFF));
			}
			case TYPE_ULONG:
			{
				return (int) ((ulong) Offset);
			}
			case TYPE_SBYTE:
			{
				return (int) ((signed char) (Offset & 0xFF));
			}
			case TYPE_SSHORT:
			{
				return (int) ((short) (Offset & 0xFFFF));
			}
			case TYPE_SLONG:
			{
				return (int) ((long) Offset);
			}
			case TYPE_ASCII:
			{
				char *s = Str();
				return (s) ? atoi(s) : 0;
			}
			case TYPE_URATIONAL:
			case TYPE_UNDEFINED:
			case TYPE_SRATIONAL:
			case TYPE_FLOAT:
			case TYPE_DOUBLE:
			{
				break;
			}
		}
	}
	return 0;
}

int IFD::ArrayValue(int i)
{
	void *p = GetData();
	if (Count > 1 AND p)
	{
		switch (Type)
		{
			case TYPE_ASCII:
			case TYPE_UBYTE:
			{
				return (int) ((uchar*)p)[i];
			}
			case TYPE_USHORT:
			{
				return (int) ((ushort*)p)[i];
			}
			case TYPE_ULONG:
			{
				return (int) ((ulong*)p)[i];
			}
			case TYPE_SBYTE:
			{
				return (int) ((signed char*)p)[i];
			}
			case TYPE_SSHORT:
			{
				return (int) ((short*)p)[i];
			}
			case TYPE_SLONG:
			{
				return (int) ((long*)p)[i];
			}
		}
	}
	return 0;
}

bool IFD::Read(GFile &f)
{
	f >> Tag;
	f >> Type;
	f >> Count;
	f >> Offset;

	int Size = Count * ElementSize();
	if (Size > 4)
	{
		Data = NEW(uchar[Size]);
		if (Data)
		{
			int Pos = f.GetPosition();
			f.Seek(Offset, SEEK_SET);
			f.Read(Data, Size);
			f.Seek(Pos, SEEK_SET);
		}
	}
	return TRUE;
}

bool IFD::Write(GFile &f)
{
	f << Tag;
	f << Type;
	f << Count;
	f << Offset;
	return TRUE;
}

////////////////////////////////////////////////////////////////////////////////////////////

GdcTiff::GdcTiff()
{
	Parent = 0;
	Props.Set("Type", "TIFF");
	Props.Set("Extension", "TIF,TIFF");
}

GdcTiff::~GdcTiff()
{
}

OsView GdcTiff::Handle()
{
	return (Parent) ? Parent->Handle() : 0;
}

IFD *GdcTiff::FindTag(ushort n)
{
	for (IFD *i = Blocks.First(); i; i = Blocks.Next())
	{
		if (i->Tag == n)
		{
			return i;
		}
	}

	return NULL;
}

bool GdcTiff::ProcessRead(GSurface *pDC)
{
	bool Status = FALSE;
	bool Error = FALSE;
	IFD *x = FindTag(TAG_ImageX);
	IFD *y = FindTag(TAG_ImageY);
	IFD *BitTag = FindTag(TAG_Bits);
	int Bits = 1;
	int Pos = GetPosition();

	if (BitTag)
	{
		switch (BitTag->Count)
		{
			case 1:
			{
				Bits = BitTag->Value();
				break;
			}
			case 3:
			{
				ushort *Depth = (ushort*) BitTag->GetData();
				if (Depth AND
					BitTag->Type == TYPE_USHORT AND
					Depth[0] == 8 AND
					Depth[1] == 8 AND
					Depth[2] == 8)
				{
					// Baseline RGB image
					Bits = 24;
				}
				break;
			}
			case 4:
			{
				// maybe a 32-bit image?
				break;
			}
		}
	}

	if (x AND
		y AND
		pDC AND
		pDC->Create(x->Value(), y->Value(), max(Bits, 8)))
	{
		// Read Palette
		IFD *Palette = FindTag(TAG_Palette);
		if (Palette)
		{
			ushort *s = (ushort*) Palette->GetData();
			if (s AND Palette->Type == TYPE_USHORT)
			{
				int Colours = Palette->Count / 3;
				GdcPalette *Pal = NEW(GdcPalette(0, Colours));
				if (Pal)
				{
					for (int i=0; i<Colours; i++)
					{
						GdcRGB *d = (*Pal)[i];
						if (d)
						{
							d->R = s[(0*Colours)+i] >> 8;
							d->G = s[(1*Colours)+i] >> 8;
							d->B = s[(2*Colours)+i] >> 8;
						}
					}

					pDC->Palette(Pal);
				}
			}

		}

		// Read Image
		IFD *Compression = FindTag(TAG_Compression);
		IFD *RowsPerStrip = FindTag(TAG_RowsPerStrip);
		IFD *StripOffsets = FindTag(TAG_StripOffsets);
		IFD *StripByteCounts = FindTag(TAG_StripByteOffsets);
		IFD *PhotometricInterpretation = FindTag(TAG_PhotometricInterpretation);
		if (Compression AND
			RowsPerStrip AND
			StripOffsets AND
			StripByteCounts)
		{
			int Comp = Compression->Value();
			int Rows = RowsPerStrip->Value();
			int Interpretation = PhotometricInterpretation->Value();

			switch (Bits)
			{
				case 8:
				case 24:
				{
					int Strip = 0;
					int Bpp = Bits / 8;
					ScanLength = pDC->X() * Bpp;

					for (	int y=0;
							NOT Error AND
							y<pDC->Y() AND
							Strip < StripOffsets->Count;
							Strip++)
					{
						int Offset = StripOffsets->ArrayValue(Strip);
						Lib.SetBufSize(StripByteCounts->ArrayValue(Strip));
						int Limit = min(pDC->Y() - y, Rows);
						Seek(Offset, SEEK_SET);

						switch (Comp)
						{
							case 1: // uncompressed
							{
								for (int i=0; i<Limit; i++, y++)
								{
									uchar *d = (*pDC)[y];
									if (d)
									{
										Read(d, ScanLength);
									}
								}
								break;
							}
							case 32773: // Pack-bits RLE
							{
								for (int i=0; i<Limit; i++, y++)
								{
									char n;
									uchar c;
									uchar *d = (*pDC)[y];
									if (d)
									{
										// unpack data
										for (uchar *e=d+ScanLength; d<e; )
										{
											(*this) >> n;
											if (n & 0x80)
											{
												// copy run
												(*this) >> c;
												n = -n+1;
												memset(d, c, n);
											}
											else
											{
												// copy data
												n++;
												Read(d, n);
											}
											d += n;
										}
									}
								}
								break;
							}
							case 2: // modified huffman
							{
								Props.Set(	"Error",
											"This image uses a modified huffman compression,\r\n"
											"Image does not support this.\r\n");
								Error = TRUE;
								break;
							}
							case 5: // LZW compression
							{
								#ifdef TIFF_USE_LZW
								BytePipe Out;
								if (Lib.Decompress(&Out, 0, this))
								{
									for (; Out.Sizeof() >= ScanLength; y++)
									{
										uchar *o = (*pDC)[y];
										if (o)
										{
											Out.Pop(o, ScanLength);
										}
									}
								}
								#else
								MessageBox(	MainWnd->Handle(),
											"This image uses LZW, which is unsupported due to\r\n"
											"UniSys requiring a royalty for it's patented algorithm.\r\n"
											"\r\n"
											"Use PNG instead.",
											"i.Mage: Tiff filter",
											MB_OK);
								Error = TRUE;
								#endif
								break;
							}
							default:
							{
								char Msg[256];
								sprintf(Msg, "This image uses an unknown compression method: %i", Comp);
								Props.Set("Error", Msg);
								Error = TRUE;
								break;
							}
						}
					}

					if (Bits == 24)
					{
						// swap R and B
						for (int y=0; y<pDC->Y(); y++)
						{
							uchar *d = (*pDC)[y];
							uchar n;
							for (int x=0; x<pDC->X(); x++, d+=3)
							{
								n = d[0];
								d[0] = d[2];
								d[2] = n;
							}
						}
					}

					IFD *Predictor = FindTag(TAG_Predictor);
					if (Predictor AND Predictor->Value() == 2)
					{
						switch (Bits)
						{
							case 8:
							{
								for (int y=0; y<pDC->Y(); y++)
								{
									uchar *d = (*pDC)[y];
									for (int x=1; x<pDC->X(); x++, d++)
									{
										d[0] += d[-1];
									}
								}
								break;
							}
							case 24:
							{
								for (int y=0; y<pDC->Y(); y++)
								{
									uchar *d = (*pDC)[y] + 3;
									for (int x=1; x<pDC->X(); x++, d+=3)
									{
										d[0] += d[-3];
										d[1] += d[-2];
										d[2] += d[-1];
									}
								}
								break;
							}
						}
					}
					break;
				}
				default:
				{
					char Msg[256];
					sprintf(Msg, "Image currently doesn't support %i bit TIFF files", Bits);
					Props.Set("Error", Msg);
					Error = TRUE;
					break;
				}
			}

			Status = NOT Error;
		}
	}

	Seek(Pos, SEEK_SET);

	return Status;
}

bool GdcTiff::ReadImage(GSurface *pDC)
{
	bool Status = FALSE;
	ushort n16;
	ulong n32;

	(*this) >> n16;
	if (n16 == 0x4D4D)
	{
		// file is big endian so we swap
		// Intel is little endian (i think)
		// TODO_PPC: change this
		SetSwap(TRUE);
	}

	(*this) >> n16;
	if (n16 == 42) // check magic number
	{
		bool Done = FALSE;
		bool Error = FALSE;

		ulong Offset;
		(*this) >> Offset;

		// loop through all the IFD's
		while (	NOT Done AND
			NOT Error AND
			NOT Eof())
		{
			if (Offset == 0)
			{
				Done = TRUE;
			}
			else
			{
				Seek(Offset, SEEK_SET);

				ushort IFDEntries;
				(*this) >> IFDEntries;

				for (int i=0; i<IFDEntries; i++)
				{
					IFD *Ifd = NEW(IFD);
					if (Ifd)
					{
						Blocks.Insert(Ifd);
						Ifd->Read(*this);
					}
					else
					{
						Error = TRUE;
					}
				}
			}

			if (NOT Error)
			{
				// process sub file
				Status = ProcessRead(pDC);
				Error = NOT Status;
			}

			(*this) >> Offset;
		}
	}

	return Status;
}

bool GdcTiff::ProcessWrite(GSurface *pDC)
{
	return FALSE;
}

bool GdcTiff::WriteImage(GSurface *pDC)
{
	return FALSE;
}
