/*hdr
**	FILE:			GdcFont.cpp
**	AUTHOR:		Matthew Allen
**	DATE:			5/5/97
**	DESCRIPTION:	GDC v2.xx True Type Font Support
**
**	Copyright (C) 1997-2001, Matthew Allen
**		fret@ozemail.com.au
*/

//////////////////////////////////////////////////////////////////////
// Includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "Lgi.h"

//////////////////////////////////////////////////////////////////////
// Helpers
#ifdef WIN32
int WinPointToHeight(int Pt)
{
	int Ht = 0;
	HWND hWnd = GetDesktopWindow();
	HDC hDC = GetDC(hWnd);
	if (hDC)
	{
		Ht = -MulDiv(Pt, GetDeviceCaps(hDC, LOGPIXELSY), 72);
		ReleaseDC(hWnd, hDC);
	}

	return Ht;
}

int WinHeightToPoint(int Ht)
{
	int Pt = 0;
	HWND hWnd = GetDesktopWindow();
	HDC hDC = GetDC(hWnd);
	if (hDC)
	{
		Pt = -MulDiv(Ht, 72, GetDeviceCaps(hDC, LOGPIXELSY));
		ReleaseDC(hWnd, hDC);
	}

	return Pt;
}
#endif

//////////////////////////////////////////////////////////////////////
// Functions
char *GdcGetCodePageDescription(int CodePage)
{
	static char *Description[GDCF_MAX] =
	{
		"Ascii",
		"ISO 8859-1 (West Europe)",
		"ISO 8859-2 (East Europe)",
		"ISO 8859-3",
		"ISO 8859-4 (Baltic)",
		"ISO 8859-5 (Russian)",
		"ISO 8859-6 (Arabic)",
		"ISO 8859-7 (Greek)",
		"ISO 8859-8 (Hebrew)",
		"ISO 8859-9 (Turkish)",
		"Shift JIS",
		"Chinese BIG5",
		"Thai",
		"Apple Mac"
	};

	return (CodePage >= 0 AND CodePage < GDCF_MAX) ? Description[CodePage] : 0;
}

char *GdcGetCodePageMimeType(int CodePage)
{
	static char *MimeType[GDCF_MAX] =
	{
		"us-ascii",
		"iso-8859-1",
		"iso-8859-2",
		"iso-8859-3",
		"iso-8859-4",
		"iso-8859-5",
		"iso-8859-6",
		"iso-8859-7",
		"iso-8859-8",
		"iso-8859-9",
		0, // "SHIFTJIS"
		0, // "CHINESEBIG5"
		0, // "THAI"
		0 // "MAC"
	};

	return (CodePage >= 0 AND CodePage < GDCF_MAX) ? MimeType[CodePage] : 0;
}

/*
short _gdc_usascii_mapping[128] =
{
	//80-8F
	0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
	//90-9F
	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
	//A0-AF
	0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
	//B0-BF
	0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
	//C0-CF
	0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
	//D0-DF
	0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
	//E0-EF
	0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
	//F0-FF
	0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff,
};
*/

short _gdc_ISO_8859_2_mapping[128] =
{
	//80-8F
	0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
	//90-9F
	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
	//A0-AF
	0xa0, 0xa5, 0xa2, 0xa3, 0xa4, 0xbc, 0x8c, 0xa7, 0xa8, 0x8a, 0xaa, 0x8d, 0x8f, 0xad, 0x8e, 0xaf,
	//B0-BF
	0xb0, 0xb9, 0xb2, 0xb3, 0xb4, 0xbe, 0x9c, 0xa1, 0xb8, 0x9a, 0xba, 0x9d, 0x9f, 0xbd, 0x9e, 0xbf,
	//C0-CF
	0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
	//D0-DF
	0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
	//E0-EF
	0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
	//F0-FF
	0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff,
};

bool GdcGetCodePageMappingTable(int CodePage, short *&Table)
{
	switch (CodePage)
	{
		case GDCF_ISO_8859_2:
		{
			Table = _gdc_ISO_8859_2_mapping;
			return true;
		}
	}

	return false;
}

bool GdcIsCodePageImplemented(int c)
{
	#ifdef _DEBUG
		return true;
	#else
		#ifdef WIN32
			return (c<GDCF_SHIFTJIS AND c!=GDCF_ISO_8859_3);
		#else
			return c == GDCF_ASCII;
		#endif
	#endif
}

//////////////////////////////////////////////////////////////////////
// Classes
GdcFont::GdcFont()
{
	_Face = 0;
	_Height = 8;
	_Width = 0;
	_Bold = false;
	_Italic = false;
	_Underline = false;
	_CodePage = GDCF_ASCII;
	_Fore = Rgb24(0, 0, 0);
	_Back = Rgb24(255, 255, 255);
	_TabSize = 32;
	_Transparent = false;
}

GdcFont::~GdcFont()
{
	DeleteArray(_Face);
}

// set
void GdcFont::Face(char *s)
{
	if (s AND
		s != _Face AND
		stricmp(s, _Face?_Face:(char*)"") != 0)
	{
		DeleteArray(_Face);
		_Face = NewStr(s);
		_OnPropChange(true);
	}
}

void GdcFont::Height(int i)
{
	_Height = i;
	_OnPropChange(true);
}

void GdcFont::Width(int i)
{
	_Width = i;
	_OnPropChange(true);
}

void GdcFont::CodePage(int i)
{
	_CodePage = i;
	_OnPropChange(true);
}

void GdcFont::TabSize(int i)
{
	_TabSize = i;
	_OnPropChange(false);
}

void GdcFont::Quality(int i)
{
	_Quality = i;
	_OnPropChange(true);
}

void GdcFont::Fore(COLOUR c)
{
	_Fore = c;
	_OnPropChange(false);
}

void GdcFont::Back(COLOUR c)
{
	_Back = c;
	_OnPropChange(false);
}

void GdcFont::Bold(bool i)
{
	_Bold = i;
	_OnPropChange(true);
}

void GdcFont::Italic(bool i)
{
	_Italic = i;
	_OnPropChange(true);
}

void GdcFont::Underline(bool i)
{
	_Underline = i;
	_OnPropChange(true);
}

void GdcFont::Transparent(bool i)
{
	_Transparent = i;
	_OnPropChange(false);
}

void GdcFont::Colour(COLOUR Fore, COLOUR Back)
{
	_Fore = Fore;
	_Back = Back;
	_OnPropChange(false);
}

////////////////////////
// get
char *GdcFont::Face()
{
	return _Face;
}

int GdcFont::Height()
{
	return _Height;
}

int GdcFont::Width()
{
	return _Width;
}

int GdcFont::CodePage()
{
	return _CodePage;
}

int GdcFont::TabSize()
{
	return _TabSize;
}

int GdcFont::Quality()
{
	return _Quality;
}

COLOUR GdcFont::Fore()
{
	return _Fore;
}

COLOUR GdcFont::Back()
{
	return _Back;
}

bool GdcFont::Bold()
{
	return _Bold;
}

bool GdcFont::Italic()
{
	return _Italic;
}

bool GdcFont::Underline()
{
	return _Underline;
}

bool GdcFont::Transparent()
{
	return _Transparent;
}

////////////////////////////////////////////////////////////////////
// GFont class, the implemention
GFont::GFont(char *face, int point)
{
	_New();
	if (face AND point > 0)
	{
		Create(face, point);
	}
}

GFont::GFont(GFontRef Handle)
{
	_New();

	GdcFontType Type;
	if (Type.GetFromRef(Handle))
	{
		Create(&Type);
	}
}

GFont::GFont(GdcFontType &Type)
{
	_New();

	Create(&Type);
}

GFont::GFont(GFont &Fnt)
{
	_New();
	*this = Fnt;
}

GFont::~GFont()
{
	_OnPropChange(true);
	DeleteObj(hFont);
}

void GFont::_New()
{
	#ifdef BEOS
	hFont = NEW(BFont);
	#else
	hFont = 0;
	#endif
	Dirty = true;
	Widths = 0;
	SizeofChar = 1;
}

void GFont::_OnPropChange(bool FontChange)
{
	if (FontChange)
	{
		#ifdef WIN32
		if (hFont)
		{
			DeleteObject(hFont);
			hFont = 0;
		}
		#else
		Dirty = true;
		#endif
		DeleteArray(Widths);
	}
}

bool GFont::IsValid()
{
	bool Status = false;

	// recreate font
	#ifdef WIN32
	if (NOT hFont)
	#else
	if (Dirty)
	#endif
	{
		Status = Create(Face(), Height());
		#ifdef BEOS
		Dirty = false;
		#endif
	}
	else
	{
		Status = true;
	}

	// find character widths
	if (NOT Widths)
	{
		int Len = 1 << 8;
		Widths = NEW(float[Len]);
		if (Widths)
		{
			#if defined WIN32

			if (hFont)
			{
				HDC hDC = CreateCompatibleDC(0);
				HFONT hOld = (HFONT) SelectObject(hDC, hFont);

				for (int i=0; i<Len; i++)
				{
					uchar c = i;
					SIZE Size;

					if (i == 146)
					{
						int n=0;
					}

					if (GetTextExtentPoint32(hDC, (const char*)&c, 1, &Size))
					{
						Widths[c] = Size.cx;
					}
					else
					{
						Widths[c] = 0;
					}
				}

				SelectObject(hDC, hOld);
				DeleteDC(hDC);
			}

			#elif defined BEOS

			float s = hFont->Size();
			for (int i=0; i<256; i++)
			{
				uchar Str[2] = { i, 0 };
				hFont->GetEscapements((char*)Str, 1, Widths+i);
				Widths[i] *= s;
			}

			#endif
		}
	}

	return Status;
}

bool GFont::Create(char *face, int height, int Param)
{
	if (face) Face(face);
	if (height > 0) Height(height);

	#if defined WIN32

	int CodePageMap[] =
	{
		ANSI_CHARSET,
		WESTEUROPE_CHARSET,
		EASTEUROPE_CHARSET,
		EASTEUROPE_CHARSET, // incorrect but I don't know what goes here
		BALTIC_CHARSET,
		RUSSIAN_CHARSET,
		ARABIC_CHARSET,
		GREEK_CHARSET,
		HEBREW_CHARSET,
		TURKISH_CHARSET,
		SHIFTJIS_CHARSET,
		CHINESEBIG5_CHARSET,
		THAI_CHARSET,
		ANSI_CHARSET // actually this is the MAC charset...
	};		

	if (hFont)
	{
		DeleteObject(hFont);
		hFont = 0;
	}

	HDC hDC = (Param) ? (HDC)Param : GetDC(0);
	int Win32Height = -MulDiv(_Height, GetDeviceCaps(hDC, LOGPIXELSY), 72);
	if (NOT Param) ReleaseDC(0, hDC);

	DeleteArray(Widths);

	hFont = ::CreateFont(	Win32Height,
							_Width,
							0,
							0,
							_Bold ? FW_BOLD : FW_NORMAL,
							_Italic,
							_Underline,
							false,
							CodePageMap[_CodePage],
							OUT_DEFAULT_PRECIS,
							CLIP_DEFAULT_PRECIS,
							_Quality,
							FF_DONTCARE,
							_Face);

	return (hFont != 0);
	
	#elif defined BEOS

	hFont->SetSize(_Height);

	if (Bold())
	{
		hFont->SetFamilyAndFace(_Face, B_BOLD_FACE);
	}
	else if (Italic())
	{
		hFont->SetFamilyAndFace(_Face, B_ITALIC_FACE);
	}
	else if (Underline())
	{
		hFont->SetFamilyAndFace(_Face, B_UNDERSCORE_FACE);
	}
	else
	{
		hFont->SetFamilyAndFace(_Face, B_REGULAR_FACE);
	}
	
	return true;

	#endif
}

#if defined WIN32
bool GFont::Create(GdcFontType *LogFont, int Param)
{
	if (hFont)
	{
		DeleteObject(hFont);
		hFont = 0;
	}
	DeleteArray(Widths);

	if (LogFont)
	{
		// set props
		HDC hDC = GetDC(0);
		Height(-MulDiv(LogFont->Info.lfHeight, 72, GetDeviceCaps(hDC, LOGPIXELSY)));
		ReleaseDC(0, hDC);

		Face(LogFont->Info.lfFaceName);
		Width(LogFont->Info.lfWidth);
		CodePage(LogFont->Info.lfCharSet);
		Quality(LogFont->Info.lfQuality);
		Bold(LogFont->Info.lfWeight >= FW_BOLD);
		Italic(LogFont->Info.lfItalic);
		Underline(LogFont->Info.lfUnderline);

		// create the handle
		Create(0, 0, Param);
	}

	return (hFont != 0);
}
#elif defined BEOS

bool GFont::Create(GdcFontType *LogFont, int Param)
{
	bool Status = false;

	if (LogFont)
	{
		hFont->SetFamilyAndFace(LogFont->Face, B_REGULAR_FACE);
		hFont->SetSize(LogFont->PointSize);
		
		// match our variables to the font
		_Italic = false;
		_Underline = false;
		_Bold = false;
		switch (hFont->Face())
		{
			case B_ITALIC_FACE:
			{
				_Italic = true;
				break;
			}
			case B_UNDERSCORE_FACE:
			{
				_Underline = true;
				break;
			}
			case B_BOLD_FACE:
			{
				_Bold = true;
				break;
			}
			default:
			case B_REGULAR_FACE:
			case B_NEGATIVE_FACE:
			case B_OUTLINED_FACE:
			case B_STRIKEOUT_FACE:
			{
				break;
			}
		}
		
		font_family family;
		font_style style;
		hFont->GetFamilyAndStyle(&family, &style);
		Status = true;

		DeleteArray(_Face);
		_Face = NewStr(family);
		_Height = (int) hFont->Size();
		_Width = 0;
		_CodePage = hFont->Encoding();
		_Quality = 0;
	}
	
	return Status;
}

#else
bool GFont::Create(GdcFontType *LogFont, int Param)
{
	return true;
}
#endif

void GFont::Text(GSurface *pDC, int x, int y, char *Str, int Len, GRegion *r, int TabOrigin)
{
	if (pDC AND
		Str AND
		IsValid())
	{
		#if defined WIN32

		if (Len < 0)
		{
			Len = strlen(Str);
		}
		if (Len > 0)
		{
			HDC hDC = pDC->StartDC();
			HFONT hOldFont = (HFONT) SelectObject(hDC, hFont);

			SetTextColor(hDC, _Fore & 0xFFFFFF);
			SetBkColor(hDC, _Back & 0xFFFFFF);
			SetBkMode(hDC, _Transparent ? TRANSPARENT : OPAQUE);

			if (_TabSize)
			{
				// tabbed text mode
				if (r)
				{
					if (NOT _Transparent)
					{
						pDC->Colour(_Back, 24);
						pDC->Rectangle(r);
					}

					// adjust clipping region to origin
					GRegion c = r;
					int x, y;
					pDC->GetOrigin(x, y);
					c.Offset(-x, -y);
					
					// clip the text.
					pDC->ClipRgn(&c);
				}

				TabbedTextOut(hDC, x, y, Str, Len, 1, &_TabSize, TabOrigin);

				if (r)
				{
					pDC->ClipRgn(0);
				}
			}
			else
			{
				// non tabbed text
				RECT rc = {0, 0, 0, 0};
				if (r) rc = *r;

				ExtTextOut(	hDC,
							x,
							y,
							((_Transparent) ? 0 : ETO_OPAQUE) |
							((r) ? ETO_CLIPPED : 0),
							(r) ? &rc : 0,
							Str,
							Len,
							NULL);
			}

			SelectObject(hDC, hOldFont);
			pDC->EndDC();
		}
		else if (r)
		{
			pDC->Colour(_Back, 24);
			pDC->Rectangle(r);
		}
	
		#elif defined BEOS

		bool OwnView = false;
		bool Lock = false;
		
		BView *View = 0;
		BBitmap *Bmp = 0;
		GMemDC *pMemDC = dynamic_cast<GMemDC*>(pDC);
		GdcViewDC *pViewDC = dynamic_cast<GdcViewDC*>(pDC);
		if (pViewDC)
		{
			View = pViewDC->Handle();
		}
		else if (pMemDC)
		{
			BRect f(0, 0, pDC->X()-1, pDC->Y()-1);
			View = NEW(BView(f, "Font View", B_FOLLOW_LEFT|B_FOLLOW_TOP, B_WILL_DRAW));
			Bmp = pMemDC->Handle();
			if (View AND
				Bmp AND
				Bmp->Lock())
			{
				Bmp->AddChild(View);
				OwnView = true;
			}
		}
		
		if (View)
		{
			Lock = View->LockLooper();

			// store old values
			rgb_color OldHigh = View->HighColor();
			rgb_color OldLow = View->LowColor();
			BFont Old;
			View->GetFont(&Old);

			// get the parameters ready
			font_height Height;
			hFont->GetHeight(&Height);
			View->SetFont(hFont);
			// BPoint Pt(x, y + Height.ascent);
			if (Len < 0) Len = strlen(Str);
			if (r) pDC->ClipRgn(r);
			
			if (NOT Transparent())
			{
				// draw the background
				pDC->Colour(_Back, 24);
				if (r)
				{
					pDC->Rectangle(r);
				}
				else
				{
					int dx, dy;
					Size(&dx, &dy, Str, Len, 0);
					pDC->Rectangle(x, y, x+dx, y+dy);
				}
			}

			if (r)
			{
				BRegion c(*r);
				View->ConstrainClippingRegion(&c);
			}

			for (char *s = Str; *s AND Len > 0; )
			{
				// find end of segment
				char *e;
				for (e = s; *e AND *e != 9 AND Len > 0; e++)
				{
					Len--;
				}
				int l = (int)e-(int)s;
				if (l > 0)
				{
					// draw the string
					View->SetHighColor(R24(_Fore), G24(_Fore), B24(_Fore));
					View->SetLowColor(R24(_Back), G24(_Back), B24(_Back));

					// draw segment
					BPoint Pt(x, y + Height.ascent);
					View->DrawString(s, l, Pt);

					int Dx = hFont->StringWidth(s, l);
					if (Underline())
					{
						View->StrokeLine(BPoint(x, y + Height.ascent + 1),	
										 BPoint(x+Dx, y + Height.ascent + 1));
					}
					
					x += Dx;					
				}

				while (*e AND *e == 9 AND Len-- > 0)
				{
					// draw tab space
					x += _TabSize;
					e++;
				}
				
				s = e;
			}
			
			if (r)
			{
				View->ConstrainClippingRegion(0);
			}

			// restore old settings
			View->SetHighColor(OldHigh);
			View->SetLowColor(OldLow);
			if (r) pDC->ClipRgn(NULL);
			View->SetFont(&Old);
		}
		
		if (OwnView)
		{
			View->Sync();
		}
		
		if (Lock) View->UnlockLooper();
		
		if (OwnView)
		{
			Bmp->RemoveChild(View);
			Bmp->Unlock();
			DeleteObj(View);
		}

		#endif
	}
}

void GFont::Size(int *x, int *y, char *Str, int Len, int Flags)
{
	if (Str AND
		IsValid())
	{
		#if defined WIN32
		HDC hDC = CreateCompatibleDC(NULL);
		HFONT hFnt = (HFONT) SelectObject(hDC, hFont);
		int Sx = 0;
		SIZE Size = {0, 0};
	
		if (Len < 0) Len = strlen(Str);

		if (Len == 0)
		{
			GetTextExtentPoint32(hDC, "A", 1, &Size);
		}
		else if (_TabSize)
		{
			for (int i=0; i<Len; i++)
			{
				if (Str[i] == 9)
				{
					Sx = ((Sx + _TabSize) / _TabSize) * _TabSize;
				}
				else
				{
					GetTextExtentPoint32(hDC, Str+i, 1, &Size);
					Sx += Size.cx;
				}
			}
		}
		else
		{
			GetTextExtentPoint32(hDC, Str, Len, &Size);
			Sx = Size.cx;
		}

		if (x) *x = Sx;
		if (y) *y = Size.cy;

		SelectObject(hDC, hFnt);
		DeleteDC(hDC);

		#elif defined BEOS

		if (x)
		{
			if (Len < 0)
			{
				Len = strlen(Str);
			}
			
			*x = 0;
			for (char *s = Str; *s AND Len > 0; )
			{
				char *e;
				for (e = s; *e AND *e != 9 AND Len > 0; e++)
				{
					Len--;
				}
				int l = (int)e-(int)s;
				if (l > 0)
				{
					*x += hFont->StringWidth(s, l);
				}
				
				while (*e AND *e == 9 AND Len-- > 0)
				{
					*x += _TabSize;
					e++;
				}
				s = e;
			}
		}
		
		if (y)
		{
			font_height Height;
			hFont->GetHeight(&Height);
			*y = Height.ascent + Height.descent + Height.leading + 0.9999;
		}
		#endif
	}
}

int GFont::CharAt(int x, char *Str, int Len)
{
	if (x < 0)
	{
		return 0;
	}

	if (IsValid())
	{
		float Cx = 0.0;
		int Chars = -1;
		for (int i=0; i<Len; i++)
		{
			float Nx;
			if (Str[i] == 9)
			{
				Nx = (Cx - ((int)Cx % _TabSize)) + _TabSize;
			}
			else
			{
				Nx = Cx + Widths[(uchar)Str[i]];
			}

			if (x >= Cx AND x < Nx)
			{
				Chars = i;
				break;
			}
			Cx = Nx;
		}

		return (Chars < 0) ? Len : Chars;
	}

	return 0;
}

int GFont::X(char *Str, int Len)
{
	int x = 0;
	Size(&x, 0, Str, Len, 0);
	return x;
}

int GFont::Y(char *Str, int Len)
{
	int y = 0;
	Size(0, &y, Str, Len, 0);
	return y;
}

GFont &GFont::operator =(GFont &f)
{
	Face(f.Face());
	Height(f.Height());
	Width(f.Width());
	CodePage(f.CodePage());
	TabSize(f.TabSize());
	Quality(f.Quality());
	Fore(f.Fore());
	Back(f.Back());
	Bold(f.Bold());
	Italic(f.Italic());
	Underline(f.Underline());
	Transparent(f.Transparent());
	
	return *this;
}

float *GFont::GetWidths()
{
	return Widths;
}

///////////////////////////////////////////////////////////////////////
// Unicode versions
void GFont::Text(	GSurface *pDC,
					int x, int y,
					ushort *Str,
					int Len,
					GRegion *r,
					int TabOrigin)
{
	if (pDC AND
		Str AND
		IsValid())
	{
		#if defined WIN32

		if (Len < 0)
		{
			Len = wcslen(Str);
		}
		if (Len > 0)
		{
			HDC hDC = pDC->StartDC();
			HFONT hOldFont = (HFONT) SelectObject(hDC, hFont);

			SetTextColor(hDC, _Fore & 0xFFFFFF);
			SetBkColor(hDC, _Back & 0xFFFFFF);
			SetBkMode(hDC, _Transparent ? TRANSPARENT : OPAQUE);

			if (_TabSize)
			{
				// tabbed text mode
				if (r)
				{
					if (NOT _Transparent)
					{
						pDC->Colour(_Back, 24);
						pDC->Rectangle(r);
					}

					// adjust clipping region to origin
					GRegion c = r;
					int x, y;
					pDC->GetOrigin(x, y);
					c.Offset(-x, -y);
					
					// clip the text.
					pDC->ClipRgn(&c);
				}

				TabbedTextOutW(hDC, x, y, Str, Len, 1, &_TabSize, TabOrigin);

				if (r)
				{
					pDC->ClipRgn(0);
				}
			}
			else
			{
				// non tabbed text
				RECT rc = {0, 0, 0, 0};
				if (r) rc = *r;

				ExtTextOutW(hDC,
							x,
							y,
							((_Transparent) ? 0 : ETO_OPAQUE) |
							((r) ? ETO_CLIPPED : 0),
							(r) ? &rc : 0,
							Str,
							Len,
							NULL);
			}

			SelectObject(hDC, hOldFont);
			pDC->EndDC();
		}
		else if (r)
		{
			pDC->Colour(_Back, 24);
			pDC->Rectangle(r);
		}

		#else

		LgiAssert(0);

		#endif
	}
}

void GFont::Size(int *x, int *y, ushort *Str, int Len, int Flags)
{
	if (Str AND
		IsValid())
	{
		#if defined WIN32
		HDC hDC = CreateCompatibleDC(NULL);
		HFONT hFnt = (HFONT) SelectObject(hDC, hFont);
		int Sx = 0;
		SIZE Size = {0, 0};
	
		if (Len < 0) Len = wcslen(Str);

		if (Len == 0)
		{
			GetTextExtentPoint32W(hDC, L"A", 1, &Size);
		}
		else if (_TabSize)
		{
			for (int i=0; i<Len; i++)
			{
				if (Str[i] == 9)
				{
					Sx = ((Sx + _TabSize) / _TabSize) * _TabSize;
				}
				else
				{
					GetTextExtentPoint32W(hDC, Str+i, 1, &Size);
					Sx += Size.cx;
				}
			}
		}
		else
		{
			GetTextExtentPoint32W(hDC, Str, Len, &Size);
			Sx = Size.cx;
		}

		if (x) *x = Sx;
		if (y) *y = Size.cy;

		SelectObject(hDC, hFnt);
		DeleteDC(hDC);

		#elif defined BEOS

		LgiAssert(0);

		#endif
	}
}

int GFont::X(ushort *Str, int Len)
{
	int x;
	Size(&x, 0, Str, Len);
	return x;
}

int GFont::Y(ushort *Str, int Len)
{
	int y;
	Size(0, &y, Str, Len);
	return y;
}

int GFont::CharAt(int x, ushort *Str, int Len)
{
	if (x < 0)
	{
		return 0;
	}

	if (IsValid())
	{
		float Cx = 0.0;
		int Chars = -1;
		for (int i=0; i<Len; i++)
		{
			float Nx;
			if (Str[i] == 9)
			{
				Nx = (Cx - ((int)Cx % _TabSize)) + _TabSize;
			}
			else
			{
				Nx = Cx + Widths[Str[i]];
			}

			if (x >= Cx AND x < Nx)
			{
				Chars = i;
				break;
			}
			Cx = Nx;
		}

		return (Chars < 0) ? Len : Chars;
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////
GdcFontType::GdcFontType(char *face, int pointsize)
{
	#if defined WIN32

	ZeroObj(Info);
	if (face)
	{
		strcpy(Info.lfFaceName, face);
	}
	if (pointsize)
	{
		Info.lfHeight = WinHeightToPoint(pointsize);
	}

	#else

	Face[0] = 0;
	if (face)
	{
		strcpy(Face, face);
	}
	PointSize = pointsize;

	#endif
}

GdcFontType::~GdcFontType()
{
}

bool GdcFontType::DoUI(GView *Parent)
{
	bool Status = false;

	#if defined WIN32

	CHOOSEFONT Dlg;

	ZeroObj(Dlg);
	Dlg.lStructSize = sizeof(Dlg);
	Dlg.hwndOwner = Parent ? Parent->Handle() : 0;
	Dlg.Flags = CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT;
	Dlg.lpLogFont = &Info;
	
	Status = ChooseFont(&Dlg);

	#elif defined BEOS

	char Temp[128];
	sprintf(Temp, "%s,%i", Face, PointSize);
	GSelectFont f(Parent, Temp);
	if (f.DoModal())
	{
		strcpy(Face, f.Font);
		PointSize = f.Size;
		Status = true;
	}

	#endif

	return Status;
}

bool GdcFontType::GetDescription(char *Str)
{
	#if defined WIN32
	if (Str AND ValidStr(Info.lfFaceName))
	{
		sprintf(Str, "%s, %i pt", Info.lfFaceName, WinHeightToPoint(Info.lfHeight));
		return true;
	}
	#else
	if (Str AND ValidStr(Face))
	{
		sprintf(Str, "%s, %i pt", Face, PointSize);
		return true;
	}
	#endif

	return false;
}

bool GdcFontType::Serialize(ObjProperties *Options, char *OptName, bool Write)
{
	bool Status = false;

	if (Options AND OptName)
	{
		#if defined WIN32
		if (Write)
		{
			Status = Options->Set(OptName, &Info, sizeof(Info));
		}
		else
		{
			void *Data = 0;
			int Len = 0;
			if (Options->Get(OptName, Data, Len) AND
				Len == sizeof(Info) AND
				Data)
			{
				memcpy(&Info, Data, Len);
				Status = true;
			}
		}
		#else
		if (Write)
		{
			char Temp[128];
			sprintf(Temp, "%s,%i pt", Face, PointSize);
			Status = Options->Set(OptName, Temp);
		}
		else
		{
			char *Temp = 0;
			if (Options->Get(OptName, Temp) AND ValidStr(Temp))
			{
				char *OurCopy = NewStr(Temp);
				if (OurCopy)
				{
					char *Comma = strchr(OurCopy, ',');
					if (Comma)
					{
						*Comma++ = 0;
						strcpy(Face, OurCopy);
						PointSize = atoi(Comma);
						Status = true;
					}

					DeleteArray(OurCopy);
				}
			}
		}
		#endif
	}

	return Status;
}

bool GdcFontType::GetConfigFont(char *Tag)
{
	bool Status = false;

	XmlTag *Font = LgiApp->GetConfig(Tag);
	if (Font)
	{
		// read from config file
		char *f, *p;
		if (Font->Get("Face", f) AND Font->Get("Point", p))
		{
			#if defined WIN32
			strcpy(Info.lfFaceName, f);
			Info.lfHeight = WinPointToHeight(atoi(p));
			#elif defined BEOS
			strcpy(Face, f);
			PointSize = atoi(p);
			#endif
			Status = true;
		}
	}

	return Status;
}

bool GdcFontType::GetSystemFont(char *Which)
{
	bool Status = false;

	if (Which)
	{
		#ifdef WIN32
		NONCLIENTMETRICS info;
		info.cbSize = sizeof(info);
		bool InfoOk = SystemParametersInfo(	SPI_GETNONCLIENTMETRICS,
											sizeof(info),
											&info,
											0);

		// Convert windows font height into points
		int Height = WinHeightToPoint(info.lfMessageFont.lfHeight);
		#endif

		if (NOT stricmp(Which, "System"))
		{
			Status = GetConfigFont("Font.System");
			if (NOT Status)
			{
				// read from system
				#if defined WIN32

				if (InfoOk)
				{
					// Copy the font metrics
					memcpy(&Info, &info.lfMessageFont, sizeof(Info));
					Status = true;
				}

				#elif defined BEOS || defined ATHEOS

				// BeOS has no system wide setting so give a valid default
				strcpy(Face, "Swis721 BT");
				PointSize = 11;
				Status = true;

				#endif
			}
		}
		else if (NOT stricmp(Which, "Menu"))
		{
			Status = GetConfigFont("Font.Menu");
			if (NOT Status)
			{
				#if defined WIN32
				if (InfoOk)
				{
					// Copy the font metrics
					memcpy(&Info, &info.lfMenuFont, sizeof(Info));
					Status = true;
				}
				#endif
			}
		}
		else if (NOT stricmp(Which, "Caption"))
		{
			Status = GetConfigFont("Font.Caption");
			if (NOT Status)
			{
				#if defined WIN32
				if (InfoOk)
				{
					// Copy the font metrics
					memcpy(&Info, &info.lfCaptionFont, sizeof(Info));
					Status = true;
				}
				#endif
			}
		}
		else if (NOT stricmp(Which, "Status"))
		{
			Status = GetConfigFont("Font.Status");
			if (NOT Status)
			{
				#if defined WIN32
				if (InfoOk)
				{
					// Copy the font metrics
					memcpy(&Info, &info.lfStatusFont, sizeof(Info));
					Status = true;
				}
				#endif
			}
		}
		else if (NOT stricmp(Which, "Small"))
		{
			Status = GetConfigFont("Font.Small");
			if (NOT Status)
			{
				#if defined WIN32
				if (InfoOk)
				{
					// Copy the font metrics
					memcpy(&Info, &info.lfSmCaptionFont, sizeof(Info));

					// Make it a bit smaller than the system font
					Info.lfHeight = WinPointToHeight(WinHeightToPoint(info.lfSmCaptionFont.lfHeight)-1);

					Status = true;
				}
				#elif defined BEOS

				// BeOS has no system wide setting so give a valid default
				strcpy(Face, "Swis721 BT");
				PointSize = 9;
				Status = true;

				#endif
			}
		}
	}

	return Status;
}

bool GdcFontType::GetFromRef(GFontRef Handle)
{
	#if defined WIN32
	return GetObject(Handle, sizeof(Info), &Info) == sizeof(Info);
	#else
	// FIXME
	LgiAssert(0);
	return false;
	#endif
}

GFont *GdcFontType::CreateFont(int Param)
{
	GFont *New = NEW(GFont);
	if (New AND
		NOT New->Create(this, Param))
	{
		DeleteObj(New);
	}
	return New;
}


