/*
**	FILE:		GDateTime.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		11/11/98
**	DESCRIPTION:	Scribe Date Time Object
**
**	Copyright (C) 1998, Matthew Allen
**		fret@ozemail.com.au
*/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <stdarg.h>

#include "Gdc2.h"
#include "GDateTime.h"

//////////////////////////////////////////////////////////////////////////////
GDateTime::GDateTime()
{
	_Day = 0;
	_Month = 0;
	_Year = 0;
	_Thousands = 0;
	_Seconds = 0;
	_Minutes = 0;
	_Hours = 0;
}

GDateTime::~GDateTime()
{
}

void GDateTime::SetNow()
{
	time_t now;
	time(&now);
	struct tm *time = localtime(&now);

    _Seconds = time->tm_sec;
    _Minutes = time->tm_min;
    _Hours = time->tm_hour;
    _Day = time->tm_mday;
    _Month = time->tm_mon + 1;
    _Year = time->tm_year + 1900;
}

#define Convert24HrTo12Hr(h)			( (h) == 0 ? 12 : (h) > 12 ? (h) % 12 : (h) )
#define Convert24HrToAmPm(h)		( (h) >= 12 ? "p" : "a" )

void GDateTime::GetDate(char *Str)
{
	if (Str)
	{
		#ifdef WIN32
		char Data[2] = {'1', 0};
		GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_IDATE, Data, sizeof(Data));
		switch (atoi(Data))
		{
			case 0:
				sprintf(Str, "%i/%i/%i", _Month, _Day, _Year);
				break;
			default:
			case 1:
				sprintf(Str, "%i/%i/%i", _Day, _Month, _Year);
				break;
			case 2:
				sprintf(Str, "%i/%i/%i", _Year, _Month, _Day);
				break;
		}
		#else
		sprintf(Str, "%i/%i/%i", _Day, _Month, _Year);
		#endif
	}
}

void GDateTime::GetTime(char *Str)
{
	if (Str)
	{
		sprintf(Str, "%i:%i:%i", _Hours, _Minutes, _Seconds);
	}
}

void GDateTime::Get(char *Str)
{
	if (Str)
	{
		const char *m = Convert24HrToAmPm(_Hours);

		#ifdef WIN32
		char Data[2] = {'1', 0};
		GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_IDATE, Data, sizeof(Data));
		switch (atoi(Data))
		{
			case 0:
			{
				sprintf(Str, "%i/%i/%i %i:%02.2i:%02.2i%s", _Month, _Day, _Year, Convert24HrTo12Hr(_Hours), _Minutes, _Seconds, m);
				break;
			}
			default:
			case 1:
			{
				sprintf(Str, "%i/%i/%i %i:%02.2i:%02.2i%s", _Day, _Month, _Year, Convert24HrTo12Hr(_Hours), _Minutes, _Seconds, m);
				break;
			}
			case 2:
			{
				sprintf(Str, "%i/%i/%i %i:%02.2i:%02.2i%s", _Year, _Month, _Day, Convert24HrTo12Hr(_Hours), _Minutes, _Seconds, m);
				break;
			}
		}
		#else
		sprintf(Str, "%i/%i/%i %i:%02.2i:%02.2i%s", _Day, _Month, _Year, Convert24HrTo12Hr(_Hours), _Minutes, _Seconds, m);
		#endif
	}
}

int GDateTime::Sizeof()
{
	return	sizeof(_Day) +
			sizeof(_Month) +
			sizeof(_Year) +
			sizeof(_Thousands) +
			sizeof(_Seconds) +
			sizeof(_Minutes) +
			sizeof(_Hours);
}

bool GDateTime::Serialize(GFile &f, bool Write)
{
	if (Write)
	{
		f << _Day;
		f << _Month;
		f << _Year;
		f << _Thousands;
		f << _Seconds;
		f << _Minutes;
		f << _Hours;
	}
	else
	{
		f >> _Day;
		f >> _Month;
		f >> _Year;
		f >> _Thousands;
		f >> _Seconds;
		f >> _Minutes;
		f >> _Hours;
	}

	return TRUE;
}

int GDateTime::Compare(GDateTime *d)
{
	int c = 0;
	if (d)
	{
		c = _Year - d->_Year;
		if (NOT c)
		{
			c = _Month - d->_Month;
			if (NOT c)
			{
				c = _Day - d->_Day;
				if (NOT c)
				{
					c = _Hours - d->_Hours;
					if (NOT c)
					{
						c = _Minutes - d->_Minutes;
						if (NOT c)
						{
							c = _Seconds - d->_Seconds;
							if (NOT c)
							{
								c = _Thousands - d->_Thousands;
							}
						}
					}
				}
			}
		}
	}
	return c;
}

bool GDateTime::operator <(GDateTime &dt)
{
	if (_Year < dt._Year) return TRUE;
	else if (_Year > dt._Year) return FALSE;

	if (_Month < dt._Month) return TRUE;
	else if (_Month > dt._Month) return FALSE;

	if (_Day < dt._Day) return TRUE;
	else if (_Day > dt._Day) return FALSE;

	if (_Hours < dt._Hours) return TRUE;
	else if (_Hours > dt._Hours) return FALSE;

	if (_Minutes < dt._Minutes) return TRUE;
	else if (_Minutes > dt._Minutes) return FALSE;

	if (_Seconds < dt._Seconds) return TRUE;
	else if (_Seconds > dt._Seconds) return FALSE;

	if (_Thousands < dt._Thousands) return TRUE;
	else if (_Thousands > dt._Thousands) return FALSE;

	return FALSE;
}

bool GDateTime::operator <=(GDateTime &dt)
{
	return NOT (*this > dt);
}

bool GDateTime::operator >(GDateTime &dt)
{
	if (_Year > dt._Year) return TRUE;
	else if (_Year < dt._Year) return FALSE;

	if (_Month > dt._Month) return TRUE;
	else if (_Month < dt._Month) return FALSE;

	if (_Day > dt._Day) return TRUE;
	else if (_Day < dt._Day) return FALSE;

	if (_Hours > dt._Hours) return TRUE;
	else if (_Hours < dt._Hours) return FALSE;

	if (_Minutes > dt._Minutes) return TRUE;
	else if (_Minutes < dt._Minutes) return FALSE;

	if (_Seconds > dt._Seconds) return TRUE;
	else if (_Seconds < dt._Seconds) return FALSE;

	if (_Thousands > dt._Thousands) return TRUE;
	else if (_Thousands < dt._Thousands) return FALSE;

	return FALSE;
}

bool GDateTime::operator >=(GDateTime &dt)
{
	return NOT (*this < dt);
}

bool GDateTime::operator ==(GDateTime &dt)
{
	return	_Year == dt._Year AND
			_Month == dt._Month AND
			_Day == dt._Day AND
			_Hours == dt._Hours AND
			_Minutes == dt._Minutes AND
			_Seconds == dt._Seconds AND
			_Thousands == dt._Thousands;
}

bool GDateTime::operator !=(GDateTime &dt)
{
	return	_Year != dt._Year OR
			_Month != dt._Month OR
			_Day != dt._Day OR
			_Hours != dt._Hours OR
			_Minutes != dt._Minutes OR
			_Seconds != dt._Seconds OR
			_Thousands != dt._Thousands;
}

bool GDateTime::IsLeapYear(int Year)
{
	if (Year < 0) Year = _Year;

	if (Year % 4 != 0)
	{
		return false;
	}

	if (Year % 400 == 0)
	{
		return true;
	}

	if (Year % 100 == 0)
	{
		return false;
	}

	return true;
}

int GDateTime::DaysInMonth(int Month, int Year)
{
	if (Month == 2 AND
		IsLeapYear(Year))
	{
		return 29;
	}

	short DaysInMonth[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	return DaysInMonth[Month-1];
}

#define MinutesInDay (60*24)

void GDateTime::AddMinutes(int Minutes)
{
	_Minutes += Minutes;

	do
	{
		if (_Minutes < 0)
		{
			AddHours(-1);
			_Minutes += 60;
		}
		else if (_Minutes >= 60)
		{
			AddHours(1);
			_Minutes -= 60;
		}
		else
		{
			break;
		}
	}
	while (1);
}

void GDateTime::AddHours(int Hours)
{
	_Hours += Hours;

	do
	{
		if (_Hours < 0)
		{
			AddDays(-1);
			_Hours += 24;
		}
		else if (_Hours >= 24)
		{
			AddDays(1);
			_Hours -= 24;
		}
		else
		{
			break;
		}
	}
	while (1);
}

void GDateTime::AddDays(int Days)
{
	_Day += Days;

	do
	{
		if (_Day < 1)
		{
			AddMonths(-1);
			_Day += DaysInMonth(_Month);
		}
		else if (_Day > DaysInMonth(_Month))
		{
			_Day -= DaysInMonth(_Month);
			AddMonths(1);
		}
		else
		{
			break;
		}
	}
	while (1);
}

void GDateTime::AddMonths(int Months)
{
	_Month += Months;

	do
	{
		if (_Month < 1)
		{
			_Year--;
			_Month += 12;
		}
		else if (_Month > 12)
		{
			_Year++;
			_Month -= 12;
		}
		else
		{
			break;
		}
	}
	while (1);
}


