/*
**	FILE:			FtpApp.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			7/9/99
**	DESCRIPTION:	i.Ftp application
**
**	Copyright (C) 1999, Matthew Allen
**		fret@ozemail.com.au
*/

// Debug defines

// Includes
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <stdarg.h>

#include "Lgi.h"
#include "FileTransferProgress.h"

//////////////////////////////////////////////////////////////////
#define THROTTLE_TEXT_WIDTH		70
#define SLIDER_ID				100

int PipeSize[] = { 29491, 57344, 65536, 131072, 1048576, 0 };

class GPaneThrottle : public GStatusPane
{
	GSlider *Slider;
	FileTransferWindow *App;
	int Pipe;

	void Set()
	{
		GRegion r(0, 0, THROTTLE_TEXT_WIDTH, Y()-1);
		GView::Invalidate(&r);

		if (App)
		{
			App->SetDataRate(	(Value() == 100) ? -1 :
								(Value() * (PipeSize[Pipe]/8)) / 100 );

			ObjProperties *Options = App->GetOptions();
			if (Options)
			{
				Options->Set(OPT_Throttle, Value());
				Options->Set(OPT_PipeSize, Pipe);
			}
		}
	}

public:
	GPaneThrottle(FileTransferWindow *app);

	int OnNotify(GView *Ctrl, int Flags);
	void OnPaint(GSurface *pDC);
	int Value();
	void OnMouseClick(GMouse &m);
};

GPaneThrottle::GPaneThrottle(FileTransferWindow *app)
{
	Pipe = -1;
	App = app;
	Width = 80+THROTTLE_TEXT_WIDTH;
	Slider = NEW(GSlider(SLIDER_ID, 0, 0, 100, 20, "Throttle", false));
}

int GPaneThrottle::OnNotify(GView *Ctrl, int Flags)
{
	if (Ctrl AND Ctrl->GetId() == SLIDER_ID)
	{
		Set();
	}

	return 0;
}

void GPaneThrottle::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);

	int x, y;
	pDC->GetOrigin(x, y);

	LgiThinBorder(pDC, r, SUNKEN);
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(r.x1, r.y1, r.x1+THROTTLE_TEXT_WIDTH, r.y2);

	if (Slider)
	{
		GRegion Sr = r;
		#ifdef WIN32
		Sr.Offset(Pos.x1, Pos.y1);
		#endif
		Sr.x1 += THROTTLE_TEXT_WIDTH+1;
		Slider->SetPos(Sr);

		if (NOT Slider->GetParent())
		{
			#ifdef BEOS
			Slider->GView::Attach(this);
			#else if WIN32
			Slider->Attach(Parent);
			#endif
			Slider->SetNotify(this);
			Slider->SetLimits(0, 100);
			Slider->Value(100);

			if (Pipe < 0)
			{
				ObjProperties *Options = (App)?App->GetOptions():0;
				if (Options)
				{
					int i = 100;
					Options->Get(OPT_PipeSize, Pipe);
					if (Slider)
					{
						Options->Get(OPT_Throttle, i);
						Slider->Value(i);
					}
				}
			}
		}

		char Str[256];
		int v = Value();
		if (v < 100)
		{
			sprintf(Str, "%i%% of %iK", v, PipeSize[Pipe]>>10);
		}
		else
		{
			strcpy(Str, "No limit");
		}
		SysFont->Colour(0, LC_MED);
		SysFont->Text(pDC, r.x1+2, r.y1, Str);
	}
}

int GPaneThrottle::Value()
{
	return (Slider) ? Slider->Value() : 100;
}

void GPaneThrottle::OnMouseClick(GMouse &m)
{
	if (m.Down())
	{
		GSubMenu *RClick = NEW(GSubMenu);
		if (RClick)
		{
			for (int i=0; PipeSize[i]; i++)
			{
				char Str[256];
				sprintf(Str, "%iK", PipeSize[i]>>10);
				GMenuItem *Item = RClick->AppendItem(Str, 100+i, true);
				if (Item AND i == Pipe)
				{
					Item->Checked(true);
				}
			}

			if (GetMouse(m, true))
			{
				int Cmd = RClick->Float(this, m.x, m.y, m.Left());
				if (Cmd >= 100)
				{
					Pipe = Cmd - 100;
					Set();
				}
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
#define HISTORY_TEXT_WIDTH		60
#define MAX_SAMPLE				70
#define HISTORY_DIV				1000
class GPaneHistory : public GStatusPane
{
	FileTransferWindow *App;
	int Cur;
	int Max;
	int Last;
	int PartialTime;
	int Bytes[MAX_SAMPLE];

	GSurface *pMemDC;

	void Push()
	{
		for (int i=MAX_SAMPLE-1; i>0; i--)
		{
			Bytes[i] = Bytes[i-1];
		}
		Bytes[0] = 0;
	}

	void Clear()
	{
		for (int i=0; i<MAX_SAMPLE; i++)
		{
			Bytes[i] = 0;
		}
	}

public:
	GPaneHistory(FileTransferWindow *app);
	~GPaneHistory();

	void OnPaint(GSurface *pDC);
	void Value(int i);
};

GPaneHistory::GPaneHistory(FileTransferWindow *app)
{
	Width = HISTORY_TEXT_WIDTH + MAX_SAMPLE + 2;
	App = app;
	pMemDC = 0;
	Cur = 0;
	Max = 0;
	Last = 0;
	Clear();
}

GPaneHistory::~GPaneHistory()
{
	DeleteObj(pMemDC);
}

void GPaneHistory::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);

	LgiThinBorder(pDC, r, SUNKEN);
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(r.x1, r.y1, r.x1+HISTORY_TEXT_WIDTH, r.y2);

	char Str[256];
	sprintf(Str, "%.1f K/s", Max / 1024.0);
	SysFont->Colour(0, LC_MED);
	SysFont->Text(pDC, r.x1+2, r.y1, Str);

	r.x1 += HISTORY_TEXT_WIDTH+1;

	if (NOT pMemDC OR
		(r.X() != pMemDC->X()))
	{
		DeleteObj(pMemDC);
		pMemDC = NEW(GMemDC);
		if (pMemDC)
		{
			if (pMemDC->Create(r.X(), r.Y(), GdcD->GetBits()))
			{
				pMemDC->Colour(0);
				pMemDC->Rectangle();
			}
		}
	}

	if (pMemDC)
	{
		pDC->Blt(r.x1, r.y1, pMemDC);
	}
	else
	{
		pDC->Colour(LC_LOW, 24);
		pDC->Rectangle(&r);
	}
}

void GPaneHistory::Value(int i)
{
	if (i <= 0)
	{
		Max = 0;
		Last = LgiCurrentTime();
		Cur = -i;
		PartialTime = 0;
		Clear();
	}
	else
	{
		int Now = LgiCurrentTime();
		int MilliSecs = Now-Last;
		Last = Now;
		double Seconds = (double)MilliSecs / HISTORY_DIV;
		if (Seconds > 0.0)
		{
			int NewData = i - Cur;
			Cur = i;
			int Time = MilliSecs;
			int Data = NewData;

			// do any fractional second
			if (PartialTime > 0)
			{
				int ThisTime = HISTORY_DIV - PartialTime;
				ThisTime = min(ThisTime, MilliSecs);
				int ThisData = (ThisTime * NewData) / MilliSecs;
				Bytes[0] += ThisData;
				Max = max(Bytes[0], Max);
				PartialTime += ThisTime;

				Data -= ThisData;
				Time -= ThisTime;

				if (PartialTime >= HISTORY_DIV)
				{
					Push();
					PartialTime = 0;
				}

			}

			if (Time > 0) // there is time left to process
			{
				// do any whole seconds
				while (Time > HISTORY_DIV)
				{
					Bytes[0] = (HISTORY_DIV * NewData) / MilliSecs;
					Max = max(Bytes[0], Max);

					Data -= Bytes[0];
					Time -= HISTORY_DIV;

					Push();
				}

				// remember any remainder for next time
				PartialTime = Time;
				Bytes[0] = Data;
			}

			// draw
			if (pMemDC AND Max > 0)
			{
				pMemDC->Colour(0);
				pMemDC->Rectangle();
				pMemDC->Colour(Rgb24(0, 255, 0), 24);

				int m=0;
				for (int x=0; x<pMemDC->X() AND x<MAX_SAMPLE; x++)
				{
					m = max(Bytes[x], m);
					int Sample = (Bytes[x] * pMemDC->Y()) / Max;
					pMemDC->Line(x, pMemDC->Y()-Sample, x, pMemDC->Y());
				}

				if (m < Max)
				{
					Max = m;
				}
			}
		}
	}

	GView::Invalidate();
}

//////////////////////////////////////////////////////////////////////////////////////
FileTransferProgress::FileTransferProgress(	FileTransferWindow *App,
											GStatusBar *Status,
											GStatusPane *Info[_STATUS_MAX],
											bool Limit)
{
	StatusInfo = Info;
	StartTime = StartPos = 0;
	ProgressPane = 0;

	// Download throttle
	StatusInfo[_STATUS_THROTTLE]	= (Limit) ? NEW(GPaneThrottle(App)) : 0;
	if (StatusInfo[_STATUS_THROTTLE]) Status->AppendPane(StatusInfo[_STATUS_THROTTLE]);
	
	// Historical graph
	StatusInfo[_STATUS_HISTORY]		= NEW(GPaneHistory(App));
	if (StatusInfo[_STATUS_HISTORY]) Status->AppendPane(StatusInfo[_STATUS_HISTORY]);

	// current download position
	StatusInfo[_STATUS_POSITION]	= Status->AppendPane("", 90);
	if (StatusInfo[_STATUS_POSITION]) StatusInfo[_STATUS_POSITION]->Sunken(TRUE);
	
	// progress meter
	StatusInfo[_STATUS_PROGRESS]	= ProgressPane = NEW(GProgressStatusPane);
	if (StatusInfo[_STATUS_PROGRESS]) Status->AppendPane(ProgressPane);
	
	// download rate
	StatusInfo[_STATUS_RATE]		= Status->AppendPane("", 60);
	if (StatusInfo[_STATUS_RATE]) StatusInfo[_STATUS_RATE]->Sunken(TRUE);
	
	// estimated time left
	StatusInfo[_STATUS_TIME_LEFT]	= Status->AppendPane("", 60);
	if (StatusInfo[_STATUS_TIME_LEFT]) StatusInfo[_STATUS_TIME_LEFT]->Sunken(TRUE);
}

void FileTransferProgress::SetLimits(int l, int h)
{
	Progress::SetLimits(l, h);
	if (ProgressPane)
	{
		ProgressPane->SetLimits(l, h);
	}
}

void FileTransferProgress::Value(int v)
{
	bool Reset = v < Val;
	bool Start = Val == 0;
	if (Start)
	{
		StartTime = LgiCurrentTime();
	}

	// Tell everyone about the new value
	Progress::Value(v);
	if (ProgressPane) ProgressPane->Value(v);
	if (StatusInfo[_STATUS_HISTORY]) StatusInfo[_STATUS_HISTORY]->Value((Start) ? -v : v);
	
	if (Reset)
	{
		StatusInfo[_STATUS_POSITION]->Name("");
		StatusInfo[_STATUS_TIME_LEFT]->Name("");
		StatusInfo[_STATUS_RATE]->Name("");
	}
	else if (High > 0)
	{
		if (StatusInfo[_STATUS_POSITION])
		{
			char Str[256];
			sprintf(Str, "%i of %i K", Val>>10, High>>10);
			StatusInfo[_STATUS_POSITION]->Name(Str);
		}

		double Rate = 0.0;
		int Now = LgiCurrentTime();
		double Seconds = ((double)Now-(double)StartTime)/CLOCKS_PER_SEC;
		if (StatusInfo[_STATUS_RATE] AND Seconds > 0.0)
		{
			char Str[256];
			Rate = ((double)(Val-StartPos))/Seconds;
			sprintf(Str, "%.2f K/s", Rate/1024.0);
			StatusInfo[_STATUS_RATE]->Name(Str);
		}

		if (StatusInfo[_STATUS_TIME_LEFT] AND Rate > 0.0)
		{
			char Str[256];
			double Time = ((double) (High-StartPos) / Rate) - Seconds + 0.5;
			sprintf(Str, "%i:%02.2i:%02.2i", (int)(Time/3600), ((int)(Time/60))%60, ((int)Time)%60);
			StatusInfo[_STATUS_TIME_LEFT]->Name(Str);
		}
	}
	else
	{
		if (StatusInfo[_STATUS_POSITION])
		{
			char Str[256];
			sprintf(Str, "%i K", Val>>10);
			StatusInfo[_STATUS_POSITION]->Name(Str);
		}
	}
}

void FileTransferProgress::SetParameter(int Which, int What)
{
	switch (Which)
	{
		case PARM_START_VALUE:
		{
			Progress::Value(StartPos = What);
			if (StatusInfo[_STATUS_HISTORY]) StatusInfo[_STATUS_HISTORY]->Value(-Val);
		}
	}
}
