/*hdr
**      FILE:           Mail.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           28/5/98
**      DESCRIPTION:    Mail app
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <stdlib.h>
#include <sys/timeb.h>

#include "Lgi.h"
#include "Mail.h"
#include "GToken.h"
#include "Base64.h"

//////////////////////////////////////////////////////////////////////////////////////////////////
#define LIB_NAME "fReT-Mail v2.00"

//////////////////////////////////////////////////////////////////////////////////////////////////
// return true if there are any characters with the 0x80 bit set
bool Is8Bit(char *Text)
{
	for (char *c = Text; *c; c++)
	{
		if (*c & 0x80)
		{
			return true;
		}
	}

	return false;
}

// returns the maximum length of the lines contained in the string
int MaxLineLen(char *Text)
{
	int Max = 0;
	int i = 0;

	for (char *c = Text; *c; c++)
	{
		if (*c == '\r')
		{
			// return
		}
		else if (*c == '\n')
		{
			// eol
			Max = max(i, Max);
			i = 0;
		}
		else
		{
			// normal char
			i++;
		}
	}

	return Max;
}

bool IsDotLined(char *Text)
{
	if (Text)
	{
		for (char *l = Text; l AND *l; )
		{
			if (l[0] == '.')
			{
				if (l[1] == '\n' OR l[1] == 0)
				{
					return true;
				}
			}

			l = strchr(l, '\n');
			if (l) l++;
		}
	}

	return false;
}

char ConvHexToBin(char c)
{
	if (c >= '0' AND c <= '9')
		return c - '0';
	if (c >= 'a' AND c <= 'f')
		return c + 10 - 'a';
	if (c >= 'A' AND c <= 'F')
		return c + 10 - 'A';
	return 0;
}

// Is s a valid non-whitespace string?
bool ValidNonWSStr(char *s)
{
	if (s AND *s)
	{
		while (*s AND strchr(" \r\t\n", *s))
		{
			s++;
		}

		if (*s)
		{
			return TRUE;
		}
	}

	return false;
}

void TokeniseStrList(char *Str, List<char> &Output, char *Delim)
{
	if (Str AND Delim)
	{
		char *s = Str;
		while (*s)
		{
			char *e = s;
			for (; *e; e++)
			{
				if (strchr("\'\"", *e))
				{
					// handle string constant
					char EndChar = *e++;
					while (*e AND *e != EndChar) e++;
				}

				if (strchr(Delim, *e))
				{
					break;
				}
			}

			int Len = (int)e - (int)s;
			if (Len > 0)
			{
				char *Temp = NEW(char[Len+1]);
				if (Temp)
				{
					memcpy(Temp, s, Len);
					Temp[Len] = 0;
					Output.Insert(Temp);
				}
			}

			s = e;
			for (; *s AND strchr(Delim, *s); s++);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
char *DecodeBase64Str(char *Str, int Len)
{
	if (Str)
	{
		int B64Len = (Len < 0) ? strlen(Str) : Len;
		int BinLen = BufferLen_64ToBin(B64Len);
		char *s = NEW(char[BinLen+1]);
		if (s)
		{
			ConvertBase64ToBinary((uchar*)s, BinLen, Str, B64Len);
			s[BinLen] = 0;
			DeleteArray(Str);
			Str = s;
		}
	}
	return Str;
}

char *DecodeQuotedPrintableStr(char *Str, int Len)
{
	if (Str)
	{
		if (Len < 0) Len = strlen(Str);
		uchar *s = NEW(uchar[Len+1]);
		if (s)
		{
			char *Out = (char*) s;
			char *Text = Str;

			for (int i=0; i<Len; i++)
			{
				if (Text[i] == '=')
				{
					if (Text[i+1] == '\r' OR Text[i+1] == '\n')
					{
						if (Text[i+1] == '\r') i++;
						if (Text[i+1] == '\n') i++;
					}
					else
					{
						i++;
						char c = ConvHexToBin(Text[i]) << 4;
						i++;
						c |= ConvHexToBin(Text[i]) & 0xF;

						*Out++ = c;
					}
				}
				else
				{
					*Out++ = Text[i];
				}
			}

			*Out++ = 0;
			DeleteArray(Str);
			Str = (char*) s;
		}
	}
	return Str;
}

#ifndef ToUpper
#define ToUpper(c)						( ((c)>='a'AND(c)<='z')?(c)-'a'+'A':(c) )
#endif

char *DecodeRfc2047(char *Str)
{
	if (Str AND strstr(Str, "=?"))
	{
		BytePipe Temp;
		char *s = Str;
		while (*s)
		{
			// is there a word remaining
			bool Encoded = false;
			char *p = strstr(s, "=?");
			char *First = 0;
			char *Second = 0;
			char *End = 0;
			if (p)
			{
				First = strchr(p+2, '?');
				if (First)
				{
					Second = strchr(First+1, '?');
					if (Second)
					{
						End = strstr(Second+1, "?=");
						Encoded = End != 0;
					}
				}
			}

			if (Encoded)
			{
				int Type = CONTENT_NONE;
				bool StripUnderscores = false;
				if (ToUpper(First[1]) == 'B')
				{
					// Base64 encoding
					Type = CONTENT_BASE64;
				}
				else if (ToUpper(First[1]) == 'Q')
				{
					// Quoted printable
					Type = CONTENT_QUOTED_PRINTABLE;
					StripUnderscores = true;
				}

				if (Type != CONTENT_NONE)
				{
					char *Block = NewStr(Second+1, (int)End-(int)Second-1);
					if (Block)
					{
						switch (Type)
						{
							case CONTENT_BASE64:
								Block = DecodeBase64Str(Block);
								break;
							case CONTENT_QUOTED_PRINTABLE:
								Block = DecodeQuotedPrintableStr(Block);
								break;
						}

						int Len = strlen(Block);
						if (StripUnderscores)
						{
							for (int i=0; i<Len; i++)
							{
								if (Block[i] == '_') Block[i] = ' ';
							}
						}

						Temp.Push((uchar*)s, (int)p-(int)s);
						Temp.Push((uchar*)Block, Len);

						DeleteArray(Block);
					}
					s = End + 2;
				}
			}

			if (NOT Encoded)
			{
				int Len = strlen(s);
				Temp.Push((uchar*) s, Len);
				s += Len;
			}
		}

		Temp.Push((uchar*) "", 1);
		int Len = Temp.Sizeof();
		char *r = NEW(char[Len+1]);
		if (r)
		{
			Temp.Pop((uchar*) r, Len);
			r[Len] = 0;
		}

		DeleteArray(Str);
		return r;
	}

	return Str;
}

char *EncodeRfc2047(char *Str)
{
	if (Str AND Is8Bit(Str))
	{
		StringPipe p;
		char *Ws = " \r\t\n";

		char *Start = Str;
		char *End = Str;
		while (*End)
		{
			char *WordStart = Start;

			// search through till the next 8 bit char
			for (; *End AND (*End&0x80)==0; End++)
			{
				if (strchr(Ws, *End))
				{
					// Is whitespace
					WordStart = 0;
				}
				else
				{
					// Not whitespace
					if (NOT WordStart)
					{
						WordStart = End;
					}
				}
			}

			if (NOT *End OR NOT WordStart)
			{
				p.Push(Start);
			}
			else
			{
				// push any chars before the 8 bit one
				p.BytePipe::Push((uchar*) Start, (int)WordStart-(int)Start);

				// find the end of the word
				End = WordStart;
				char *WordEnd = End;
				for (; *WordEnd AND NOT strchr(Ws, *WordEnd); WordEnd++);

				// encode the word
				char s[256], *o = s;;
				int Len = (int)WordEnd-(int)End;
				for (int i=0; i<Len; i++)
				{
					if (End[i] & 0x80)
					{
						sprintf(o, "=%02.2X", (uchar)End[i]);
						o += 3;
					}
					else
					{
						*o++ = End[i];
					}
				}
				*o++ = 0;

				p.Push("=?us-ascii?Q?");
				p.Push(s);
				p.Push("?=");

				End = WordEnd;
			}

			Start = End;
		}

		DeleteArray(Str);
		Str = p.NewStr();
	}

	return Str;
}

//////////////////////////////////////////////////////////////////////////////
void DecodeAddrName(char *Start, char *&Name, char *&Addr, char *DefaultDomain)
{
	/* Testing code
	char *Input[] =
	{
		"\"'Matthew Allen'\" <fret@ozemail.com.au>",
		"Matthew Allen",
		"fret@ozemail.com.au",
		"\"<Matthew Allen>\" <fret@ozemail.com.au>",
		"<Matthew Allen> (fret@ozemail.com.au)",
		"Matthew Allen <fret@ozemail.com.au>",
		"\"Matthew, Allen\" (fret@ozemail.com.au)",
		0
	};

	char *Name = 0;
	char *Addr = 0;

	for (char **i = Input; *i; i++)
	{
		DecodeAddrName(*i, Name, Addr, "ozemail.com.au");
		int y=0;
	}
	*/

	if (Start)
	{
		Name = Addr = 0;

		int NextStr = 0;
		char *Str[2] = {0, 0};

		for (char *c = Start; *c AND NextStr < 2; )
		{
			if (strchr(" \t\r\n", *c))
			{
				// skip whitespace
			}
			else if (strchr("\"'", *c))
			{
				// string delim
				char End = *c;
				char *Start = c;
				for (c++; *c AND *c != End; c++);
				Str[NextStr++] = NewStr(Start, (int)c-(int)Start);
			}
			else if (strchr("<(", *c))
			{
				// brackets
				char End = (*c == '<') ? '>' : ')';
				char *Start = ++c;
				for (; *c AND *c != End; c++);
				Str[NextStr++] = NewStr(Start, (int)c-(int)Start);
			}
			else
			{
				// Some string
				char *Start = c;
				for (; *c AND NOT strchr("<(\"'", *c); c++);
				Str[NextStr++] = NewStr(Start, (int)c-(int)Start);
			}

			if (*c)
			{
				c++;
			}
			else break;
		}

		char *Delim = " \r\n\t<>()\"'";
		if (Str[1])
		{
			if (strchr(Str[1], '@'))
			{
				Addr = TrimStr(Str[1], Delim);
			}
			else
			{
				Name = TrimStr(Str[1], Delim);
			}
		}

		if (Str[0])
		{
			if (Addr OR NOT strchr(Str[0], '@')) // already filled in
			{
				DeleteArray(Name);
				Name = TrimStr(Str[0], Delim);
			}
			else if (NOT Addr)
			{
				Addr = TrimStr(Str[0], Delim);
			}
		}

		DeleteArray(Str[0]);
		DeleteArray(Str[1]);
	}
}

void StrCopyToEOL(char *d, char *s)
{
	if (d AND s)
	{
		while (*s AND *s != '\r' AND *s != '\n')
		{
			*d++ = *s++;
		}

		*d = 0;
	}
}

/*
		int Found = 0;
		int Ints[8];
		char Num[32];
		char *In = Buffer;
		char *Out = 0;

		while (*In)
		{
			if (Out)
			{
				if (IsDigit(*In))
				{
					*Out++ = *In;
				}
				else
				{
					*Out++ = 0;
					Ints[Found++] = atoi(Num);
					Out = NULL;
				}
			}
			else
			{
				if (IsDigit(*In))
				{
					Out = Num;
					*Out++ = *In;
				}
			}

			In++;
		}

		int Size = 0;
		for (int i=0; i<Found; i++)
		{
			if (Ints[i] != Message)
			{
				// probably the size
				Size = Ints[i];
			}
		}

*/
//////////////////////////////////////////////////////////////////////////////////////////////////
FileDescriptor::FileDescriptor()
{
	Embeded = 0;
	Offset = 0;
	Size = 0;
	Data = 0;
	MimeType = 0;
}

FileDescriptor::FileDescriptor(GFile *embed, int offset, int size, char *name)
{
	Embeded = embed;
	Offset = offset;
	Size = size;
	Data = 0;
	MimeType = 0;

	if (name)
	{
		Name(name);
	}
}

FileDescriptor::FileDescriptor(char *name)
{
	Embeded = 0;
	Offset = 0;
	Size = 0;
	Data = 0;
	MimeType = 0;

	if (name)
	{
		Name(name);
		if (File.Open(name, O_READ))
		{
			Size = File.GetSize();
			File.Close();
		}
	}
}

FileDescriptor::~FileDescriptor()
{
	DeleteArray(MimeType);
	DeleteArray(Data);
}

GFile *FileDescriptor::GotoObject()
{
	if (Embeded)
	{
		Embeded->Seek(Offset, SEEK_SET);
		return Embeded;
	}
	else if (Name() AND File.Open(Name(), O_READ))
	{
		return &File;
	}

	return NULL;
}

int FileDescriptor::Sizeof()
{
	return Size;
}

uchar *FileDescriptor::GetData()
{
	return Data;
}

bool FileDescriptor::Decode(	char *ContentType,
				char *ContentTransferEncoding,
				char *MimeData,
				int MimeDataLen)
{
	bool Status = false;
	int Content = CONTENT_NONE;

	if (ContentType AND ContentTransferEncoding)
	{
		// Content-Type: application/octet-stream; name="Scribe.opt"
		Content = CONTENT_OCTET_STREAM;

		if (strnistr(ContentTransferEncoding, "base64", 1000))
		{
			Content = CONTENT_BASE64;
		}

		if (strnistr(ContentTransferEncoding, "quoted-printable", 1000))
		{
			Content = CONTENT_QUOTED_PRINTABLE;
		}

		if (Content != CONTENT_NONE);
		{
			char *NameKey = "name";
			char *n = strnistr(ContentType, NameKey, 1000);
			if (n)
			{
				char *Equal = strchr(n, '=');
				if (Equal)
				{
					Equal++;
					while (*Equal AND *Equal == '\"')
					{
						Equal++;
					}

					char *End = strchr(Equal, '\"');
					if (End)
					{
						*End = 0;
					}

					Name(Equal);
					Status = TRUE;
				}
			}
		}
	}

	if (Status AND MimeData AND MimeDataLen > 0 AND Content != CONTENT_NONE)
	{
		Status = false;
		char *Base64 = NEW(char[MimeDataLen]);

		switch (Content)
		{
			case CONTENT_OCTET_STREAM:
			{
				Size = 0;
				DeleteObj(Data);
				Data = NEW(uchar[MimeDataLen]);
				if (Data)
				{
					Size = MimeDataLen;
					memcpy(Data, MimeData, Size);
					Status = TRUE;
				}
				break;
			}
			case CONTENT_QUOTED_PRINTABLE:
			{
				Size = 0;
				DeleteObj(Data);
				Data = NEW(uchar[MimeDataLen+1]);
				if (Data)
				{
					char *Out = (char*) Data;
					for (int i=0; i<MimeDataLen; i++)
					{
						if (MimeData[i] == '=')
						{
							if (MimeData[i+1] == '\r' OR MimeData[i+1] == '\n')
							{
								if (MimeData[i+1] == '\r') i++;
								if (MimeData[i+1] == '\n') i++;
							}
							else
							{
								i++;
								char c = ConvHexToBin(MimeData[i]) << 4;
								i++;
								c |= ConvHexToBin(MimeData[i]) & 0xF;

								Size++;
								*Out++ = c;
							}
						}
						else
						{
							Size++;
							*Out++ = MimeData[i];
						}
					}

					*Out++ = 0;
					Status = TRUE;
				}
				break;
			}
			case CONTENT_BASE64:
			{
				int InCount = 0;
				int OutCount = 0;
				char *In = MimeData;
				char *Out = Base64;
				while (InCount < MimeDataLen)
				{
					if (*In != '\r' AND *In != '\n')
					{
						*Out++ = *In;
						OutCount++;
					}
					InCount++;
					In++;
				}

				Size = BufferLen_64ToBin(OutCount);
				Data = NEW(uchar[Size]);
				if (Data)
				{
					int Converted = ConvertBase64ToBinary(Data, Size, Base64, OutCount);
					Status = Converted <= Size AND Converted >= Size - 3;
					if (Status)
					{
						Size = Converted;
					}
					else
					{
						DeleteArray(Data);
						Size = 0;
					}
				}
				break;
			}
		}
	}
	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
AddressDescriptor::AddressDescriptor(AddressDescriptor *Copy)
{
	Status = Copy ? Copy->Status : false;
	CC = Copy ? Copy->CC : false;
	Addr = Copy ? NewStr(Copy->Addr) : 0;
	Name = Copy ? NewStr(Copy->Name) : 0;
}

AddressDescriptor::~AddressDescriptor()
{
	DeleteArray(Name);
	DeleteArray(Addr);
}

void AddressDescriptor::Print(char *Str)
{
	if (Str)
	{
		if (Addr AND Name)
		{
			sprintf(Str, "%s (%s)", Addr, Name);
		}
		else if (Addr)
		{
			strcpy(Str, Addr);
		}
		else if (Name)
		{
			sprintf(Str, "(%s)", Name);
		}
	}
}

int AddressDescriptor::Sizeof()
{
	return	SizeofStr(Name) +
		SizeofStr(Addr);
}

bool AddressDescriptor::Serialize(GFile &f, bool Write)
{
	bool Status = TRUE;
	if (Write)
	{
		WriteStr(f, Name);
		WriteStr(f, Addr);
	}
	else
	{
		DeleteArray(Name);
		Name = ReadStr(f);
		DeleteArray(Addr);
		Addr = ReadStr(f);
	}
	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
MailProtocol::MailProtocol()
{
	Socket = 0;
	ServerMsg = 0;
	Buffer[0] = 0;
	Info = 0;
	Lock = 0;
	Log = 0;
}

MailProtocol::~MailProtocol()
{
	if (_Lock())
	{
		DeleteObj(Info);
		_Unlock();
	}

	DeleteArray(ServerMsg);
}

bool MailProtocol::_Lock()
{
	if (Lock)
	{
		return Lock->Lock();
	}

	return true;
}

void MailProtocol::_Unlock()
{
	if (Lock)
	{
		Lock->Unlock();
	}
}

bool MailProtocol::_Cancel()
{
	bool Status = false;

	if (_Lock())
	{
		if (Info)
		{
			Status = Info->Cancel();
		}
		_Unlock();
	}

	return Status;
}

void MailProtocol::LogStr(char *Str)
{
	// we need to allocate our own array here because some OS's can't write
	// to strings in the data segment. so we copy it into dynamically allocated
	// memory to do our logging
	if (Log)
	{
		char *MyBuf = NewStr(Str);
		if (MyBuf)
		{
			char *Cr = strchr(MyBuf, '\r');
			if (Cr) *Cr = 0;

			Log->Insert(NEW(LogEntry(MyBuf, MAIL_SEND_LINE, MAIL_SEND_COLOUR)));
			DeleteArray(MyBuf);
		}
	}
}

bool MailProtocol::Read()
{
	bool Status = false;
	if (Socket)
	{
		Status = Socket->Read(Buffer, sizeof(Buffer), 0) > 0;
	}
	return Status;
}

bool MailProtocol::Write(char *Buf, bool LogWrite)
{
	bool Status = false;
	if (Socket)
	{
		char *p = (Buf) ? Buf : Buffer;
		Status = Socket->Write(p, strlen(p), 0);
		
		if (LogWrite)
		{
			LogStr(p);
		}
	}
	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
#define	VERIFY_RET_VAL(Arg)					\
{											\
	if (NOT Arg)							\
	{										\
		Socket = 0;							\
		return false;						\
	}										\
}

MailSmtp::MailSmtp()
{
	ProgramName = "MailSmtp Object";
}

MailSmtp::~MailSmtp()
{
}

bool MailSmtp::Open(GSocket *S, char *RemoteHost, char *UserName, char *Password, int Port)
{
	char Str[256] = "";
	bool Status = false;

	if (RemoteHost)
	{
		strcpy(Str, RemoteHost);
		char *Colon = strchr(Str, ':');
		if (Colon)
		{
			*Colon = 0;
			Colon++;
			Port = atoi(Colon);
		}

		char *Server = TrimStr(Str);
		if (Server)
		{
			Socket = S;
			if (Socket->Open(Server, Port))
			{
				// receive signon message
				VERIFY_RET_VAL(ReadReply("220"));

				if (Password)
				{
					char Str[256];

					// Rfc 2554 ESMTP authentication
					sprintf(Buffer, "EHLO %s\r\n", (ValidNonWSStr(UserName)) ? UserName : "default");
					VERIFY_RET_VAL(Write(0, true));
					VERIFY_RET_VAL(ReadReply("250", Str));

					// walk through AUTH types looking for ones we recognise
					bool Authed = false;
					char AuthTypes[256] = "";
					for (char *l=strtok(Str, "\n"); l AND NOT Authed; l=strtok(NULL, "\n"))
					{
						char *AuthStr = strstr(l, "AUTH");
						if (AuthStr)
						{
							// is AUTH type
							if (strstr(l, "LOGIN"))
							{
								// LOGIN auth type
								char User64[128];
								char Pass64[128];

								ZeroObj(User64);
								ZeroObj(Pass64);

								ConvertBinaryToBase64(User64, sizeof(User64), (uchar*)UserName, strlen(UserName));
								ConvertBinaryToBase64(Pass64, sizeof(Pass64), (uchar*)Password, strlen(Password));

								sprintf(Buffer, "AUTH LOGIN %s\r\n", User64);
								VERIFY_RET_VAL(Write(0, true));
								VERIFY_RET_VAL(ReadReply("334"));

								sprintf(Buffer, "%s\r\n", Pass64);
								VERIFY_RET_VAL(Write(0, true));
								VERIFY_RET_VAL(ReadReply("235"));

								Authed = true;
							}
							else
							{
								// unknown AUTH type
								if (AuthTypes[0]) strcat(AuthTypes, ", ");
								strcat(AuthTypes, AuthStr+4);
							}
						}
					}

					if (NOT Authed)
					{
						sprintf(Buffer, "Only the following unsupported methods are available for authentication:\n%s", AuthTypes);
						DeleteArray(ServerMsg);
						ServerMsg = NewStr(Buffer);
					}

					Status = Authed;
				}
				else
				{
					// Normal SMTP login
					// send HELO message
					sprintf(Buffer, "HELO %s\r\n", (ValidNonWSStr(UserName)) ? UserName : "default");
					VERIFY_RET_VAL(Write(0, true));
					VERIFY_RET_VAL(ReadReply("250"));

					Status = true;
				}
			}

			DeleteArray(Server);
		}
	}

	return Status;
}

bool MailSmtp::WriteText(char *Str)
{
	// we have to convert all strings to CRLF in here
	bool Status = false;
	if (Str)
	{
		BytePipe Temp;

		char *Start = Str;
		while (*Str)
		{
			if (*Str == '\n')
			{
				// send a string
				int Size = (int)Str-(int)Start;
				if (Str[-1] == '\r') Size--;
				Temp.Push((uchar*) Start, Size);
				Temp.Push((uchar*) "\r\n", 2);
				Start = Str + 1;
			}

			Str++;
		}

		// send the final string
		int Size = (int)Str-(int)Start;
		if (Str[-1] == '\r') Size--;
		Temp.Push((uchar*) Start, Size);

		Size = Temp.Sizeof();
		char *Data = NEW(char[Size]);
		if (Data)
		{
			Temp.Pop((uchar*) Data, Size);
			Status = Socket->Write(Data, Size, 0);
			DeleteArray(Data);
		}
	}

	return Status;
}

char *CreateAddressTag(List<AddressDescriptor> &l, int Type)
{
	char *Result = 0;
	List<AddressDescriptor> Addr;
	AddressDescriptor *a;
	for (a = l.First(); a; a = l.Next())
	{
		if (a->CC == Type)
		{
			Addr.Insert(a);
		}
	}
	
	if (Addr.GetItems() > 0)
	{
		StringPipe StrBuf;
		StrBuf.Push((Type == 0) ? (char*)"To: " : (char*)"Cc: ");

		for (a = Addr.First(); a; )
		{
			AddressDescriptor *NextA = Addr.Next();
			char Buffer[256] = "";

			if (a->Name)
			{
				sprintf(Buffer, "\"%s\" <%s>", a->Name, a->Addr);
			}
			else if (a->Addr)
			{
				sprintf(Buffer, "<%s>", a->Addr);
			}

			if (NextA) strcat(Buffer, ",\r\n\t");
			StrBuf.Push(Buffer);

			a = NextA;
		}

		StrBuf.Push("\r\n");
		Result = StrBuf.NewStr();
	}

	return Result;
}

bool MailSmtp::Send(MailMessage *Msg, bool Mime)
{
	bool Status = false;

	if (Socket AND Msg)
	{
		// send MAIL message
		if (Msg->From.Addr)
		{
			sprintf(Buffer, "MAIL FROM: <%s>\r\n", Msg->From.Addr);
		}
		else
		{
			ServerMsg = NewStr("No 'from' address in email.");
			return false;
		}

		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadReply("250"));

		// send RCPT message
		for (AddressDescriptor *a = Msg->To.First(); a; a = Msg->To.Next())
		{
			sprintf(Buffer, "RCPT TO: <%s>\r\n", a->Addr);
			VERIFY_RET_VAL(Write(0, true));
			a->Status = ReadReply("25");
		}

		// send DATA message
		sprintf(Buffer, "DATA\r\n");
		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadReply("354"));

		// track progress
		if (_Lock())
		{
			Info = NEW(Progress);
			if (Info)
			{
				Info->SetDescription("Encoding email...");
				Socket->SetParameter(PARAMETER_PROGRESS, (int) Info);
			}
			_Unlock();
		}

		// encode message for transport
		StringPipe Buf;
		if (Msg->Encode(Buf, this))
		{
			if (_Lock())
			{
				if (Info)
				{
					Info->SetDescription("Allocating output buffer...");
				}
				_Unlock();
			}

			char *b = Buf.NewStr();
			if (b)
			{
				Buf.Empty(); // we don't need 2 copies of the data in memory
				int Length = strlen(b); // get the length of the encoded email
				
				// setup a gui progress meter to send the email
				if (_Lock())
				{
					if (Info)
					{
						Info->SetLimits(0, Length-1);
						Info->SetScale(1.0 / 1024.0);
						Info->SetType("K");
						Info->SetDescription("Sending email...");
						Info->Value(0);
					}

					_Unlock();
				}

				// write the data to the socket
				int Written = 0;
				while (Written < Length)
				{
					int Len = min(8 << 10, Length - Written);
					int Result = Socket->Write(b+Written, Len, 0);
					if (Result < 0)
					{
						// error
						break;
					}

					if (_Cancel())
					{
						// cancelled
						Socket->Cancel(true);
						break;
					}

					if (Result > 0)
					{
						// another bit written
						Written += Result;
					}

					LgiSleep(10);
				}

				Status = Length == Written;

				// delete buffer for the encoded message
				DeleteObj(b);
			}
		}

		// close the progress meter
		if (_Lock())
		{
			if (Info)
			{
				Socket->SetParameter(PARAMETER_PROGRESS, 0);
				DeleteObj(Info);
			}
			_Unlock();
		}

		if (Status)
		{
			// send message terminator and receive reply
			strcpy(Buffer, "\r\n.\r\n");
			VERIFY_RET_VAL(Write(0, true));
			VERIFY_RET_VAL(ReadReply("250"));
		}
		// else just close the connection on them
		// so nothing gets sent
	}

	return Status;
}

bool MailSmtp::Close()
{
	if (Socket)
	{
		// send QUIT message
		sprintf(Buffer, "QUIT\r\n");
		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadReply("221"));

		Socket->Close();
		Socket = NULL;

		return TRUE;
	}

	return false;
}

bool MailSmtp::ReadReply(char *Str, char *Line)
{
	bool Status = false;
	if (Socket AND Str)
	{
		int Pos = 0;
		char *Start = Buffer;

		ZeroObj(Buffer);
		if (Line) Line[0] = 0;

		while (Pos<sizeof(Buffer))
		{
			int Len = Socket->Read(Buffer+Pos, sizeof(Buffer)-Pos, 0);
			if (Len > 0)
			{
				char *Eol = strstr(Start, "\r\n");
				while (Eol)
				{
					// wipe EOL chars
					*Eol++ = 0;
					*Eol++ = 0;

					// log it
					LogEntry *Entry = 0;
					if (Log)
					{
						Log->Insert(Entry = NEW(LogEntry(Start, MAIL_RECEIVE_LINE, MAIL_RECEIVE_COLOUR)));
					}

					// process
					if (Line)
					{
						if (Line[0]) strcat(Line, "\n");
						strcat(Line, Start);
					}

					if (Start[3] == ' ')
					{
						// end of response
						if (NOT strncmp(Start, Str, strlen(Str)))
						{
							Status = TRUE;
						}

						if (Entry)
						{
							// get code
							Start[3] = 0;
							if (atoi(Start) >= 400)
							{
								// error
								Entry->c = MAIL_ERROR_COLOUR;
								Entry->Data = MAIL_ERROR_LINE;
							}
							Start[3] = ' ';
						}

						// exit loop
						Pos = sizeof(Buffer);
						break;
					}
					else
					{
						// more lines follow
						Start = Eol;
						Eol = strstr(Start, "\r\n");
					}
				}

				Pos += Len;
   			}
		}

		if (NOT Status)
		{
			DeleteArray(ServerMsg);
			ServerMsg = NewStr(Buffer);
		}
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
MailPop3::MailPop3()
{
	End = "\r\n.\r\n";
	Marker = End;
	Messages = 0;
}

MailPop3::~MailPop3()
{
}

int MailPop3::GetInt()
{
	char Buf[32];
	char *Start = strchr(Buffer, ' ');
	if (Start)
	{
		Start++;
		char *End = strchr(Start, ' ');
		if (End)
		{
			int Len = (int) (End - Start);
			memcpy(Buf, Start, Len);
			Buf[Len] = 0;
			return atoi(Buf);
		}
	}
	return 0;
}

bool MailPop3::ReadReply()
{
	bool Status = false;
	if (Socket)
	{
		int StartRead = LgiCurrentTime();
		// int Timeout = 15; // seconds

		int Pos = 0;
		ZeroObj(Buffer);
		do
		{
			int Result = Socket->Read(Buffer+Pos, sizeof(Buffer)-Pos, 0);
			if (Result <= 0)
			{
				break;
			}

			Pos += Result;
    	}
		while (	NOT strstr(Buffer, "\r\n") AND
				sizeof(Buffer)-Pos > 0);

		Status = (Buffer[0] == '+') AND strstr(Buffer, "\r\n");

		if (Log)
		{
			char *Cr = strchr(Buffer, '\r');
			if (Cr) *Cr = 0;
			Log->Insert(NEW(LogEntry(Buffer,
									(Status) ? MAIL_RECEIVE_LINE : MAIL_ERROR_LINE,
									(Status) ? MAIL_RECEIVE_COLOUR : MAIL_ERROR_COLOUR)));
			if (Cr) *Cr = '\r';
		}

		if (NOT Status)
		{
			DeleteArray(ServerMsg);
			ServerMsg = NewStr(Buffer);
		}
	}

	return Status;
}

bool MailPop3::Open(GSocket *S, char *RemoteHost, int Port, char *User, char *Password)
{
	bool Status = false;
	bool RemoveMail = false;
	char Str[256] = "";

	if (RemoteHost)
	{
		if (Port < 1) Port = POP3_PORT;

		strcpy(Str, RemoteHost);
		char *Colon = strchr(Str, ':');
		if (Colon)
		{
			*Colon = 0;
			Colon++;
			Port = atoi(Colon);
		}

		if (S AND
			User AND
			Password)
		{
			char *Server = TrimStr(Str);
			if (Server)
			{
				Socket = S;
				if (Socket->Open(Server, Port))
				{
					char Key[256];
					bool GotKey = false;

					// receive signon message
					VERIFY_RET_VAL(ReadReply());

					// get key here
					char *Start = strchr(Buffer, '<');
					if (Start)
					{
						char *End = strchr(Start, '>');
						if (End)
						{
							int Len = ((int) (End - Start)) + 1;
							memcpy(Key, Start, Len);
							Key[Len] = 0;

							GotKey = TRUE;
						}
					}

					// login
					if (false) // GotKey, not implemented
					{
						// using encrypted password
						unsigned char Digest[16];
						char HexDigest[33];

						// append password
						strcat(Key, Password);

						ZeroObj(Digest);
						MDStringToDigest(Digest, Key);
						for (int i = 0; i < 16; i++) sprintf(HexDigest + (i*2), "%02.2x", Digest[i]);
						HexDigest[32] = 0;

						sprintf(Buffer, "APOP %s %s\r\n", User, HexDigest);
						VERIFY_RET_VAL(Write(0, true));
						VERIFY_RET_VAL(ReadReply());
					}
					else
					{
						// have to use non-key method
						sprintf(Buffer, "USER %s\r\n", User);
						VERIFY_RET_VAL(Write(0, true));
						VERIFY_RET_VAL(ReadReply());

						sprintf(Buffer, "PASS %s\r\n", Password);
						VERIFY_RET_VAL(Write(0, false));
						LogStr("PASS *******");

						if (NOT ReadReply())
						{
							if (Socket)
							{
								Socket->Close();
								Socket = NULL;
							}
						}
						else
						{
							Status = TRUE;

							// see how many messages there are
							VERIFY_RET_VAL(Write("STAT\r\n", true));
							VERIFY_RET_VAL(ReadReply());

							Messages = GetInt();
						}
					}
				}

				DeleteArray(Server);
			}
		}
	}

	return Status;
}

bool MailPop3::MailIsEnd(char *Ptr, int Len)
{
	for (char *c = Ptr; Len--; c++)
	{
		if (*c != *Marker)
		{
			Marker = End;
		}

		if (*c == *Marker)
		{
			Marker++;
			if (NOT *Marker)
			{
				return true;
			}
		}
	}

	return false;
}

#define IsDigit(c) ((c) >= '0' AND (c) <= '9')

bool MailPop3::Receive(int Message, MailMessage *Msg, bool Delete, int MaxSize)
{
	bool Status = false;
	Marker = End;

	if (Msg AND
		Socket AND
		Message >= 0 AND
		Message < Messages)
	{
		// get messages
		sprintf(Buffer, "LIST %i\r\n", Message + 1);
		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadReply());

		int Size = 0;
		char *s = strchr(Buffer, ' ');
		if (s)
		{
			s = strchr(s+1, ' ');
			if (s)
			{
				Size = atoi(s);

				if (Size > 0 AND
					(MaxSize < 1 OR Size <= MaxSize) )
				{
					BytePipe Temp;

					sprintf(Buffer, "RETR %i\r\n", Message + 1);
					VERIFY_RET_VAL(Write(0, true));

					int ReadLen = Socket->Read(Buffer, sizeof(Buffer), 0);
					if (ReadLen > 0 AND Buffer[0] == '+')
					{
						char *Eol = strchr(Buffer, '\n');
						if (Eol)
						{
							char *Ptr = Eol + 1;
							ReadLen -= ((int)Ptr-(int)Buffer);

							Temp.Push((uchar*) Ptr, ReadLen);

							if (_Lock())
							{
								Info = NEW(Progress);
								if (Info)
								{
									Info->SetLimits(0, Size);
									Info->SetScale(1.0 / 1024.0);
									Info->SetType("K");
									Info->SetDescription("Receiving email...");
									Info->Value(0);
									Socket->SetParameter(PARAMETER_PROGRESS, (int) Info);
								}
								_Unlock();
							}

							Status = true;
							while (NOT MailIsEnd(Ptr, ReadLen))
							{
								ReadLen = Socket->Read(Buffer, sizeof(Buffer), 0);
								if (ReadLen >= 0)
								{
									Temp.Push((uchar*) Buffer, ReadLen);
									Ptr = Buffer;
								}
								else
								{
									Status = false;
									break;
								}

								if (_Cancel())
								{
									// cancelled
									Socket->Cancel(true);
									Status = false;
									break;
								}
							}

							if (Status)
							{
								Size = max(Temp.Sizeof() - 3, 0);
								Msg->Text = NEW(char[Size+1]);

								if (Msg->Text)
								{
									Temp.Pop((uchar*) Msg->Text, Size);
									Msg->Text[Size] = 0;
								}
							}

							if (_Lock())
							{
								if (Info)
								{
									Socket->SetParameter(PARAMETER_PROGRESS, 0);
									DeleteObj(Info);
								}
								_Unlock();
							}

							if (Delete AND Status)
							{
								MailPop3::Delete(Message);
							}
						}
					}
				}
			}
		}
	}

	return Status;
}

int MailPop3::Sizeof(int Message)
{
	int Size = 0;
	if (Socket)
	{
		sprintf(Buffer, "LIST %i\r\n", Message + 1);
		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadReply());

		char *s = strchr(Buffer, ' ');
		if (s)
		{
			s = strchr(s+1, ' ');
			if (s)
			{
				Size = atoi(s);
			}
		}
	}

	return Size;
}

bool MailPop3::Delete(int Message)
{
	if (Socket)
	{
		sprintf(Buffer, "DELE %i\r\n", Message + 1);
		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadReply());

		return TRUE;
	}

	return false;
}

bool MailPop3::GetUid(int Index, char *Id)
{
	if (Socket AND Id)
	{
		sprintf(Buffer, "UIDL %i\r\n", Index + 1);
		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadReply());

		char *Space = strchr(Buffer, ' ');
		if (Space)
		{
			Space = strchr(Space+1, ' ');
			if (Space)
			{
				for (char *s = Space+1; *s; s++)
				{
					if (*s == '\r' OR *s == '\n')
					{
						*s = 0;
						break;
					}
				}
				
				strcpy(Id, Space+1);
				return TRUE;
			}
		}
	}

	return false;
}

bool MailPop3::GetUidList(List<char> &Id)
{
	bool Status = false;
	if (Socket)
	{
		char *Str = 0;

		sprintf(Buffer, "UIDL\r\n");
		VERIFY_RET_VAL(Write(0, true));
		VERIFY_RET_VAL(ReadMultiLineReply(Str));
		if (Str)
		{
			Status = true;
			TokenParser T(Str, "\r\n");
			for (int i=0; i<T.Tokens; i++)
			{
				char *s = T.Get(i);
				if (s AND *s != '.')
				{
					char *Space = strchr(s, ' ');
					if (Space)
					{
						Id.Insert(NewStr(Space+1));
					}
				}
			}

			DeleteArray(Str);
		}
	}

	return Status;
}

char *MailPop3::GetHeaders(int Message)
{
	char *Header = 0;
	if (Socket)
	{
		sprintf(Buffer, "TOP %i 0\r\n", Message + 1);
		if (Write(0, true))
		{
			ReadMultiLineReply(Header);
		}
	}

	return Header;
}

bool MailPop3::ReadMultiLineReply(char *&Str)
{
	bool Status = false;
	if (Socket)
	{
		BytePipe Temp;
		int ReadLen = Socket->Read(Buffer, sizeof(Buffer), 0);
		if (ReadLen > 0 AND Buffer[0] == '+')
		{
			// positive response
			char *Eol = strchr(Buffer, '\n');
			if (Eol)
			{
				char *Ptr = Eol + 1;
				ReadLen -= ((int)Ptr-(int)Buffer);
				memmove(Buffer, Ptr, ReadLen);
				Temp.Push((uchar*) Buffer, ReadLen);

				while (	NOT MailIsEnd(Buffer, ReadLen) AND
						NOT Socket->Cancel())
				{
					ReadLen = Socket->Read(Buffer, sizeof(Buffer), 0);
					Temp.Push((uchar*) Buffer, ReadLen);
				}

				int Len = Temp.Sizeof();
				Str = NEW(char[Len+1]);
				if (Str)
				{
					Temp.Pop((uchar*)Str, Len);
					Str[Len] = 0;
					Status = true;
				}
			}
		}
	}

	return Status;
}

bool MailPop3::Close()
{
	if (Socket)
	{
		// logout
		VERIFY_RET_VAL(Write("QUIT\r\n", true));
		VERIFY_RET_VAL(ReadReply());

		Socket->Close();
		Socket = 0;
		Messages = 0;
		return TRUE;
	}
	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
MailMessage::MailMessage()
{
	Subject = 0;
	Text = 0;
	MessageID = 0;
	InternetHeader = 0;
	Priority = MAIL_PRIORITY_NORMAL;
	CodePage = 0;
	HtmlText = 0;
}

MailMessage::~MailMessage()
{
	Empty();
}

void MailMessage::Empty()
{
	DeleteArray(Subject);
	DeleteArray(Text);
	DeleteArray(MessageID);
	DeleteArray(InternetHeader);
	DeleteArray(HtmlText);

	for (AddressDescriptor *a = To.First(); a; a = To.Next())
	{
		DeleteObj(a);
	}
	To.Empty();

	for (FileDescriptor *f = Attachments.First(); f; f = Attachments.Next())
	{
		DeleteObj(f);
	}
	Attachments.Empty();
}

void MailMessage::EncodeQuotedPrintable(StringPipe &Out, char *In)
{
	if (In)
	{
		char Buffer[100];
		char *d = Buffer;
		for (char *s = In; true; )
		{
			if (*s == '\n' OR NOT *s)
			{
				strcpy(d, "\r\n");
				d = Buffer;
				Out.Push(Buffer);

				if (NOT *s) break;
			}
			else if (*s & 0x80 OR
					 *s == '.' OR
					 *s == '=')
			{
				sprintf(d, "=%02.2X", (uchar)*s);
				d += 3;
			}
			else if (*s != '\r')
			{
				*d++ = *s;
			}
			s++;

			if ((int)d-(int)Buffer > 73)
			{
				// time for a new line.
				strcpy(d, "=\r\n");
				d = Buffer;
				Out.Push(Buffer);
			}
		}
	}
}

void MailMessage::EncodeText(StringPipe &Out, char *In)
{
	char *s;
	for (s = In; s AND *s; )
	{
		char *NewLine = strchr(s, '\n');
		int Len = (NewLine) ? (int)NewLine-(int)s : strlen(s);
		if (Len > 0 AND s[Len-1] == '\r') Len--;
		if (Len == 1 AND s[0] == '.')
		{
			// this removes the sequence ".\n"
			// which is the END OF MAIL in the SMTP protocol.
			Out.Push(". ");
		}
		else
		{
			Out.BytePipe::Push((uchar*)s, Len);
		}
		Out.Push("\r\n");
		s = (NewLine) ? NewLine + 1 : 0;
	}
}

/*
// find time zone offset and correct for DST
if (tbTime.dstflag)
{
	tbTime.timezone -= 60;
}
*/

// Encode the whole email
bool MailMessage::Encode(StringPipe &Out, MailProtocol *Protocol, bool Mime)
{
	bool Status = EncodeHeaders(Out, Protocol, Mime);
	if (Status)
	{
		Status = EncodeBody(Out, Protocol, Mime);
	}

	return Status;
}

// This encodes the main headers but not the headers relating to the 
// actual content. Thats done by the ::EncodeBody function.
bool MailMessage::EncodeHeaders(StringPipe &Out, MailProtocol *Protocol, bool Mime)
{
	// Setup
	char Buffer[1025];
	time_t Time = time(NULL);
	timeb tbTime;
	tm *ptm = localtime(&Time);

	// Construct date
	char TimeStr[MAX_NAME_SIZE+1];
	strftime(TimeStr, MAX_NAME_SIZE, "%a, %d %b %Y %H:%M:%S", ptm);
	ftime(&tbTime);

	// Date:
	sprintf(Buffer,
			"Date: %s %s%2.2d%2.2d\r\n",
			TimeStr,
			(tbTime.timezone < 0) ? "+" : "",
			-tbTime.timezone / 60,
			tbTime.timezone % 60);
	Out.Push(Buffer);

	// X-Mailer:
	sprintf(Buffer, "X-Mailer: %s\r\n", (Protocol) ? Protocol->ProgramName : LIB_NAME);
	Out.Push(Buffer);

	if (Priority != MAIL_PRIORITY_NORMAL)
	{
		// X-Priority:
		sprintf(Buffer, "X-Priority: %i\r\n", Priority);
		Out.Push(Buffer);
	}

	// Message-ID:
	if (MessageID)
	{
		sprintf(Buffer, "Message-ID: %s\r\n", MessageID);
		Out.Push(Buffer);
	}

	// To:
	char *ToAddr = CreateAddressTag(To, 0);
	if (ToAddr)
	{
		Out.Push(ToAddr);
		DeleteArray(ToAddr);
	}

	char *CcAddr = CreateAddressTag(To, 1);
	if (CcAddr)
	{
		Out.Push(CcAddr);
		DeleteArray(CcAddr);
	}

	// From:
	if (From.Addr)
	{
		strcpy(Buffer, "From: ");

		if (From.Name)
		{
			sprintf(Buffer + strlen(Buffer), "\"%s\" ", From.Name);
		}

		sprintf(Buffer + strlen(Buffer), "<%s>\r\n", From.Addr);
	}
	else
	{
		return false;
	}
	Out.Push(Buffer);

	// Reply-To:
	if (ValidStr(Reply.Addr))
	{
		if (ValidStr(Reply.Name))
		{
			sprintf(Buffer, "Reply-To: \"%s\" <%s>\r\n", Reply.Name, Reply.Addr);
		}
		else
		{
			sprintf(Buffer, "Reply-To: <%s>", Reply.Addr);
		}

		Out.Push(Buffer);
	}

	// Subject:
	char *Subj = EncodeRfc2047(NewStr(Subject));
	sprintf(Buffer, "Subject: %s\r\n", (Subj) ? Subj : "");
	Out.Push(Buffer);
	DeleteArray(Subj);

	return true;
}

bool MailMessage::EncodeBody(StringPipe &Out, MailProtocol *Protocol, bool Mime)
{
	char Buffer[1025];
	char *CodePageName = GdcGetCodePageMimeType(CodePage);
	if (NOT CodePageName) CodePageName = "us-ascii";

	if (Mime)
	{
		bool MultiPart = (Text) AND (Attachments.GetItems() > 0);
		char Separator[256];
		#ifdef WIN32
		sprintf(Separator, "---- =_NextPart_000_%08.8X.%08.8X", time(NULL), GetCurrentProcessId());
		#else if BEOS
		sprintf(Separator, "---- =_NextPart_000_%08.8X", time(NULL));
		#endif

		sprintf(Buffer, "MIME-Version: 1.0\r\n");
		Out.Push(Buffer);
		
		if (MultiPart)
		{
			sprintf(Buffer, "Content-Type: multipart/mixed; boundary=\"%s\"\r\n", Separator);
			Out.Push(Buffer);
			Out.Push("Status: RO\r\n\r\n");
		}
		
		if (Text)
		{
			if (MultiPart)
			{
				sprintf(Buffer, "\r\n--%s\r\n", Separator);
				Out.Push(Buffer);
			}

			// Content type
			sprintf(Buffer, "Content-Type: text/plain; charset=\"%s\"\r\n", CodePageName);
			Out.Push(Buffer);

			// Transfer encoding
			if (Is8Bit(Text) OR
				MaxLineLen(Text) >= 80)
			{
				Out.Push("Content-Transfer-Encoding: quoted-printable\r\n\r\n");
				EncodeQuotedPrintable(Out, Text);
			}
			else
			{
				Out.Push("Content-Transfer-Encoding: 7bit\r\n\r\n");
				EncodeText(Out, Text);
			}
		}

		FileDescriptor *FileDes = Attachments.First();
		while (FileDes)
		{
			GFile *F = FileDes->GotoObject();
			if (F)
			{
				int InLen = FileDes->Sizeof();
				uchar *InBuf = NEW(uchar[InLen]);
				if (InBuf)
				{
					F->Read(InBuf, InLen);
					int OutLen = BufferLen_BinTo64(InLen);
					char *OutBuf = NEW(char[OutLen]);
					if (OutBuf)
					{
						int Len = ConvertBinaryToBase64(OutBuf, OutLen, InBuf, InLen);
						if (Len > 0)
						{
							// successfully encoded attachment
							// so write a MIME segment for this
							if (MultiPart)
							{
								sprintf(Buffer, "\r\n--%s\r\n", Separator);
								Out.Push(Buffer);
							}

							char FileName[256];
							char *s = FileDes->Name(), *d = FileName;
							while (*s)
							{
								if (*s != '\\')
								{
									*d++ = *s++;
								}
								else
								{
									*d++ = '\\';
									*d++ = '\\';
									s++;
								}
							}
							*d = 0;

							char *FName = EncodeRfc2047(NewStr(FileName));
							sprintf(Buffer,
									"Content-Type: %s; "
									"name=\"%s\"\r\n",
									FileDes->GetMimeType() ? FileDes->GetMimeType() : "application/x-zip-compressed",
									(FName) ? FName : FileName);
							Out.Push(Buffer);
							DeleteArray(FName);

							sprintf(Buffer, "Content-Transfer-Encoding: base64\r\n\r\n");
							Out.Push(Buffer);

							int Pos = 0;
							while (Pos < Len)
							{
								int Size = Len - Pos;
								int WriteLen = min(Size, EMAIL_LINE_SIZE);

								memcpy(Buffer, OutBuf + Pos, WriteLen);
								strcpy(Buffer + WriteLen, "\r\n");
								Out.Push(Buffer);
								Pos += WriteLen;
							}
						}

						DeleteArray(OutBuf);
					}

					DeleteArray(InBuf);
				}
			}

			FileDes = Attachments.Next();
		}

		if (MultiPart)
		{
			// write final separator
			sprintf(Buffer, "\r\n--%s--\r\n\r\n", Separator);
			Out.Push(Buffer);
		}
	}
	else
	{
		// send content type
		sprintf(Buffer, "Content-Type: text/plain; charset=\"%s\"\r\n", CodePageName);
		Out.Push(Buffer);

		if (Is8Bit(Text))
		{
			// send the encoding and a blank line
			Out.Push("Content-Transfer-Encoding: quoted-printable\r\n\r\n");
			
			// send message text
			EncodeQuotedPrintable(Out, Text);
		}
		else
		{
			// send a blank line
			Out.Push("\r\n");
			
			// send message text
			EncodeText(Out, Text);
		}
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////
#define DEFAULT_IMAP_PORT				143
#ifndef WIN32
#include "AuthNtlm/Ntlm.h"
#else
#include "AuthNtlm\ntlm.h"
#endif

void ClearStrList(List<char> &l)
{
	for (char *s=l.First(); s; s=l.Next())
	{
		DeleteArray(s);
	}
	l.Empty();
}

void ChopNewLine(char *Str)
{
	char *End = Str+strlen(Str)-1;
	if (*End == '\n')
	{
		*End-- = 0;
	}
	if (*End == '\r')
	{
		*End-- = 0;
	}
}

MailIMap::MailIMap()
{
	NextCmd = 1;
	Buffer[0] = 0;
}

MailIMap::~MailIMap()
{
	ClearDialog();
	ClearUid();
}

void MailIMap::ClearUid()
{
	ClearStrList(Uid);
}

void MailIMap::ClearDialog()
{
	ClearStrList(Dialog);
}

bool MailIMap::WriteBuf()
{
	if (S)
	{
		if (S->Write(Buf, strlen(Buf), 0) == strlen(Buf))
		{
			return true;
		}
	}

	return false;
}

bool MailIMap::ReadResponse(int Cmd)
{
	bool Done = false;
	bool Status = false;
	if (S)
	{
		int Len = strlen(Buffer);
		while (NOT Done)
		{
			int Read = S->Read(Buffer+Len, sizeof(Buffer)-Len-1, 0);
			if (Read > 0)
			{
				Buffer[Len+Read] = 0;
				char *Start = Buffer;

				for (char *p=Buffer+Len; p<Buffer+(Len+Read); p++)
				{
					if (p[0] == '\r' AND p[1] == '\n')
					{
						*p++ = 0;
						/*
						if (Start[0] == '*' OR
							Start[0] == '+')
						{
						*/
							Dialog.Insert(NewStr(Start));

							if (Cmd < 0)
							{
								Status = Done = true;
							}
						/*
						}
						else
						*/
						{
							char Num[8];
							sprintf(Num, "A%04.4i ", Cmd);
							if (strnicmp(Start, Num, 6) == 0)
							{
								Done = true;
								Status = strnicmp(Start+6, "OK", 2) == 0;
								if (NOT Status)
								{
									DeleteArray(ServerMsg);
									ServerMsg = NewStr(Start+6);
								}
							}
						}

						Start = p+1;
					}
				}

				char *End = Buffer + (Len + Read);
				if (Start < End)
				{
					memmove(Buffer, Start, (int)End-(int)Start + 1);
					Len = strlen(Buffer);
				}
				else
				{
					Buffer[0] = 0;
					Len = 0;
				}
			}
		}
	}

	return Status;
}

bool MailIMap::Open(GSocket *s, char *RemoteHost, int Port, char *User, char *Password)
{
	bool Status = false;

	S = s;
	if (S AND
		RemoteHost AND
		User AND
		Password)
	{
		// prepare address
		if (Port < 1) Port = DEFAULT_IMAP_PORT;

		char Remote[256];
		strcpy(Remote, RemoteHost);
		char *Colon = strchr(Remote, ':');
		if (Colon)
		{
			*Colon++ = 0;
			Port = atoi(Colon);
		}

		// connect
		if (S->Open(Remote, Port))
		{
			bool IMAP4Server = false;
			List<char> Auths;

			// check capability
			int CapCmd = NextCmd++;
			sprintf(Buf, "A%04.4i CAPABILITY\r\n", CapCmd);
			if (WriteBuf())
			{
				if (ReadResponse(CapCmd))
				{
					for (char *r=Dialog.First(); r; r=Dialog.Next())
					{
						TokenParser T(r, " ");
						if (T.Tokens > 1 AND
							stricmp(T.Get(1), "CAPABILITY") == 0)
						{
							for (int i=2; i<T.Tokens; i++)
							{
								if (stricmp(T.Get(i), "IMAP4") == 0 OR
									stricmp(T.Get(i), "IMAP4rev1") == 0)
								{
									IMAP4Server = true;
								}
								if (strnicmp(T.Get(i), "AUTH=", 5) == 0)
								{
									Auths.Insert(NewStr(T.Get(i)+5));
								}
							}
						}
					}

					ClearDialog();
				}
			}

			if (IMAP4Server)
			{
				char *DefaultAuthType = 0;
				if (Auths.GetItems() == 0)
				{
					// no auths???
					Auths.Insert(DefaultAuthType = NewStr("LOGIN"));
				}

				// login
				bool LoggedIn = false;
				char AuthTypeStr[256] = "";
				char *AuthType = 0;
				for (AuthType=Auths.First(); AuthType; AuthType=Auths.Next())
				{
					if (DefaultAuthType != AuthType)
					{
						// put a string of auth types together
						if (strlen(AuthTypeStr) > 0)
						{
							strcat(AuthTypeStr, ", ");
						}
						strcat(AuthTypeStr, AuthType);
					}

					// do auth
					if (stricmp(AuthType, "NTLM") == 0)
					{
						// NT Lan Man authentication
						
						// Username is in the format: User[@Domain]
						char UserDom[256];
						strcpy(UserDom, User);
						char *Domain = strchr(UserDom, '@');
						if (Domain)
						{
							*Domain++ = 0;
						}
						else
						{
							Domain = UserDom + strlen(UserDom);
						}

						int AuthCmd = NextCmd++;
						sprintf(Buf, "A%04.4i AUTHENTICATE NTLM\r\n", AuthCmd);
						if (WriteBuf())
						{
							if (ReadResponse())
							{
								tSmbNtlmAuthRequest   request;              
								tSmbNtlmAuthChallenge challenge;
								tSmbNtlmAuthResponse  response;
								char tmpstr[32];

								buildSmbNtlmAuthRequest(&request,UserDom,Domain);
								ZeroObj(Buf);
								ConvertBinaryToBase64(Buf, sizeof(Buf), (uchar*) &request, SmbLength(&request));
								strcat(Buf, "\r\n");
								WriteBuf();

								/* read challange data from server, convert from base64 */

								Buf[0] = 0;
								ClearDialog();
								if (ReadResponse())
								{
									/* buffer should contain the string "+ [base 64 data]" */

									char *Line = Dialog.First();
									ASSERT(Line);
									ChopNewLine(Line);
									ConvertBase64ToBinary((uchar*) &challenge, sizeof(challenge), Line+2, strlen(Line)-2);

									/* prepare response, convert to base64, send to server */

									buildSmbNtlmAuthResponse(&challenge, &response, User, Password);
									ZeroObj(Buf);
									ConvertBinaryToBase64(Buf, sizeof(Buf), (uchar*) &response, SmbLength(&response));
									strcat(Buf, "\r\n");
									WriteBuf();

									/* read line from server, it should be "[seq] OK blah blah blah" */

									LoggedIn = ReadResponse(AuthCmd);
								}
							}
						}

						ClearDialog();
					}
					else if (stricmp(AuthType, "KERBEROS_V4") == 0)
					{
						// kerberos v4 authenication
					}
					else if (stricmp(AuthType, "LOGIN") == 0)
					{
						// clear text authentication
						char *uUser = NewStr(User);
						char *uPassword = NewStr(Password);
						if (uUser AND uPassword)
						{
							strupr(uUser);
							strupr(uPassword);

							int AuthCmd = NextCmd++;
							sprintf(Buf, "A%04.4i LOGIN %s %s\r\n", AuthCmd, uUser, uPassword);
							if (WriteBuf())
							{
								LoggedIn = ReadResponse(AuthCmd);
							}
						}

						DeleteArray(uUser);
						DeleteArray(uPassword);
					}
				}

				if (LoggedIn)
				{
					Status = true;
				}
				else
				{
					char Str[300];
					sprintf(Str,
							"Authentication failed, types available:\n%s",
							(strlen(AuthTypeStr)>0) ? AuthTypeStr : "(none)");
					ServerMsg = NewStr(Str);
				}
			}
		}
	}

	return Status;
}

bool MailIMap::Close()
{
	bool Status = false;
	if (S)
	{
		int Cmd = NextCmd++;
		sprintf(Buf, "A%04.4i EXPUNGE\r\n", Cmd);
		if (WriteBuf())
		{
			ClearDialog();
			ReadResponse(Cmd);
		}
		
		Cmd = NextCmd++;
		sprintf(Buf, "A%04.4i LOGOUT\r\n", Cmd);
		if (WriteBuf())
		{
			Status = true;
		}
	}
	return Status;
}

int MailIMap::GetMessages()
{
	int Status = 0;

	if (S)
	{
		int Cmd = NextCmd++;
		sprintf(Buf, "A%04.4i SELECT INBOX\r\n", Cmd);
		if (WriteBuf())
		{
			ClearDialog();
			if (ReadResponse(Cmd))
			{
				for (char *d = Dialog.First(); d AND NOT Status; d=Dialog.Next())
				{
					if (strstr(d, "EXISTS"))
					{
						TokenParser T(d, " ");
						if (T.Tokens == 3)
						{
							Status = atoi(T.Get(1));
						}
					}
				}
			}
		}
	}

	return Status;
}

bool MailIMap::GetParts(int Message, char *&Ptr, char *Parts)
{
	bool Status = false;

	if (Parts)
	{
		int Cmd = NextCmd++;
		sprintf(Buf, "A%04.4i FETCH %i (%s)\r\n", Cmd, Message+1, Parts);
		if (WriteBuf())
		{
			ClearDialog();
			Buf[0] = 0;
			if (ReadResponse(Cmd))
			{
				StringPipe Text;
				int Chars = 0;

				for (char *d = Dialog.First(); d AND NOT Status; d=Dialog.Next())
				{
					if (Chars < 1)
					{
						char *s = strchr(d, '{');
						if (s)
						{
							char *e = strchr(s, '}');
							if (e)
							{
								*e = 0;
								Chars = atoi(s+1);
								continue;
							}
						}

						// error
						break;
					}

					Text.Push(d);
					Text.Push("\r\n");
					Chars -= strlen(d) + 2;
				}

				Status = (Ptr = Text.NewStr()) != 0;
			}
		}
	}

	return Status;
}

bool MailIMap::Receive(int Message, MailMessage *Msg, bool Delete, int MaxSize)
{
	if (Msg)
	{
		return GetParts(Message, Msg->Text, "RFC822.HEADER RFC822.TEXT");
	}
	return false;
}

bool MailIMap::Delete(int Message)
{
	bool Status = false;

	if (S)
	{
		int Cmd = NextCmd++;
		sprintf(Buf, "A%04.4i STORE %i +FLAGS.SILENT (\\deleted)\r\n", Cmd, Message+1);
		if (WriteBuf())
		{
			ClearDialog();
			Status = ReadResponse(Cmd);
		}
	}

	return Status;
}

int MailIMap::Sizeof(int Message)
{
	int Status = 0;

	if (S)
	{
		char *Tag = "RFC822.SIZE";

		int Cmd = NextCmd++;
		sprintf(Buf, "A%04.4i FETCH %i (%s)\r\n", Cmd, Message+1, Tag);
		if (WriteBuf())
		{
			ClearDialog();
			Buf[0] = 0;
			if (ReadResponse(Cmd))
			{
				char *d = Dialog.First();
				if (d)
				{
					char *t = strstr(d, Tag);
					if (t)
					{
						t += strlen(Tag) + 1;
						Status = atoi(t);
					}
				}
			}
		}
	}

	return Status;
}

bool MailIMap::GetUid(int Message, char *Id)
{
	if (FillUidList())
	{
		char *s = Uid.ItemAt(Message);
		if (s AND Id)
		{
			strcpy(Id, s);
		}
	}

	return false;
}

bool MailIMap::FillUidList()
{
	bool Status = false;

	if (S)
	{
		if (Uid.GetItems() == 0)
		{
			int Cmd = NextCmd++;
			sprintf(Buf, "A%04.4i UID SEARCH ALL\r\n", Cmd);
			if (WriteBuf())
			{
				ClearDialog();
				if (ReadResponse(Cmd))
				{
					for (char *d = Dialog.First(); d AND NOT Status; d=Dialog.Next())
					{
						TokenParser T(d, " ");
						if (T.Get(1) AND strcmp(T.Get(1), "SEARCH") == 0)
						{
							for (int i=2; i<T.Tokens; i++)
							{
								Uid.Insert(NewStr(T.Get(i)));
							}
						}
					}

					Status = true;
				}
			}
		}
		else
		{
			Status = true;
		}
	}

	return Status;
}

bool MailIMap::GetUidList(List<char> &Id)
{
	bool Status = false;

	if (FillUidList())
	{
		for (char *s=Uid.First(); s; s=Uid.Next())
		{
			Id.Insert(NewStr(s));
		}

		Status = true;
	}

	return Status;
}

char *MailIMap::GetHeaders(int Message)
{
	char *Text;
	if (GetParts(Message, Text, "RFC822.HEADER"))
	{
		return Text;
	}
	return 0;
}
