/*hdr
**	FILE:			Effect.h
**	AUTHOR:			Matthew Allen
**	DATE:			4/3/98
**	DESCRIPTION:	Draw mode effects
**
**		Copyright (C) 1998, Matthew Allen
**			fret@ozemail.com.au
*/

class GAlphaApp : public GApplicator {
protected:
	uchar a, oma;
	int Bits, Bytes;
	uchar *Ptr;

public:
	GAlphaApp()
	{
		Op = GDC_ALPHA;
		a = 0xFF;
		oma = 0;
		Bits = 8;
		Bytes = 1;
		Ptr = 0;
	}

	uchar GetAlpha() { return a; }
	virtual void SetAlpha(uchar A, GdcPalette *Pal)
	{ 
		a = A;
		oma = 0xFF - A;
	}

	bool SetSurface(GdcSurface *d, GdcPalette *p, GdcSurface *a)
	{
		if (d AND d->Bits == Bits)
		{
			Dest = d;
			Pal = p;
			Ptr = d->Base;
			Alpha = a;
			return TRUE;
		}
		return FALSE;
	}

	void SetPtr(int x, int y)
	{
		ASSERT(Dest AND Dest->Base);
		Ptr = Dest->Base + ((y * Dest->Line) + (x * Bytes));
	}

	void IncX()
	{
		Ptr += Bytes;
	}

	void IncY()
	{
		Ptr += Dest->Line;
	}

	void IncPtr(int X, int Y)
	{
		Ptr += (Y * Dest->Line) + (X * Bytes);
	}

	COLOUR Get()
	{
		switch (Bytes)
		{
			case 1:
				return *((uchar*)Ptr);
			case 2:
				return *((ushort*)Ptr);
			case 3:
				return Rgb24(Ptr[2], Ptr[1], Ptr[0]);
			case 4:
				return *((ulong*)Ptr);
		}
		return 0;
	}
};

class GdcApp8Alpha : public GAlphaApp {

	char Remap[256];

public:
	GdcApp8Alpha();
	void SetAlpha(uchar A, GdcPalette *Pal);

	void Set();
	void VLine(int height);
	void Rectangle(int x, int y);
	bool Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha);
};

class GdcApp15Alpha : public GAlphaApp {
public:
	GdcApp15Alpha()
	{
		Bits = 15; Bytes = 2;
	}

	void Set();
	void VLine(int height);
	void Rectangle(int x, int y);
	bool Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha);
};

class GdcApp16Alpha : public GAlphaApp {
public:
	GdcApp16Alpha()
	{
		Bits = 16; Bytes = 2;
	}

	void Set();
	void VLine(int height);
	void Rectangle(int x, int y);
	bool Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha);
};

class GdcApp24Alpha : public GAlphaApp {
public:
	GdcApp24Alpha()
	{
		Bits = 24; Bytes = 3;
	}

	void Set();
	void VLine(int height);
	void Rectangle(int x, int y);
	bool Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha);
};

class GdcApp32Alpha : public GAlphaApp {
public:
	GdcApp32Alpha()
	{
		Bits = 32; Bytes = 4;
	}

	void Set();
	void VLine(int height);
	void Rectangle(int x, int y);
	bool Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha);
};

class GdcAlphaServer : public GdcApplicatorServer {
public:
	int Op() { return GDC_ALPHA; }
	GApplicator *Create(int Bits);
};

