
#ifndef __DATA_DLG_TOOLS
#define __DATA_DLG_TOOLS

#include "GProperties.h"

#define DATA_STR					1
#define DATA_BOOL					2
#define DATA_INT					3
#define DATA_FLOAT					4

class DataDlgTools;

class LgiClass DataDlgField {

	friend class DataDlgTools;

protected:
	int Type;
	int CtrlId;
	char *Option;
	char *Desc;

public:
	DataDlgField(int type, int ctrlid, char *opt, char *desc = 0);
	~DataDlgField();

	int GetType() { return Type; }
	int GetCtrl() { return CtrlId; }
	char *GetOption() { return Option; }
	char *GetDesc() { return Desc; }
};

typedef List<DataDlgField> DataDlgFieldList;

class LgiClass DataDlgTools {
protected:
	GView *Dlg;
	ObjProperties *Options;

public:
	DataDlgTools();

	void Set(GView *Dlg, ObjProperties *Options);
	ObjProperties *GetOptions();

	void SaveInt(GEdit *Ed, char *Name);
	void InitInt(GEdit *Ed, char *Name);
	void SaveEdit(GEdit *Ed, char *Name);
	void InitEdit(GEdit *Ed, char *Name);
	void SaveCheck(GCheckBox *Chk, char *Name);
	void InitCheck(GCheckBox *Chk, char *Name);
	bool ProcessField(DataDlgField *f, bool Write, char *OptionOverride = 0);
	bool ProcessFields(DataDlgFieldList &Field, bool Write);
	bool LoadFields(DataDlgFieldList &Field);
	bool SaveFields(DataDlgFieldList &Field);
};

class LgiClass DRecordSetObj
{
public:
	virtual bool Serialize(ObjProperties &f, bool Write)
	{
		return false;
	}
};

class LgiClass DRecordSetCtrls : public DataDlgTools
{
	// Controls
	GText *Description;
	GScrollBar *Scroll;
	int NewRecordId;
	int DeleteRecordId;

	DataDlgFieldList *Fields;
	
protected:
	DRecordSetObj *Current;
	List<DRecordSetObj> *Records;

public:
	DRecordSetCtrls(	GView *window,
						DataDlgFieldList *fields,
						List<DRecordSetObj> *records,
						int DescId,
						int ScrollId,
						int NewId,
						int DelId);


	virtual DRecordSetObj *NewObj() { return 0; }
	virtual void DelObj(DRecordSetObj *Obj) {}

	int GetCurrentIndex();
	void Serialize(bool Write);
	void OnMoveRecord(DRecordSetObj *r);
	int OnNotify(GView *Col, int Flags);
};

template <class Record>
class LgiClass DRecordSet : public DRecordSetCtrls
{
public:
	DRecordSet(	GView *window, DataDlgFieldList *fields, List<Record> *records, int DescId, int ScrollId, int NewId, int DelId) :
		DRecordSetCtrls(window, fields, (List<DRecordSetObj>*) records, DescId, ScrollId, NewId, DelId)
	{
	}

	DRecordSetObj *NewObj()
	{
		return NEW(Record);
	}

	void DelObj(DRecordSetObj *Obj)
	{
		Record *r = (Record*) Obj;
		if (r)
		{
			DeleteObj(r);
		}
	}

	Record *ItemAt(int i=-1)
	{
		return dynamic_cast<Record*>((i<0 OR NOT Records) ? Current : Records->ItemAt(i));
	}

	void OnMoveRecord(Record *r)
	{
		DRecordSetCtrls::OnMoveRecord((DRecordSetObj*)r);
	}
};

#endif