/*hdr
**      FILE:           GDb.h
**      AUTHOR:         Matthew Allen
**      DATE:           8/2/2000
**      DESCRIPTION:    Db API definitions
**
**      Copyright (C) 2000 Matthew Allen
**						fret@ozemail.com.au
*/

#ifndef __GDB_H
#define __GDB_H

// Data types
#define GDbVoid						0
#define GDbInt						1
#define GDbFloat					2
#define GDbString					3
#define GDbDate						4
#define GDbBoolean					5
#define GDbMemo						6
#define GDbBinary					7

// Forward decl
class GDatabase;

// Classes
class GDbField
{
public:
	virtual ~GDbField() {};

	// Properties
	virtual char *Name() = 0;
	virtual bool Name(char *Str) = 0;
	virtual int Type() = 0;
	virtual bool Type(int NewType) = 0;
	virtual int Length() = 0;
	virtual bool Length(int NewLength) = 0;
	virtual char *Description() = 0;
	virtual bool Description(char *NewDesc) = 0;

	// Setting the data
	virtual bool Set(char *Str) = 0;
	virtual bool Set(int Int) = 0;
	virtual bool Set(double Dbl) = 0;
	virtual bool Set(void *Data, int Length) = 0;

	// Getting the data
	virtual bool Get(char *&Str) = 0;
	virtual bool Get(int &Int) = 0;
	virtual bool Get(double &Dbl) = 0;
	virtual bool Get(void *&Data, int &Length) = 0;
};

class GDbRecordset
{
public:
	virtual ~GDbRecordset() {};

	// Fields
	virtual GDbField *Field(int Index) = 0;
	virtual GDbField *Field(char *Name) = 0;
	virtual bool InsertField(	GDbField *Fld,
								int Index = -1) = 0; // default: append to the end
	virtual bool DeleteField(GDbField *Fld) = 0;
	virtual int CountFields() = 0; // return number of fields

	// Markers
	virtual bool Bof() = 0;
	virtual bool Eof() = 0;

	// New/Edit
	virtual bool AddNew() = 0;
	virtual bool Edit() = 0;
	virtual bool Update() = 0;

	// Moving around the records
	virtual bool MoveFirst() = 0;
	virtual bool MoveNext() = 0;
	virtual bool MovePrev() = 0;
	virtual bool MoveLast() = 0;
	virtual int SeekRecord(int i) = 0; // from the beginning

	// Numbers/Counts
	virtual int RecordIndex() = 0;
	virtual int CountRecords() = 0;
	
	// Deleting
	virtual bool DeleteRecord() = 0; // delete's the current record
};

class GDb
{
public:
	virtual ~GDb() {}

	// Attaching to data
	virtual bool Connect(char *Init) = 0;
	virtual bool Disconnect() = 0;

	// Retrieving data
	virtual GDbRecordset *OpenRecordset(char *Name) = 0;

	// Design
	virtual GDbRecordset *GetTables() = 0;
	virtual GDbRecordset *GetFields(char *Table) = 0;
};

// Actual implementations
extern GDb *OpenDbfDatabase(char *Dir);		// add 'GDb-Dbf.cpp' to your project
extern GDb *OpenAdoDatabase(char *Name);	// add 'GDb-Ado.cpp' to your project

#endif