/*hdr
**      FILE:           GDocApp.h
**      AUTHOR:         Matthew Allen
**      DATE:           20/4/01
**      DESCRIPTION:    Generic Document Application Header
**
*/

#ifndef __GDOCAPP_H
#define __GDOCAPP_H

#include <stdio.h>

#include "Lgi.h"
#include "GXml.h"
#include "GMru.h"

////////////////////////////////////////////////////////////////////////////////////////////
// Defines

// window messages
#define		IDM_SAVE			(IDM_GMRU_BASE+2)
#define		IDM_CLOSE			(IDM_GMRU_BASE+3)
#define		IDM_EXIT			(IDM_GMRU_BASE+4)

// Misc
#define		MainWnd				((AppWnd*)GApp::ObjInstance()->AppWnd)

////////////////////////////////////////////////////////////////////////////////////////////
// Classes
class App : public GApp {
public:
	#if defined WIN32
	App(HINSTANCE inst, char *cmd, int show) :
		GApp(inst, cmd, show) {}
	#elif defined BEOS
	App(char *MimeSig, char *CmdLine) :
		GApp(MimeSig, CmdLine) {}
	#elif defined LINUX
	App(int Args, char **Arg) :
		GApp(Args, Arg) {}
	#endif
};

class GDocApp :
	public GWindow,
	public GMru
{
	#if defined WIN32
	static GWin32Class Win32Class;
	#endif

private:
	// Data
	char			*_OptionsFile;
	ObjProperties	*_Options;
	char			*_AppName;
	char			*_CurFile;
	bool			_Dirty;

	// Methods
	void			_Close();
	char			*_OptionsFileName();

protected:
	// Methods
	bool			_LoadMenu(char *Resource = 0);
	bool			_Create();
	bool			_Destroy();
	bool			_OpenFile(char *File);
	bool			_SaveFile(char *File);
	bool			_DoSerialize(bool Write);
	bool			_SerializeFile(GFile &f, bool Write);

	// Data
	GSubMenu		*_FileMenu;

public:
	GDocApp(char *appname = 0, char *icon = 0);
	~GDocApp();

	// Property access
	void SetCurFile(char *f);
	char *GetCurFile();
	bool SetDirty(bool Dirty);
	bool GetDirty();
	ObjProperties *GetOptions();
	char *GetAppName();

	// Override
	virtual void Empty() {}
	virtual bool SerializeOptions(ObjProperties *Options, bool Write) { return false; }

	// Impl
	void OnReceiveFiles(List<char> &Files);
	bool OnRequestClose();
	int OnCommand(int Cmd, int Event, OsView Window);
	int OnEvent(GMessage *m);
};

#endif
