/*hdr
**      FILE:           Filter2.h
**      AUTHOR:         Matthew Allen
**      DATE:           1/4/97
**      DESCRIPTION:    Graphics file filters
**
**      Copyright (C) 1997, Matthew Allen
**              fret@ozemail.com.au
*/

#ifndef FILTER2_H
#define FILTER2_H

#include <string.h>
#include "Contain.h"
#include "GProperties.h"

LgiFunc int FindHeader(int Offset, char *Str, GFile &f);

#define FILTER_CAP_READ			0x0001
#define FILTER_CAP_WRITE		0x0002
#define FILTER_CAP_INFO			0x0004

class LgiClass GFilter : public GFile
{
protected:
	GdcSurface *GetSurface(GDeviceContext *pDC) { return pDC->pMem; }
	GRegion *GetClip(GDeviceContext *pDC) { return &pDC->Clip; }
	int GetLineLength(GDeviceContext *pDC) { return (pDC->pMem) ? pDC->pMem->Line : 0; }

	Progress *Meter;

public:
	ObjProperties Props;

	GFilter() { Meter = 0; }
	virtual ~GFilter() {}

	// properties
	Progress *GetProgress() { return Meter; }
	void SetProgress(Progress *Prg) { Meter = Prg; }

	// image read write
	virtual int GetCapabilites() { return 0; }
	virtual int GetImages() { return 1; }
	virtual bool ReadImage(GSurface *pDC) { return FALSE; }
	virtual bool WriteImage(GSurface *pDC) { return FALSE; }
};

#define GDC_RLE_COLOUR				0x0001
#define GDC_RLE_MONO				0x0002
#define GDC_RLE_READONLY			0x0004
#define GDC_RLE_KEY					0x0008		// set if valid

class LgiClass GdcRleDC : public GMemDC {
protected:
	COLOUR	Key;

	int	Flags;
	int	Length;
	int	Alloc;
	uchar	*Data;
	uchar	**ScanLine;

	bool SetLength(int Len);
	bool FindScanLines();
	void Empty();

public:
	GdcRleDC();
	virtual ~GdcRleDC();

	void Move(GdcRleDC *pDC);

	bool Create(int x, int y, int Bits, int LineLen = 0);
	bool CreateInfo(int x, int y, int Bits);
	void ReadOnly(bool Read);
	bool ReadOnly();
	void Mono(bool Mono);
	bool Mono();

	void Update(int Flags);
	void Draw(GSurface *Dest, int x, int y);

	int SizeOf();
	bool Read(GFile &F);
	bool Write(GFile &F);
};


////////////////////////////////////////////////////////////////
class LgiClass GFilterFactory
{
	static class GFilterFactory *First;
	class GFilterFactory *Next;

	// Override these in your filter library
	virtual bool CheckFile(char *File) { return false; }
	virtual GFilter *NewObject() { return NULL; }

public:
	GFilterFactory();
	virtual ~GFilterFactory();

	static GFilter *New(char *File);
	static GFilter *NewAt(int i);
	static int GetItems();
};

#endif
