/*hdr
**	FILE:			GdcFont.h
**	AUTHOR:			Matthew Allen
**	DATE:			5/5/97
**	DESCRIPTION:	GDC v2.xx True Type Font Support
**
**	Copyright (C) 1997-2001, Matthew Allen
**		fret@ozemail.com.au
*/

#ifndef _GDCFONT_H_
#define _GDCFONT_H_

#include "string.h"
#include "GRegion.h"
#include "GProperties.h"

//////////////////////////////////////////////////////////////
// Defines

// Char sets
#define GDCF_ASCII						0	// Default charset
#define GDCF_ISO_8859_1					1	// West europe
#define GDCF_ISO_8859_2					2	// East europe
#define GDCF_ISO_8859_3					3
#define GDCF_ISO_8859_4					4	// Baltic
#define GDCF_ISO_8859_5					5	// Russian
#define GDCF_ISO_8859_6					6	// Arabic
#define GDCF_ISO_8859_7					7	// Greek
#define GDCF_ISO_8859_8					8	// Hebrew
#define GDCF_ISO_8859_9					9	// Turkish

// not implemented
#define GDCF_SHIFTJIS					10	// Japanese
#define GDCF_CHINESEBIG5				11	// Chinese
#define GDCF_THAI						12	// Thai
#define GDCF_MAC						13	// Apple Mac

#define GDCF_MAX						14

#ifndef WIN32

// font weights
#define FW_DONTCARE						0 
#define FW_THIN							100
#define FW_EXTRALIGHT					200
#define FW_ULTRALIGHT					200
#define FW_LIGHT						300
#define FW_NORMAL						400
#define FW_REGULAR						400
#define FW_MEDIUM						500
#define FW_SEMIBOLD						600
#define FW_DEMIBOLD						600
#define FW_BOLD							700
#define FW_EXTRABOLD					800
#define FW_ULTRABOLD					800
#define FW_HEAVY						900
#define FW_BLACK						900

// quality
#define DEFAULT_QUALITY					0
#define ANTIALIASED_QUALITY				1
#define NONANTIALIASED_QUALITY			2

#elif defined WIN32

#define WESTEUROPE_CHARSET				BALTIC_CHARSET		// ??? don't know

#endif

// OS specific font type
#if defined WIN32
typedef HFONT GFontRef;
#elif defined BEOS
typedef BFont *GFontRef;
#else
typedef void *GFontRef;
#endif

//////////////////////////////////////////////////////////////
// Functions

// CodePage is one of the GDCF_xxx defines
LgiFunc char *GdcGetCodePageDescription(int CodePage);
LgiFunc char *GdcGetCodePageMimeType(int CodePage);
LgiFunc bool GdcGetCodePageMappingTable(int CodePage, short *&Table); // Table is always an array of 128 shorts
LgiFunc bool GdcIsCodePageImplemented(int CodePage);

//////////////////////////////////////////////////////////////
// Classes
class GdcFontType;

class LgiClass GdcFont {
protected:
	// Type
	char *_Face;			// type face
	int _Height;			// point size
	int _Width;
	bool _Bold;
	bool _Italic;
	bool _Underline;
	int _CodePage;
	int _Quality;

	// Output
	COLOUR _Fore;
	COLOUR _Back;
	int _TabSize;
	bool _Transparent;

	// Methods
	virtual void _OnPropChange(bool c) {} // if false then it's just a property change

public:
	GdcFont();
	virtual ~GdcFont();

	// Set
	void Face(char *s);
	void Height(int i);
	void Width(int i);
	void CodePage(int i);
	void TabSize(int i);
	void Quality(int i);
	void Fore(COLOUR c);
	void Back(COLOUR c);
	void Bold(bool i);
	void Italic(bool i);
	void Underline(bool i);
	void Transparent(bool i);

	// Get
	char *Face();
	int Height();
	int Width();
	int CodePage();
	int TabSize();
	int Quality();
	COLOUR Fore();
	COLOUR Back();
	bool Bold();
	bool Italic();
	bool Underline();
	bool Transparent();

	// Virtuals
	virtual void Colour(COLOUR Fore, COLOUR Back = 0xFFFFFFFF);
};

class LgiClass GFont :
	public GdcFont
{
	GFontRef	hFont;
	int			SizeofChar;
	bool		Dirty;
	float		*Widths;

	bool IsValid();
	void _New();
	void _OnPropChange(bool Change);

public:
	GFont(char *face = 0, int point = -1);
	GFont(GFontRef Handle);
	GFont(GdcFontType &Type);
	GFont(GFont &Fnt);
	~GFont();

	// General
	bool Create(char *Face = 0, int Height = -1, int Param = 0);
	bool Create(GdcFontType *Type, int Param = 0);
	GFontRef Handle() { return hFont; }
	GFont &operator =(GFont &f);
	float *GetWidths();

	// 8-Bit
	void Text(	GSurface *pDC,
				int x, int y,
				char *Str,
				int Len = -1,
				GRegion *r = 0,
				int TabOrigin = 0);
	void Size(int *x, int *y, char *Str, int Len = -1, int Flags = 0);
	int X(char *Str, int Len = -1);
	int Y(char *Str, int Len = -1);
	int CharAt(int x, char *Str, int Len = -1);

	// Unicode
	void Text(	GSurface *pDC,
				int x, int y,
				ushort *Str,
				int Len = -1,
				GRegion *r = 0,
				int TabOrigin = 0);
	void Size(int *x, int *y, ushort *Str, int Len = -1, int Flags = 0);
	int X(ushort *Str, int Len = -1);
	int Y(ushort *Str, int Len = -1);
	int CharAt(int x, ushort *Str, int Len = -1);
};

class LgiClass GdcFontType
{
	friend class GFont;
	friend class GdcFont;

protected:
	#if defined WIN32
	LOGFONT Info;
	#else
	char Face[64];
	int PointSize;
	#endif

public:
	GdcFontType(char *face = 0, int pointsize = 0);
	virtual ~GdcFontType();

	// Put a user interface for the user to edit the font def
	bool DoUI(GView *Parent);
	
	// Describe the font to the user as a string
	bool GetDescription(char *Str);
	
	// Read/Write the font def to storage
	bool Serialize(ObjProperties *Options, char *OptName, bool Write);

	// Read the font from the LGI config
	bool GetConfigFont(char *Tag);
	
	// Read the font from LGI config (if there) and then the system settings
	bool GetSystemFont(char *Which);

	// Read from OS reference
	bool GetFromRef(GFontRef Handle);

	// Create a font based on this font def
	GFont *CreateFont(int Param = 0);
};
#endif
