/*hdr
**	FILE:		Region.h
**	AUTHOR:		Matthew Allen
**	DATE:		22/8/97
**	DESCRIPTION:	GDC Region classes
**
**	Copyright (C) 1997, Matthew Allen
**		fret@ozemail.com.au
*/

#ifndef __GDCREGION_H
#define __GDCREGION_H

#if defined WIN32

	#define CornerOffset 1
	typedef RECT OsRect;

#elif defined BEOS

	#define CornerOffset 0
	typedef BRect OsRect;

#elif defined ATHEOS

	#include <gui/rect.h>
	#define CornerOffset 0
	typedef os::Rect OsRect;

#endif

class LgiClass GRegion {

	friend LgiClass bool operator ==(GRegion &a, GRegion &b);
	friend LgiClass bool operator !=(GRegion &a, GRegion &b);

public:
	int x1, y1, x2, y2;

	GRegion() {}
	GRegion(int X1, int Y1, int X2, int Y2)
	{
		x1 = X1;
		x2 = X2;
		y1 = Y1;
		y2 = Y2;
	}

	GRegion(GRegion *r)
	{
		x1 = r->x1;
		x2 = r->x2;
		y1 = r->y1;
		y2 = r->y2;
	}

	int X() { return x2 - x1 + 1; }
	int Y() { return y2 - y1 + 1; }
	void Set(int X1, int Y1, int X2, int Y2)
	{
		x1 = X1;
		x2 = X2;
		y1 = Y1;
		y2 = Y2;
	}

	void ZOff(int x, int y);
	void Normal();
	bool Valid();
	void Offset(int x, int y);
	void Offset(GRegion *a);
	void Size(int x, int y);
	void Size(GRegion *a);
	void Dimension(int x, int y);
	void Dimension(GRegion *a);
	void Bound(GRegion *b);
	bool Overlap(int x, int y);
	bool Overlap(GRegion *b);
	void Union(int x, int y);
	void Union(GRegion *a);
	void Union(GRegion *a, GRegion *b);
	void Intersection(GRegion *a);
	void Intersection(GRegion *a, GRegion *b);

	operator OsRect()
	{
		OsRect r;

		r.left = x1;
		r.top = y1;
		r.right = x2+CornerOffset;
		r.bottom = y2+CornerOffset;

		return r;
	}

	GRegion &operator =(OsRect &r)
	{
		x1 = (int) r.left;
		y1 = (int) r.top;
		x2 = (int) r.right - CornerOffset;
		y2 = (int) r.bottom - CornerOffset;
		return *this;
	}

	GRegion(OsRect r)
	{
		x1 = (int) r.left;
		y1 = (int) r.top;
		x2 = (int) r.right - CornerOffset;
		y2 = (int) r.bottom - CornerOffset;
	}
};

LgiClass bool operator ==(GRegion &a, GRegion &b);
LgiClass bool operator !=(GRegion &a, GRegion &b);

class LgiClass GComplexRegion : public GRegion
{
	int Size;
	int Alloc;
	int Current;
	GRegion *a;

	bool SetSize(int s);
	GRegion *NewOne() { return (SetSize(Size+1)) ? a+(Size-1) : 0; }
	bool Delete(int i);
	void Empty();

public:
	GComplexRegion();
	GComplexRegion(int X1, int Y1, int X2, int Y2);
	GComplexRegion(GRegion &r);
	GComplexRegion(OsRect &r);
	~GComplexRegion();

	int X() { return x2 - x1 + 1; }
	int Y() { return y2 - y1 + 1; }
	int GetSize() { return Size; }
	GRegion *operator [](int i) { return a+i; }
	GRegion *First();
	GRegion *Last();
	GRegion *Next();
	GRegion *Prev();

	void ZOff(int x, int y);
	void Normal();
	bool Valid();
	void Offset(int x, int y);
	void Bound(GRegion *b);
	bool Overlap(GRegion *b);
	
	// Set operations
	void Union(GRegion *a);
	void Intersect(GRegion *a);
	void Subtract(GRegion *a);

	friend bool operator ==(GComplexRegion &a, GComplexRegion &b);
	friend bool operator !=(GComplexRegion &a, GComplexRegion &b);
};

#endif


