/*hdr
**      FILE:           TextView.h
**      AUTHOR:         Matthew Allen
**      DATE:           6/11/98
**      DESCRIPTION:    TextView header
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#ifndef __TEXTVIEW_H
#define __TEXTVIEW_H

#include "Text.h"

////////////////////////////////////////////////////////////////////////////////
#define TVF_DIRTY_BORDER		0x40000000
#define TVF_FOCUS				0x00002000
#define TABSTOPS				16

#define WM_STATUS_UPDATE		(WM_USER+0x200)
#define WM_TITLE				(WM_USER+0x201)

////////////////////////////////////////////////////////////////////////////////
extern char *Untitled;
extern char *FindMiss;

extern int CountLines(char *Str, int Len);
extern void StrCpyToCrLf(char *d, char *s, int Len);
extern void StrCpyToLf(char *d, char *s, int Len);

////////////////////////////////////////////////////////////////////////////////
class GTextView :
#ifdef BEOS
	public BControl,
#endif
	public GControl,
	public GFindReplace,
	public TextView
{

protected:
	#if defined WIN32

	static GWin32Class WndClass;
	char *Class();
	uint32 Style();

	#elif defined BEOS

	void AttachedToWindow() { OnCreate(); }
	void Draw(BRect UpdateRect) { GdcViewDC DC(_View); OnPaint(&DC); }
	void FrameMoved(BPoint Point) { Pos.Offset(Point.x-Pos.x1, Point.y-Pos.y1); OnPosChange(); }
	void FrameResized(float width, float height) { Pos.Dimension(width+1, height+1); OnPosChange(); }
	void Pulse() { OnPulse(); }
	void MessageReceived(BMessage *message) { OnEvent(message); }
	void MakeFocus(bool f = true) { BControl::MakeFocus(f); OnFocus(f); }
	void MouseDown(BPoint point) { Sys_MouseDown(point); }
	void MouseMoved(BPoint point, uint32 transit, const BMessage *message) { Sys_MouseMoved(point, transit, message); }
	void KeyDown(const char *bytes, int32 numBytes);
	void KeyUp(const char *bytes, int32 numBytes);

	GScrollView *ScrollView;
	bool Attach(GView *p);

public:
	bool SetPos(GRegion &p, bool Repaint = true);
	GRegion &GetPos();
	void Visible(bool i);

protected:
	#endif

	char	*AppName;
	int		Border;
	GFont	Font;

	GCursor	Old;
	int		TopMargin;
	int		LeftMargin;
	int		LineHt;
	int		Spacer;

	int		TabSize;
	int		*TabStops;
	int		WrapAtCol;
	bool	UrlDetect;

	int		ExtraFlags;
	GScrollBar *VScroll;
	GScrollBar *HScroll;
	bool	SetScrollBars(bool x, bool y);

	bool	MouseSelect;
	void	Title(char *s);
	void	SetStatus(int i, char *s);
	void	OutputText(GSurface *pDC, int x, int y, char *Str, int Len = -1);

public:
	// Data access
	TextDocument *GetDoc() { return &Doc; }

	// Construction
	GTextView(	int Id,
				int x,
				int y,
				int cx,
				int cy,
				GView *notify = 0,
				char *appname = 0,
				GFont *LogFont = 0);
	~GTextView();

	// API
	char *Name();
	bool Name(char *n);

	// Options
	void SetBorder(int b);
	void SetWrapAtCol(int i) { WrapAtCol = i; }
	int GetWrapAtCol() { return WrapAtCol; }
	void SetURLDetect(bool b) { UrlDetect = b; }
	bool GetURLDetect() { return UrlDetect; }
	bool AcceptEdits() { return Doc.AcceptEdits(); }
	void AcceptEdits(bool i) { Doc.AcceptEdits(i); }

	// Clipboard IO
	bool Cut();
	bool Copy();
	bool Paste();

	// Actions
	void Dirty(int Type);
	bool ClearDirty(bool Ask);
	void CalcCursorPos(GCursor c, int &x, int &y);
	void CalcMousePos(GCursor &c, GMouse m);
	void DrawCursor(GSurface *pDC, GCursor &c, COLOUR col);
	bool ClipText(char *Str, int Len = -1);
	char *ClipText();
	bool Open(char *Name);
	bool UpdateHiddenCheck();
	void UpdateScrollBars();
	bool DoFind();
	void SelectWord(GMouse &m);
	void SelectAll();
	void SelectNone();

	// After...
	void AfterInsertText(char *Text, int Len);
	void AfterDeleteText();

	// Object Events
	bool OnMoveCursor(int Dx, int Dy, bool NoSelect = FALSE);
	void OnSetCursor(int X, int Y);
	bool OnFind();
	bool OnReplace();
	bool OnMultiLineTab(bool In);
	void OnSetHidden(int Hidden);
	void OnPosChange();
	void OnCreate();
	void OnEscape(GKey &K);
	void OnMouseWheel(double Lines);
	virtual void OnUrl(char *Url);

	// Window Events
	void OnFocus(bool f);
	void OnMouseClick(GMouse &m);
	void OnMouseMove(GMouse &m);
	void OnKey(GKey &k);
	void OnPaint(GSurface *pDC);
	int OnEvent(GMessage *Msg);
	void OnPulse();
};


#endif