
#ifndef __GTEXTVIEW_3_H
#define __GTEXTVIEW_3_H

// Word wrap
#define TEXTED_WRAP_NONE			0	// no word wrap
#define TEXTED_WRAP_NONREFLOW		1	// insert LF when you hit the end
#define TEXTED_WRAP_REFLOW			2	// dynamically wrap line to editor width

// Code pages
#define TVCP_US_ASCII				0
#define TVCP_ISO_8859_2				1
#define TVCP_WIN_1250				2
#define TVCP_MAX					3

// Notify flags
#define GTVN_DOC_CHANGED			0x0001
#define GTVN_CURSOR_CHANGED			0x0002
#define GTVN_CODEPAGE_CHANGED		0x0004

// use CRLF as opposed to just LF
// internally it uses LF only... this is just to remember what to
// save out as.
#define TEXTED_USES_CR				0x00000001

#define TAB_SIZE					4

#define IsDigit(c)					((c) >= '0' AND (c) <= '9')
#define IsAlpha(c)					(((c) >= 'a' AND (c) <= 'z') OR ((c) >= 'A' AND (c) <= 'Z'))
#define IsText(c)					(IsDigit(c) OR IsAlpha(c) OR (c) == '_')

#define DEBUG_TIMES_MSG				8000 // a=0 b=(char*)Str

// Typedefs
typedef ushort						gchar;

// Externs
extern gchar Delimiters[];

// Classes
class GCursor
{
public:
};

class GTextStyle;

class GTextLine {
public:
	int Start;			// Start offset
	int Len;			// length of text
	GRegion r;		// Screen location

	virtual ~GTextLine() {}
	bool Overlap(int i)
	{
		return i>=Start AND i<=Start+Len;
	}
};

class GTextView : public GLayout, public GFindReplace
{
	friend class GTextStyle;

protected:

	enum GTextViewSeek
	{
		PrevLine,
		NextLine,
		StartLine,
 		EndLine
	};

	// Options
	int CodePage;
	int WrapAtCol;
	int WrapType;
	bool UrlDetect;
	bool AcceptEdit;
	bool CrLf;
	bool Dirty;
	bool AutoIndent;
	COLOUR ReadOnlyBackColour;

	// display
	GFont *Font;
	GFont *BlueUnderline;	// URL display
	int LineY;
	int Lines;
	int LeftMargin;
	int SelStart, SelEnd;
	bool Track;
	int DocOffset;

	List<GTextLine> Line;
	List<GTextStyle> Style;		// sorted in 'Start' order

	// data
	gchar *Text;
	int Cursor;
	int CharSize;
	int Size;
	int Alloc;

	// code page mapping cache
	char *MappedChars;
	short *MappedShorts;
	int MappedLen;
	char *MapChars(char *Source, int Len);
	short *MapShorts(char *Source, int Len);

	// debug
	int _PourTime;
	int _StyleTime;
	int _PaintTime;

	// private methods
	bool Insert(int At, gchar *Data, int Len);
	bool Delete(int At, int Len);
	int HitText(int x, int y);
	GTextLine *GetLine(int Offset, int *Index = 0);
	void DeleteSelection(gchar **Cut = 0);
	int SeekLine(int Start, GTextViewSeek Where);
	int TextWidth(GFont *f, gchar *s, int Len, int x, int Origin);
	
	// styles
	void InsertStyle(GTextStyle *s);
	GTextStyle *GetNextStyle(int Where = -1);
	GTextStyle *HitStyle(int i);

	// Overridables
	virtual void PourText();
	virtual void PourStyle();
	virtual void OnFontChange();

public:
	// Construction
	GTextView(	int Id,
				int x,
				int y,
				int cx,
				int cy,
				GView *notify = 0,
				char *appname = 0,
				GdcFontType *FontInfo = 0);
	~GTextView();

	// Data
	gchar *GetText();
	bool SetText(gchar *s);

	// Options
	void SetBorder(int b);
	void SetWrapAtCol(int i) { WrapAtCol = i; }
	int GetWrapAtCol() { return WrapAtCol; }
	void SetURLDetect(bool b) { UrlDetect = b; }
	bool GetURLDetect() { return UrlDetect; }
	bool AcceptEdits() { return AcceptEdit; }
	void AcceptEdits(bool i) { AcceptEdit = i; }
	int GetCodePage() { return CodePage; }
	void SetCodePage(int i);
	int GetWrapType() { return WrapType; }
	void SetWrapType(int i) { WrapType = i; }
	bool UseCrLf() { return CrLf; }
	void UseCrLf(bool i) { CrLf = i; }
	bool GetAutoIndent() { return AutoIndent; }
	void SetAutoIndent(bool b) { AutoIndent = b; }
	
	// State / Selection
	void SetCursor(int i, bool Select, bool ForceFullUpdate = false);
	bool IsDirty() { return Dirty; }
	bool HasSelection();
	void UnSelectAll();
	void SelectWord(int From);
	void SelectAll();
	GdcPt2 GetCursor();
	int GetLines();

	// File IO
	bool Open(char *Name);
	bool Save(char *Name);

	// Clipboard IO
	bool Cut();
	bool Copy();
	bool Paste();

	// Actions
	bool ClearDirty(bool Ask, char *FileName = 0);
	void UpdateScrollBars(bool Reset = false);
	bool DoFind();

	// Object Events
	bool OnFind();
	bool OnReplace();
	bool OnMultiLineTab(bool In);
	void OnSetHidden(int Hidden);
	void OnPosChange();
	void OnCreate();
	void OnEscape(GKey &K);
	void OnMouseWheel(double Lines);

	// Window Events
	void OnFocus(bool f);
	void OnMouseClick(GMouse &m);
	void OnMouseMove(GMouse &m);
	void OnKey(GKey &k);
	void OnPaint(GSurface *pDC);
	int OnEvent(GMessage *Msg);
	int OnNotify(GView *Ctrl, int Flags);
	void OnPulse();
	int OnHitTest(int x, int y);

	// Virtuals
	virtual void OnEnter(GKey &k) {}
	virtual void OnUrl(gchar *Url) {}
};

class GTextStyle {
public:
	// data
	class GTextView *View;
	int Start;
	int Len;
	GFont *Font;
	COLOUR c;

	// attach
	char *Data;

	virtual bool OnMouseClick(GMouse *m) { return false; }
	virtual bool OnMenu(GSubMenu *m) { return false; }
	virtual void OnMenuClick(int i) {}
	virtual char *GetCursor() { return 0; }
};

#endif