#ifndef __GTOOLBAR_H
#define __GTOOLBAR_H

////////////////////////////////////////////////////////////////////////
#define IMGLST_SELECTED				0x0001
#define IMGLST_DISABLED				0x0002
#define IMGLST_GDC					0x0004

class LgiClass GImageList : public GMemDC
{
	#ifdef WIN32
	HIMAGELIST hImgList;
	#else if BEOS
	GSurface *Mask;
	#endif
	GSurface *pCreate;
	int Sx, Sy;

public:
	GImageList(int x, int y, GSurface *pDC = NULL);
	~GImageList();

	#ifdef WIN32
	operator const HIMAGELIST () { return hImgList; }
	HIMAGELIST Handle() { return hImgList; }
	#endif
	
	int TileX() { return Sx; }
	int TileY() { return Sy; }
	int GetItems() { return X() / Sx; }

	bool Create(int x, int y, int Bits, int LineLen = 0, bool KeepData = FALSE);
	void Update(int Flags);
	void Draw(GSurface *pDest, int x, int y, int Image, int Flags = 0);
};

///////////////////////////////////////////////////////////////////////////////////////////////
// Toolbar stuff

// Tool button types
#define TBT_PUSH				0
#define TBT_RADIO				1
#define TBT_TOGGLE				2

// Commands
#define IDM_NONE				0
#define IDM_SEPARATOR			-1
#define IDM_BREAK				-2

// Dimensions
#define BORDER_SHADE			1
#define BORDER_SPACER			1
#define BORDER_SEPARATOR		4
#define BORDER_BUTTON			1

class LgiClass GToolButton :
	public GView
{
	friend class GToolBar;

protected:
	int			Type;		// Button Type
	bool		Clicked;
	bool		Down;
	bool		Over;
	int			ImgIndex;
	int			TipId;

	virtual	void SetDown(bool d);

public:
	GToolButton(int Bx, int By);
	~GToolButton();

	int Value() { return Down; }
	void Value(int i);
	int Image() { return ImgIndex; }
	void Image(int i) { ImgIndex = i; }
	bool GetDown() { return Down; }
	int GetType() { return Type; }
	void SetType(int i) { Type = i; }
	bool Separator() { return GetId() == IDM_SEPARATOR; }
	void Separator(bool i) { SetId(IDM_SEPARATOR); }

	void OnPaint(GSurface *pDC);

	void OnMouseClick(GMouse &m);
	void OnMouseMove(GMouse &m);
	void OnMouseEnter(GMouse &m);
	void OnMouseExit(GMouse &m);

	virtual void OnCommand(GMouse &m);
	virtual void SendNotify(int Flags);
	virtual bool GetDimension(int &x, int &y) { return FALSE; }
};

class LgiClass GToolBar : public GLayout
{
	friend class GToolButton;

protected:
	int Bx, By;
	int Sx, Sy;
	bool Vertical;
	bool Text;
	int LastIndex;
	bool OwnImgList;
	GImageList *ImgList;
	GFont *Font;

	// bitmap cache
	GSurface *pColour;
	GSurface *pDisabled;

	void _BuildCache(GImageList *From);
	void _DrawFromCache(GSurface *pDC, int x, int y, int Index, bool Disabled);

	#if defined WIN32
	HWND hTip;
	#endif
	virtual int PostDescription(GView *Ctrl, char *Text);

public:
	GToolBar();
	~GToolBar();

	// Overridables
	virtual void OnButtonClick(GToolButton *Btn);

	// Properties
	bool IsVertical() { return Vertical; }
	void IsVertical(bool v) { Vertical = v; }
	bool TextLabels() { return Text; }
	void TextLabels(bool i) { Text = i; }
	bool SetBitmap(char *File, int Bx, int By);
	bool SetDC(GSurface *pDC, int Bx, int By);
	bool SetImageList(GImageList *l, int Bx, int By, bool Own = true);

	// Buttons
	GToolButton *AppendButton(char *Tip, int Id, int Type = TBT_PUSH, int Enabled = TRUE, int IconId = -1);
	bool AppendControl(GView *Ctrl);
	bool AppendSeparator();
	bool AppendBreak();
	void Empty();

	// Events
	int OnEvent(GMessage *Msg);
	void OnPaint(GSurface *pDC);
	void OnMouseClick(GMouse &m);
	void OnMouseEnter(GMouse &m);
	void OnMouseExit(GMouse &m);
	void OnMouseMove(GMouse &m);

	// Impl
	bool Pour(GComplexRegion &r);
};

#endif
