#ifndef __GTREE2_H
#define __GTREE2_H

#include "GList2.h"

// Tree view
class LgiClass GTreeItem :
	public GItem
{
	friend class GTree;

protected:
	// Private data
	GRegion _Pos;
	GRegion _Thumb;
	GRegion _Text;
	GRegion _Icon;
	bool _Open;
	bool _Selected;
	bool _Visible;
	bool _Last;
	int _Depth;

	// Private methods
	void _Pour(GdcPt2 *Limit, int Depth, bool Visible);
	void _Remove();
	void _MouseClick(GMouse &m);
	GTreeItem *_HitTest(int x, int y);

	virtual void _PourText(GdcPt2 &Size);
	virtual void _PaintText(GSurface *pDC, COLOUR Fore, COLOUR Back);

protected:
	char *Str;
	int Sys_Image;
	GTree *Tree;
	GTreeItem *Parent;
	List<GTreeItem> Items;

public:
	GTreeItem();
	virtual ~GTreeItem();

	// Properties
	char *GetText(int i=0);
	bool SetText(char *s, int i=0);
	int GetImage(int Flags = 0);
	void SetImage(int i);
	void Update();
	bool Selected();
	void Selected(bool b);
	bool Expanded();
	void Expanded(bool b);

	// List
	GTreeItem *Insert(GTreeItem *Obj = NULL, int Pos = -1);
	void Remove();
	GTreeItem *GetNext();
	GTreeItem *GetPrev();
	GTreeItem *GetChild();
	GTreeItem *GetParent() { return Parent; }
	int IndexOf();

	// State-Events
	virtual void OnExpand(bool b) {}

	// Impl
	void OnMouseClick(GMouse &m);
	void OnPaint(GSurface *pDC, GRegion &r, COLOUR Back);
};

class LgiClass GTree :
#if defined LGI_RES
	public ResObject,
#endif
	public GLayout,
	public GItemContainer,
	public GTreeItem
{
	friend class GTreeItem;

	#if defined WIN32
	static GWin32Class WndClass;
	#endif

	// Private data
	int _LineFlags[4];
	bool _LayoutDirty;
	GdcPt2 _Limit;
	GdcPt2 _LastClick;
	GTreeItem *_LastHit;
	List<GTreeItem> _Selection;
	GTreeItem *_DropTarget;
	GSurface *_IconCache;

	// Private methods
	void _Pour();
	void _OnSelect(GTreeItem *Item);
	GdcPt2 _ScrollPos();
	void _Update(GRegion *r = 0, bool Now = false);
	void _UpdateBelow(int y, bool Now = false);
	void _UpdateScrollBars();

protected:
	// Options
	bool Lines;
	bool Buttons;
	bool LinesAtRoot;
	bool EditLabels;
	bool MultipleSelect;
	
public:
	GTree(int id, int x, int y, int cx, int cy, char *name = 0);
	~GTree();

	// Events
	virtual void OnItemClick(GTreeItem *Item, GMouse &m);
	virtual void OnItemBeginDrag(GTreeItem *Item, int Flags);
	virtual void OnItemExpand(GTreeItem *Item, bool Expand);
	virtual void OnItemSelect(GTreeItem *Item);
	
	// Window
	void OnMouseClick(GMouse &m);
	void OnMouseMove(GMouse &m);
	void OnMouseWheel(double Lines);
	void OnPaint(GSurface *pDC);
	void OnFocus(bool b);
	void OnPosChange();
	void OnKey(GKey &k);
	int OnNotify(GView *Ctrl, int Flags);
	int OnEvent(GMessage *Msg);

	// Adding/deleting
	GTreeItem *Insert(GTreeItem *Obj = 0, int Pos = -1);
	bool Delete(GTreeItem *Obj);
	bool Remove(GTreeItem *Obj);
	GTreeItem *ItemAt(int Pos) { return Items.ItemAt(Pos); }

	// Selection/Items
	bool Select(GTreeItem *Obj);
	GTreeItem *Selection();
	GTreeItem *ItemAtPoint(int x, int y);
	void SelectDropTarget(GTreeItem *Item);

	// Container
	void Empty();		// delete all items
	void RemoveAll();	// remove reference to externally owned items

	// Impl
};

#endif