#ifndef __IFTP_H
#define __IFTP_H

#include "INet.h"

#define IFTP_READ			0x0400
#define IFTP_WRITE			0x0200
#define IFTP_EXECUTE		0x0100
#define IFTP_DIR			0x0800

#define IFTP_GRP_READ		0x0040
#define IFTP_GRP_WRITE		0x0020
#define IFTP_GRP_EXECUTE	0x0010

#define IFTP_GLOB_READ		0x0004
#define IFTP_GLOB_WRITE		0x0002
#define IFTP_GLOB_EXECUTE	0x0001

#define IFTP_ARCHIVE		0x1000
#define IFTP_HIDDEN			0x2000
#define IFTP_SYSTEM			0x4000

#define IFTP_SYM_LINK		0x8000

class IFtpDataPipe
{
public:
	virtual int WritePipe(GSocket *c, char *Data, int Len, int Flags) = 0;
	virtual int ReadPipe(GSocket *c, char *Data, int Len, int Flags) = 0;
};

class IFtpEntry
{
public:
	int Attributes;
	int Size;
	char *Name;
	char *Date;
	char *Time;

	IFtpEntry();
	IFtpEntry(char *Entry);
	IFtpEntry(IFtpEntry *Entry);
	virtual ~IFtpEntry();

	bool IsDir() { return (Attributes & IFTP_DIR) != 0; }
};

class IFtp
{
	// Sockets...
	GSocket *Socket;	// commands
	GSocket *Listen;	// listen for data
	GSocket *Data;		// get data
	GFile *F;

	char OutBuf[256];
	char InBuf[256];
	List<char> In;

	int WriteLine(char *Msg = 0);
	int ReadLine(char *Msg = 0);
	bool TransferFile(char *Local, char *Remote, int RemoteSize, bool Upload);

	// Data connections
	char Ip[64];
	int Port;
	bool PassiveMode;
	bool ForceActive;
	IFtpDataPipe *Pipe;

	// State
	int RestorePos;
	bool AbortTransfer;
	
	bool SetupData();
	bool ConnectData();

public:
	Progress *Meter;
	bool Authenticated;

	IFtp(IFtpDataPipe *OptionalPipe = 0);
	virtual ~IFtp();

	// Properties
	bool IsForceActive() { return ForceActive; }
	void IsForceActive(bool i) { ForceActive = i; }

	// Data
	GSocket *Handle() { return Socket; }
	
	// Connection
	bool Open(GSocket *S, char *RemoteHost, int Port, char *User, char *Password);
	bool Close();
	bool IsOpen();
	void Noop();

	// Directory
	bool GetDir(char *Dir);
	bool SetDir(char *Dir);
	bool CreateDir(char *Dir);
	bool DeleteDir(char *Dir);
	bool ListDir(List<IFtpEntry> *Dir);
	bool UpDir();
	
	// File
	bool DeleteFile(char *Remote);
	bool DownloadFile(char *Local, IFtpEntry *Remote);
	bool UploadFile(char *Local, char *Remote);
	bool RenameFile(char *From, char *To);
	bool SetPerms(char *File, int Perms);
	bool ResumeAt(int Pos);
	void Abort() { AbortTransfer = true; }
};

#endif