/*hdr
**      FILE:           Inet.h
**      AUTHOR:         Matthew Allen
**      DATE:           28/5/98
**      DESCRIPTION:    Internet sockets classes
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#ifndef __INET_H
#define __INET_H

#include "VMem.h"
#include "Contain.h"
#include "LgiNetInc.h"

#if defined WIN32

	#include <winsock.h>

#elif defined BEOS

	#include <socket.h>
	#include <netdb.h>

#elif defined ATHEOS

	#include <atheos/socket.h>

#else

typedef int SOCKET;

#endif

#define DEFAULT_TIMEOUT			30

// Parameters
#define PARAMETER_LOG			1
#define PARAMETER_PROGRESS		2

#if defined BEOS

	#include "NetworkKit.h"
	#define ValidSocket(s)			((s)>=0)
	#define INVALID_SOCKET			0
	typedef int SOCKET;

#elif defined ATHEOS

	#define ValidSocket(s)			((s)>=0)
	#define INVALID_SOCKET			0
	typedef int SOCKET;

#elif defined WIN32

	#define ValidSocket(s)			((s) != INVALID_SOCKET)

#endif

// Logging types..
enum
{
	NET_LOG_NONE = 0,
	NET_LOG_HEX_DUMP = 1,
	NET_LOG_ALL_BYTES = 2
};

// Functions
LgiNetFunc bool HaveNetConnection();
LgiNetFunc bool WhatsMyIp(char *IpAddr);
LgiNetFunc void MDStringToDigest(unsigned char digest[16], char *Str);

// Classes
class GNetwork;
class GSocket;

class LgiNetClass GNetwork {

	bool SocketsOpen;

public:
	GNetwork();
	virtual ~GNetwork();

	bool IsValid() { return SocketsOpen; }

	GSocket *CreateConnection();
	bool EnumInterfaces(List<char> &Lst);
};

class LgiNetClass GSocket {

	friend class GNetwork;

protected:
	// Data
	GNetwork	*Parent;
	bool		Async;
	bool		Canceled;
	int			LogType;
	char		*LogFile;

	#if defined WIN32
	SOCKET		Socket;
	#elif defined BEOS
	BNetEndpoint *Ep;
	#endif

	// Methods
	bool IsOK() { return (Parent) ? Parent->IsValid() : false; }
	void Log(char *Msg, int Ret, char *Buf, int Len);

public:
	int			BytesWritten;
	int			BytesRead;

	GSocket();
	virtual ~GSocket();
	SOCKET Handle();

	// Status
	bool IsReadable();
	bool IsOpen() { return ValidSocket(Handle()); }
	virtual bool Cancel() { return Canceled; }
	virtual void Cancel(bool i) { Canceled = i; }
	virtual bool IsAsync();
	virtual void IsAsync(bool i);

	// Client
	virtual bool Open(char *HostAddr, int port);
	virtual void Close();

	// Server
	virtual bool Listen(int Port = 0);
	virtual bool Accept(GSocket *c);

	// Data transfer
	virtual int Write(char *Data, int Len, int Flags);
	virtual int Read(char *Data, int Len, int Flags);
	virtual int Error();

	// Info
	bool WhatsMyIp(char *IpAddr);
	int WhatsMyPort();
	bool GetRemoteIp(char *IpAddr);
	void SetLogFile(char *FileName = 0, int Type = NET_LOG_HEX_DUMP);

	// Events
	virtual void OnDisconnect();

	virtual void OnRead(char *Data, int Len) {}
	virtual void OnWrite(char *Data, int Len) {}
	virtual void OnError(int ErrorCode, char *ErrorDescription) {}
	virtual void OnInformation(char *Str) {}
	virtual void SetParameter(int Which, int What) {}
};

class LgiNetClass GSocks5Socket : public GSocket
{
	char *Proxy;
	int Port;
	char *UserName;
	char *Password;

protected:
	bool Socks5Connected;

public:
	GSocks5Socket(char *proxy, int port, char *username, char *password);
	~GSocks5Socket();

	// Connection
	bool Open(char *HostAddr, int port);

	// Server
	bool Listen(int Port) { return false; }
};

#endif

