/*hdr
**	FILE:			LgiClasses.h
**	AUTHOR:		Matthew Allen
**	DATE:			19/12/97
**	DESCRIPTION:	LGI Class definitions
**	PLATFROM:		Xp
**
**
**	Copyright (C) 1997-2001, Matthew Allen
**		fret@ozemail.com.au
*/

/////////////////////////////////////////////////////////////////////////////////////
// Includes
#ifndef __GUI_H
#define __GUI_H

#if defined BEOS

	#include <string.h>

#endif

#include "GSemaphore.h"
#include "GLibrary.h"
#include "GXml.h"

/////////////////////////////////////////////////////////////////////////////////////
// Externs
extern long MouseWatcher(void *Ptr);
extern bool _lgi_check_file(char *Path);

/////////////////////////////////////////////////////////////////////////////////////
// Classes
#if defined BEOS

	class GScrollView;

	class LgiClass BViewRedir : public BView
	{
		void AttachedToWindow();
		void DetachedFromWindow();
		void Draw(BRect UpdateRect);
		void FrameMoved(BPoint Point);
		void FrameResized(float width, float height);
		void Pulse();
		void MessageReceived(BMessage *message);
		void MakeFocus(bool f = true);
		void KeyDown(const char *bytes, int32 numBytes);
		void KeyUp(const char *bytes, int32 numBytes);
		void MouseDown(BPoint point);
		void MouseMoved(BPoint point, uint32 transit, const BMessage *message);
		bool QuitRequested();
	
		GView *Wnd;
		GScrollView *ScrollView;
		
	public:
		BViewRedir(GView *Wnd);

		GView *GetWindow() { return Wnd; }

		GScrollView *GetScrollView() { return ScrollView; }
		void SetScrollView(GScrollView *v) { ScrollView = v; }
	};

#elif defined ATHEOS

	template <class a>
	class LgiClass AView : public a
	{
		GView *View;

		void _New()
		{
		}

	public:

		/*
		os::Button(Rect cFrame,
					const char *pzName,
					const char *pzLabel,
					Message *pcMessage,
					uint32 nResizeMask,
					uint32 nFlags)

		os::TextView(const Rect &cFrame,
					const char *pzTitle,
					const char *pzBuffer,
					uint32 nResizeMask,
					uint32 nFlags);
		*/

		AView(GView *v) :
			a(os::Rect(0, 0, 1, 1), std::string(""))
		{
			printf("AView contructor 1\n");
			View = v;
			_New();
		}

		AView(GView *v, char *Name, char *Label) :
			a(os::Rect(0, 0, 1, 1), Name, Label)
		{
			printf("AView contructor 2\n");
			View = v;
			_New();
		}

		AView(GView *v, char *Name, char *Label, os::alignment Align) :
			a(os::Rect(0, 0, 1, 1), Name, Label, Align)
		{
			printf("AView contructor 3\n");
			View = v;
			_New();
		}

		AView(	GView *v,
				const std::string &Name,
				const std::string &Label) :
			a(os::Rect(0, 0, 1, 1), Name, Label)
		{
			printf("AView contructor 4\n");
			View = v;
			_New();
		}

		AView(	GView *v,
				const char *Name,
				const char *Label,
				os::Message *Msg,
				int Resize = os::CF_FOLLOW_LEFT | os::CF_FOLLOW_TOP) :
			a(os::Rect(0, 0, 1, 1), Name, Label, Msg, Resize)
		{
			printf("AView contructor 5\n");
			View = v;
			_New();
		}

		AView(	GView *v,
				const std::string &Name,
				const std::string &Label,
				os::Message *Msg,
				int Resize = os::CF_FOLLOW_LEFT | os::CF_FOLLOW_TOP) :
			a(os::Rect(0, 0, 1, 1), Name, Label, Msg, Resize)
		{
			printf("AView contructor 6\n");
			View = v;
			_New();
		}

		AView(	GView *v,
				const char *Name,
				os::Message *Msg) :
			a(os::Rect(0, 0, 1, 1), Name, Msg, 0.0, 1.0)
		{
			printf("AView contructor 7\n");
			View = v;
			_New();
		}

		// From os::Handler
		void HandleMessage(os::Message *Msg)
		{
			View->OnEvent(Msg);
			a::HandleMessage(Msg);
		}

		// From os::View
		void AttachedToWindow()
		{
			View->OnCreate();
			a::AttachedToWindow();
		}

		void DetachedFromWindow()
		{
			View->OnCreate();
			a::DetachedFromWindow();
		}

		void Activated(bool b)
		{
			View->OnFocus(b);
			a::Activated(b);
		}

		void Paint(const os::Rect &UpdateRect)
		{
			GdcViewDC DC(this);
			View->_Paint(&DC);
		}

		void MouseMove(const os::Point &NewPos, int Code, uint32 Buttons, os::Message *Msg)
		{
			GMouse m;
			m.x = (int) NewPos.x;
			m.y = (int) NewPos.y;
			m.Flags = Buttons;
			View->OnMouseMove(m);
		}

		void MouseDown(const os::Point &Pos, uint32 Buttons)
		{
			GMouse m;
			m.x = (int) Pos.x;
			m.y = (int) Pos.y;
			m.Flags = Buttons;
			m.Down(true);
			View->OnMouseMove(m);
		}

		void MouseUp(const os::Point &Pos, uint32 Buttons, os::Message *Msg)
		{
			GMouse m;
			m.x = (int) Pos.x;
			m.y = (int) Pos.y;
			m.Flags = Buttons;
			m.Down(false);
			View->OnMouseClick(m);
		}

		void KeyDown(const char *String, const char *RawString, uint32 Qualifiers) {}
		void KeyUp(const char *String, const char *RawString, uint32 Qualifiers) {}

		void FrameMoved(const os::Point &Delta)
		{
			View->OnPosChange();
		}

		void FrameSized(const os::Point &Delta)
		{
			View->OnPosChange();
		}

		void ViewScrolled(const os::Point &Delta) {}
		void WheelMoved(const os::Point &Delta) {}

		// From os::Control
		void EnableStatusChanged(bool b)
		{
		}
	};

#endif

#ifdef BEOS

	// GKey flags
	#define LGI_VKEY_CTRL		B_CONTROL_KEY	// 0x0004
	#define LGI_VKEY_ALT		B_OPTION_KEY	// 0x0040
	#define LGI_VKEY_SHIFT		B_SHIFT_KEY		// 0x0001

	// GMouse flags
	#define LGI_VMOUSE_LEFT		B_PRIMARY_MOUSE_BUTTON		// 0x0001
	#define LGI_VMOUSE_MIDDLE	B_TERTIARY_MOUSE_BUTTON		// 0x0004
	#define LGI_VMOUSE_RIGHT	B_SECONDARY_MOUSE_BUTTON	// 0x0002

	#define LGI_VMOUSE_CTRL		0x08
	#define LGI_VMOUSE_ALT		0x10
	#define LGI_VMOUSE_SHIFT	0x20

	#define LGI_VMOUSE_DOWN		0x40
	#define LGI_VMOUSE_DOUBLE	0x80

#else

	// GKey flags
	#define LGI_VKEY_CTRL		MK_CONTROL
	#define LGI_VKEY_ALT		MK_ALT
	#define LGI_VKEY_SHIFT		MK_SHIFT

	// GMouse flags
	#define LGI_VMOUSE_LEFT		MK_LBUTTON		// 0x0001
	#define LGI_VMOUSE_MIDDLE	MK_MBUTTON		// 0x0010
	#define LGI_VMOUSE_RIGHT	MK_RBUTTON		// 0x0002

	#define LGI_VMOUSE_SHIFT	MK_SHIFT		// 0x0004
	#define LGI_VMOUSE_CTRL		MK_CONTROL		// 0x0008
	#define LGI_VMOUSE_ALT		0x20

	#define LGI_VMOUSE_DOWN		0x40
	#define LGI_VMOUSE_DOUBLE	0x80

#endif

class LgiClass GKey {
public:
	uchar c;
	int Flags;
	int Data;
	bool Down;
	bool IsChar;

	// Keyboard modifiers
	bool Ctrl()		{ return TestFlag(Flags, LGI_VKEY_CTRL); }
	bool Alt()		{ return TestFlag(Flags, LGI_VKEY_ALT); }
	bool Shift()	{ return TestFlag(Flags, LGI_VKEY_SHIFT); }
};

class LgiClass GMouse {
public:
	int x;
	int y;
	int Flags;

	// Get
	bool Down()		{ return TestFlag(Flags, LGI_VMOUSE_DOWN); }
	bool Double()	{ return TestFlag(Flags, LGI_VMOUSE_DOUBLE); }
	bool Left()		{ return TestFlag(Flags, LGI_VMOUSE_LEFT); }
	bool Middle()	{ return TestFlag(Flags, LGI_VMOUSE_MIDDLE); }
	bool Right()	{ return TestFlag(Flags, LGI_VMOUSE_RIGHT); }
	bool Ctrl()		{ return TestFlag(Flags, LGI_VMOUSE_CTRL); }
	bool Alt()		{ return TestFlag(Flags, LGI_VMOUSE_ALT); }
	bool Shift()	{ return TestFlag(Flags, LGI_VMOUSE_SHIFT); }

	// Set
	#define AssignFlag(f, bit, to) ( (to) ) ? f |= bit : f &= ~(bit)
	void Down(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_DOWN, i); }
	void Double(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_DOUBLE, i); }
	void Left(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_LEFT, i); }
	void Middle(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_MIDDLE, i); }
	void Right(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_RIGHT, i); }
	void Ctrl(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_CTRL, i); }
	void Alt(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_ALT, i); }
	void Shift(bool i)	{ AssignFlag(Flags, LGI_VMOUSE_SHIFT, i); }
};

class LgiClass GTarget {
public:
	virtual int OnEvent(GMessage *Msg) { return 0; }
};

class LgiClass GEvent {
public:
	// Mouse related
	virtual void OnMouseClick(GMouse &m) {}
	virtual void OnMouseEnter(GMouse &m) {}
	virtual void OnMouseExit(GMouse &m) {}
	virtual void OnMouseMove(GMouse &m) {}

	// Keyboard related
	virtual void OnKey(GKey &k) {}

	// Other
	virtual void OnPaint(GSurface *pDC) {}
	virtual void OnPosChange() {}
};

class LgiClass GFlags
{
protected:
	int WndFlags;

public:
	GFlags();

	// State
	virtual bool Enabled();
	virtual void Enabled(bool i);
	virtual bool Visible();
	virtual void Visible(bool i);
	virtual bool Focus();
	virtual void Focus(bool i);
	virtual bool DropTarget();
	virtual void DropTarget(bool t);

	virtual void OnUpdateState() {}

	// Border
	virtual bool Sunken();
	virtual void Sunken(bool i);
	virtual bool Flat();
	virtual void Flat(bool i);
	virtual bool Raised();
	virtual void Raised(bool i);

	virtual void OnUpdateBorder() {}
};

#if defined WIN32

// Win32 window class
class LgiClass GWin32Class : public GObject
{
	static LRESULT CALLBACK Redir(OsView hWnd, UINT m, WPARAM a, LPARAM b);
	static LRESULT CALLBACK SubClassRedir(OsView hWnd, UINT m, WPARAM a, LPARAM b);
	WNDPROC ParentProc;

public:
	WNDCLASS Class;

	GWin32Class();
	~GWin32Class();

	bool Create(char *ClassName);
	bool SubClass(char *ClassName, char *Parent);
	LRESULT CALLBACK CallParent(OsView hWnd, UINT m, WPARAM a, LPARAM b);
};

#endif

/////////////////////////////////////////////////////////////////////////////////
#define LgiApp						(GApp::ObjInstance())
#define SysFont						(LgiApp->SystemFont)
#define hAppInst()					(LgiApp?LgiApp->Instance():0)
#define LgiExitApp()					LgiApp->Exit(true)
#define LgiCloseApp()				LgiApp->Exit(false)

class LgiClass GApp
#if defined BEOS
	: public BApplication
#elif defined ATHEOS
	: public os::Application
#endif
{
	friend class GView;
	#ifdef ATHEOS
	friend bool LgiGetExeFile(char *Dst);
	#endif

protected:
	// private member vars
	static GApp *TheApp;

	char *CmdLine;
	Xml *Config;

	#if defined WIN32

	HINSTANCE hAppInst;
	int CmdShow;
	HCURSOR hNormalCursor;

	#elif defined ATHEOS

	char *_AppFile;

	#elif defined BEOS

	void RefsReceived(BMessage *Msg);

	#endif

public:
	// public member vars
	#ifdef WIN32
	static OsView ActiveDialog;
	#endif
	GFont *SystemFont;
	GWindow *AppWnd;

	// public methods
	#if defined WIN32

	GApp(HINSTANCE hInst, char *lpCmdLine, int nShow);

	HINSTANCE Instance() { return hAppInst; }
	int Show() { return CmdShow; }

	#else

	GApp(char *AppMime, int Args = 0, char **Arg = 0);

	#endif

	virtual ~GApp();

	// Msg loop
	virtual bool Run(bool Loop = true);

	// The one and only instance
	static GApp *ObjInstance() { return TheApp; }

	// Options etc
	bool GetOption(char *Dst, char Option);
	XmlTag *GetConfig(char *Tag);

	// Override
	virtual void OnCmdLine(char *Cmd);
	virtual void OnReceiveFiles(List<char> &Files);
	virtual void Exit(int Code = 0);
};

class LgiClass GView :
	public GObject,
	public GTarget,
	public GEvent,
	public GFlags
{
	friend		class GWindow;
	friend		class GLayout;
	friend		class GDialog;
	friend		class GControl;
	friend		class GMenu;
	friend		class GSubMenu;
	friend		class GWnd;
	friend		class GScrollView;

	friend		bool SysOnKey(GView *w, GMessage *m);
	friend		GView *GetNextTabStop(GView *Wnd, bool Back);

	#ifdef WIN32

	friend		LRESULT CALLBACK GWin32Class::Redir(OsView hWnd, UINT m, WPARAM a, LPARAM b);
	friend		LRESULT CALLBACK GWin32Class::SubClassRedir(OsView hWnd, UINT m, WPARAM a, LPARAM b);
	friend		LRESULT CALLBACK DlgRedir(OsView hWnd, UINT m, WPARAM a, LPARAM b);
	static		void CALLBACK TimerProc(OsView hwnd, UINT uMsg, UINT idEvent, uint32 dwTime);

	#else if BEOS

	friend		class GButtonRedir;
	friend		class _OsEditFrame;
	friend		class BViewRedir;
	friend		long _lgi_pulse_thread(void *ptr);

	#endif

protected:
	OsView _View; // OS specific handle to view object

	#if defined WIN32

	// Tracking handles
	static GView	*Capturing;
	static GView	*Over;
	static OsView	hPrevCapture;

	// Window data
	int				WndStyle;				// Win32 hWnd Style
	int				WndExStyle;				// Win32 hWnd ExStyle
	int				WndDlgCode;				// Win32 Dialog Code (WM_GETDLGCODE)
	char			*WndClass;
	UINT			_Win32_TimerId;

	// Methods
	virtual uint32	Style();
	virtual uint32	ExStyle();
	virtual char	*Class();
	virtual GView	*FindReal(GdcPt2 *Offset = 0);
	virtual bool	HandleCapture(GView *Wnd, bool c);

	// Stubs
	virtual int		SysOnNotify(int Code) { return 0; }

	// Create a Win32 window
	virtual bool	Create(char *Class, OsView Parent);

	#elif defined BEOS

	struct OsMouseInfo;
	friend long _lgi_mouse_thread(OsMouseInfo *Info);

	OsMouseInfo		*_MouseInfo;
	OsThread		_CaptureThread;
	OsThread		_PulseThread;
	int				_PulseRate;

	void Sys_KeyDown(const char *bytes, int32 numBytes);
	void Sys_KeyUp(const char *bytes, int32 numBytes);
	void Sys_MouseDown(BPoint point);
	void Sys_MouseMoved(BPoint point, uint32 transit, const BMessage *message);
	void Sys_FrameMoved(BPoint p);
	virtual bool QuitRequested() {}

	#endif

	// General
	int				CtrlId;
	GRegion			Pos;
	GView			*Parent;
	GView			*Notify;
	GView			*WantPulses;
	GdcPt2			_MinimumSize;
	int				_BorderSize;

	// Complex Region searches
	GRegion *FindLargest(GComplexRegion &r);
	GRegion *FindSmallestFit(GComplexRegion &r, int Sx, int Sy);
	GRegion *FindLargestEdge(GComplexRegion &r, int Edge);

#ifdef ATHEOS
public:
#endif
	// Impl
	virtual void _Paint(GSurface *pDC);

public:
	List<GView>	Children;

	#ifdef WIN32

	GView();

	#else // BEOS || ATHEOS

	GView(OsView wnd);

	#endif

	virtual ~GView();

	OsView Handle() { return _View; }
	operator OsView() { return _View; }

	#ifdef BEOS

	virtual OsWindow GetWindow() { return (_View) ? _View->Window() : 0; }

	#elif defined ATHEOS

	virtual OsWindow GetWindow() { return (_View) ? _View->GetWindow() : 0; }

	#endif

	// Attaching windows / heirarchy
	virtual bool Attach(GView *p);
	virtual bool AttachChildren();
	virtual bool Detach();
	virtual bool IsAttached();
	virtual void Quit(bool DontDelete = false);

	virtual GView *GetParent() { return Parent; }
	virtual void SetParent(GView *p) { Parent = p; }
	virtual GView *GetNotify() { return Notify; }		// the "notify" window gets told
	virtual void SetNotify(GView *n) { Notify = n; }	// via the OnNotify event handler
														// when this window has changed
	// Target
	int OnEvent(GMessage *Msg);							// generic message handler

	// Flags
	bool Enabled();
	void Enabled(bool e);
	bool Visible();
	void Visible(bool v);
	bool Focus();
	void Focus(bool f);

	// Border
	bool Sunken();
	void Sunken(bool i);
	bool Flat();
	void Flat(bool i);
	bool Raised();
	void Raised(bool i);

	// Thread
	bool InThread(); // return true if the calling thread owns the window
	bool PostEvent(int Cmd, int a, int b);

	// Non-implemented events
	virtual void OnMouseClick(GMouse &m) {}
	virtual void OnMouseEnter(GMouse &m) {}
	virtual void OnMouseExit(GMouse &m) {}
	virtual void OnMouseMove(GMouse &m) {}
	virtual void OnKey(GKey &k) {}
	virtual void OnMouseWheel(double Lines) {}
	virtual void OnCreate() {}
	virtual void OnDestroy() {}
	virtual void OnFocus(bool f) {}
	virtual void OnPulse() {}
	virtual void OnPosChange();
	virtual bool OnRequestClose() { return true; }
	virtual int OnHitTest(int x, int y) { return -1; }
	virtual int OnNotify(GView *Ctrl, int Flags);
	virtual void OnChildrenChanged(GView *Wnd, bool Attaching) {}
	virtual int OnCommand(int Cmd, int Event, OsView Wnd) { return 0; }
	virtual void OnPaint(GSurface *pDC) {}

	// Name
	bool Name(char *n);
	char *Name();

	// Font
	virtual GFont *GetFont();
	virtual void SetFont(GFont *Fnt);

	// Position
	virtual GRegion &GetPos() { return Pos; }
	virtual bool SetPos(GRegion &p, bool Repaint = FALSE);
	virtual int X() { return Pos.X(); }
	virtual int Y() { return Pos.Y(); }
	virtual GRegion &GetClient();

	void MoveOnScreen();
	void MoveToCenter();
	void MoveToMouse();

	// Mouse & Mouse capture capture
	virtual bool Capture(bool c);		// passing true to this function "captures"
										// the mouse, so that you will receive mouse
										// move and click events outside your windows
										// area. passing false stops this.
	virtual bool IsCapturing();			// returns whether capture is on.
	virtual bool GetMouse(GMouse &m, bool ScreenCoords = false);

	// Control related
	virtual int GetId() { return CtrlId; }
	virtual void SetId(int i);
	virtual int Value() { return 0; }
	virtual void Value(int i) {}
	virtual GView *FindControl(OsView hCtrl);
	virtual GView *FindControl(int Id);

	int GetCtrlValue(int CtrlId);
	void SetCtrlValue(int CtrlId, int i);
	char *GetCtrlName(int CtrlId);
	void SetCtrlName(int CtrlId, char *s);
	bool GetCtrlEnabled(int CtrlId);
	void SetCtrlEnabled(int Ctrld, bool Enabled);

	// Other
	virtual bool Pour(GComplexRegion &r) { return false; }
	virtual bool Invalidate(GRegion *r = NULL, bool Repaint = FALSE);
	virtual bool InvalidateFrame(GRegion *r = NULL, bool Repaint = FALSE);
	virtual bool IsOver(GMouse &m);
	virtual GView *WindowFromPoint(int x, int y);
	virtual void SetPulse(int Length = -1);
	#ifdef WIN32
	virtual bool WindowVirtualOffset(GdcPt2 *Offset);
	#endif
};

///////////////////////////////////////////////////////////////////////////////////////////////
// Control or View window
#define GV_EDGE_TOP				0x0001
#define GV_EDGE_RIGHT			0x0002
#define GV_EDGE_BOTTOM			0x0004
#define GV_EDGE_LEFT			0x0008

#define IDC_VSCROLL				14000
#define IDC_HSCROLL				14001

class LgiClass GLayout : public GView
{
protected:
	#ifdef WIN32

	static GWin32Class	Class;

	#else if BEOS

	bool XScroll, YScroll;
	BViewRedir *Redir;

	#endif

	// Private variables
	bool			_SettingScrollBars;
	bool			_PourLargest;
	GScrollBar		*VScroll;
	GScrollBar		*HScroll;

	// Private members
	bool SetScrollBars(bool x, bool y);

public:
	GLayout();
	~GLayout();

	// Overridables
	virtual void GetScrollPos(int &x, int &y);
	virtual void SetScrollPos(int x, int y);

	// Properties
	bool GetPourLargest();
	void SetPourLargest(bool i);

	// Events
	int OnEvent(GMessage *Msg);
	bool Pour(GComplexRegion &r);

	// Impl
	#ifdef BEOS
	bool Attach(GView *p);
	virtual GRegion &GetPos();
	bool SetPos(GRegion &p, bool Repaint = false);
	void Visible(bool i);
	bool Visible();
	#endif
};

////////////////////////////////////////////////////////////////////////////
// Dialog stuff
#include "LgiWidgets.h"

////////////////////////////////////////////////////////////////////////////
// Progress meters stuff
#include "Progress.h"
#include "GProgress.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
// Menus
#include "GMenu.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
class LgiClass GWindow :
	public GView
{
	friend class BViewRedir;
	friend class GView;

	bool _QuitOnClose;

protected:
	#if defined BEOS

	BWindow *Wnd;
	bool QuitRequested();

	#elif defined WIN32

	GRegion OldPos;
	HWND LastFocus;

	#elif defined ATHEOS

	os::Window *Wnd;

	#endif
	
	GMenu *Menu;

public:
	GWindow();
	~GWindow();

	// Methods
	virtual void Pour();

	// Properties
	GMenu *GetMenu() { return Menu; }
	void SetMenu(GMenu *m) { Menu = m; }
	bool Visible();
	void Visible(bool v);
	bool GetQuitOnClose() { return _QuitOnClose; }
	void SetQuitOnClose(bool i) { _QuitOnClose = i; }

	// Events
	void OnPosChange();
	int OnEvent(GMessage *Msg);
	void OnPaint(GSurface *pDC);
	virtual void OnReceiveFiles(List<char> &Files) {}

	// OS impl
	#ifndef WIN32

	bool Name(char *n);
	char *Name();
	GRegion &GetPos();
	bool SetPos(GRegion &p, bool Repaint = FALSE);
	void OnChildrenChanged(GView *Wnd, bool Attaching);
	void OnCreate();
	OsWindow GetWindow() { return Wnd; }

	#endif
};

////////////////////////////////////////////////////////////////////////
#include "GFileSelect.h"
#include "GFindReplaceDlg.h"
#include "GToolBar.h"
#include "GThread.h"

////////////////////////////////////////////////////////////////////////////////////////////////
class LgiClass GSplitter : public GLayout
{
	bool		Vertical;
	bool		Moving;
	int			SplitPos;
	int			SplitSet;
	bool		OverBar;
	GRegion	Bar;
	int			Offset;
	bool		SplitFollow;

	GRegion	PosA;
	GRegion	PosB;
	GView	*ViewA;
	GView	*ViewB;
	bool		BorderA;
	bool		BorderB;

	#ifdef WIN32
	HCURSOR		OldCursor;
	HCURSOR		Cursor;
	#endif

	void		CalcRegions(GRegion Rect, bool Follow = false);
	bool		OverSplit(int x, int y);

public:
	GSplitter();
	~GSplitter();

	bool Border() { return TestFlag(WndFlags, GWF_BORDER); }
	void Border(bool i) { (i) ? SetFlag(WndFlags, GWF_BORDER) : ClearFlag(WndFlags, GWF_BORDER); }
	bool IsVertical() { return Vertical; }
	void IsVertical(bool v);
	bool DoesSplitFollow() { return SplitFollow; }
	void DoesSplitFollow(bool i) { SplitFollow = i; }
	int Split() { return SplitPos; }
	void Split(int s);
	GView *GetViewA() { return ViewA; }
	GView *GetViewB() { return ViewB; }
	void SetViewA(GView *a, bool Border = true);
	void SetViewB(GView *b, bool Border = true);
	GView *FindControl(OsView hCtrl);

	bool Attach(GView *p);

	void OnPaint(GSurface *pDC);
	void OnPosChange();
	void OnMouseClick(GMouse &m);
	void OnMouseMove(GMouse &m);
	void OnMouseExit(GMouse &m);
	int OnHitTest(int x, int y);
};

////////////////////////////////////////////////////////////////////////////////////////////////
#define STATUSBAR_SEPARATOR			4
#define GSP_SUNKEN					0x0001

typedef GStatusPane *GStatusPanePtr;

class LgiClass GStatusBar : public GLayout, private PtrArray
{
	friend class GStatusPane;

protected:
	GFont *Font;
	void RePour();

public:
	GStatusBar();
	~GStatusBar();

	bool Pour(GComplexRegion &r);
	void OnPaint(GSurface *pDC);

	GStatusPane *AppendPane(char *Text, int Width);
	bool AppendPane(GStatusPane *Pane);
};

class LgiClass GStatusPane :
	public GView
{
	friend class GStatusBar;

protected:
	int		Flags;
	int		Width;
	GSurface *pDC;

public:
	GStatusPane();
	~GStatusPane();

	char *Name() { return GObject::Name(); }
	bool Name(char *n);
	void OnPaint(GSurface *pDC);

	int GetWidth();
	void SetWidth(int x);
	bool Sunken();
	void Sunken(bool i);
	GSurface *Bitmap();
	void Bitmap(GSurface *pdc);
};

/////////////////////////////////////////////////////////////////////////////////////////////
class LgiClass GCommand : public GObject, public GFlags
{
	int			Flags;
	bool		PrevValue;

public:
	int			Id;
	GToolButton	*ToolButton;
	GMenuItem	*MenuItem;
	GKey		*Accelerator;
	char		*TipHelp;
	
	GCommand();
	~GCommand();

	bool Enabled();
	void Enabled(bool e);
	bool Value();
	void Value(bool v);
};

/////////////////////////////////////////////////////////////////////////////////////////////
class LgiClass GClipBoard
{
	GView *Owner;
	bool Open;

public:
	GClipBoard(GView *o);
	~GClipBoard();

	bool IsOpen() { return true; }
	bool Empty();

	// Text
	bool Text(char *Str);
	char *Text();

	bool TextW(ushort *Str);
	ushort *TextW();

	// Bitmap
	bool Bitmap(GSurface *pDC);
	GSurface *Bitmap();
	#ifdef WIN32
	GSurface *ConvertFromPtr(void *Ptr);
	#endif

	// Binary
	bool Binary(int Format, uchar *Ptr, int Len);	// Set
	bool Binary(int Format, uchar **Ptr, int *Len);	// Get
};

/////////////////////////////////////////////////////////////////////////
class LgiClass GTrayIcon :
	public GObject,
	public GFlags
{
protected:
	static int Id;
	#ifdef WIN32
	static int TrayCreateMsg;
	#endif

	#ifdef WIN32
	HICON *hIcon;	// array of loaded icons
	#endif
	int Val;		// which icon visible
	int Icons;		// size of array
	int MyId;
	int Msg;
	GView *Notify;

public:
	#ifdef WIN32
	NOTIFYICONDATA	TrayIcon;
	#endif

	GTrayIcon(int Icons = 1);
	~GTrayIcon();

	bool LoadIcon(char *Str, int Index = 0);
	bool Visible();
	void Visible(bool v);

	int Value();
	void Value(int v);
	int NotifyMsg() { return Msg; }
	void NotifyMsg(int i) { Msg = i; }
	GView *GetNotify() { return Notify; }
	void SetNotify(GView *n) { Notify = n; }
	virtual int OnEvent(GMessage *Msg);
	virtual void OnMouseClick(GMouse &m);
};

//////////////////////////////////////////////////////////////////
#include "GInput.h"

//////////////////////////////////////////////////////////////////
typedef char *pchar;

class LgiClass GPrinter
{
	#ifdef WIN32
	PRINTDLG Info;
	#else if BEOS
	char Name[32];
	#endif

public:
	GPrinter();
	virtual ~GPrinter();

	#ifdef WIN32
	HDC StartDC(GView *Parent = 0);
	void EndDC();
	#else if BEOS
	void *StartDC(GView *Parent = 0);
	void EndDC();
	#endif
	bool Browse(GView *Parent);
	bool Serialize(pchar &Str, bool Write);
};

//////////////////////////////////////////////////////////////////
class LgiClass GAlert : public GDialog
{
public:
	GAlert(	GView *parent,
			char *Title,
			char *Text,
			char *Btn1,
			char *Btn2 = 0,
			char *Btn3 = 0);

	int OnNotify(GView *Ctrl, int Flags);
};

////////////////////////////////////////////////////////////////////////
#ifdef BEOS
class LgiClass GScroll : public BScrollBar {

	GScrollView *Notify;

public:
	GScroll(GScrollView *notify, orientation posture);
	void ValueChanged(float newValue);
};

class LgiClass GScrollView : public BView
{
	friend class GLayout;

	int Border;
	GView *Target;
	BScrollBar *X, *Y;
	GRegion Pos;

public:
	GScrollView(	char *name,
					GView *target,
					uint32 resizingMode = B_FOLLOW_LEFT | B_FOLLOW_TOP,
					uint32 flags = 0,
					bool horizontal = false,
					bool vertical = false,
					int border = SUNKEN);
	~GScrollView();

	void Draw(BRect update);
	void FrameResized(float width, float height);
	void AttachedToWindow();
	void SetPos(GRegion &r);
	GRegion &GetPos();
	void OnNotify(GScroll *Bar, float Value);
	BScrollBar *ScrollBar(orientation posture);
};

class LgiClass GSelectFont : public GDialog
{
	GCombo *CFont;
	GCombo *CSize;

public:
	char Font[64];
	int Size;

	GSelectFont(GView *parent, char *Face = 0);
	~GSelectFont();
	int OnNotify(GView *Ctrl, int Flags);
};

#endif

/////////////////////////////////////////////////////////
class LgiClass DoEvery {

	int64 LastTime;
	int64 Period;

public:
	// p = timeout in ms
	DoEvery(int p = 1000);
	void Init(int p = -1);
	bool DoNow();
};

#endif






















