#ifndef __STORE_COMMON_H
#define __STORE_COMMON_H

#include "GPassword.h"

class StorageItem;
class StorageObj;
class StorageKit;
typedef bool (*ValidateItemFunc)(StorageItem *Item, uint UserData);

class StorageItem
{
public:
	StorageObj *Object;

	virtual ~StorageItem() {}

	// Properties
	virtual int GetType() = 0;
	virtual void SetType(int i) = 0;
	virtual int GetTotalSize() = 0;
	virtual int GetObjectSize() = 0;
	virtual StorageKit *GetTree() = 0;

	// Heirarchy
	virtual StorageItem *GetNext() = 0;
	virtual StorageItem *GetPrev() = 0;
	virtual StorageItem *GetParent() = 0;
	virtual StorageItem *GetChild() = 0;
	virtual StorageItem *CreateNext(StorageObj *Obj) = 0;
	virtual StorageItem *CreateChild(StorageObj *Obj) = 0;
	virtual StorageItem *CreateSub(StorageObj *Obj) = 0;
	virtual bool DeleteChild(StorageItem *Obj) = 0;
	virtual bool DeleteAllChildren() = 0;

	// Impl
	virtual bool Save() = 0;
	virtual GFile *GotoObject() = 0;
	virtual bool EndOfObj(GFile &f) = 0;
};

class StorageObj
{
public:
	bool StoreDirty;
	StorageItem *Store;

	StorageObj()
	{
		StoreDirty = FALSE;
		Store = 0;
	}

	virtual ~StorageObj()
	{
		DeleteObj(Store);
	}

	virtual int Type() { return 0; }
	virtual int Sizeof() { return 0; }
	virtual bool Serialize(GFile &f, bool Write) { return FALSE; }
};

class StorageKit
{
public:
	virtual ~StorageKit() {}

	// Properties
	virtual int GetStatus() = 0;
	virtual int GetVersion() = 0;
	virtual quad GetFileSize() = 0;
	virtual bool GetPassword(GPassword *p) = 0;
	virtual bool SetPassword(GPassword *p) = 0;
	
	// Heirarchy
	virtual StorageItem *GetRoot() = 0;
	virtual StorageItem *CreateRoot(StorageObj *Obj) = 0;
	virtual bool DeleteItem(StorageItem *Item) = 0;
	virtual bool SeparateItem(StorageItem *Item) = 0;
	virtual bool ReparentItem(StorageItem *Item, StorageItem *NewParent, bool Next = false) = 0;

	// Methods
	virtual bool Compact(Progress *p, ValidateItemFunc func = 0, uint userdata = 0) = 0;
};

#endif