
#include <stdlib.h>
#include <stdio.h>

#include "Lgi.h"

#define COMBO_HEIGHT			20

//////////////////////////////////////////////////////////////////////////////////
#ifdef WIN32
GWin32Class GCombo::Win32Class;
#endif

GCombo::GCombo(int id, int x, int y, int cx, int cy, char *name) :
#if defined BEOS
	GView(NEW(BViewRedir(this)))
#elif defined ATHEOS
	GView(NEW(AView<os::View>(this)))
#else
	GView()
#endif
#ifdef LGI_RES
	, ResObject(Res_ComboBox)
#endif
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+COMBO_HEIGHT);
	Name(name);
	_SortItems = FALSE;
	_Current = 0;

	#ifdef WIN32
	WndStyle |= WS_CHILD | WS_VISIBLE | WS_TABSTOP;
	Win32Class.Create(WndClass = "LGI_GCombo");
	#endif
}

GCombo::~GCombo()
{
	DeleteObjList(&_Items, char);
}

bool GCombo::Name(char *n)
{
	char *c = _Items.ItemAt(_Current);
	if (c)
	{
		_Items.Delete(c);
		DeleteArray(c);
		
		c = NewStr(n);
		if (c)
		{
			_Items.Insert(NewStr(n), _Current);
			Invalidate();
			return true;
		}
	}

	return false;
}

char *GCombo::Name()
{
	return _Items.ItemAt(_Current);
}

void GCombo::Value(int i)
{
	char Str[32];
	sprintf(Str, "%i", i);
	Name(Str);
}

int GCombo::Value()
{
	char *n = Name();
	return (n)?atoi(n):0;
}

int GCombo::Index()
{
	return _Current;
}

void GCombo::Index(int i)
{
	if (_Current != i)
	{
		_Current = i;
		Invalidate();
	}
}

int GCombo::OnEvent(GMessage *Msg)
{
	return GView::OnEvent(Msg);
}

bool GCombo::Delete()
{
	return Delete(_Current);
}

bool GCombo::Delete(int i)
{
	char *c = _Items.ItemAt(i);
	if (c)
	{
		_Items.Delete(c);
		DeleteArray(c);
		return true;
	}
	return false;
}

bool GCombo::Delete(char *p)
{
	if (p AND _Items.HasItem(p))
	{
		_Items.Delete(p);
		DeleteArray(p);
		return true;
	}
	return false;
}

bool GCombo::Insert(char *p, int Index)
{
	int i = (Index < 0) ? _Items.GetItems() : Index;
	bool Status = _Items.Insert(NewStr(p), Index);
	if (Status AND i == _Current)
	{
		Invalidate();
	}
	return Status;
}

void GCombo::OnMouseClick(GMouse &m)
{
	if (m.Down() AND
		Enabled())
	{
		Focus(true);

		if (m.Left() AND
			_Arrow.Overlap(m.x, m.y))
		{

			GSubMenu *RClick = NEW(GSubMenu);
			if (RClick)
			{
				int Base = 1000;
				int i=0;
				char *c = _Items.First();
				if (c)
				{
					for (; c; c = _Items.Next(), i++)
					{
						RClick->AppendItem(c, Base+i, i != _Current);
					}
				}
				else
				{
					RClick->AppendItem("", Base+i, false);
				}

				#if defined WIN32
				POINT p = {_Arrow.x1, _Arrow.y1};
				ClientToScreen(Handle(), &p);
				#else
				GMouse p;
				GetMouse(p, true);
				#endif

				int Result = RClick->Float(this, p.x, p.y, true);
				if (Result >= Base)
				{
					_Current = Result - Base;
					Invalidate();

					GView *n = (GetNotify()) ? GetNotify() : GetParent();
					if (n)
					{
						n->OnNotify(this, _Current);
					}
				}
			}
		}
	}
}

void GCombo::OnKey(GKey &k)
{
	if (k.Down)
	{
		switch (k.c)
		{
			case VK_UP:
			{
				int i = Index();
				if (i > 0)
				{
					Index(i-1);
				}
				break;
			}
			case VK_DOWN:
			{
				int i = Index();
				if (i < _Items.GetItems()-1)
				{
					Index(i+1);
				}
				break;
			}
		}
	}
}

void GCombo::OnFocus(bool f)
{
	Invalidate();
}

void GCombo::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);

	#ifdef WIN32
	if (r.Y() != COMBO_HEIGHT)
	{
		GRegion t = GetPos();
		t.y2 = t.y1 + COMBO_HEIGHT - 1;
		SetPos(t, true);
		r.ZOff(t.X()-1, t.Y()-1);
	}
	#endif

	// draw shadowed border
	pDC->Colour(LC_SHADOW, 24);
	pDC->Line(r.x1+2, r.y2, r.x2, r.y2);
	pDC->Line(r.x2, r.y1+2, r.x2, r.y2);

	pDC->Colour(LC_MED, 24);
	pDC->Set(r.x1, r.y2);
	pDC->Set(r.x1+1, r.y2);
	pDC->Set(r.x2, r.y1);
	pDC->Set(r.x2, r.y1+1);

	r.x2--;
	r.y2--;
	pDC->Colour(LC_LOW, 24);
	pDC->Box(&r);
	r.Size(1, 1);
	pDC->Colour(LC_HIGH, 24);
	pDC->Line(r.x1, r.y1, r.x2, r.y1);
	pDC->Line(r.x1, r.y1+1, r.x1, r.y2);
	r.x1++;
	r.y1++;

	// fill the background
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&r);

	// draw the drop down arrow
	_Arrow = r;
	_Arrow.x1 = _Arrow.x2 - 10;
	pDC->Colour(Enabled() ? LC_SHADOW : LC_LOW, 24);
	int Cx = r.x2 - 5;
	int Cy = r.y1 + (r.Y()/2) + 2;
	for (int i=0; i<4; i++)
	{
		pDC->Line(Cx-i, Cy, Cx+i, Cy);
		Cy--;
	}
	r.x2 = _Arrow.x1 - 1;
	pDC->Colour(LC_LOW, 24);
	pDC->Line(r.x2, r.y1+1, r.x2, r.y2-1);
	r.x2--;

	// draw the text
	r.Size(1, 1);
	if (r.Valid())
	{
		char *Text = Name();
		if (Text)
		{
			if (Enabled())
			{
				bool f = Focus();
				SysFont->Colour(f ? LC_SEL_TEXT : LC_TEXT, f ? LC_SELECTION : LC_MED);
				SysFont->Transparent(false);
				SysFont->Text(pDC, r.x1, r.y1, Text, -1, &r);
			}
			else
			{
				SysFont->Transparent(false);
				SysFont->Colour(LC_LIGHT, LC_MED);
				SysFont->Text(pDC, r.x1+1, r.y1+1, Text, -1, &r);

				SysFont->Transparent(true);
				SysFont->Colour(LC_LOW, 0);
				SysFont->Text(pDC, r.x1, r.y1, Text, -1, &r);
			}
		}
		else
		{
			pDC->Colour(Focus() ? LC_SELECTION : LC_MED, 24);
			pDC->Rectangle(&r);
		}
	}
}

