/*
**	FILE:		DragAndDrop.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		30/11/98
**	DESCRIPTION:	Drag and drop support
**
**	Copyright (C) 1998, Matthew Allen
**		fret@ozemail.com.au
*/

#include <stdio.h>

#include "Lgi.h"
#include "DragAndDrop.h"

/////////////////////////////////////////////////////////////////////////////////////////
GDragDropSource::GDragDropSource()
{
	bool Status = FALSE;

	#ifdef WIN32
	Index = 0;
	DdFormat = RegisterClipboardFormat(CF_GDragDropSource);
	if (NOT DdFormat)
	{
		Status = true;
	}
	#endif

	OnDragInit(Status);
}

GDragDropSource::~GDragDropSource()
{
	#ifdef WIN32
	#else if BEOS
	#endif
}

bool GDragDropSource::GetFormats(List<int> &Formats)
{
	#ifdef WIN32

	Formats.Insert(NEW(int(DdFormat)));

	#else if BEOS

	return 0;

	#endif

	return Formats.GetItems() > 0;
}

int GDragDropSource::Drag(GView *SourceWnd, int Effect)
{
	#if defined WIN32
	DWORD dwEffect;
	if (::DoDragDrop(this, this, Effect, &dwEffect) == DRAGDROP_S_DROP)
	{
		return Effect;
	}
	return DROPEFFECT_NONE;
	#elif defined BEOS
	if (SourceWnd)
	{
		BMessage Msg(WM_DRAG_DROP);
		Msg.AddPointer("GDragDropSource", this);
		
		BPoint Cursor;
		uint32 Buttons;
		SourceWnd->Handle()->GetMouse(&Cursor, &Buttons);
		
		SourceWnd->Handle()->DragMessage(&Msg, BRect(Cursor.x-10, Cursor.y-10, Cursor.x+10, Cursor.y+10));
		return DROPEFFECT_COPY;
	}
	return DROPEFFECT_NONE;
	#endif
}

#ifdef WIN32
ULONG STDMETHODCALLTYPE
GDragDropSource::AddRef()
{
	return 1;
}

ULONG STDMETHODCALLTYPE
GDragDropSource::Release()
{
	return 0;
}

HRESULT
STDMETHODCALLTYPE
GDragDropSource::QueryInterface(REFIID iid, void **ppv)
{
	*ppv=NULL;
	if (IID_IUnknown==iid)
	{
		*ppv=(void*)(IUnknown*)(IDataObject*)this;
	}

	if (IID_IEnumFORMATETC==iid)
	{
		*ppv=(void*)(IEnumFORMATETC*) this;
	}

	if (IID_IDataObject==iid)
	{
		*ppv=(void*)(IDataObject*) this;
	}

	if (IID_IDropSource==iid)
	{
		*ppv=(void*)(IDropSource*) this;
	}

	if (NULL==*ppv)
	{
		return E_NOINTERFACE;
	}

	AddRef();
	
	return NOERROR;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::Next(ULONG celt, FORMATETC *rgelt, ULONG *pceltFetched)
{
	List<int> Formats;
	if (rgelt AND
		GetFormats(Formats))
	{
		int *i = Formats.ItemAt(Index);
		if (i)
		{
			rgelt->cfFormat = *i;
			rgelt->ptd = 0;
			rgelt->dwAspect = DVASPECT_CONTENT;
			rgelt->lindex = -1;
			rgelt->tymed = TYMED_HGLOBAL;
			Index++;

			if (pceltFetched)
			{
				*pceltFetched = 1;
				return (celt == *pceltFetched) ? S_OK : S_FALSE;
			}

			return S_OK;
		}
	}

	return S_FALSE;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::Skip(ULONG celt)
{
	Index += celt;
	return S_FALSE;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::Reset()
{
	Index = 0;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::Clone(IEnumFORMATETC **ppenum)
{
	if (ppenum)
	{
		*ppenum = NULL;
		if (*ppenum)
		{
			return S_OK;
		}
		return E_OUTOFMEMORY;
	}
	return E_INVALIDARG;
}

HRESULT
STDMETHODCALLTYPE
GDragDropSource::QueryContinueDrag(BOOL fEscapePressed, DWORD InputState)
{
	if (fEscapePressed)
	{
		return DRAGDROP_S_CANCEL;
	}

	if (NOT (InputState & (MK_LBUTTON | MK_MBUTTON | MK_RBUTTON)))
	{
		return DRAGDROP_S_DROP;
	}

	return S_OK;
}

HRESULT
STDMETHODCALLTYPE
GDragDropSource::GiveFeedback(DWORD dwEffect)
{
	return DRAGDROP_S_USEDEFAULTCURSORS;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::GetData(FORMATETC *pFormatEtc, STGMEDIUM *PMedium)
{
	List<int> Formats;
	CurrentFormat = -1;
	GetFormats(Formats);
	for (int *i=Formats.First(); i; i = Formats.Next())
	{
		if (*i == pFormatEtc->cfFormat)
		{
			CurrentFormat = pFormatEtc->cfFormat;
		}
		// DeleteObj(i);
	}

	if (pFormatEtc AND PMedium AND CurrentFormat > 0)
	{
		uchar *DataPtr = 0;
		int DataSize = 0;

		// the users don't HAVE to use this...
		GetData(&DataPtr, &DataSize);

		ZeroObj(*PMedium);
		PMedium->tymed = TYMED_HGLOBAL;

		if (CurrentFormat == DdFormat)
		{
			PMedium->hGlobal = GlobalAlloc(GHND, sizeof(this) + sizeof(int) + DataSize);
			if (PMedium->hGlobal)
			{
				GDragDropSource **Ptr = (GDragDropSource**) GlobalLock(PMedium->hGlobal);
				if (Ptr)
				{
					Ptr[0] = this;
					*((int*) (Ptr+1)) = DataSize;
					if (DataPtr)
					{
						uchar *p = (uchar*) (Ptr+2);
						memcpy(p, DataPtr, DataSize);
					}

					GlobalUnlock(PMedium->hGlobal);

					return S_OK;
				}
			}
		}
		else
		{
			PMedium->hGlobal = GlobalAlloc(GHND, DataSize);
			if (PMedium->hGlobal)
			{
				GDragDropSource **Ptr = (GDragDropSource**) GlobalLock(PMedium->hGlobal);
				if (Ptr)
				{
					memcpy(Ptr, DataPtr, DataSize);
					GlobalUnlock(PMedium->hGlobal);

					return S_OK;
				}
			}
		}
	}

	return E_INVALIDARG;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::GetDataHere(FORMATETC *pFormatEtc, STGMEDIUM *PMedium)
{
	return E_INVALIDARG;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::QueryGetData(FORMATETC *pFormatEtc)
{
	if (pFormatEtc)
	{
		List<int> Formats;
		GetFormats(Formats);
		bool HaveFormat = false;

		char name[256] = "";
		GetClipboardFormatName(pFormatEtc->cfFormat, name, sizeof(name));

		for (int *i = Formats.First(); i; i = Formats.Next())
		{
			if (pFormatEtc->cfFormat == *i)
			{
				HaveFormat = true;
			}

			DeleteObj(i);
		}
		if (NOT HaveFormat)
		{
			return DV_E_FORMATETC;
		}

		if (pFormatEtc->tymed != TYMED_HGLOBAL)
		{
			return DV_E_TYMED;
		}

		return S_OK;
	}
	
	return E_INVALIDARG;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::EnumFormatEtc(DWORD dwDirection, IEnumFORMATETC **ppFormatEtc)
{
	if (ppFormatEtc)
	{
		*ppFormatEtc = (IEnumFORMATETC*) this;
		return (*ppFormatEtc) ? S_OK : E_OUTOFMEMORY;
	}

	return E_INVALIDARG;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::GetCanonicalFormatEtc(FORMATETC * pFormatetcIn, FORMATETC * pFormatetcOut)
{
	return E_UNEXPECTED;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::SetData(FORMATETC * pFormatetc, STGMEDIUM * pmedium, BOOL fRelease)
{
	return E_UNEXPECTED;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::DAdvise(FORMATETC * pFormatetc, DWORD advf, IAdviseSink * pAdvSink, DWORD * pdwConnection)
{
	return E_UNEXPECTED;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::DUnadvise(DWORD dwConnection)
{
	return E_UNEXPECTED;
}

HRESULT STDMETHODCALLTYPE
GDragDropSource::EnumDAdvise(IEnumSTATDATA **ppenumAdvise)
{
	return E_UNEXPECTED;
}
#endif

////////////////////////////////////////////////////////////////////////////////////////////
GDragDropTarget::GDragDropTarget()
{
	DragDropObj = 0;
	DragDropData = 0;
	DragDropLength = 0;
	To = 0;
}

GDragDropTarget::~GDragDropTarget()
{
	DragDropObj = 0;
	DragDropLength = 0;
	DeleteArray(DragDropData);
}

void GDragDropTarget::SetWindow(GView *to)
{
	bool Status = false;
	#ifdef WIN32
	To = to;
	if (RegisterDragDrop(To->Handle(), (IDropTarget*) this) == S_OK)
	{
		Status = true;
	}
	#else if BEOS
	#endif
	OnDragInit(Status);
}

#ifdef WIN32
ULONG STDMETHODCALLTYPE GDragDropTarget::AddRef()
{
	return 1;
}

ULONG STDMETHODCALLTYPE GDragDropTarget::Release()
{
	return 0;
}

HRESULT STDMETHODCALLTYPE GDragDropTarget::QueryInterface(REFIID iid, void **ppv)
{
	*ppv=NULL;

	if (IID_IUnknown==iid)
	{
		*ppv=(void*)(IUnknown*)(IDataObject*) this;
	}

	if (IID_IDropTarget==iid)
	{
		*ppv=(void*)(IDropTarget*) this;
	}

	if (NULL==*ppv)
	{
		return E_NOINTERFACE;
	}

	AddRef();
	
	return NOERROR;
}

HGLOBAL GDragDropTarget::GetDataHandle(IDataObject *pDataObject)
{
	if (pDataObject)
	{
		FORMATETC Format;
		STGMEDIUM Medium;

		ZeroObj(Format);
		Format.cfFormat = RegisterClipboardFormat(CF_GDragDropSource);
		Format.dwAspect = DVASPECT_CONTENT;
		Format.lindex = -1;
		Format.tymed = TYMED_HGLOBAL;

		ZeroObj(Medium);
		if (pDataObject->GetData(&Format, &Medium) == S_OK)
		{
			if (Medium.tymed == TYMED_HGLOBAL AND
				Medium.hGlobal)
			{
				return Medium.hGlobal;
			}
		}
	}

	return 0;
}

bool GDragDropTarget::UnwrapData(HGLOBAL hMem)
{
	bool Status = false;

	if (hMem)
	{
		GDragDropSource **Ptr = (GDragDropSource**) GlobalLock(hMem);
		if (Ptr)
		{
			DragDropObj = Ptr[0];
			if (DragDropObj)
			{
				DragDropLength = (int) Ptr[1];
				DragDropData = NEW(uchar[DragDropLength]);
				if (DragDropData)
				{
					memcpy(DragDropData, (void*) (Ptr + 2), DragDropLength);
					Status = true;
				}
			}

			GlobalUnlock(Ptr);
			GlobalFree(hMem);
		}
	}

	return Status;
}

HRESULT STDMETHODCALLTYPE GDragDropTarget::DragEnter(IDataObject *pDataObject, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect)
{
	HRESULT Result = E_UNEXPECTED;

	DragDropObj = 0;
	DragDropLength = 0;
	DeleteArray(DragDropData);

	HGLOBAL hMem = GetDataHandle(pDataObject);
	if (hMem)
	{
		// LGI drag'n'drop operation
		if (UnwrapData(hMem))
		{
			OnDragEnter();
			DragOver(grfKeyState, pt, pdwEffect);
		}

		Result = S_OK;
	}
	else
	{
		// Something from another app
		IEnumFORMATETC *FormatETC = 0;
		if (pDataObject->EnumFormatEtc(DATADIR_GET, &FormatETC) == S_OK AND
			FormatETC)
		{
			// Enumerate the formats...
			ULONG Fetched = 0;
			FORMATETC Format;

			// Clean out format list
			DeleteObjList(&Formats, int);

			// Ask what formats are being dropped
			while (FormatETC->Next(1, &Format, &Fetched) == S_OK)
			{
				if (Fetched == 1)
				{
					Formats.Insert(NEW(int(Format.cfFormat)));
				}
			}

			// Process the format list
			if (Formats.GetItems() > 0)
			{
				POINT p;
				p.x = pt.x;
				p.y = pt.y;
				ScreenToClient(To->Handle(), &p);
				GdcPt2 Pt(p.x, p.y);

				// Ask the app what formats it supports.
				// It deletes those it doesn't and leaves the formats it can handle
				int Effect = WillAccept(Formats, Pt, grfKeyState);
				if (Effect != DROPEFFECT_NONE)
				{
					*pdwEffect |= Effect;
				}
				else
				{
					*pdwEffect = DROPEFFECT_NONE;
				}

				Result = S_OK;
			}

			FormatETC->Release();
		}
	}

	return Result;
}

HRESULT STDMETHODCALLTYPE GDragDropTarget::DragOver(DWORD grfKeyState, POINTL pt, DWORD *pdwEffect)
{
	if (DragDropObj AND pdwEffect)
	{
		ASSERT(To);

		POINT p;
		p.x = pt.x;
		p.y = pt.y;
		ScreenToClient(To->Handle(), &p);
		GdcPt2 Pt(p.x, p.y);
		int Effect = WillAccept(DragDropObj, Pt, grfKeyState);
		if (Effect != DROPEFFECT_NONE)
		{
			*pdwEffect |= Effect;
		}
		else
		{
			*pdwEffect = DROPEFFECT_NONE;
		}
	}

	return S_OK;
}

HRESULT STDMETHODCALLTYPE GDragDropTarget::DragLeave()
{
	OnDragExit();
	return S_OK;
}

HRESULT STDMETHODCALLTYPE GDragDropTarget::Drop(IDataObject *pDataObject, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect)
{
	HRESULT Result = E_UNEXPECTED;

	ASSERT(To);

	POINT p;
	p.x = pt.x;
	p.y = pt.y;
	ScreenToClient(To->Handle(), &p);
	GdcPt2 Pt(p.x, p.y);

	HGLOBAL hMem = GetDataHandle(pDataObject);
	if (hMem)
	{
		// Lgi Drop
		if (UnwrapData(hMem))
		{
			int Effect = OnDrop(DragDropData, DragDropLength, Pt, grfKeyState);
			if (pdwEffect)
			{
				*pdwEffect = Effect;
			}
		}

		Result = S_OK;
	}
	else
	{
		// OS drop
		if (Formats.GetItems() > 0)
		{
			FORMATETC Format;
			Format.cfFormat = *Formats[0];
			Format.dwAspect = DVASPECT_CONTENT;
			Format.lindex = -1;
			Format.tymed = TYMED_HGLOBAL;
			Format.ptd = 0;

			STGMEDIUM Medium;
			ZeroObj(Medium);

			if (pDataObject->GetData(&Format, &Medium) == S_OK)
			{
				Result = S_OK;
				switch (Medium.tymed)
				{
					case TYMED_HGLOBAL:
					{
						int Size = GlobalSize(Medium.hGlobal);
						void *Ptr = GlobalLock(Medium.hGlobal);
						if (Ptr)
						{
							OnDrop((uchar*)Ptr, Size, Pt, grfKeyState);
							GlobalUnlock(Ptr);
						}
						break;
					}
					default:
					{
						// unsupported TYMED
						Result = E_UNEXPECTED;
						break;
					}
				}
			}
		}
	}

	return Result;
}
#endif
