#include <stdio.h>

#include "Lgi.h"
#include "GMru.h"

////////////////////////////////////////////////////////////////////
#define IDM_MRU_BASE				(WM_USER+0x3500)

////////////////////////////////////////////////////////////////////
GMru::GMru()
{
	_Parent = 0;
	_Size = 10;
	_SelectedType = 0;
}

GMru::~GMru()
{
	DeleteObjList(&_Items, char);
}

void GMru::GetFileTypes(GFileSelect *Dlg, bool Write)
{
	Dlg->Type("All Files", "*.*");
}

bool GMru::_OpenFile(char *File)
{
	bool Status = OpenFile(File);

	if (Status)
	{
		AddFile(File);
	}
	else
	{
		RemoveFile(File);
	}

	return Status;
}

bool GMru::_SaveFile(char *File)
{
	bool Status = false;

	if (File)
	{
		if (GetSelectedType() AND
			GetSelectedType()->Extension())
		{
			// extract extension
			char Ext[64];
			char *s = strchr(GetSelectedType()->Extension(), '.'), *d = Ext;
			if (s)
			{
				while (*s AND *s != ';')
				{
					*d++ = *s++;
				}
				*d++ = 0;
				strlwr(Ext);

				// bung the extension from the file type if not there
				char *Dot = strrchr(File, '.');
				if (NOT Dot OR strchr(Dot, DIR_CHAR))
				{
					strcat(File, Ext);
				}
			}
		}
		
		Status = SaveFile(File);

		if (Status)
		{
			AddFile(File);
		}
		else
		{
			RemoveFile(File);
		}
	}

	return Status;
}

void GMru::_Update()
{
	while (_Items.GetItems() > _Size)
	{
		char *s = _Items.Last();
		_Items.Delete(s);
		DeleteArray(s);
	}

	if (_Parent)
	{
		// remove existing items.
		GMenuItem *Item;
		while (Item = _Parent->ItemAt(0))
		{
			_Parent->RemoveItem(0);
			DeleteObj(Item);
		}

		// add current items
		if (_Items.GetItems() > 0)
		{
			int i = IDM_MRU_BASE;
			for (char *c = _Items.First(); c; c = _Items.Next())
			{
				_Parent->AppendItem(c, i++, true);
			}
		}
		else
		{
			_Parent->AppendItem("(none)", -1, false);
		}
	}
}

bool GMru::Set(GSubMenu *parent, int size)
{
	_Parent = parent;
	if (size > 0)
	{
		_Size = size;
	}

	_Update();

	return true;
}

char *GMru::AddFile(char *FileName, bool Update)
{
	char *Status = FileName;
	char *c = 0;
	for (c = _Items.First(); c; c = _Items.Next())
	{
		if (c == FileName)
		{
			// exact string being added.. just move to the top
			// no need to reallocate
			_Items.Delete(c);
			_Items.Insert(c, 0);
			break;
		}
	}

	if (NOT c)
	{
		Status = NewStr(FileName);
		RemoveFile(FileName, false); // may have same path
		_Items.Insert(Status, 0);
	}

	if (Update)
	{
		// update
		_Update();
	}

	return Status;
}

void GMru::RemoveFile(char *FileName, bool Update)
{
	// remove from list if there
	for (char *c = _Items.First(); c; c = _Items.Next())
	{
		if (stricmp(c, FileName) == 0)
		{
			_Items.Delete(c);
			DeleteArray(c);
			break;
		}
	}

	if (Update)
	{
		_Update();
	}
}

int GMru::OnEvent(GMessage *Msg)
{
	if (_Parent AND
		MsgCode(Msg) == WM_COMMAND)
	{
		#ifdef BEOS
		int32 Cmd = 0;
		int32 Event = 0;
		Msg->FindInt32("Cmd", &Cmd);
		Msg->FindInt32("Event", &Event);
		#else
		int Cmd = MsgA(Msg) & 0xffff;
		#endif

		GView *Wnd = _Parent->GetMenu() ? _Parent->GetMenu()->GetWindow() : 0;
		if (Wnd)
		{
			_Select.Parent(Wnd);
			_Select.ClearTypes();
			_SelectedType = 0;

			if (Cmd == IDM_OPEN)
			{
				GetFileTypes(&_Select, false);
				if (_Select.Open())
				{
					_SelectedType = _Select.TypeAt(_Select.SelectedType());
					_OpenFile(_Select.Name());
				}
			}
			else if (Cmd == IDM_SAVEAS)
			{
				GetFileTypes(&_Select, true);
				if (_Select.Save())
				{
					_SelectedType = _Select.TypeAt(_Select.SelectedType());
					_SaveFile(_Select.Name());
				}
			}
		}

		if (Cmd >= IDM_MRU_BASE AND
			Cmd < IDM_MRU_BASE + _Items.GetItems())
		{
			int Index = Cmd - IDM_MRU_BASE;
			char *c = _Items.ItemAt(Index);
			if (c)
			{
				_OpenFile(c);
			}
		}
	}

	return false;
}

bool GMru::Serialize(ObjProperties *Props, char *Prefix, bool Write)
{
	bool Status = false;

	if (Props AND Prefix)
	{
		if (Write)
		{
			// clear existing keys
			for (bool b=Props->FirstKey(); b; b=Props->NextKey())
			{
				if (Props->KeyName() AND
					strnicmp(Props->KeyName(), "Prefix", strlen(Prefix)) == 0)
				{
					Props->DeleteKey();
					b = Props->KeyType();
				}
				else
				{
					b=Props->NextKey();
				}
			}

			// add our keys
			char Key[64];
			sprintf(Key, "%s.Items", Prefix);
			Props->Set(Key, _Items.GetItems());

			int i=0;
			for (char *c=_Items.First(); c; c=_Items.Next())
			{
				sprintf(Key, "%s.%i", Prefix, i++);
				Props->Set(Key, c);
			}
		}
		else
		{
			// clear ourself
			DeleteObjList(&_Items, char);

			// read our keys in
			char Key[64];
			sprintf(Key, "%s.Items", Prefix);
			int i = 0;
			if (Props->Get(Key, i))
			{
				for (int n=0; n<i; n++)
				{
					sprintf(Key, "%s.%i", Prefix, n);

					char *File;
					if (Props->Get(Key, File))
					{
						_Items.Insert(NewStr(File));
					}
				}
			}

			_Update();
		}
	}

	return Status;
}

