/*
**	FILE:			GPanel.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			29/8/99
**	DESCRIPTION:	Scribe Mail Object and UI
**
**	Copyright (C) 1999, Matthew Allen
**		fret@ozemail.com.au
*/
#include <stdio.h>
#include "Lgi.h"
#include "GPanel.h"

//////////////////////////////////////////////////////////////////////////////
GPanel::GPanel(char *name, int size, bool open)
{
	if (name) Name(name);
	IsOpen = open;
	ClosedSize = SysFont->Y("A") + 3;
	OpenSize = size;
	Align = GV_EDGE_TOP;
	
	#if defined BEOS
	Handle()->SetViewColor(216, 216, 216);
	#endif
}

GPanel::~GPanel()
{
	int n=0;
}

int GPanel::CalcWidth()
{
	return 30 + SysFont->X(Name());
}

bool GPanel::Open()
{
	return IsOpen;
}

void GPanel::Open(bool i)
{
	if (i != IsOpen)
	{
		IsOpen = i;
		SetChildrenVisibility(IsOpen);
		RePour();
	}
}

int GPanel::Alignment()
{
	return Align;
}

void GPanel::Alignment(int i)
{
	Align = i;
}

int GPanel::GetClosedSize()
{
	return ClosedSize;
}

void GPanel::SetClosedSize(int i)
{
	ClosedSize = i;
}

int GPanel::GetOpenSize()
{
	return OpenSize;
}

void GPanel::SetOpenSize(int i)
{
	OpenSize = i;
}

bool GPanel::Attach(GView *Wnd)
{
	bool Status = GLayout::Attach(Wnd);
	if (Status)
	{
		AttachChildren();
	}

	SetChildrenVisibility(IsOpen);
	return Status;
}

bool GPanel::Pour(GComplexRegion &r)
{
	int Sx = CalcWidth();
	GRegion *Best = 0;
	if (Open())
	{
		Best = FindLargest(r);
	}
	else
	{
		Best = FindSmallestFit(r, Sx, ClosedSize);
	}

	if (Best)
	{
		GRegion r = *Best;
		if (OpenSize > 0)
		{
			int Size = ((Open()) ? OpenSize : ClosedSize);
			int Limit = 30;
			if (TestFlag(Align, GV_EDGE_RIGHT) OR
				TestFlag(Align, GV_EDGE_LEFT))
			{
				Limit = min(Size, r.X()-1);
			}
			else /* if (TestFlag(Align, GV_EDGE_BOTTOM) OR
					 TextFlag(Align, GV_EDGE_TOP)) */
			{
				Limit = min(Size, r.Y()-1);
			}
			
			if (Align & GV_EDGE_RIGHT)
			{
				r.x1 = r.x2 - Limit;
			}
			else if (Align & GV_EDGE_BOTTOM)
			{
				r.y1 = r.y2 - Limit;
			}
			else if (Align & GV_EDGE_LEFT)
			{
				r.x2 = r.x1 + Limit;
			}
			else // if (Align & GV_EDGE_TOP)
			{
				r.y2 = r.y1 + Limit;
			}
			
			if (NOT Open())
			{
				r.x2 = r.x1 + Sx - 1;
			}
		}
		else
		{
			r.y2 = r.y1 - OpenSize;
		}
		SetPos(r);
		return TRUE;
	}
	return FALSE;
}

int GPanel::OnNotify(GView *Ctrl, int Flags)
{
	if (GetParent())
	{
		return GetParent()->OnNotify(Ctrl, Flags);
	}

	return 0;
}

void GPanel::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, Pos.X()-1, Pos.Y()-1);

	LgiThinBorder(pDC, r, RAISED);

	#ifndef BEOS // BeOS view is set to paint itself
	pDC->Colour(LC_MED, 24);
	GComplexRegion Cr(r);

	#ifdef WIN32
	if (Open())
	{
		for (GView *c = Children.First(); c; c = Children.Next())
		{
			Cr.Subtract(&c->GetPos());
		}
	}
	#endif
	for (GRegion *i = Cr.First(); i; i = Cr.Next())
	{
		pDC->Rectangle(i);
	}
	#endif

	SysFont->Transparent(false);
	SysFont->Fore(0);
	SysFont->Back(LC_MED);
	if (NOT Open())
	{
		// title
		SysFont->Text(pDC, r.x1 + 20, r.y1 + 1, Name());
	}

	if (OpenSize > 0)
	{
		// Draw thumb
		ThumbPos.ZOff(8, 8);
		ThumbPos.Offset(r.x1 + 3, r.y1 + 3);
		
		pDC->Colour(LC_LOW, 24);
		pDC->Box(&ThumbPos);
		pDC->Colour(LC_WHITE, 24);
		pDC->Rectangle(ThumbPos.x1+1, ThumbPos.y1+1, ThumbPos.x2-1, ThumbPos.y2-1);
		pDC->Colour(LC_SHADOW, 24);
		pDC->Line(	ThumbPos.x1+2,
					ThumbPos.y1+4,
					ThumbPos.x1+6,
					ThumbPos.y1+4);
		
		if (NOT Open())
		{
			pDC->Line(	ThumbPos.x1+4,
						ThumbPos.y1+2,
						ThumbPos.x1+4,
						ThumbPos.y1+6);
		}
	}
}

void GPanel::OnMouseClick(GMouse &m)
{
	if (OpenSize > 0 AND
		m.Left() AND
		m.Down() AND
		ThumbPos.Overlap(m.x, m.y))
	{
		Open(NOT IsOpen);
		if (GetParent())
		{
			OnNotify(this, 0);
		}
	}
}

void GPanel::RePour()
{
	for (GView *p = GetParent(); p; p = p->GetParent())
	{
		GWindow *App = dynamic_cast<GWindow*>(p);
		if (App)
		{
			App->Pour();
			break;
		}
	}
}

void GPanel::SetChildrenVisibility(bool i)
{
	for (GView *w = Children.First(); w; w = Children.Next())
	{
		bool Vis = w->Visible();
		
		w->Visible(i);
	}
}
