/*hdr
**      FILE:           GTabView.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           20/10/2000
**      DESCRIPTION:    Lgi self-drawn tab control
**
**      Copyright (C) 2000 Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

#ifdef WIN32
GWin32Class GTabView::W32Class;
#endif

#ifdef WIN32

#ifdef TOOL_VLOW
#undef TOOL_VLOW
#endif

#ifdef TOOL_LOW
#undef TOOL_LOW
#endif

#ifdef TOOL_HIGH
#undef TOOL_HIGH
#endif

#ifdef TOOL_VHIGH
#undef TOOL_VHIGH
#endif


#define TOOL_VLOW	GetSysColor(COLOR_3DDKSHADOW)
#define TOOL_LOW	GetSysColor(COLOR_3DSHADOW)
#define TOOL_HIGH	GetSysColor(COLOR_3DLIGHT)
#define TOOL_VHIGH	GetSysColor(COLOR_3DHILIGHT)

#endif

////////////////////////////////////////////////////////////////////////////////////////////
// Tab Control
GTabView::GTabView(int id, int x, int y, int cx, int cy, char *name, int Init) :
#if defined BEOS
	GControl(NEW(BViewRedir(this)))
#elif defined ATHEOS
	GControl(NEW(AView<os::View>(this)))
#else
	GControl()
#endif
#ifdef LGI_RES
	, ResObject(Res_TabView)
#endif
{
	SetId(id);
	Pos.ZOff(cx, cy);
	Pos.Offset(x, y);
	Name(name);
	Current = Init;
	TabClient.ZOff(-1, -1);
	_BorderSize = 0;
	
	#if defined WIN32
	WndClass = LGI_TABCONTROL;
	WndDlgCode = DLGC_WANTARROWS;
	// WndExStyle |= WS_EX_CONTROLPARENT;
	WndStyle |= WS_TABSTOP;
	W32Class.Create(Class());
	#elif defined BEOS
	Handle()->SetFlags(	Handle()->Flags() |
						B_NAVIGABLE |
						B_FULL_UPDATE_ON_RESIZE);
	#endif
}

GTabView::~GTabView()
{
	List<GTabPage> Temp;
	for (GTabPage *p=Tabs.First(); p; p=Tabs.Next())
	{
		Temp.Insert(p);
	}
	Tabs.Empty();
	DeleteObjList(&Temp, GTabPage);
}

int GTabView::TabY()
{
	return max(16, SysFont->Y("A")) + 4;
}

void GTabView::OnChildrenChanged(GView *Wnd, bool Attaching)
{
	/*
	if (NOT Attaching)
	{
		for (	GTabPage *Page = Tabs.First();
				Page;
				Page = Tabs.Next())
		{
			Page->Children.Delete(Wnd);
		}
	}
	*/
}

#ifdef WIN32
GView *GTabView::FindControl(HWND hCtrl)
{
	GView *Ctrl = 0;

	if (hCtrl == Handle())
	{
		return this;
	}

	for (	GTabPage *Page = Tabs.First();
			Page AND NOT Ctrl;
			Page = Tabs.Next())
	{
		Ctrl = Page->FindControl(hCtrl);
	}
	return Ctrl;
}

#endif

GView *GTabView::FindControl(int Id)
{
	GView *Ctrl = 0;

	if (GetId() == Id)
	{
		return this;
	}

	for (	GTabPage *Page = Tabs.First();
			Page AND NOT Ctrl;
			Page = Tabs.Next())
	{
		Iterator<GView> Ctrls(&Page->Children);
		for (GView *c = Ctrls.First(); c AND NOT Ctrl; c = Ctrls.Next())
		{
			Ctrl = c->FindControl(Id);
		}
	}
	return Ctrl;
}

bool GTabView::Attach(GView *parent)
{
	bool Status = GControl::Attach(parent);
	if (Status)
	{
		GTabPage *p = Tabs.ItemAt(Current);
		if (p)
		{
			p->Attach(this);
		}
		
		OnPosChange();
	}

	return Status;
}

int GTabView::Value()
{
	return Current;
}

void GTabView::Value(int i)
{
	if (i != Current)
	{
		// change tab
		GTabPage *Old = Tabs.ItemAt(Current);
		if (Old)
		{
			Old->Visible(false);
		}
		
		Current = i;
		
		GTabPage *p = Tabs.ItemAt(Current);
		if (p)
		{
			if (NOT p->IsAttached())
			{
				p->Attach(this);
			}
			p->Visible(true);
		}
		
		OnPosChange();
		Invalidate();

		GView *n = (Notify) ? Notify : Parent;
		if (n)
		{
			n->OnNotify(this, Current);
		}
	}
}

int GTabView::OnEvent(GMessage *Msg)
{
	return GView::OnEvent(Msg);
}

int GTabView::OnNotify(GView *Ctrl, int Flags)
{
	if (GetParent())
	{
		return GetParent()->OnNotify(Ctrl, Flags);
	}
	
	return 0;
}

bool GTabView::Append(GTabPage *Page, int Where)
{
	if (Page)
	{
		Page->TabCtrl = this;
		Tabs.Insert(Page, Where);

		if (IsAttached() AND Tabs.GetItems() == 1)
		{
			Page->Attach(this);
			OnPosChange();
		}

		Invalidate();

		return true;
	}
	return false;
}

GTabPage *GTabView::Append(char *name, int Where)
{
	GTabPage *Page = NEW(GTabPage(name));
	if (Page)
	{
		Page->TabCtrl = this;
		Tabs.Insert(Page, Where);

		if (IsAttached() AND Tabs.GetItems() == 1)
		{
			Page->Attach(this);
			OnPosChange();
		}

		Invalidate();
	}
	return Page;
}

bool GTabView::Delete(GTabPage *Page)
{
	bool Status = false;
	int Index = Tabs.IndexOf(Page);
	if (Index >= 0)
	{
		if (Index == Current)
		{
			Page->Detach();
		}
		
		Tabs.Delete(Page);
		Status = false;
	}
	
	return Status;
}

GRegion &GTabView::GetTabClient()
{
	return TabClient;
}

void GTabView::OnMouseClick(GMouse &m)
{
	if (m.Down() AND m.Left())
	{
		int Hit = -1;
		GTabPage *p = 0;
		for (p = Tabs.First(); p; p = Tabs.Next())
		{
			if (p->TabPos.Overlap(m.x, m.y))
			{
				Hit = Tabs.IndexOf(p);
				break;
			}
		}
		
		if (Hit >= 0)
		{
			Value(Hit);
		}

		Focus(true);
	}
}

void GTabView::OnKey(GKey &k)
{
	if (k.Down)
	{
		switch (k.c)
		{
			case VK_LEFT:
			{
				if (Current > 0)
				{
					Value(Current - 1);
				}
				break;
			}
			case VK_RIGHT:
			{
				if (Current < Tabs.GetItems() - 1)
				{
					Value(Current + 1);
				}
				break;
			}
		}
	}
}

void GTabView::OnFocus(bool f)
{
	GTabPage *p = Tabs.ItemAt(Current);
	if (p)
	{
		GRegion r = p->TabPos;
		r.y2 += 2;
		Invalidate(&r);
	}
}

void GTabView::OnPaint(GSurface *pDC)
{
	GRegion r = TabClient;

	r.Size(-2, -2);
	LgiWideBorder(pDC, r, RAISED);
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(0, 0, X()-1, TabClient.y1-3);

	GTabPage *Sel = 0;
	int x = r.x1;
	int n=0;
	for (GTabPage *p = Tabs.First(); p; p = Tabs.Next(), n++)
	{
		char *Text = p->Name();
		int Wid = ((Text) ? SysFont->X(Text) : 0) + 13;
		p->TabPos.ZOff(Wid, TabY()-3);
		p->TabPos.Offset(x, 2);
		if (Current != n)
		{
			p->PaintTab(pDC, false);
		}
		else
		{
			Sel = p;
		}
		x += Wid+1;
	}
	
	if (Sel)
	{
		Sel->PaintTab(pDC, true);
	}
}

void GTabView::OnPosChange()
{
	TabClient = GView::GetClient();
	TabClient.Offset(-TabClient.x1, -TabClient.y1);
	TabClient.Size(2, 2);
	TabClient.y1 += TabY();
	
	GTabPage *p = Tabs.ItemAt(Current);
	if (p)
	{
		p->SetPos(TabClient, true);
	}
}

//////////////////////////////////////////////////////////////////////////////////
char *_lgi_gview_cmp(GView *a, GView *b)
{
	static char Str[256];
	if (a AND b)
	{
		sprintf(Str,
				"GView: %p,%p Hnd: %x,%x",
				dynamic_cast<GView*>(a),
				dynamic_cast<GView*>(b),
				a->Handle(),
				b->Handle());
	}
	else
	{
		Str[0] = 0;
	}
	return Str;
}


#ifdef WIN32
GWin32Class PageClass;
#endif

GTabPage::GTabPage(char *name) :
#if defined BEOS
	BView(BRect(0, 0, 100, 100), "GTabPage", B_FOLLOW_ALL_SIDES, B_FRAME_EVENTS),
	GView(this)
#elif defined ATHEOS
	GView(NEW(AView<os::View>(this)))
#else
	GView()
#endif
#ifdef LGI_RES
	, ResObject(Res_Tab)
#endif
{
	Pos.ZOff(100, 100);
	Name(name);
	TabCtrl = 0;

	#if defined BEOS
	SetViewColor(216, 216, 216);
	#elif defined WIN32
	PageClass.Create(Class());
	#endif
}

GTabPage::~GTabPage()
{
}

#ifdef WIN32
char *GTabPage::Class()
{
	return "LGI_TabPage";
}
#endif

void GTabPage::PaintTab(GSurface *pDC, bool Selected)
{
	GRegion r = TabPos;
	if (Selected)
	{
		r.Size(-2, -2);
	}
	
	pDC->Colour(LC_LIGHT, 24);
	pDC->Line(r.x1, r.y1+1, r.x1, r.y2);
	pDC->Line(r.x1+1, r.y1, r.x2-1, r.y1);

	pDC->Colour(LC_HIGH, 24);
	pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2);
	pDC->Line(r.x1+1, r.y1+1, r.x2-1, r.y1+1);

	pDC->Colour(LC_LOW, 24);
	pDC->Line(r.x2-1, r.y1+1, r.x2-1, r.y2);

	pDC->Colour(LC_SHADOW, 24);
	pDC->Line(r.x2, r.y1+1, r.x2, r.y2);

	r.x2 -= 2;
	r.x1 += 2;
	r.y1 += 2;
	
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&r);
	pDC->Set(r.x1, r.y1);
	pDC->Set(r.x2, r.y1);
	
	char *t = Name();
	if (t)
	{
		int Tx, Ty;
		SysFont->Size(&Tx, &Ty, t);
		SysFont->Colour(0, LC_MED);
		SysFont->Transparent(true);
		
		int x = r.x1 + ((r.X()-Tx)/2), y = r.y1 + ((r.Y()-Ty)/2);
		SysFont->Text(pDC, x, y, t);
		
		if (TabCtrl->Focus() AND Selected)
		{
			r.Set(x, y, x+Tx, y+Ty);
			r.Size(-2, -2);
			pDC->Colour(LC_LOW, 24);
			pDC->Box(&r);
		}
	}
}

bool GTabPage::Attach(GView *parent)
{
	bool Status = false;
	if (TabCtrl AND
		GView::Attach(parent))
	{
		for (GView *w = Children.First(); w; w = Children.Next())
		{
			w->Attach(dynamic_cast<GView*>(this));
			w->SetNotify(TabCtrl->GetParent());
		}

		Status = true;
	}

	return Status;
}

char *GTabPage::Name()
{
	return GObject::Name();
}

bool GTabPage::Name(char *name)
{
	return GView::Name(name);
}

int GTabPage::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		#ifdef WIN32
		#endif
	}

	return GView::OnEvent(Msg);
}

bool GTabPage::Enabled()
{
	return true;
}

void GTabPage::Enabled(bool e)
{
}

bool GTabPage::Visible()
{
	return GView::Visible();
}

void GTabPage::Visible(bool v)
{
	GView::Visible(v);
}

bool GTabPage::IsAttached()
{
	#if defined WIN32
	return TabCtrl AND Handle() != 0;
	#elif defined BEOS
	return TabCtrl AND Handle()->Parent() != 0;
	#else
	return false;
	#endif
}

void GTabPage::Append(GView *Wnd)
{
	if (Wnd)
	{
		Wnd->SetParent(this);
		Children.Insert(Wnd);

		if (IsAttached() AND
			TabCtrl AND
			TabCtrl->Value() == TabCtrl->Tabs.IndexOf(this))
		{
			Wnd->Attach(this);
			Wnd->SetNotify(TabCtrl);
		}
	}
}

bool GTabPage::Remove(GView *Wnd)
{
	if (Wnd)
	{
		Children.Delete(Wnd);
		if (Wnd->GetParent())
		{
			Wnd->Detach();
		}
		return true;
	}
	return false;
}

void GTabPage::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);
	GComplexRegion Cr(r);
	pDC->Colour(LC_MED, 24);
	
	for (GView *c = Children.First(); c; c = Children.Next())
	{
		if (NOT dynamic_cast<GRadioGroup*>(c))
			Cr.Subtract(&c->GetPos());
	}
	for (GRegion *i = Cr.First(); i; i = Cr.Next())
	{
		pDC->Rectangle(i);
	}
}

void GTabPage::OnFocus(bool b)
{
}

void GTabPage::OnKey(GKey &k)
{
}

bool GTabPage::AppendDialogFromResource(char *Name)
{
	List<GView> l;
	bool Status = GLgiRes::LoadFromResource(Name, l);

	for (GView *v=l.First(); v; v=l.Next())
	{
		Append(v);
	}

	return Status;
}




