/*
**	FILE:			GToolbar.cpp
**	AUTHOR:			Matthew Allen
**	DATE:				18/10/2001
**	DESCRIPTION:	Toolbar classes
**
**	Copyright (C) 2001, Matthew Allen
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"
#include "Alpha.h"

#ifdef WIN32

HPALETTE GetSystemPalette();
bool BltBmpToBmp(HBITMAP hDest, int xDst, int yDst, int cx, int cy, HBITMAP hSrc, int xSrc, int ySrc, DWORD dwRop);
bool BltBmpToDc(HDC DestDC, int xDst, int yDst, int cx, int cy, HBITMAP hSrc, int xSrc, int ySrc,  DWORD dwRop);
bool BltDcToBmp(HBITMAP hDest, int xDst, int yDst, int cx, int cy, HDC SrcDC, int xSrc, int ySrc, DWORD dwRop);

#define AttachButton(b)	Children.Insert(b);

#else

#define AttachButton(b) b->Attach(this);

#endif
COLOUR Map(GSurface *pDC, COLOUR c);

////////////////////////////////////////////////////////////////////////
GToolBar *LgiLoadToolbar(GView *Parent, char *File, int x, int y)
{
printf("LgiLoadToolbar(0)\n");

	GToolBar *Toolbar = NEW(GToolBar);
	if (Toolbar)
	{
		char *FileName = LgiFindFile(File);

printf("LgiLoadToolbar(1) filename='%s'\n", FileName?FileName:"(null)");

		#if defined BEOS
		int Old = GdcD->SetOption(GDC_IMPORT_8_AS_16, true);
		#endif

		bool Success = FileName AND Toolbar->SetBitmap(FileName, x, y);

printf("LgiLoadToolbar(2) Success=%i\n", Success);

		#if defined BEOS
		GdcD->SetOption(GDC_IMPORT_8_AS_16, Old);
		#endif

		if (NOT Success)
		{
printf("LgiLoadToolbar(3)\n");

			LgiMsg(Parent,
					"Can't find the graphic '%s' for the toolbar.\n"
					"You can find it in this program's archive.",
					"Lgi::LgiLoadToolbar",
					MB_OK,
					File);
			DeleteObj(Toolbar);
		}

printf("LgiLoadToolbar(4)\n");

		DeleteArray(FileName);
	}

printf("LgiLoadToolbar(5)\n");

	return Toolbar;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
#define ImgLst_Empty	0x40000000
#define IgmLst_Add		0x80000000

GImageList::GImageList(int x, int y, GSurface *pDC)
{
	Sx = x;
	Sy = y;
	pCreate = pDC;

	#if defined WIN32

	hImgList = 0;

	#elif defined BEOS

	if (pDC)
	{
		int Bits = (GdcD->GetBits() == 32) ? 32 : 15;
		if (Create(pDC->X(), pDC->Y(), Bits))
		{
			GdcPalette *Pal = pDC->Palette();
			if (Pal)
			{
				Palette(NEW(GdcPalette(Pal)));
			}

			Blt(0, 0, pDC);
			
			COLOUR InitKey = Get(0, 0);

			for (int cy=0; cy<Y(); cy++)
			{
				switch (GetBits())
				{
					case 8:
					{
						uchar Key = InitKey;
						uchar *p = (*this)[cy];
						for (int cx=0; cx<X(); cx++)
						{
							if (p[cx] == Key)
							{
								p[cx] = B_TRANSPARENT_MAGIC_CMAP8;
							}
						}
						break;
					}
					case 15:
					{
						ushort Key = InitKey;
						ushort *p = (ushort*) (*this)[cy];
						ushort Trans = B_TRANSPARENT_MAGIC_RGBA15;
						
						for (int cx=0; cx<X(); cx++)
						{
							if (p[cx] == Trans)
							{
								p[cx] = Trans+1;
							}
							else if (p[cx] == Key)
							{
								p[cx] = Trans;
							}
						}
						break;
					}
					case 32:
					{
						ulong Key = InitKey;
						ulong *p = (ulong*) ((*this)[cy]);
						for (int cx=0; cx<X(); cx++)
						{
							if (	R32(p[cx]) == R32(Key) AND
									G32(p[cx]) == G32(Key) AND
									B32(p[cx]) == B32(Key))
							{
								p[cx] = B_TRANSPARENT_MAGIC_RGBA32;
							}
						}
						break;
					}
				}

			}
		}
	}

	#endif
}

GImageList::~GImageList()
{
	#ifdef WIN32
	if (hImgList)
	{
		ImageList_Destroy(hImgList);
		hImgList = 0;
	}
	#endif
}

bool GImageList::Create(int x, int y, int Bits, int LineLen, bool KeepData)
{
	#ifdef WIN32

	bool Status = GMemDC::Create(x, y, (Bits < 8) ? 24 : Bits, LineLen, KeepData);
	if (Status)
	{
		if (pCreate)
		{
			RGBQUAD *Rgb = Info->bmiColors;

			if (Bits <= 8)
			{
				// create palette system palette
				PALETTEENTRY Entries[256];
				HDC hDC = CreateCompatibleDC(0);
				GetSystemPaletteEntries(hDC, 0, 256, Entries);
				DeleteDC(hDC);

				GdcPalette *Pal = NEW(GdcPalette(0, 256));
				if (Pal)
				{
					for (int i=0; i<256; i++)
					{
						GdcRGB *p = (*Pal)[i];
						if (p)
						{
							Rgb[i].rgbRed = p->R = Entries[i].peRed;
							Rgb[i].rgbGreen = p->G = Entries[i].peGreen;
							Rgb[i].rgbBlue = p->B = Entries[i].peBlue;
							Rgb[i].rgbReserved = 0;
						}
					}
	
					Palette(Pal);
				}
			}

			if (pCreate->GetBits() == 8 AND GetBits() == 8)
			{
				// Set the palette
				HDC hDC = StartDC();
				if (hDC)
				{
					SetDIBColorTable(hDC, 0, 256, Rgb);
					EndDC();
				}

				// gotta remap here
				char Remap[256];
				GdcPalette *SrcPal = pCreate->Palette();
				GdcPalette *DstPal = Palette();
				if (SrcPal AND DstPal)
				{
					// build remap table
					ZeroObj(Remap);
					for (int i=0; i<256; i++)
					{
						GdcRGB *From = (*SrcPal)[i];
						if (From)
						{
							Remap[i] = DstPal->MatchRgb(Rgb24(From->R, From->G, From->B));
						}
					}

					// do blt
					for (int y=0; y<Y(); y++)
					{
						uchar *s = (*pCreate)[y];
						uchar *d = (*this)[y];
						for (int x=0; x<X(); x++)
						{
							*d++ = Remap[*s++];
						}
					}
				}
			}
			else
			{
				// just standard copy blt
				Blt(0, 0, pCreate);
			}
		}

		int Flags = 0;
		switch (GetBits())
		{
			case 8:
			{
				Flags = ILC_COLOR8;
				break;
			}
			case 16:
			{
				Flags = ILC_COLOR16;
				break;
			}
			case 24:
			{
				Flags = ILC_COLOR24;
				break;
			}
			case 32:
			{
				Flags = ILC_COLOR32;
				break;
			}
		}

		int Images = X() / Sx;
		hImgList = ImageList_Create(Sx,
									Sy,
									Flags | ILC_MASK,
									Images,
									0);
		Update(IgmLst_Add);
	}
	return Status;

	#else

	return GMemDC::Create(x, y, Bits, LineLen, KeepData);

	#endif
}

void GImageList::Update(int Flags)
{
	#ifdef WIN32
	if (hImgList)
	{
		if (Flags & ImgLst_Empty)
		{
			ImageList_RemoveAll(hImgList);
		}

		if (Flags & IgmLst_Add)
		{
			int Width = ((X() + 15) / 16) * 2;
			uchar *Data = NEW(uchar[Width * Y()]);
			if (Data)
			{
				// prepare some 2-byte aligned monochrome data for CreateBitmap
				memset(Data, 0, Width * Y());

				COLOUR Key = Get(0, 0);
				for (int y = 0; y < Y(); y++)
				{
					int Iy = Y() - y - 1;
					uchar *d = Data + (y * Width);
					uchar Mask = 0x80;

					for (int x = 0; x < X(); x++)
					{
						COLOUR c = Get(x, y);
						if (c == Key)
						{
							*d |= Mask;
						}

						Mask >>= 1;
						if (NOT Mask)
						{
							Mask = 0x80;
							d++;
						}
					}
				}
			}

			int Index = -1;
			HDC hDC = StartDC();
			HBITMAP bmp = CreateCompatibleBitmap(hDC, X(), Y());
			HBITMAP mask = CreateBitmap(X(), Y(), 1, 1, Data);
			if (bmp)
			{
				BltDcToBmp(bmp, 0, 0, X(), Y(), hDC, 0, 0, SRCCOPY);
				EndDC();
				Index = ImageList_Add(hImgList, bmp, mask);
				DeleteObject(bmp);
				if (mask) DeleteObject(mask);
			}

			DeleteArray(Data);
		}
	}
	#endif
}

void GImageList::Draw(GSurface *pDest, int Dx, int Dy, int Image, int Flags)
{
	#ifdef WIN32
	if (Flags & IMGLST_GDC)
	#endif
	{
		if (pDest)
		{
			COLOUR Key = Get(0, 0);
			GRegion r;
			r.ZOff(Sx-1, Sy-1);
			r.Offset(Image * Sx, 0);

			if (Flags & IMGLST_DISABLED)
			{
				COLOUR Low = CBit(pDest->GetBits(), LC_LOW);
				COLOUR High = CBit(pDest->GetBits(), LC_LIGHT);
				
				for (int y=0; y<r.Y(); y++)
				{
					for (int x=0; x<r.X(); x++)
					{
						if (Get(r.x1 + x, r.y1 + y) != Key)
						{
							pDest->Colour(Low);
							pDest->Set(Dx + x, Dy + y);
							pDest->Colour(High);
							pDest->Set(Dx + x + 1, Dy + y + 1);
						}
					}
				}
			}
			else
			{
				#ifdef BEOS
				GdcViewDC *pViewDC = dynamic_cast<GdcViewDC*>(pDest);
				if (pDest->IsScreen() AND pViewDC)
				{
					BView *owner = pViewDC->Handle();
					BBitmap *Bmp = Handle();

					if (owner AND Bmp)
					{
						int x = Image * Sx;
						BRect S(x, 0, x+Sx-1, Sy-1);
						BRect D(Dx, Dy, Dx+(Sx-1), Dy+(Sx-1));

						drawing_mode Mode = owner->DrawingMode();
						owner->SetDrawingMode(B_OP_OVER);
						owner->DrawBitmap(Bmp, S, D);
						owner->SetDrawingMode(Mode);
					}
				}
				else
				{
				#endif

					for (int y=0; y<r.Y(); y++)
					{
						for (int x=0; x<r.X(); x++)
						{
							COLOUR c = Get(r.x1 + x, r.y1 + y);
							if (c != Key)
							{
								pDest->Colour(c);
								pDest->Set(Dx + x, Dy + y);
							}
						}
					}

				#ifdef BEOS
				}
				#endif
			}
		}
	}
	#ifdef WIN32
	else
	{
		HDC hdc = pDest->StartDC();
		if (hdc)
		{
			int Mode = ILD_NORMAL;
			if (Flags & IMGLST_SELECTED)
			{
				Mode = ILD_BLEND50;
			}
			if (Flags & IMGLST_DISABLED)
			{
				Mode = ILD_FOCUS;
			}

			bool Status = ImageList_DrawEx(	hImgList,
											Image,
											hdc,
											Dx,
											Dy,
											Sx,
											Sy,
											CLR_NONE, // Background colour 
											LC_LIGHT, // Foreground colour
											Mode);

			pDest->EndDC();
		}
	}
	#endif
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GToolButton::GToolButton(int Bx, int By)
#if defined ATHEOS
	: GView(NEW(AView<os::View>(this)))
#elif defined BEOS
	: GView(NEW(BViewRedir(this)))
#endif
{
	Name("");
	Type = TBT_PUSH;
	SetId(IDM_NONE);
	SetDown(false);
	Clicked = false;
	Over = false;
	ImgIndex = 0;

	Pos.ZOff(Bx+1, By+1);
	SetParent(0);
	TipId = -1;

	#ifdef BEOS
	_BorderSize = 0;
	Handle()->SetViewColor(B_TRANSPARENT_COLOR);
	#endif
}

GToolButton::~GToolButton()
{
}

void GToolButton::OnPaint(GSurface *pDC)
{
	GToolBar *Par = (GToolBar*) Parent;

	if (NOT Enabled() AND Over)
	{
		Over = false;
	}

	if (Par AND Par->ImgList)
	{
		#ifdef TOOLBAR_DOUBLE_BUFFER
		GMemDC *pMemDC = NEW(GMemDC);
		int MemBits = GdcD->GetBits();

		if (pMemDC AND
			pMemDC->Create(	Pos.X(),
							Pos.Y(),
							(MemBits < 8) ? 24 : MemBits))
		{
		#else
		GSurface *pMemDC = pDC;
		#endif
			int Px = Pos.X()-1;
			int Py = Pos.Y()-1;

			// Draw Background
			COLOUR Grey = Over ? LC_HIGH : LC_MED;
			pMemDC->Colour(Grey, 24);
			pMemDC->Rectangle();

			int BaseX = (Pos.X() - (2*BORDER_BUTTON) - Par->Bx) / 2;
			int Off = BORDER_BUTTON + ((Down) ? 1 : 0);
			
			if (GetId() >= 0)
			{
				// button
				int Flags = IMGLST_GDC;

				if (Over)
				{
					// draw transparent
					Par->ImgList->Draw(pMemDC, BaseX+Off, Off, ImgIndex, Flags);
				}
				else
				{
					// draw cached
					Par->_DrawFromCache(pMemDC, BaseX+Off, Off, ImgIndex, NOT Enabled());
				}

				// draw border
				if (Down)
				{
					pMemDC->Colour(Map(pMemDC, (Down) ? LC_LIGHT : LC_LOW), 24);
					pMemDC->Line(Px, Py, Px, 0);
					pMemDC->Line(Px, Py, 0, Py);

					pMemDC->Colour(Map(pMemDC, (Down) ? LC_LOW : LC_LIGHT), 24);
					pMemDC->Line(0, 0, 0, Py);
					pMemDC->Line(0, 0, Px, 0);
				}

				// text
				char *n = Name();
				if (Par->Text AND
					Par->Font AND
					n)
				{
					// write each word centered on a different line
					int TextY = Off+Par->By;
					Par->Font->Colour(Enabled() ? LC_TEXT : LC_LOW, LC_MED);

					while (*n)
					{
						char *e;
						for (e=n; *e; e++)
						{
							if (*e == ' ')
							{
								if (e[1] AND e[2] == ' ')
								{
									continue;
								}
								break;
							}
						}

						int l = (int)e-(int)n;
						int _x = 0;
						int _y = 0;

						Par->Font->Size(&_x, &_y, n, l);
						Par->Font->Text(	pMemDC,
											Off + ((Pos.X()-_x)/2),
											TextY,
											n,
											l);
						TextY += _y;
						n = e;
						if (*n == ' ') n++;
					}
				}
			}
			else
			{
				// separator
				if (X() > Y())
				{
					int c = Y()/2-1;
					pMemDC->Colour(Map(pMemDC, LC_LOW), 24);
					pMemDC->Line(2, c, Px-2, c);
					pMemDC->Colour(Map(pMemDC, LC_LIGHT), 24);
					pMemDC->Line(2, c+1, Px-2, c+1);
				}
				else
				{
					int c = X()/2-1;
					pMemDC->Colour(Map(pMemDC, LC_LOW), 24);
					pMemDC->Line(c, 2, c, Py-2);
					pMemDC->Colour(Map(pMemDC, LC_LIGHT), 24);
					pMemDC->Line(c+1, 2, c+1, Py-2);
				}
			}

		#ifdef TOOLBAR_DOUBLE_BUFFER
			pDC->Blt(0, 0, pMemDC);
		}

		DeleteObj(pMemDC);
		#endif
	}
}

void GToolButton::Value(int b)
{
	switch (Type)
	{
		case TBT_PUSH:
		{
			// do nothing... can't set value
			break;
		}

		case TBT_TOGGLE:
		{
			SetDown(b);
			Invalidate();
			break;
		}

		case TBT_RADIO:
		{
			if (Parent AND b)
			{
				// Clear any other radio buttons that are down
				GToolButton *But;
				for (	But = (GToolButton*) Parent->Children.Find(this);
						But AND But->GetId() >= 0;
						But = (GToolButton*) Parent->Children.Next())
				{
					if (But->Type == TBT_RADIO AND
						But != this AND
						But->Down)
					{
						But->SetDown(false);
						But->Invalidate();
					}
				}

				for (	But = (GToolButton*) Parent->Children.Find(this);
						But AND But->GetId() >= 0;
						But = (GToolButton*) Parent->Children.Prev())
				{
					if (But->Type == TBT_RADIO AND
						But != this AND
						But->Down)
					{
						But->SetDown(false);
						But->Invalidate();
					}
				}
			}

			SetDown(b);
			Invalidate();
			break;
		}
	}
}

void GToolButton::OnCommand(GMouse &m)
{
	if (Parent)
	{
		((GToolBar*) Parent)->OnButtonClick(this);
	}
}

void GToolButton::SendNotify(int Flags)
{
	GView *w = (GetNotify()) ? GetNotify() : GetParent();
	if (w)
	{
		w->PostEvent(WM_CHANGE, (int)((GWindow*)this), Flags);
	}
}

void GToolButton::OnMouseClick(GMouse &m)
{
	// if (m.Left())
	{
		// left click action...
		if (GetId() >= 0 AND Enabled())
		{
			switch (Type)
			{
				case TBT_PUSH:
				{
					bool Old = Down;

					Clicked = m.Down();
					Capture(m.Down());

					if (Old AND IsOver(m))
					{
						if (m.Left())
						{
							OnCommand(m);
						}
						SendNotify(m.Flags);
					}

					SetDown(m.Down());
					if (Old != Down)
					{
						Invalidate();
					}
					break;
				}

				case TBT_TOGGLE:
				{
					if (m.Down())
					{
						if (m.Left())
						{
							Value(NOT Down);
							OnCommand(m);
						}
						SendNotify(m.Flags);
					}
					break;
				}

				case TBT_RADIO:
				{
					if (m.Down())
					{
						if (NOT Down AND m.Left())
						{
							Value(true);
							OnCommand(m);
						}
						SendNotify(m.Flags);
					}
					break;
				}
			}
		}
	}
}

void GToolButton::OnMouseEnter(GMouse &m)
{
	if (NOT Separator() AND Enabled())
	{
		Over = true;
		Invalidate();
	}

	if (Clicked)
	{
		SetDown(true);
		Invalidate();
	}
	else
	{
		#ifdef WIN32
		if (TipId < 0)
		{
			GToolBar *Bar = (GToolBar*) Parent;
			if (Bar AND Bar->hTip)
			{
				TOOLINFO ti;

				ZeroObj(ti);
				ti.cbSize = sizeof(ti);
				ti.uFlags = TTF_SUBCLASS;
				ti.hwnd = GetParent()->Handle();
				ti.rect = GetPos();
				ti.lpszText = Name();
				ti.uId = TipId = Bar->Children.IndexOf(this);

				SendMessage(Bar->hTip, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
			}
		}
		#endif

		((GToolBar*) Parent)->PostDescription(this, Name());
	}
}

void GToolButton::SetDown(bool d)
{
	Down = d;
}

void GToolButton::OnMouseMove(GMouse &m)
{
	#ifdef BEOS
	if (GetParent())
	{
		((GToolBar*) GetParent())->PostDescription(this, Name());
	}
	#endif
}

void GToolButton::OnMouseExit(GMouse &m)
{
	if (Over)
	{
		Over = false;
		Invalidate();
	}

	if (Clicked)
	{
		SetDown(false);
		Invalidate();
	}
	else
	{
		((GToolBar*) Parent)->PostDescription(this, "");
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GToolBar::GToolBar()
{
	Name("LGI_Toolbar");
	Bx = By = 0;
	Sx = Sy = 10;
	Vertical = false;
	Text = false;

	// Setup tool button font
	GdcFontType SysFontType;
	if (SysFontType.GetSystemFont("Small"))
	{
		Font = SysFontType.CreateFont();
		if (Font)
		{
			Font->Colour(0);
			Font->Bold(false);
			Font->Transparent(true);
		}
	}

	LastIndex = 0;
	OwnImgList = false;
	ImgList = 0;
	pColour = 0;
	pDisabled = 0;

	#if defined WIN32
	hTip = NULL;
	#elif defined BEOS
	Handle()->SetViewColor(B_TRANSPARENT_COLOR);
	#endif
}

GToolBar::~GToolBar()
{
	_BuildCache(0);

	if (OwnImgList)
	{
		DeleteObj(ImgList);
	}

	#ifdef WIN32
	if (hTip)
	{
		DestroyWindow(hTip);
	}
	#endif
}

void GToolBar::_BuildCache(GImageList *From)
{
	DeleteObj(pColour);
	DeleteObj(pDisabled);

	if (From)
	{
		#ifdef BEOS
		int Bits = From->GetBits();
		#else
		int Bits = GdcD->GetBits();
		#endif
		
		pColour = NEW(GMemDC);
		if (pColour AND
			pColour->Create(From->X(), From->Y(), Bits))
		{
			pColour->Colour(LC_MED, 24);
			pColour->Rectangle();

			for (int i=0; i<From->GetItems(); i++)
			{
				From->Draw(pColour, i * From->TileX(), 0, i);
			}
		}

		pDisabled = NEW(GMemDC);
		if (pDisabled AND
			pDisabled->Create(From->X(), From->Y(), Bits))
		{
			pDisabled->Colour(LC_MED, 24);
			pDisabled->Rectangle();
			pDisabled->Op(GDC_ALPHA);
			GAlphaApp *pApp = (GAlphaApp*) pDisabled->Applicator();
			if (pApp)
			{
				pApp->SetAlpha(40, pDisabled->Palette());
			}

			for (int i=0; i<From->GetItems(); i++)
			{
				From->Draw(pDisabled, i * From->TileX(), 0, i, IMGLST_GDC);
			}
		}

	}
}

void GToolBar::_DrawFromCache(GSurface *pDC, int x, int y, int Index, bool Disabled)
{
	if (pDC AND
		pDisabled AND
		pColour AND
		ImgList)
	{
		GRegion s;
		s.ZOff(ImgList->TileX()-1, ImgList->TileY()-1);
		s.Offset(ImgList->TileX()*Index, 0);
		pDC->Blt(x, y, Disabled ? pDisabled : pColour, &s);
	}
}

bool GToolBar::Pour(GComplexRegion &r)
{
	if (ImgList)
	{
		int PosX = BORDER_SHADE + BORDER_SPACER;
		int PosY = BORDER_SHADE + BORDER_SPACER;
		int SrcX = 0;
		int SrcY = 0;
		int EndX = 0;
		int EndY = 0;
		int MaxDim = 0;

		for (GView *w = Children.First(); w; w = Children.Next())
		{
			MaxDim = max(MaxDim, Vertical ? w->X() : w->Y());
		}

		GRegion ButPos;
		GView *But = Children.First();
		while (But)
		{
			ButPos = But->GetPos();
			if (Vertical)
			{
				if (ButPos.X() < MaxDim)
				{
					ButPos.Dimension(MaxDim, ButPos.Y());
				}
			}
			else
			{
				if (ButPos.Y() < MaxDim)
				{
					ButPos.Dimension(ButPos.X(), MaxDim);
				}
			}

			if (But->GetId() == IDM_SEPARATOR)
			{
				// separator
				if (Vertical)
				{
					ButPos.ZOff(MaxDim-1, BORDER_SEPARATOR+1);
				}
				else
				{
					ButPos.ZOff(BORDER_SEPARATOR+1, MaxDim-1); // +(2*BORDER_BUTTON)-1
				}
			}
			else if (But->GetId() >= 0)
			{
				// other control
			}

			ButPos.Offset(PosX - ButPos.x1, PosY - ButPos.y1);

			if (But->GetId() == IDM_BREAK)
			{
				ButPos.ZOff(0, 0);
				if (Vertical)
				{
					PosX = EndX;
					PosY = BORDER_SHADE + BORDER_SPACER;
				}
				else
				{
					PosX = BORDER_SHADE + BORDER_SPACER;
					PosY = EndY;
				}
			}
			else
			{
				if (Vertical)
				{
					EndX = max(EndX, ButPos.x2 + 1);
					PosY = ButPos.y2 + 1;
				}
				else
				{
					PosX = ButPos.x2 + 1;
					EndY = max(EndY, ButPos.y2 + 1);
				}
			}

			But->SetPos(ButPos);
			But = Children.Next();
		}

		if (Vertical)
		{
			PosX = EndX;
		}
		else
		{
			PosY = EndY;
		}

		Sx = PosX + BORDER_SHADE + BORDER_SPACER;
		Sy = PosY + BORDER_SHADE + BORDER_SPACER;
		Pos.ZOff(Sx-1, Sy-1);

		// GRegion *Best = FindLargest(r);
		GRegion *Best = FindLargestEdge(r, GV_EDGE_TOP);
		if (Best)
		{
			Pos.Offset(Best->x1, Best->y1);
			Pos.Bound(Best);
			SetPos(Pos);

			return TRUE;
		}
	}

	return FALSE;
}

void GToolBar::OnButtonClick(GToolButton *Btn)
{
	GView *w = (GetNotify()) ? GetNotify() : GetParent();
	if (w AND Btn)
	{
		w->PostEvent(WM_COMMAND, Btn->GetId(), (int) Handle());
	}
}

int GToolBar::PostDescription(GView *Ctrl, char *Text)
{
	if (GetParent())
	{
		return GetParent()->PostEvent(WM_DESCRIBE, (int) Ctrl, (int) Text);
	}
	return 0;
}

int GToolBar::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case WM_CHANGE:
		{
			if (GetParent())
			{
				return GetParent()->OnEvent(Msg);
			}
			break;
		}
	}

	return GView::OnEvent(Msg);
}

void GToolBar::OnPaint(GSurface *pScr)
{
	GRegion r(0, 0, X()-1, Y()-1);

	// Draw a border
	LgiThinBorder(pScr, r, RAISED);
	pScr->Colour(LC_MED, 24);
	pScr->Box(&r);
	r.Size(1, 1);

	// Draw all the children
	GView::OnPaint(pScr);
}

void GToolBar::OnMouseClick(GMouse &m)
{
}

void GToolBar::OnMouseEnter(GMouse &m)
{
	#ifdef WIN32
	if (NOT hTip)
	{
		hTip = CreateWindowEx(	NULL, TOOLTIPS_CLASS, NULL,
								WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP,
								CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
								GetParent()->Handle(),
								NULL,
								NULL,
								NULL);

		if (hTip)
		{
			SetWindowPos(		hTip,
								HWND_TOPMOST,
								0, 0, 0, 0,
								SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
		}
	}
	#endif
}

void GToolBar::OnMouseExit(GMouse &m)
{
}

void GToolBar::OnMouseMove(GMouse &m)
{
}

bool GToolBar::SetBitmap(char *File, int bx, int by)
{
	bool Status = false;
	
	GSurface *pDC = LoadDC(File);
	if (pDC)
	{
		Status = SetDC(pDC, bx, by);
		DeleteObj(pDC);
	}

	return Status;
}

bool GToolBar::SetDC(GSurface *pNewDC, int bx, int by)
{
	if (OwnImgList)
	{
		DeleteObj(ImgList);
	}
	Bx = bx;
	By = by;

	if (pNewDC)
	{
		ImgList = NEW(GImageList(bx, by, pNewDC));
		if (ImgList)
		{
			#ifdef WIN32
			if (ImgList->Create(pNewDC->X(), pNewDC->Y(), GdcD->GetBits()))
			#endif
			{
				_BuildCache(ImgList);
				return true;
			}
		}
	}
	return false;
}

bool GToolBar::SetImageList(GImageList *l, int bx, int by, bool Own)
{
	if (OwnImgList)
	{
		DeleteObj(ImgList);
	}
	OwnImgList = Own;
	Bx = bx;
	By = by;

	ImgList = l;
	_BuildCache(ImgList);

	return ImgList != 0;
}

GToolButton *GToolBar::AppendButton(char *Tip, int Id, int Type, int Enabled, int IconId)
{
	int x = 0, y = 0;

	#if defined ATHEOS

	// FIXME
	char *name = Tip;

	#else

	char *name = 0;

	if (Tip)
	{
		// Process '&'s
		name = NewStr(Tip);
		if (name)
		{
			char *Src = Tip;
			char *Dest = name;
			while (*Src)
			{
				if (Src[0] == '&')
				{
					if (Src[1] == '&') *Dest++ = '&';
					Src++;
				}
				else
				{
					*Dest++ = *Src++;
				}
			}
			*Dest++ = 0;
		}

		// work out the space required to
		// write each word centered on a different line
		if (Font AND
			Text)
		{
			char *n = name;
			while (*n)
			{
				char *e;
				for (e=n; *e; e++)
				{
					if (*e == ' ')
					{
						if (e[1] AND e[2] == ' ')
						{
							continue;
						}
						break;
					}
				}
				int l = (int)e-(int)n;
				int _x = 0, _y = 0;

				Font->Size(&_x, &_y, n, l);
				x = max(x, _x);
				y += _y;

				n = e;
				if (*n == ' ') n++;
			}
		}
	}

	#endif

	GToolButton *But = NEW(GToolButton(max(Bx, x + 4), By + y));
	if (But)
	{
		But->Name(name);
		But->SetId(Id);
		But->Type = Type;
		But->Parent = this;
		But->Enabled(Enabled);
		
		if (IconId >= 0)
		{
			But->ImgIndex = IconId;
		}
		else
		{
			But->ImgIndex = LastIndex++;
		}

		AttachButton(But);
	}

	// DeleteArray(name);

	return But;
}

bool GToolBar::AppendSeparator()
{
	GToolButton *But = NEW(GToolButton(Bx, By));
	if (But)
	{
		But->SetId(IDM_SEPARATOR);
		But->Parent = this;
		AttachButton(But);
		return TRUE;
	}
	return FALSE;
}

bool GToolBar::AppendBreak()
{
	GToolButton *But = NEW(GToolButton(Bx, By));
	if (But)
	{
		But->SetId(IDM_BREAK);
		But->Parent = this;
		AttachButton(But);
		return true;
	}
	return false;
}

bool GToolBar::AppendControl(GView *Ctrl)
{
	bool Status = false;
	if (Ctrl)
	{
		Ctrl->SetParent(this);
		AttachButton(Ctrl);
		Status = true;
	}
	return Status;
}

void GToolBar::Empty()
{
	for (GView *But = Children.First(); But; But = Children.Next())
	{
		DeleteObj(But);
	}
}

///////////////////////////////////////////////////////////////////////
COLOUR Map(GSurface *pDC, COLOUR c)
{
	if (pDC AND pDC->GetBits() <= 8)
	{
		if (pDC->IsScreen())
		{
			c = CBit(24, c);
		}
		#ifdef WIN32
		else
		{
			HPALETTE hPal = GetSystemPalette();
			if (hPal)
			{
				c = GetNearestPaletteIndex(hPal, c);
				DeleteObject(hPal);
			}
		}
		#endif
	}
	return c;
}

#ifdef WIN32
HPALETTE GetSystemPalette()
{
	HPALETTE hPal = 0;
	LOGPALETTE *Log = (LOGPALETTE*) NEW(uchar[sizeof(LOGPALETTE) + (sizeof(PALETTEENTRY) * 255)]);
	if (Log)
	{
		Log->palVersion = 0x300;
		Log->palNumEntries = 256;

		HDC hDC = CreateCompatibleDC(0);
		GetSystemPaletteEntries(hDC, 0, 256, Log->palPalEntry);
		DeleteDC(hDC);

		hPal = CreatePalette(Log);
	}

	return hPal;
}

bool BltBmpToBmp(HBITMAP hDest,
		int xDst,
		int yDst,
		int cx,
		int cy,
		HBITMAP hSrc,
		int xSrc,
		int ySrc,
		DWORD dwRop)
{
	bool Status = FALSE;
	HDC DestDC = CreateCompatibleDC(0);
	HDC SrcDC = CreateCompatibleDC(0);

	if (DestDC AND SrcDC)
	{
		hDest = (HBITMAP) SelectObject(DestDC, hDest);
		hSrc = (HBITMAP) SelectObject(SrcDC, hSrc);

		Status = BitBlt(DestDC,
				xDst,
				yDst,
				cx,
				cy,
				SrcDC,
				xSrc,
				ySrc,
				dwRop);

		hDest = (HBITMAP) SelectObject(DestDC, hDest);
		hSrc = (HBITMAP) SelectObject(SrcDC, hSrc);
	}

	if (DestDC)
	{
		DeleteDC(DestDC);
	}
	if (SrcDC)
	{
		DeleteDC(SrcDC);
	}

	return Status;
}

bool BltBmpToDc(HDC DestDC,
		int xDst,
		int yDst,
		int cx,
		int cy,
		HBITMAP hSrc,
		int xSrc,
		int ySrc,
		DWORD dwRop)
{
	bool Status = FALSE;
	HDC SrcDC = CreateCompatibleDC(0);

	if (DestDC AND SrcDC)
	{
		hSrc = (HBITMAP) SelectObject(SrcDC, hSrc);

		Status = BitBlt(DestDC,
				xDst,
				yDst,
				cx,
				cy,
				SrcDC,
				xSrc,
				ySrc,
				dwRop);

		hSrc = (HBITMAP) SelectObject(SrcDC, hSrc);
	}

	if (SrcDC)
	{
		DeleteDC(SrcDC);
	}

	return Status;
}

bool BltDcToBmp(HBITMAP hDest,
		int xDst,
		int yDst,
		int cx,
		int cy,
		HDC SrcDC,
		int xSrc,
		int ySrc,
		DWORD dwRop)
{
	bool Status = FALSE;
	HDC DestDC = CreateCompatibleDC(0);

	if (DestDC AND SrcDC)
	{
		hDest = (HBITMAP) SelectObject(DestDC, hDest);

		Status = BitBlt(DestDC,
				xDst,
				yDst,
				cx,
				cy,
				SrcDC,
				xSrc,
				ySrc,
				dwRop);

		hDest = (HBITMAP) SelectObject(DestDC, hDest);
	}

	if (DestDC)
	{
		DeleteDC(DestDC);
	}

	return Status;
}
#endif






