#include "Lgi.h"

#define IDC_H				10
#define IDC_V				11

#define TREE_BLOCK			16

#define DRAG_THRESHOLD		4

//////////////////////////////////////////////////////////////////////////////
GTreeItem::GTreeItem()
{
	Tree = 0;
	Parent = 0;
	Str = 0;
	Sys_Image = -1;

	_Pos.ZOff(-1, -1);
	_Open = false;
	_Last = false;
	_Selected = false;
	_Depth = 0;
}

GTreeItem::~GTreeItem()
{
	int y = Parent ? Parent->_Pos.y1 : _Pos.y1;

	_Remove();
	DeleteArray(Str);

	if (Tree)
	{
		Tree->_UpdateBelow(y);
	}
}

void GTreeItem::_Remove()
{
	for (GTreeItem *i=Items.First(); i; i=Items.First())
	{
		i->_Remove();
	}

	if (Parent)
	{
		Parent->Items.Delete(this);
	}
	else
	{
		Tree->Items.Delete(this);
	}
	if (_Selected)
	{
		Tree->_Selection.Delete(this);
		_Selected = false;
	}

	if (Tree)
	{
		Tree->_LayoutDirty = true;
	}
}

void GTreeItem::_PourText(GdcPt2 &Size)
{
	char *t = GetText();
	Size.x = (t ? SysFont->X(t) : 0) + 4;
	Size.y = 0;
}

void GTreeItem::_PaintText(GSurface *pDC, COLOUR Fore, COLOUR Back)
{
	char *Text = GetText();
	if (Text)
	{
		int Tab = SysFont->TabSize();
		SysFont->TabSize(0);
		SysFont->Transparent(false);
		SysFont->Colour(Fore, Back);
		SysFont->Text(pDC, _Text.x1 + 2, _Text.y1 + 1, Text, -1, &_Text);
		SysFont->TabSize(Tab);
	}
	else
	{
		pDC->Colour(Back, 24);
		pDC->Rectangle(&_Text);
	}
}

void GTreeItem::_Pour(GdcPt2 *Limit, int Depth, bool Visible)
{
	_Visible = Visible;
	_Depth = Depth;

	if (_Visible)
	{
		GdcPt2 TextSize;
		_PourText(TextSize);
		int IconX = (GetImage() >= 0) ? TREE_BLOCK : 0;

		_Pos.ZOff(	(TREE_BLOCK*_Depth) +	// trunk
					TREE_BLOCK +			// node
					IconX +					// icon if present
					TextSize.x,				// text if present
					max(TREE_BLOCK, SysFont->Y("A"))-1);
		_Pos.Offset(0, Limit->y);

		Limit->x = max(Limit->x, _Pos.x2 + 1);
		Limit->y = max(Limit->y, _Pos.y2 + 1);
	}
	else
	{
		_Pos.ZOff(-1, -1);
	}

	GTreeItem *n;
	for (GTreeItem *i=Items.First(); i; i=n)
	{
		n=Items.Next();
		i->_Last = n == 0;
		i->_Pour(Limit, Depth+1, _Open AND _Visible);
	}
}

char *GTreeItem::GetText(int i)
{
	return Str;
}

int GTreeItem::IndexOf()
{
	if (Parent)
	{
		return Parent->Items.IndexOf(this);
	}
	else if (Tree)
	{
		return Tree->Items.IndexOf(this);
	}

	return -1;
}

GTreeItem *GTreeItem::GetChild()
{
	return Items.First();
}

GTreeItem *GTreeItem::GetPrev()
{
	List<GTreeItem> *l = (Parent) ? &Parent->Items : &Tree->Items;
	int Index = l->IndexOf(this);
	if (Index >= 0)
	{
		return l->ItemAt(Index-1);
	}

	return 0;
}

GTreeItem *GTreeItem::GetNext()
{
	List<GTreeItem> *l = (Parent) ? &Parent->Items : &Tree->Items;
	int Index = l->IndexOf(this);
	if (Index >= 0)
	{
		return l->ItemAt(Index+1);
	}

	return 0;
}

bool GTreeItem::SetText(char *s, int i)
{
	char *New = NewStr(s);
	DeleteArray(Str);
	Str = New;

	if (Tree)
	{
		Update();
	}

	return TRUE;
}

int GTreeItem::GetImage(int Flags)
{
	return Sys_Image;
}

void GTreeItem::SetImage(int i)
{
	Sys_Image = i;
}

GTreeItem *GTreeItem::Insert(GTreeItem *Obj, int Pos)
{
	GTreeItem *NewObj = (Obj) ? Obj : NEW(GTreeItem);
	if (NewObj)
	{
		NewObj->Parent = this;
		NewObj->Tree = Tree;

		Items.Delete(NewObj);
		Items.Insert(NewObj, Pos);

		if (Tree)
		{
			Tree->_LayoutDirty = true;
			if (_Pos.Y() > 0)
			{
				Tree->_UpdateBelow(_Pos.y1);
			}
			else
			{
				Tree->Invalidate();
			}
		}
	}

	return NewObj;
}

void GTreeItem::Remove()
{
	_Remove();

	if (Tree)
	{
		Tree->_UpdateBelow(_Pos.y1);
	}
}


void GTreeItem::Update()
{
	if (Tree)
	{
		Tree->_Update(&_Pos, true);
	}
}

bool GTreeItem::Selected()
{
	return _Selected;
}

void GTreeItem::Selected(bool b)
{
	if (_Selected != b)
	{
		_Selected = b;
		Update();

		if (b AND Tree)
		{
			Tree->_OnSelect(this);
			Tree->OnItemSelect(this);
		}
	}
}

bool GTreeItem::Expanded()
{
	return _Open;
}

void GTreeItem::Expanded(bool b)
{
	if (_Open != b)
	{
		_Open = b;

		if (Items.GetItems() > 0)
		{
			if (Tree)
			{
				Tree->_LayoutDirty = true;
				Tree->_UpdateBelow(_Pos.y1);
			}
			OnExpand(b);
		}
	}
}

GTreeItem *GTreeItem::_HitTest(int x, int y)
{
	GTreeItem *Status = 0;

	if (_Pos.Overlap(x, y) AND
		x > (_Depth*TREE_BLOCK))
	{
		Status = this;
	}

	if (_Open)
	{
		for (GTreeItem *i=Items.First(); i AND NOT Status; i=Items.Next())
		{
			Status = i->_HitTest(x, y);
		}
	}

	return Status;
}

void GTreeItem::_MouseClick(GMouse &m)
{
	if (m.Down())
	{
		if (Items.GetItems() > 0 AND
			_Thumb.Overlap(m.x, m.y))
		{
			Expanded(NOT Expanded());
		}

		if (_Text.Overlap(m.x, m.y) OR
			_Icon.Overlap(m.x, m.y))
		{
			Selected(true);
		}
	}

	if (Tree)
	{
		Tree->OnItemClick(this, m);
	}
}

void GTreeItem::OnMouseClick(GMouse &m)
{
}

void GTreeItem::OnPaint(GSurface *pDC, GRegion &r, COLOUR Back)
{
	bool IsSelected = (Tree->_DropTarget == this) OR (Tree->_DropTarget == 0 AND _Selected);

	// background up to text
	pDC->Colour(LC_WORKSPACE, 24);
	pDC->Rectangle(0, _Pos.y1, (_Depth*TREE_BLOCK)+TREE_BLOCK, _Pos.y2);

	// draw trunk
	int x = 0;
	pDC->Colour(LC_MED, 24);
	for (int i=0; i<_Depth; i++)
	{
		if (Tree->_LineFlags[0] & (1 << i))
		{
			pDC->Line(x + 8, _Pos.y1, x + 8, _Pos.y2);
		}
		x += TREE_BLOCK;
	}

	// draw node
	if (Items.GetItems() > 0)
	{
		// plus/minus symbol
		_Thumb.ZOff(8, 8);
		_Thumb.Offset(x + 4, _Pos.y1 + 4);

		pDC->Colour(LC_LOW, 24);
		pDC->Box(&_Thumb);
		pDC->Colour(LC_WHITE, 24);
		pDC->Rectangle(_Thumb.x1+1, _Thumb.y1+1, _Thumb.x2-1, _Thumb.y2-1);
		pDC->Colour(LC_SHADOW, 24);
		pDC->Line(	_Thumb.x1+2,
					_Thumb.y1+4,
					_Thumb.x1+6,
					_Thumb.y1+4);

		pDC->Colour(LC_MED, 24);

		if (Parent OR IndexOf() > 0)
		{
			// draw line to item above
			pDC->Line(x + 8, _Pos.y1, x + 8, _Thumb.y1-1);
		}

		// draw line to leaf beside
		pDC->Line(_Thumb.x2+1, _Pos.y1 + 8, x + (TREE_BLOCK-1), _Pos.y1 + 8);

		if (NOT _Last)
		{
			// draw line to item below
			pDC->Line(x + 8, _Thumb.y2+1, x + 8, _Pos.y2);
		}

		if (NOT _Open)
		{
			pDC->Colour(LC_SHADOW, 24);
			pDC->Line(	_Thumb.x1+4,
						_Thumb.y1+2,
						_Thumb.x1+4,
						_Thumb.y1+6);
		}
	}
	else
	{
		// leaf node
		pDC->Colour(LC_MED, 24);
		if (_Last)
		{
			pDC->Rectangle(x + 8, _Pos.y1, x + 8, _Pos.y1 + 8);
		}
		else
		{
			pDC->Rectangle(x + 8, _Pos.y1, x + 8, _Pos.y2);
		}
		pDC->Rectangle(x + 8, _Pos.y1 + 8, x + (TREE_BLOCK-1), _Pos.y1 + 8);
	}
	x += TREE_BLOCK;

	// draw icon
	int Image = GetImage(Selected());
	if (Image >= 0 AND
		Tree->GetImageList())
	{
		_Icon.ZOff(TREE_BLOCK-1, TREE_BLOCK-1);
		_Icon.Offset(x, _Pos.y1);

		if (Tree->_IconCache)
		{
			// no flicker
			GRegion From;
			From.ZOff(TREE_BLOCK-1, TREE_BLOCK-1);
			From.Offset(Tree->GetImageList()->TileX()*Image, 0);
			pDC->Blt(_Icon.x1, _Icon.y1, Tree->_IconCache, &From);
		}
		else
		{
			// flickers...
			Tree->GetImageList()->Draw(pDC, x, _Pos.y1, Image);
		}

		if (_Icon.y2 < _Pos.y2)
		{
			pDC->Colour(LC_WORKSPACE, 24);
			pDC->Rectangle(x, _Icon.y2, _Icon.x2, _Pos.y2);
		}

		x += TREE_BLOCK;
	}

	// text
	char *Text = GetText();
	if (Text)
	{
		GdcPt2 TextSize;
		_PourText(TextSize);

		_Text.ZOff(TextSize.x, _Pos.Y()-1);
		_Text.Offset(x, _Pos.y1);
		_PaintText(	pDC,
					(IsSelected) ? LC_SEL_TEXT : LC_TEXT,
					(IsSelected) ? Back : LC_WORKSPACE);
		_Pos.x2 = _Text.x2;
	}

	// background after text
	pDC->Colour(LC_WORKSPACE, 24);
	pDC->Rectangle(_Text.x2, _Pos.y1, max(Tree->X(), Tree->_Limit.x), _Pos.y2);

	// children
	if (_Open)
	{
		if (NOT _Last)
		{
			Tree->_LineFlags[0] |= 1 << _Depth;
		}

		for (GTreeItem *i=Items.First(); i; i=Items.Next())
		{
			i->OnPaint(pDC, r, Back);
		}

		Tree->_LineFlags[0] &= ~(1 << _Depth);
	}
}

//////////////////////////////////////////////////////////////////////////////
GTree::GTree(int id, int x, int y, int cx, int cy, char *name) :
#ifdef LGI_RES
	ResObject(Res_TreeView)
#endif
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	if (name) Name(name);

	Lines = true;
	Buttons = true;
	LinesAtRoot = true;
	EditLabels = false;
	Tree = this;
	MultipleSelect = false;

	_LastHit = 0;
	_DropTarget = 0;
	_IconCache = 0;
	_LayoutDirty = true;

	#ifdef WIN32
	WndStyle |= WS_CHILD | WS_VISIBLE | WS_TABSTOP;
	#endif
}

GTree::~GTree()
{
	Empty();
}

// Internal tree methods
void GTree::_Update(GRegion *r, bool Now)
{
	if (r)
	{
		GRegion u = *r;
		GdcPt2 s = _ScrollPos();
		u.Offset(-s.x, -s.y);
		Invalidate(&u);
	}
	else
	{
		Invalidate();
	}

	#ifdef WIN32
	if (Now)
	{
		UpdateWindow(Handle());
	}
	#endif
}

void GTree::_UpdateBelow(int y, bool Now)
{
	GdcPt2 s = _ScrollPos();
	GRegion u(0, y - s.y, X()-1, Y()-1);
	Invalidate(&u);

	#ifdef WIN32
	if (Now)
	{
		UpdateWindow(Handle());
	}
	#endif
}

GdcPt2 GTree::_ScrollPos()
{
	int All = _Limit.y / TREE_BLOCK - 1;
	int Visible = Y() / TREE_BLOCK;
	GdcPt2 Status;

	Status.x = (HScroll) ? HScroll->Pos() : 0;
	Status.y = (VScroll) ? min(VScroll->Pos(), All-Visible+1) * TREE_BLOCK : 0;

	return Status;
}

void GTree::_UpdateScrollBars()
{
	static bool Processing = false;
	if (NOT Processing)
	{
		Processing = true;

		GRegion Client = GetClient();
		SetScrollBars(_Limit.x > Client.X(), _Limit.y > Client.Y());

		// x scroll... in pixels
		if (HScroll)
		{
			HScroll->SetId(IDC_H);
			HScroll->SetLimits(0, _Limit.x-1);
			HScroll->SetPage(X());

			int Max = _Limit.x - X();
			if (HScroll->Pos() > Max)
			{
				HScroll->SetPos(Max);
				Invalidate();
			}
		}

		// y scroll... in items
		if (VScroll)
		{
			int All = _Limit.y / TREE_BLOCK - 1;
			int Visible = Y() / TREE_BLOCK;

			VScroll->SetId(IDC_V);
			VScroll->SetLimits(0, All);
			VScroll->SetPage(Visible);

			int Max = All - Visible;
			if (VScroll->Pos() > Max)
			{
				VScroll->SetPos(Max+1);
				Invalidate();
			}
		}

		Processing = false;
	}
}

void GTree::_OnSelect(GTreeItem *Item)
{
	if (NOT MultipleSelect)
	{
		for (GTreeItem *i=_Selection.First(); i; i=_Selection.Next())
		{
			if (i != Item)
			{
				i->Selected(false);
			}
		}

		_Selection.Empty();
	}
	else
	{
		_Selection.Delete(Item);
	}

	_Selection.Insert(Item);
}

void GTree::_Pour()
{
	_Limit.x = _Limit.y = 0;

	GTreeItem *n;
	for (GTreeItem *i=Items.First(); i; i=n)
	{
		n=Items.Next();
		i->_Last = n == 0;
		i->_Pour(&_Limit, 0, true);
		_UpdateScrollBars();
	}

	_LayoutDirty = false;
}

// External methods and events
void GTree::OnItemSelect(GTreeItem *Item)
{
	if (Item)
	{
		Item->OnSelect();
	}
}

void GTree::OnItemExpand(GTreeItem *Item, bool Expand)
{
	if (Item)
	{
		Item->OnExpand(Expand);
	}
}

void GTree::OnKey(GKey &k)
{
	if (k.Down)
	{
		GTreeItem *i = _Selection.First();

		switch (k.c)
		{
			case VK_LEFT:
			{
				if (i)
				{
					i->Expanded(false);
					if (_LayoutDirty)
					{
						_Pour();
						break;
					}
				}
				// fall thru
			}
			case VK_UP:
			{
				if (i)
				{
					GTreeItem *p = i->GetParent() ? i->GetParent() : this;
					if (p)
					{
						int Index = i->IndexOf();
						GTreeItem *n = p;
						if (Index > 0)
						{
							n = i->GetPrev();
							while (	n->GetChild() AND
									n->GetChild()->_Visible)
							{
								n = n->Items.ItemAt(n->Items.GetItems()-1);
							}
						}

						if (n)
						{
							n->Selected(true);
						}
					}
				}
				break;
			}
			case VK_RIGHT:
			{
				if (i)
				{
					i->Expanded(true);
					if (_LayoutDirty)
					{
						_Pour();
						break;
					}
				}
				// fall thru
			}
			case VK_DOWN:
			{
				if (i)
				{
					GTreeItem *n = i->GetChild();
					if (NOT n OR
						NOT n->_Visible)
					{
						GTreeItem *p = i->GetParent() ? i->GetParent() : this;
						if (p)
						{
							int Index = i->IndexOf();
							n = (Index < p->Items.GetItems()-1) ? i->GetNext() : p->GetNext();
						}
					}

					if (n)
					{
						n->Selected(true);
					}
				}
				break;
			}
		}
	}
}

GTreeItem *GTree::ItemAtPoint(int x, int y)
{
	GdcPt2 s = _ScrollPos();

	for (GTreeItem *i = Items.First(); i; i=Items.Next())
	{
		GTreeItem *Hit = i->_HitTest(s.x + x, s.y + y);
		if (Hit)
		{
			return Hit;
		}
	}

	return 0;
}

void GTree::OnMouseWheel(double Lines)
{
	if (VScroll)
	{
		VScroll->SetPos(VScroll->Pos() + Lines);
		OnNotify(VScroll, 0);
	}
}

void GTree::OnMouseClick(GMouse &m)
{
	if (m.Down())
	{
		Focus(true);
	}
	Capture(m.Down());
	if (m.Down())
	{
		_LastClick.x = m.x;
		_LastClick.y = m.y;
	}

	_LastHit = ItemAtPoint(m.x, m.y);
	if (_LastHit)
	{
		GdcPt2 c = _ScrollPos();
		m.x += c.x;
		m.y += c.y;
		_LastHit->_MouseClick(m);
	}
}

void GTree::OnMouseMove(GMouse &m)
{
	if (IsCapturing())
	{
		if (abs(_LastClick.x - m.x) > DRAG_THRESHOLD OR
			abs(_LastClick.y - m.y) > DRAG_THRESHOLD)
		{
			OnItemBeginDrag(_LastHit, m.Flags);
			Capture(false);
		}
	}
}

void GTree::OnPosChange()
{
	_UpdateScrollBars();
}

void GTree::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);

	// icon cache
	if (GetImageList() AND
		NOT _IconCache)
	{
		#ifdef BEOS
		int Bits = GetImageList()->GetBits();
		#else
		int Bits = GdcD->GetBits();
		#endif
		
		_IconCache = NEW(GMemDC);
		if (_IconCache AND
			_IconCache->Create(GetImageList()->X(), GetImageList()->Y(), Bits))
		{
			_IconCache->Colour(LC_WORKSPACE, 24);
			_IconCache->Rectangle();

			for (int i=0; i<GetImageList()->GetItems(); i++)
			{
				GetImageList()->Draw(_IconCache, i * GetImageList()->TileX(), 0, i);
			}
		}
	}

	// scroll
	GdcPt2 s = _ScrollPos();
	pDC->SetOrigin(s.x, s.y);

	// selection colour
	COLOUR SelectedCol = Focus() ? LC_SELECTION : GdcMixColour(LC_SELECTION, LC_WORKSPACE);

	// layout items
	if (_LayoutDirty)
	{
		_Pour();
	}

	// paint itmes
	ZeroObj(_LineFlags);
	for (GTreeItem *i = Items.First(); i; i=Items.Next())
	{
		GRegion Item;
		i->OnPaint(pDC, Item, SelectedCol);
	}

	pDC->SetOrigin(0, 0);
	if (_Limit.y-s.y < Y())
	{
		// paint after items
		pDC->Colour(LC_WORKSPACE, 24);
		pDC->Rectangle(0, _Limit.y-s.y, X()-1, Y()-1);
	}
}

int GTree::OnNotify(GView *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
		case IDC_V:
		case IDC_H:
		{
			Invalidate();
			break;
		}
	}

	return GLayout::OnNotify(Ctrl, Flags);
}

int GTree::OnEvent(GMessage *Msg)
{
	return GLayout::OnEvent(Msg);
}

GTreeItem *GTree::Insert(GTreeItem *Obj, int Pos)
{
	GTreeItem *NewObj = GTreeItem::Insert(Obj, Pos);
	if (NewObj)
	{
		NewObj->Parent = 0;
		NewObj->Tree = this;
	}
	return NewObj;
}

bool GTree::Remove(GTreeItem *Obj)
{
	if (Obj AND
		Obj->Tree == this)
	{
		Obj->Remove();
		return true;
	}
	return false;
}

void GTree::RemoveAll()
{
	for (GTreeItem *i=Items.First(); i; i=Items.First())
	{
		i->_Remove();
	}

	Invalidate();
}

void GTree::Empty()
{
	GTreeItem *i;
	while (i=Items.First())
	{
		Delete(i);
	}
}

bool GTree::Delete(GTreeItem *Obj)
{
	bool Status = FALSE;
	if (Obj AND Remove(Obj))
	{
		DeleteObj(Obj);
		Status = true;
	}
	return Status;
}

void GTree::SelectDropTarget(GTreeItem *Item)
{
	bool Update = (_DropTarget != 0) ^ (Item != 0);
	GTreeItem *Old = _DropTarget;

	_DropTarget = Item;
	if (Old)
	{
		Old->Update();
	}
	if (_DropTarget)
	{
		_DropTarget->Update();
	}
	if (Update)
	{
		OnFocus(true);
	}
}

bool GTree::Select(GTreeItem *Obj)
{
	bool Status = FALSE;
	if (Obj AND IsAttached())
	{
		Obj->Selected(true);
		Status = TRUE;
	}
	return Status;
}

GTreeItem *GTree::Selection()
{
	return _Selection.First();
}

void GTree::OnItemClick(GTreeItem *Item, GMouse &m)
{
	if (Item)
	{
		Item->OnMouseClick(m);
	}
}

void GTree::OnItemBeginDrag(GTreeItem *Item, int Flags)
{
	if (Item)
	{
		GMouse m;
		ZeroObj(m);
		m.Flags = Flags;
		Item->OnBeginDrag(m);
	}
}

void GTree::OnFocus(bool b)
{
	for (GTreeItem *i=_Selection.First(); i; i=_Selection.Next())
	{
		i->Update();
	}
}
