// Lgi.cpp
#include <time.h>
#include <stdio.h>

#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////
//#define XP_CTRLS
COLOUR _LgiColours[LC_MAXIMUM];

#define ReadColourConfig(def)	_lgi_read_colour_config("Colour."#def, _LgiColours+i++)

int _hex_to_int(char h)
{
	if (h >= '0' AND h <= '9')
	{
		return h - '0';
	}
	else if (h >= 'A' AND h <= 'F')
	{
		return h - 'A' + 10;
	}
	else if (h >= 'a' AND h <= 'f')
	{
		return h - 'a' + 10;
	}

	LgiAssert(0);
	return 0;
}

void _lgi_read_colour_config(char *Tag, COLOUR *c)
{
	XmlTag *Col = LgiApp->GetConfig(Tag);
	if (Col)
	{
		char *h;
		if (Col->Get("Hex", h))
		{
			int n = 0;
			if (*h == '#') h++;
			
			for (int i=0; i<5; i++)
			{
				n |= _hex_to_int(*h++);
				n <<= 4;
			}
			n |= _hex_to_int(*h++);

			if (c)
			{
				*c = Rgb24( n>>16, n>>8, n );
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////
void LgiInitColours()
{
	int i = 0;

	// Basic colours
	_LgiColours[i++] = Rgb24(0, 0, 0); // LC_BLACK
	_LgiColours[i++] = Rgb24(0x40, 0x40, 0x40); // LC_DKGREY
	_LgiColours[i++] = Rgb24(0x80, 0x80, 0x80); // LC_MIDGREY
	_LgiColours[i++] = Rgb24(0xc0, 0xc0, 0xc0); // LC_LTGREY
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_WHITE

	// Variable colours
	#if defined XP_CTRLS
	_LgiColours[i++] = Rgb24(0x42, 0x27, 0x63); // LC_SHADOW
	_LgiColours[i++] = Rgb24(0x7a, 0x54, 0xa9); // LC_LOW
	_LgiColours[i++] = Rgb24(0xbc, 0xa9, 0xd4); // LC_MED
	_LgiColours[i++] = Rgb24(0xdd, 0xd4, 0xe9); // LC_HIGH
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_LIGHT
	_LgiColours[i++] = Rgb24(0xbc, 0xa9, 0xd4); // LC_DIALOG
	_LgiColours[i++] = Rgb24(0xeb, 0xe6, 0xf2); // LC_WORKSPACE
	_LgiColours[i++] = Rgb24(0x35, 0x1f, 0x4f); // LC_TEXT
	_LgiColours[i++] = Rgb24(0xbf, 0x67, 0x93); // LC_SELECTION
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_SEL_TEXT
	_LgiColours[i++] = Rgb24(0x70, 0x3a, 0xec); // LC_ACTIVE_TITLE
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_ACTIVE_TITLE_TEXT
	_LgiColours[i++] = Rgb24(0x80, 0x80, 0x80); // LC_INACTIVE_TITLE
	_LgiColours[i++] = Rgb24(0x40, 0x40, 0x40); // LC_INACTIVE_TITLE_TEXT
	#elif defined WIN32
	_LgiColours[i++] = GetSysColor(COLOR_3DDKSHADOW); // LC_SHADOW
	_LgiColours[i++] = GetSysColor(COLOR_3DSHADOW); // LC_LOW
	_LgiColours[i++] = GetSysColor(COLOR_3DFACE); // LC_MED
	_LgiColours[i++] = GetSysColor(COLOR_3DLIGHT); // LC_HIGH, 
	_LgiColours[i++] = GetSysColor(COLOR_3DHIGHLIGHT); // LC_LIGHT
	_LgiColours[i++] = GetSysColor(COLOR_3DFACE); // LC_DIALOG
	_LgiColours[i++] = GetSysColor(COLOR_WINDOW); // LC_WORKSPACE
	_LgiColours[i++] = GetSysColor(COLOR_BTNTEXT); // LC_TEXT
	_LgiColours[i++] = GetSysColor(COLOR_HIGHLIGHT); // LC_SELECTION
	_LgiColours[i++] = GetSysColor(COLOR_HIGHLIGHTTEXT); // LC_SEL_TEXT
	_LgiColours[i++] = GetSysColor(COLOR_ACTIVECAPTION); // LC_ACTIVE_TITLE
	_LgiColours[i++] = GetSysColor(COLOR_CAPTIONTEXT); // LC_ACTIVE_TITLE_TEXT
	_LgiColours[i++] = GetSysColor(COLOR_INACTIVECAPTION); // LC_INACTIVE_TITLE
	_LgiColours[i++] = GetSysColor(COLOR_INACTIVECAPTIONTEXT); // LC_INACTIVE_TITLE_TEXT
	#else // defaults for non-windows, plain greys
	_LgiColours[i++] = Rgb24(128, 128, 128); // LC_SHADOW
	_LgiColours[i++] = Rgb24(164, 164, 164); // LC_LOW
	_LgiColours[i++] = Rgb24(216, 216, 216); // LC_MED
	_LgiColours[i++] = Rgb24(236, 236, 236); // LC_HIGH
	_LgiColours[i++] = Rgb24(255, 255, 255); // LC_LIGHT
	_LgiColours[i++] = Rgb24(216, 216, 216); // LC_DIALOG
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_WORKSPACE
	_LgiColours[i++] = Rgb24(0, 0, 0); // LC_TEXT
	_LgiColours[i++] = Rgb24(0x80, 0x80, 0x80); // LC_SELECTION
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_SEL_TEXT
	_LgiColours[i++] = Rgb24(0, 0, 0xff); // LC_ACTIVE_TITLE
	_LgiColours[i++] = Rgb24(0xff, 0xff, 0xff); // LC_ACTIVE_TITLE_TEXT
	_LgiColours[i++] = Rgb24(0x80, 0x80, 0x80); // LC_INACTIVE_TITLE
	_LgiColours[i++] = Rgb24(0x40, 0x40, 0x40); // LC_INACTIVE_TITLE_TEXT
	#endif

	// Tweak
	if (LgiGetOs() == LGI_OS_WINNT OR
		LgiGetOs() == LGI_OS_WIN9X)
	{
		// Win32 doesn't seem to get this right, so we just tweak it here
		#define MixComp(a)		( (a(_LgiColours[7])+a(_LgiColours[9])) / 2 )
		_LgiColours[8] = Rgb24(MixComp(R24), MixComp(G24), MixComp(B24));
	}

	// Read any settings out of config
	i = 5;
	ReadColourConfig(LC_SHADOW);
	ReadColourConfig(LC_LOW);
	ReadColourConfig(LC_MED);
	ReadColourConfig(LC_HIGH);
	ReadColourConfig(LC_LIGHT);
	ReadColourConfig(LC_DIALOG);
	ReadColourConfig(LC_WORKSPACE);
	ReadColourConfig(LC_TEXT);
	ReadColourConfig(LC_SELECTION);
	ReadColourConfig(LC_SEL_TEXT);
	ReadColourConfig(LC_ACTIVE_TITLE);
	ReadColourConfig(LC_ACTIVE_TITLE_TEXT);
	ReadColourConfig(LC_INACTIVE_TITLE);
	ReadColourConfig(LC_INACTIVE_TITLE_TEXT);
}

COLOUR LgiColour(int i)
{
	return i<LC_MAXIMUM ? _LgiColours[i] : 0;
}

//////////////////////////////////////////////////////////////////////////
int LgiGetOs(int *Ver, int *Rev)
{
	#if defined WIN32

	OSVERSIONINFO v;
	v.dwOSVersionInfoSize=sizeof(v);
	GetVersionEx(&v);

	if (Ver) *Ver = v.dwMajorVersion;
	if (Rev) *Rev = v.dwMinorVersion;

	return (v.dwPlatformId == VER_PLATFORM_WIN32_NT) ? LGI_OS_WINNT : LGI_OS_WIN9X;

	#elif defined BEOS

	if (Ver) *Ver = 5;
	if (Rev) *Rev = 0;
	
	return LGI_OS_BEOS;

	#elif defined LINUX

	return LGI_OS_LINUX;

	#elif defined ATHEOS

	if (Ver) *Ver = 0;
	if (Rev) *Rev = 0x307;
	
	return LGI_OS_ATHEOS;

	#else

	return LGI_OS_UNKNOWN;

	#endif
}

char *LgiGetOsName()
{
	char *Str[LGI_OS_MAX] =
	{
		"Unknown",
		"Win32",
		"Win32",
		"BeOS",
		"Linux",
		"AtheOS"
	};

	return Str[LgiGetOs()];
}

#ifdef WIN32
#define RecursiveFileSearch_Wildcard "*.*"
#else // unix'ish OS
#define RecursiveFileSearch_Wildcard "*"
#endif

bool LgiRecursiveFileSearch(char *Root,
							List<char> &Ext,
							List<char> *Files,
							quad *Size,
							quad *Count,
							RecursiveFileSearch_Callback Callback)
{
	bool Status = false;
	char Buf[256];

	// validate args
	if (NOT Root) return 0;

	// setup search string
	strcpy(Buf, Root);
	if (Buf[strlen(Buf)-1] != DIR_CHAR)
	{
		strcat(Buf, DIR_STR);
	}
	char *Wildcard = Buf+strlen(Buf);
	strcpy(Wildcard, RecursiveFileSearch_Wildcard);

	// get directory enumerator
	VDirectory *Dir = FileDev->GetDir(0);
	if (Dir)
	{
		Status = true;

		// enumerate the directory contents
		for (bool Found = Dir->First(Buf, 0, -1); Found; Found = Dir->Next())
		{
			char *Name = Dir->GetName();
			if (Name AND
				stricmp(Name, ".") != 0 AND
				stricmp(Name, "..") != 0)
			{
				strcpy(Wildcard, Name);

				if (Dir->IsDir())
				{
					// dir
					LgiRecursiveFileSearch(	Buf,
											Ext,
											Files,
											Size,
											Count,
											Callback);

				}
				else
				{
					// process file
					bool Match = true; // if no Ext's then default to match
					for (char *e = Ext.First(); e; e = Ext.Next())
					{
						if (Match = MatchStr(e, Name))
						{
							break;
						}
					}

					if (Match)
					{
						// file matched... process:
						if (Files)
						{
							Files->Insert(NewStr(Buf));
						}

						if (Size)
						{
							*Size += Dir->GetSize();
						}

						if (Count)
						{
							*Count++;
						}

						if (Callback)
						{
							if (NOT Callback(Buf, Dir))
							{
								Status = false;
								break;
							}
						}
					}
				}

				strcpy(Wildcard, RecursiveFileSearch_Wildcard);
			}
		}

		DeleteObj(Dir);
	}

	return Status;
}

#define LGI_TRACE_TO_FILE

void LgiTrace(char *Msg, ...)
{
	if (Msg)
	{
		char Buffer[1025];
		LgiGetExeFile(Buffer);
		char *Dot = strrchr(Buffer, '.');
		if (Dot AND NOT strchr(Dot, DIR_CHAR))
		{
			strcpy(Dot+1, "txt");
		}
		else
		{
			strcat(Buffer, ".txt");
		}

		#ifdef LGI_TRACE_TO_FILE
		GFile f;
		bool FileOpen = f.Open(Buffer, O_WRITE);
		#endif

		va_list Arg;
		va_start(Arg, Msg);
		vsprintf(Buffer, Msg, Arg);
		va_end(Arg);

		#ifdef LGI_TRACE_TO_FILE
		f.Seek(0, SEEK_END);
		f.Write(Buffer, strlen(Buffer));
		f.Write((char*)"\n", 1);
		f.Close();
		#endif

		#if defined WIN32
		OutputDebugString(Buffer);
		#elif defined BEOS || defined ATHEOS
		printf("Trace: %s", Buffer);
		#endif
	}
}


bool LgiTrimDir(char *Path)
{
	if (Path)
	{
		char *p = strrchr(Path, DIR_CHAR);
		if (p)
		{
			*p = 0;
			return true;
		}
	}
	return false;
}

bool LgiGetExeFile(char *Dst)
{
	if (Dst)
	{
		#if defined WIN32

		char Exe[256];
		if (GetModuleFileName(NULL, Exe, sizeof(Exe)) > 0)
		{
			strlwr(Exe);
			strcpy(Dst, Exe);
			return TRUE;
		}

		#elif defined BEOS

		app_info Info;
		if (LgiApp->GetAppInfo(&Info) == B_OK)
		{
			BEntry e(&Info.ref);
			BPath p;
			if (e.GetPath(&p) == B_OK)
			{
				strcpy(Dst, p.Path());
				return TRUE;
			}
		}

		#elif defined ATHEOS

		os::Directory AppPath;
		if (AppPath.SetTo("^/.") == 0)
		{
			std::string p;
			if (AppPath.GetPath(&p) == 0)
			{
				sprintf(Dst, "%s%s%s", p.c_str(), DIR_STR, LgiApp->_AppFile);
				return true;
			}
		}

		#endif
	}
	return false;
}

bool LgiGetExePath(char *Dst)
{
	if (Dst)
	{
		if (LgiGetExeFile(Dst))
		{
			LgiTrimDir(Dst);
			return true;
		}
	}
	return false;
}

char *LgiFindFile(char *Name)
{
	char *Result = 0;

	if (Name)
	{
		char *JustFile = strrchr(Name, DIR_CHAR);
		if (JustFile) JustFile++;
		else JustFile = Name;

		char ExePath[256];
		char File[256];
		if (LgiGetExePath(ExePath))
		{
			#if defined WIN32 && defined _DEBUG
			LgiTrimDir(ExePath);
			#endif

			sprintf(File, "%s%s%s", ExePath, DIR_STR, JustFile);
			if (_lgi_check_file(File))
			{
				Result = NewStr(File);
			}
			else
			{
				// ..
				sprintf(File, "%s%sCode%s%s", ExePath, DIR_STR, DIR_STR, JustFile);
				if (_lgi_check_file(File))
				{
					Result = NewStr(File);
				}
			}
		}
	}

	return Result;
}

/////////////////////////////////////////////////////////////////////////////////////////////
// Drawing functions
void LgiDrawBox(GSurface *pDC, GRegion &r, bool Sunken, bool Fill)
{
	if (Fill)
	{
		pDC->Colour(LC_MED, 24);
		pDC->Rectangle(r.x1+1, r.y1+1, r.x2-1, r.y2-1);
	}

	pDC->Colour((Sunken) ? LC_LIGHT : LC_LOW, 24);
	pDC->Line(r.x2, r.y2, r.x2, r.y1);
	pDC->Line(r.x2, r.y2, r.x1, r.y2);

	pDC->Colour((Sunken) ? LC_LOW : LC_LIGHT, 24);
	pDC->Line(r.x1, r.y1, r.x1, r.y2);
	pDC->Line(r.x1, r.y1, r.x2, r.y1);
}

void LgiWideBorder(GSurface *pDC, GRegion &r, int Type)
{
	if (NOT pDC) return;
	COLOUR Old = pDC->Colour();
	COLOUR VLow = LC_SHADOW;
	COLOUR Low = LC_LOW;
	COLOUR High = LC_HIGH;
	COLOUR VHigh = LC_LIGHT;

	switch (Type)
	{
		case SUNKEN:
		{
			pDC->Colour(Low, 24);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(VLow, 24);
			pDC->Line(r.x1+1, r.y1+1, r.x2-2, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-2);

			pDC->Colour(High, 24);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(VHigh, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
		case RAISED:
		{
			pDC->Colour(VHigh, 24);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(High, 24);
			pDC->Line(r.x1+1, r.y1+1, r.x2-1, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-1);

			pDC->Colour(Low, 24);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(VLow, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
		case CHISEL:
		{
			pDC->Colour(Low, 24);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(VHigh, 24);
			pDC->Line(r.x1+1, r.y1+1, r.x2-2, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-2);

			pDC->Colour(Low, 24);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(VHigh, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
	}

	r.Size(2, 2);
	pDC->Colour(Old);
}

void LgiThinBorder(GSurface *pDC, GRegion &r, int Type)
{
	if (NOT pDC) return;
	COLOUR Old = pDC->Colour();

	switch (Type)
	{
		case SUNKEN:
		{
			pDC->Colour(LC_LIGHT, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);

			pDC->Colour(LC_LOW, 24);
			pDC->Line(r.x1, r.y1, r.x1, r.y2);
			pDC->Line(r.x1, r.y1, r.x2, r.y1);

			r.Size(1, 1);
			break;
		}
		case RAISED:
		{
			pDC->Colour(LC_LOW, 24);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);

			pDC->Colour(LC_LIGHT, 24);
			pDC->Line(r.x1, r.y1, r.x1, r.y2);
			pDC->Line(r.x1, r.y1, r.x2, r.y1);

			r.Size(1, 1);
			break;
		}
	}

	pDC->Colour(Old);
}

void LgiFlatBorder(GSurface *pDC, GRegion &r, int Width)
{
	pDC->Colour(LC_MED, 24);
	if (Width < 1 OR r.X() < (2 * Width) OR r.Y() < (2 * Width))
	{
		pDC->Rectangle(&r);
		r.ZOff(-1, -1);
	}
	else
	{
		pDC->Rectangle(r.x1, r.y1, r.x2, r.y1+Width-1);
		pDC->Rectangle(r.x1, r.y2-Width+1, r.x2, r.y2);
		pDC->Rectangle(r.x1, r.y1+Width, r.x1+Width-1, r.y2-Width);
		pDC->Rectangle(r.x2-Width+1, r.y1+Width, r.x2, r.y2-Width);
		r.Size(Width, Width);
	}
}

//////////////////////////////////////////////////////////////////////////
GObject::GObject()
{
	ObjName = 0;
}

GObject::~GObject()
{
	DeleteArray(ObjName);
}

char *GObject::Name()
{
	return ObjName;
}

bool GObject::Name(char *n)
{
	if (n == ObjName) return TRUE;
	DeleteObj(ObjName);
	if (n)
	{
		ObjName = NEW(char[strlen(n)+1]);
		if (ObjName)
		{
			strcpy(ObjName, n);
			return TRUE;
		}
	}
	return FALSE;
}


//////////////////////////////////////////////////////////////////////////////
DoEvery::DoEvery(int p) // p = timeout in ms
{
	Init(p);
}

void DoEvery::Init(int p) // p = timeout in ms
{
	LastTime = LgiCurrentTime();
	if (p > 0)
	{
		Period = (p * CLOCKS_PER_SEC) / 1000;
	}
}

bool DoEvery::DoNow()
{
	int Now = LgiCurrentTime();
	if (LastTime + Period < Now)
	{
		LastTime = Now;
		return true;
	}
	return false;
}







