/*hdr
**      FILE:           LgiRes.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           12/3/2000
**      DESCRIPTION:    Xp resource interfaces
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>

#include "Lgi.h"
#include "GToken.h"

// If it is defined it will use the cross platform 
// "res" library distributed with the LGI library.

#include "GXml.h"

#define DeleteAllListObjects(pList, Type)	{ for (Type *o = (pList)->First(); o; o = (pList)->Next()) { DeleteObj(o); } }
#define CastToGWnd(RObj)					((RObj != 0) ? dynamic_cast<GView*>(RObj) : 0)

LgiStringRes::LgiStringRes()
{
	Ref = 0;
	Id = 0;
	Str = 0;
	IsString = false;
}

LgiStringRes::~LgiStringRes()
{
	DeleteArray(Str);
}

bool LgiStringRes::Read(List<XmlTag> &Tags)
{
	XmlTag *t = Tags.Current();
	if (t AND stricmp(t->Name, "string") == 0)
	{
		char *n = 0;
		if (t->Get("Id", n))
		{
			Id = atoi(n);
		}
		if (t->Get("Ref", n))
		{
			Ref = atoi(n);
		}
		if (t->Get("Define", n))
		{
			if (strcmp(n, "IDOK") == 0)
			{
				Id = IDOK;
			}
			else if (strcmp(n, "IDCANCEL") == 0)
			{
				Id = IDCANCEL;
			}
			else if (strcmp(n, "IDC_STATIC") == 0)
			{
				Id = -1;
			}
		}

		int CurLang = LgiGetLanguageId();
		
		char Name[256];
		sprintf(Name, "Text(%i)", CurLang);
		n = 0;

		if (t->Get(Name, n) AND
			n AND
			strlen(n) > 0)
		{
			// language string ok
		}
		else if (CurLang != 13)
		{
			// no string, try english
			t->Get("Text(13)", n);
		}

		if (n)
		{
			DeleteArray(Str);
			Str = NewStr(n);
			if (Str)
			{
				char *d = Str;
				for (char *s = Str; s AND *s;)
				{
					if (*s == '\\')
					{
						if (*(s+1) == 'n')
						{
							*d++ = '\n';
						}
						else if (*(s+1) == 't')
						{
							*d++ = '\t';
						}
						s += 2;
					}
					else 
					{
						*d++ = *s++;
					}
				}
				*d++ = 0;
			}
		}

		return true;
	}
	return false;
}

///////////////////////////////////////////////////////////////////
// Lgi resource factory
List<LgiResources> LgiResources::ResourceContainers;

LgiResources::LgiResources(char *FileName, bool Warn)
{
	// global pointer list
	ResourceContainers.Insert(this);

	char File[256] = "(app)";
	if (FileName)
	{
		char *f = strrchr(FileName, DIR_CHAR);
		if (f)
		{
			strcpy(File, f + 1);
		}
		else
		{
			strcpy(File, FileName);
		}
	}
	else
	{
		char Str[256];
		if (LgiGetExeFile(Str))
		{
			char *f = strrchr(Str, DIR_CHAR);
			if (f)
			{
				f++;
				
				#if defined WIN32
				char *Period = strrchr(f, '.');
				if (Period)
				{
					*Period = 0;
				}
				#endif

				strcpy(File, f);
			}
		}
	}

	strcat(File, ".lr");

	char *FullPath = LgiFindFile(File);
	#ifdef WIN32
	if (NOT FullPath)
	{
		// maybe theres a link to it?
		strcat(File, ".lnk");
		
		char *Link = LgiFindFile(File);
		if (Link)
		{
			bool IsLink = ResolveShortcut(Link, File, strlen(File));
			DeleteArray(Link);
			if (IsLink)
			{
				FullPath = NewStr(File);
			}
		}
	}
	#endif

	if (FullPath)
	{
		Load(FullPath);
		DeleteArray(FullPath);
	}
	else if (Warn)
	{
		char Msg[256];
		char Exe[256];

		// Prepare data
		strlwr(File);
		sprintf(Msg, "Couldn't find the file '%s' required to run this application", File);
		LgiGetExeFile(Exe);
		strlwr(Exe);

		// Dialog
		GAlert Dlg(0, Exe, Msg, "Ok");
		Dlg.DoModal();

		// Exit
		LgiExitApp();
	}
}

LgiResources::~LgiResources()
{
	ResourceContainers.Delete(this);

	DeleteAllListObjects(&Dialogs, LgiDialogRes);
	DeleteAllListObjects(&Strings, LgiStringRes);
	DeleteAllListObjects(&Menus, LgiMenuRes);
}

bool LgiResources::Load(char *FileName)
{
	bool Status = false;

	if (FileName)
	{
		GFile F;
		if (F.Open(FileName, O_READ))
		{
			Xml X;
			if (X.ParseXmlFile(F) > 0)
			{
				Status = true;

				for (XmlTag *t = X.Values.First(); t; t = X.Values.Next())
				{
					if (t->Name)
					{
						if (stricmp(t->Name, "string-group") == 0)
						{
							bool IsString = true;
							char *Name = 0;
							if (t->Get("Name", Name))
							{
								IsString = stricmp("_Dialog Symbols_", Name) != 0;
							}

							// search through group, adding strings to database
							for (	t = X.Values.Next();
									t AND t->Name AND stricmp(t->Name, "/string-group") != 0;
									t = X.Values.Next())
							{
								LgiStringRes *s = NEW(LgiStringRes);
								if (s AND s->Read(X.Values))
								{
									Strings.Insert(s);
									s->IsString = IsString;
								}
								else
								{
									DeleteObj(s);
								}
							}
						}
						else if (stricmp(t->Name, "dialog") == 0)
						{
							LgiDialogRes *d = NEW(LgiDialogRes);
							if (d AND d->Read(X.Values))
							{
								Dialogs.Insert(d);
							}
							else
							{
								DeleteObj(d);
							}
						}
						else if (stricmp(t->Name, "menu") == 0)
						{
							LgiMenuRes *m = NEW(LgiMenuRes);
							if (m AND m->Read(X.Values))
							{
								Menus.Insert(m);
							}
							else
							{
								DeleteObj(m);
							}
						}
					}
				}
			}
		}
	}

	return Status;
}

LgiStringRes *LgiResources::StrFromRef(int Ref)
{
	for (LgiStringRes *s = Strings.First(); s; s = Strings.Next())
	{
		if (s->Ref == Ref)
		{
			return s;
		}
	}

	return 0;
}

char *LgiResources::StringFromRef(int Ref)
{
	for (LgiStringRes *s = Strings.First(); s; s = Strings.Next())
	{
		if (s->Ref == Ref)
		{
			return s->Str;
		}
	}

	return 0;
}

ResObject *LgiResources::CreateObject(char *Name, ResObject *Parent)
{
	ResObject *Wnd = 0;
	if (Name)
	{
		if (stricmp(Name, Res_StaticText) == 0)
		{
			Wnd = NEW(GText(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_EditBox) == 0)
		{
			Wnd = NEW(GEdit(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_CheckBox) == 0)
		{
			Wnd = NEW(GCheckBox(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_Button) == 0)
		{
			Wnd = NEW(GButton(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_Group) == 0)
		{
			Wnd = NEW(GRadioGroup(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_RadioBox) == 0)
		{
			Wnd = NEW(GRadioButton(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_TabView) == 0)
		{
			Wnd = NEW(GTabView(0, 10, 10, 100, 100, "GTabView"));
		}
		else if (stricmp(Name, Res_Tab) == 0)
		{
			if (Parent)
			{
				// GView *p = (GView*) Parent;
				GTabView *Ctrl = dynamic_cast<GTabView*>(Parent);
				if (Ctrl)
				{
					Wnd = Ctrl->Append("<error>");
				}
			}
		}
		else if (stricmp(Name, Res_ListView) == 0)
		{
			GList *w;
			Wnd = w = NEW(GList(0, 0, 0, -1, -1, ""));
			if (w)
			{
				w->Sunken(true);
			}
		}
		else if (stricmp(Name, Res_Column) == 0)
		{
			if (Parent)
			{
				GList *Lst = dynamic_cast<GList*>(Parent);

				assert(Lst);

				if (Lst)
				{
					Wnd = Lst->AddColumn("");
				}
			}
		}
		else if (stricmp(Name, Res_ComboBox) == 0)
		{
			Wnd = NEW(GCombo(0, 0, 0, 100, 20, ""));
		}
		else if (stricmp(Name, Res_Bitmap) == 0)
		{
			Wnd = NEW(GBitmap(0, 0, 0, 0));
		}
		else if (stricmp(Name, Res_Progress) == 0)
		{
			Wnd = NEW(GProgress(0, 0, 0, -1, -1, ""));
		}
		else if (stricmp(Name, Res_Slider) == 0)
		{
			Wnd = NEW(GSlider(0, 0, 0, -1, -1, "", false));
		}
		else if (stricmp(Name, Res_ScrollBar) == 0)
		{
			Wnd = NEW(GScrollBar(0, 0, 0, 20, 100, ""));
		}
		else if (stricmp(Name, Res_Progress) == 0)
		{
			Wnd = NEW(GProgress(0, 0, 0, 1, 1, ""));
		}
		else if (stricmp(Name, Res_TreeView) == 0)
		{
			Wnd = NEW(GTree(0, 0, 0, 1, 1, ""));
		}
	}

	ASSERT(Wnd);

	return Wnd;
}

void LgiResources::Res_SetPos(ResObject *Obj, int x1, int y1, int x2, int y2)
{
	GListColumn *Col = dynamic_cast<GListColumn*>(Obj);
	if (Col)
	{
		Col->Width(x2-x1);
	}
	else
	{
		GView *w = CastToGWnd(Obj);
		if (w)
		{
			#ifdef WIN32
			GCombo *Cbo = dynamic_cast<GCombo*>(w);
			if (Cbo)
			{
				int y = y2-y1+1;
				y = max(y, 100);
				w->SetPos(GRegion(x1, y1, x2, y1+y));
			}
			else
			#endif
			{
				GRegion n(x1, y1, x2, y2);
				w->SetPos(n);
			}
		}
	}
}

void LgiResources::Res_SetPos(ResObject *Obj, char *s)
{
	if (Obj AND s)
	{
		TokenParser T(s, ",");
		if (T.Tokens == 4)
		{
			Res_SetPos(Obj, atoi(T.Get(0)), atoi(T.Get(1)), atoi(T.Get(2)), atoi(T.Get(3)));
		}
	}
}

bool LgiResources::Res_GetProperties(ResObject *Obj, ObjProperties *Props)
{
	// this is a read-only system...
	return false;
}

bool LgiResources::Res_SetProperties(ResObject *Obj, ObjProperties *Props)
{
	GEdit *e = dynamic_cast<GEdit*>(Obj);
	if (e AND Props)
	{
		int i;
		if (Props->Get("password", i))
		{
			e->Password(i);
		}
		if (Props->Get("enabled", i))
		{
			e->Enabled(i);
		}
		if (Props->Get("multiline", i))
		{
			e->MultiLine(i);
		}
		return true;
	}
	return false;
}

GRegion LgiResources::Res_GetPos(ResObject *Obj)
{
	GView *w = CastToGWnd(Obj);
	if (w)
	{
		return w->GetPos();
	}

	return GRegion(0, 0, 0, 0);
}

int LgiResources::Res_GetStrRef(ResObject *Obj)
{
	return 0;
}

void LgiResources::Res_SetStrRef(ResObject *Obj, int Ref)
{
	for (LgiStringRes *s = Strings.First(); s; s = Strings.Next())
	{
		if (s->Ref == Ref)
		{
			GView *w = CastToGWnd(Obj);
			if (w)
			{
				w->Name(s->Str);
				w->SetId(s->Id);
			}
			else if (Obj)
			{
				GListColumn *Col = dynamic_cast<GListColumn*>(Obj);
				if (Col)
				{
					Col->Name(s->Str);
					return;
				}

				GTabPage *Page = dynamic_cast<GTabPage*>(Obj);
				if (Page)
				{
					Page->Name(s->Str);
				}
			}
			break;
		}
	}
}

void LgiResources::Res_Attach(ResObject *Obj, ResObject *Parent)
{
	GView *o = CastToGWnd(Obj);
	GView *p = CastToGWnd(Parent);
	GTabPage *Tab = dynamic_cast<GTabPage*>(Parent);
	if (o)
	{
		if (Tab)
		{
			if (Tab)
			{
				GRegion r = o->GetPos();
				r.Offset(-4, -24);
				o->SetPos(r);

				Tab->Append(o);
			}
		}
		else if (p)
		{
			if (NOT o->Attach(p))
			{
				p->Children.Insert(o);
				o->SetParent(p);
			}
		}
		else
		{
			ASSERT(p);
		}
	}
}

bool LgiResources::Res_GetChildren(ResObject *Obj, List<ResObject> *l, bool Deep)
{
	GView *o = CastToGWnd(Obj);
	if (o AND l)
	{
		for (GView *w = o->Children.First(); w; w = o->Children.Next())
		{
			ResObject *n = dynamic_cast<ResObject*>(w);
			if (n)
			{
				l->Insert(n);
			}
		}
		return true;
	}
	return false;
}

void LgiResources::Res_Append(ResObject *Obj, ResObject *Parent)
{
	if (Obj AND Parent)
	{
		GListColumn *Col = dynamic_cast<GListColumn*>(Obj);
		GList *Lst = dynamic_cast<GList*>(Parent);
		if (Lst AND Col)
		{
			Lst->AddColumn(Col);
		}

		GTabPage *Tab = dynamic_cast<GTabPage*>(Obj);
		GTabView *Tabs = dynamic_cast<GTabView*>(Parent);
		if (Tab AND Tabs)
		{
			Tabs->Append(Tab);
		}
	}
}

bool LgiResources::Res_GetItems(ResObject *Obj, List<ResObject> *l)
{
	if (Obj AND l)
	{
		GList *Lst = dynamic_cast<GList*>(Obj);
		if (Lst)
		{
			for (int i=0; i<Lst->GetColumns(); i++)
			{
				l->Insert(Lst->ColumnAt(i));
			}
			return true;
		}

		GTabView *Tabs = dynamic_cast<GTabView*>(Obj);
		if (Tabs)
		{
			for (int i=0; i<Tabs->GetTabs(); i++)
			{
				l->Insert(Tabs->TabAt(i));
			}
			return true;
		}
	}
	return false;
}

///////////////////////////////////////////////////////
LgiDialogRes::LgiDialogRes()
{
	Dialog = 0;
	Str = 0;
}

LgiDialogRes::~LgiDialogRes()
{
	DeleteAllListObjects(&Tags, XmlTag);
}

bool LgiDialogRes::Read(List<XmlTag> &l)
{
	XmlTag *t = Dialog = l.Current();
	if (Dialog)
	{
		char *n = 0;
		if (Dialog->Get("ref", n))
		{
			int Ref = atoi(n);
			Str = LgiGetResObj()->StrFromRef(Ref);
		}
		if (Dialog->Get("pos", n))
		{
			TokenParser T(n, ",");
			if (T.Tokens == 4)
			{
				Pos.x1 = atoi(T.Get(0));
				Pos.y1 = atoi(T.Get(1));
				Pos.x2 = atoi(T.Get(2));
				Pos.y2 = atoi(T.Get(3));
			}
		}
	}

	for (	t = l.Next();
			t AND t->Name AND
			stricmp(t->Name, "/dialog") != 0;
			t = l.Next())
	{
		Tags.Insert(t);
	}

	for (	XmlTag *d = Tags.First();
			d;
			d = Tags.Next())
	{
		l.Delete(d);
	}

	if (t)
	{
		l.Find(t);
	}

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////
LgiMenuRes::LgiMenuRes()
{
}

LgiMenuRes::~LgiMenuRes()
{
	DeleteAllListObjects(&Tags, XmlTag);
}

bool LgiMenuRes::Read(List<XmlTag> &l)
{
	XmlTag *t = l.Current();
	if (t AND stricmp(t->Name, "menu") == 0)
	{
		char *n;
		if (t->Get("name", n))
		{
			GObject::Name(n);
		}

		for (; t AND t->Name AND stricmp(t->Name, "/menu") != 0; t = l.Next())
		{
			Tags.Insert(t);
		}

		for (	XmlTag *d = Tags.First();
				d;
				d = Tags.Next())
		{
			l.Delete(d);
		}

		if (t)
		{
			l.Find(t);
		}

		return true;
	}

	return false;
}

/////////////////////////////////////////////////////////////////////////////////////////
// Dialog
bool GLgiRes::LoadFromResource(char *Resource, List<GView> &Children, GRegion *Pos, char *Name)
{
	bool Status = false;

	LgiResources *r = LgiGetResObj();
	if (r)
	{
		for (LgiDialogRes *Dlg = r->Dialogs.First(); Dlg; Dlg = r->Dialogs.Next())
		{
			if (Dlg->Id() == ((int) Resource))
			{
				// found the dialog to load, set properties
				if (Name AND Dlg->Name())
				{
					strcpy(Name, Dlg->Name());
				}
				
				if (Pos)
				{
					Pos->ZOff(Dlg->X()-1, Dlg->Y()-1);
				}

				// instantiate control list
				for (XmlTag *t = Dlg->Tags.First(); t; t = Dlg->Tags.Current())
				{
					ResObject *Obj = r->CreateObject(t->Name, 0);
					if (Obj)
					{
						if (r->Res_Read(Obj, Dlg->Tags))
						{
							GView *w = dynamic_cast<GView*>(Obj);
							if (w)
							{
								GRegion r = w->GetPos();
								#ifdef WIN32
								r.Offset(-3, -17);
								#else if BEOS
								r.Offset(0, -10);
								#endif
								w->SetPos(r);
								Children.Insert(w);
							}
						}
						else
						{
							LgiMsg(NULL, "Resource read error, tag: %s", "GDlgFunc::LoadFromResource", MB_OK, t->Name);
							break;
						}
					}
					else
					{
						// oh well... next
						Dlg->Tags.Next();
					}
				}

				Status = true;
			}
		}
	}

	return Status;
}

////////////////////////////////////////////////////////////////////////
// Get language
int LgiGetLanguageId()
{
	// Check for command line override
	char Buf[64];
	if (LgiApp->GetOption(Buf, 'i'))
	{
		LangInfo *i = LangFind(0, atoi(Buf));
		if (i)
		{
			return i->Id;
		}
	}

	// Check for config setting
	XmlTag *LangConfig = LgiApp->GetConfig("Language");
	if (LangConfig)
	{
		char *Id;
		if (LangConfig->Get("Id", Id) AND
			ValidStr(Id) AND
			isdigit(*Id))
		{
			return atoi(Id);
		}
	}

	// Use a system setting
	#ifdef WIN32
	int Lang = PRIMARYLANGID(GetSystemDefaultLCID());

	TCHAR b[256];
	if (GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_IDEFAULTLANGUAGE, b, sizeof(b)) > 0)
	{
		Lang = PRIMARYLANGID(htoi(b));
	}

	for (LangInfo *i = AllLanguages; i->Id; i++)
	{
		if (i->Win32Id == Lang)
		{
			return i->Id;
		}
	}
	#endif

	// Return a default of English
	return LangIdEnglish;
}

///////////////////////////////////////////////////////////////////////////////////////
// Load string
char *LgiLoadString(int Res)
{
	LgiResources *r = LgiGetResObj();
	if (r)
	{
		#if 0 // use to test for non-resource strings
		return "@@@@";
		#else
		for (LgiStringRes *s = r->Strings.First(); s; s = r->Strings.Next())
		{
			if (s->IsString AND s->Id == Res)
			{
				return s->Str;
			}
		}
		#endif
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////
// Menus

LgiStringRes *NewMenuString(XmlTag *Tag, List<XmlTag> &Tags)
{
	if (Tag)
	{
		char *n = 0;
		if (Tag->Get("ref", n))
		{
			int Ref = atoi(n);
			XmlTag *Cur = Tags.Current();
			for (XmlTag *t = Tags.First(); t; t = Tags.Next())
			{
				if (stricmp(t->Name, "string") == 0)
				{
					if (t->Get("ref", n) AND atoi(n) == Ref)
					{
						List<XmlTag> l;
						l.Insert(t);
						l.First();

						LgiStringRes *Str = NEW(LgiStringRes);
						if (Str AND Str->Read(l))
						{
							if (Cur) Tags.Find(Cur);
							return Str;
						}
						else
						{
							DeleteObj(Str);
						}
						break;
					}
				}
			}

			if (Cur) Tags.Find(Cur);
		}
	}

	return 0;
}

/*
#ifdef WIN32
bool GSubMenu::Load(List<XmlTag> &Tags)
#else if BEOS
*/
bool GMenuLoader::Load(List<XmlTag> &Tags)
{
	bool Status = false;
	XmlTag *t = Tags.Current();
	if (t AND t->Name)
	{
		char *EndTag = 0;
		if (stricmp(t->Name, "menu") == 0)
		{
			#ifdef WIN32
			hMenu = CreateMenu();
			#endif
			EndTag = "/menu";
		}
		else
		{
			EndTag = "/submenu";
		}

		#ifdef WIN32
		if (hMenu)
		#endif
		{
			Status = true;
			for (t = Tags.Next(); t; t = Tags.Next())
			{
				if (stricmp(t->Name, EndTag) == 0)
				{
					break;
				}
				else if (Status)
				{
					if (stricmp(t->Name, "submenu") == 0)
					{
						LgiStringRes *Str = NewMenuString(t, Tags);
						if (Str AND Str->Str)
						{
							GSubMenu *Sub = AppendSub(Str->Str);
							if (Sub)
							{
								Status = Sub->Load(Tags);
							}
							DeleteObj(Str);
						}
					}
					else if (stricmp(t->Name, "menuitem") == 0)
					{
						char *n = 0;
						if (t->Get("sep", n) AND atoi(n) != 0)
						{
							AppendSeparator();
						}
						else
						{
							LgiStringRes *Str = NewMenuString(t, Tags);
							if (Str AND Str->Str)
							{
								int Enabled = true;
								if (t->Get("enabled", n))
								{
									Enabled = atoi(n);
								}
								Status = AppendItem(Str->Str, Str->Id, Enabled) != 0;
							}

							DeleteObj(Str);
						}
					}
				}
			}
		}
	}
	return Status;
}

bool GMenu::Load(GView *w, char *Res)
{
	bool Status = FALSE;

	LgiResources *r = LgiGetResObj();
	if (r)
	{
		for (LgiMenuRes *m = r->Menus.First(); m; m = r->Menus.Next())
		{
			if (stricmp(m->Name(), Res) == 0)
			{
				m->Tags.First();
				#ifdef WIN32
				Status = GSubMenu::Load(m->Tags);
				#else if BEOS
				Status = GMenuLoader::Load(m->Tags);
				#endif
				break;
			}
		}
	}

	return Status;
}

LgiResources *LgiGetResObj(bool Warn)
{
	if (LgiResources::ResourceContainers.GetItems() == 0)
	{
		NEW(LgiResources(0, Warn));
	}

	return LgiResources::ResourceContainers.First();
}
