#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "VMem.h"

char WhiteSpace[] = " \t\r\n";

// misc
char *strnchr(char *s, char c, int Len)
{
	if (s)
	{
		for (int i=0; i<Len; i++)
		{
			if (s[i] == c)
			{
				return s + i;
			}
		}
	}

	return 0;
}

char *strnstr(char *a, char *b, int n)
{
	if (a AND b)
	{
		int SLen = strlen(b);
		int DLen = strlen(a);

		while (SLen <= DLen AND n >= SLen)
		{
			int i;
			for (i=0; i<SLen AND a[i] == b[i]; i++);
			if (i == SLen) return a;

			n--;
			DLen--;
			a++;
		}
	}

	return NULL;
}

char *strnistr(char *a, char *b, int n)
{
	if (a AND b)
	{
		int SLen = strlen(b);
		int DLen = strlen(a);

		while (SLen <= DLen AND n >= SLen)
		{
			int i;
			for (i=0; i<SLen AND toupper(a[i]) == toupper(b[i]); i++);
			if (i == SLen) return a;

			n--;
			DLen--;
			a++;
		}
	}

	return NULL;
}

int strnicmp(char *a, char *b, int i)
{
	int Cmp = -1;
	if (a AND b AND i > 0)
	{
		for (Cmp = 0; i-- AND Cmp == 0; )
		{
			Cmp += tolower(*a) - tolower(*b);
			if (NOT *a OR NOT *b) break;
			a++;
			b++;
		}
	}
	
	return Cmp;
}

char *stristr(char *a, char *b)
{
	if (a AND b)
	{
		int SLen = strlen(b);
		int DLen = strlen(a);

		while (SLen <= DLen)
		{
			int i;
			for (i=0; i<SLen AND toupper(a[i]) == toupper(b[i]); i++);
			if (i == SLen) return a;

			DLen--;
			a++;
		}
	}

	return NULL;
}

int strcmp(char *a, char *b)
{
	int c = -1;
	if (a AND b)
	{
		c = 0;
		while (NOT c)
		{
			c = *a - *b;
			if (NOT *a OR NOT *b) break;
			a++;
			b++;
		}
	}
	return c;
}

int stricmp(char *a, char *b)
{
	int c = -1;
	if (a AND b)
	{
		c = 0;
		while (NOT c)
		{
			c = toupper(*a) - toupper(*b);
			if (NOT *a OR NOT *b) break;
			a++;
			b++;
		}
	}
	return c;
}

#ifndef WIN32
void strupr(char *a)
{
	for (; a AND *a; a++)
	{
		*a = toupper(*a);
	}
}

void strlwr(char *a)
{
	for (; a AND *a; a++)
	{
		*a = tolower(*a);
	}
}
#endif

char *TrimStr(char *s, char *Delim)
{
	if (s)
	{
		char *Start = s;
		while (*Start AND strchr(Delim, *Start))
		{
			Start++;
		}

		char *End = Start + strlen(Start) - 1;
		while (*End AND End > Start AND strchr(Delim, *End))
		{
			End--;
		}

		if (*Start)
		{
			int Len = (uint) End - (uint) Start + 1;
			char *n = NEW(char[Len+1]);
			if (n)
			{
				memcpy(n, Start, Len);
				n[Len] = 0;
				return n;
			}
		}
	}

	return NULL;
}

bool ValidStr(char *s)
{
	int NonWhite = 0;
	if (s)
	{
		for (; *s; s++)
		{
			if (strchr(" \t\r\n", *s) == 0)
			{
				NonWhite++;
			}
		}
	}

	return (NonWhite > 0);
}

void FormatSize(char *Str, quad Size)
{
	int K = 1024;
	int M = K * K;
	int G = K * M;

	if (Size < K)
	{
		sprintf(Str, "%i bytes", Size);
	}
	else if (Size < 10 * K)
	{
		sprintf(Str, "%.2f K", (double) Size / K);
	}
	else if (Size < M)
	{
		sprintf(Str, "%i K", Size / K);
	}
	else if (Size < G)
	{
		sprintf(Str, "%.2f M", (double) Size / M);
	}
	/*
	else if (Size < G)
	{
		sprintf(Str, "%i M", Size / M);
	}
	*/
	else
	{
		sprintf(Str, "%.2f G", (double) Size / G);
	}
}

char *NewStr(char *s, int Len)
{
	if (s)
	{
		if (Len < 0) Len = strlen(s);
		char *Ret = NEW(char[Len+1]);
		if (Ret)
		{
			if (Len > 0) memcpy(Ret, s, Len);
			Ret[Len] = 0;
			return Ret;
		}
	}
	return NULL;
}

#ifdef BEOS
#define toupper(c) ( ((c)>='a' AND (c)<='z') ? (c)-'a'+'A' : (c) )
#endif

bool MatchStr(	char *a, // the template
				char *b  // the data being tested
				)
{
	if (NOT a OR NOT b)
	{
		return false;
	}

	if (stricmp(a, "*") == 0)
	{
		// matches anything
		return true;
	}

	while (*a AND *b)
	{
		if (*a == '*')
		{
			a++;

			if (*a)
			{
				char *EndA;
				for (EndA = a; *EndA AND *EndA!='?' AND *EndA!='*'; EndA++);
				int SegLen = (int)EndA-(int)a;
				char *Seg = NewStr(a, SegLen);
				if (NOT Seg) return false;

				// find end of non match
				while (*b)
				{
					if (strnicmp(b, Seg, SegLen) == 0)
					{
						break;
					}
					b++;
				}

				DeleteArray(Seg);

				if (*b)
				{
					a += SegLen;
					b += SegLen;
				}
				else
				{
					// can't find any matching part in b
					return false;
				}
			}
			else
			{
				// '*' matches everything
				return true;
			}

		}
		else if (*a == '?')
		{
			a++;
			b++;
		}
		else
		{
			if (toupper(*a) != toupper(*b))
			{
				return false;
			}

			a++;
			b++;
		}
	}

	return	((*a == 0) OR (stricmp(a, "*") == 0)) AND
			(*b == 0);
}

int htoi(char *a)
{
	int Status = 0;

	for (; a AND *a; a++)
	{
		if (*a >= '0' AND *a <= '9')
		{
			Status <<= 4;
			Status |= *a - '0';
		}
		else if (*a >= 'a' AND *a <= 'f')
		{
			Status <<= 4;
			Status |= *a - 'a' + 10;
		}
		else if (*a >= 'A' AND *a <= 'F')
		{
			Status <<= 4;
			Status |= *a - 'A' + 10;
		}
		// else is not a HEX char
	}

	return Status;
}
