/*hdr
**	FILE:		TedView.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		23/9/98
**	DESCRIPTION:	Text editor view pane
**
**	Copyright (C) 1998, Matthew Allen
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include <string.h>

#include "Lgi.h"
#include "GTextView.h"

// Defines
#define IDM_CUT					100
#define IDM_COPY				101
#define IDM_PASTE				102
#define IDM_SAVE_TO				103

// #define toupper(c) (((c)>='a'AND(c)<='z')?(c)-'a'+'A':(c))

// Strings
char *Untitled = "<Untitled>";
char *FindMiss = "Couldn't find text.";


///////////////////////////////////////////////////////////////////////////////////////////////
bool AlphaOrDigit(char c)
{
	return IsDigit(c) OR IsAlpha(c);
}

int CountLines(char *Str, int Len)
{
	int i = 1;
	for (; Str AND *Str AND Len>=0; Len--, Str++)
	{
		if (*Str == '\n') i++;
	}
	return i;
}

void StrCpyToCrLf(char *d, char *s, int Len)
{
	char *Start = d;

	if (s AND d)
	{
		while (*s AND Len--)
		{
			if (*s == 10)
			{
				*d++ = 13;
				*d++ = 10;
			}
			else if (*s != 13)
			{
				*d++ = *s;
			}
			s++;
		}
		*d = 0;
	}
}

void StrCpyToLf(char *d, char *s, int Len)
{
	if (s AND d)
	{
		while (*s AND Len--)
		{
			if (*s == 10)
			{
				*d++ = 10;
			}
			else if (*s != 13)
			{
				*d++ = *s;
			}
			s++;
		}
		*d = 0;
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////
#ifdef WIN32
GWin32Class GTextView::WndClass;
#endif

GTextView::GTextView(	int Id,
						int x,
						int y,
						int cx,
						int cy,
						GView *notify,
						char *appname,
						GFont *LogFont)
#ifdef BEOS
						:
	GControl(			this),
	BControl(			BRect(x, y, x+cx, y+cy),
						NULL,
						"Text Editor",
						NULL,
						B_FOLLOW_ALL_SIDES,
						LGI_DRAW_VIEW_FLAGS) ///B_POINTER_EVENTS
#endif
{
	SetId(Id);
	Pos.Set(x, y, x+cx, y+cy);

	Notify = notify;
	AppName = appname;
	TopMargin = 4;
	LeftMargin = 3;
	Spacer = 2;
	LineHt = 20;
	Border = FLAT;
	ExtraFlags = 0;
	VScroll = 0;
	HScroll = 0;
	WrapAtCol = 80;
	UrlDetect = false;
	
	Sunken(false);

	#if defined WIN32
	WndDlgCode = DLGC_WANTALLKEYS;
	WndClass.Create(Class());

	// BeOS does this in the ::Attach function
	SetScrollBars(false, true);
	
	#elif defined BEOS

	ScrollView = 0;
	SetViewColor(B_TRANSPARENT_COLOR);

	#endif


	MouseSelect = FALSE;
	bool FontCreated = FALSE;

	if (LogFont)
	{
#if defined WIN32
		FontCreated = Font.CreateFont((LOGFONT*)LogFont);
#elif defined BEOS

		font_family family;
		font_style style;
		LogFont->GetFamilyAndStyle(&family, &style);
		Font.SetFamilyAndStyle(family, style);
		Font.SetSize(LogFont->BFont::Size());

		FontCreated = true;
#endif
	}
	else
	{
#if defined WIN32
		FontCreated = Font.CreateFont(	-12,
										0, 0, 0, 0, 0, 0, 0,
										ANSI_CHARSET,
										OUT_TT_PRECIS,
										CLIP_DEFAULT_PRECIS,
										DEFAULT_QUALITY,
										DEFAULT_PITCH,
										"Courier New"); //"Verdana");
#elif defined BEOS
		FontCreated = Font.CreateFont(	12,
										0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
										"Courier10 BT");
										
		SetViewColor(B_TRANSPARENT_COLOR);
#endif
	}

	if (FontCreated)
	{
		Font.Transparent(FALSE);
		LineHt = Font.Y("Ap");
		
		TextView::Flags |= TVF_DIRTY_BORDER | TVF_DIRTY_ALL;

		char Spaces[33];
		TabSize = TAB_SIZE;
		memset(Spaces, ' ', TabSize);
		Spaces[TabSize] = 0;
		TabStops = NEW(int[TABSTOPS]);
		if (TabStops)
		{
			int n = Font.X(Spaces, TabSize);
			for (int i=0; i<TABSTOPS; i++)
			{
				TabStops[i] = i * n;
			}
		}
	}

	Dirty(0);
	Doc.CursorCreate(&User, 0, 0);
	Old = User;
}

GTextView::~GTextView()
{
	DeleteObj(VScroll);
	DeleteArray(TabStops);
}

#if defined WIN32
char *GTextView::Class()
{
	return LGI_TEXT;
}

uint32 GTextView::Style()
{
	return WS_VISIBLE | WS_CHILD | WS_TABSTOP | ExtraFlags;
}

#elif defined BEOS

void GTextView::KeyDown(const char *bytes, int32 numBytes)
{
	Sys_KeyDown(bytes, numBytes);
}

void GTextView::KeyUp(const char *bytes, int32 numBytes)
{
	Sys_KeyUp(bytes, numBytes);
}

bool GTextView::Attach(GView *p)
{
	bool Status = GView::Attach(p);
	SetScrollBars(false, true);
	UpdateScrollBars();
	if (NOT GView::Visible())
	{
		ScrollView->Hide();
	}
	return Status;
}

GRegion &GTextView::GetPos()
{
	return (ScrollView) ? ScrollView->GetPos() : GView::GetPos();
}

bool GTextView::SetPos(GRegion &p, bool Repaint)
{
	if (ScrollView)
	{
		ScrollView->SetPos(p);
	}
	else
	{
		Pos = p;
		OnPosChange();
	}
	
	return TRUE;
}

void GTextView::Visible(bool i)
{
	if (GView::Visible() ^ i)
	{
		if (ScrollView)
		{
			bool Lock = ScrollView->LockLooper();
			if (i)
			{
				ScrollView->Show();
			}
			else
			{
				ScrollView->Hide();
			}
			if (Lock) ScrollView->UnlockLooper();

			if (i) SetFlag(WndFlags, GWF_VISIBLE);
			else ClearFlag(WndFlags, GWF_VISIBLE);
		}
		else
		{
			GView::Visible(i);
		}
	}
}

#endif

void GTextView::SetBorder(int b)
{
	#ifdef WIN32
	Border = b;
	#endif
}

void GTextView::Title(char *s)
{
	if (Notify)
	{
		GMessage m = CreateMsg(WM_TITLE, 0, (int) s);
		Notify->OnEvent(&m);
	}
}

void GTextView::SetStatus(int i, char *s)
{
	if (Notify)
	{
		GMessage m = CreateMsg(WM_STATUS_UPDATE, i, (int) s);
		Notify->OnEvent(&m);
	}
}

void GTextView::Dirty(int Type)
{
	TextView::Dirty(Type);

	if (TextView::Flags & TVF_DIRTY_BORDER)
	{
		if (Doc.GetName())
		{
			Title(Doc.GetName());
		}
		else
		{
			Title(Untitled);
		}
		TextView::Flags &= ~TVF_DIRTY_BORDER;
	}

	if (TextView::Flags & (TVF_DIRTY_SELECTION | TVF_DIRTY_ALL))
	{
		GView::Invalidate(NULL);
		return;
	}

	if (TextView::Flags & (TVF_DIRTY_TO_EOL | TVF_DIRTY_TO_EOP))
	{
		int x, y;
		CalcCursorPos(User, x, y);
		GRegion r(0, y, X(), y+Spacer+LineHt);
		GView::Invalidate(&r);

		if (TextView::Flags & TVF_DIRTY_TO_EOP)
		{
			r.y1 = r.y2 + 1;
			r.y2 = Y();
			GView::Invalidate(&r);
		}
		return;
	}

	if (TextView::Flags & TVF_DIRTY_CURSOR)
	{
		int x, y, Ox = 0, Oy = 0;

		CalcCursorPos(Old, x, y);
		GRegion OldR(x+Ox, y+Oy, x+Ox, y+Spacer+LineHt+Oy+1);
		GView::Invalidate(&OldR);

		CalcCursorPos(User, x, y);
		GRegion NewR(x+Ox, y+Oy, x+Ox, y+Spacer+LineHt+Oy+1);
		GView::Invalidate(&NewR);
	}
}

bool GTextView::ClearDirty(bool Ask)
{
	if (Doc.IsDirty())
	{
		int Result = IDYES;

		if (Ask)
		{
			char Str[256];
			char *Name = (Doc.GetName()) ? Doc.GetName() : Untitled;
			sprintf(Str, "Do you want to save '%s'?", Name);
			Result = LgiMsg(this, Str, "Save?", MB_YESNOCANCEL);
		}

		if (Result == IDYES)
		{
			// save the file
			if (Doc.GetName())
			{
				Save(Doc.GetName());
			}
			else
			{
				GFileSelect Select;

				Select.Parent(this);

				if (Doc.GetName())
				{
					Select.Name(Doc.GetName());
				}

				Select.Type("Code", "*.cpp;*.h;*.c");
				Select.Type("Text files", "*.txt;*.html;*.htm");
				Select.Type("All files", "*.*");

				if (Select.Save())
				{
					Save(Select.Name());
				}
			}
		}
		else if (Result == IDNO)
		{
			// do nothing
		}
		else
		{
			return FALSE;
		}
	}
	
	return TRUE;
}

void GTextView::CalcCursorPos(GCursor c, int &x, int &y)
{
	int Sy = Spacer + LineHt;
	char *Str = c.GetStr();
	if (Str)
	{
		x = LeftMargin + Font.X(Str, c.X()) - 1;
	}
	else
	{
		x = 0;
	}

	y = TopMargin + (Sy * (c.Y() - HiddenLines));
}

void GTextView::CalcMousePos(GCursor &c, GMouse m)
{
	int x = m.x;
	int y = m.y;

	#ifdef BEOS
	switch (Border)
	{
		case RAISED:
		case SUNKEN:
		{
			x -= 2;
			y -= 2;
		}
	}
	#endif

	int Sy = Spacer + LineHt;
	int Line = HiddenLines + (y / Sy);

	Doc.CursorCreate(&c, 0, limit(Line, 0, Doc.GetLines()-1));
	int CharX = Font.CharAt(x-LeftMargin, c.GetStr(), c.LineLength());

	if (Line < 0)
	{
		CharX = 0;
	}

	if (Line >= Doc.GetLines())
	{
		CharX = c.LineLength();
	}

	c.SetX(CharX);
}

bool GTextView::OnMoveCursor(int Dx, int Dy, bool NoSelect)
{
	Old = User;
	int Status = TextView::OnMoveCursor(Dx, Dy, NoSelect);
	if (Old.X() != User.X() OR Old.Y() != User.Y())
	{
		char Msg[64];
		sprintf(Msg, "(%i, %i)", User.X(), User.Y());
		SetStatus(1, Msg);
	}
	return Status;
}

void GTextView::OnSetCursor(int X, int Y)
{
	Old = User;
	TextView::OnSetCursor(X, Y);
	if (Old.X() != User.X() OR Old.Y() != User.Y())
	{
		Dirty(TVF_DIRTY_CURSOR);
		
		char Msg[64];
		sprintf(Msg, "(%i, %i)", User.X(), User.Y());
		SetStatus(1, Msg);
	}
}

void GTextView::DrawCursor(GSurface *pDC, GCursor &c, COLOUR col)
{
	if (pDC)
	{
		int x, y;
		CalcCursorPos(c, x, y);
		pDC->Colour(col, 24);
		pDC->Line(x, y, x, y + Spacer + LineHt - 2);
	}
}

bool GTextView::ClipText(char *Str, int Len)
{
	bool Status = FALSE;
	if (Str)
	{
		if (Len < 0)
		{
			Len = strlen(Str);
		}

		int Lines = CountLines(Str, Len);
		char *Text = NEW(char[Len+Lines+1]);
		if (Text)
		{
			StrCpyToCrLf(Text, Str, Len);

			GClipBoard Clip(this);
			Clip.Empty();
			Clip.Text(Text);

			DeleteObj(Text);
		}
	}
	return Status;
}

char *GTextView::ClipText()
{
	GClipBoard Clip(this);

	DeleteObj(ClipData);
	char *Str = Clip.Text();
	if (Str)
	{
		ClipData = NEW(char[strlen(Str)+1]);
		if (ClipData)
		{
			StrCpyToLf(ClipData, Str, strlen(Str));
		}

		DeleteObj(Str);
	}

	return ClipData;
}

bool GTextView::Cut()
{
	bool Status = TextView::Cut();
	UpdateScrollBars();
	return Status;
}

bool GTextView::Paste()
{
	bool Status = TextView::Paste();
	UpdateScrollBars();
	return Status;
}

bool GTextView::Copy()
{
	return TextView::Copy();
}

bool GTextView::UpdateHiddenCheck()
{
	bool Status = TextView::UpdateHiddenCheck();
	if (Status)
	{
		UpdateScrollBars();
	}
	return Status;
}

void GTextView::OnSetHidden(int Hidden)
{
	TextView::OnSetHidden(Hidden);
	UpdateScrollBars();
}

void GTextView::UpdateScrollBars()
{
	int Sy = Spacer + LineHt;

	#if defined WIN32
	
	RECT Client;
	if (hWnd AND GetClientRect(hWnd, &Client))
	{
		DisplayLines = ((Client.bottom - Client.top) / Sy);
	}

	// SetScrollBars(FALSE, Doc.GetLines() > DisplayLines);
	if (VScroll)
	{
		VScroll->SetLimits(0, Doc.GetLines()-1);
		VScroll->SetPage(DisplayLines);
		VScroll->SetPos(HiddenLines);
	}
	
	#elif defined BEOS
	
	bool Lock = LockLooper();
		
	DisplayLines = Pos.Y() / Sy;
	
	BScrollBar *Bar = (ScrollView)?ScrollView->ScrollBar(B_VERTICAL):0;
	if (Bar)
	{
		// printf("Hidden(%i) Display(%i) Lines(%i)\n", HiddenLines, DisplayLines, Doc.GetLines());
	
		Bar->SetRange(0, max(Doc.GetLines() - DisplayLines, 0));
		Bar->SetSteps(1.0, DisplayLines);
		if (DisplayLines < Doc.GetLines())
		{
			float p = (float) DisplayLines / Doc.GetLines();
			Bar->SetProportion(p);
		}
		Bar->SetValue(HiddenLines);
	}

	if (Lock) UnlockLooper();
	
	#endif
}

void GTextView::OnPosChange()
{
	UpdateScrollBars();
	Dirty(TVF_DIRTY_ALL);
}

void GTextView::OnCreate()
{
	UpdateScrollBars();
}

bool GTextView::OnMultiLineTab(bool In)
{
	if ((TextView::Flags & TVF_SELECTION) AND Start.Y() != End.Y())
	{
		GCursor *a, *b;

		if (Start > End)
		{
			a = &End;
			b = &Start;
		}
		else
		{
			a = &Start;
			b = &End;
		}

		a->SetX(0);
		GCursor c = *a;

		for (; c.Y()<=b->Y(); c.MoveY(1))
		{
			if (c.Y() == b->Y() AND b->X() == 0)
			{
				break;
			}

			if (In)
			{
				if (c.LineLength() > 0)
				{
					char Tab[2] = { 9, 0 };
					Doc.Insert(&c, Tab, 1);
				}
			}
			else
			{
				char *Str = c;
				int Len = 0;

				if (*Str == 9)
				{
					Len = 1;
				}
				else
				{
					for (; *Str == ' ' AND Len <= 8; Str++)
					{
						Len++;
					}
				}

				if (Len)
				{
					Doc.Delete(&c, Len);
				}
			}

			if (c.Y() >= Doc.GetLines() - 1)
			{
				break;
			}
		}

		Dirty(TVF_DIRTY_ALL);

		bool MoveLastToEnd = (b->X() > 0);


		Doc.CursorCreate(a, 0, a->Y());
		a->SetX(0);
		Doc.CursorCreate(b, 0, b->Y());
		if (MoveLastToEnd) b->SetX(b->LineLength());
		Doc.CursorCreate(&User, User.X(), User.Y());

		return TRUE;
	}

	return FALSE;
}

void GTextView::AfterInsertText(char *Text, int Len)
{
	if (Text)
	{
		if (Len < 0)
		{
			Len = strlen(Text);
		}

		if (WrapType == TEXTED_WRAP_NONREFLOW AND User.X() >= WrapAtCol AND Len == 1)
		{
			GCursor Ms = User;
			int CharsToRight = User.LineLength() - User.X() - 1;

			OnMoveCursor(-1, 0, TRUE);
			
 			char c = *User;
			bool White = strchr(WhiteSpace, c) != 0;
			bool Delimit = strchr(Delimiters, c) != 0;
			bool Alpha = IsDigit(c) OR IsAlpha(c);

			if (NOT White)
			{
				while (	(NOT strchr(WhiteSpace, *User)) AND
						User.X() > 0 AND
						OnMoveCursor(-1, 0, TRUE))
				{
				}
			}

			int OldY = User.Y();
			OnMoveCursor(1, 0, TRUE);
			Doc.Insert(&User, "\n", 1);
			Doc.CursorCreate(&User, 0, OldY+1);
			OnMoveCursor(User.LineLength() - CharsToRight - 1, 0, TRUE);

			Dirty(TVF_DIRTY_ALL);
		}
	}

	UpdateScrollBars();
}

void GTextView::AfterDeleteText()
{
	UpdateScrollBars();
}

bool GTextView::OnFind()
{
	if (BtnFindNext() AND strlen(Find))
	{
		bool Match = DoFind();
		if (Match)
		{
			return GFindReplace::OnFind();
		}
		else
		{
			LgiMsg(this, FindMiss, (AppName)?AppName:(char*)"Find", MB_OK);
		}
	}

	return 0;
}

bool GTextView::OnReplace()
{
	int FindLen = strlen(Find);
	int ReplaceLen = strlen(Replace);

	if (FindLen AND ReplaceLen)
	{
		if (BtnFindNext())
		{
			bool Match = DoFind();
			if (NOT Match)
			{
				LgiMsg(this, FindMiss, (AppName)?AppName:(char*)"Replace", MB_OK);
			}
		}
		else if (BtnReplace())
		{
			GCursor StartPos = User;

			if (TextView::Flags & TVF_SELECTION)
			{
				if (Start < End)
				{
					User = Start;
				}
				else
				{
					User = End;
				}

				OnEndSelection();
			}

			if (DoFind())
			{
				if (StartPos == User)
				{
					// The current selection was ready to be
					// replaced
					OnDeleteSelection(FALSE);
					Doc.Insert(&User, Replace);
					Doc.CursorCreate(&User, User.X(), User.Y());
					User.MoveX(ReplaceLen);

					if (NOT DoFind())
					{
						return GFindReplace::OnReplace();
					}
				}
			}
			else
			{
				LgiMsg(this, FindMiss, (AppName)?AppName:(char*)"Replace", MB_OK);
			}
		}
		else if (BtnReplaceAll())
		{
			if (TextView::Flags & TVF_SELECTION)
			{
				if (Start < End)
				{
					User = Start;
				}
				else
				{
					User = End;
				}

				OnEndSelection();
			}

			int Items = 0;
			while (DoFind())
			{
				OnDeleteSelection(FALSE);
				Doc.Insert(&User, Replace);
				Doc.CursorCreate(&User, User.X(), User.Y());
				User.MoveX(ReplaceLen);

				Items++;
			}

			if (NOT Items)
			{
				LgiMsg(this, FindMiss, (AppName)?AppName:(char*)"Replace", MB_OK);
			}
			else
			{
				return GFindReplace::OnReplace();
			}
		}
	}

	return 0;
}

bool GTextView::DoFind()
{
	bool Match = FALSE;

	int Length = strlen(Find);
	if (Length > 0)
	{
		int XOff = User.X();

		// Find
		for (int Line=User.Y(); NOT Match AND Line<Doc.GetLines(); Line += 32)
		{
			int EndLine = min(Line+31, Doc.GetLines());
			TextLock Lock;
			if (Doc.Lock(&Lock, Line, EndLine, GetCodePage()))
			{
				for (int l=Line; NOT Match AND l<=EndLine; l++)
				{
					char *c = Lock[l];
					if (c)
					{
						char *Str = 0;
						int LineLength = StrLen(c);
						if (XOff < LineLength)
						{
							if (CaseSensitive())
							{
								Str = strnstr(c+XOff, Find, LineLength-XOff);
							}
							else
							{
								Str = strnistr(c+XOff, Find, LineLength-XOff);
							}

							if (Str)
							{
								Match = TRUE;
								int X = (int) (Str - c);

								GCursor Found;
								Doc.CursorCreate(&Found, X, l);

								if (WholeWord())
								{
									if (X > 0)
									{
										char Before = c[X-1];

										if (	NOT strchr(WhiteSpace, Before) AND
											NOT strchr(Delimiters, Before))
										{
											Match = FALSE;
										}
									}

									if (X + Length < LineLength - 1)
									{
										char After = c[X+Length];

										if (	NOT strchr(WhiteSpace, After) AND
											NOT strchr(Delimiters, After))
										{
											Match = FALSE;
										}
									}
								}

								if (Match)
								{
									if (TextView::Flags & TVF_SELECTION)
									{
										OnEndSelection();
									}

									OnStartSelection(&Found);
									End.MoveX(Length);
									User = End;
									UpdateHiddenCheck();
								}
							}
						}
					}

					XOff = 0;
				}

				Doc.UnLock(&Lock);
			}
		}
	}

	return Match;
}

#define UrlChar(c) (strchr(UrlDelim, (c)) OR AlphaOrDigit((c)))

// Selects a word or URL
void GTextView::SelectWord(GMouse &m)
{
	GCursor Ms;

	CalcMousePos(Ms, m);
	OnSetCursor(Ms.X(), Ms.Y());
	char c = *User;
	bool White = strchr(WhiteSpace, c) != 0;
	bool Delimit = strchr(Delimiters, c) != 0;
	bool Alpha = IsDigit(c) OR IsAlpha(c);

	// Url detect algorithm gets first crack at it
	if (UrlDetect AND NOT White)
	{
		bool UrlFound = FALSE;
		char *UrlDelim = "~/:%+-?@&$#._=,";

		// seek back to the start of a possible URL
		while (UrlChar(*User) AND OnMoveCursor(-1, 0));

		// move back onto the start of the string
		if (NOT UrlChar(*User)) OnMoveCursor(1, 0);
		
		OnStartSelection(&User);
		TextView::Flags |= TVF_SHIFT;

		// find the end of the string
		while (UrlChar(*User) AND OnMoveCursor(1, 0));

		GCursor S;
		int Len;

		if (Start < End)
		{
			S = Start;
			Len = End - Start;
		}
		else
		{
			S = End;
			Len = Start - End;
		}
		
		char *Url = NEW(char[Len+1]);
		if (Url)
		{
			memcpy(Url, (char*) S, Len);
			Url[Len] = 0;

			// see if it's got the valid prefix
			char *Match[] = {	"http://",
								"https://",
								"ftp://",
								"mailto:",
								"www.",
								"ftp.",
								0};
			char *Prefix[] = {	"",
								"",
								"",
								"",
								"http://",
								"ftp://",
								0};
			for (int i=0; Match[i]; i++)
			{
				if (strnicmp(Url, Match[i], strlen(Match[i])) == 0)
				{
					// Ok so trigger the event

					char *Temp = NEW(char[strlen(Prefix[i])+strlen(Url)+1]);
					if (Temp)
					{
						*Temp = 0;
						strcat(Temp, Prefix[i]);
						strcat(Temp, Url);
						OnUrl(Url);
						DeleteArray(Temp);
					}
					UrlFound = TRUE;
					break;
				}
			}

			if (NOT UrlFound AND strchr(Url, '@'))
			{
				// possibly a email address without the "mailto:"
				// some better detection may be in order here.
				OnUrl(Url);
				UrlFound = TRUE;
			}

			DeleteArray(Url);
		}

		if (UrlFound)
		{
			// no need to hang around
			return;
		}

		// put the cursor back to it's original location for the normal word find
		OnSetCursor(Ms.X(), Ms.Y());
	}

	// Normal select word algorithm
	while (	((White AND strchr(WhiteSpace, *User)) OR 
			(Delimit AND strchr(Delimiters, *User)) OR 
			(Alpha	AND NOT strchr(WhiteSpace, *User)
			AND NOT strchr(Delimiters, *User))) AND
			OnMoveCursor(-1, 0))
	{
	}
	OnMoveCursor(1, 0);
	OnStartSelection(&User);
	TextView::Flags |= TVF_SHIFT;

	while (	((White AND strchr(WhiteSpace, *User)) OR 
			(Delimit AND strchr(Delimiters, *User)) OR 
			(Alpha	AND NOT strchr(WhiteSpace, *User)
			AND NOT strchr(Delimiters, *User))) AND
			OnMoveCursor(1, 0))
	{
	}
}

void GTextView::OnMouseWheel(double Lines)
{
	if (Doc.GetLines() > DisplayLines)
	{
		int NewHidden = (int) limit((int) HiddenLines + Lines,
0,
Doc.GetLines()- DisplayLines);
		if (NewHidden != HiddenLines)
		{
			OnSetHidden(NewHidden);
		}
	}
}

void GTextView::OnKey(GKey &K)
{
	if (NOT K.IsChar AND K.Down)
	{
		switch (K.c)
		{
			case 'a':
			case 'A':
			{
				// select all
				if (K.Ctrl())
				{
					SelectAll();
				}
				break;
			}
			case 'x':
			case 'X':
			{
				if (K.Ctrl())
				{
					Cut();
				}
				return;
			}
			case 'c':
			case 'C':
			{
				if (K.Ctrl())
				{
					Copy();
				}					
				return;
			}
			case 'v':
			case 'V':
			{
				if (K.Ctrl())
				{
					Paste();
				}
				return;
			}
			/*
			case 's':
			case 'S':
			{
				if (K.Ctrl())
				{
					OnSave();
				}
				return;
			}
			*/
			case VK_F3:
			{
				if (NOT DoFind())
				{
					LgiMsg(this, FindMiss, (AppName)?AppName:(char*)"Find", MB_OK);
				}
				return;
			}
			case 'f':
			case 'F':
			{
				if (K.Ctrl())
				{
					OpenFind();
				}
				break;
			}
			case 'h':
			case 'H':
			{
				if (K.Ctrl())
				{
					OpenReplace();
				}
				break;
			}
		}
	}

	ProcessKey(K);
}

void GTextView::OnPulse()
{
	if (MouseSelect)
	{
		static GdcPt2 Last(-1, -1);
		GdcPt2 Pt(0, 0);

		#if defined WIN32

		LgiGetCursorPos(&Pt);
		POINT p;
		p.x = Pt.x;
		p.y = Pt.y;
		ScreenToClient(hWnd, &p);
		Pt.x = p.x;
		Pt.y = p.y;

		#elif defined BEOS

		BPoint Cur;
		uint32 Button;
		GetBView()->GetMouse(&Cur, &Button);
		Pt.x = Cur.x;
		Pt.y = Cur.y;

		#endif

		GMouse m;
		ZeroObj(m);
		m.x = Pt.x;
		m.y = Pt.y;

		GCursor Ms;
		CalcMousePos(Ms, m);
		if (Last.x != Ms.X() OR Last.y != Ms.Y())
		{
			OnSetCursor(Ms.X(), Ms.Y());
			Last.x = Ms.X();
			Last.y = Ms.Y();
		}
	}
}

#define TVCP_BASE			1000

void GTextView::OnMouseClick(GMouse &m)
{
	if (m.Down())
	{
		Focus(true);

		if (m.Flags & MK_LEFT)
		{
			if (m.Double)
			{
				SelectWord(m);
			}
			else
			{
				// start selection with mouse
				GCursor Ms;
				CalcMousePos(Ms, m);
				OnSetCursor(Ms.X(), Ms.Y());
				OnStartSelection(&User);

				Capture(TRUE);
				MouseSelect = TRUE;
				SetPulse(100);
			}
		}
		
		if (m.Flags & MK_RIGHT)
		{
			// context menu
			GdcPt2 Pt;
			if (LgiGetCursorPos(&Pt))
			{
				GSubMenu *RClick = NEW(GSubMenu);
				if (RClick)
				{
					RClick->AppendItem("Cut", IDM_CUT, true);
					RClick->AppendItem("Copy", IDM_COPY, true);
					RClick->AppendItem("Paste", IDM_PASTE, true);
					RClick->AppendSeparator();

					RClick->AppendItem("Save text to file", IDM_SAVE_TO, true);
					RClick->AppendSeparator();

					GSubMenu *Cp = RClick->AppendSub("Code Page");
					if (Cp)
					{
						GMenuItem *Items[] = {
							Cp->AppendItem("us-ascii", TVCP_BASE+TVCP_US_ASCII, true),
							Cp->AppendItem("iso-8859-2", TVCP_BASE+TVCP_ISO_8859_2, false),
							Cp->AppendItem("win-1250", TVCP_BASE+TVCP_WIN_1250, true)};
						if (Items[GetCodePage()])
						{
							Items[GetCodePage()]->Checked(true);
						}
					}
				}

				int Id = RClick->Float(this, Pt.x, Pt.y);
				switch (Id)
				{
					case IDM_CUT:
					{
						Cut();
						break;
					}
					case IDM_COPY:
					{
						Copy();
						break;
					}
					case IDM_PASTE:
					{
						Paste();
						break;
					}
					case IDM_SAVE_TO:
					{
						GFileSelect Select;

						Select.Parent(this);
						Select.Type("Text", "*.txt;*.html;*.htm");
						Select.Type("All files", "*.*");

						if (Select.Save())
						{
							char Str[256];
							strcpy(Str, Select.Name());
							if (NOT strchr(Str, '.'))
							{
								strcat(Str, ".txt");
							}

							if (NOT Save(Str))
							{
								LgiMsg(this,	"Couldn't save to '%s'\n"
												"File may be read only or locked by another process", "Error", MB_OK, Select.Name());
							}
						}
						break;
					}
					default:
					{
						if (Id >= TVCP_BASE AND Id < TVCP_BASE+TVCP_MAX)
						{
							if (GetCodePage() != Id - TVCP_BASE)
							{
								SetCodePage(Id - TVCP_BASE);
								Dirty(TVF_DIRTY_ALL);

								/*
								GFile f;
								if (f.Open("c:\\temp\\cp.txt", O_WRITE))
								{
									for (int i=1; i<256; i++)
									{
										char Str[256];
										sprintf(Str, "%i:\t%c\r\n", i, i);
										f.Write(Str, strlen(Str));
									}
								}
								*/
							}
						}
						break;
					}
				}
			}
		}
	}
	else
	{
		if (MouseSelect)
		{
			Capture(false);
			SetPulse();
			MouseSelect = FALSE;

			if (Start == End)
			{
				OnEndSelection();
			}
		}
	}
}

void GTextView::OnMouseMove(GMouse &m)
{
}

void GTextView::OutputText(GSurface *pDC, int x, int y, char *Str, int Len)
{
	#ifdef WIN32
	ushort *Map = GetCodePageMap();
	if (Map)
	{
		if (Len < 0) Len = strlen(Str);
		ushort *Mapped = NEW(ushort[Len+1]);
		if (Mapped)
		{
			int i;
			for (i=0; i<Len; i++)
			{
				Mapped[i] = (Str[i] & 0x80) ? Map[Str[i]&0x7F] : Str[i];
			}
			Mapped[i] = 0;

			Font.TextW(pDC, x, y, Mapped, Len);
			DeleteArray(Mapped);
		}
	}
	else
	#endif
	{
		Font.Text(pDC, x, y, Str, Len);
	}
}

void GTextView::OnPaint(GSurface *pDC)
{
	#ifdef WIN32
	COLOUR Fore = GetSysColor(COLOR_WINDOWTEXT);
	COLOUR Back = GetSysColor(COLOR_WINDOW);
	#else if BEOS
	COLOUR Fore = Rgb24(0, 0, 0);
	COLOUR Back = Rgb24(255, 255, 255);
	#endif
	COLOUR Cursor = 0, DiabledCursor = Rgb24(192, 192, 192);
	int Sy = Spacer + LineHt;

	Font.Colour(Fore, Back);
	Font.SetTabs(TABSTOPS, TabStops, LeftMargin);

	TextLock Lock;
	Doc.Lock(&Lock, HiddenLines, DisplayLines+1);

	// update everthing
	char *StartPtr = 0, *EndPtr = 0;
	int Sel = 0;
	if (TextView::Flags & TVF_SELECTION)
	{
		Sel = 1;
		if (Start > End)
		{
			StartPtr = End;
			EndPtr = Start;
		}
		else
		{
			StartPtr = Start;
			EndPtr = End;
		}
	}

	char *Next = StartPtr;
	char *StartLock = Lock[HiddenLines];
	if (StartPtr AND (StartLock > StartPtr) AND (NOT (StartLock > EndPtr)))
	{
		Font.Colour(Font.Back(), Font.Fore());

		Next = EndPtr;
		Sel = 2;
	}

	// Set clip region
	GRegion r(0, 0, Pos.X()-1, Pos.Y()-1);

	#if defined BEOS

	if (Raised() OR Sunken())
	{
		LgiWideBorder(pDC, r, Raised() ? RAISED : SUNKEN);
		pDC->ClipRgn(&r);
	}

	#elif defined WIN32

	RECT rc;
	GetClientRect(hWnd, &rc);
	r = rc;

	#endif

	int ScrY = TopMargin;
	int EndY = ScrY + Sy - 1;

	// Draw top margin
	pDC->Colour(Back, 24);
	pDC->Rectangle(0, 0, pDC->X(), TopMargin-1);

	for (int i=0; i<=DisplayLines; i++, ScrY+=Sy, EndY+=Sy)
	{
		char *BeginLine = Lock[HiddenLines+i];
		char *c = BeginLine;

		pDC->Colour(Back, 24);
		if (c)
		{	
			int Len = StrLen(c);
			int SpacerY = ScrY + LineHt;
			int DrawX = LeftMargin;
			int TextX = Font.X(c, Len);

			// to left margin
			pDC->Rectangle(0, ScrY, LeftMargin-1, EndY);
			
			while (Sel AND Next >= c AND Next <= c + Len)
			{
				int ThisSide = (int) (Next - c);
				int Width = Font.X(BeginLine, (int) (Next - BeginLine)) - Font.X(BeginLine, (int) (c - BeginLine));

				OutputText(pDC, DrawX, ScrY, c, ThisSide);

				// spacer below text
				pDC->Colour(Font.Back(), 24);
				pDC->Rectangle(DrawX, SpacerY, DrawX+Width, EndY);

				Font.Colour(Font.Back(), Font.Fore());
				DrawX += Width;
				TextX -= Width;
				c += ThisSide;
				Len -= ThisSide;
				
				if (Sel == 1)
				{
					Sel = 2;
					Next = EndPtr;
				}
				else
				{
					Sel = 0;
					Next = 0;
				}
			}

			OutputText(pDC, DrawX, ScrY, c, Len);

			// draw space below text
			pDC->Colour(Font.Back(), 24);
			pDC->Rectangle(DrawX, SpacerY, DrawX+TextX, EndY);

			int RightMargin = DrawX + TextX;
			if (Sel == 2)
			{
				// the EOL char needs to be seen when selected
				RightMargin += 5;
				pDC->Rectangle(DrawX+TextX, ScrY, RightMargin-1, EndY);
			}

			pDC->Colour(Back, 24);

			// from right margin
			pDC->Rectangle(RightMargin, ScrY, pDC->X(), EndY);
		}
		else
		{
			pDC->Rectangle(0, ScrY, pDC->X()-1, EndY);
		}
	}

	// draw from the last text down
	pDC->Colour(Back, 24);
	pDC->Rectangle(0, ScrY, pDC->X(), pDC->Y());

	if (Focus())
	{
		// put the cursor back
		DrawCursor(pDC, User, Doc.AcceptEdits() ? Cursor : DiabledCursor);
	}

	Doc.UnLock(&Lock);
	Clean();
}

void GTextView::OnFocus(bool f)
{
	if (f)
	{
		SetFlag(TextView::Flags, TVF_FOCUS);
	}
	else
	{
		ClearFlag(TextView::Flags, TVF_FOCUS);
	}
	Dirty(TVF_DIRTY_CURSOR);
}

int GTextView::OnEvent(GMessage *Msg)
{
	GFindReplace::OnEvent(Msg);

	switch (Message(Msg))
	{
		#if defined WIN32
		case WM_NCPAINT:
		{
			switch (Border)
			{
				case SUNKEN:
				case RAISED:
				{
					HDC hDC = GetWindowDC(hWnd);
					if (hDC)
					{
						GScreenDC DC(hDC, hWnd);
						GRegion r(0, 0, Pos.X()-1, Pos.Y()-1);
						LgiWideBorder(&DC, r, Border);
						ReleaseDC(hWnd, hDC);
					}
					break;
				}
			}
			break;
		}
		case WM_NCCALCSIZE:
		{
			// do default processing on the message
			// int Status = DefWindowProc(hWnd, Msg->Msg, Msg->a, Msg->b);
			RECT *i = (RECT*) Msg->b;
			switch (Border)
			{
				case SUNKEN:
				case RAISED:
				{
					i->top += 2;
					i->left += 2;
					i->right -= 2;
					i->bottom -= 2;
					break;
				}
			}
			break;
		}
		case WM_CREATE:
		{
			GFindReplace::SetParent(this);
			GTextView::OnCreate();
			break;
		}
		case WM_VSCROLL:
		{
			if (Doc.GetLines() > DisplayLines)
			{
				int nScrollCode = (int) LOWORD(Msg->a);
				int nPos = (short int) HIWORD(Msg->a);
				HWND hwndScrollBar = (HWND) Msg->b;

				switch (nScrollCode)
				{
					case SB_BOTTOM:
					case SB_LINEDOWN:
					{
						if (HiddenLines < Doc.GetLines()-DisplayLines)
						{
							OnSetHidden(HiddenLines+1);
						}
						break;
					}
					case SB_TOP:
					case SB_LINEUP:
					{
						if (HiddenLines > 0)
						{
							OnSetHidden(HiddenLines-1);
						}
						break;
					}
					case SB_PAGEDOWN:
					{
						int Move = DisplayLines-1;
						OnSetHidden(HiddenLines+Move);
						OnMoveCursor(0, Move);
						break;
					}
					case SB_PAGEUP:
					{
						int Move = 1-DisplayLines;
						OnSetHidden(HiddenLines+Move);
						OnMoveCursor(0, Move);
						break;
					}
					case SB_THUMBPOSITION:
					case SB_THUMBTRACK:
					{
						OnSetHidden(nPos);
						break;
					}
				}
			}
			break;
		}

		// Win32 edit control compatibility messages
		case WM_GETTEXTLENGTH:
		{
			return max(0, Doc.Length-1);
			break;
		}
		case WM_GETTEXT:
		{
			char *Text = Name();
			if (Text AND MsgA(Msg) > 0 AND MsgB(Msg))
			{
				int Copy = min(MsgA(Msg) - 1, strlen(Text));
				char *Dest = (char*) MsgB(Msg);
				memcpy(Dest, Text, Copy);
				Dest[Copy] = 0;
				return Copy;
			}
			return 0;
			break;
		}
		case WM_SETTEXT:
		{
			if (MsgB(Msg))
			{
				Name((char*) MsgB(Msg));
				return true;
			}
			return false;
			break;
		}
		#elif defined BEOS
		case B_CUT:
		{
			Cut();
			break;
		}
		case B_COPY:
		{
			Copy();
			break;
		}
		case B_PASTE:
		{
			Paste();
			break;
		}
		case WM_VSCROLL:
		{
			BScrollBar *Bar = (ScrollView)?ScrollView->ScrollBar(B_VERTICAL):0;
			if (Bar AND HiddenLines != Bar->Value())
			{
				OnSetHidden(Bar->Value());
			}
			break;
		}
		#endif
	}

	return GView::OnEvent(Msg);
}

bool GTextView::Open(char *Name)
{
	bool Status = TextView::Open(Name);
	if (Status)
	{
		Doc.CursorCreate(&User, 0, 0);
		SetStatus(1, "(0, 0)");

		HiddenLines = 0;
		UpdateScrollBars();
	}
	return Status;
}

bool GTextView::Name(char *Name)
{
	bool Status = FALSE;
	
	HiddenLines = 0;
	Status = Doc.Import((Name)?Name:(char*)"");
	Doc.CursorCreate(&User, 0, 0);

	SetStatus(1, "(0, 0)");
	UpdateScrollBars();
	GView::Invalidate();

	return Status;
}

char *GTextView::Name()
{
	GCursor C;
	if (Doc.CursorCreate(&C, 0, 0))
	{
		return (char*) C;
	}
	return NULL;
	return 0;
}

bool GTextView::SetScrollBars(bool x, bool y)
{
	bool Status = FALSE;
	static bool Processing = FALSE;

	#if defined WIN32
	if (NOT Processing)
	{
		Processing = TRUE;

		if (y AND NOT VScroll)
		{
			VScroll = NEW(GScrollBar);
			if (VScroll)
			{
				ExtraFlags |= WS_VSCROLL;
				VScroll->SetVertical(TRUE);
				VScroll->SetParent(this);
				Status = TRUE;
			}
		}
		if (NOT y AND VScroll)
		{
			ExtraFlags &= ~WS_VSCROLL;
			DeleteObj(VScroll);
			Status = TRUE;
		}
		
		if (x AND NOT HScroll)
		{
			HScroll = NEW(GScrollBar);
			if (HScroll)
			{
				ExtraFlags |= WS_HSCROLL;
				HScroll->SetVertical(FALSE);
				HScroll->SetParent(this);
				Status = TRUE;
			}
		}
		if (NOT x AND HScroll)
		{
			ExtraFlags &= ~WS_HSCROLL;
			DeleteObj(HScroll);
			Status = TRUE;
		}

		Processing = FALSE;
	}
	#elif defined BEOS
	DeleteObj(ScrollView);
	ScrollView = NEW(GScrollView("GViewScroller", this, B_FOLLOW_LEFT | B_FOLLOW_TOP, 0, x, y));
	#endif
	
	return Status;
}

void GTextView::OnUrl(char *Url)
{
}

void GTextView::OnEscape(GKey &K)
{
}

void GTextView::SelectAll()
{
	OnSetCursor(0, 0);
	OnStartSelection(&User);
	OnSetCursor(1024, Doc.GetLines());
}

void GTextView::SelectNone()
{
	OnEndSelection();
}
