#include <stdlib.h>
#include <stdio.h>

#include "Lgi.h"
#include "GTextView3.h"
#include "GInput.h"

#ifdef GTEXTVIEW_RES
#include "resdefs.h"
#else
#define GTEXTVIEW_BASE				29000
#define GTEXTVIEW_EMAIL_TO			(GTEXTVIEW_BASE+1)
#define GTEXTVIEW_AUTO_INDENT		(GTEXTVIEW_BASE+2)
#define GTEXTVIEW_CODEPAGE			(GTEXTVIEW_BASE+3)
#define GTEXTVIEW_OPENURL			(GTEXTVIEW_BASE+4)
#define GTEXTVIEW_COPYLINK			(GTEXTVIEW_BASE+5)
#define GTEXTVIEW_CUT				(GTEXTVIEW_BASE+6)
#define GTEXTVIEW_COPY				(GTEXTVIEW_BASE+7)
#define GTEXTVIEW_PASTE				(GTEXTVIEW_BASE+8)
#endif

#ifdef BEOS
#define DOUBLE_BUFFER_PAINT
#endif

#define ALLOC_BLOCK					64
#define IDC_VS						1000

#define IsWhiteSpace(c)				(Strchr(WhiteSp, c) != 0)
#define IsDelimiter(c)				(Strchr(Delimiters, c) != 0)
#define IsWordBoundry(c)			(Strchr(WhiteSp, c) OR Strchr(Delimiters, c))
#define UrlChar(c)					(Strchr(UrlDelim, (c)) OR AlphaOrDigit((c)))
#define EmailChar(c)				(Strchr(G("._-:"), (c)) OR AlphaOrDigit((c)))

#ifndef IDM_OPEN
#define IDM_OPEN					1
#endif
#ifndef IDM_NEW
#define	IDM_NEW						2
#endif
#ifndef IDM_COPY
#define IDM_COPY					3
#endif
#ifndef IDM_CUT
#define IDM_CUT						4
#endif
#ifndef IDM_PASTE
#define IDM_PASTE					5
#endif
#define IDM_COPY_URL				6
#define IDM_AUTO_INDENT				7

#define CODEPAGE_BASE				100

#ifndef WIN32
#define toupper(c)					(((c)>='a'AND(c)<='z') ? (c)-'a'+'A' : (c))
#endif

//////////////////////////////////////////////////////////////////////

#if 1

#include <wchar.h>

// Unicode string support
#define G(s)					L##s
#define Strnicmp(a, b, n)		_wcsnicmp(a, b, n)
#define Strncmp(a, b, n)		wcsncmp(a, b, n)
#define Strlen(a)				wcslen(a)
#define Strchr(s, c)			wcschr(s, c)

void Str2native(gchar *a, char *b)
{
	if (a AND b)
	{
		while (*a)
		{
			*a++ = *b++;
		}
		*a++ = 0;
	}
}

gchar *Strnew(gchar *s, int l=-1)
{
	if (l<0)
	{
		l = (s) ? Strlen(s) : 0;
	}

	gchar *r = NEW(gchar[l+1]);
	if (r)
	{
		memcpy(r, s, sizeof(gchar) * l);
		r[l] = 0;
	}

	return r;
}

#else

// 8bit string support
#define G(c)					##c
#define Strnicmp(a, b, n)		strnicmp(a, b, n)
#define Strncmp(a, b, n)		strncmp(a, b, n)
#define Strlen(a)				strlen(a)
#define Strnew					NewStr
#define Str2native(a, b)		strcpy(a, b)
#define Strchr(s, c)			strchr(s, c)

#endif

////////////////////////////////////////////////////////////////////////
gchar WhiteSp[] = G(" \t\r\n");
gchar Delimiters[] = G("!@#$%^&*()'\":;,.<>/?[]{}-=+\\|`~");
gchar *UrlDelim = G("!~/:%+-?@&$#._=");

bool AlphaOrDigit(gchar c)
{
	return IsDigit(c) OR IsAlpha(c);
}


//////////////////////////////////////////////////////////////////////
char *TvLoadString(int Id)
{
	// check resource for strings
	char *s = LgiLoadString(Id);
	if (NOT s)
	{
		// setup some defaults
		switch (Id)
		{
			case GTEXTVIEW_EMAIL_TO:
				s = "New Email to...";
				break;
			case GTEXTVIEW_AUTO_INDENT:
				s = "Auto Indent";
				break;
			case GTEXTVIEW_CODEPAGE:
				s = "Code page";
				break;
			case GTEXTVIEW_OPENURL:
				s = "Open URL";
				break;
			case GTEXTVIEW_COPYLINK:
				s = "Copy link location";
				break;
			case GTEXTVIEW_CUT:
				s = "Cut";
				break;
			case GTEXTVIEW_COPY:
				s = "Copy";
				break;
			case GTEXTVIEW_PASTE:
				s = "Paste";
				break;
		}
	}

	return s;
}

//////////////////////////////////////////////////////////////////////
GTextView::GTextView(	int Id,
						int x, int y, int cx, int cy,
						GView *notify,
						char *appname,
						GdcFontType *FontType)
{
	// init vars
	MappedChars = 0;
	MappedShorts = 0;
	MappedLen = 0;
	LineY = 1;

	// setup window
	SetId(Id);
	Pos.ZOff(cx-1, cy-1);
	Pos.Offset(x, y);
	SetNotify(notify);
	GFindReplace::SetParent(this);

	// default options
	CodePage = GDCF_ASCII;
	WrapAtCol = 0;
	WrapType = TEXTED_WRAP_NONE;
	UrlDetect = true;
	AcceptEdit = true;
	Dirty = false;
	#ifdef WIN32
	CrLf = true;
	#else
	CrLf = false;
	#endif
	BlueUnderline = 0;
	AutoIndent = false;
	ReadOnlyBackColour = LC_WORKSPACE;

	// debug times
	_PourTime = 0;
	_StyleTime = 0;
	_PaintTime = 0;

	// data
	Alloc = ALLOC_BLOCK;
	Text = NEW(gchar[Alloc]);
	if (Text) *Text = 0;
	Cursor = 0;
	CharSize = 1; // or 2
	Size = 0;

	// display
	Lines = 1;
	LeftMargin = 4;
	SelStart = SelEnd = -1;
	DocOffset = 0;
	Track = false;

	if (FontType)
	{
		Font = FontType->CreateFont();
	}
	else
	{
		Font = NEW(GFont);
		if (Font)
		{
			#if defined WIN32
			Font->Create("Courier New", 10);
			#elif defined BEOS
			Font->Create("Courier10 BT", 12);
			#endif
		}
	}
	if (Font)
	{
		#if defined WIN32
		WndStyle |= WS_TABSTOP;
		WndDlgCode = DLGC_WANTALLKEYS;
		#elif defined BEOS
		GetBView()->SetFlags(GetBView()->Flags() | B_NAVIGABLE);
		GetBView()->SetViewColor(B_TRANSPARENT_COLOR);
		#endif

		BlueUnderline = NEW(GFont);
		if (BlueUnderline)
		{
			*BlueUnderline = *Font;
			BlueUnderline->Underline(true);
			BlueUnderline->Fore(Rgb24(0, 0, 255));
			BlueUnderline->Create();
		}

		OnFontChange();
	}
}

GTextView::~GTextView()
{
	DeleteArray(Text);
	DeleteArray(MappedChars);
	DeleteArray(MappedShorts);
}

void GTextView::OnFontChange()
{
	if (Font)
	{
		// get line height
		int OldLineY = LineY;
		Font->Size(0, &LineY, "A", 1);

		// get tab size
		char Spaces[TAB_SIZE+1];
		memset(Spaces, 'A', TAB_SIZE);
		Spaces[TAB_SIZE] = 0;
		Font->TabSize(Font->X(Spaces));

		// repour doc if line height has changed
		if (OldLineY > 0 AND
			LineY != OldLineY)
		{
			PourText();
		}

		// validate blue underline font
		if (BlueUnderline)
		{
			BlueUnderline->X("a");
		}
	}
}

char *GTextView::MapChars(char *Source, int Len)
{
	short *Table = 0;
	
	if (Source AND
		CodePage != GDCF_ASCII AND
		GdcGetCodePageMappingTable(CodePage, Table))
	{
		if (Len < 0)
		{
			Len = strlen(Source);
		}

		if (NOT MappedChars OR MappedLen < Len)
		{
			DeleteArray(MappedChars);
			DeleteArray(MappedShorts);
			MappedChars = NEW(char[Len]);
			MappedLen = Len;
		}

		if (MappedChars)
		{
			for (int i=0; i<Len; i++)
			{
				if ((uchar)Source[i] == 0xa1)
				{
					int csd=0;
				}
				MappedChars[i] = (Source[i] & 0x80) ? Table[(uchar)Source[i]-0x80] : Source[i];
			}

			return MappedChars;
		}
	}

	return Source; // no mapping available... go with the source
}

short *GTextView::MapShorts(char *Source, int Len)
{
	short *Table = 0;
	
	if (Source)
	{
		if (Len < 0)
		{
			Len = strlen(Source);
		}

		if (NOT MappedShorts OR MappedLen < Len)
		{
			DeleteArray(MappedChars);
			DeleteArray(MappedShorts);
			MappedShorts = NEW(short[Len]);
			MappedLen = Len;
		}

		if (MappedShorts)
		{
			if (Table)
			{
				// do mapping
				for (int i=0; i<Len; i++)
				{
					MappedShorts[i] = (Source[i] & 0x80) ? Table[(uchar)Source[i]-0x80] : Source[i];
				}
			}
			else
			{
				// no mapping available, convert chars to shorts.
				for (int i=0; i<Len; i++)
				{
					MappedShorts[i] = Source[i];
				}
			}

			return MappedShorts;
		}
	}

	LgiAssert(0);
	return 0; // allocation or argument error
}

void GTextView::PourText()
{
	#ifdef _DEBUG
	int StartTime = LgiCurrentTime();
	#endif

	GRegion Client = GetClient();
	int Mx = Client.X() - LeftMargin;
	int y = 0;
	int Start = 0;

	// empty list
	for (GTextLine *l=Line.First(); l; l=Line.First())
	{
		Line.Delete(l);
		DeleteObj(l);
	}

	if (Text AND Font)
	{
		// break array, break out of loop when we hit these chars
		bool Break[1 << (sizeof(gchar)*8)];
		ZeroObj(Break);
		Break[0] = true;
		Break[' '] = true;
		Break[9] = true;
		Break['\n'] = true;

		// tracking vars
		int e;
		int Cx = 0, Cy = 0;
		int LastX = 0;
		int LastChar = 0;
		int Fy = Font->Y("A");
		bool Wrap = false;

		// char width stuff
		float *Widths = Font->GetWidths();
		short *MappingTable = 0;
		bool OwnWidths = false;

		/*
		// map glyph widths if we have a codepage mapping
		if (Widths AND
			CodePage != GDCF_ASCII AND
			GdcGetCodePageMappingTable(CodePage, MappingTable))
		{
			float *MappedWidths = NEW(float[256]);
			if (MappedWidths)
			{
				int i;

				// first 128 chars are as is
				for (i=0; i<128; i++)
				{
					MappedWidths[i] = Widths[i];
				}

				// map the widths for the extended charset
				for (; i<256; i++)
				{
					MappedWidths[i] = Widths[ MappingTable[i-128] & 0xff ];
				}

				Widths = MappedWidths;
				OwnWidths = true;
			}
		}
		*/

		// no widths array? we're outta here
		if (NOT Widths) return;

		// alright... lets pour!
		for (int i=Start; i<=Size; i = e)
		{
			// seek till next char of interest
			for (e=i; NOT Break[(gchar)Text[e]]; e++)
			{
				Cx += (int) Widths[(gchar)Text[e]];
			}

			if (Text[e] == 9)
			{
				// tab processing
				Cx = ((Cx + Font->TabSize()) / Font->TabSize()) * Font->TabSize();
				e++;
				continue;
			}
			else if (Text[e] == ' ')
			{
				// space processing
				Cx += (int) Widths[Text[e]];
				e++;

				if (WrapAtCol > 0)
				{
					// wrap at the column specified
					if (i - LastChar < WrapAtCol)
					{
						LastX = Cx;
						continue;
					}

					Wrap = true;
				}
				else
				{
					// wrap at the end of the page
					if (Cx < Mx)
					{
						LastX = Cx;
						continue;
					}

					Wrap = true;
				}
			}
			else if (Text[e] == '\n' OR Text[e] == 0)
			{
				if (WrapAtCol > 0)
				{
					// wrap at the column specified
					Wrap = (i - LastChar >= WrapAtCol);
				}
				else
				{
					// wrap at the end of the page
					Wrap = (Cx >= Mx);
				}
			}

			GTextLine *l = NEW(GTextLine);
			if (l)
			{
				l->Start = LastChar;
				l->r.x1 = LeftMargin;

				if (Wrap)
				{
					if (i > LastChar)
					{
						// wrap at last space
						l->Len = i - LastChar - 1;
						l->r.x2 = l->r.x1 + LastX - Widths[' '];
						LastChar = e = i;
					}
					else
					{
						// no space on this line,
						// wrap at char that is at the edge of page
						l->Len = Font->CharAt(Mx, Text + LastChar, e - LastChar);
						l->r.x2 = l->r.x1 + Mx;
						LastChar = ++e;
					}

					Wrap = false;
				}
				else
				{
					// end of line
					l->Len = e - LastChar;
					l->r.x2 = l->r.x1 + Cx;
					LastChar = ++e;
				}

				l->r.y1 = Cy;
				l->r.y2 = l->r.y1 + Fy - 1;

				Line.Insert(l);

				LastX = Cx = 0;
				Cy += Fy;
			}
		}

		// Update lines var
		Lines = Line.GetItems();

		if (OwnWidths)
		{
			DeleteArray(Widths);
		}
	}

	#ifdef _DEBUG
	_PourTime = LgiCurrentTime() - StartTime;
	#endif

	GRegion c = GetClient();
	SetScrollBars(false, c.Y() < Lines * LineY);
	if (VScroll)
	{
		VScroll->SetId(IDC_VS);
	}
}

void GTextView::InsertStyle(GTextStyle *s)
{
	if (s)
	{
		int Last = 0;
		int n = 0;

		for (GTextStyle *i=Style.First(); i; i=Style.Next(), n++)
		{
			if (s->Start >= Last AND s->Start < i->Start)
			{
				Style.Insert(s, n);
				return;
			}

			Last = i->Start;
		}

		Style.Insert(s);
	}
}

GTextStyle *GTextView::GetNextStyle(int Where)
{
	GTextStyle *s = (Where >= 0) ? Style.First() : Style.Next();
	while (s)
	{
		// determin whether style is relevent..
		// styles in the selected region are ignored
		int Min = min(SelStart, SelEnd);
		int Max = max(SelStart, SelEnd);
		if (SelStart >= 0 AND
			s->Start >= Min AND
			s->Start+s->Len < Max)
		{
			// style is completely inside selection: ignore
			s = Style.Next();
		}
		else if (Where >= 0 AND
			s->Start+s->Len < Where)
		{
			s = Style.Next();
		}
		else
		{
			return s;
		}
	}

	return 0;
}

class GUrl : public GTextStyle
{
public:
	bool Email;

	bool OnMouseClick(GMouse *m)
	{
		if (View)
		{
			if ( (m AND m->Left() AND m->Double()) OR (NOT m) )
			{
				gchar *Url = Strnew(View->GetText() + Start, Len);
				if (Url)
				{
					View->OnUrl(Url);
					DeleteObj(Url);
					return true;
				}
			}
		}

		return false;
	}

	bool OnMenu(GSubMenu *m)
	{
		if (m)
		{
			if (Email)
			{
				m->AppendItem(TvLoadString(GTEXTVIEW_EMAIL_TO), IDM_NEW, true);
			}
			else
			{
				m->AppendItem(TvLoadString(GTEXTVIEW_OPENURL), IDM_OPEN, true);
			}

			m->AppendItem(TvLoadString(GTEXTVIEW_COPYLINK), IDM_COPY_URL, true);

			return true;
		}

		return false;
	}

	void OnMenuClick(int i)
	{
		switch (i)
		{
			case IDM_NEW:
			case IDM_OPEN:
			{
				OnMouseClick(0);
				break;
			}
			case IDM_COPY_URL:
			{
				gchar *Url = Strnew(View->GetText() + Start, Len);
				if (Url)
				{
					GClipBoard Clip(View);
					Clip.TextW(Url);
					DeleteObj(Url);
				}
				break;
			}
		}
	}

	char *GetCursor()
	{
		#ifdef WIN32
		int Ver, Rev;
		if (LgiGetOs(&Ver, &Rev) == LGI_OS_WINNT AND
			Ver >= 5)
		{
			return MAKEINTRESOURCE(32649); // hand
		}
		else
		{
			return IDC_ARROW;
		}
		#endif
		return 0;
	}
};

GTextStyle *GTextView::HitStyle(int i)
{
	for (GTextStyle *s=Style.First(); s; s=Style.Next())
	{
		if (i >= s->Start AND i < s->Start+s->Len)
		{
			return s;
		}
	}

	return 0;	
}

void GTextView::PourStyle()
{
	#ifdef _DEBUG
	int StartTime = LgiCurrentTime();
	#endif

	// empty list
	for (GTextStyle *s=Style.First(); s; s=Style.First())
	{
		Style.Delete(s);
		DeleteObj(s);
	}

	if (UrlDetect)
	{
		char Map[1 << (sizeof(gchar) * 8)];

		ZeroObj(Map);
		Map['h'] = true;
		Map['@'] = true;

		for (int i=0; i<Size; i++)
		{
			if (Map[Text[i]])
			{
				switch (Text[i])
				{
					case 'h':
					{
						if (Strnicmp(Text+i, G("http://"), 6) == 0)
						{
							// find end
							gchar *s = Text + i;
							gchar *e = s + 6;
							for ( ; (((int)e-(int)Text) < Size) AND 
									UrlChar(*e); e++);

							GUrl *Url = NEW(GUrl);
							if (Url)
							{
								Url->Email = false;
								Url->View = this;
								Url->Start = (int)s - (int)Text;
								Url->Len = (int)e - (int)s;
								Url->Font = BlueUnderline;
								Url->c = Rgb24(0, 0, 255);

								InsertStyle(Url);
							}
							i = (int)e-(int)Text;
						}
						break;
					}
					case '@':
					{
						// find start
						gchar *s = Text + (max(i, 1) - 1);
						for ( ; (s>=Text) AND 
								EmailChar(*s); s--);
						if (s < Text + i)
						{
							if (s>=Text) s++;
							bool FoundDot = false;
							gchar *e = Text + i + 1;
							for ( ; (((int)e-(int)Text) < Size) AND 
									UrlChar(*e); e++)
							{
								if (*e == '.') FoundDot = true;
							}

							if (FoundDot)
							{
								GUrl *Url = NEW(GUrl);
								if (Url)
								{
									Url->Email = true;
									Url->View = this;
									Url->Start = (int)s - (int)Text;
									Url->Len = (int)e - (int)s;
									Url->Font = BlueUnderline;
									Url->c = Rgb24(0, 0, 255);

									InsertStyle(Url);
								}
								i = (int)e-(int)Text;
							}
						}
						break;
					}
				}
			}
		}
	}

	#ifdef _DEBUG
	_StyleTime = LgiCurrentTime() - StartTime;
	#endif
}

bool GTextView::Insert(int At, gchar *Data, int Len)
{
	if (AcceptEdit)
	{
		// limit input to valid data
		At = min(Size, At);

		// make sure we have enough memory
		int NewAlloc = Size + Len + 1;
		NewAlloc += ALLOC_BLOCK - (NewAlloc % ALLOC_BLOCK);
		if (NewAlloc != Alloc)
		{
			gchar *NewText = NEW(gchar[NewAlloc]);
			if (NewText)
			{
				if (Text)
				{
					// copy any existing data across
					memcpy(NewText, Text, Size);
				}

				DeleteArray(Text);
				Text = NewText;
				Alloc = NewAlloc;
			}
			else
			{
				// memory allocation error
				return false;
			}
		}

		if (Text)
		{
			// insert the data
			// move the section after the insert
			memmove(Text+(At+Len), Text+At, Size-At);

			if (Data)
			{
				// copy new data in
				memcpy(Text+At, Data, Len);
				Size += Len;
			}
			else
			{
				return false;
			}

			Text[Size] = 0;
			Dirty = true;
			PourText();
			PourStyle();

			GView *n = GetNotify()?GetNotify():GetParent();
			if (n)
			{
				n->OnNotify(this, GTVN_DOC_CHANGED);
			}

			return true;
		}
	}

	return false;
}

bool GTextView::Delete(int At, int Len)
{
	bool Status = false;

	if (AcceptEdit)
	{
		// limit input
		At = max(At, 0);
		At = min(At, Size);
		Len = min(Size-At, Len);

		if (Len > 0)
		{
			// do delete
			memmove(Text+At, Text+(At+Len), Size-At-Len);
			Size -= Len;
			Text[Size] = 0;

			Dirty = true;
			Status = true;
			PourText();
			PourStyle();

			GView *n = GetNotify()?GetNotify():GetParent();
			if (n)
			{
				n->OnNotify(this, GTVN_DOC_CHANGED);
			}

			Status = true;
		}
	}

	return Status;
}

void GTextView::DeleteSelection(gchar **Cut)
{
	if (SelStart >= 0)
	{
		int Min = min(SelStart, SelEnd);
		int Max = max(SelStart, SelEnd);

		if (Cut)
		{
			*Cut = Strnew(Text + Min, Max - Min);
		}

		Delete(Min, Max - Min);
		SetCursor(Min, false, true);
	}
}

GTextLine *GTextView::GetLine(int Offset, int *Index)
{
	int i = 0;
	for (GTextLine *l=Line.First(); l; l=Line.Next(), i++)
	{
		if (Offset >= l->Start AND Offset <= l->Start+l->Len)
		{
			if (Index)
			{
				*Index = i;
			}

			return l;
		}
	}

	return 0;
}

gchar *GTextView::GetText()
{
	return Text;
}

bool GTextView::SetText(gchar *s)
{
	bool Status = false;
	if (s)
	{
		DeleteArray(Text);
		Size = Strlen(s);
		Alloc = Size + 1;
		Text = NEW(gchar[Alloc]);
		Cursor = 0;
		if (Text)
		{
			memcpy(Text, s, Size);

			// remove LF's
			int In = 0, Out = 0;
			CrLf = false;
			for (; In<Size; In++)
			{
				if (Text[In] != '\r')
				{
					Text[Out++] = Text[In];
				}
				else
				{
					CrLf = true;
				}
			}
			Size = Out;
			Text[Size] = 0;

			// update everything else
			PourText();
			PourStyle();
			UpdateScrollBars(true);

			Status = true;
		}

		Invalidate();
	}

	return Status;
}

bool GTextView::HasSelection()
{
	return (SelStart >= 0) AND (SelStart != SelEnd);
}

void GTextView::SelectAll()
{
	SelStart = 0;
	SelEnd = Size;
	Invalidate();
}

void GTextView::UnSelectAll()
{
	bool Update = HasSelection();

	SelStart = -1;
	SelEnd = -1;

	if (Update)
	{
		Invalidate();
	}
}

int GTextView::GetLines()
{
	return Lines;
}

GdcPt2 GTextView::GetCursor()
{
	GdcPt2 p(0, 0);
	int FromIndex = 0;
	GTextLine *From = GetLine(Cursor, &FromIndex);
	if (From)
	{
		p.x = Cursor - From->Start;
		p.y = FromIndex;
	}

	return p;
}

void GTextView::SetCursor(int i, bool Select, bool ForceFullUpdate)
{
	if (i >= 0 AND i<=Size)
	{
		int s = SelStart, e = SelEnd, c = Cursor;
		
		if (Select AND i != SelStart)
		{
			if (SelStart < 0)
			{
				SelStart = Cursor;
			}

			SelEnd = i;
		}
		else
		{
			SelStart = SelEnd = -1;
		}

		int FromIndex = 0;
		GTextLine *From = GetLine(Cursor, &FromIndex);

		Cursor = i;

		// check the cursor is on the screen
		int ToIndex = 0;
		GTextLine *To = GetLine(Cursor, &ToIndex);
		if (VScroll AND To)
		{
			if (ToIndex < VScroll->Pos())
			{
				// above the visible region...
				VScroll->SetPos(ToIndex);
				ForceFullUpdate = true;
			}

			int DisplayLines = GetClient().Y() / LineY;
			if (ToIndex >= VScroll->Pos() + DisplayLines)
			{
				VScroll->SetPos(ToIndex - DisplayLines + 1);
				ForceFullUpdate = true;
			}
		}

		// check whether we need to update the screen
		if (ForceFullUpdate OR
			SelStart != s OR
			SelEnd != e OR
			NOT To OR
			NOT From)
		{
			// need full update
			Invalidate();
		}
		else if (Cursor != c)
		{
			// just the cursor has moved

			// update the line the cursor moved to
			GRegion r = To->r;
			r.Offset(0, -DocOffset);
			r.x2 = X();
			Invalidate(&r);

			if (To != From)
			{
				// update the line the cursor came from,
				// if it's a different line from the "to"
				r = From->r;
				r.Offset(0, -DocOffset);
				r.x2 = X();
				Invalidate(&r);
			}
		}

		if (c != Cursor)
		{
			GView *n = GetNotify()?GetNotify():GetParent();
			if (n)
			{
				n->OnNotify(this, GTVN_CURSOR_CHANGED);
			}
		}
	}
}

void GTextView::SetBorder(int b)
{
}

gchar *ConvertToCrLf(gchar *Text)
{
	if (Text)
	{
		// add '\r's
		int Lfs = 0;
		int Len = 0;
		gchar *s=Text;
		for (; *s; s++)
		{
			if (*s == '\n') Lfs++;
			Len++;
		}

		gchar *Temp = NEW(gchar[Len+Lfs+1]);
		if (Temp)
		{
			gchar *d=Temp;
			s = Text;
			for (; *s; s++)
			{
				if (*s == '\n')
				{
					*d++ = 0x0d;
					*d++ = 0x0a;
				}
				else if (*s == '\r')
				{
					// ignore
				}
				else
				{
					*d++ = *s;
				}
			}
			*d++ = 0;

			DeleteObj(Text);
			return Temp;
		}
	}

	return Text;
}

bool GTextView::Cut()
{
	bool Status = false;
	gchar *t = 0;
	
	DeleteSelection(&t);
	GClipBoard Clip(this);
	if (t)
	{
		t = ConvertToCrLf(t);
		Status = Clip.TextW(t);
		DeleteArray(t);
	}

	return Status;
}

bool GTextView::Copy()
{
	bool Status = true;

	if (SelStart >= 0)
	{
		int Min = min(SelStart, SelEnd);
		int Max = max(SelStart, SelEnd);

		gchar *t = Strnew(Text+Min, Max-Min);
		t = ConvertToCrLf(t);

		GClipBoard Clip(this);
		Status = Clip.TextW(t);
		DeleteArray(t);
	}

	return Status;
}

bool GTextView::Paste()
{
	GClipBoard Clip(this);
	gchar *t = Clip.TextW();
	if (t)
	{
		if (SelStart >= 0)
		{
			DeleteSelection();
		}

		// remove '\r's
		bool Multiline = false;
		gchar *s = t, *d = t;
		for (; *s; s++)
		{
			if (*s == '\n')
			{
				Multiline = true;
			}

			if (*s != '\r')
			{
				*d++ = *s;
			}
		}
		*d++ = 0;

		// insert text
		int Len = Strlen(t);
		Insert(Cursor, t, Len);
		SetCursor(Cursor+Len, false, true); // Multiline

		// clean up
		DeleteArray(t);
	}
	return false;
}

bool GTextView::ClearDirty(bool Ask, char *FileName)
{
	if (Dirty)
	{
		int Answer = (Ask) ? LgiMsg(this, "Do you want to save your changes to this document?", "Save", MB_YESNOCANCEL) : IDYES;
		if (Answer == IDYES)
		{
			GFileSelect Select;
			Select.Parent(this);
			if (NOT FileName AND
				Select.Save())
			{
				FileName = Select.Name();
			}

			Save(FileName);
		}
		else if (Answer == IDCANCEL)
		{
			return false;
		}
	}

	return true;
}

bool GTextView::Open(char *Name)
{
	bool Status = false;
	GFile f;

	if (f.Open(Name, O_READ))
	{
		DeleteArray(Text);
		Size = f.GetSize();
		Alloc = Size + 1;
		Text = NEW(gchar[Alloc]);
		SetCursor(0, false);
		if (Text)
		{
			if (f.Read(Text, Size) == Size)
			{
				// remove LF's
				int In = 0, Out = 0;
				CrLf = false;
				for (; In<Size; In++)
				{
					if (Text[In] != '\r')
					{
						Text[Out++] = Text[In];
					}
					else
					{
						CrLf = true;
					}
				}

				Size = Out;
				Text[Size] = 0;
				Dirty = false;

				PourText();
				PourStyle();
				UpdateScrollBars(true);

				Status = true;
			}
		}

		Invalidate();
	}

	return Status;
}

bool GTextView::Save(char *Name)
{
	GFile f;
	if (f.Open(Name, O_WRITE))
	{
		f.SetSize(0);
		if (Text)
		{
			bool Status = false;
			
			if (CrLf)
			{
				int Start = 0;
				Status = true;
				for (int i=0; i<=Size AND Status; i++)
				{
					if (Text[i] == '\n' OR i >= Size)
					{
						Status = f.Write(Text + Start, i - Start) == (i - Start);
						if (i < Size)
						{
							Status = f.Write((char*)"\r\n", 2) == 2;
						}
						Start = i+1;
					}
				}
			}
			else
			{
				Status = f.Write(Text, Size) == Size;
			}

			Dirty = false;
			return Status;
		}
	}
	return false;
}

void GTextView::UpdateScrollBars(bool Reset)
{
	if (VScroll)
	{
		GRegion Before = GetClient();

		int DisplayLines = Y() / LineY;
		VScroll->SetLimits(0, Lines);
		VScroll->SetPage(DisplayLines);

		if (Reset)
		{
			VScroll->SetPos(0);
			SelStart = SelEnd = -1;
		}

		GRegion After = GetClient();

		if (Before != After)
		{
			PourText();
			Invalidate();
		}
	}
}

bool GTextView::DoFind()
{
	return false;
}

void GTextView::SelectWord(int From)
{
	char Delim[] = " \t\n.,()[]<>=?/\\{}\"\';:+=-|!@#$%^&*";

	for (SelStart = From; SelStart >= 0; SelStart--)
	{
		if (strchr(Delim, Text[SelStart]))
		{
			SelStart++;
			break;
		}
	}

	for (SelEnd = From; SelEnd < Size; SelEnd++)
	{
		if (strchr(Delim, Text[SelEnd]))
		{
			break;
		}
	}

	Invalidate();
}

bool GTextView::OnFind()
{
	int Len = strlen(Find);
	if (Len > 0)
	{
		bool Found = false;
		int i=Cursor;

		if (CaseSensitive())
		{
			for (; i<Size-Len; i++)
			{
				if (Text[i] == Find[0])
				{
					gchar Buf[256];
					Str2native(Buf, Find);
					Found = Strncmp(Text+i, Buf, Len) == 0;
					if (Found)
					{
						break;
					}
				}
			}
		}
		else
		{
			for (; i<Size-Len; i++)
			{
				if (toupper(Text[i]) == toupper(Find[0]))
				{
					gchar Buf[256];
					Str2native(Buf, Find);
					Found = Strnicmp(Text+i, Buf, Len) == 0;
					if (Found)
					{
						break;
					}
				}
			}
		}

		if (Found)
		{
			SetCursor(i, false);
			SetCursor(i + Len, true);
			return true;
		}
	}


	return false;
}

bool GTextView::OnReplace()
{
	return false;
}

int GTextView::SeekLine(int i, GTextViewSeek Where)
{
	switch (Where)
	{
		case PrevLine:
		{
			for (; i > 0 AND Text[i] != '\n'; i--);
			if (i > 0) i--;
			for (; i > 0 AND Text[i] != '\n'; i--);
			if (i > 0) i++;
			break;
		}
		case NextLine:
		{
			for (; i < Size AND Text[i] != '\n'; i++);
			i++;
			break;
		}
		case StartLine:
		{
			for (; i > 0 AND Text[i] != '\n'; i--);
			if (i > 0) i++;
			break;
		}
 		case EndLine:
		{
			for (; i < Size AND Text[i] != '\n'; i++);
			break;
		}
		default:
		{
			ASSERT(false);
			break;
		}
	}

	return i;
}

bool GTextView::OnMultiLineTab(bool In)
{
	bool Status = false;
	int Min = min(SelStart, SelEnd);
	int Max = max(SelStart, SelEnd);

	Min = SeekLine(Min, StartLine);

	BytePipe p;
	int Ls = 0, i;
	for (i=Min; i<Max AND i<Size; i=SeekLine(i, NextLine))
	{
		p.Push((uchar*)&i, sizeof(i));
		Ls++;
	}
	if (Max < i)
	{
		Max = SeekLine(Max, EndLine);
	}

	int *Indexes = NEW(int[Ls]);
	if (Indexes)
	{
		p.Pop((uchar*)Indexes, Ls*sizeof(int));

		SelStart = Min;

		for (i=Ls-1; i>=0; i--)
		{
			if (In)
			{
				// <-
				int n = Indexes[i], Space = 0;
				for (; Space<TAB_SIZE AND n<Size; n++)
				{
					if (Text[n] == 9)
					{
						Space += TAB_SIZE;
					}
					else if (Text[n] == ' ')
					{
						Space += 1;
					}
					else
					{
						break;
					}
				}

				int Chs = n-Indexes[i];
				Delete(Indexes[i], Chs);
				Max -= Chs;
			}
			else
			{
				// ->
				int Len = Indexes[i];
				for (; Text[Len] != '\n' AND Len<Size; Len++);
				if (Len > Indexes[i])
				{
					Insert(Indexes[i], G("\t"), 1);
					Max++;
				}
			}
		}

		SelEnd = Cursor = Max;

		DeleteArray(Indexes);
	}

	PourText();
	Invalidate();
	Status = true;

	return Status;
}

void GTextView::OnSetHidden(int Hidden)
{
}

void GTextView::OnPosChange()
{
	PourText();
	UpdateScrollBars(false);
}

void GTextView::OnCreate()
{
}

void GTextView::OnEscape(GKey &K)
{
}

void GTextView::OnMouseWheel(double l)
{
	if (VScroll)
	{
		int NewPos = VScroll->Pos() + (int) l;
		NewPos = limit(NewPos, 0, Lines);
		VScroll->SetPos(NewPos);
		Invalidate();
	}
}

void GTextView::OnFocus(bool f)
{
	Invalidate();
}

int GTextView::HitText(int x, int y)
{
	int Y = (VScroll) ? VScroll->Pos() : 0;
	GTextLine *l = Line.ItemAt(Y);
	y += (l) ? l->r.y1 : 0;

	for (; l; l=Line.Next())
	{
		if (l->r.Overlap(x, y))
		{
			// Click in a line
			return l->Start + Font->CharAt(x-l->r.x1, Text+l->Start, l->Len);
		}
		else if (y >= l->r.y1 AND y <= l->r.y2)
		{
			// Click horizontally before of after line
			if (x < l->r.x1)
			{
				return l->Start;
			}
			else if (x > l->r.x2)
			{
				return l->Start + l->Len;
			}
		}
	}

	// outside text area
	l = Line.Last();
	if (l)
	{
		if (y > l->r.y2)
		{
			// end of document
			return Size;
		}
	}

	return -1;
}

void GTextView::OnMouseClick(GMouse &m)
{
	if (m.Down())
	{
		if (m.Left())
		{
			bool Processed = false;

			Focus(true);

			int Hit = HitText(m.x, m.y);
			if (Hit >= 0)
			{
				SetCursor(Hit, m.Shift());

				GTextStyle *s = HitStyle(Hit);
				if (s)
				{
					Processed = s->OnMouseClick(&m);
				}
			}

			if (NOT Processed AND m.Double())
			{
				SelectWord(Cursor);
			}
		}
		else if (m.Right())
		{
			GSubMenu *RClick = NEW(GSubMenu);
			if (RClick)
			{
				char *ClipText = 0;
				{
					GClipBoard Clip(this);
					ClipText = Clip.Text();
				}

				GTextStyle *s = HitStyle(HitText(m.x, m.y));
				if (s)
				{
					if (s->OnMenu(RClick))
					{
						RClick->AppendSeparator();
					}
				}

				RClick->AppendItem(TvLoadString(GTEXTVIEW_CUT), IDM_CUT, HasSelection());
				RClick->AppendItem(TvLoadString(GTEXTVIEW_COPY), IDM_COPY, HasSelection());
				RClick->AppendItem(TvLoadString(GTEXTVIEW_PASTE), IDM_PASTE, ClipText != 0);
				RClick->AppendSeparator();

				GMenuItem *Indent = RClick->AppendItem(TvLoadString(GTEXTVIEW_AUTO_INDENT), IDM_AUTO_INDENT, true);
				if (Indent)
				{
					Indent->Checked(AutoIndent);
				}				

				GSubMenu *CodePageMenu = RClick->AppendSub(TvLoadString(GTEXTVIEW_CODEPAGE));
				if (CodePageMenu)
				{
					GMenuItem *Item[GDCF_MAX];
					for (int c=GDCF_ASCII; c<GDCF_MAX; c++)
					{
						char *n = GdcGetCodePageDescription(c);
						Item[c] = CodePageMenu->AppendItem(n?n:(char*)"<NoDescription>", CODEPAGE_BASE+c, GdcIsCodePageImplemented(c));
					}

					if (Item[CodePage])
					{
						Item[CodePage]->Checked(true);
					}
				}

				if (GetMouse(m, true))
				{
					int Id = 0;
					switch (Id = RClick->Float(this, m.x, m.y))
					{
						case IDM_CUT:
						{
							Cut();
							break;
						}
						case IDM_COPY:
						{
							Copy();
							break;
						}
						case IDM_PASTE:
						{
							Paste();
							break;
						}
						case IDM_AUTO_INDENT:
						{
							AutoIndent = NOT AutoIndent;
							break;
						}
						default:
						{
							if (s)
							{
								s->OnMenuClick(Id);
							}

							if (Id >= CODEPAGE_BASE AND
								Id < CODEPAGE_BASE + GDCF_MAX)
							{
								SetCodePage(Id - CODEPAGE_BASE);
							}
							break;
						}
					}
				}

				DeleteObj(RClick);
			}

			return;
		}
	}

	Track = m.Down();
	Capture(m.Down());
}

void GTextView::SetCodePage(int i)
{
	CodePage = i;

	if (Font)
	{
		Font->CodePage(CodePage);
	}
	if (BlueUnderline)
	{
		BlueUnderline->CodePage(CodePage);
	}

	GView *n = GetNotify()?GetNotify():GetParent();
	if (n)
	{
		n->OnNotify(this, GTVN_CODEPAGE_CHANGED);
	}

	OnFontChange();
	Invalidate();
}

int GTextView::OnHitTest(int x, int y)
{
	#ifdef WIN32
	if (GetClient().Overlap(x, y))
	{
		return HTCLIENT;
	}
	#endif
	return GView::OnHitTest(x, y);
}

void GTextView::OnMouseMove(GMouse &m)
{
	int Hit = HitText(m.x, m.y);

	if (Track)
	{
		SetCursor(Hit, m.Down());
	}

	#ifdef WIN32
	GRegion c = GetClient();
	c.Offset(-c.x1, -c.y1);
	if (c.Overlap(m.x, m.y))
	{
		GTextStyle *s = HitStyle(Hit);
		char *c = (s) ? s->GetCursor() : 0;
		if (NOT c) c = IDC_IBEAM;
		::SetCursor(LoadCursor(0, MAKEINTRESOURCE(c)));
	}
	#endif
}

void GTextView::OnKey(GKey &k)
{
	if (k.Down)
	{
		if (k.IsChar)
		{
			switch (k.c)
			{
				default:
				{
					// process single char input
					if (k.c >= ' ' OR
						k.c == 9)
					{
						// letter/number etc
						if (SelStart >= 0)
						{
							if (k.c == 9)
							{
								if (OnMultiLineTab(k.Shift()))
								{
									break;
								}
							}
							else
							{
								DeleteSelection();
							}
						}
						
						GTextLine *l = GetLine(Cursor);
						int Len = (l) ? l->Len : 0;

						gchar kc = k.c;

						if (Insert(Cursor, &kc, 1))
						{
							l = GetLine(Cursor);
							int NewLen = (l) ? l->Len : 0;
							SetCursor(Cursor + 1, false, Len != NewLen - 1);
						}
					}
					break;
				}
				case VK_RETURN:
				{
					// enter
					if (SelStart >= 0)
					{
						DeleteSelection();
					}

					gchar InsertStr[256] = G("\n");

					GTextLine *CurLine = GetLine(Cursor);
					if (CurLine AND AutoIndent)
					{
						int WsLen = 0;
						for (;	WsLen < CurLine->Len AND
								WsLen < (Cursor - CurLine->Start) AND
								strchr(" \t", Text[CurLine->Start + WsLen]); WsLen++);
						if (WsLen > 0)
						{
							memcpy(InsertStr+1, Text+CurLine->Start, WsLen);
							InsertStr[WsLen+1] = 0;
						}
					}
					
					if (Insert(Cursor, InsertStr, Strlen(InsertStr)))
					{
						SetCursor(Cursor + Strlen(InsertStr), false, true);
					}

					OnEnter(k);
					break;
				}
				case VK_BACK:
				{
					if (SelStart >= 0)
					{
						// delete selection
						DeleteSelection();
					}
					else
					{
						char Del = Text[Cursor-1];

						if (Cursor > 0 AND
							Delete(Cursor - 1, 1))
						{
							// normal backspace
							SetCursor(Cursor - 1, false, Del == '\n');
						}
					}
					break;
				}
			}
		}
		else // not a char
		{
			switch (k.c)
			{
				case VK_LEFT:
				{
					if (SelStart >= 0 AND
						NOT k.Shift())
					{
						SetCursor(min(SelStart, SelEnd), false);
					}
					else if (Cursor > 0)
					{
						int n = Cursor;

						if (k.Ctrl())
						{
							// word move/select
							bool StartWhiteSpace = IsWhiteSpace(Text[n]);
							bool LeftWhiteSpace = n > 0 AND IsWhiteSpace(Text[n-1]);

							if (NOT StartWhiteSpace OR
								Text[n] == '\n') n--;
							for (; n > 0 AND strchr(" \t", Text[n]); n--);
							if (Text[n] == '\n')
							{
								n--;
							}
							else if (NOT StartWhiteSpace OR NOT LeftWhiteSpace)
							{
								if (IsDelimiter(Text[n]))
								{
									for (; n > 0 AND IsDelimiter(Text[n]); n--);
								}
								else
								{
									for (; n > 0 AND NOT IsWordBoundry(Text[n]); n--);
								}
							}
							if (n > 0) n++;
						}
						else
						{
							// single char
							n--;
						}

						SetCursor(n, k.Shift());
					}
					break;
				}
				case VK_RIGHT:
				{
					if (SelStart >= 0 AND
						NOT k.Shift())
					{
						SetCursor(max(SelStart, SelEnd), false);
					}
					else if (Cursor < Size)
					{
						int n = Cursor;

						if (k.Ctrl())
						{
							// word move/select
							if (IsWhiteSpace(Text[n]))
							{
								for (; n<Size AND IsWhiteSpace(Text[n]); n++);
							}
							else
							{
								if (IsDelimiter(Text[n]))
								{
									n++;
								}
								else
								{
									for (; n<Size AND (IsAlpha(Text[n]) OR IsDigit(Text[n])); n++);
								}

								if (n < Size AND
									Text[n] != '\n')
								{
									if (IsWhiteSpace(Text[n]))
									{
										n++;
									}
								}
							}
						}
						else
						{
							// single char
							n++;
						}

						SetCursor(n, k.Shift());
					}
					break;
				}
				case VK_UP:
				{
					int Old = Cursor;
					GTextLine *l = GetLine(Cursor);
					if (l)
					{
						GTextLine *Prev = Line.Prev();
						if (Prev)
						{
							SetCursor(Prev->Start + min(Old-l->Start, Prev->Len), k.Shift());
						}
					}
					break;
				}
				case VK_DOWN:
				{
					int Old = Cursor;
					GTextLine *l = GetLine(Cursor);
					if (l)
					{
						GTextLine *Next = Line.Next();
						if (Next)
						{
							SetCursor(Next->Start + min(Old-l->Start, Next->Len), k.Shift());
						}
					}
					break;
				}
				case VK_END:
				{
					if (k.Ctrl())
					{
						SetCursor(Size, k.Shift());
					}
					else
					{
						GTextLine *l = GetLine(Cursor);
						if (l)
						{
							SetCursor(l->Start + l->Len, k.Shift());
						}
					}
					break;
				}
				case VK_HOME:
				{
					if (k.Ctrl())
					{
						SetCursor(0, k.Shift());
					}
					else
					{
						GTextLine *l = GetLine(Cursor);
						if (l)
						{
							gchar *Line = Text + l->Start;
							gchar *s;
							for (s = Line; ((int)s-(int)Line) < l->Len AND strchr(" \t", *s); s++);
							int Whitespace = (int)s-(int)Line;

							if (l->Start + Whitespace == Cursor)
							{
								SetCursor(l->Start, k.Shift());
							}
							else
							{
								SetCursor(l->Start + Whitespace, k.Shift());
							}
						}
					}
					break;
				}
				case VK_PRIOR:
				{
					GTextLine *l = GetLine(Cursor);
					if (l)
					{
						int DisplayLines = Y() / LineY;
						int CurLine = Line.IndexOf(l);

						GTextLine *New = Line.ItemAt(max(CurLine - DisplayLines, 0));
						if (New)
						{
							SetCursor(New->Start + min(Cursor - l->Start, New->Len), k.Shift());
						}
					}				
					break;
				}
				case VK_NEXT:
				{
					GTextLine *l = GetLine(Cursor);
					if (l)
					{
						int DisplayLines = Y() / LineY;
						int CurLine = Line.IndexOf(l);

						GTextLine *New = Line.ItemAt(min(CurLine + DisplayLines, Lines-1));
						if (New)
						{
							SetCursor(New->Start + min(Cursor - l->Start, New->Len), k.Shift());
						}
					}				
					break;
				}
				case VK_INSERT:
				{
					if (k.Ctrl())
					{
						Copy();
					}
					else if (k.Shift())
					{
						Paste();
					}
					break;
				}
				case VK_DELETE:
				{
					if (SelStart >= 0)
					{
						if (k.Shift())
						{
							Cut();
						}
						else
						{
							DeleteSelection();
						}
					}
					else if (Cursor < Size AND
							Delete(Cursor, 1))
					{
						Invalidate();
					}
					break;
				}
				default:
				{
					if (k.Ctrl())
					{
						switch (k.c)
						{
							case 'a':
							case 'A':
							{
								// select all
								SelStart = 0;
								SelEnd = Size;
								Invalidate();
								break;
							}
							case 'x':
							case 'X':
							{
								Cut();
								break;
							}
							case 'c':
							case 'C':
							{
								Copy();
								break;
							}
							case 'v':
							case 'V':
							{
								Paste();
								break;
							}
							case 'f':
							case 'F':
							{
								OpenFind();
								break;
							}
							case 'g':
							case 'G':
							{
								GInput Dlg(this, "", "Goto line:", "Text");
								if (Dlg.DoModal() == IDOK AND
									Dlg.Str)
								{
									GTextLine *l = Line.ItemAt(atoi(Dlg.Str) - 1);
									if (l)
									{
										SetCursor(l->Start, false);
									}
								}
								break;
							}
							case 'h':
							case 'H':
							{
								OpenReplace();
								break;
							}
							case VK_RETURN:
							{
								OnEnter(k);
								break;
							}
						}
					}
					break;
				}
			}
		}
	}
}

int GTextView::TextWidth(GFont *f, gchar *s, int Len, int x, int Origin)
{
	int Status = 0;

	if (f)
	{
		int cx = x;
		float *Widths = f->GetWidths();
		int _TabSize = f->TabSize();

		if (Widths)
		{
			for (; *s AND Len > 0; s++, Len--)
			{
				if (*s == '\t')
				{
					// tab processing
					cx = ((cx + _TabSize) / _TabSize) * _TabSize;

				}
				else
				{
					// normal char
					cx += Widths[(gchar)*s];
				}
			}

			Status = cx - x;
		}
	}

	return Status;
}

void GTextView::OnPaint(GSurface *pDC)
{
	#ifdef _DEBUG
	int StartTime = LgiCurrentTime();
	#endif

	GRegion r(0, 0, X()-1, Y()-1);
	GSurface *pOut = pDC;
	bool DrawSel = false;
	COLOUR Fore = LC_TEXT;
	COLOUR Back = (AcceptEdit) ? LC_WORKSPACE  : ReadOnlyBackColour;
	COLOUR Selected = LC_SELECTION;
	if (NOT Focus())
	{
		COLOUR Work = LC_WORKSPACE;
		Selected = Rgb24(	(R24(Work)+R24(Selected))/2,
							(G24(Work)+G24(Selected))/2,
							(B24(Work)+B24(Selected))/2);
	}

	#ifdef DOUBLE_BUFFER_PAINT
	GMemDC *pMem = NEW(GMemDC);
	pOut = pMem;
	#endif
	if (Text AND
		Font
		#ifdef DOUBLE_BUFFER_PAINT
		AND
		pMem AND
		pMem->Create(r.X()-LeftMargin, LineY, GdcD->GetBits())
		#endif
		)
	{
		int y = 0;
		int x = LeftMargin;
		int SelMin = min(SelStart, SelEnd);
		int SelMax = max(SelStart, SelEnd);

		// font properties
		Font->Colour(Fore, Back);
		Font->Transparent(false);

		// draw left margin
		pDC->Colour(Back, 24);
		pDC->Rectangle(0, 0, x-1, r.y2);
	
		// draw lines of text
		int k = (VScroll) ? VScroll->Pos() : 0;
		GTextLine *l=Line.ItemAt(k);
		int Dy = (l) ? -l->r.y1 : 0;
		int NextSelection = (SelStart != SelEnd) ? SelMin : -1; // offset where selection next changes
		if (l AND
			SelStart >= 0 AND
			SelStart < l->Start AND
			SelEnd > l->Start)
		{
			// start of visible area is in selection
			// init to selection colour
			DrawSel = true;
			Font->Colour(LC_SEL_TEXT, Selected);
			NextSelection = SelMax;
		}

		GTextStyle *NextStyle = GetNextStyle((l) ? l->Start : 0);

		DocOffset = (l) ? l->r.y1 : 0;

		// loop through all visible lines
		for (; l AND l->r.y1+Dy < r.Y(); l=Line.Next())
		{
			GRegion Tr = l->r;
			#ifdef DOUBLE_BUFFER_PAINT
			Tr.Offset(-Tr.x1, -Tr.y1);
			#else
			Tr.Offset(0, Dy);
			#endif
			int TabOri = Tr.x1;

			// deal with selection change on beginning of line
			if (NextSelection == l->Start)
			{
				// selection change
				DrawSel = NOT DrawSel;
				if (DrawSel)
				{
					Font->Colour(LC_SEL_TEXT, Selected);
				}
				else
				{
					Font->Colour(Fore, Back);
				}
				NextSelection = (NextSelection == SelMin) ? SelMax : -1;
			}

			// draw text
			int Done = 0;	// how many chars on this line have we 
							// processed so far
			int TextX = 0;	// pixels we have moved so far
			
			// loop through all sections of similar text on a line
			while (Done < l->Len)
			{
				// decide how big this block is
				int Cur = l->Start + Done;
				int Block = l->Len - Done;
				
				// check for style change
				if (NextStyle)
				{
					// start
					if (l->Overlap(NextStyle->Start) AND
						NextStyle->Start > Cur AND
						NextStyle->Start - Cur < Block)
					{
						Block = NextStyle->Start - Cur;
					}

					// end
					int StyleEnd = NextStyle->Start + NextStyle->Len;
					if (l->Overlap(StyleEnd) AND
						StyleEnd > Cur AND
						StyleEnd - Cur < Block)
					{
						Block = StyleEnd - Cur;
					}
				}

				// check for next selection change
				// this may truncate the style
				if (l->Overlap(NextSelection) AND
					((NextSelection - Cur < Block) OR (Cur >= SelMin)))
				{
					Block = NextSelection - Cur;
				}

				ASSERT(Block != 0);	// sanity check

				// Do code page mapping
				gchar *_temp = Text+(l->Start+Done); // MapChars(Text+(l->Start+Done), Block);
				LgiAssert(_temp); // code page mapping failed

				if (NextStyle AND							// There is a style
					Cur != SelMin AND						// AND we're not drawing a selection block
					Cur >= NextStyle->Start AND				// AND we're inside the styled area
					Cur < NextStyle->Start+NextStyle->Len)
				{
					if (NextStyle->Font)
					{
						// draw styled text
						NextStyle->Font->Colour(NextStyle->c, Back);
						NextStyle->Font->Transparent(false);

						/* old tab insensitive text size calc
						NextStyle->Font->Size(&TextX, 0, Text+(l->Start+Done), Block);
						*/
						TextX = TextWidth(NextStyle->Font, Text+(l->Start+Done), Block, Tr.x1 - LeftMargin, TabOri);

						NextStyle->Font->Text(	pOut,
												Tr.x1,
												Tr.y1,
												_temp,
												Block,
												&Tr,
												TabOri);
					}
				}
				else
				{
					// draw a block of normal text
					
					/* old tab insensitive text size calc
					Font->Size(&TextX, 0, Text+(l->Start+Done), Block);
					*/

					TextX = TextWidth(Font, Text+(l->Start+Done), Block, Tr.x1 - LeftMargin, TabOri);
					
					Font->Text(	pOut,
								Tr.x1,
								Tr.y1,
								_temp,
								Block,
								0,
								TabOri); // &Tr);
				}

				if (NextStyle AND
					Cur+Block >= NextStyle->Start+NextStyle->Len)
				{
					// end of this styled block
					NextStyle = GetNextStyle();
				}

				if (NextSelection == Cur+Block)
				{
					// selection change
					DrawSel = NOT DrawSel;
					if (DrawSel)
					{
						Font->Colour(LC_SEL_TEXT, Selected);
					}
					else
					{
						Font->Colour(Fore, Back);
					}
					NextSelection = (NextSelection == SelMin) ? SelMax : -1;
				}

				Tr.x1 += TextX;
				Done += Block;
			
			} // end block loop

			// eol processing
			int EndOfLine = l->Start+l->Len;
			if (EndOfLine >= SelMin AND EndOfLine < SelMax)
			{
				// draw the '\n' at the end of the line as selected
				pOut->Colour(Font->Back(), 24);
				pOut->Rectangle(Tr.x2, Tr.y1, Tr.x2+7, Tr.y2);
				Tr.x2 += 7;
			}
			// draw any space after text
			pOut->Colour(Back, 24);
			pOut->Rectangle(Tr.x2, Tr.y1, r.x2, Tr.y2);

			// cursor?
			if (Focus())
			{
				// draw the cursor if on this line
				if (Cursor >= l->Start AND Cursor <= l->Start+l->Len)
				{
					int Cx = 0;
					#ifdef DOUBLE_BUFFER_PAINT
					int Cy = 0;
					#else
					int Cy = y;
					#endif
					// Font->Size(&Cx, 0, Text+l->Start, Cursor-l->Start);
					Cx = TextWidth(Font, Text+l->Start, Cursor-l->Start, 0, TabOri) + LeftMargin;
					
					pOut->Colour((AcceptEdit) ? Fore : Rgb24(192, 192, 192), 24);
					pOut->Rectangle(Cx, Cy, Cx+1, Cy+LineY-1);
				}
			}

			#ifdef DOUBLE_BUFFER_PAINT
			// dump to screen
			pDC->Blt(LeftMargin, y, pOut);
			#endif
			y += LineY;

		} // end of line loop

		// draw any space under the lines
		if (y < r.y2)
		{
			pDC->Colour(Back, 24);
			pDC->Rectangle(x, y, r.x2, r.y2);
		}

		#ifdef DOUBLE_BUFFER_PAINT
		DeleteObj(pMem);
		#endif
	}
	else
	{
		// default drawing: nothing
		pDC->Colour(Back, 24);
		pDC->Rectangle(&r);
	}

	UpdateScrollBars();

	#ifdef _DEBUG
	_PaintTime = LgiCurrentTime() - StartTime;
	if (GetNotify())
	{
		char s[256];
		sprintf(s, "Pour:%i Style:%i Paint:%i ms", _PourTime, _StyleTime, _PaintTime);
		GMessage m = CreateMsg(DEBUG_TIMES_MSG, 0, (int)s);
		GetNotify()->OnEvent(&m);
	}
	#endif
}

int GTextView::OnEvent(GMessage *Msg)
{
	switch (Message(Msg))
	{
		#if defined WIN32
		case WM_ACTIVATE:
		{
			if (GFindReplace::Handle())
			{
				// this makes the find/replace window active
				// when we're alt-tabbing to the text window
				SetActiveWindow(GFindReplace::Handle());
				return 0;
			}
			break;
		}
		#endif
	}

	GFindReplace::OnEvent(Msg);
	return GLayout::OnEvent(Msg);
}

int GTextView::OnNotify(GView *Ctrl, int Flags)
{
	if (Ctrl->GetId() == IDC_VS AND VScroll)
	{
		Invalidate();
	}

	return 0;
}

void GTextView::OnPulse()
{
}

///////////////////////////////////////////////////////////////////////////////
