/*hdr
**      FILE:           GuiDlg.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           8/9/1998
**      DESCRIPTION:    Dialog components
**
**      Copyright (C) 1998 Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdlib.h>
#include <stdio.h>
#include "Lgi.h"

///////////////////////////////////////////////////////////////////////////////////////////
GDialog::GDialog()
#ifdef LGI_RES
	: ResObject(Res_Dialog)
#endif
{
	Name("Dialog");
	IsModal = true;
}

GDialog::~GDialog()
{
}

bool GDialog::LoadFromResource(char *Resource)
{
	char n[256];
	bool Status = GLgiRes::LoadFromResource(Resource, Children, &Pos, n);
	if (Status)
	{
		Name(n);
	}
	return Status;
}

int GDialog::DoModal()
{
	int Status = -1;
	return Status;
}

int GDialog::DoModeless()
{
	int Status = -1;
	return Status;
}

extern GButton *FindDefault(GView *w);

int GDialog::OnEvent(GMessage *Msg)
{
	switch (Msg->Msg)
	{
	}

	return GView::OnEvent(Msg);
}

void GDialog::EndModal(int Code)
{
}

void GDialog::EndModeless(int Code)
{
	Quit(Code);
}

///////////////////////////////////////////////////////////////////////////////////////////
GControl::GControl()
{
	Pos.ZOff(10, 10);
}

GControl::~GControl()
{
}

int GControl::OnEvent(GMessage *Msg)
{
	switch (Msg->Msg)
	{
	}
	return 0;
}

GdcPt2 GControl::SizeOfStr(char *Str)
{
	GdcPt2 Pt(0, 0);
	if (Str)
	{
	}

	return Pt;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Button
GButton::GButton(int id, int x, int y, int cx, int cy, char *name)
#ifdef LGI_RES
	: ResObject(Res_Button)
#endif
{
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
}

GButton::~GButton()
{
}

bool GButton::Default()
{
	return false;
}

void GButton::Default(bool b)
{
}

int GButton::OnEvent(GMessage *Msg)
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Edit
GEdit::GEdit(int id, int x, int y, int cx, int cy, char *name)
#ifdef LGI_RES
	: ResObject(Res_EditBox)
#endif
{
	GdcPt2 Size = SizeOfStr((name)?name:"A");
	if (cx < 0) cx = Size.x + 6;
	if (cy < 0) cy = Size.y + 4;

	MultiLineEdit = false;
	PasswordEdit = false;
	Enter = false;
	SetId(id);
	Sunken(true);
	if (name) Name(name);
	Pos.Set(x, y, x+max(cx, 10), y+max(cy, 10));
}

GEdit::~GEdit()
{
}

void GEdit::MultiLine(bool m)
{
	MultiLineEdit = m;
}

void GEdit::Password(bool m)
{
	PasswordEdit = m;
}

void GEdit::Value(int i)
{
	char Str[32];
	sprintf(Str, "%i", i);
	Name(Str);
}

int GEdit::Value()
{
	char *n = Name();
	return (n) ? atoi(n) : 0;
}

int GEdit::OnEvent(GMessage *Msg)
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Check box
GCheckBox::GCheckBox(int id, int x, int y, int cx, int cy, char *name, int InitState)
#ifdef LGI_RES
	: ResObject(Res_CheckBox)
#endif
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	Val = InitState;
	Three = FALSE;
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
}

GCheckBox::~GCheckBox()
{
}

int GCheckBox::Value()
{
	return Val;
}

void GCheckBox::Value(int b)
{
	Val = b;
}

int GCheckBox::OnEvent(GMessage *Msg)
{
	return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
// Text
GText::GText(int id, int x, int y, int cx, int cy, char *name)
#ifdef LGI_RES
	: ResObject(Res_StaticText)
#endif
{
	GdcPt2 Size = SizeOfStr(name);
	int NewLines = 0;
	for (char *c = name; c AND *c; c++)
	{
		if (*c == '\n') NewLines++;
	}
	if (cx < 0) cx = Size.x;
	if (cy < 0) cy = Size.y * (NewLines + 1);

	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
}

GText::~GText()
{
}

int GText::OnEvent(GMessage *Msg)
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio group
int GRadioGroup::NextId = 20000;

GRadioGroup::GRadioGroup(int id, int x, int y, int cx, int cy, char *name, int Init)
#ifdef LGI_RES
	: ResObject(Res_Group)
#endif
{
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	InitialValue = Init;
}

GRadioGroup::~GRadioGroup()
{
}

bool GRadioGroup::Attach(GView *parent)
{
	bool Status = GControl::Attach(parent);
	if (Status)
	{
		AttachChildren();
	}
	return Status;
}

int GRadioGroup::Value()
{
	int i=0;
	for (	GView *w = Children.First();
			w;
			w = Children.Next(), i++)
	{
		GRadioButton *But = dynamic_cast<GRadioButton*>(w);
		if (But AND But->Value()) return i;
	}

	return -1;
}

void GRadioGroup::Value(int Which)
{
	InitialValue = Which;
}

int GRadioGroup::OnNotify(GView *Ctrl, int Flags)
{
	GView *n = GetNotify() ? GetNotify() : GetParent();
	if (n)
	{
		return n->OnNotify(Ctrl, Flags);
	}
	return 0;
}

int GRadioGroup::OnEvent(GMessage *Msg)
{
	return 0;
}

GRadioButton *GRadioGroup::Append(int x, int y, char *name)
{
	GRadioButton *But = NEW(GRadioButton(NextId++, x, y, -1, -1, name, Children.GetItems() == 0));
	if (But)
	{
		Children.Insert(But);
	}
	return But;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio button
GRadioButton::GRadioButton(int id, int x, int y, int cx, int cy, char *name, bool first)
#ifdef LGI_RES
	: ResObject(Res_RadioBox)
#endif
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
}

GRadioButton::~GRadioButton()
{
}

int GRadioButton::Value()
{
	return 0;
}

void GRadioButton::Value(int i)
{
}

int GRadioButton::OnEvent(GMessage *Msg)
{
	return 0;
}


//////////////////////////////////////////////////////////////////////////////////
// Slider control
GSlider::GSlider(int id, int x, int y, int cx, int cy, char *name, bool vert)
#ifdef LGI_RES
	: ResObject(Res_Slider)
#endif
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	Name(name);
	Vertical = vert;
	Min = Max = 0;
	Val = 0;
}

GSlider::~GSlider()
{
}

void GSlider::Value(int i)
{
	Val = i;
}

int GSlider::Value()
{
	return Val;
}

void GSlider::GetLimits(int &min, int &max)
{
	min = Min;
	max = Max;
}

void GSlider::SetLimits(int min, int max)
{
	Min = min;
	Max = max;
}

int GSlider::OnEvent(GMessage *Msg)
{
	return 0;
}

//////////////////////////////////////////////////////////////////////////////////
#ifdef _MT
class GBitmapThread : public GThread
{
	GBitmap *Bmp;
	char *File;
	GThread **Owner;

public:
	GBitmapThread(GBitmap *bmp, char *file, GThread **owner) : GThread(true)
	{
		Bmp = bmp;
		File = NewStr(file);
		Owner = owner;
		if (Owner)
		{
			*Owner = this;
		}
		Run(true);
	}

	~GBitmapThread()
	{
		if (Owner)
		{
			*Owner = 0;
		}
		DeleteArray(File);
	}

	int Main()
	{
		if (Bmp)
		{
			GSurface *pDC = LoadDC(File);
			if (pDC)
			{
				Bmp->SetDC(pDC);
				GRegion r = Bmp->GetPos();
				r.Dimension(pDC->X()+4, pDC->Y()+4);
				Bmp->SetPos(r);
				Bmp->Invalidate();
			}
		}

		return 0;
	}
};
#endif

GBitmap::GBitmap(int id, int x, int y, char *FileName, bool Async)
#ifdef LGI_RES
	: ResObject(Res_Bitmap)
#endif
{
	pDC = 0;
	pThread = 0;

	SetId(id);
	Pos.ZOff(1, 1);
	Pos.Offset(x, y);

	#ifdef _MT
	if (NOT Async)
	{
	#endif
		pDC = LoadDC(FileName);
		if (pDC)
		{
			Pos.Dimension(pDC->X()+4, pDC->Y()+4);
		}
	#ifdef _MT
	}
	else
	{
		NEW(GBitmapThread(this, FileName, &pThread));
	}
	#endif
}

GBitmap::~GBitmap()
{
	#ifdef _MT
	if (pThread)
	{
		pThread->Terminate();
		DeleteObj(pThread);
	}
	#endif
	DeleteObj(pDC);
}

void GBitmap::SetDC(GSurface *pNewDC)
{
	DeleteObj(pDC);
	pDC = pNewDC;
}

GSurface *GBitmap::GetDC()
{
	return pDC;
}

int GBitmap::OnEvent(GMessage *Msg)
{
	return GView::OnEvent(Msg);
}

void GBitmap::OnPaint(GSurface *pScreen)
{
	GRegion a(0, 0, Pos.X()-1, Pos.Y()-1);
	LgiWideBorder(pScreen, a, SUNKEN);
	if (pDC)
	{
		pScreen->Blt(2, 2, pDC);
	}
	else
	{
		pScreen->Colour(CBit(pScreen->GetBits(), Rgb24(255, 255, 255) ));
		pScreen->Rectangle(&a);
	}
}

void GBitmap::OnMouseClick(GMouse &m)
{
	if (NOT m.Down() AND Parent)
	{
		((GDialog*) Parent)->OnNotify(this, 0);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////
GItemContainer::GItemContainer()
{
	Flags = 0;
	ImageList = 0;
}

GItemContainer::~GItemContainer()
{
	if (OwnList())
	{
		DeleteObj(ImageList);
	}
	else
	{
		ImageList = 0;
	}
}

bool GItemContainer::SetImageList(GImageList *list, bool Own)
{
	ImageList = list;
	OwnList(Own);
	AskImage(true);
	return ImageList != NULL;
}

bool GItemContainer::Load(char *File)
{
	GSurface *pDC = LoadDC(File);
	if (pDC)
	{
		ImageList = NEW(GImageList(24, 24, pDC));
		if (ImageList)
		{
			#ifdef WIN32
			ImageList->Create(pDC->X(), pDC->Y(), pDC->GetBits());
			#endif
		}
	}
	return pDC != 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
GProgress::GProgress(int id, int x, int y, int cx, int cy, char *name)
#ifdef LGI_RES
	: ResObject(Res_Progress)
#endif
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	if (name) GControl::Name(name);
}

GProgress::~GProgress()
{
}

void GProgress::SetLimits(int l, int h)
{
	Low = l;
	High = h;
}

void GProgress::Value(int v)
{
	Val = v;
}

int GProgress::OnEvent(GMessage *Msg)
{
	return 0;
}


