/*hdr
**	FILE:		File.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		11/7/95
**	DESCRIPTION:	The new file subsystem
**
**	Copyright (C) 1995, Matthew Allen
**		fret@ozemail.com.au
*/

/****************************** Includes ************************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>
#include <shlobj.h>
#include <direct.h>

#include "LibDefs.h"
#include "GFile.h"

/****************************** Defines *************************************************************************************/

#define FILEDEBUG

#define FLOPPY_360K				0x0001
#define FLOPPY_720K				0x0002
#define FLOPPY_1_2M				0x0004
#define FLOPPY_1_4M				0x0008
#define FLOPPY_5_25				(FLOPPY_360K | FLOPPY_1_2M)
#define FLOPPY_3_5				(FLOPPY_720K | FLOPPY_1_4M)

/****************************** Globals *************************************************************************************/

FileDevice *FileDevice::Instance = 0;

/****************************** Helper Functions ****************************************************************************/

int FileSize(char *FileName)
{
	VDirectory *Dir = FileDev->GetDir(0);
	if (Dir AND
		Dir->First(FileName, 0, 0))
	{
		return Dir->GetSize();
	}
	return 0;
}

bool FileExists(char *Name)
{
	WIN32_FIND_DATA Info;
	HANDLE hFind = FindFirstFile(Name, &Info);
	if (hFind != INVALID_HANDLE_VALUE)
	{
		FindClose(hFind);
		return true;
	}
	return false;
}

bool DirExists(char *Dir)
{
	int Val = SetCurrentDirectory(Dir);
	return Val;
}

int _GetLongPathName(char *Short, char *Long, int Buf)
{
	bool Status = false;
	int Len = 0;

	if (Short AND Long)
	{
		HMODULE hDll = 0;

		hDll = LoadLibrary("kernel32.dll");
		if (hDll)
		{
			// Win98/ME/2K... short->long conversion supported
			typedef DWORD (__stdcall *Proc_GetLongPathName)(LPCSTR, LPSTR, DWORD);
			Proc_GetLongPathName Proc = (Proc_GetLongPathName)GetProcAddress(hDll, "GetLongPathNameA");
			if (Proc)
			{
				Len = Proc(Short, Long, Buf);
				Status = true;
			}
			FreeLibrary(hDll);
		}
		
		if (NOT Status)
		{
			// Win95/NT4 doesn't support this function... so we do the conversion ourselves
			char *In = Short;
			char *Out = Long;

			*Out = 0;

			while (*In)
			{
				// find end of segment
				char *e = strchr(In, DIR_CHAR);
				if (NOT e) e = In + strlen(In);

				// process segment
				char Old = *e;
				*e = 0;
				if (strchr(In, ':'))
				{
					// drive specification
					strcat(Out, In); // just copy over
				}
				else
				{
					strcat(Out, DIR_STR);

					if (strlen(In) > 0)
					{
						// has segment to work on
						WIN32_FIND_DATA Info;
						HANDLE Search = FindFirstFile(Short, &Info);
						if (Search != INVALID_HANDLE_VALUE)
						{
							// have long name segment, copy over
							strcat(Out, Info.cFileName);
							FindClose(Search);
						}
						else
						{
							// no long name segment... copy over short segment :(
							strcat(Out, In);
						}
					}
					// else is a double path char... i.e. as in a M$ network path
				}

				In = (Old) ? e + 1 : e;
				Out += strlen(Out);
				*e = Old;
			}
		}
	}

	return Len;
}

bool ResolveShortcut(char *LinkFile, char *Path, int Len) 
{
	bool Status = false;
	/*
	HMODULE hDll = LoadLibrary("ole32.dll");
	if (hDll)
	{
		typedef HRESULT (__stdcall *Proc_CoInitialize)(LPVOID pvReserved);
		typedef HRESULT (__stdcall *Proc_CoCreateInstance)(REFCLSID rclsid, LPUNKNOWN pUnkOuter, DWORD dwClsContext, REFIID riid, LPVOID FAR* ppv);
		typedef void (__stdcall *Proc_CoUninitialize)(void);

		Proc_CoInitialize pCoInitialize = (Proc_CoInitialize)GetProcAddress(hDll, "CoInitialize");
		Proc_CoCreateInstance pCoCreateInstance = (Proc_CoCreateInstance)GetProcAddress(hDll, "CoCreateInstance");
		Proc_CoUninitialize pCoUninitialize = (Proc_CoUninitialize)GetProcAddress(hDll, "CoUninitialize");

		if (pCoInitialize AND
			pCoCreateInstance AND
			pCoUninitialize)
		{
		*/
			HWND hwnd = NULL;
			HRESULT hres;
			IShellLink* psl;
			char szGotPath[MAX_PATH] = "";
			WIN32_FIND_DATA wfd;

			*Path = 0;

			CoInitialize(0);
			hres = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER, IID_IShellLink, (void**) &psl);

			if (SUCCEEDED(hres))
			{
				IPersistFile* ppf;

				hres = psl->QueryInterface(IID_IPersistFile, (void**) &ppf);
				if (SUCCEEDED(hres))
				{
					WORD wsz[MAX_PATH];

					MultiByteToWideChar(CP_ACP, 0, LinkFile, -1, wsz, MAX_PATH);

					hres = ppf->Load(wsz, STGM_READ);
					if (SUCCEEDED(hres))
					{
						hres = psl->Resolve(hwnd, SLR_ANY_MATCH);
						if (SUCCEEDED(hres))
						{
							hres = psl->GetPath(szGotPath, MAX_PATH, (WIN32_FIND_DATA *)&wfd, SLGP_SHORTPATH );

							if (SUCCEEDED(hres) AND strlen(szGotPath) > 0)
							{
								// lstrcpy(Path, szGotPath);
								_GetLongPathName(szGotPath, Path, Len);
								Status = true;
							}
						}
					}

					ppf->Release();
				}

				psl->Release();
			}

			CoUninitialize();

		/*
		}

		FreeLibrary(hDll);
	}
	*/

	return Status;
}

void WriteStr(GFile &f, char *s)
{
	ulong Len = (s) ? strlen(s) : 0;
	f << Len;
	if (Len > 0)
	{
		f.Write(s, Len);
	}
}

char *ReadStr(GFile &f)
{
	char *s = 0;

	// read the strings length...
	ulong Len;
	f >> Len;

	if (Len > 0)
	{
		// 16mb sanity check.... anything over this
		// is _probably_ an error
		if (Len >= (16 << 20))
		{
			LgiAssert(0);
			return 0;
		}

		// allocate the memory buffer
		s = NEW(char[Len+1]);
		if (s)
		{
			// read the bytes from disk
			f.Read(s, Len);
			s[Len] = 0;
		}
		else
		{
			// memory allocation error, skip the data
			// on disk so the caller is where they think 
			// they are in the file.
			f.Seek(Len, SEEK_CUR);
		}
	}

	return s;
}

int SizeofStr(char *s)
{
	return sizeof(ulong) + ((s) ? strlen(s) : 0);
}

/****************************** Classes *************************************************************************************/

FileDevice::FileDevice()
{
	Instance = this;
	NumDrive = 0;
	DriveList = 0;
}

FileDevice::~FileDevice()
{
	DeleteArray(DriveList);
}

int FloppyType(int Letter)
{
	uchar MaxTrack;
	uchar SecPerTrack;

	_asm {
		mov eax, 0800h
		mov edx, Letter
		int 13h
		mov MaxTrack, ch
		mov SecPerTrack, cl
	}
	
	if (MaxTrack > 39)
	{
		switch (SecPerTrack)
		{
			case 9:
			{
				return FLOPPY_720K;
			}
			case 15:
			{
				return FLOPPY_1_2M;
			}
			case 18:
			{
				return FLOPPY_1_4M;
			}
		}
	}
	else
	{
		return FLOPPY_360K;
	}

	return 0;
}

int FileDevice::GetNumDrives()
{
	if (NOT DriveList)
	{
		FillDiskList();
	}

	return NumDrive;
}

DriveInfo *FileDevice::GetDriveInfo(int Drive)
{
	if (NOT DriveList)
	{
		FillDiskList();
	}

	return (Drive >= 0 AND Drive < NumDrive AND DriveList) ? DriveList + Drive : 0;
}

int FileDevice::FillDiskList()
{
	NumDrive = 0;
	DriveList = NEW(DriveInfo[26]);
	if (DriveList)
	{
		memset(DriveList, 0, sizeof(DriveInfo[26]));

		DriveInfo *d = DriveList;
		char Str[256];
		if (GetLogicalDriveStrings(sizeof(Str), Str) > 0)
		{
			for (char *s=Str; *s; s+=strlen(s)+1)
			{
				strcpy(d->Name, s);
				d->Type = GetDriveType(s);
				NumDrive++;
				d++;
			}
		}
	}

	return true;
}

VDirectory *FileDevice::GetDir(char *DirName)
{
	return NEW(VWin32Dir);
}

HANDLE FileDevice::OpenFile(char *Name, int Mode, int Attrib)
{
	bool Create = false;

	if (Mode & GENERIC_WRITE)
	{
		Create = true;
	}

	HANDLE hFile = CreateFile(	Name,
								Mode,
								FILE_SHARE_READ,
								0,
								(Create) ? OPEN_ALWAYS : OPEN_EXISTING,
								Attrib,
								NULL);
	if (NOT ValidHandle(hFile))
	{
		DWORD Error = GetLastError();
		switch (Error)
		{
			case 2:
			case 3:
			{
				// Path/File not found;
				int i=0;
				break;
			}
			case 5:
			{
				// Access is denied:
				// Read-Only or another process has the file open
				int i=0;
				break;
			}
		}
	}

	return hFile;
}

bool FileDevice::DeleteFile(char *FileName)
{
	return ::DeleteFile(FileName);
}

bool FileDevice::CreateDirectory(char *PathName)
{
	return ::CreateDirectory(PathName, NULL);
}

bool FileDevice::RemoveDirectory(char *PathName)
{
	return ::RemoveDirectory(PathName);
}

bool FileDevice::SetCurrentDirectory(char *PathName)
{
	return ::SetCurrentDirectory(PathName);
}

bool FileDevice::GetCurrentDirectory(char *PathName, int Length)
{
	return (::GetCurrentDirectory(Length, PathName) > 0);
}

bool FileDevice::MoveFile(char *OldName, char *NewName)
{
	return ::MoveFile(OldName, NewName);
}

bool FileDevice::GetVolumeInfomation(char Drive, VolumeInfo *pInfo)
{
	bool Status = false;

	if (pInfo)
	{
		char Name[] = "A:\\";
		ulong Serial;

		Name[0] = Drive + 'A';

		if (GetVolumeInformation(	Name,
						pInfo->Name,
						sizeof(pInfo->Name),
						&Serial,
						&pInfo->MaxPath,
						&pInfo->Flags,
						pInfo->System,
						sizeof(pInfo->System)))
		{
			pInfo->Drive = Drive;
			Status = true;
		}
	}

	return Status;
}

void GetFlagsStr(char *flagstr, short flags)
{
	if (flagstr)
	{
		flagstr[0] = 0;

		if (flags & O_WRONLY)
		{
			strcat(flagstr,"w");
		}
		else if (flags & O_APPEND)
		{
			strcat(flagstr,"a");
		}
		else
		{
			strcat(flagstr,"r");
		}

		if ((flags & O_RDWR) == O_RDWR)
		{
			strcat(flagstr,"+");
		}

		strcat(flagstr,"b");
	}
}

bool Match(char *Name, char *Mask)
{
	strupr(Name);
	strupr(Mask);

	while (*Name AND *Mask)
	{
		if (*Mask == '*')
		{
			if (*Name == *(Mask+1))
			{
				Mask++;
			}
			else
			{
				Name++;
			}
		}
		else if (*Mask == '?' OR *Mask == *Name)
		{
			Mask++;
			Name++;
		}
		else
		{
			return false;
		}
	}

	while (*Mask AND ((*Mask == '*') OR (*Mask == '.'))) Mask++;

	return (*Name == 0 AND *Mask == 0);
}

short DaysInMonth[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
int LeapYear(int year)
{
	if (year & 3)
	{
		return 0;
	}
	if ((year % 100 == 0) && !(year % 400 == 0))
	{
		return 0;
	}
	
	return 1;
}

// Disk Infomation
typedef struct
{
    uchar	bsJump[3];
    uchar	bsOemName[8];
    ushort	bsBytePerSec;
    uchar	bsSecPerCluster;
    ushort	bsResSectores;
    uchar	bsFAT;
    ushort	bsRootDirEnts;
    ushort	bsSectors;
    uchar	bsMedia;
    ushort	bsFATsecs;
    ushort	bsSecPerTrack;
    ushort	bsHeads;
    ulong	bsHiddenSecs;
    ulong	bsHugeSectoes;
    uchar	bsDriveNumber;
    uchar	bsReserved;
    uchar	bsBootSig;
    ulong	bsVolumeID; // serial number
    uchar	bsVolumeLabel[11];
    uchar	bsFileSysType[8];

} BOOTSECTOR;

typedef struct
{
    ulong	diStartSector;
    ushort	diSectors;
    char	*diBuffer;
} DISKIO;

/////////////////////////////////////////////////////////////////////////////////
bool VDirectory::ConvertToTime(char *Str, quad Time)
{
	if (Str)
	{
		FILETIME t;
		SYSTEMTIME s;
		t.dwHighDateTime = Time >> 32;
		t.dwLowDateTime = Time & 0xFFFFFFFF;
		if (FileTimeToSystemTime(&t, &s))
		{
			int Hour = (s.wHour < 1) ? s.wHour + 23 : s.wHour - 1;
			char AP = (Hour >= 12) ? 'a' : 'p';
			Hour %= 12;
			if (Hour == 0) Hour = 12;
			sprintf(Str, "%i:%02.2i:%02.2i %c", Hour, s.wMinute, s.wSecond, AP);
		}
	}
	return false;
}

bool VDirectory::ConvertToDate(char *Str, quad Time)
{
	if (Str)
	{
		FILETIME t;
		SYSTEMTIME s;
		t.dwHighDateTime = Time >> 32;
		t.dwLowDateTime = Time & 0xFFFFFFFF;
		if (FileTimeToSystemTime(&t, &s))
		{
			sprintf(Str, "%i/%i/%i", s.wDay, s.wMonth, s.wYear);
		}
	}
	return false;
}

/////////////////////////////////////////////////////////////////////////////////
//////////////////////////// Directory //////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

VWin32Dir::VWin32Dir()
{
	Open = false;
}

VWin32Dir::~VWin32Dir()
{
	Close();
}

int VWin32Dir::First(char *Name, char Must, char Search)
{
	Close();

	if (Name)
	{
		char *End = 0;
		for (End=Name+strlen(Name)-1; End>Name AND *End!=DIR_CHAR; End--);
		if (End > Name)
		{
			char Buffer[260];

			if (End[-1] == ':')
			{
				// raw drive
				strcpy(Buffer, Name);
			}
			else
			{
				// dir
				*End = 0;
				GetFullPathName(Name, sizeof(Buffer), Buffer, NULL);
				*End = DIR_CHAR;
				if (Buffer[strlen(Buffer)-1] != DIR_CHAR) strcat(Buffer, DIR_STR);
				strcat(Buffer, End+1);
			}

			Handle = FindFirstFile(Buffer, &Data);
			if (Handle != INVALID_HANDLE_VALUE)
			{
				Open = true;
			}
		}
	}

	return Open;
}

int VWin32Dir::Next()
{
	int Status = false;
	
	if (Open)
	{
		Status = FindNextFile(Handle, &Data);
	}

	return Status;
}

int VWin32Dir::Close()
{
	if (Open)
	{
		FindClose(Handle);
		Open = false;
		Handle = INVALID_HANDLE_VALUE;
	}

	return true;
}

bool VWin32Dir::IsDir()
{
	return (Data.dwFileAttributes & FA_DIRECTORY) != 0;
}

long VWin32Dir::GetAttributes()
{
	return Data.dwFileAttributes;
}

char *VWin32Dir::GetName()
{
	return Data.cFileName;
}

void VWin32Dir::GetName(char *Buf, int Len)
{
	strncpy(Buf, Data.cFileName, Len);
}

const quad VWin32Dir::GetCreationTime()
{
	return ((quad) Data.ftCreationTime.dwHighDateTime) << 32 | Data.ftCreationTime.dwLowDateTime;
}

const quad VWin32Dir::GetLastAccessTime()
{
	return ((quad) Data.ftLastAccessTime.dwHighDateTime) << 32 | Data.ftLastAccessTime.dwLowDateTime;
}

const quad VWin32Dir::GetLastWriteTime()
{
	return ((quad) Data.ftLastWriteTime.dwHighDateTime) << 32 | Data.ftLastWriteTime.dwLowDateTime;
}

const quad VWin32Dir::GetSize()
{
	return ((quad) Data.nFileSizeHigh) << 32 | Data.nFileSizeLow;
}

/////////////////////////////////////////////////////////////////////////////////
//////////////////////////// File ///////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

GFile::GFile()
{
	hFile = INVALID_HANDLE;
	Name = 0;
	Attributes = 0;
	Swap = false;
	Status = false;
}

GFile::~GFile()
{
	if (ValidHandle(hFile))
	{
		Close();
	}
	DeleteArray(Name);
}

bool GFile::Open(char *File, int Mode)
{
	int Status = false;

	if (File)
	{
		Close();
		hFile = FileDev->OpenFile(File, Mode, 0);

		if (ValidHandle(hFile))
		{
			Attributes = Mode;
			Name = NEW(char[strlen(File)+1]);
			if (Name)
			{
				strcpy(Name, File);
			}
			
			Status = true;
			GFile::Status = false;
		}
	}

	return Status;
}

bool GFile::Close()
{
	int Status = false;

	if (ValidHandle(hFile))
	{
		::CloseHandle(hFile);
		hFile = INVALID_HANDLE;
	}

	DeleteArray(Name);

	return Status;
}

int GFile::Print(char *Format, ...)
{
	int Chars = 0;

	if (Format)
	{
		char Buffer[1024];
		va_list Arg;

		va_start(Arg, Format);
		Chars = vsprintf(Buffer, Format, Arg);
		va_end(Arg);

		if (Chars > 0)
		{
			Write(Buffer, Chars);
		}
	}

	return Chars;
}

#define CHUNK		0xFFF0

int GFile::Read(void *Buffer, int Size)
{
	DWORD Red;
	Status |= NOT ReadFile(hFile, Buffer, Size, &Red, NULL);
	return Red;
}

int GFile::Write(void *Buffer, int Size)
{
	DWORD Written = 0;
	Status |= NOT WriteFile(hFile, Buffer, Size, &Written, NULL);
	return Written;
}

int GFile::Seek(int To, int Whence)
{
	int Mode;
	switch (Whence)
	{
		default:
		case SEEK_SET:
			Mode = FILE_BEGIN;
			break;
		case SEEK_CUR:
			Mode = FILE_CURRENT;
			break;
		case SEEK_END:
			Mode = FILE_END;
			break;
	}

	return SetFilePointer(hFile, To, NULL, Mode) != 0xFFFFFFFF;
}

int GFile::GetPosition()
{
	return SetFilePointer(hFile, 0, NULL, FILE_CURRENT);
}

int GFile::GetSize()
{
	return GetFileSize(hFile, NULL);
}

bool GFile::SetSize(int Size)
{
	int OldPos = SetFilePointer(hFile, 0, NULL, SEEK_CUR);
	SetFilePointer(hFile, Size, NULL, FILE_BEGIN);
	SetEndOfFile(hFile);
	SetFilePointer(hFile, OldPos, NULL, FILE_BEGIN);

	return GetFileSize(hFile, NULL) == (uint) Size;
}

bool GFile::Eof()
{
	return GetPosition() >= GetSize();
}

int GFile::SwapRead(uchar *Buf, int Size)
{
	int i = 0;
	DWORD r;
	Buf = &Buf[Size-1];
	while (Size--)
	{
		ReadFile(hFile, Buf--, 1, &r, NULL);
		i += r;
	}
	return i;
}

int GFile::SwapWrite(uchar *Buf, int Size)
{
	int i = 0;
	DWORD w;
	Buf = &Buf[Size-1];
	while (Size--)
	{
		WriteFile(hFile, Buf--, 1, &w, NULL);
		i += w;
	}
	return i;
}

int GFile::ReadStr(char *Buf, int Size)
{
	int i = 0;
	DWORD r;
	if (Buf AND Size > 0)
	{
		char c;

		Size--;

		do
		{
			ReadFile(hFile, &c, 1, &r, NULL);
			if (Eof())
			{
				break;
			}
			
			*Buf++ = c;
			i++;
		
		} while (i < Size - 1 AND c != '\n');

		*Buf = 0;
	}

	return i;
}

int GFile::WriteStr(char *Buf, int Size)
{
	int i = 0;
	DWORD w;

	while (i <= Size)
	{
		WriteFile(hFile, Buf, 1, &w, NULL);
		Buf++;
		i++;
		if (*Buf == '\n') break;
	}

	return i;
}

void GFile::SetStatus(bool s)
{
	Status = s;
}

bool GFile::GetStatus()
{
	return Status;
}

#define RdIO { Status |= ((Swap) ? SwapRead((uchar*) &i, sizeof(i)) : Read(&i, sizeof(i))) != sizeof(i); return *this; }
#define WrIO { Status |= ((Swap) ? SwapWrite((uchar*) &i, sizeof(i)) : Write(&i, sizeof(i))) != sizeof(i); return *this; }

GFile &GFile::operator >> (char &i) RdIO
GFile &GFile::operator >> (signed char &i) RdIO
GFile &GFile::operator >> (unsigned char &i) RdIO
GFile &GFile::operator >> (signed short &i) RdIO
GFile &GFile::operator >> (unsigned short &i) RdIO
GFile &GFile::operator >> (signed int &i) RdIO
GFile &GFile::operator >> (unsigned int &i) RdIO
GFile &GFile::operator >> (signed long &i) RdIO
GFile &GFile::operator >> (unsigned long &i) RdIO
GFile &GFile::operator >> (float &i) RdIO
GFile &GFile::operator >> (double &i) RdIO
GFile &GFile::operator >> (long double &i) RdIO

GFile &GFile::operator << (char i) WrIO
GFile &GFile::operator << (signed char i) WrIO
GFile &GFile::operator << (unsigned char i) WrIO
GFile &GFile::operator << (signed short i) WrIO
GFile &GFile::operator << (unsigned short i) WrIO
GFile &GFile::operator << (signed int i) WrIO
GFile &GFile::operator << (unsigned int i) WrIO
GFile &GFile::operator << (signed long i) WrIO
GFile &GFile::operator << (unsigned long i) WrIO
GFile &GFile::operator << (float i) WrIO
GFile &GFile::operator << (double i) WrIO
GFile &GFile::operator << (long double i) WrIO

VFileBuf::VFileBuf()
{
	BufSize = 0;
	BufPos = 0;
	Data = 0;
}

VFileBuf::~VFileBuf()
{
	Close();
}

bool VFileBuf::Open(char *Name, int Attrib)
{
	bool Status = false;

	if (GFile::Open(Name, Attrib))
	{
		BufPos = 0;
		BufSize = GFile::GetSize();
		Data = NEW(uchar[BufSize]);
		if (Data)
		{
			GFile::Read(Data, BufSize);
			Status = true;
		}
	}

	return Status;
}

bool VFileBuf::Open()
{
	return (Data != 0) AND ValidHandle(hFile);
}

bool VFileBuf::Close()
{
	BufSize = 0;
	BufPos = 0;
	DeleteArray(Data);
	GFile::Close();
	return true;
}

int VFileBuf::Read(void *Buffer, int Size)
{
	int Len = 0;

	if (Buffer AND Data AND BufPos >= 0 AND BufPos + 1 < BufSize)
	{
		Len = min(Size, BufSize - BufPos);
		memcpy(Buffer, Data + BufPos, Len);
		BufPos += Len;
	}

	return Len;
}

int VFileBuf::Write(void *Buffer, int Size)
{
	int Len = 0;

	if (Buffer AND BufPos >= 0 AND BufPos + 1 < BufSize)
	{
		Len = min(Size, BufSize - BufPos);
		memcpy(Data + BufPos, Buffer, Len);
		BufPos += Len;

		Seek(BufPos, SEEK_SET);
		Len = GFile::Write(Buffer, Size);
	}

	return Len;
}

int VFileBuf::Seek(int To, int Whence)
{
	int Start = BufPos;

	switch (Whence)
	{
		case SEEK_SET:
		{
			BufPos = To;
			break;
		}
		case SEEK_CUR:
		{
			BufPos += To;
			break;
		}
		case SEEK_END:
		{
			BufPos = BufSize + To;
			break;
		}
	}

	if (BufPos != Start)
	{
		GFile::Seek(To, Whence);
	}
	
	return BufPos;
}

int VFileBuf::GetPosition()
{
	return BufPos;
}

int VFileBuf::GetSize()
{
	return BufSize;
}

bool VFileBuf::SetSize(int Size)
{
	bool Status = false;

	if (Size != BufSize)
	{
		int Common = min(Size, BufSize);
		uchar *Temp = NEW(uchar[Size]);
		if (Temp)
		{
			memcpy(Temp, Data, Common);
			if (Common < Size)
			{
				memset(Temp + Common, 0, Size - Common);
			}

			DeleteArray(Data);
			Data = Temp;
			BufSize = Size;
			Status = GFile::SetSize(Size);
		}
	}

	return Status;
}

bool VFileBuf::Eof()
{
	return (BufPos >= 0 AND BufPos < BufSize);
}

int VFileBuf::ReadStr(char *Buf, int Size)
{
	int i;

	for (i=0; i<Size-1 AND Data[BufPos++] != '\n'; i++);
	if (Buf)
	{
		memcpy(Buf, Data+BufPos, i);
		Buf[Size-1] = 0;
		return i;
	}
	return 0;
}

int VFileBuf::WriteStr(char *Buf, int Size)
{
	if (Buf)
	{
		int i;
		uchar *p = Data + BufPos;
		for (i=0; i<Size AND Buf[BufPos] != '\n'; i++)
			*p++ = Buf[BufPos++];

		Seek(BufPos, SEEK_SET);
		GFile::Write(Buf, i);

		return i;
	}
	return 0;
}

int VFileBuf::SwapRead(uchar *Buf, int Size)
{
	int i=0;
	Buf = &Buf[Size-1];
	while (Size-- AND BufPos < BufSize)
	{
		*Buf-- = Data[BufPos++];
		i++;
	}
	return i;
}

int VFileBuf::SwapWrite(uchar *Buf, int Size)
{
	int i=0;
	int Start = BufPos;

	Buf = &Buf[Size-1];
	while (Size-- AND BufPos < BufSize)
	{
		Data[BufPos++] = *Buf--;
		i++;
	}

	Seek(Start, SEEK_SET);
	GFile::Write(Data + Start, i);

	return i;
}

#define BufRdIO { Status |= ((Swap) ? SwapRead((uchar*) &i, sizeof(i)) : Read(&i, sizeof(i))) != sizeof(i); return *this; }
#define BufWrIO { Status |= ((Swap) ? SwapWrite((uchar*) &i, sizeof(i)) : Write(&i, sizeof(i))) != sizeof(i); return *this; }

GFile &VFileBuf::operator >> (char &i) BufRdIO
GFile &VFileBuf::operator >> (signed char &i) BufRdIO
GFile &VFileBuf::operator >> (unsigned char &i) BufRdIO
GFile &VFileBuf::operator >> (signed short &i) BufRdIO
GFile &VFileBuf::operator >> (unsigned short &i) BufRdIO
GFile &VFileBuf::operator >> (signed int &i) BufRdIO
GFile &VFileBuf::operator >> (unsigned int &i) BufRdIO
GFile &VFileBuf::operator >> (signed long &i) BufRdIO
GFile &VFileBuf::operator >> (unsigned long &i) BufRdIO
GFile &VFileBuf::operator >> (float &i) BufRdIO
GFile &VFileBuf::operator >> (double &i) BufRdIO
GFile &VFileBuf::operator >> (long double &i) BufRdIO

GFile &VFileBuf::operator << (char i) BufWrIO
GFile &VFileBuf::operator << (signed char i) BufWrIO
GFile &VFileBuf::operator << (unsigned char i) BufWrIO
GFile &VFileBuf::operator << (signed short i) BufWrIO
GFile &VFileBuf::operator << (unsigned short i) BufWrIO
GFile &VFileBuf::operator << (signed int i) BufWrIO
GFile &VFileBuf::operator << (unsigned int i) BufWrIO
GFile &VFileBuf::operator << (signed long i) BufWrIO
GFile &VFileBuf::operator << (unsigned long i) BufWrIO
GFile &VFileBuf::operator << (float i) BufWrIO
GFile &VFileBuf::operator << (double i) BufWrIO
GFile &VFileBuf::operator << (long double i) BufWrIO

/////////////////////////////////////////////////////////////////////////////////
//////////////////////////// VDirView ///////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

VDirView::VDirView()
{
	Items = 0;
	Files = 0;
	Dirs = 0;
	strcpy(CurrentDir, ".");


	Index = 0;
	Root = 0;
}

VDirView::~VDirView()
{
	DeleteObj(Root);
	DeleteArray(Index);
}

bool VDirView::Read(char *Dir)
{
	if (NOT Dir) Dir = ".";
	FileDev->SetCurrentDirectory(Dir);
	FileDev->GetCurrentDirectory(CurrentDir, sizeof(CurrentDir));

	if (Root)
	{
		delete Root;
		Root = 0;
	}

	Items = 0;
	Files = 0;
	Dirs = 0;
	DeleteArray(Index);

	Node *Current;
	if (strlen(CurrentDir) > 3)
	{
		Current = NEW(Node("..", FA_DIRECTORY, 0));
		if (Current)
		{
			if (NOT Root)
			{
				Root = Current;
			}
			else
			{
				Root->Add(Current);
			}

			Dirs++;
			Items++;
		}
	}

	bool Done = First("*.*", 0x10, 0x1F);
	while (NOT Done)
	{
		if (strcmp(GetName(), ".") AND strcmp(GetName(), ".."))
		{
			Current = NEW(Node(GetName(), GetAttributes(), GetSize()));
			if (Current)
			{
				if (NOT Root)
				{
					Root = Current;
				}
				else
				{
					Root->Add(Current);
				}

				Dirs++;
				Items++;
			}
		}
		Done = Next();
	}

	Done = First("*.*", 0x0, 0x2F);
	while (NOT Done)
	{
		Current = NEW(Node(GetName(), GetAttributes(), GetSize()));
		if (Current)
		{
			if (NOT Root)
			{
				Root = Current;
			}
			else
			{
				Root->Add(Current);
			}

			Files++;
			Items++;
		}
		Done = Next();
	}

	if (Root AND Items)
	{
		Index = NEW(Node*[Items]);
		if (Index)
		{
			Node **Ptr = Root->Traverse(Index, true);
			Root->Traverse(Ptr, false);
		}
	}

	return true;
}

VDirView::Node::Node()
{
	Left = Right = 0;
}

VDirView::Node::Node(char *n, int a, quad s) : NodeInfo(n, a, s)
{
	Left = Right = 0;
}

VDirView::Node::~Node()
{
	DeleteObj(Left);
	DeleteObj(Right);
}

VDirView::Node **VDirView::Node::Traverse(VDirView::Node **ppNode, int Dir)
{
	if (Left)
	{
		ppNode = Left->Traverse(ppNode, Dir);
	}

	if (Dir)
	{
		if (GetAttributes() & FA_DIRECTORY)
		{
			*ppNode++ = this;
		}
	}
	else
	{
		if (NOT (GetAttributes() & FA_DIRECTORY))
		{
			*ppNode++ = this;
		}
	}

	if (Right)
	{
		ppNode = Right->Traverse(ppNode, Dir);
	}

	return ppNode;
}

void VDirView::Node::Add(Node *Info)
{
	int Result = *this ^ *Info;

	if (Result < 0)
	{
		if (Right)
		{
			Right->Add(Info);
		}
		else
		{
			Right = Info;
		}
	}
	else if (Result > 0)
	{
		if (Left)
		{
			Left->Add(Info);
		}
		else
		{
			Left = Info;
		}
	}
}

NodeInfo::NodeInfo()
{
	Name = 0;
	Attributes = 0;
	Size = 0;
}

NodeInfo::NodeInfo(char *n, int a, quad s)
{
	if (n)
	{
		Name = NEW(char[strlen(n)+1]);
		if (Name)
		{
			strcpy(Name, n);
		}
	}
	else
	{
		Name = 0;
	}
	Attributes = a;
	Size = s;
}

NodeInfo::~NodeInfo()
{
	DeleteArray(Name);
}

int NodeInfo::operator ^(NodeInfo &N)
{
	return stricmp(Name, N.Name);
}

