/*hdr
**      FILE:           Memory.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           10/4/98
**      DESCRIPTION:    Memory subsystem
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <string.h>
#include <crtdbg.h>

/*
#include "LibDefs.h"
#include "LgiOsDefs.h"
#include "VMem.h"
*/

#include "Lgi.h"

#ifdef MEMORY_DEBUG

// #define _VMEM

#ifdef _VMEM

#define MEM_MAGIC				0xDEFA0012
#define MEM_PAD					64
#define PAD_VALUE				0xAA
#define PAD_UNALLOC				0xBB

struct _Mem
{
	_Mem *Next, *Prev;
	uint Magic;
	char File[24];
	int Line;
	int Size;
	uchar Pad[MEM_PAD];

} *_First = 0, *_Last = 0;

#endif

char *_vmem_file = 0;
int _vmem_line = 0;
int _vmem_allocs = 0;

#ifdef _VMEM
class _VMem : public GSemaphore
{
public:
	~_VMem()
	{
		if (Lock())
		{
			char File[256];
			LgiGetExePath(File);
			if (File[strlen(File)-1] != DIR_CHAR) strcat(File, DIR_STR);
			strcat(File, "vmemdump.txt");
			GFile f;
			if (f.Open(File, O_WRITE))
			{
				f.SetSize(0);
				f.Print("Blocks leaked: %i\n", _vmem_allocs);
				for (_Mem *m = _First; m; m = m->Next)
				{
					f.Print("Block, size: %i, alloc: %s(%i)\n", m->Size, m->File, m->Line);
				}
			}
		}
	}
} _MemoryDumper;

bool _vmem_check(uchar *p, int v, int len)
{
	while (len--)
	{
		if (*p++ != v) return false;
	}

	return true;
}
#endif

bool LgiCheckHeap()
{
	int a = 0;
	int Error = 0;

	#ifdef _VMEM

	if (_MemoryDumper.Lock())
	{
		for (_Mem *m = _First; m; m = m->Next)
		{
			// check magic
			if (m->Magic != MEM_MAGIC)
			{
				Error = 1;
				break;
			}

			// check for under/over writes
			if (NOT _vmem_check(m->Pad, PAD_VALUE, MEM_PAD)) // under write
			{
				Error = 2;
				break;
			}

			if (NOT _vmem_check( ((uchar*)(m+1)) + m->Size, PAD_VALUE, MEM_PAD)) // over write
			{
				Error = 3;
				break;
			}

			// count blocks
			a++;
		}

		if (_vmem_allocs != a)
		{
			Error = 4;
		}

		LgiAssert(Error == 0);

		_MemoryDumper.Unlock();
	}

	#endif

	return Error == 0;
}

void *_vmem_alloc(size_t Size, char *file, int line)
{
	#ifdef _VMEM
	
	_Mem *m = 0;

	if (_MemoryDumper.Lock())
	{
		// allocate memory
		m = (_Mem*) malloc(Size + sizeof(_Mem) + MEM_PAD);
		
		// check we got the block
		LgiAssert(m);

		// insert in block list...
		m->Next = 0;
		if (_Last)
		{
			_Last->Next = m;
			m->Prev = _Last;
			_Last = m;
		}
		else
		{
			_First = _Last = m;
			m->Prev = 0;
		}

		// set data items
		char *f = file ? strrchr(file, DIR_CHAR) : 0;

		m->Magic = MEM_MAGIC;
		memset(m->Pad, PAD_VALUE, sizeof(m->Pad));
		memset( ((char*)(m+1)) + Size, PAD_VALUE, sizeof(m->Pad));
		m->Line = line;
		strncpy(m->File, (f) ? f+1 : file, sizeof(m->File));
		m->Size = Size;

		// inc block count
		_vmem_allocs++;

		_MemoryDumper.Unlock();
	}

	// return block of memory
	return (m) ? m + 1 : 0;

	#else

	return malloc(Size);

	#endif
}

void _vmem_free(void *p)
{
	#ifdef _VMEM
	
	if (_MemoryDumper.Lock())
	{
		// check for NULL delete
		LgiAssert(p);

		// get pointer to the mem info block
		_Mem *m = ((_Mem*)p) - 1;

		// check majic
		LgiAssert(m->Magic == MEM_MAGIC);

		// dec block count
		LgiAssert(_vmem_allocs > 0);
		_vmem_allocs--;

		// take block out of list
		if (m->Next)
		{
			m->Next->Prev = m->Prev;
		}
		if (m->Prev)
		{
			m->Prev->Next = m->Next;
		}
		if (_First == m)
		{
			_First = m->Next;
		}
		if (_Last == m)
		{
			_Last = m->Prev;
		}

		// check for under/over writes
		LgiAssert(_vmem_check(m->Pad, PAD_VALUE, MEM_PAD)); // under write
		LgiAssert(_vmem_check( ((uchar*)(m+1)) + m->Size, PAD_VALUE, MEM_PAD)); // over write

		// give memory back to OS
		free(m);

		_MemoryDumper.Unlock();
	}

	#else

	free(p);

	#endif
}

#else

bool CheckHeap()
{
	return true;
}

#endif

void MemorySizeToStr(char *Str, quad Size)
{
	#define K				1024.0
	#define M				(K*K)
	#define G				(M*K)

	if (Size >= G)
	{
		sprintf(Str, "%.2 G", Size / G);
	}
	else if (Size >= M)
	{
		sprintf(Str, "%.2f M", Size / M);
	}
	else if (Size >= (10 * K))
	{
		sprintf(Str, "%.2f K", Size / K);
	}
	else
	{
		sprintf(Str, "%ld bytes", Size);
	}

	#undef K
	#undef M
	#undef G
}

