/*hdr
**	FILE:		Gdc2.h
**	AUTHOR:		Matthew Allen
**	DATE:		20/2/97
**	DESCRIPTION:	GDC v2.xx header
**
**	Copyright (C) 1997, Matthew Allen
**		fret@ozemail.com.au
*/

#ifndef _GDC2_H_
#define _GDC2_H_

// sub-system headers
#include "LibDefs.h"				// Xp
#include "LgiOsDefs.h"				// Platform specific
#include "Progress.h"				// Xp
#include "GFile.h"					// Platform specific
#include "VMem.h"					// Platform specific
#include "Core.h"					// Platform specific
#include "Contain.h"				// Xp

//				Defines
#define DEFAULT_GAMMA				1.0
#define PI							3.141592654
#define DegToRad(i)					((i)*PI/180)
#define RadToDeg(i)					((i)*180/PI)
#define GDC_USE_ASM					// Use intel assembly instructions
									// comment out for porting

// GDC2 Compile time options
#define ADJUST_ORIGIN				// allows you the set the origin of a DC

// Drawing operation types
#define GDC_SET						0
#define GDC_AND						1
#define GDC_OR						2
#define GDC_XOR						3
#define GDC_ALPHA					4
#define GDC_REMAP					5
#define GDC_MAXOP					6
#define GDC_CACHE_SIZE				4

// Channel types
#define GDCC_MONO					0
#define GDCC_GREY					1
#define GDCC_INDEX					2
#define GDCC_R						3
#define GDCC_G						4
#define GDCC_B						5
#define GDCC_ALPHA					6

// Native data formats
#define GDC_8BIT					0
#define GDC_16BIT					1
#define GDC_24BIT					2
#define GDC_32BIT					3
#define GDC_MAXFMT					4

// Colour spaces
#define GDC_8I						0	// 8 bit paletted
#define GDC_5R6G5B					1	// 16 bit
#define GDC_8B8G8B					2	// 24 bit
#define GDC_8A8R8G8B				3	// 32 bit
#define GDC_MAXSPACE				4

// Update types
#define GDC_PAL_CHANGE				0x1
#define GDC_BITS_CHANGE				0x2

// Flood fill types
#define GDC_FILL_TO_DIFFERENT		0
#define GDC_FILL_TO_BORDER			1
#define GDC_FILL_NEAR				2

// Gdc options
#define GDC_REDUCE_TYPE				0
	#define REDUCE_NONE					0
	#define REDUCE_NEAREST				1
	#define REDUCE_HALFTONE				2
	#define REDUCE_ERROR_DIFFUSION		3
	#define REDUCE_DL1					4
#define GDC_HALFTONE_BASE_INDEX		1
#define GDC_PALETTE_TYPE			2
	#define PALTYPE_ALLOC				0
	#define PALTYPE_RGB_CUBE			1
	#define PALTYPE_HSL					2
#define GDC_IMPORT_8_AS_16			3
#define GDC_MAX_OPTION				4

//	Flags
// Surface
#define GDC_OWN_MEMORY				0x0001
#define GDC_ON_SCREEN				0x0002
#define GDC_ALPHA_CHANNEL			0x0004
#define GDC_UPDATED_PALETTE			0x0008

// BaseDC
#define GDC_OWN_APPLICATOR			0x0001
#define GDC_CACHED_APPLICATOR		0x0002
#define GDC_OWN_PALETTE				0x0004
#define GDC_DRAW_ON_ALPHA			0x0008

// Region types
#define GDC_RGN_NONE				0	// No clipping
#define GDC_RGN_SIMPLE				1	// Single rectangle
#define GDC_RGN_COMPLEX				2	// Many rectangles

// Error codes
#define GDCERR_NONE					0
#define GDCERR_ERROR				1
#define GDCERR_CANT_SET_SCAN_WIDTH	2
#define GDC_INVALIDMODE				-1

// Font display flags
#define GDCFNT_TRANSPARENT			0x0001	// not set - SOLID
#define GDCFNT_UNDERLINE			0x0002

// Palette file types
#define GDCPAL_JASC					1
#define GDCPAL_MICROSOFT			2

// Misc
#define BMPWIDTH(bits)				((((bits)+31)/32)*4)


//				Typedefs

//	COLOUR typedef description
//
//				31              24 23              16 15              8  7               0
//				|-|-|-|-|-|-|-|-|  |-|-|-|-|-|-|-|-|  |-|-|-|-|-|-|-|-|  |-|-|-|-|-|-|-|-|
//
//  Standard:
//	8bit mode:	|------------------------ unused ---------------------|  |---- index ----|
//	15bit mode:	|----------------- unused ------------|T|-- red --|-- green ---|-- blue -|
//	16bit mode:	|------------- unused -------------|  |-- red --|--- green ----|-- blue -|
//	24bit mode:	|---- unused ---|  |----- blue ----|  |---- green ----|  |----- red -----|
//	32bit mode:	|---- alpha ----|  |----- red -----|  |---- green ----|  |----- blue ----|
//		
//  Note:		In 15bit mode the T bit can toggle between 8-bit indexing and RGB
//	
typedef unsigned long				COLOUR;

// alpha value, 0 - transparent, 255 - solid
typedef unsigned char				ALPHA;

// colour conversion defines
// RGB colour space
#define BitWidth(bits, cropbits)	( (((bits)+(cropbits)-1)/(cropbits)) * 4 )
#define Rgb32(r, g, b)				( (b&0xFF) | ((g&0xFF)<<8) | ((r&0xFF)<<16) )
#define Rgb24(r, g, b)				( (r&0xFF) | ((g&0xFF)<<8) | ((b&0xFF)<<16) )
#define Rgb15(r, g, b)				( ((r&0xF8)<<7) | ((g&0xF8)<<2) | ((b&0xF8)>>3))
#define Rgb16(r, g, b)				( ((r&0xF8)<<8) | ((g&0xFC)<<3) | ((b&0xF8)>>3))

#define Rgb32To24(c32)				( ((c32&0xFF0000)>>16) | ((c32&0xFF00)) | ((c32&0xFF)<<16))
#define Rgb32To16(c32)				( ((c32&0xF8)>>3) | ((c32&0xFC00)>>5) | ((c32&0xF80000)>>8) )
#define Rgb32To15(c32)				( ((c32&0xF8)>>3) | ((c32&0xF800)>>6) | ((c32&0xF80000)>>9) )
#define Rgb24To32(c24)				( ((c24&0xFF0000)>>16) | ((c24&0xFF00)) | ((c24&0xFF)<<16))
#define Rgb24To16(c24)				( ((c24>>19)&0x1F) | ((c24>>5)&0x7E0) | ((c24<<8)&0xF800) )
#define Rgb24To15(c24)				( ((c24>>19)&0x1F) | ((c24>>6)&0x3E0) | ((c24<<7)&0x7C00) )
#define Rgb15To32(c15)				( ((c15&0x7C00)<<9) | ((c15&0x3E0)<<6) | ((c15&0x1F)<<3) )
#define Rgb15To16(c15)				( ((c15&0x7FE)<<1) | (c15&0x1F) )
#define Rgb15To24(c15)				( ((c15&0x7C00)>>7) | ((c15&0x3E0)<<6) | ((c15&0x1F)<<19) )
#define Rgb16To32(c16)				( ((c16&0xF800)<<8) | ((c16&0x7E0)<<5) | ((c16&0x1F)<<3) )
#define Rgb16To15(c16)				( ((c16&0xFFC0)>>1) | (c16&0x1F) )
#define Rgb16To24(c16)				( ((c16&0xF800)>>8) | ((c16&0x7E0)<<5) | ((c16&0x1F)<<19) )

#define R32(c32)					( (uchar) (c32>>16) )
#define G32(c32)					( (uchar) (c32>>8) )
#define B32(c32)					( (uchar) (c32) )

#define R24(c24)					( (uchar) (c24) )
#define G24(c24)					( (uchar) (c24>>8) )
#define B24(c24)					( (uchar) (c24>>16) )

#define R16(c16)					( (uchar) ((c16>>11)&0x1F) )
#define G16(c16)					( (uchar) ((c16>>5)&0x3F) )
#define B16(c16)					( (uchar) ((c16)&0x1F) )

#define R15(c15)					( (uchar) ((c15>>10)&0x1F) )
#define G15(c15)					( (uchar) ((c15>>5)&0x1F) )
#define B15(c15)					( (uchar) ((c15)&0x1F) )

// HLS colour Space
#define Hls32(h, l, s)				( ((h)<<16) | ((l)<<8) | (s) )
#define H32(c)						( ((c)>>16)&0xFFFF )
#define L32(c)						( ((c)>>8)&0xFF )
#define S32(c)						( (c)&0xFF )
#define HUE_UNDEFINED				1024
#define HlsIsUndefined(Hls)			(H32(Hls) == HUE_UNDEFINED)

LgiFunc COLOUR RgbToHls(COLOUR Rgb24);
LgiFunc COLOUR HlsToRgb(COLOUR Hsl32);

#pragma pack(push, before_pack)
#pragma pack(1)

class LgiClass Pixel24 {
public:
	uchar B;
	uchar G;
	uchar R;
};

class LgiClass Pixel32 {
public:
	uchar A;
	uchar R;
	uchar G;
	uchar B;
};

#pragma pack(pop, before_pack)

//				Classes
class GFilter;
class GSurface;

#include "GRegion.h"
#include "GFont.h"

class LgiClass GdcPt2 {
public:
	int x, y;

	GdcPt2(int Ix = 0, int Iy = 0)
	{
		x = Ix;
		y = Iy;
	}

	GdcPt2(const GdcPt2 &p)
	{
		x = p.x;
		y = p.y;
	}

	bool Inside(GRegion &r)
	{
		return	(x >= r.x1) AND
			(x <= r.x2) AND
			(y >= r.y1) AND
			(y <= r.y2);
	}
};

class LgiClass GdcPt3 {
public:
	int x, y, z;
};

#pragma pack(push, before_pack)
#pragma pack(1)

class LgiClass GdcRGB {
public:
	uchar R, G, B, Flags;
	void Set(uchar r, uchar g, uchar b, uchar f = 0) { R = r; G = g; B = b; Flags = f; }
};

#pragma pack(pop, before_pack)

class LgiClass GdcPalette {
protected:
	HPALETTE	hPal;
	LOGPALETTE	*Data;

public:
	GdcPalette();
	virtual ~GdcPalette();
	
	HPALETTE Handle() { return hPal; }

	GdcPalette(GdcPalette *pPal);	
	GdcPalette(uchar *pPal, int s = 256);
	void Set(GdcPalette *pPal);
	void Set(uchar *pPal, int s = 256);

	int GetSize() { return (Data) ? Data->palNumEntries : 0; }
	GdcRGB *operator [](int i) { return (i >= 0 AND i < GetSize() AND Data) ? (GdcRGB*) (Data->palPalEntry + i) : 0; }
	bool Update();
	bool SetSize(int s = 256);
	void SwapRAndB();
	int MatchRgb(COLOUR Rgb);
	void CreateCube();
	bool Load(GFile &F);
	bool Save(GFile &F, int Format);

	bool operator ==(GdcPalette &p);
	bool operator !=(GdcPalette &p);
};

class LgiClass GdcSurface {
public:
	uchar *Base;
	int x, y, Bits, Line;
	int Flags;

	GdcSurface();
	~GdcSurface();
};

class LgiClass GApplicator {
protected:
	GdcSurface *Dest;
	GdcSurface *Alpha;
	GdcPalette *Pal;
	int Op;

public:
	COLOUR c;				// main colour

	GApplicator() { c = 0; }
	GApplicator(COLOUR Colour) { c = Colour; }
	virtual ~GApplicator() { }

	virtual int OnEvent(int m, int a, int b) { return 0; }
	void SetOp(int o) { Op = o; }
	int GetOp() { return Op; }
	int GetBits() { return (Dest) ? Dest->Bits : 0; }
	int GetFlags() { return (Dest) ? Dest->Flags : 0; }
	GdcPalette *GetPal() { return Pal; }

	virtual bool SetSurface(GdcSurface *d, GdcPalette *p = 0, GdcSurface *a = 0) = 0; // sets Dest, returns FALSE on error
	virtual void SetPtr(int x, int y) = 0;	// calculates Ptr from x, y
	virtual void IncX() = 0;
	virtual void IncY() = 0;
	virtual void IncPtr(int X, int Y) = 0;

	virtual void Set() = 0;			// sets pixel at Ptr
	virtual COLOUR Get() = 0;		// get pixel at Ptr
	virtual void VLine(int height) = 0;	// vertical line from Ptr down
						// rectangle from Ptr down
	virtual void Rectangle(int x, int y) = 0;
						// blt from Src to Dest
	virtual bool Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha = 0) = 0;
};

class LgiClass GdcApplicatorServer {
public:
	GdcApplicatorServer() {}
	virtual ~GdcApplicatorServer() {}

	virtual int Op() { return 0; }
	virtual GApplicator *Create(int Bits) { return NULL; }
};

typedef List<GdcApplicatorServer> GdcApplicatorServerList;

#define OrgX(x)			x += OriginX
#define OrgY(y)			y += OriginY
#define OrgXy(x, y)		x += OriginX; y += OriginY
#define OrgPt(p)		p.x += OriginX; p.y += OriginY
#define OrgRgn(r)		r.x1 += OriginX; r.y1 += OriginY; r.x2 += OriginX; r.y2 += OriginY

class LgiClass GDeviceContext {

	friend class GFilter;
	friend class GDeviceContext;
	friend class GSurface;
	friend class GdcDibSection;

protected:
	int					Flags;
	int					PrevOp;
	GRegion			Clip;
	GdcSurface			*pMem;
	GSurface	*pAlphaDC;
	GdcPalette			*pPalette;
	GApplicator		*pApp;
	GApplicator		*pAppCache[GDC_CACHE_SIZE];

	#ifdef WIN32
	HRGN				hRgn;
	#endif

	#ifdef ADJUST_ORIGIN
	int					OriginX, OriginY;
	#endif

	// Protected functions
	GApplicator		*CreateApplicator(int Op = GDC_SET, int Bits = 0);
	bool				AddChild(GDeviceContext *c, int Pos = 0);
	void				RemoveFromTree();

public:
	GDeviceContext();
	virtual ~GDeviceContext();

	virtual bool Create(int x, int y, int Bits, int LineLen = 0, bool KeepData = FALSE) { return FALSE; }
	virtual void Update(int Flags) {}

	// Alpha channel
	virtual bool IsAlpha() { return pAlphaDC != 0; }
	virtual bool IsAlpha(bool b);
	bool DrawOnAlpha() { return ((Flags & GDC_DRAW_ON_ALPHA) != 0); }
	bool DrawOnAlpha(bool Draw);
	GSurface *AlphaDC() { return pAlphaDC; }

	// Applicator
	virtual bool Applicator(GApplicator *pApp);
	virtual GApplicator *Applicator();

	// Palette
	virtual GdcPalette *Palette();
	virtual void Palette(GdcPalette *pPal, bool bOwnIt = TRUE);

	// Clip region
	virtual GRegion ClipRgn(GRegion *Rgn);
	virtual GRegion ClipRgn();

	// Set properties
	virtual COLOUR Colour(COLOUR c, int Bits = 0);
	virtual int Op(int Op);

	// Get properties
	virtual COLOUR Colour() { return pApp->c; }
	virtual int Op() { return (pApp) ? pApp->GetOp() : 0; }
	virtual int X() { return (pMem) ? pMem->x : 0; }
	virtual int Y() { return (pMem) ? pMem->y : 0; }
	virtual int GetBits() { return (pMem) ? pMem->Bits : 0; }
	virtual int GetFlags() { return Flags; }
	virtual bool IsScreen() { return (pApp) ? pApp->GetFlags() & GDC_ON_SCREEN : 0; }
	virtual uchar *operator[](int y) { return (pMem) ? pMem->Base + (pMem->Line * y) : 0; }

	#ifdef ADJUST_ORIGIN
	// Origin
	virtual void GetOrigin(int &x, int &y) { x = -OriginX; y = -OriginY; }
	virtual void SetOrigin(int x, int y) { OriginX = -x; OriginY = -y; }
	#endif
};

class LgiClass GSurface : public GDeviceContext {
protected:
	uint		LineBits;
	uint		LineMask;

	HDC			hDC;
	HBITMAP		hBmp;

public:
	GSurface();
	GSurface(GSurface *pDC);

	// Win32
	#ifdef WIN32

	virtual HDC StartDC() { return hDC; }
	virtual void EndDC() {}
	HBITMAP GetBitmap() { return hBmp; }
	HDC Handle() { return hDC; }

	#endif

	// Pixel
	virtual void Set(int x, int y);
	virtual COLOUR Get(int x, int y);

	// Line
	virtual void HLine(int x1, int x2, int y);
	virtual void VLine(int x, int y1, int y2);
	virtual void Line(int x1, int y1, int x2, int y2);
	
	virtual uint LineStyle(uint Bits) { uint B = LineBits; LineBits = Bits; return B; }
	virtual uint LineStyle() { return LineBits; }

	// Curve
	virtual void Circle(double cx, double cy, double radius);
	virtual void FilledCircle(double cx, double cy, double radius);
	virtual void Arc(double cx, double cy, double radius, double start, double end);
	virtual void FilledArc(double cx, double cy, double radius, double start, double end);
	virtual void Ellipse(double cx, double cy, double x, double y);
	virtual void FilledEllipse(double cx, double cy, double x, double y);

	// Rectangular
	virtual void Box(int x1, int y1, int x2, int y2);
	virtual void Box(GRegion *a = NULL);
	virtual void Rectangle(int x1, int y1, int x2, int y2);
	virtual void Rectangle(GRegion *a = NULL);
	virtual void Blt(int x, int y, GSurface *Src, GRegion *a = NULL);
	virtual void StretchBlt(GRegion *d, GSurface *Src, GRegion *s);

	// Other
	virtual void Polygon(int Points, GdcPt2 *Data);
	virtual void Text(GdcFont *pFont, int x, int y, char *Str, int Len);
	virtual void Bezier(int Threshold, GdcPt2 *Pt);
	virtual void FloodFill(int x, int y, int Mode, COLOUR Border = 0, GRegion *Bounds = NULL);
};

class LgiClass GScreenDC : public GSurface
{
	bool		OwnDC;
	HWND		hWnd;
	PAINTSTRUCT	Ps;
	HPEN		hPen;
	HBRUSH		hBrush;
	HPEN		hNullPen;
	HBRUSH		hNullBrush;
	HBITMAP		hOldBitmap;

	int			Mode;
	COLOUR		Col;
	int			Sx, Sy;
	
	static GdcPalette *LastRealized;

	void Init();

public:
	GScreenDC();
	GScreenDC(HWND hwnd);
	GScreenDC(HDC hdc, HWND hwnd);
	GScreenDC(HBITMAP hBmp, int Sx, int Sy);
	virtual ~GScreenDC();

	bool Create(HDC hdc);

	// Properties
	void SetSize(int x, int y);
	#ifdef ADJUST_ORIGIN
	void GetOrigin(int &x, int &y);
	void SetOrigin(int x, int y);
	#endif
	bool Applicator(GApplicator *pApp);
	GApplicator *Applicator();
	GdcPalette *Palette();
	void Palette(GdcPalette *pPal, bool bOwnIt = TRUE);
	GRegion ClipRgn(GRegion *Rgn);
	GRegion ClipRgn();
	COLOUR Colour(COLOUR c, int Bits = 0);
	int Op(int Op);
	COLOUR Colour();
	int Op();
	int X();
	int Y();
	int GetBits();
	bool IsScreen();
	uchar *operator[](int y) { return NULL; }
	uint LineStyle(uint Bits);
	uint LineStyle();

	// Primitives
	void Set(int x, int y);
	COLOUR Get(int x, int y);
	void HLine(int x1, int x2, int y);
	void VLine(int x, int y1, int y2);
	void Line(int x1, int y1, int x2, int y2);
	void Circle(double cx, double cy, double radius);
	void FilledCircle(double cx, double cy, double radius);
	void Arc(double cx, double cy, double radius, double start, double end);
	void FilledArc(double cx, double cy, double radius, double start, double end);
	void Ellipse(double cx, double cy, double x, double y);
	void FilledEllipse(double cx, double cy, double x, double y);
	void Box(int x1, int y1, int x2, int y2);
	void Box(GRegion *a);
	void Rectangle(int x1, int y1, int x2, int y2);
	void Rectangle(GRegion *a = NULL);
	void Blt(int x, int y, GSurface *Src, GRegion *a = NULL);
	void StretchBlt(GRegion *d, GSurface *Src, GRegion *s = NULL);
	void Polygon(int Points, int *Data);
	void Text(GdcFont *pFont, int x, int y, char *Str, int Len);
	void Bezier(int Threshold, GdcPt2 *Pt);
	void FloodFill(int x, int y, int Mode, COLOUR Border = 0);
};

class LgiClass GMemDC : public GSurface {
protected:
	void		*pBits;
	PBITMAPINFO	Info;
	HPALETTE	OldPal;

public:
	GMemDC();
	GMemDC(GSurface *pDC);
	virtual ~GMemDC();

	HDC StartDC();
	void EndDC();
	void Update(int Flags);
	GRegion ClipRgn(GRegion *Rgn);	// need this to tell the HDC about the
										// new clipping region
	GRegion ClipRgn() { return Clip; }


	#ifdef ADJUST_ORIGIN
	// Origin
	virtual void SetOrigin(int x, int y);
	#endif

	bool Create(int x, int y, int Bits, int LineLen = 0, bool KeepData = FALSE);
	void Blt(int x, int y, GSurface *Src, GRegion *a = NULL);
	void StretchBlt(GRegion *d, GSurface *Src, GRegion *s = NULL);

	void HLine(int x1, int x2, int y, COLOUR a, COLOUR b);
	void VLine(int x, int y1, int y2, COLOUR a, COLOUR b);
};

class LgiClass GSprite {
protected:
	// Dimensions
	int Sx, Sy, Bits;
	
	// Screen relation
	bool Visible;
	int PosX, PosY;
	int HotX, HotY;

	// Mode
	int DrawMode;			// See access functions for description

	// Destination DC, usually the screen
	// Don't own and must not be NULL
	GSurface *pScreen;
	
	// Internal sprite DC's
	// TODO: combine all these into one DC 3 times bigger
	GSurface *pBack;
	GSurface *pMask;
	GSurface *pColour;
	GSurface *pTemp;

	// This function gets called to update any sprite position dependent
	// graphics when the sprite moves.
	virtual void DrawOnMove(	GSurface *pDC,
					int Nx, int Ny,
					int Ox, int Oy) {}

public:
	GSprite();
	virtual ~GSprite();

	int X() { return Sx; }
	int Y() { return Sy; }
	int GetPosX() { return PosX; }
	int GetPosY() { return PosY; }
	int GetHotX() { return HotX; }
	int GetHotY() { return HotY; }
	int GetBits() { return Bits; }
	bool IsVisible() { return Visible; }

	void SetXY(int x, int y);

	bool Create(GSurface *pScreen, int x, int y, int Bits, uchar *Colour, uchar *Mask);
	bool CreateKey(GSurface *pScreen, int x, int y, int Bits, uchar *Colour, int Key);
	bool Create(GSurface *pScreen, GSprite *pSpr);
	bool SetSize(int x, int y, int Bits, GSurface *pScr = NULL);
	void Delete();

	int Mode() { return DrawMode; }
	int Mode(int m) { int Old = DrawMode; DrawMode = m; return Old; }
	// Mode:
	//	0 - Transparent background, AND mask, XOR colour
	//	1 - Transparent background, AND mask, OR colour
	//	3 - COPY colour

	void SetHotPoint(int x, int y);
	void SetVisible(bool v);
	void Move(int x, int y, bool WaitRetrace);

	void Draw(int x, int y, COLOUR Back, int Mode = 0, GSurface *pDC = NULL);
	// Mode:
	//	0 - Solid background, AND colour, XOR mask
	//	1 - Solid background, AND mask, OR colour
	//	2 - Transparent background, AND mask, OR colour
	//	3 - COPY colour

	void MoveToScreen(GSurface *Scr)
	{
		bool v = Visible;
		SetVisible(FALSE);
		pScreen = Scr;
		SetVisible(v);
	}
	GSurface *GetMask() { return pMask; }
	GSurface *GetColour() { return pColour; }
};

class LgiClass GdcDevice
{
	static GdcDevice *pInstance;

protected:
	bool PrevVerbose;

	// Lookups
	ulong *CharSquareData;

	// Palette
	double GammaCorrection;
	uchar GammaTable[256];
	GdcPalette *pSysPal;

	// Current mode info
	int ScrX;
	int ScrY;
	int ScrBits;

	// Options
	int OptVal[GDC_MAX_OPTION];

	// Applicator server
	GdcApplicatorServerList AppServerList;

public:
	GdcDevice();
	~GdcDevice();
	static GdcDevice *GetInst() { return pInstance; }

	// Mode
	int GetBits();
	int X();
	int Y();

	// Options
	int GetOption(int Opt) { return (Opt >= 0 AND Opt < GDC_MAX_OPTION) ? OptVal[Opt] : 0; }
	int SetOption(int Opt, int Value)
	{
		int Prev = OptVal[Opt];
		if (Opt >= 0 AND Opt < GDC_MAX_OPTION)
			OptVal[Opt] = Value;
		return Prev;
	}

	// Lookups
	ulong *GetCharSquares() { return (CharSquareData) ? CharSquareData + 255 : 0; }

	// Applicators
	bool AddAppServer(GdcApplicatorServer *pServer);
	GdcApplicatorServer *GetAppServer(int Op);

	// Palette/Colour
	void SetGamma(double Gamma);
	double GetGamma();

	void SetSystemPalette(int Start, int Size, GdcPalette *Pal);
	GdcPalette *GetSystemPalette();

	void SetColourPaletteType(int Type);	// Type = PALTYPE_xxx define
	COLOUR GetColour(COLOUR Rgb24, GDeviceContext *pDC = NULL);
};

// file filter support
#include "GFilter.h"

// applicators
#include "GApplicator.h"

// globals
#define GdcD			GdcDevice::GetInst()
LgiFunc GSurface *ConvertDC(GSurface *pDC, int Bits);
LgiFunc GSurface *LoadDC(char *Name);
LgiFunc bool WriteDC(char *Name, GSurface *pDC);
LgiFunc COLOUR CBit(int DstBits, COLOUR c, int SrcBits = 24, GdcPalette *Pal = 0);
LgiFunc COLOUR GdcMixColour(COLOUR a, COLOUR b, double HowMuchA = 0.5);

#endif