#include <time.h>
#include "Lgi.h"

/////////////////////////////////////////////////////////////////////////////
extern int MouseRollMsg;
extern HHOOK hMouseHook;
extern HWND hMouseOver;
extern LRESULT CALLBACK MouseHook(int nCode, WPARAM wParam, LPARAM lParam);

GApp *GApp::TheApp = 0;

GApp::GApp(HINSTANCE hInst, char *lpCmdLine, int nShow)
{
	// Vars...
	Config = 0;
	AppWnd = 0;
	CmdShow = nShow;
	hAppInst = hInst;
	CmdLine = NewStr(lpCmdLine);

	// GApp instance
	LgiAssert(TheApp == 0);
	TheApp = this;

	// System font
	SystemFont = 0;
	GdcFontType SysFontType;
	if (SysFontType.GetSystemFont("System"))
	{
		SystemFont = SysFontType.CreateFont();
		if (SystemFont)
		{
			SystemFont->Transparent(true);
		}
	}

	if (NOT SystemFont)
	{
		LgiMsg(0, "Error: Couldn't create system font.", "Lgi Error: GApp::GApp", MB_OK);
		LgiExitApp();
	}

	// Other vars and init
	hNormalCursor = LoadCursor(NULL, IDC_ARROW);
	srand(((unsigned)time(NULL))*((uint)GetCurrentThread()));
	MouseRollMsg = RegisterWindowMessage("MSWHEEL_ROLLMSG");
	hMouseHook = SetWindowsHookEx(WH_MOUSE, (HOOKPROC) MouseHook, NULL, (DWORD) GetCurrentThreadId());
	LgiInitColours();
}

GApp::~GApp()
{
	DeleteArray(CmdLine);
	DeleteObj(AppWnd);	
	DeleteObj(SystemFont);
	if (hMouseHook)
	{
		UnhookWindowsHookEx(hMouseHook);
	}
	TheApp = 0;
}

XmlTag *GApp::GetConfig(char *Tag)
{
	if (NOT Config)
	{
		char Path[256];
		if (LgiGetExePath(Path))
		{
			if (Path[strlen(Path)-1] != DIR_CHAR) strcat(Path, DIR_STR);
			strcat(Path, "lgi.conf");
			if (FileExists(Path))
			{
				Config = NEW(Xml);
				if (Config)
				{
					if (Config->ParseXmlFile(Path) <= 0)
					{
						DeleteObj(Config);
					}
				}
			}
		}
	}

	if (Tag AND Config)
	{
		for (XmlTag *t=Config->Values.First(); t; t=Config->Values.Next())
		{
			if (t->Name AND
				stricmp(t->Name, Tag) == 0)
			{
				return t;
			}
		}
	}

	return 0;
}

bool GApp::GetOption(char *Dest, char Option)
{
	char WhiteSpace[] = " \t\r\n";
	char *c = CmdLine;

	while (c AND *c)
	{
		if (*c == '/' OR *c == '-')
		{
			c++;
			if (*c == Option)
			{
				c++;
				if (*c)
				{
					// skip leading whitespace
					while (*c AND strchr(WhiteSpace, *c))
					{
						c++;
					}

					// write str out if they want it
					if (Dest)
					{
						char End = (*c == '\'') ? *c++ : ' ';

						while (*c AND *c != End)
						{
							*Dest++ = *c++;
						}

						*Dest++ = 0;
					}
				}

				// yeah we got the option
				return true;
			}
		}

		c = strchr(c, ' ');
		if (c) c++;
	}

	return false;
}

#include "ginput.h"

void GApp::OnCmdLine(char *Cmd)
{
	char WhiteSpace[] = " \r\n\t";

	DeleteArray(CmdLine);
	CmdLine = NewStr(Cmd);

	/*
	GInput dlg(0, CmdLine);
	dlg.DoModal();
	*/
	
	/*
	if (Cmd AND strlen(Cmd) > 0)
	{
		List<char> l;

		// search for args
		for (char *c = Cmd; *c; )
		{
			// skip white
			while (strchr(WhiteSpace, *c)) c++;
			if (*c)
			{
				char End = (*c == '\"') ? '\"' : ' ';
				if (End == '\"') c++;
				char *Start = c;
				while (*c AND *c != End) c++;
				char *Str = NewStr(Start, (int)c-(int)Start);
				if (Str) l.Insert(Str);
				if (End == '\"') c++;
			}
		}

		if (l.GetItems() > 0)
		{
			// read args into an array
			char **Arg = NEW(char*[l.GetItems()]);
			if (Arg)
			{
				int n=0;
				for (char *a = l.First(); a; a = l.Next())
				{
					Arg[n++] = a;
				}

				// call the client
				OnArgs(l.GetItems(), Arg);

				// clean up
				for (n=0; n<l.GetItems(); n++)
				{
					DeleteArray(Arg[n]);
				}
				DeleteArray(Arg);
			}
		}
	}
	*/

	if (CmdLine)
	{
		List<char> Files;
		
		/*
		char File[256] = "";

		for (int i=0; i<Args; i++)
		{
			if (strlen(File) == 0)
			{
				strcpy(File, Arg[i]);
				if (FileExists(File) OR
					DirExists(File))
				{
					Files.Insert(NewStr(File));
					File[0] = 0;
				}
			}
			else
			{
				strcat(File, " ");
				strcat(File, Arg[i]);
				if (FileExists(File))
				{
					Files.Insert(NewStr(File));
					File[0] = 0;
				}
			}
		}
		*/

		char *Delim = "\'\"";
		char *s;
		for (s = CmdLine; *s; )
		{
			// skip ws
			while (strchr(WhiteSpace, *s)) s++;

			// read to end of token
			char *e = s;
			if (strchr(Delim, *s))
			{
				for (e = ++s; *e AND NOT strchr(Delim, *e); e++);
			}
			else
			{
				for (; *e AND NOT strchr(WhiteSpace, *e); e++);
			}

			char *Arg = NewStr(s, (int)e-(int)s);
			if (Arg)
			{
				if (FileExists(Arg))
				{
					Files.Insert(Arg);
				}
				else
				{
					DeleteArray(Arg);
				}
			}

			// next
			s = (*e) ? e + 1 : e;
		}

		// call app
		if (Files.GetItems() > 0)
		{
			OnReceiveFiles(Files);
		}

		// clear up
		DeleteArrayList(&Files, char);
	}
}

void GApp::OnReceiveFiles(List<char> &Files)
{
	if (AppWnd)
	{
		AppWnd->OnReceiveFiles(Files);
	}
}

bool IsTabStop(HWND hWnd)
{
	int Style = GetWindowLong(hWnd, GWL_STYLE);
	return ((Style & WS_TABSTOP) != 0) AND ((Style & WS_VISIBLE) != 0);
}

HWND GetTabControl(HWND hWnd)
{
	// Hack for selecting TabControls first even though they
	// are last in the list
	HWND hTab = GetWindow(hWnd, GW_HWNDLAST);

	// Ok we're not the first
	// Get the class.
	char ClassName[256];
	if (GetClassName(hTab, ClassName, sizeof(ClassName)) > 0)
	{
		if (stricmp(ClassName, LGI_TABCONTROL) == 0)
		{
			return hTab;
		}
	}

	return NULL;
}


HWND GetTabstop(HWND hWnd, bool Forward = true)
{
	HWND h = GetWindow(hWnd, (Forward) ? GW_HWNDNEXT : GW_HWNDPREV);
	while (h)
	{
		if (h == hWnd)
		{
			return NULL;
		}

		if (IsTabStop(h))
		{
			// Is a tabstop...
			return h;
		}

		h = GetWindow(h, (Forward) ? GW_HWNDNEXT : GW_HWNDPREV);
	}

	return NULL;
}

/*
if (hTab == hWnd)
{
	if (Shifted)
	{
		hNext = 0;
	}
	else
	{
		hNext = GetWindow(hWnd, GW_CHILD);
		if (NOT IsTabStop(hNext))
		{
			hNext = GetTabstop(hNext, NOT Shifted);
		}
	}
}
else

BOOL ProcessDialogMessage(HWND hDlg, MSG *Msg)
{
	if (Msg->message == WM_KEYDOWN AND
		Msg->wParam == VK_TAB AND
		NOT (Msg->lParam & 0x80000000))
	{
		bool Shifted = (GetAsyncKeyState(VK_SHIFT) & 0x8000) != 0;
		HWND hWnd = GetFocus();
		if (hWnd)
		{
			int DlgCode = SendMessage(hWnd, WM_GETDLGCODE, 0, 0);
			if (NOT (DlgCode & (DLGC_WANTTAB | DLGC_WANTALLKEYS)))
			{
				// Doesn't want tab's so we process the tab
				// HWND hTab = GetTabControl(hWnd);
				HWND hNext = 0;

				// step down to child level
				hNext = GetWindow(hWnd, GW_CHILD);
				if (hNext)
				{
					if (Shifted)
					{
						hNext = GetWindow(hNext, GW_HWNDLAST);
					}

					if (NOT IsTabStop(hNext))
					{
						hNext = GetTabstop(hNext, NOT Shifted);
					}
				}

				if (NOT hNext)
				{
					// no children... so along
					hNext = GetTabstop(hWnd, NOT Shifted);
				}

				if (hNext)
				{
					SetFocus(hNext);
				}
				else
				{
					// end of list
					
					// go up till we find a "Next"
					bool ParentExit = false;
					HWND hParent = GetParent(hWnd);
					while (	hParent AND
							NOT (hNext = GetWindow(hParent, (Shifted) ? GW_HWNDPREV : GW_HWNDNEXT)) AND
							GetParent(hParent))
					{
						HWND p = GetParent(hParent);
						if (GetParent(p))
						{
							hParent = p;
						}
						else
						{
							ParentExit = true;
							break;
						}
					}

					if (ParentExit AND NOT hNext)
					{
						hNext = GetWindow(hParent, (Shifted) ? GW_HWNDLAST : GW_HWNDFIRST);
					}

					while (hNext)
					{
						// search through all the children for tabstops
						HWND hChild = GetWindow(hNext, GW_CHILD);
						if (hChild)
						{
							hChild = GetWindow(hChild, (Shifted) ? GW_HWNDLAST : GW_HWNDFIRST);
							HWND hTab = GetTabControl(hChild);
							if (hTab)
							{
								hChild = hTab;
							}
							else if (NOT IsTabStop(hChild))
							{
								hChild = GetTabstop(hChild, NOT Shifted);
							}
						}
						else
						{
							if (IsTabStop(hNext))
							{
								hChild = hNext;
							}
							else
							{
								hChild = GetTabstop(hNext, NOT Shifted);
							}
						}

						if (hChild)
						{
							SetFocus(hChild);
							return 0;
						}

						HWND n = GetWindow(hNext, (Shifted) ? GW_HWNDPREV : GW_HWNDNEXT);
						if (NOT n)
						{
							hNext = GetWindow(hNext, (Shifted) ? GW_HWNDLAST : GW_HWNDFIRST);
						}
						else
						{
							hNext = n;
						}
					}
				}
				return true;
			}
		}
	}

	return IsDialogMessage(hDlg, Msg);
}
*/

/*
BOOL ProcessDialogMessage(HWND hDlg, MSG *Msg)
{
	switch (Msg->message)
	{
		case WM_KEYDOWN:
		{
			if (Msg->wParam == VK_RETURN)
			{
				int i = SendMessage(Msg->hwnd, WM_GETDLGCODE, 0, 0);
				if (i & DLGC_WANTALLKEYS == 0)
				{
					int DefId = SendMessage(hDlg, DM_GETDEFID, 0, 0);
					Msg->hwnd = GetDlgItem(hDlg, DefId);

					TranslateMessage(Msg);
					DispatchMessage(Msg);
					return true;
				}
			}
			break;
		}
	}

	return IsDialogMessage(hDlg, Msg);
}
*/

HWND GApp::ActiveDialog = NULL;
bool GApp::Run(bool Loop)
{
	MSG Msg;
	bool status = true;

	if (Loop)
	{
		while (GetMessage(&Msg, NULL, 0, 0) > 0)
		{
			/*
			if (ActiveDialog)
			{
				// if (ProcessDialogMessage(ActiveDialog, &Msg))
				if (IsDialogMessage(ActiveDialog, &Msg))
				{
					continue;
				}
			}
			*/

			if (Msg.message == WM_KEYDOWN)
			{
				HWND hTop = Msg.hwnd, hNext;
				while (hNext = GetParent(hTop))
				{
					hTop = hNext;
				}

				if (hTop)
				{
					SendMessage(hTop, WM_TRANSLATE_ACCELERATOR, 0, (LPARAM) &Msg);
				}
			}

			TranslateMessage(&Msg);
			DispatchMessage(&Msg);
		}
	}
	else
	{
		while (	PeekMessage(&Msg, NULL, 0, 0, PM_REMOVE) AND
				Msg.message != WM_QUIT)
		{
			/*
			if (ActiveDialog)
			{
				// if (ProcessDialogMessage(ActiveDialog, &Msg))
				if (IsDialogMessage(ActiveDialog, &Msg))
				{
					continue;
				}
			}
			*/

			if (Msg.message == WM_KEYDOWN)
			{
				HWND hTop = Msg.hwnd, hNext;
				while (hNext = GetParent(hTop))
				{
					hTop = hNext;
				}
				if (hTop)
				{
					SendMessage(hTop, WM_TRANSLATE_ACCELERATOR, 0, (LPARAM) &Msg);
				}
			}

			TranslateMessage(&Msg);
			DispatchMessage(&Msg);
		}
	}

	return Msg.message != WM_QUIT;
}

void GApp::Exit(int Code)
{
	if (Code)
	{
		exit(0);
	}
	else
	{
		PostQuitMessage(Code);
	}
}

