/*hdr
**      FILE:           GFileSelect.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           26/9/2001
**      DESCRIPTION:    Common file/directory selection dialog
**
**      Copyright (C) 1998-2001, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
GFileSelect::GFileSelect()
{
	Size = 4096;
	FileStr = NEW(char[Size]);
	if (FileStr)
	{
		*FileStr = 0;
	}
	else
	{
		Size = 0;
	}

	ZeroObj(Info);
	Info.lStructSize = sizeof(Info);
	Info.hInstance = hAppInst();
}

GFileSelect::~GFileSelect()
{
	DeleteArray(FileStr);
	ClearTypes();
}

char *GFileSelect::TypeStr()
{
	char *Str = NEW(char[4096]);
	if (Str)
	{
		char *pstr = Str;

		for (GFileType *Type = TypeList.First(); Type; Type = TypeList.Next())
		{
			strcpy(pstr, Type->Description());
			pstr += strlen(Type->Description()) + 1;
			strcpy(pstr, Type->Extension());
			pstr += strlen(Type->Extension()) + 1;
		}

		*pstr = 0;
		Info.lpstrFilter = Str;
	}

	return Str;
}

char *GFileSelect::Name()
{
	return FileStr;
}

bool GFileSelect::Name(char *n)
{
	bool Status = FALSE;

	if (n AND FileStr)
	{
		strncpy(FileStr, n, Size);
		Info.lpstrFile = FileStr;
		Info.nMaxFile = Size;
	}

	return Status;
}

char *GFileSelect::operator [](int i)
{
	if (Info.nFileOffset > strlen(FileStr))
	{
		// multi-select
		char *File = FileStr + Info.nFileOffset;
		for (int n=0; n<i AND *File; n++)
		{
			File = File + strlen(File) + 1;
		}

		static char Buf[256];
		sprintf(Buf, "%s%s%s", FileStr, DIR_STR, File);
		return Buf;
	}
	else
	{
		// single select
		return FileStr;
	}
}

int GFileSelect::Names()
{
	if (Info.nFileOffset > strlen(FileStr))
	{
		// multi-select
		char *File = FileStr + Info.nFileOffset;
		int n;
		for (n = 0; *File; n++)
		{
			File = File + strlen(File) + 1;
		}

		return n;
	}
	else
	{
		// single select
		return 1;
	}
}

char *GFileSelect::File()
{
	return FileStr + Info.nFileOffset;
}

char *GFileSelect::Extension()
{
	return FileStr + Info.nFileExtension;
}

int GFileSelect::Types()
{
	return TypeList.GetItems();
}

int GFileSelect::SelectedType()
{
	return Info.nFilterIndex - 1;
}

GFileType *GFileSelect::TypeAt(int n)
{
	return TypeList.ItemAt(n);
}

void GFileSelect::ClearTypes()
{
	DeleteObjList(&TypeList, GFileType);
}

bool GFileSelect::Type(char *Description, char *Extension, int Data)
{
	GFileType *Type = NEW(GFileType);
	if (Type)
	{
		Type->Description(Description);
		Type->Extension(Extension);
		Type->Data(Data);
		TypeList.Insert(Type);
	}

	return Type != 0;
}

GView *GFileSelect::Parent()
{
	return ParentWnd;
}

void GFileSelect::Parent(GView *Window)
{
	ParentWnd = Window;
	if (ParentWnd)
	{
		Info.hwndOwner = Window->Handle();
	}
}

bool GFileSelect::MultiSelect()
{
	return (Info.Flags & OFN_ALLOWMULTISELECT) != 0;
}

void GFileSelect::MultiSelect(bool Multi)
{
	if (Multi)
	{
		Info.Flags |= OFN_ALLOWMULTISELECT;
	}
	else
	{
		Info.Flags &= ~OFN_ALLOWMULTISELECT;
	}

	Info.Flags |= OFN_EXPLORER;
}

char *GFileSelect::InitialDir()
{
	return (char*) Info.lpstrInitialDir;
}

void GFileSelect::InitialDir(char *InitDir)
{
	Info.lpstrInitialDir = InitDir;
}

char *GFileSelect::Title()
{
	return (char*) Info.lpstrTitle;
}

void GFileSelect::Title(char *Title)
{
	Info.lpstrTitle = Title;
}

char *GFileSelect::DefaultExtension()
{
	return (char*) Info.lpstrDefExt;
}

void GFileSelect::DefaultExtension(char *DefExt)
{
	Info.lpstrDefExt = DefExt;
}

bool GFileSelect::Open()
{
	bool Status = FALSE;

	char *Str = TypeStr();
	if (Str)
	{
		Info.lpstrFilter = Str;		
		Info.lpstrFile = FileStr;
		Info.nMaxFile = Size;

		Status = GetOpenFileName(&Info);

		DeleteObj(Str);
	}

	return Status AND strlen(FileStr) > 0;
}

#include "shlobj.h"

char *_GFileSelect_DefaultFolder = 0;
int CALLBACK GFileSelectBrowseCallback(HWND hwnd, UINT uMsg, LPARAM lParam, LPARAM lpData)
{
	if (uMsg == BFFM_INITIALIZED AND
		_GFileSelect_DefaultFolder)
	{
		PostMessage(hwnd, BFFM_SETSELECTION, true, (uint)_GFileSelect_DefaultFolder);
	}

	return 0;
}

bool GFileSelect::OpenFolder()
{
	LPMALLOC pMalloc; // Gets the Shell's default allocator
	bool Status = false;
	_GFileSelect_DefaultFolder = Name();

	if  (FileStr AND ::SHGetMalloc(&pMalloc) == NOERROR)
	{
		BROWSEINFO      bi;
		char            pszBuffer[MAX_PATH];
		LPITEMIDLIST    pidl;

		ZeroObj(bi);

		/* Set the root folder?
		if (Name() AND strlen(Name()))
		{
			OLECHAR			olepath[256];
			ulong			cheaten;
			ulong			dwattributes;
			HRESULT			hr;
			LPSHELLFOLDER	pdesktopfolder;

			// // get a pointer to the desktop's ishellfolder interface. //
			if (SUCCEEDED(SHGetDesktopFolder(&pdesktopfolder)))
			{
				// ishellfolder::parsedisplayname requires the file name be in unicode.
				MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, Name(), -1, olepath, CountOf(olepath));

				// convert the path to an itemidlist.
				hr = pdesktopfolder->ParseDisplayName(	0,
														0,
														olepath,
														&cheaten,
														&pidl,
														&dwattributes);
				bi.pidlRoot = pidl;
			}
		}
		*/

		// Get help on BROWSEINFO struct - it's got all the bit settings.
		bi.hwndOwner        = (ParentWnd) ? ParentWnd->Handle() : 0;
		bi.pszDisplayName   = pszBuffer;
		bi.lpszTitle        = "Select a target Directory";
		bi.ulFlags          = BIF_RETURNFSANCESTORS | BIF_RETURNONLYFSDIRS;
		bi.lpfn				= GFileSelectBrowseCallback;

		// This next call issues the dialog box.
		if  (pidl = ::SHBrowseForFolder(&bi))
		{
			if  (::SHGetPathFromIDList(pidl, pszBuffer))
			{
				// At this point pszBuffer contains the selected path
				strcpy(FileStr, pszBuffer);
				Status = true;
			}

			// Free the PIDL allocated by SHBrowseForFolder.
			pMalloc->Free(pidl);
		}

		// Release the shell's allocator
		pMalloc->Release();
	}

	return Status;
}

bool GFileSelect::Save()
{
	bool Status = FALSE;

	char *Str = TypeStr();
	if (Str)
	{
		Info.lpstrFilter = Str;		
		Info.lpstrFile = FileStr;
		Info.nMaxFile = Size;

		Status = GetSaveFileName(&Info);

		DeleteObj(Str);
	}

	return Status AND strlen(FileStr) > 0;
}
