/*hdr
**      FILE:           GuiCommon.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           18/7/98
**      DESCRIPTION:    Common windows and controls
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

///////////////////////////////////////////////////////////////////////////////////////////////
GFindReplace::GFindReplace(GView *parent)
{
	Parent = parent;
	FindOp = true;
	FindMsgID = RegisterWindowMessage(FINDMSGSTRING);
	Find[0] = 0;
	Replace[0] = 0;
	hDlg = 0;

	ZeroObj(FindInfo);
	FindInfo.lStructSize = sizeof(FindInfo);
	FindInfo.hInstance = hAppInst();
	FindInfo.Flags = FR_HIDEUPDOWN;
	FindInfo.hwndOwner = Parent ? Parent->Handle() : 0;
	FindInfo.lpstrFindWhat = Find;
	FindInfo.lpstrReplaceWith = Replace;
	FindInfo.wFindWhatLen = sizeof(Find);
	FindInfo.wReplaceWithLen = sizeof(Replace);
}

GFindReplace::~GFindReplace()
{
	if (hDlg)
	{
		DestroyWindow(hDlg);
		hDlg = 0;
	}
}

void GFindReplace::SetParent(GView *parent)
{
	Parent = parent;
}

bool GFindReplace::OpenFind()
{
	FindOp = true;
	FindInfo.hwndOwner = Parent ? Parent->Handle() : 0;
	hDlg = FindText(&FindInfo);
	if (hDlg)
	{
		MSG Msg;
		while (GetMessage(&Msg, NULL, 0, 0) > 0 AND IsWindow(hDlg))
		{
			if (NOT hDlg OR NOT IsDialogMessage(hDlg, &Msg))
			{
				TranslateMessage(&Msg);
				DispatchMessage(&Msg);
			}
		}

		return TRUE;
	}
	else
	{
		DWORD e = CommDlgExtendedError();
		int k=0;
	}

	return FALSE;
}

bool GFindReplace::OpenReplace()
{
	FindOp = FALSE;
	hDlg = ReplaceText(&FindInfo);
	if (hDlg)
	{
		MSG Msg;
		while (GetMessage(&Msg, NULL, 0, 0) > 0 AND IsWindow(hDlg))
		{
			if (NOT hDlg OR NOT IsDialogMessage(hDlg, &Msg))
			{
				TranslateMessage(&Msg);
				DispatchMessage(&Msg);
			}
		}

		return TRUE;
	}

	return FALSE;
}

int GFindReplace::OnEvent(GMessage *Msg)
{
	int Status = FALSE;

	if (Msg->Msg == FindMsgID)
	{
		if (FindOp)
		{
			Status = OnFind();
		}
		else
		{
			Status = OnReplace();
		}
	}

	return Status;
}

bool GFindReplace::OnFind()
{
	DestroyWindow(hDlg);
	hDlg = 0;
	return TRUE;
}

bool GFindReplace::OnReplace()
{
	DestroyWindow(hDlg);
	hDlg = 0;
	return TRUE;
}

bool GFindReplace::CaseSensitive()
{
	return (FindInfo.Flags & FR_MATCHCASE) != 0;
}

void GFindReplace::CaseSensitive(bool i)
{
	if (i)
	{
		FindInfo.Flags |= FR_MATCHCASE;
	}
	else
	{
		FindInfo.Flags &= ~FR_MATCHCASE;
	}
}

bool GFindReplace::WholeWord()
{
	return (FindInfo.Flags & FR_WHOLEWORD) != 0;
}

void GFindReplace::WholeWord(bool i)
{
	if (i)
	{
		FindInfo.Flags |= FR_WHOLEWORD;
	}
	else
	{
		FindInfo.Flags &= ~FR_WHOLEWORD;
	}
}

bool GFindReplace::BtnFindNext()
{
	return (FindInfo.Flags & FR_FINDNEXT) != 0;
}

bool GFindReplace::BtnReplace()
{
	return (FindInfo.Flags & FR_REPLACE) != 0;
}

bool GFindReplace::BtnReplaceAll()
{
	return (FindInfo.Flags & FR_REPLACEALL) != 0;
}

