// Win32 Implementation of General LGI functions
#include <stdio.h>

#include "Lgi.h"

////////////////////////////////////////////////////////////////
// Local helper functions
bool _lgi_check_file(char *Path)
{
	if (Path)
	{
		if (FileExists(Path))
		{
			// file is there
			return true;
		}
		else
		{
			// shortcut?
			char *e = Path + strlen(Path);
			strcpy(e, ".lnk");
			if (FileExists(Path))
			{
				// resolve shortcut
				char Link[256];
				if (ResolveShortcut(Path, Link, sizeof(Link)))
				{
					// check destination of link
					if (FileExists(Link))
					{
						strcpy(Path, Link);
						return true;
					}
				}
			}
			*e = 0;
		}
	}

	return false;
}

////////////////////////////////////////////////////////////////
// Implementations
bool LgiGetFileMimeType(char *File, char *Mime)
{
	bool Status = false;

	if (File AND Mime)
	{
		char *Dot = strrchr(File, '.');
		if (Dot)
		{
			HKEY hClass;
			if (RegOpenKey(HKEY_CLASSES_ROOT, Dot, &hClass) == ERROR_SUCCESS)
			{
				char Str[256];
				DWORD StrLen = sizeof(Str), Type = 0;
				if (RegQueryValueEx(hClass, "Content Type", 0, &Type, (uchar*)Str, &StrLen) == ERROR_SUCCESS)
				{
					strcpy(Mime, Str);
					Status = true;
				}

				RegCloseKey(hClass);
			}
		}

		if (NOT Status)
		{
			// no extension?
			// no registry entry for file type?
			strcpy(Mime, "application/octet-stream");
			Status = true;
		}
	}

	return Status;
}

bool LgiGetAppForMimeType(char *Mime, char *AppPath)
{
	bool Status = false;

	if (AppPath AND Mime)
	{
		if (stricmp(Mime, "application/email") == 0)
		{
			// get email app
			HKEY hKey;
			if (RegOpenKey(HKEY_CLASSES_ROOT, "mailto\\shell\\open\\command", &hKey) == ERROR_SUCCESS)
			{
				char Str[256];
				DWORD StrLen = sizeof(Str), Type = 0;

				// get app path
				if (RegQueryValueEx(hKey, 0, 0, &Type, (uchar*)Str, &StrLen) == ERROR_SUCCESS)
				{
					strcpy(AppPath, Str);
					Status = true;
				}

				RegCloseKey(hKey);
			}
		}
		else
		{
			// get generic app
			HKEY hKey;
			char KeyName[256];
			sprintf(KeyName, "MIME\\Database\\Content Type\\%s", Mime);
			if (RegOpenKey(HKEY_CLASSES_ROOT, KeyName, &hKey) == ERROR_SUCCESS)
			{
				char Str[256];
				DWORD StrLen = sizeof(Str), Type = 0;

				// get extension
				if (RegQueryValueEx(hKey, "Extension", 0, &Type, (uchar*)Str, &StrLen) == ERROR_SUCCESS)
				{
					RegCloseKey(hKey);

					// get classes location
					if (RegOpenKey(HKEY_CLASSES_ROOT, Str, &hKey) == ERROR_SUCCESS)
					{
						StrLen = sizeof(Str);
						if (RegQueryValueEx(hKey, 0, 0, &Type, (uchar*)Str, &StrLen) == ERROR_SUCCESS)
						{
							RegCloseKey(hKey);
							
							// get app path
							strcat(Str, "\\shell\\open\\command");
							if (RegOpenKey(HKEY_CLASSES_ROOT, Str, &hKey) == ERROR_SUCCESS)
							{
								StrLen = sizeof(Str);
								if (RegQueryValueEx(hKey, 0, 0, &Type, (uchar*)Str, &StrLen) == ERROR_SUCCESS)
								{
									strcpy(AppPath, Str);
									Status = true;
								}

								RegCloseKey(hKey);
							}
						}
						else
						{
							RegCloseKey(hKey);
						}
					}
				}
				else
				{
					RegCloseKey(hKey);
				}
			}
		}

		char *RootVar = "%SystemRoot%";
		char *SysRoot = strstr(RootVar, AppPath);
		if (SysRoot)
		{
			// correct path for variables
			char Temp[256];
			GetWindowsDirectory(Temp, sizeof(Temp));
			if (Temp[strlen(Temp)-1] != DIR_CHAR) strcat(Temp, DIR_STR);
			char *End = AppPath + strlen(RootVar);
			strcat(Temp, (*End == DIR_CHAR) ? End + 1 : End);
			strcpy(AppPath, Temp);
		}
	}

	return Status;
}

int LgiRand(int i)
{
	return (rand() % i);
}

int LgiMsg(GView *Parent, char *Msg, char *Title, int Type, ...)
{
	int Result = 0;
	if (Msg)
	{
		char Buffer[1025];

		va_list Arg;
		va_start(Arg, Type);
		vsprintf(Buffer, Msg, Arg);
		va_end(Arg);

		Result = MessageBox((Parent)?Parent->Handle():0,
							Buffer,
							(Title)?Title:"Message",
							Type);
	}

	return Result;
}

bool LgiPlaySound(char *FileName, int Flags)
{
	bool Status = false;

	HMODULE hDll = LoadLibrary("winmm.dll");
	if (hDll)
	{
		typedef BOOL (__stdcall *Proc_sndPlaySound)(LPCSTR pszSound, UINT fuSound);
		Proc_sndPlaySound psndPlaySound = (Proc_sndPlaySound)GetProcAddress(hDll, "sndPlaySoundA");

		if (psndPlaySound)
		{
			if (LgiGetOs() == LGI_OS_WIN9X)
			{
				// async broken on 98?
				Flags = 0;
			}

			Status = psndPlaySound(FileName, Flags);
		}

		FreeLibrary(hDll);
	}

	return Status;
}

bool LgiExecute(char *File, char *Arguments, char *Dir)
{
	int Status = 0;
	
	if (File)
	{
		Status = (int) ShellExecute(NULL, "open", File, Arguments, Dir, 5);
	}
	
	return Status > 32;
}
