/*
**	FILE:		GuiViews.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		18/7/98
**	DESCRIPTION:	Standard Views
**
**	Copyright (C) 1998, Matthew Allen
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
/*
bool IsWindowThread(GView *w)
{
	return	w AND
			GetCurrentThreadId() == GetWindowThreadProcessId(w->Handle(), NULL);
}
*/

void HashRectangle(GView *Wnd, GRegion *r)
{
	if (r AND Wnd AND Wnd->Handle())
	{
		HDC hdc = GetDC(Wnd->Handle());
		if (hdc)
		{
			BITMAPINFO *bmp = (BITMAPINFO*) NEW(char[sizeof(BITMAPINFO)+sizeof(RGBQUAD)+32]);
			int Pat[8] = {0xAAAAAAAA, 0x55555555, 0xAAAAAAAA, 0x55555555, 0xAAAAAAAA, 0x55555555, 0xAAAAAAAA, 0x55555555};
			if (bmp)
			{
				bmp->bmiHeader.biSize = sizeof(bmp->bmiHeader);
				bmp->bmiHeader.biWidth = 8;
				bmp->bmiHeader.biHeight = 8;
				bmp->bmiHeader.biPlanes = 1;
				bmp->bmiHeader.biBitCount = 1;
				bmp->bmiHeader.biCompression = BI_RGB;
				bmp->bmiHeader.biSizeImage = 32;
				bmp->bmiHeader.biXPelsPerMeter = 3000;
				bmp->bmiHeader.biYPelsPerMeter = 3000;
				bmp->bmiHeader.biClrUsed = 0;
				bmp->bmiHeader.biClrImportant = 0;

				bmp->bmiColors[0].rgbRed = 255;
				bmp->bmiColors[0].rgbGreen = 255;
				bmp->bmiColors[0].rgbBlue = 255;
				bmp->bmiColors[1].rgbRed = 0;
				bmp->bmiColors[1].rgbGreen = 0;
				bmp->bmiColors[1].rgbBlue = 0;

				memcpy(bmp->bmiColors + 2, Pat, 32);
				
				LOGBRUSH brush;
				brush.lbStyle = BS_DIBPATTERNPT;
				brush.lbColor = DIB_RGB_COLORS;
				brush.lbHatch = (LONG) bmp;

				HBRUSH hbrush = CreateBrushIndirect(&brush);
				if (hbrush)
				{
					HPEN hpen = (HPEN) GetStockObject(NULL_PEN);
					hpen = (HPEN) SelectObject(hdc, hpen);
					hbrush = (HBRUSH) SelectObject(hdc, hbrush);

					SetTextColor(hdc, RGB(255, 255, 255));
					SetBkColor(hdc, RGB(0, 0, 0));
					SetROP2(hdc, R2_XORPEN);
					SetBkMode(hdc, TRANSPARENT);
					::Rectangle(hdc, r->x1, r->y1, r->x2+1, r->y2+1);

					hbrush = (HBRUSH) SelectObject(hdc, hbrush);
					hpen = (HPEN) SelectObject(hdc, hpen);
					DeleteObject(hbrush);
				}

				ReleaseDC(Wnd->Handle(), hdc);
				DeleteArray(bmp);
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
GLayout::GLayout()
{
	_SettingScrollBars = 0;
	_PourLargest = 0;
	VScroll = 0;
	HScroll = 0;

	// Class.Class.style |= CS_HREDRAW | CS_VREDRAW;
	Class.Create("GLayout");
	WndClass = Class.Name();
	WndStyle |= WS_CHILD | WS_CLIPSIBLINGS; // | WS_CLIPCHILDREN
	WndExStyle |= WS_EX_CONTROLPARENT;
}

GLayout::~GLayout()
{
	DeleteObj(VScroll);
	DeleteObj(HScroll);
}

void GLayout::GetScrollPos(int &x, int &y)
{
	if (HScroll)
	{
		x = HScroll->Pos();
	}
	else
	{
		x = 0;
	}

	if (VScroll)
	{
		y = VScroll->Pos();
	}
	else
	{
		y = 0;
	}
}

void GLayout::SetScrollPos(int x, int y)
{
	if (HScroll)
	{
		HScroll->SetPos(x);
	}

	if (VScroll)
	{
		VScroll->SetPos(y);
	}
}

bool GLayout::SetScrollBars(bool x, bool y)
{
	bool Status = FALSE;

	if (NOT _SettingScrollBars)
	{
		_SettingScrollBars = true;

		/*
		if (Handle())
		{
			WndStyle = GetWindowLong(Handle(), GWL_STYLE);
		}
		*/

		if (y)
		{
			if (NOT VScroll)
			{
				VScroll = NEW(GScrollBar);
				if (VScroll)
				{
					// WndStyle |= WS_VSCROLL;
					VScroll->SetVertical(TRUE);
					VScroll->SetParent(this);
					Status = TRUE;
				}
			}
		}
		else
		{
			if (VScroll)
			{
				// WndStyle &= ~WS_VSCROLL;
				DeleteObj(VScroll);
				Status = TRUE;
			}
		}
		
		if (x)
		{
			if (NOT HScroll)
			{
				HScroll = NEW(GScrollBar);
				if (HScroll)
				{
					// WndStyle |= WS_HSCROLL;
					HScroll->SetVertical(FALSE);
					HScroll->SetParent(this);
					Status = TRUE;
				}
			}
		}
		else
		{
			if (HScroll)
			{
				// WndStyle &= ~WS_HSCROLL;
				DeleteObj(HScroll);
				Status = TRUE;
			}
		}

		/*
		if (Handle())
		{
			SetWindowLong(Handle(), GWL_STYLE, WndStyle);
			SetWindowPos(	Handle(),
							0,
							Pos.x1, Pos.y1, Pos.X(), Pos.Y(),
							SWP_NOMOVE | SWP_NOZORDER | SWP_NOSIZE | SWP_FRAMECHANGED);
		}
		*/

		_SettingScrollBars = false;
	}

	return Status;
}

bool GLayout::GetPourLargest()
{
	return _PourLargest;
}

void GLayout::SetPourLargest(bool i)
{
	_PourLargest = i;
}

bool GLayout::Pour(GComplexRegion &r)
{
	if (_PourLargest)
	{
		GRegion *Best = FindLargest(r);
		if (Best)
		{
			SetPos(*Best);
			return true;
		}
	}
	return false;
}

int GLayout::OnEvent(GMessage *Msg)
{
	if (MsgCode(Msg) == WM_CHANGE AND
		Parent)
	{
		Parent->OnEvent(Msg);
	}

	if (VScroll) VScroll->OnEvent(Msg);
	if (HScroll) HScroll->OnEvent(Msg);
	return GView::OnEvent(Msg);
}

//////////////////////////////////////////////////////////////////////////////////////
GWin32Class GScrollBar::WndClass;

GScrollBar::GScrollBar()
#ifdef LGI_RES
	: ResObject(Res_ScrollBar)
#endif
{
	Type = 0;
	Parent = 0;
	ZeroObj(Info);
	Info.cbSize = sizeof(Info);
	Info.fMask = SIF_ALL; // | SIF_DISABLENOSCROLL;
	Info.nMax = -1;
}

GScrollBar::GScrollBar(int id, int x, int y, int cx, int cy, char *name)
#ifdef LGI_RES
	: ResObject(Res_ScrollBar)
#endif
{
	SetId(id);
	GView::Pos.Set(x, y, x+cx, y+cy);
	if (name) GControl::Name(name);
	WndClass.SubClass(Class(), "SCROLLBAR");

	Type = SB_HORZ;
	ZeroObj(Info);
	Info.cbSize = sizeof(Info);
	Info.fMask = SIF_ALL | SIF_DISABLENOSCROLL;
	Info.nMax = -1;
}

GScrollBar::~GScrollBar()
{
	if (NOT Handle())
	{
		SetParentFlag(false);
	}
}

uint32 GScrollBar::Style()
{
	return	WS_CHILD |
			WS_TABSTOP |
			WS_VISIBLE |
			GView::Style() |
			((Type == SB_VERT) ? SBS_VERT : SBS_HORZ);
}

bool GScrollBar::Vertical()
{
	return Type == SB_VERT;
}

void GScrollBar::SetVertical(bool v)
{
	if (v)
	{
		Type = SB_VERT;
	}
	else
	{
		Type = SB_HORZ;
	}
}

void GScrollBar::SetParentFlag(bool Set)
{
	if (NOT Handle() AND
		Parent AND
		Parent->Handle())
	{
		DWORD Flags = GetWindowLong(Parent->Handle(), GWL_STYLE);
		int Flag = (Vertical()) ? WS_VSCROLL : WS_HSCROLL;

		if (Set)
		{
			Flags |= Flag;
		}
		else
		{
			Flags &= ~Flag;

			// clear limits.
			Info.nMin = 0;
			Info.nMax = -1;
			Update();
		}

		// check that we're in thread...
		if (Parent->InThread())
		{
			// you can't call 'SetWindowLong' from a thread other than
			// the one that created the window.
			SetWindowLong(Parent->Handle(), GWL_STYLE, Flags);
			SetWindowPos(	Parent->Handle(),
							0, 0, 0, 0, 0,
							SWP_NOMOVE | SWP_NOZORDER | SWP_NOSIZE | SWP_FRAMECHANGED);

			if (Set)
			{
				Update();
			}
		}
	}
}

void GScrollBar::Update()
{
	if (Parent AND Parent->Handle())
	{
		if (Handle())
		{
			if (InThread())
			{
				SetScrollInfo(Handle(), SB_CTL, &Info, true);
				EnableScrollBar(Handle(), SB_CTL, (Info.nMax < Info.nMin) ? ESB_DISABLE_BOTH : ESB_ENABLE_BOTH);
			}
		}
		else if (Parent)
		{
			if (Parent->InThread())
			{
				Info.fMask = SIF_ALL;
				SetScrollInfo(Parent->Handle(), Type, &Info, true);
			}
		}
	}
}

void GScrollBar::SetParent(GView *p)
{
	Parent = p;
	SetParentFlag(true);
}

int GScrollBar::Pos()
{
	return Info.nPos;
}

void GScrollBar::SetPos(int p)
{
	Info.nPos = limit(p, Info.nMin, Info.nMax);
	Update();
}

void GScrollBar::Limits(int &Low, int &High)
{
	Low = Info.nMin;
	High = Info.nMax;
}

void GScrollBar::SetLimits(int Low, int High)
{
	Info.nMin = Low;
	Info.nMax = High;
	Update();
}

int GScrollBar::Page()
{
	return Info.nPage;
}

void GScrollBar::SetPage(int p)
{
	Info.nPage = p;
	Update();
}

void GScrollBar::Invalidate()
{
	SetScrollInfo(Parent->Handle(), Type, &Info, true);
}

int GScrollBar::OnEvent(GMessage *Msg)
{
	int Status = WndClass.CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);

	if ((Msg->Msg == WM_HSCROLL AND Type == SB_HORZ) OR
		(Msg->Msg == WM_VSCROLL AND Type == SB_VERT))
	{
		int OldPos = Info.nPos;
		int NewPos = Info.nPos;
		int ScrollCode = (int) LOWORD(Msg->a);
		int nPos = (short) HIWORD(Msg->a);

		switch (ScrollCode)
		{
			case SB_BOTTOM:
			case SB_LINEDOWN:
			{
				NewPos = limit(NewPos+1, Info.nMin, Info.nMax);
				break;
			}
			case SB_TOP:
			case SB_LINEUP:
			{
				NewPos = limit(NewPos-1, Info.nMin, Info.nMax);
				break;
			}
			case SB_PAGEDOWN:
			{
				NewPos = limit(NewPos+(int)Info.nPage, Info.nMin, Info.nMax);
				break;
			}
			case SB_PAGEUP:
			{
				NewPos = limit(NewPos-(int)Info.nPage, Info.nMin, Info.nMax);
				break;
			}
			case SB_THUMBPOSITION:
			case SB_THUMBTRACK:
			{
				NewPos = limit(nPos, Info.nMin, Info.nMax);
				break;
			}
		}

		if (OldPos != NewPos)
		{
			Info.nPos = NewPos;
			Update();
			OnChange(NewPos);

			if (Parent)
			{
				GView *w = dynamic_cast<GView*>(this);
				GMessage m(WM_CHANGE, (int)w, NewPos);
				
				Parent->OnEvent(&m);
			}
		}
	}

	if (MsgCode(Msg) == WM_CREATE)
	{
		Update();
	}

	return Status;
}

