/*hdr
**      FILE:           GuiMenu.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           18/7/98
**      DESCRIPTION:    Gui menu system
**
**      Copyright (C) 1998, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

///////////////////////////////////////////////////////////////////////////////////////////////
#define LGI_OWNER_DRAW_MENUS

///////////////////////////////////////////////////////////////////////////////////////////////
GSubMenu::GSubMenu(char *name, bool Popup)
{
	if (Popup)
	{
		hMenu = CreatePopupMenu();
	}
	else
	{
		hMenu = CreateMenu();
	}

	Menu = 0;
	Parent = 0;
	TrackHandle = 0;
	Window = 0;
}

GSubMenu::~GSubMenu()
{
	DeleteObjList(&Items, GMenuItem);

	if (TrackHandle)
	{
		// At least in Win98, if you call InvalidateRect(...) on a window
		// multiple times just after you use TracePopupMenu(...) on that
		// window the invalid region isn't updated.
		//
		// This behaviour is not by design, it is a bug. However this
		// update forces the window to repaint thus working around the
		// bug in windows.
		UpdateWindow(TrackHandle);
		TrackHandle = 0;
	}

	if (hMenu)
	{
		DestroyMenu(hMenu);
		hMenu = NULL;
	}
}

GMenuItem *GSubMenu::AppendItem(char *Str, int Id, bool Enabled, int Where)
{
	GMenuItem *Item = NEW(GMenuItem(Menu, this, Items.GetItems()));
	if (Item)
	{
		Item->Name(Str);
		Item->Id(Id);
		Item->Enabled(Enabled);
		Item->Parent = this;
		Items.Insert(Item, Where);
		Item->ScanForAccel();

		Item->Insert(Items.IndexOf(Item));
	}
	return Item;
}

GMenuItem *GSubMenu::AppendSeparator(int Where)
{
	GMenuItem *Item = NEW(GMenuItem(Menu, this, Items.GetItems()));
	if (Item)
	{
		Item->Separator(true);
		Item->Parent = this;
		Items.Insert(Item, Where);

		Item->Insert(Items.IndexOf(Item));
	}
	return Item;
}

GSubMenu *GSubMenu::AppendSub(char *Str, int Where)
{
	GMenuItem *Item = NEW(GMenuItem(Menu, this, Items.GetItems()));
	GSubMenu *Sub = NEW(GSubMenu);

	if (Item AND Sub)
	{
		Item->Parent = this;
		Item->Name(Str);
		Item->Sub(Sub);
		Items.Insert(Item, Where);

		Item->Insert(Items.IndexOf(Item));
	}
	else
	{
		DeleteObj(Item);
		DeleteObj(Sub);
	}

	return Sub;
}

void GSubMenu::Empty()
{
	while (RemoveItem(0));
}

bool GSubMenu::RemoveItem(int i)
{
	bool Status = false;
	GMenuItem *Item = Items[i];
	if (Item)
	{
		Status = Item->Remove();
	}
	return Status;
}

bool GSubMenu::RemoveItem(GMenuItem *Item)
{
	if (Item AND Items.HasItem(Item))
	{
		return Item->Remove();
	}

	return FALSE;
}

int GSubMenu::Float(GView *From, int x, int y, bool Left)
{
	int Cmd = 0;

	if (From AND hMenu)
	{
		GView *Wnd = From;
		while (Wnd AND NOT Wnd->Handle())
		{
			Wnd = Wnd->GetParent();
		}
		TrackHandle = Wnd ? Wnd->Handle() : 0;

		Cmd = TrackPopupMenu(	hMenu,
								TPM_LEFTALIGN |
								TPM_TOPALIGN |
								TPM_RETURNCMD |
								((Left)?TPM_LEFTBUTTON:TPM_RIGHTBUTTON),
								x,
								y,
								0,
								TrackHandle,
								NULL);
	}

	return Cmd;
}

GMenuItem *GSubMenu::ItemAt(int Id)
{
	return Items.ItemAt(Id);
}

GMenuItem *GSubMenu::FindItem(int Id)
{
	for (GMenuItem *i = Items.First(); i; i = Items.Next())
	{
		GSubMenu *Sub = i->Sub();

		if (i->Id() == Id)
		{
			return i;
		}
		else if (Sub)
		{
			i = Sub->FindItem(Id);
			if (i)
			{
				return i;
			}
		}
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////
GMenuItem::GMenuItem()
{
	Menu = 0;
	Parent = 0;
	Child = 0;
	Position = 0;
	_Icon = -1;

	ZeroObj(Info);
	Info.cbSize = sizeof(Info);

	#ifdef LGI_OWNER_DRAW_MENUS
	Info.fMask = MIIM_DATA;
	Info.fType = MFT_OWNERDRAW;
	Info.dwItemData = (int)this;
	#endif

	Enabled(true);
}

GMenuItem::GMenuItem(GMenu *m, GSubMenu *p, int Pos)
{
	Position = Pos;
	Menu = m;
	Parent = p;
	Child = 0;

	ZeroObj(Info);
	Info.cbSize = sizeof(Info);

	#ifdef LGI_OWNER_DRAW_MENUS
	Info.fMask = MIIM_DATA;
	Info.fType = MFT_OWNERDRAW;
	Info.dwItemData = (int)this;
	#endif

	Name("<error>");
	Id(0);
	Enabled(true);
}

GMenuItem::~GMenuItem()
{
	DeleteObj(Child);
}

// the following 3 functions paint the menus according the to
// windows standard. but also allow for correct drawing of menuitem
// icons. some implementations of windows force the program back
// to the 8-bit palette when specifying the icon graphic, thus removing
// control over the colours displayed. these functions remove that
// limitation and also provide the application the ability to override
// the default painting behaviour if desired.
void GMenuItem::_Measure(GdcPt2 &Size)
{
	bool BaseMenu = Parent == Menu; // true if attached to a windows menu
									// else is a submenu
	int Ht = Menu->GetFont()->Y("A");
	// int IconX = BaseMenu ? ((24-Ht)/2)-Ht : 20;
	int IconX = BaseMenu ? 0 : 20;

	if (Separator())
	{
		Size.x = 8;
		Size.y = 8;
	}
	else
	{
		// remove '&' chars for string measurement
		char Str[256];
		char *n = Name(), *i = n, *o = Str;

		while (i AND *i)
		{
			if (*i == '&')
			{
				if (i[1] == '&')
				{
					*o++ = *i++;
				}
			}
			else
			{
				*o++ = *i;
			}

			i++;
		}
		*o++ = 0;

		// check for accelerators
		char *Tab = strchr(Str, '\t');
		if (Tab)
		{
			// string with accel
			int Mx, Tx;
			Menu->GetFont()->Size(&Mx, 0, Str, (int)Tab-(int)Str);
			Menu->GetFont()->Size(&Tx, 0, Tab+1);
			Size.x = IconX + 32 + Mx + Tx;
		}
		else
		{
			// normal string
			Size.x = IconX + Menu->GetFont()->X(Str);
		}

		if (NOT BaseMenu)
		{
			// leave room for child pointer
			Size.x += Child ? 8 : 0;
		}

		Size.y = BaseMenu ? GetSystemMetrics(SM_CYMENU)+1 : Ht+2;
	}
}

void GMenuItem::_PaintText(GSurface *pDC, int x, int y, int Width)
{
	char *n = Name();
	if (n)
	{
		bool Underline = false;
		int CharY = Menu->GetFont()->Y("A")-1;

		char *e = 0;
		for (char *s=n; *s; s = (*e) ? e + 1 : e)
		{
			// find end of segment
			for (e = s; *e; e++)
			{
				if (*e == '\t') break;
				if (e[0] == '&' AND e[1] != '&') break;
			}
			int Len = (int)e-(int)s;

			if (Len > 0)
			{
				// paint text till that point
				int SegX;
				
				Menu->GetFont()->Size(&SegX, 0, s, Len);
				Menu->GetFont()->Text(pDC, x, y, s, Len);

				if (Underline)
				{
					float *Widths = Menu->GetFont()->GetWidths();
					if (Widths)
					{
						pDC->Colour(Menu->GetFont()->Fore(), 24);
						pDC->Line(x, y+CharY, x+max(Widths[*s]-1, 2), y+CharY);
					}
					Underline = false;
				}

				x += SegX;
			}

			if (*e == '\t')
			{
				x = Width - Menu->GetFont()->X(e + 1) - 8;
			}
			if (*e == '&')
			{
				Underline = true;
			}
		}

	}
}

void GMenuItem::_Paint(GSurface *pDC, int Flags)
{
	bool BaseMenu = Parent == Menu;
	int IconX = BaseMenu ? 5 : 20;
	bool Selected = TestFlag(Flags, ODS_SELECTED);
	bool Disabled = TestFlag(Flags, ODS_DISABLED);
	bool Checked = TestFlag(Flags, ODS_CHECKED);
	GRegion r(0, 0, pDC->X()-1, pDC->Y()-1);
	char *Text = Name();

	if (Separator())
	{
		// paint a separator
		int Cy = pDC->Y() / 2;

		pDC->Colour(LC_MED, 24);
		pDC->Rectangle();

		pDC->Colour(LC_LOW, 24);
		pDC->Line(0, Cy-1, pDC->X()-1, Cy-1);
		
		pDC->Colour(LC_LIGHT, 24);
		pDC->Line(0, Cy, pDC->X()-1, Cy);
	}
	else
	{
		// paint a text menu item
		COLOUR Fore = Selected ? LC_SEL_TEXT : LC_TEXT;
		COLOUR Back = Selected ? LC_SELECTION : LC_MED;
		int x = IconX;
		int y = 1;

		// paint the background
		if (BaseMenu)
		{
			// for a menu, sunken on selected
			GRegion rgn = r;
			if (Selected)
			{
				LgiThinBorder(pDC, rgn, SUNKEN);
				Fore = LC_TEXT;
				x++;
				y++;
			}
			
			// always dialog colour
			pDC->Colour(LC_MED, 24);
			pDC->Rectangle(&rgn);
		}
		else
		{
			// for a submenu
			pDC->Colour(Back, 24);
			pDC->Rectangle();
		}

		// draw the text on top
		Menu->GetFont()->Transparent(true);
		if (Disabled)
		{
			// disabled text
			if (NOT Selected)
			{
				Menu->GetFont()->Colour(LC_LIGHT, 0);
				_PaintText(pDC, x+1, y+1, r.X());
			}
			// else selected... don't draw the hilight

			// "greyed" text...
			Menu->GetFont()->Colour(LC_LOW, 0);
			_PaintText(pDC, x, y, r.X());
		}
		else
		{
			// normal coloured text
			Menu->GetFont()->Colour(Fore, 0);
			_PaintText(pDC, x, y, r.X());
		}

		GImageList *ImgLst = (Menu AND Menu->GetImageList()) ? Menu->GetImageList() : Parent ? Parent->GetImageList() : 0;

		// draw icon/check mark
		if (Checked AND IconX > 0)
		{
			// it's a check!
			int x = 4;
			int y = 6;
			
			pDC->Colour(Fore, 24);
			pDC->Line(x, y, x+2, y+2);
			pDC->Line(x+2, y+2, x+6, y-2);
			y++;
			pDC->Line(x, y, x+2, y+2);
			pDC->Line(x+2, y+2, x+6, y-2);
			y++;
			pDC->Line(x, y, x+2, y+2);
			pDC->Line(x+2, y+2, x+6, y-2);
		}
		else if (ImgLst AND
				_Icon >= 0)
		{
			// it's an icon!
			ImgLst->Draw(pDC, 0, 0, _Icon);
		}
	}
}

bool GMenuItem::ScanForAccel()
{
	char *n = GObject::Name();
	if (n AND Menu)
	{
		char *Tab = strchr(n, '\t');
		if (Tab)
		{
			if (strlen(Tab+1) >= 6 AND
				strnicmp(Tab+1, "Ctrl+", 5) == 0)
			{
				Menu->Accel.Insert(NEW(GAccelerator(VK_CONTROL, VkKeyScan(Tab[6]) & 0xFF, Id())));
				return true;
			}
			else if (stricmp(Tab+1, "Del") == 0)
			{
				Menu->Accel.Insert(NEW(GAccelerator(0, VK_DELETE, Id())));
				return true;
			}
		}
	}

	return false;
}

GSubMenu *GMenuItem::GetParent()
{
	return Parent;
}

bool GMenuItem::Remove()
{
	if (Parent)
	{
		int Index = Parent->Items.IndexOf(this);
		Parent->Items.Delete(this);
		return RemoveMenu(Parent->Handle(), Index, MF_BYPOSITION); 
	}
	return FALSE;
}

bool GMenuItem::Update()
{
	bool Status = FALSE;

	if (Parent AND Parent->Handle())
	{
		Status = SetMenuItemInfo(Parent->Handle(), Position, true, &Info);
	}

	return Status;
}

void GMenuItem::Icon(int i)
{
	_Icon = i;
}

int GMenuItem::Icon()
{
	return _Icon;
}

void GMenuItem::Id(int i)
{
	Info.wID = i;
	Info.fMask |= MIIM_ID;
	Update();
}

void GMenuItem::Separator(bool s)
{
	if (s)
	{
		Info.fType |= MFT_SEPARATOR;
	}
	else
	{
		Info.fType &= ~MFT_SEPARATOR;
	}

	Info.fMask |= MIIM_TYPE;
	Update();
}

void GMenuItem::Checked(bool c)
{
	if (c)
	{
		Info.fState |= MFS_CHECKED;
	}
	else
	{
		Info.fState &= ~MFS_CHECKED;
	}
	
	Info.fMask |= MIIM_STATE;
	Update();
}

bool GMenuItem::Name(char *n)
{
	bool Status = GObject::Name(n);
	if (Status)
	{
		Info.dwTypeData = GObject::Name();
		Info.cch = strlen(GObject::Name());
		Info.fType |= MFT_STRING;
		Info.fMask |= MIIM_TYPE | MIIM_DATA;
		Update();
	}
	return Status;
}

void GMenuItem::Enabled(bool e)
{
	Info.fState &= ~(MFS_ENABLED|MFS_DISABLED|MFS_GRAYED);

	if (e)
	{
		Info.fState |= MFS_ENABLED;
	}
	else
	{
		Info.fState |= MFS_DISABLED|MFS_GRAYED;
	}

	Info.fMask |= MIIM_STATE;
	Update();
}

void GMenuItem::Focus(bool f)
{
	if (f)
	{
		Info.fState |= MFS_HILITE;
	}
	else
	{
		Info.fState &= ~MFS_HILITE;
	}

	Info.fMask |= MIIM_STATE;
	Update();
}

void GMenuItem::Sub(GSubMenu *s)
{
	Child = s;
	if (Child)
	{
		Info.hSubMenu = Child->Handle();
		s->Menu = Menu;
		s->Parent = this;
	}
	else
	{
		Info.hSubMenu = 0;
	}
	Info.fMask |= MIIM_SUBMENU;
	Update();
}

bool GMenuItem::Insert(int Pos)
{
	bool Status = false;

	if (Parent)
	{
		Position = Pos;
		Status = InsertMenuItem(Parent->Handle(),
								Position,
								true,
								&Info);
	}

	return Status;
}

void GMenuItem::Visible(bool i)
{
}

int GMenuItem::Id() { return Info.wID; }
char *GMenuItem::Name() { return (NOT (Info.fType & MFT_SEPARATOR)) ? Info.dwTypeData : 0; }
bool GMenuItem::Separator() { return (Info.fType & MFT_SEPARATOR) != 0; }
bool GMenuItem::Checked() { return (Info.fState & MF_CHECKED) != 0; }
bool GMenuItem::Enabled() { return (Info.fState & MFS_ENABLED) != 0; }
bool GMenuItem::Visible() { return true; }
bool GMenuItem::Focus() { return (Info.fState & MFS_HILITE) != 0; }
GSubMenu *GMenuItem::Sub() { return Child; }

///////////////////////////////////////////////////////////////////////////////////////////////
int _GMenuFontRef = 0;
GFont *GMenu::_Font = 0;

GMenu::GMenu() : GSubMenu("", false)
{
	Menu = this;
	Window = NULL;
	_GMenuFontRef++;
}

GMenu::~GMenu()
{
	DeleteObjList(&Accel, GAccelerator);
	if (--_GMenuFontRef == 0)
	{
		DeleteObj(_Font);
	}
}

GFont *GMenu::GetFont()
{
	if (NOT _Font)
	{
		GdcFontType Type;
		if (Type.GetSystemFont("Menu"))
		{
			_Font = Type.CreateFont();
		}
	}

	return _Font ? _Font : SysFont;
}

bool GMenu::Attach(GView *p)
{
	bool Status = FALSE;
	Window = p;
	if (Window)
	{
		Status = SetMenu(Window->Handle(), hMenu);
	}
	return Status;
}

bool GMenu::Detach()
{
	bool Status = FALSE;

	if (Window)
	{
		HMENU hWndMenu = ::GetMenu(Window->Handle());
		if (hWndMenu == hMenu)
		{
			Status = SetMenu(Window->Handle(), NULL);
			if (Status)
			{
				Window = NULL;
			}
		}
	}

	return Status;
}

int GMenu::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case WM_TRANSLATE_ACCELERATOR:
		{
			MSG *m = (MSG*) MsgB(Msg);
			if (m AND
				m->message == WM_KEYDOWN)
			{
				for (GAccelerator *a = Accel.First(); a; a = Accel.Next())
				{
					if (a->Match(m->wParam, m->lParam))
					{
						Window->OnCommand(a->GetId(), 0, 0);
						break;
					}
				}
			}
			break;
		}
	}
	return 0;
}

int GMenu::_OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case WM_MENUCHAR:
		{
			short Key = MsgA(Msg) & 0xffff;
			HMENU hMenu = (HMENU)MsgB(Msg);
			MENUITEMINFO Info;

			ZeroObj(Info);
			Info.cbSize = sizeof(Info);
			Info.fMask = MIIM_DATA;
			if (GetMenuItemInfo(hMenu, 0, true, &Info))
			{
				GMenuItem *Item = (GMenuItem*)Info.dwItemData;
				if (Item)
				{
					GSubMenu *Menu = Item->Parent;
					if (Menu)
					{
						int Index=0;
						for (GMenuItem *i=Menu->Items.First(); i; i=Menu->Items.Next(), Index++)
						{
							char *n = i->Name();
							if (n)
							{
								char *Amp = strchr(n, '&');
								if (Amp AND
									toupper(Amp[1]) == toupper(Key))
								{
									return (MNC_EXECUTE << 16) | Index;
								}
							}
						}
					}
				}
			}			
			break;
		}
		case WM_MEASUREITEM:
		{
			LPMEASUREITEMSTRUCT Item = (LPMEASUREITEMSTRUCT)Msg->b;
			if (Item)
			{
				GdcPt2 Size;

				((GMenuItem*)Item->itemData)->_Measure(Size);
				Item->itemWidth = Size.x;
				Item->itemHeight = Size.y;
				return true;
			}
			break;
		}
		case WM_DRAWITEM:
		{
			LPDRAWITEMSTRUCT Item = (LPDRAWITEMSTRUCT)Msg->b;
			if (Item)
			{
				GRegion r = Item->rcItem;

				GScreenDC Dc(Item->hDC, Item->hwndItem);
				Dc.SetOrigin(-r.x1, -r.y1);
				Dc.SetSize(r.X(), r.Y());

				((GMenuItem*)Item->itemData)->_Paint(&Dc, Item->itemState);

				Dc.SetOrigin(0, 0);

				return true;
			}
			break;
		}
	}

	return 0;
}

////////////////////////////////////////////////////////////////////////////
GAccelerator::GAccelerator(int flags, int key, int id)
{
	Flags = flags;
	Key = key;
	Id = id;
}

bool GAccelerator::Match(int a, int b)
{
	if (a == Key)
	{
		if (NOT Flags OR
			(GetAsyncKeyState(Flags) & 0x8000) != 0)
		{
			return true;
		}
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////
GCommand::GCommand()
{
	Flags = GWF_ENABLED | GWF_VISIBLE;
	Id = 0;
	ToolButton = 0;
	MenuItem = 0;
	Accelerator = 0;
	TipHelp = 0;
	PrevValue = false;
}

GCommand::~GCommand()
{
	DeleteArray(Accelerator);
	DeleteArray(TipHelp);
}

bool GCommand::Enabled()
{
	if (ToolButton)
		return ToolButton->Enabled();
	if (MenuItem)
		return MenuItem->Enabled();
	return false;
}

void GCommand::Enabled(bool e)
{
	if (ToolButton)
	{
		ToolButton->Enabled(e);
	}
	if (MenuItem)
	{
		MenuItem->Enabled(e);
	}
}

bool GCommand::Value()
{
	bool HasChanged = false;

	if (ToolButton)
	{
		HasChanged |= (ToolButton->Value() != 0) ^ PrevValue;
	}
	if (MenuItem)
	{
		HasChanged |= (MenuItem->Checked() != 0) ^ PrevValue;
	}

	if (HasChanged)
	{
		Value(NOT PrevValue);
	}

	return PrevValue;
}

void GCommand::Value(bool v)
{
	if (ToolButton)
	{
		ToolButton->Value(v);
	}
	if (MenuItem)
	{
		MenuItem->Checked(v);
	}
	PrevValue = v;
}
