#include "Lgi.h"
#include "Base64.h"

//////////////////////////////////////////////////////////////////////
GPrinter::GPrinter()
{
	ZeroObj(Info);
	Info.lStructSize = sizeof(Info);
}

GPrinter::~GPrinter()
{
}

HDC GPrinter::StartDC(GView *Parent)
{
	Info.hwndOwner = (Parent) ? Parent->Handle() : 0;
	Info.Flags = PD_RETURNDC;
	if (NOT Parent OR PrintDlg(&Info))
	{
		if (NOT Info.hDC)
		{
			GMem DevMode(Info.hDevMode);
			GMem DevNames(Info.hDevNames);
			DEVNAMES *Names = (DEVNAMES*) DevNames.Lock();
			if (Names)
			{
				Info.hDC = ::CreateDC(	NULL,
										((char*) Names) + Names->wDeviceOffset,
										NULL,
										(DEVMODE*) DevMode.Lock());
			}
		}
	}

	return Info.hDC;
}

void GPrinter::EndDC()
{
	if (Info.hDC)
	{
		DeleteDC(Info.hDC);
		Info.hDC = 0;
	}
}

bool GPrinter::Browse(GView *Parent)
{
	Info.hwndOwner = (Parent) ? Parent->Handle() : 0;
	Info.Flags = PD_PRINTSETUP;
	return PrintDlg(&Info);
}

#define MAGIC_PRINTDLG					0xAAFF0100
#define MAGIC_DEVMODE					0xAAFF0101
#define MAGIC_DEVNAMES					0xAAFF0102

bool GPrinter::Serialize(pchar &Str, bool Write)
{
	int Size = sizeof(Info);
	if (Write)
	{
		GMem DevMode(Info.hDevMode);
		GMem DevNames(Info.hDevNames);
		DEVMODE *Mode = (DEVMODE*) DevMode.Lock();
		DEVNAMES *Names = (DEVNAMES*) DevNames.Lock();
		BytePipe Temp;
		int m;

		// Dlg
		m = MAGIC_PRINTDLG;
		Temp.Push(m);
		m = sizeof(Info);
		Temp.Push(m);
		Temp.Push((uchar*)&Info, m);

		// Mode
		if (Mode)
		{
			m = MAGIC_DEVMODE;
			Temp.Push(m);
			m = DevMode.GetSize();
			Temp.Push(m);
			Temp.Push((uchar*) Mode, m);
		}

		// Names
		if (Names)
		{
			m = MAGIC_DEVNAMES;
			Temp.Push(m);
			m = DevNames.GetSize();
			Temp.Push(m);
			Temp.Push((uchar*) Names, m);
		}

		// Convert to base64
		int BinLen = Temp.Sizeof();
		uchar *Bin = NEW(uchar[BinLen]);
		if (Bin AND Temp.Pop(Bin, BinLen))
		{
			int Base64Len = BufferLen_BinTo64(BinLen);
			char *Base64 = NEW(char[Base64Len+1]);
			if (Base64)
			{
				memset(Base64, 0, Base64Len+1);
				ConvertBinaryToBase64(Base64, Base64Len, Bin, BinLen);
				
				DeleteArray(Str);
				Str = Base64;
			}
			return true;
		}
	}
	else
	{
		bool Status = false;

		ZeroObj(Info);
		Info.lStructSize = Size;
		if (Str)
		{
			// Convert to binary
			BytePipe Temp;

			int Base64Len = strlen(Str);
			int BinaryLen = BufferLen_64ToBin(Base64Len);
			uchar *Binary = NEW(uchar[BinaryLen]);
			if (Binary)
			{
				int Len = ConvertBase64ToBinary(Binary, BinaryLen, Str, Base64Len);
				Temp.Push(Binary, Len);

				bool Done = false;
				do
				{
					int Type = 0;
					int Size = 0;

					if (Temp.Pop(Type) AND
						Temp.Pop(Size))
					{
						switch (Type)
						{
							case MAGIC_PRINTDLG:
							{
								if (Size == sizeof(Info))
								{
									Temp.Pop((uchar*) &Info, Size);
									Status = true;
								}
								else
								{
									// Grrrr: Error
									Done = true;
								}
								break;
							}
							case MAGIC_DEVMODE:
							{
								GMem DevMode(Size);
								uchar *Mode = (uchar*) DevMode.Lock();
								if (Mode)
								{
									Temp.Pop(Mode, Size);
									Info.hDevMode = DevMode.Handle();
									DevMode.Detach();
								}
								break;
							}
							case MAGIC_DEVNAMES:
							{
								GMem DevNames(Size);
								uchar *Names = (uchar*) DevNames.Lock();
								if (Names)
								{
									Temp.Pop(Names, Size);
									Info.hDevNames = DevNames.Handle();
									DevNames.Detach();
								}
								break;
							}
							default:
							{
								// Eof
								Done = true;
								break;
							}
						}
					}
					else
					{
						Done = true;
					}
				}
				while (NOT Done);
			}
		}

		return Status;
	}

	return false;
}

