#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
#define SPLITER_BORDER			4
#define SPLITER_BAR				5
#define SPLITER_INSET			2
#define SPLITER_MIN_X			(2*SPLITER_INSET)
#define SPLITER_MIN_Y			(2*SPLITER_INSET)

#define SPLITER_FULL_DRAG

GSplitter::GSplitter()
{
	Name("LGI_Spliter");
	WndFlags |= WS_CLIPCHILDREN;
	SplitSet = SplitPos = -1;
	Moving = FALSE;
	Cursor = 0;

	PosA.ZOff(0, 0);
	PosB.ZOff(0, 0);
	ViewA = 0;
	ViewB = 0;
	BorderA = true;
	BorderB = true;
	OverBar = false;
	SplitFollow = false;

	OldCursor = 0;
	IsVertical(true);
	Border(true);
	Raised(true);
	Split(0);

	SetPourLargest(true);
	// WndStyle |= WS_CLIPCHILDREN;
}

GSplitter::~GSplitter()
{
	DeleteObj(ViewA);
	DeleteObj(ViewB);
}

void GSplitter::IsVertical(bool v)
{
	Vertical = v;
	if (Cursor)
	{
		DeleteObject(Cursor);
		Cursor = 0;
	}
}

void GSplitter::SetViewA(GView *a, bool Border)
{
	BorderA = Border;
	ViewA = a;
	if (a)
	{
		a->Attach(this);
		a->Visible(true);

		CalcRegions(Pos);
		if (Visible())
		{
			Invalidate();
		}
	}
}

void GSplitter::SetViewB(GView *b, bool Border)
{
	BorderB = Border;
	ViewB = b;
	if (b)
	{
		b->Attach(this);
		b->Visible(true);

		CalcRegions(Pos);
		if (Visible())
		{
			Invalidate();
		}
	}
}

void GSplitter::Split(int s)
{
	SplitSet = s;
	int Limit = ((Vertical) ? Pos.X() : Pos.Y()) - 18;
	int NewPos = (s > 0) ? max(s, 4) : max(4, Limit+s);
	if (NewPos != SplitPos)
	{
		SplitPos = NewPos;
		CalcRegions(Pos);
		if (Visible())
		{
			Invalidate();
		}
	}
}

GView *GSplitter::FindControl(HWND hCtrl)
{
	GView *c = 0;
	if (ViewA) c = ViewA->FindControl(hCtrl);
	if (c) return c;

	if (ViewB) c = ViewB->FindControl(hCtrl);
	return c;
}

void GSplitter::CalcRegions(GRegion Rect, bool Follow)
{
	PosA = Rect;
	PosA.Offset(-PosA.x1, -PosA.y1);
	if (Border())
	{
		PosA.Size(SPLITER_BORDER, SPLITER_BORDER);
	}

	PosB = PosA;

	if (SplitFollow AND Follow)
	{
		SplitPos = max(1, ((Vertical) ? Pos.X()-10 : Pos.Y()-10) + SplitSet);
	}

	if (Vertical)
	{
		PosA.x2 = PosA.x1 + SplitPos;
		PosB.x1 = PosA.x2 + SPLITER_BAR;
	}
	else
	{
		PosA.y2 = PosA.y1 + SplitPos;
		PosB.y1 = PosA.y2 + SPLITER_BAR;
	}

	if (ViewA)
	{
		GRegion r = PosA;
		if (BorderA)
		{
			r.Size(SPLITER_INSET, SPLITER_INSET);
		}
		ViewA->SetPos(r);
		
		GComplexRegion Rgn(r);
		ViewA->Pour(Rgn);
	}
	if (ViewB)
	{
		GRegion r = PosB;
		if (BorderB)
		{
			r.Size(SPLITER_INSET, SPLITER_INSET);
		}
		ViewB->SetPos(r);

		GComplexRegion Rgn(r);
		ViewB->Pour(Rgn);
	}
}

void GSplitter::OnPosChange()
{
	CalcRegions(Pos, true);
}

void ClipDC(HDC hDC, RECT rc)
{
	HRGN hRgn = CreateRectRgn(rc.left, rc.top, rc.right, rc.bottom);
	SelectClipRgn(hDC, hRgn);
}

void GSplitter::OnPaint(GSurface *pDC)
{
	GRegion r;
	r.ZOff(pDC->X()-1, pDC->Y()-1);

	if (Border())
	{
		if (Sunken())
		{
			LgiThinBorder(pDC, r, SUNKEN);
		}
		else if (Raised())
		{
			LgiThinBorder(pDC, r, RAISED);
		}

		pDC->Colour(LC_MED, 24);
		LgiFlatBorder(pDC, r, SPLITER_BORDER);
	}
	else
	{
		pDC->Colour(LC_MED, 24);
	}

	// bar
	if (Vertical)
	{
		pDC->Rectangle(PosA.x2, PosA.y1, PosB.x1, PosA.y2);
	}
	else
	{
		pDC->Rectangle(PosA.x1, PosA.y2, PosA.x2, PosB.y1);
	}

	// PosA region
	r = PosA;
	if (r.X() > SPLITER_MIN_X AND r.Y() > SPLITER_MIN_Y)
	{
		if (BorderA)
		{
			LgiWideBorder(pDC, r, SUNKEN);
		}

		if (ViewA)
		{
			if (NOT ViewA->Handle())
			{
				pDC->SetOrigin(PosA.x1, PosA.y1);
				ViewA->OnPaint(pDC);
				pDC->SetOrigin(0, 0);
			}
			else
			{
				// View is real and will paint itself
			}
		}
		else
		{
			pDC->Colour(LC_WORKSPACE, 24);
			pDC->Rectangle(&r);
		}
	}
	else
	{
		// too small
	}

	r = PosB;
	if (r.X() > SPLITER_MIN_X AND r.Y() > SPLITER_MIN_Y)
	{
		if (BorderB)
		{
			LgiWideBorder(pDC, r, SUNKEN);
		}

		if (ViewB)
		{
			if (NOT ViewB->Handle())
			{
				pDC->SetOrigin(PosB.x1, PosB.y1);
				ViewB->OnPaint(pDC);
				pDC->SetOrigin(0, 0);
			}
			else
			{
				// View is real and will paint itself
			}
		}
		else
		{
			pDC->Colour(LC_WORKSPACE, 24);
			pDC->Rectangle(&r);
		}
	}
	else
	{
		// too small
	}

	pDC->EndDC();
}

void GSplitter::OnMouseClick(GMouse &m)
{
	if (m.Down())
	{
		if (OverSplit(m.x, m.y))
		{
			if (Vertical)
			{
				Capture(TRUE);
				Moving = TRUE;

				Bar.x1 = PosA.x2;
				Bar.y1 = PosA.y1;
				Bar.x2 = PosB.x1;
				Bar.y2 = PosA.y2;
				Offset = m.x - Bar.x1;

				#ifndef SPLITER_FULL_DRAG
				HashRectangle(this, &Bar);
				#endif
			}
			else
			{
				Capture(TRUE);
				Moving = TRUE;

				Bar.x1 = PosA.x1;
				Bar.y1 = PosA.y2;
				Bar.x2 = PosA.x2;
				Bar.y2 = PosB.y1;
				Offset = m.y - Bar.y1;

				#ifndef SPLITER_FULL_DRAG
				HashRectangle(this, &Bar);
				#endif
			}
		}
	}
	else
	{
		#ifdef SPLITER_FULL_DRAG
		Capture(FALSE);
		Moving = FALSE;
		#else
		if (Moving)
		{
			HashRectangle(this, &Bar);
			Capture(FALSE);
			Moving = FALSE;

			int NewPos;
			if (Vertical)
			{
				NewPos = limit(Bar.x1-4, 4, Pos.X()-18);
			}
			else
			{
				NewPos = limit(Bar.y1-4, 4, Pos.Y()-18);
			}

			if (NewPos != SplitPos)
			{
				SplitPos = NewPos;
				CalcRegions(Pos);
				Invalidate();
			}
		}
		#endif
	}
}

bool GSplitter::OverSplit(int x, int y)
{
	if (Vertical)
	{
		return (x > PosA.x2 AND x < PosB.x1 AND y > PosA.y1 AND y < PosA.y2);
	}
	else
	{
		return (x > PosA.x1 AND x < PosA.x2 AND y > PosA.y2 AND y < PosB.y1);
	}
}

void GSplitter::OnMouseExit(GMouse &m)
{
	if (OldCursor)
	{
		SetCursor(OldCursor);
	}
}

void GSplitter::OnMouseMove(GMouse &m)
{
	if (Moving)
	{
		#ifdef SPLITER_FULL_DRAG
		if (Vertical)
		{
			Bar.Offset((m.x - Offset) - Bar.x1, 0);
		}
		else
		{
			Bar.Offset(0, (m.y - Offset) - Bar.y1);
		}

		int NewPos;
		if (Vertical)
		{
			NewPos = limit(Bar.x1-4, 4, Pos.X()-18);
		}
		else
		{
			NewPos = limit(Bar.y1, 4, Pos.Y()-18);
		}

		if (NewPos != SplitPos)
		{
			SplitPos = NewPos;
			CalcRegions(Pos);
			Invalidate();
			#ifdef WIN32
			UpdateWindow(Handle());
			#endif
		}
		#else
		HashRectangle(this, &Bar);
		if (Vertical)
		{
			Bar.Offset((m.x - Offset) - Bar.x1, 0);
		}
		else
		{
			Bar.Offset(0, (m.y - Offset) - Bar.y1);
		}
		HashRectangle(this, &Bar);
		#endif
	}
}

#ifdef WIN32
int GSplitter::OnHitTest(int x, int y)
{
	if (OverSplit(x, y))
	{
		if (NOT Cursor)
		{
			Cursor = LoadCursor(NULL, MAKEINTRESOURCE((Vertical) ? IDC_SIZEWE : IDC_SIZENS));
		}

		SetCursor(Cursor);
		return 1;
	}

	return -1;
}
#endif

bool GSplitter::Attach(GView *p)
{
	bool Status = GLayout::Attach(p);
	if (Status)
	{
		if (ViewA AND NOT ViewA->Handle())
		{
			ViewA->Attach(this);
		}

		if (ViewB AND NOT ViewB->Handle())
		{
			ViewB->Attach(this);
		}
	}

	return Status;
}
