#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////////////
GStatusBar::GStatusBar()
{
	Name("LGI_StatusBar");

	Font = NEW(GFont);
	if (Font)
	{
		Font->Height(8);
		Font->Face("Ms Sans Serif");
		Font->Transparent(true);
		ASSERT(Font->Create());
	}
}

GStatusBar::~GStatusBar()
{
	DeleteObj(Font);
}

bool GStatusBar::Pour(GComplexRegion &r)
{
	GRegion *Best = FindLargestEdge(r, GV_EDGE_BOTTOM);
	if (Best)
	{
		GRegion Take = *Best;
		if (Take.Y() > 21)
		{
			Take.y1 = max((Take.y2 - 21), Take.y1);
			SetPos(Take);
			RePour();
			return TRUE;
		}
	}
	return FALSE;
}

void GStatusBar::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);

	LgiThinBorder(pDC, r, RAISED);

	GComplexRegion Cr(r);
	#ifdef WIN32
	for (GView *c = Children.First(); c; c = Children.Next())
	{
		GRegion n = c->GetPos();
		Cr.Subtract(&n);
	}
	#endif
	pDC->Colour(LC_MED, 24);
	for (GRegion *i = Cr.First(); i; i = Cr.Next())
	{
		pDC->Rectangle(i);
	}

	GView::OnPaint(pDC);
}

void GStatusBar::RePour()
{
	int X = Pos.X() - 5;
	int i = Children.GetItems()-1;

	for (	GView *w = Children.Last();
			w;
			w = Children.Prev(), i--)
	{
		GStatusPane *Pane = dynamic_cast<GStatusPane*>(w);
		if (Pane)
		{
			if (i)
			{
				// not first
				X -= Pane->Width; 
				Pane->Pos.ZOff(Pane->Width, Pos.Y()-7);
				Pane->Pos.Offset(X, 3);
				X -= STATUSBAR_SEPARATOR;
			}
			else
			{
				// first
				Pane->Pos.ZOff(X-2, Pos.Y()-4);
				Pane->Pos.Offset(2, 2);
				X = 0;
			}
		}
	}
}

GStatusPane *GStatusBar::AppendPane(char *Text, int Width)
{
	GStatusPane *Pane = 0;
	if (Text)
	{
		Pane = NEW(GStatusPane);
		if (Pane)
		{
			Pane->Parent = this;
			Pane->Name(Text);
			Pane->Width = Width;
			Children.Insert(Pane);
		}
	}
	return Pane;
}

bool GStatusBar::AppendPane(GStatusPane *Pane)
{
	if (Pane)
	{
		Pane->Parent = this;
		Children.Insert(Pane);
	}
	return TRUE;
}

// Pane
GStatusPane::GStatusPane()
{
	Parent = 0;
	Flags = 0;
	Width = 32;
	Pos.ZOff(Width-1, 20);
	pDC = 0;
}

GStatusPane::~GStatusPane()
{
	DeleteObj(pDC)
}

bool GStatusPane::Name(char *n)
{
	bool Status = GObject::Name(n);
	GRegion p(0, 0, Pos.X()-1, Pos.Y()-1);
	p.Size(1, 1);
	Invalidate(&p);
	return Status;
}

void GStatusPane::OnPaint(GSurface *pDC)
{
	GRegion r(0, 0, X()-1, Y()-1);
	int Off = 1;

	if (Sunken() OR Raised())
	{
		Off = 0;
		LgiThinBorder(pDC, r, Sunken()?SUNKEN:RAISED);
	}

	GStatusBar *Par = (GStatusBar*) Parent;
	if (Par AND Par->Font)
	{
		char *t = Name();
		if (t AND strlen(t)>0)
		{
			int TabSize = Par->Font->TabSize();
			Par->Font->TabSize(0);
			Par->Font->Colour(Rgb24(0, 0, 0), LC_MED);
			Par->Font->Transparent(false);
			Par->Font->Text(pDC, 3+Off, 1+Off, Name(), -1, &r);
			Par->Font->TabSize(TabSize);
			return;
		}
	}

	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&r);
}

int GStatusPane::GetWidth()
{
	return Width;
}

void GStatusPane::SetWidth(int x)
{
	Width = x;
	if (Parent)
	{
		((GStatusBar*) Parent)->RePour();
	}
}

bool GStatusPane::Sunken()
{
	return TestFlag(Flags, GSP_SUNKEN);
}

void GStatusPane::Sunken(bool s)
{
	bool Old = Sunken();
	if (s) SetFlag(Flags, GSP_SUNKEN);
	else ClearFlag(Flags, GSP_SUNKEN);
	if (Old != Sunken() AND Parent)
	{
		Parent->Invalidate(&Pos);
	}
}

GSurface *GStatusPane::Bitmap()
{
	return pDC;
}

void GStatusPane::Bitmap(GSurface *pNewDC)
{
	DeleteObj(pDC);
	if (pNewDC)
	{
		pDC = NEW(GMemDC);
		if (pDC)
		{
			if (pDC->Create(pNewDC->X(), pNewDC->Y(), pNewDC->GetBits()))
			{
				pDC->Blt(0, 0, pNewDC);
			}
			else
			{
				DeleteObj(pDC);
			}
		}
	}

	if (Parent)
	{
		Parent->Invalidate(&Pos);
	}
}

