#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
#include <process.h>

uint WINAPI ThreadEntryPoint(int i)
{
#ifdef	_MT
	GThread *Thread = (GThread*) i;
	if (Thread)
	{
		Thread->OnBeforeMain();
		Thread->State = LGITHREAD_RUNNING;
		int Status = Thread->ReturnValue = Thread->Main();
		Thread->State = LGITHREAD_EXITED;
		Thread->OnAfterMain();

		if (Thread->DeleteOnExit)
		{
			DeleteObj(Thread);
		}

		_endthreadex(Status);
	}
#endif

	return 0;
}

GThread::GThread()
{
	State = LGITHREAD_INIT;
	hThread = 0;
	ThreadId = 0;
	ReturnValue = 0;
	DeleteOnExit = false;

#ifdef	_MT
	hThread = (HANDLE) _beginthreadex(NULL,
									16<<10,
									(unsigned int (__stdcall *)(void *)) ThreadEntryPoint,
									(LPVOID) this,
									CREATE_SUSPENDED,
									&ThreadId);
#else
	#error "You are tryed to compile GThread support without the WIN32 multithreaded libs.";
#endif
}

GThread::~GThread()
{
}

bool GThread::IsExited()
{
	return (ExitCode() != STILL_ACTIVE);
}

bool GThread::IsRunning()
{
	return State == LGITHREAD_RUNNING;
}

void GThread::Run(bool b)
{
	DWORD Status = 0;
	DWORD Error = 0;

	if (b)
	{
		Status = ResumeThread(hThread);
		State = LGITHREAD_RUNNING;
	}
	else
	{
		State = LGITHREAD_ASLEEP;	// set before we sleep otherwise it can't 
									// be called from the actual thread.
		Status = SuspendThread(hThread);
	}

	OnRun(b);
}

void GThread::Terminate()
{
	if (hThread)
	{
		TerminateThread(hThread, 0);
		while (NOT IsExited());
	}
}

int GThread::ExitCode()
{
	DWORD d;
	GetExitCodeThread(hThread, &d);
	ReturnValue = d;
	return ReturnValue;
}

int GThread::Main()
{
	return 0;
}

int GThread::Priority()
{
	int p = 0;
	if (hThread)
	{
		p = GetThreadPriority(hThread);
	}
	return p;
}

void GThread::Priority(int p)
{
	if (hThread)
	{
		SetThreadPriority(hThread, p);
	}
}

