#include "Lgi.h"
#include "GTree.h"

//////////////////////////////////////////////////////////////////////////////
GTreeItem::GTreeItem()
{
	hTreeItem = 0;
	Tree = 0;
	Parent = 0;
	Str = 0;
	Sys_Image = -1;
}

GTreeItem::~GTreeItem()
{
	if (Tree)
	{
		Tree->Remove(this);
	}

	DeleteArray(Str);
}

char *GTreeItem::GetText(int i)
{
	return Str;
}

int GTreeItem::IndexOf()
{
	if (Parent)
	{
		int n=0;
		for (GTreeItem *c = Parent->GetChild(); c; c = c->GetNext())
		{
			if (c == this) return n;
			n++;
		}
	}
	else if (Tree)
	{
		return Tree->Items.IndexOf(this);
	}

	return -1;
}

GTreeItem *GTreeItem::GetChild()
{
	/*
	HTREEITEM hItem = TreeView_GetChild(Tree->Handle(), hTreeItem);
	if (hItem)
	{
		TV_ITEM Item;
		ZeroObj(Item);
		Item.mask = TVIF_PARAM;
		Item.hItem = hItem;

		if (TreeView_GetItem(Tree->Handle(), &Item))
		{
			return (GTreeItem*) Item.lParam;
		}
	}
	return 0;
	*/

	return Items.First();
}

GTreeItem *GTreeItem::GetNext()
{
	/*
	HTREEITEM hItem = TreeView_GetNextSibling(Tree->Handle(), hTreeItem);
	if (hItem)
	{
		TV_ITEM Item;
		ZeroObj(Item);
		Item.mask = TVIF_PARAM;
		Item.hItem = hItem;

		if (TreeView_GetItem(Tree->Handle(), &Item))
		{
			return (GTreeItem*) Item.lParam;
		}
	}
	return 0;
	*/

	if (Parent)
	{
		int Index = Parent->Items.IndexOf(this);
		if (Index >= 0)
		{
			return Parent->Items.ItemAt(Index+1);
		}
	}
	else if (Tree)
	{
		int Index = Tree->Items.IndexOf(this);
		if (Index >= 0)
		{
			return Tree->Items.ItemAt(Index+1);
		}
	}

	return 0;
}

bool GTreeItem::SetText(char *s, int i)
{
	DeleteArray(Str);
	Str = NewStr(s);
	return TRUE;
}

int GTreeItem::GetImage(int Flags)
{
	return Sys_Image;
}

void GTreeItem::SetImage(int i)
{
	Sys_Image = i;
}

GTreeItem *GTreeItem::Insert(GTreeItem *Obj, int Pos)
{
	GTreeItem *NewObj = (Obj) ? Obj : NEW(GTreeItem);
	if (NewObj)
	{
		Items.Delete(NewObj);
		Items.Insert(NewObj, Pos);
		NewObj->Parent = this;
		NewObj->Tree = Tree;
		NewObj->Attach((Pos>0) ? Items.ItemAt(Pos-1) : 0);

		for (GTreeItem *i=NewObj->Items.First(); i; i=NewObj->Items.Next())
		{
			NewObj->Insert(i);
		}
	}

	return NewObj;
}

void GTreeItem::Remove()
{
	if (Tree)
	{
		Tree->Remove(this);
	}
}

bool GTreeItem::Attach(GTreeItem *After)
{
	bool Status = FALSE;
	if (Tree)
	{
		TV_INSERTSTRUCT Item;

		Item.hParent = (Parent) ? Parent->ItemHandle() : TVI_ROOT;
		if (Parent AND Parent->Items.IndexOf(this) == 0)
		{
			Item.hInsertAfter = TVI_FIRST;
		}
		else
		{
			Item.hInsertAfter = (After) ? After->ItemHandle() : TVI_LAST;
		}

		Item.item.mask = TVIF_TEXT | TVIF_PARAM;
		Item.item.state = 0;
		Item.item.stateMask = 0;
		Item.item.pszText = LPSTR_TEXTCALLBACK;
		Item.item.cchTextMax = 0;
		Item.item.lParam = (DWORD) this;

		if (Tree AND Tree->ImageList)
		{
			Item.item.mask |= TVIF_IMAGE | TVIF_SELECTEDIMAGE;
			Item.item.iImage = I_IMAGECALLBACK;
			Item.item.iSelectedImage = I_IMAGECALLBACK;
		}

		hTreeItem = TreeView_InsertItem(Tree->Handle(), &Item);
		Status = (hTreeItem != NULL);
	}
	return Status;
}

bool GTreeItem::AttachAll()
{
	bool Status = TRUE;
	GTreeItem *Prev = 0;
	for (GTreeItem *i=Items.First(); i; i = Items.Next())
	{
		Status &= i->Attach(Prev);
		i->AttachAll();
		Prev = i;
	}

	return Status;
}

void GTreeItem::Update()
{
	if (Tree)
	{
		RECT rc;
		if (TreeView_GetItemRect(Tree->Handle(), hTreeItem, &rc, false))
		{
			Tree->Invalidate(&GRegion(rc), true);
		}
	}
}

bool GTreeItem::Selected()
{
	if (Tree)
	{
		TV_ITEM Item;
		
		ZeroObj(Item);
		Item.mask = TVIF_STATE;
		Item.hItem = hTreeItem;
		Item.stateMask = TVIS_SELECTED;
		
		if (TreeView_GetItem(Tree->Handle(), &Item))
		{
			return (Item.state & TVIS_SELECTED) != 0;
		}
	}
	return FALSE;
}

void GTreeItem::Selected(bool b)
{
	if (Tree)
	{
		TV_ITEM Item;
		
		ZeroObj(Item);
		Item.mask = TVIF_STATE;
		Item.hItem = hTreeItem;
		Item.state = (b) ? TVIS_SELECTED : 0;
		Item.stateMask = TVIS_SELECTED;
		
		TreeView_SetItem(Tree->Handle(), &Item);
	}
}

bool GTreeItem::Expanded()
{
	if (Tree)
	{
		TV_ITEM It;

		ZeroObj(It);
		It.mask = TVIF_STATE;
		It.stateMask = TVIS_EXPANDED;
		It.hItem = hTreeItem;
		if (TreeView_GetItem(Tree->Handle(), &It))
		{
			return (It.state & TVIS_EXPANDED) != 0;
		}
	}

	return FALSE;
}

void GTreeItem::Expanded(bool b)
{
	if (Tree)
	{
		TreeView_Expand(Tree->Handle(), hTreeItem, (b) ? TVE_EXPAND : TVE_COLLAPSE); 
	}
}

//////////////////////////////////////////////////////////////////////////////
GWin32Class GTree::WndClass;

GTree::GTree(int id, int x, int y, int cx, int cy, char *name) :
	GControl(&WndClass)
#ifdef LGI_RES
	, ResObject(Res_TreeView)
#endif
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	if (name) GControl::Name(name);

	Lines = TRUE;
	Buttons = TRUE;
	LinesAtRoot = TRUE;
	EditLabels = FALSE;
	Tree = this;

	WndClass.SubClass(Class(), WC_TREEVIEW);
}

GTree::~GTree()
{
	Tree = 0;

	/*
	if (hWnd)
	{
		HTREEITEM hItem = TreeView_GetRoot(hWnd);
		TV_ITEM Item;

		ZeroObj(Item);
		Item.mask = TVIF_PARAM;
		Item.hItem = hItem;
		if (hItem AND TreeView_GetItem(hWnd, &Item))
		{
			GTreeItem *i = (GTreeItem*) Item.lParam;
			if (i)
			{
				DeleteObj(i);
			}
		}
	}
	*/
}

int GTree::IndexOf(GTreeItem *Item)
{
	return Items.IndexOf(Item);
}

GTreeItem *GTree::ItemAt(int i)
{
	return Items.ItemAt(i);
}

void GTree::OnItemSelect(GTreeItem *Item)
{
	if (Item)
	{
		Item->OnSelect();
	}
}

void GTree::OnItemExpand(GTreeItem *Item, bool Expand)
{
	if (Item)
	{
		Item->OnExpand(Expand);
	}
}

void GTree::OnMouseClick(GMouse &m)
{
	TV_HITTESTINFO Info;

	Info.pt.x = m.x;
	Info.pt.y = m.y;
 
 	HTREEITEM hItem = TreeView_HitTest(hWnd, &Info);

	if (Info.flags & TVHT_ONITEM)
	{
		// over item
		TV_ITEM Item;

		ZeroObj(Item);
		Item.mask = TVIF_HANDLE | TVIF_PARAM;
		Item.hItem = Info.hItem;

		if (TreeView_GetItem(hWnd, &Item))
		{
			OnItemClick((GTreeItem*) Item.lParam, m);
		}		
	}
}

int GTree::OnEvent(GMessage *Msg)
{
	switch (Message(Msg))
	{
		case WM_DESTROY:
		{
			if (ImageList)
			{
				TreeView_SetImageList(hWnd, 0, LVSIL_SMALL);
			}
			break;
		}
	}

	int Status = GControl::OnEvent(Msg);

	switch (Msg->Msg)
	{
		case WM_CREATE:
		{
			if (ImageList)
			{
				TreeView_SetImageList(hWnd, ImageList->Handle(), TVSIL_NORMAL);
			}
			AttachAll();
			break;
		}
		case WM_LBUTTONDBLCLK:
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_RBUTTONDBLCLK:
		case WM_RBUTTONDOWN:
		case WM_RBUTTONUP:
		case WM_MBUTTONDBLCLK:
		case WM_MBUTTONDOWN:
		case WM_MBUTTONUP:
		{
			GView::OnEvent(Msg);
			break;
		}
	}

	return Status;
}

GTreeItem *GTree::Insert(GTreeItem *Obj, int Pos)
{
	GTreeItem *NewObj = GTreeItem::Insert(Obj, Pos);
	if (NewObj)
	{
		NewObj->Parent = 0;
		NewObj->Tree = this;
	}
	return NewObj;
}

bool GTree::Remove(GTreeItem *Obj)
{
	if (hWnd AND Obj)
	{
		if (Obj->Parent)
		{
			Obj->Parent->Items.Delete(Obj);
		}
		else
		{
			Items.Delete(Obj);
		}

		Obj->Tree = 0;
		Obj->Parent = 0;

		return TreeView_DeleteItem(hWnd, Obj->hTreeItem); 
	}
	return FALSE;
}

void GTree::RemoveAll()
{
	if (hWnd)
	{
		Items.Empty();
		TreeView_DeleteAllItems(hWnd); 
	}
}

bool GTree::Delete(GTreeItem *Obj)
{
	bool Status = FALSE;
	if (Obj AND Remove(Obj))
	{
		DeleteObj(Obj);
		Status = true;
	}
	return Status;
}

bool GTree::Select(GTreeItem *Obj)
{
	bool Status = FALSE;
	if (Obj AND hWnd)
	{
		TreeView_SelectItem(hWnd, Obj->ItemHandle());
		Status = TRUE;
	}
	return Status;
}

GTreeItem *GTree::Selection()
{
	GTreeItem *Item = 0;
	if (hWnd)
	{
		HTREEITEM hItem = TreeView_GetSelection(Tree->Handle());
		if (hItem)
		{
			TV_ITEM It;

			ZeroObj(It);
			It.mask = TVIF_PARAM;
			It.hItem = hItem;
			if (TreeView_GetItem(Tree->Handle(), &It))
			{
				Item = (GTreeItem*) It.lParam;
			}
		}
	}
	return Item;	
}

void GTree::OnItemClick(GTreeItem *Item, GMouse &m)
{
	if (Item)
	{
		Item->OnMouseClick(m);
	}
}

void GTree::OnItemBeginDrag(GTreeItem *Item, int Flags)
{
	if (Item)
	{
		GMouse m;
		ZeroObj(m);
		m.Flags = Flags;
		Item->OnBeginDrag(m);
	}
}

