/*hdr
**      FILE:           Gui.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           23/4/98
**      DESCRIPTION:    Win32 View Implementation
**
**      Copyright (C) 1998-2001, Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <time.h>
#include "Lgi.h"
#include "base64.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
// Helper Stuff
#include "zmouse.h"

#define SetKeyFlag(v, k, f)		if (GetKeyState(k)&0xFF00) { v |= f; }

GWin32Class GLayout::Class;
int MouseRollMsg = 0;
HHOOK hMouseHook = 0;
HWND hMouseOver = 0;

int _lgi_mouse_wheel_lines()
{
	OSVERSIONINFO Info;
	ZeroObj(Info);
	Info.dwOSVersionInfoSize = sizeof(Info);
	if (GetVersionEx(&Info) AND
		Info.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS AND
		Info.dwMajorVersion == 4 AND
		Info.dwMinorVersion == 0)
	{
       HWND hdlMSHWheel=NULL;
       UINT msgMSHWheelGetScrollLines=NULL;
       UINT uiMsh_WheelScrollLines;

       msgMSHWheelGetScrollLines = 
               RegisterWindowMessage(MSH_SCROLL_LINES);
       hdlMSHWheel = FindWindow(MSH_WHEELMODULE_CLASS, 
                                MSH_WHEELMODULE_TITLE);
       if (hdlMSHWheel && msgMSHWheelGetScrollLines)
       {
			return SendMessage(hdlMSHWheel, msgMSHWheelGetScrollLines, 0, 0);
       }
	}
	else
	{
		UINT nScrollLines;
		if (SystemParametersInfo(	SPI_GETWHEELSCROLLLINES, 
									0, 
									(PVOID) &nScrollLines, 
									0))
		{
			return nScrollLines;
		}
	}

	return 3;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK MouseHook(int nCode, WPARAM wParam, LPARAM lParam)
{
	// check whether the mouse has moved off the last window
	MOUSEHOOKSTRUCT *Ms = (MOUSEHOOKSTRUCT*) lParam;
	if (Ms)
	{
		RECT rc;
		GetClientRect(Ms->hwnd, &rc);
		ScreenToClient(Ms->hwnd, &Ms->pt);

		// is the mouse inside the client area?
		bool Inside = NOT	(Ms->pt.x < 0 OR
							Ms->pt.y < 0 OR
							Ms->pt.x >= rc.right-rc.left OR
							Ms->pt.y >= rc.bottom-rc.top);

		HWND hWnd = (Inside) ? Ms->hwnd : 0;

		if (hWnd != hMouseOver)
		{
			// window has changed
			if (hMouseOver AND IsWindow(hMouseOver))
			{
				// window is LOSING mouse
				SendMessage(hMouseOver, WM_MOUSEEXIT, 0, MAKELONG((short) Ms->pt.x, (short) Ms->pt.y));
			}

			// set current window
			hMouseOver = hWnd;

			if (hMouseOver AND IsWindow(hMouseOver))
			{
				// window is GETTING mouse
				SendMessage(hMouseOver, WM_MOUSEENTER, Inside, MAKELONG((short) Ms->pt.x, (short) Ms->pt.y));
			}
		}
	}

	return CallNextHookEx(hMouseHook, nCode, wParam, lParam);
}

GMouse _lgi_adjust_click_for_window(GMouse &Info, GView *Wnd)
{
	GMouse Temp = Info;
	if (Wnd)
	{
		GdcPt2 Offset(0, 0);
		if (Wnd->WindowVirtualOffset(&Offset))
		{
			Temp.x -= Offset.x;
			Temp.y -= Offset.y;
		}
	}
	return Temp;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK GWin32Class::Redir(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (m == WM_NCCREATE)
	{
		LPCREATESTRUCT Info = (LPCREATESTRUCT) b;
		
		GView *NewWnd = (GView*) Info->lpCreateParams;
		NewWnd->_View = hWnd;
		SetWindowLong(hWnd, GWL_USERDATA, (int) NewWnd);
	}

	GView *Wnd = (GView*) GetWindowLong(hWnd, GWL_USERDATA);
	if (Wnd)
	{
		GMessage Msg(m, a, b);
		return Wnd->OnEvent(&Msg);
	}

	return DefWindowProc(hWnd, m, a, b);
}

LRESULT CALLBACK GWin32Class::SubClassRedir(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (m == WM_NCCREATE)
	{
		LPCREATESTRUCT Info = (LPCREATESTRUCT) b;
		GView *NewWnd = 0;
		if (Info->lpCreateParams)
		{
			NewWnd = (GView*) Info->lpCreateParams;
			NewWnd->_View = hWnd;
		}
		SetWindowLong(hWnd, GWL_USERDATA, (int) NewWnd);
	}

	GView *Wnd = (GView*) GetWindowLong(hWnd, GWL_USERDATA);
	if (Wnd)
	{
		GMessage Msg(m, a, b);
		int Status = Wnd->OnEvent(&Msg);
		return Status;
	}

	return DefWindowProc(hWnd, m, a, b);
}

GWin32Class::GWin32Class()
{
	Class.style = CS_DBLCLKS; //  | CS_HREDRAW | CS_VREDRAW
	Class.lpfnWndProc = (WNDPROC) Redir;
	Class.cbClsExtra = 0;
	Class.cbWndExtra = 0;
	Class.hInstance = 0;
	Class.hIcon = NULL;
	Class.hCursor = 0; // LoadCursor(hAppInst(), MAKEINTRESOURCE(IDC_ARROW));
	Class.hbrBackground = NULL;
	Class.lpszMenuName = NULL;
	Class.lpszClassName = NULL;

	ParentProc = 0;
}

GWin32Class::~GWin32Class()
{
	UnregisterClass(Name(), hAppInst());
	Class.lpszClassName = NULL;
}

bool GWin32Class::Create(char *ClassName)
{
	bool Status = FALSE;

	if (NOT Class.lpszClassName)
	{
		Name(ClassName);
		Class.hInstance = hAppInst();
		Class.lpszClassName = ClassName;
		Status = RegisterClass(&Class) != 0;
	}
	else
	{
		Status = TRUE;
	}

	return Status;
}

bool GWin32Class::SubClass(char *ClassName, char *Parent)
{
	bool Status = FALSE;

	if (NOT Class.lpszClassName)
	{
		HBRUSH hBr = Class.hbrBackground;

		if (GetClassInfo(hAppInst(), Parent, &Class))
		{
			Name(ClassName);
			ParentProc = Class.lpfnWndProc;
			if (hBr)
			{
				Class.hbrBackground = hBr;
			}
			Class.hInstance = hAppInst();
			Class.lpfnWndProc = (WNDPROC) SubClassRedir;
			Class.lpszClassName = ClassName;
			Status = RegisterClass(&Class) != 0;
		}
	}
	else
	{
		return TRUE;
	}

	return Status;
}

LRESULT CALLBACK GWin32Class::CallParent(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (NOT ParentProc) return 0;
	#if _MSC_VER == 1100
	return CallWindowProc((FARPROC) ParentProc, hWnd, m, a, b);
	#else
	return CallWindowProc(ParentProc, hWnd, m, a, b);
	#endif
}

//////////////////////////////////////////////////////////////////////////////
GView *GWindowFromHandle(HWND hWnd)
{
	return dynamic_cast<GView*>((GView*) GetWindowLong(hWnd, GWL_USERDATA));
}

//////////////////////////////////////////////////////////////////////////////
GView *GView::Capturing = 0;
GView *GView::Over = 0;
HWND GView::hPrevCapture = 0;

GView::GView()
{
	_View = 0;
	CtrlId = -1;

	WndFlags = 0;
	WndStyle = WS_CHILD | WS_VISIBLE;
	WndExStyle = 0;
	WndClass = 0;
	WndDlgCode = 0;
	_Win32_TimerId = 0;

	Pos.ZOff(0, 0);
	Parent = 0;
	Notify = 0;
	_MinimumSize.x = _MinimumSize.y = 0;
}

GView::~GView()
{
	// Delete myself out of my parent's list
	if (Parent)
	{
		Parent->Children.Delete(this);
	}

	// Delete all children
	GView *c;
	for (c = Children.First(); c; c = Children.Next())
	{
		ASSERT(c->Parent == this OR c->Parent == 0);
	}

	while (c = Children.First())
	{
		ASSERT(c->Parent == this);
		DeleteObj(c);
	}

	// Delete the OS representation of myself
	if (_View AND IsWindow(_View))
	{
		WndFlags |= GWF_DESTRUCTOR;
		DestroyWindow(_View);
	}
	
	// NULL my handles and flags
	_View = 0;
	WndFlags = 0;

	// Remove static references to myself
	if (Over == this) Over = 0;
	if (Capturing == this) Capturing = 0;
	if (LgiApp->AppWnd == this)
	{
		LgiApp->AppWnd = 0;
	}
}

void GView::OnPosChange()
{
}

void GView::_Paint(GSurface *pDC)
{
	// Paint me
	OnPaint(pDC);

	// Paint any virtual children
	int x, y;
	pDC->GetOrigin(x, y);
	for (GView *w = Children.First(); w; w = Children.Next())
	{
		if (NOT w->Handle())
		{
			GRegion r = w->GetPos();
			GRegion c = r;

			c.Offset(-x, -y);
			c.Intersection(&pDC->ClipRgn());
			pDC->ClipRgn(&c);
			pDC->SetOrigin(x-r.x1, y-r.y1);

			w->_Paint(pDC);

			pDC->SetOrigin(x, y);
			pDC->ClipRgn(NULL);
		}
	}
}

#define ADJ_LEFT			1
#define ADJ_RIGHT			2
#define ADJ_UP				3
#define ADJ_DOWN			4

int IsAdjacent(GRegion &a, GRegion &b)
{
	if (	(a.x1 == b.x2 + 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_LEFT;
	}

	if (	(a.x2 == b.x1 - 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_RIGHT;
	}

	if (	(a.y1 == b.y2 + 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_UP;
	}

	if (	(a.y2 == b.y1 - 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_DOWN;
	}

	return 0;
}

GRegion JoinAdjacent(GRegion &a, GRegion &b, int Adj)
{
	GRegion t;

	switch (Adj)
	{
		case ADJ_LEFT:
		case ADJ_RIGHT:
		{
			t.y1 = max(a.y1, b.y1);
			t.y2 = min(a.y2, b.y2);
			t.x1 = min(a.x1, b.x1);
			t.x2 = max(a.x2, b.x2);
			break;
		}
		case ADJ_UP:
		case ADJ_DOWN:
		{
			t.y1 = min(a.y1, b.y1);
			t.y2 = max(a.y2, b.y2);
			t.x1 = max(a.x1, b.x1);
			t.x2 = min(a.x2, b.x2);
			break;
		}
	}

	return t;
}

GRegion *GView::FindLargest(GComplexRegion &r)
{
	int Pixels = 0;
	GRegion *Best = 0;
	static GRegion Final;

	// do initial run through the list to find largest single region
	for (GRegion *i = r.First(); i; i = r.Next())
	{
		int Pix = i->X() * i->Y();
		if (Pix > Pixels)
		{
			Pixels = Pix;
			Best = i;
		}
	}

	if (Best)
	{
		Final = *Best;
		Pixels = Final.X() * Final.Y();

		int LastPixels = Pixels;
		GRegion LastRgn = Final;
		int ThisPixels = Pixels;
		GRegion ThisRgn = Final;

		GComplexRegion TempList;
		for (GRegion *i = r.First(); i; i = r.Next())
		{
			TempList.Union(i);
		}
		TempList.Subtract(Best);

		do
		{
			LastPixels = ThisPixels;
			LastRgn = ThisRgn;

			// search for adjoining rectangles that maybe we can add
			for (GRegion *i = TempList.First(); i; i = TempList.Next())
			{
				int Adj = IsAdjacent(ThisRgn, *i);
				if (Adj)
				{
					GRegion t = JoinAdjacent(ThisRgn, *i, Adj);
					int Pix = t.X() * t.Y();
					if (Pix > ThisPixels)
					{
						ThisPixels = Pix;
						ThisRgn = t;

						TempList.Subtract(i);
					}
				}
			}
		} while (LastPixels < ThisPixels);

		Final = ThisRgn;
	}

	return &Final;
}

GRegion *GView::FindSmallestFit(GComplexRegion &r, int Sx, int Sy)
{
	int X = 1000000;
	int Y = 1000000;
	GRegion *Best = 0;

	for (GRegion *i = r.First(); i; i = r.Next())
	{
		if (	(i->X() >= Sx AND i->Y() >= Sy) AND
			(i->X() < X OR i->Y() < Y))
		{
			X = i->X();
			Y = i->Y();
			Best = i;
		}
	}

	return Best;
}

GRegion *GView::FindLargestEdge(GComplexRegion &r, int Edge)
{
	GRegion *Best = 0;

	for (GRegion *i = r.First(); i; i = r.Next())
	{
		if (NOT Best)
		{
			Best = i;
		}

		if ((Edge & GV_EDGE_TOP) AND (i->y1 < Best->y1) OR
			(Edge & GV_EDGE_RIGHT) AND (i->x2 > Best->x2) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 > Best->y2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 < Best->x1))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_TOP) AND (i->y1 == Best->y1) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 == Best->y2)) AND
			(i->X() > Best->X()))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_RIGHT) AND (i->x2 == Best->x2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 == Best->x1)) AND
			(i->Y() > Best->Y()))
		{
			Best = i;
		}
	}

	return Best;
}

void GView::Quit(bool DontDelete)
{
	if (_View)
	{
		if (NOT DontDelete)
		{
			WndFlags |= GWF_QUIT_WND;
		}

		DestroyWindow(_View);
	}
}

uint32 GView::Style()
{
	return WndStyle;
}

uint32 GView::ExStyle()
{
	return WndExStyle;
}

char *GView::Class()
{
	return WndClass;
}

GView *GView::FindReal(GdcPt2 *Offset)
{
	if (Offset)
	{
		Offset->x = 0;
		Offset->y = 0;
	}

	GView *p = Parent;
	while (p AND NOT p->Handle())
	{
		if (Offset)
		{
			Offset->x += Pos.x1;
			Offset->y += Pos.y1;
		}

		p = p->Parent;
	}
	
	if (p AND p->Handle())
	{
		return p;
	}

	return NULL;
}

bool GView::Capture(bool c)
{
	return HandleCapture(this, c);
}

bool GView::IsCapturing()
{
	return Capturing != 0;
}

bool GView::HandleCapture(GView *Wnd, bool c)
{
	bool Status = FALSE;

	if (_View)
	{
		if (c)
		{
			if (Capturing)
			{
				Capturing->HandleCapture(NULL, FALSE);
			}

			hPrevCapture = SetCapture(_View);
			Status = TRUE;
			Capturing = Wnd;
		}
		else
		{
			if (Capturing)
			{
				ReleaseCapture();
				Capturing = 0;
			}
		}
	}
	else
	{
		if (Parent)
		{
			Status = Parent->HandleCapture(Wnd, c);
		}
	}

	return Status;
}

bool GView::Enabled()
{
	if (_View)
	{
		return IsWindowEnabled(_View);
	}
	return NOT TestFlag(WndStyle, WS_DISABLED);
}

void GView::Enabled(bool i)
{
	if (NOT i) SetFlag(WndStyle, WS_DISABLED);
	else ClearFlag(WndStyle, WS_DISABLED);

	if (_View)
	{
		EnableWindow(_View, i);
	}
	else
	{
		Invalidate();
	}
}

bool GView::Visible()
{
	if (_View)
	{
		return IsWindowVisible(_View);
	}

	return TestFlag(WndStyle, WS_VISIBLE);
}

void GView::Visible(bool i)
{
	if (i) SetFlag(WndStyle, WS_VISIBLE);
	else ClearFlag(WndStyle, WS_VISIBLE);

	if (_View)
	{
		ShowWindow(_View, (i) ? SW_SHOWNORMAL : SW_HIDE);
	}
	Invalidate();
}

bool GView::Focus()
{
	if (_View)
	{
		return GetFocus() == _View;
	}
	return TestFlag(WndFlags, GWF_FOCUS);
}

void GView::Focus(bool i)
{
	if (i) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);

	if (_View)
	{
		SetFocus((i) ? _View : GetDesktopWindow());
	}
	else
	{
		Invalidate();
	}
}

bool GView::Sunken()
{
	return TestFlag(WndExStyle, WS_EX_CLIENTEDGE);
}

void GView::Sunken(bool i)
{
	ClearFlag(WndExStyle, WS_EX_WINDOWEDGE);
	
	if (i) SetFlag(WndExStyle, WS_EX_CLIENTEDGE);
	else ClearFlag(WndExStyle, WS_EX_CLIENTEDGE);
}

bool GView::Flat()
{
	return	NOT TestFlag(WndExStyle, WS_EX_CLIENTEDGE) AND
			NOT TestFlag(WndExStyle, WS_EX_WINDOWEDGE);
}

void GView::Flat(bool i)
{
	ClearFlag(WndExStyle, WS_EX_CLIENTEDGE);
	ClearFlag(WndExStyle, WS_EX_WINDOWEDGE);
}

bool GView::Raised()
{
	return TestFlag(WndExStyle, WS_EX_WINDOWEDGE);
}

void GView::Raised(bool i)
{
	ClearFlag(WndExStyle, WS_EX_CLIENTEDGE);
	
	if (i) SetFlag(WndExStyle, WS_EX_WINDOWEDGE);
	else ClearFlag(WndExStyle, WS_EX_WINDOWEDGE);
}

int GView::GetCtrlValue(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Value() : -1;
}

void GView::SetCtrlValue(int Id, int i)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Value(i);
	}
}

char *GView::GetCtrlName(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Name() : 0;
}

void GView::SetCtrlName(int Id, char *s)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Name(s);
	}
}

bool GView::GetCtrlEnabled(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Enabled() : 0;
}

void GView::SetCtrlEnabled(int Id, bool Enabled)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Enabled(Enabled);
	}
}

void GView::SetId(int i)
{
	CtrlId = i;
	if (_View)
	{
		SetWindowLong(_View, GWL_ID, CtrlId);
	}
}

bool GView::Create(char *Class, OsView Parent)
{
	CreateWindowEx(	ExStyle(),
					Class,
					Name(),
					Style(),
					Pos.x1, Pos.y1,
					Pos.X(), Pos.Y(),
					Parent,
					NULL,
					hAppInst(),
					this);

	return _View != NULL;
}

bool GView::IsAttached()
{
	return _View AND IsWindow(_View);
}

bool GView::Attach(GView *p)
{
	bool Status = false;
	if (p)
	{
		Parent = p;

		if (Class())
		{
			// Real window with HWND
			if (Create(Class(), p->Handle()))
			{
				Status = (_View != 0);
			}
		}
		else
		{
			// Virtual window (no HWND)
			Status = true;
		}

		if (Status)
		{
			if (NOT p->Children.HasItem(this))
			{
				Parent->Children.Insert(this);
			}
			Parent->OnChildrenChanged(this, true);
		}
		else
		{
			Parent = 0;
		}
	}
	return Status;
}

bool GView::AttachChildren()
{
	Iterator<GView> ChildLst(&Children);
	for (GView *c = ChildLst.First(); c; c = ChildLst.Next())
	{
		if (NOT c->Attach(this))
		{
			return false;
		}
	}

	return true;
}

bool GView::Detach()
{
	bool Status = false;
	if (Parent)
	{
		Visible(false);
		Parent->Children.Delete(this);
		Parent->OnChildrenChanged(this, false);
		Parent = 0;
		Status = true;

		if (_View)
		{
			WndFlags &= ~GWF_QUIT_WND;
			DestroyWindow(_View);
		}
	}
	return Status;
}

GRegion &GView::GetClient()
{
	RECT rc;

	if (_View)
	{
		GetClientRect(_View, &rc);
	}
	else
	{
		rc = Pos;
	}

	static GRegion Client;
	Client = rc;
	return Client;
}

GFont *GView::GetFont()
{
	/*
	GFontRef Fnt = 0; 
	if (hWnd)
	{
		Fnt = (GFontRef) SendMessage(hWnd, WM_GETFONT, 0, 0);
	}

	return Fnt;
	*/

	return SysFont;
}

void GView::SetFont(GFont *Font)
{
	if (_View)
	{
		SendMessage(_View, WM_SETFONT, (WPARAM) Font->Handle(), 0);
	}
}

bool GView::IsOver(GMouse &m)
{
	return	(m.x >= 0) AND
			(m.y >= 0) AND
			(m.x < Pos.X()) AND
			(m.y < Pos.Y());
}

bool GView::GetMouse(GMouse &m, bool ScreenCoords)
{
	// position
	POINT p;
	GetCursorPos(&p);
	if (NOT ScreenCoords)
	{
		ScreenToClient(_View, &p);
	}
	m.x = p.x;
	m.y = p.y;

	// buttons
	m.Flags =	((GetAsyncKeyState(VK_LBUTTON)&0x8000) ? MK_LEFT : 0) |
				((GetAsyncKeyState(VK_MBUTTON)&0x8000) ? MK_MIDDLE : 0) |
				((GetAsyncKeyState(VK_RBUTTON)&0x8000) ? MK_RIGHT : 0) |
				((GetAsyncKeyState(VK_CONTROL)&0x8000) ? MK_CTRL : 0) |
				((GetAsyncKeyState(VK_MENU)&0x8000) ? MK_ALT : 0) |
				((GetAsyncKeyState(VK_SHIFT)&0x8000) ? MK_SHIFT : 0);

	return true;
}

bool GView::WindowVirtualOffset(GdcPt2 *Offset)
{
	bool Status = false;

	if (Offset)
	{
		Offset->x = 0;
		Offset->y = 0;
		GView *Wnd = this;
		while (Wnd)
		{
			GRegion r = Wnd->GetPos();
			GView *Parent = Wnd->GetParent();
			if (NOT Wnd->Handle())
			{
				Offset->x += r.x1;
				Offset->y += r.y1;
				Wnd = Parent;
				Status = true;
			}
			else
			{
				Wnd = 0;
			}
		}
	}

	return Status;
}

GView *GView::WindowFromPoint(int x, int y)
{
	for (GView *c = Children.First(); c; c = Children.Next())
	{
		GRegion r = c->GetPos();
		if (r.Overlap(x, y) AND
			c->Visible())
		{
			GView *Child = c->WindowFromPoint(x - r.x1, y - r.y1);
			if (Child)
			{
				return Child;
			}
		}
	}

	if (x >= 0 AND y >= 0 AND x < Pos.X() AND y < Pos.Y())
	{
		return this;
	}

	return NULL;
}

bool GView::SetPos(GRegion &p, bool Repaint)
{
	bool Status = TRUE;
	Pos = p;
	if (_View)
	{
		Status = SetWindowPos(	_View,
								NULL,
								Pos.x1,
								Pos.y1,
								Pos.X(),
								Pos.Y(),
								((Repaint) ? 0 : SWP_NOREDRAW) |
								SWP_NOZORDER);
	}
	else if (Repaint)
	{
		Invalidate();
	}
	return Status;
}

bool GView::InThread()
{
	DWORD CurThread = GetCurrentThreadId();
	DWORD WndThread = GetWindowThreadProcessId(_View, NULL);
	
	return CurThread == WndThread;
}

bool GView::PostEvent(int Cmd, int a, int b)
{
	return PostMessage(_View, Cmd, a, b);
}

bool GView::Invalidate(GRegion *r, bool Repaint)
{
	if (_View)
	{
		bool Status = InvalidateRect(_View, (r) ? &((RECT)*r) : NULL, Repaint);

		if (Repaint)
		{
			UpdateWindow(_View);
		}

		return Status;

		/*
		else
		{
			int a = (r) ? (r->x1 << 16) | (r->y1 & 0xffff) : 0xffffffff;
			int b = (r) ? (r->x2 << 16) | (r->y2 & 0xffff) : 0xffffffff;
			PostEvent(WM_INVALIDATE, a, b);
			if (r)
			{
				LgiTrace("sending WM_INVALIDATE from thread %X, hWnd %i, region(%i, %i, %i, %i)", LgiGetCurrentThread(), hWnd, r->x1, r->y1, r->x2, r->y2);
			}
			else
			{
				LgiTrace("sending WM_INVALIDATE from thread %X, hWnd %i, region(0)", LgiGetCurrentThread(), hWnd);
			}
		}
		*/
	}
	else
	{
		GRegion Up;
		GView *p = this;

		if (r)
		{
			Up = *r;
		}
		else
		{
			Up.Set(0, 0, Pos.X()-1, Pos.Y()-1);
		}

		while (p AND NOT p->Handle())
		{
			Up.Offset(p->Pos.x1, p->Pos.y1);
			p = p->Parent;
		}

		if (p AND p->Handle())
		{
			return p->Invalidate(&Up);
		}
	}
	return FALSE;
}

bool GView::InvalidateFrame(GRegion *r, bool Repaint)
{
	if (_View)
	{
		RedrawWindow(	_View,
						NULL,
						NULL,
						RDW_FRAME |
						RDW_INVALIDATE |
						RDW_ALLCHILDREN |
						((Repaint) ? RDW_UPDATENOW : 0));

		return TRUE;
	}

	return FALSE;
}

void
CALLBACK
GView::TimerProc(HWND hwnd, UINT uMsg, UINT idEvent, uint32 dwTime)
{
	GView *View = (GView*) idEvent;
	if (View)
	{
		View->OnPulse();
	}
}

void GView::SetPulse(int Length)
{
	if (_View)
	{
		if (Length > 0)
		{
			_Win32_TimerId = SetTimer(_View, (UINT) this, Length, (TIMERPROC) TimerProc);
		}
		else
		{
			KillTimer(_View, _Win32_TimerId);
			_Win32_TimerId = 0;
		}
	}
}

int GView::OnNotify(GView *Ctrl, int Flags)
{
	if (Ctrl AND Parent)
	{
		// default behaviour is just to pass the 
		// notification up to the parent
		return Parent->OnNotify(Ctrl, Flags);
	}

	return 0;
}

GButton *FindDefault(GView *w)
{
	for (GView *c = w->Children.First(); c; c = w->Children.Next())
	{
		GButton *But = dynamic_cast<GButton*>(c);
		if (But AND But->Default())
		{
			return But;
		}
		
		But = FindDefault(c);
		if (But)
		{
			return But;
		}
	}

	return NULL;
}

List<GView> *GetChildrenList(GView *w)
{
	if (w)
	{
		GTabPage *Page = dynamic_cast<GTabPage*>(w);
		GTabView *TabCtrl = dynamic_cast<GTabView*>(Page ? Page->GetParent() : w);
		if (TabCtrl)
		{
			GTabPage *CurTab = TabCtrl->TabAt(TabCtrl->Value());
			if (CurTab)
			{
				return &CurTab->Children;
			}
		}
		else
		{
			return &w->Children;
		}
	}

	return 0;
}

GView *GetNextTabStop(GView *Wnd, bool Back)
{
	GView *w = Wnd;
	do
	{
		// move
		if (w AND w->GetParent())
		{
			List<GView> *CList = GetChildrenList(w);
			List<GView> *PList = GetChildrenList(w->GetParent());
			if (CList AND PList)
			{
				int Me = PList->IndexOf(w);
				if (Back)
				{
					GView *Prev = PList->ItemAt(Me-1);
					while (Prev)
					{
						GView *w = Prev->Children.ItemAt(Prev->Children.GetItems()-1);
						if (w)
						{
							Prev = w;
						}
						else break;
					}

					w = Prev;
				}
				else
				{
					GView *Next = CList->First();
					if (Next)
					{
						// Make child the next focus
						w = Next;
					}
					else
					{
						// No children
						Next = PList->ItemAt(Me+1);
						if (Next)
						{
							// Make next sibling the focus...
							w = Next;
						}
						else
						{
							// No sibling
							while (w)
							{
								GView *Parent = w->GetParent();
								if (dynamic_cast<GWindow*>(Parent) OR
									dynamic_cast<GDialog*>(Parent))
								{
									int Index = Parent->Children.IndexOf(w);
									Next = Parent->Children.ItemAt(Index+1);
									if (NOT Next)
									{
										Next = Parent->Children.ItemAt(0);
									}
									if (Next)
									{
										w = Next;
									}
									break;
								}
								else if (Parent)
								{
									List<GView> *Lst = GetChildrenList(Parent);
									int Index = (Lst) ? Lst->IndexOf(w) : -1;
									Next = (Lst AND Index >= 0) ? Lst->ItemAt(Index+1) : 0;
									if (Next AND Lst)
									{
										w = Next;
										break;
									}
									else
									{
										w = Parent;
									}
								}
								else break;
							}
						}
					}
				}
			}
			else
			{
				w = 0;
			}
		}
		else
		{
			break;
		}

		// test
		if (w AND
			w->Enabled() AND
			TestFlag(w->Style(), WS_TABSTOP))
		{
			return w;
		}
	}
	while (w AND w != Wnd);

	return 0;
}

bool SysOnKey(GView *w, GMessage *m)
{
	if (m->a == VK_TAB AND
		(m->Msg == WM_KEYDOWN OR
		m->Msg == WM_SYSKEYDOWN) )
	{
		if (NOT TestFlag(w->WndDlgCode, DLGC_WANTTAB) AND
			NOT TestFlag(w->WndDlgCode, DLGC_WANTALLKEYS))
		{
			// push the focus to the next control
			bool Shifted = (GetAsyncKeyState(VK_SHIFT) & 0x8000) != 0;
			GView *Wnd = GetNextTabStop(w, Shifted);
			if (Wnd)
			{
				::SetFocus(Wnd->Handle());
				return true;
			}
		}
	}

	return false;
}


#define KEY_FLAGS		(~(MK_LBUTTON | MK_MBUTTON | MK_RBUTTON))

int GView::OnEvent(GMessage *Msg)
{
	int Status = 0;

	if (MsgCode(Msg) == MouseRollMsg)
	{
		HWND hFocus = GetFocus();
		if (_View)
		{
			int Flags = ((GetKeyState(VK_SHIFT)&0xF000) ? VK_SHIFT : 0) | 
						((GetKeyState(VK_CONTROL)&0xF000) ? VK_CONTROL : 0);
			
			PostMessage(hFocus, WM_MOUSEWHEEL, MAKELONG(Flags, (short)Msg->a), Msg->b);
		}
		return 0;
	}

	if (_View)
	{
		switch (Msg->Msg)
		{
			/*
			case WM_INVALIDATE:
			{
				LgiTrace("receiving WM_INVALIDATE in thread %X, hWnd %i", LgiGetCurrentThread(), hWnd);
				if (Msg->a != 0xffffffff)
				{
					GRegion r;
					r.x1 = Msg->a >> 16;
					r.y1 = Msg->a & 0xffff;
					r.x2 = Msg->b >> 16;
					r.y2 = Msg->b & 0xffff;

					RECT rc = r;
					InvalidateRect(hWnd, &rc, true);
				}
				else
				{
					InvalidateRect(hWnd, 0, true);
				}
				break;
			}
			*/
			case WM_MENUCHAR:
			case WM_MEASUREITEM:
			case WM_DRAWITEM:
			{
				return GMenu::_OnEvent(Msg);
				break;
			}
			case WM_ENABLE:
			{
				Invalidate();
				break;
			}
			case WM_HSCROLL:
			case WM_VSCROLL:
			{
				GView *Wnd = dynamic_cast<GView*>(FindControl((HWND) Msg->b));
				if (Wnd)
				{
					Wnd->OnEvent(Msg);
				}
				break;
			}
			case WM_GETDLGCODE:
			{
				// we handle all tab control stuff
				return WndDlgCode | DLGC_WANTTAB;
			}
			case WM_MOUSEWHEEL:
			{
				short fwKeys = LOWORD(Msg->a);			// key flags
				short zDelta = (short) HIWORD(Msg->a);	// wheel rotation
				short xPos = (short) LOWORD(Msg->b);	// horizontal position of pointer
				short yPos = (short) HIWORD(Msg->b);	// vertical position of pointer
				int nScrollLines = - _lgi_mouse_wheel_lines();
				OnMouseWheel( ((int) zDelta / WHEEL_DELTA) * (int) nScrollLines );
				return 0;
			}
			case WM_CHANGE:
			{
				GView *Ctrl = dynamic_cast<GView*>((GView*) Msg->a);
				if (Ctrl)
				{
					OnNotify(Ctrl, Msg->b);
				}
				break;
			}
			case WM_NOTIFY:
			{
				LPNMHDR Hdr = (LPNMHDR) Msg->b;
				GView *Ctrl = FindControl(Hdr->hwndFrom);

				int i1 = LVN_GETDISPINFO;
				if (Ctrl) 
				{
					switch (Hdr->code)
					{
						#ifdef __GLIST_H
						// List View Stuff
						//
						// this is only required when using the Win32 ListView control
						// in GList.h/GList.cpp
						// for xp compatibility use the owner draw list control in
						// GList2.h/GList2.cpp
						case LVN_GETDISPINFOW:
						case LVN_GETDISPINFOA:
						{
							bool Unicode = Hdr->code == LVN_GETDISPINFOW;
							LV_DISPINFO *Info = (LV_DISPINFO*) Msg->b;
							GListItem *Item = (GListItem*) Info->item.lParam;

							if (Item)
							{
								if (Info->item.mask & LVIF_TEXT)
								{
									if (Info->item.pszText)
									{
										char *Text = Item->GetText(Info->item.iSubItem);
										if (Text)
										{
											if (Unicode)
											{
												short *d = (short*) Info->item.pszText;
												char *s = Text;
												int Size = Info->item.cchTextMax - 2;
												while (*s AND Size > 1)
												{
													*d++ = *s++;
													Size -= 2;
												}
												*d++ = 0;
											}
											else
											{
												strcpy(Info->item.pszText, Text);
											}
										}
										else
										{
											Info->item.pszText[0] = 0;
										}
									}
									else
									{
										if (Unicode)
										{
											static short Error[] = { 'E', 'r', 'r', 'o', 'r', 0 };
											Info->item.pszText = (char*) Error;
										}
										else
										{
											// what the?
											// don't know what causes windows to call using a NULL
											// pointer. Any ideas ppl?
											Info->item.pszText = Item->GetText(Info->item.iSubItem);
										}
									}
								}

								if (Info->item.mask & LVIF_IMAGE)
								{
									Info->item.iImage = Item->GetImage();
								}
							}
							break;
						}
						case LVN_ENDLABELEDIT:
						{
							LV_DISPINFO *Info = (LV_DISPINFO*) Msg->b;
							GListItem *Item = (GListItem*) Info->item.lParam;
							if (Info AND Item)
							{
								if (Info->item.mask & LVIF_TEXT)
								{
									return Item->SetText(Info->item.pszText);
								}
							}
							break;
						}
						case LVN_BEGINDRAG:
						case LVN_BEGINRDRAG:
						{
							NM_LISTVIEW *Info = (NM_LISTVIEW*) Msg->b;
							LV_ITEM Item;

							ZeroObj(Item);
							Item.mask = LVIF_PARAM;
							Item.iItem = Info->iItem;
							if (ListView_GetItem(Ctrl->Handle(), &Item))
							{
								GMouse m;
								m.Flags = (Hdr->code == LVN_BEGINDRAG) ? MK_LBUTTON : MK_RBUTTON;
								((GList*) Ctrl)->OnItemBeginDrag((GListItem*) Item.lParam, m);
							}
							break;
						}
						case LVN_COLUMNCLICK:
						{
							NM_LISTVIEW *Info = (NM_LISTVIEW*) Msg->b;
							((GList*) Ctrl)->OnColumnClick(Info->iSubItem);
							break;
						}
						case LVN_ITEMCHANGED:
						{
							LPNMLISTVIEW Info = (LPNMLISTVIEW) Msg->b;
							if ((Info->uNewState | Info->uOldState) & LVIS_SELECTED)
							{
								GList *Lst = (GList*) Ctrl;
								if (Info->uNewState & LVIS_SELECTED)
								{
									Lst->OnItemSelect((GListItem*) Info->lParam);
								}
								else
								{
									Lst->OnItemSelect(0);
								}
							}
							break;
						}
						#endif

						// Tab Control Stuff
						case TCN_SELCHANGING:
						{
							Status = Ctrl->SysOnNotify(Hdr->code);
							break;
						}
						case TCN_SELCHANGE:
						{
							Ctrl->SysOnNotify(Hdr->code);
							OnNotify(Ctrl, 0);
							break;
						}

						#ifdef __GTREE_H
						// Tree View Control
						//
						// this is only relevent when using the Win32 Tree control
						// which resides in GTree.h and GTree.cpp
						// for xp compatibility use the owner draw tree control
						// in GTree2.h and GTree2.cpp
	 					case TVN_BEGINRDRAG:
						case TVN_BEGINDRAG:
						{
							NM_TREEVIEW *Item = (NM_TREEVIEW*) Msg->b;
							((GTree*) Ctrl)->OnItemBeginDrag((GTreeItem*) Item->itemNew.lParam, (Hdr->code == TVN_BEGINDRAG) ? MK_LBUTTON : MK_RBUTTON);
							break;
						}
						case TVN_SELCHANGED:
						{
							NM_TREEVIEW *Item = (NM_TREEVIEW*) Msg->b;
							((GTree*) Ctrl)->OnItemSelect((GTreeItem*) Item->itemNew.lParam);
							break;
						}
						case TVN_ITEMEXPANDED:
						{
							NM_TREEVIEW *Item = (NM_TREEVIEW*) Msg->b;
							((GTree*) Ctrl)->OnItemExpand((GTreeItem*) Item->itemNew.lParam,
															(Item->itemNew.state & TVIS_EXPANDED) != 0);
							break;
						}
						case TVN_GETDISPINFO:
						{
							TV_DISPINFO *Info = (TV_DISPINFO*) Msg->b;
							GTreeItem *Item = (GTreeItem*) Info->item.lParam;
							if (Item)
							{
								if (Info->item.mask & TVIF_TEXT)
								{
									char *Text = Item->GetText();
									if (NOT Text) Text = "";
									if (Info->item.pszText)
									{
										strcpy(Info->item.pszText, Text);
									}
									else
									{
										Info->item.pszText = Text;
									}								
								}

								if (Info->item.mask & TVIF_IMAGE)
								{
									Info->item.iImage = Item->GetImage();
								}
								
								if (Info->item.mask & TVIF_SELECTEDIMAGE)
								{
									Info->item.iSelectedImage = Item->GetImage(1);
								}							
							}
							break;
						}
						case TVN_DELETEITEM:
						{
							NM_TREEVIEW *pnmtv = (NM_TREEVIEW*) Msg->b;
							if (pnmtv)
							{
								GTreeItem *Item = (GTreeItem*) pnmtv->itemOld.lParam;
								if (Item)
								{
									Item->hTreeItem = 0;
								}
							}
							break;
						}
						#endif
					}
				}
				break;
			}
			case WM_COMMAND:
			{
				GView *Ctrl = FindControl((HWND) Msg->b);
				if (Ctrl)
				{
					short Code = HIWORD(Msg->a);
					switch (Code)
					{
						case BN_CLICKED: // BUTTON
						{
							OnNotify(Ctrl, 0);
							break;
						}
						case EN_CHANGE: // EDIT
						{
							int n = Ctrl->SysOnNotify(Code);
							OnNotify(Ctrl, n);
							break;
						}
						case CBN_CLOSEUP:
						{
							PostMessage(_View, WM_COMMAND, MAKELONG(Ctrl->GetId(), CBN_EDITCHANGE), Msg->b);
							break;
						}
						case CBN_EDITCHANGE: // COMBO
						{
							Ctrl->SysOnNotify(Code);
							OnNotify(Ctrl, 0);
							break;
						}
					}
				}
				break;
			}
			case WM_NCDESTROY:
			{
				SetWindowLong(_View, GWL_USERDATA, 0);
				_View = 0;
				if (WndFlags & GWF_QUIT_WND)
				{
					delete this;
				}
				break;
			}
	 		case WM_CLOSE:
			{
				if (OnRequestClose())
				{
					Quit();
				}
				break;
			}
			case WM_DESTROY:
			{
				OnDestroy();
				break;
			}
			case WM_CREATE:
			{
				SetId(CtrlId);
				OnCreate();
				break;
			}
			case WM_SETFOCUS:
			{
				OnFocus(true);
				break;
			}
			case WM_KILLFOCUS:
			{
				OnFocus(false);
				break;
			}
			case WM_WINDOWPOSCHANGED:
			{
				WINDOWPOS *Info = (LPWINDOWPOS) Msg->b;
				if (Info)
				{
					Pos.ZOff(Info->cx-1, Info->cy-1);
					Pos.Offset(Info->x, Info->y);
				}
				OnPosChange();

				if (NOT (WndFlags & GWF_DIALOG))
				{
					return DefWindowProc(_View, Msg->Msg, Msg->a, Msg->b);
				}
				break;
			}
			case WM_CAPTURECHANGED:
			{
				Capturing = 0;
				break;
			}
			case WM_MOUSEENTER:
			{
				if (Msg->a AND NOT TestFlag(WndFlags, GWF_OVER))
				{
					GMouse Ms;
					Ms.x = (short) (Msg->b&0xFFFF);
					Ms.y = (short) (Msg->b>>16);
					Ms.Flags = 0;

					GView *MouseOver = WindowFromPoint(Ms.x, Ms.y);
					if (Capturing)
					{
						if (MouseOver == Capturing)
						{
							Capturing->OnMouseEnter(_lgi_adjust_click_for_window(Ms, Capturing));
						}
					}
					else
					{
						if (Over != MouseOver)
						{
							if (Over)
							{
								Over->OnMouseExit(_lgi_adjust_click_for_window(Ms, Over));
							}

							Over = MouseOver;

							if (Over)
							{
								Over->OnMouseEnter(_lgi_adjust_click_for_window(Ms, Over));
							}
						}
					}

					SetFlag(WndFlags, GWF_OVER);
				}
				break;
			}
			case WM_MOUSEEXIT:
			{
				if (NOT Msg->a AND TestFlag(WndFlags, GWF_OVER))
				{
					GMouse Ms;
					Ms.x = (short) (Msg->b&0xFFFF);
					Ms.y = (short) (Msg->b>>16);
					Ms.Flags = 0;

					if (Capturing)
					{
						Capturing->OnMouseExit(_lgi_adjust_click_for_window(Ms, Capturing));
					}
					else
					{
						if (Over)
						{
							Over->OnMouseExit(_lgi_adjust_click_for_window(Ms, Over));
							Over = 0;
						}
						else
						{
							OnMouseExit(Ms);
						}
					}

					ClearFlag(WndFlags, GWF_OVER);
				}
				break;
			}
			case WM_MOUSEMOVE:
			{
				GMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = Msg->a;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down((Msg->a & (MK_LBUTTON|MK_MBUTTON|MK_RBUTTON)) != 0);

				int Hit = OnHitTest(Ms.x, Ms.y);
				if (Hit < 0)
				{
					SetCursor(LgiApp->hNormalCursor);
				}

				GView *MouseOver = WindowFromPoint(Ms.x, Ms.y);
				if (Over != MouseOver)
				{
					if (Over)
					{
						Over->OnMouseExit(_lgi_adjust_click_for_window(Ms, Over));
					}

					Over = MouseOver;

					if (Over)
					{
						Over->OnMouseEnter(_lgi_adjust_click_for_window(Ms, Over));
					}
				}

				if (Capturing)
				{
					Capturing->OnMouseMove(_lgi_adjust_click_for_window(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseMove(_lgi_adjust_click_for_window(Ms, Over));
				}
				break;
			}
			/*
			case WM_NCHITTEST:
			{
				POINT Pt = { LOWORD(Msg->b), HIWORD(Msg->b) };
				ScreenToClient(hWnd, &Pt);
				int Hit = OnHitTest(Pt.x, Pt.y);
				if (Hit >= 0)
				{
					return Hit;
				}
				if (NOT (WndFlags & GWF_DIALOG))
				{
					return DefWindowProc(hWnd, Msg->Msg, Msg->a, Msg->b);
				}
				break;
			}
			*/
			case WM_LBUTTONDBLCLK:
			case WM_LBUTTONDOWN:
			case WM_LBUTTONUP:
			{
				GMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = (Msg->a & KEY_FLAGS) | MK_LEFT;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down(Msg->Msg != WM_LBUTTONUP);
				Ms.Double(Msg->Msg == WM_LBUTTONDBLCLK);

				if (Capturing)
				{
					Capturing->OnMouseClick(_lgi_adjust_click_for_window(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseClick(_lgi_adjust_click_for_window(Ms, Over));
				}
				else
				{
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_RBUTTONDBLCLK:
			case WM_RBUTTONDOWN:
			case WM_RBUTTONUP:
			{
				GMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = (Msg->a & KEY_FLAGS) | MK_RIGHT;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down(Msg->Msg != WM_RBUTTONUP);
				Ms.Double(Msg->Msg == WM_RBUTTONDBLCLK);

				if (Capturing)
				{
					Capturing->OnMouseClick(_lgi_adjust_click_for_window(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseClick(_lgi_adjust_click_for_window(Ms, Over));
				}
				else
				{
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_MBUTTONDBLCLK:
			case WM_MBUTTONDOWN:
			case WM_MBUTTONUP:
			{
				GMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = (Msg->a & KEY_FLAGS) | MK_MIDDLE;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
 				Ms.Down(Msg->Msg != WM_MBUTTONUP);
				Ms.Double(Msg->Msg == WM_MBUTTONDBLCLK);

				if (Capturing)
				{
					Capturing->OnMouseClick(_lgi_adjust_click_for_window(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseClick(_lgi_adjust_click_for_window(Ms, Over));
				}
				else
				{
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_SYSKEYUP:
			case WM_SYSKEYDOWN:
			case WM_KEYDOWN:
			case WM_KEYUP:
			{
				bool IsDialog = TestFlag(WndFlags, GWF_DIALOG);
				HWND hwnd = _View;

				if (NOT SysOnKey(this, Msg))
				{
					GKey Key;
					Key.c = Msg->a;		// virtual-key code 
					Key.Flags = 0;
					SetKeyFlag(Key.Flags, VK_MENU, MK_ALT);
					SetKeyFlag(Key.Flags, VK_SHIFT, MK_SHIFT);
					SetKeyFlag(Key.Flags, VK_CONTROL, MK_CTRL);
					Key.Data = Msg->b;
					Key.Down =	Msg->Msg == WM_KEYDOWN OR
								Msg->Msg == WM_SYSKEYDOWN;
					Key.IsChar = FALSE;

					OnKey(Key);
				}

				if (NOT IsDialog)
				{
					// required for Alt-Key function (eg Alt-F4 closes window)
					return DefWindowProc(hwnd, Msg->Msg, Msg->a, Msg->b);
				}
				break;
			}
			case WM_CHAR:
			{
				GKey Key;
				Key.c = Msg->a;		// virtual-key code 
				Key.Flags = 0;
				SetKeyFlag(Key.Flags, VK_MENU, MK_ALT);
				SetKeyFlag(Key.Flags, VK_SHIFT, MK_SHIFT);
				SetKeyFlag(Key.Flags, VK_CONTROL, MK_CTRL);
				Key.Data = Msg->b;
				Key.Down = TRUE;
				Key.IsChar = TRUE;

				OnKey(Key);
				break;
			}
			case WM_PAINT:
			{
				GScreenDC DeviceContext(_View);
				_Paint(&DeviceContext);
				break;
			}
			default:
			{
				if (NOT (WndFlags & GWF_DIALOG))
				{
					return DefWindowProc(_View, Msg->Msg, Msg->a, Msg->b);
				}
				break;
			}
		}
	}

	return 0;
}

bool GView::Name(char *n)
{
	bool Status = GObject::Name(n);
	if (_View) SetWindowText(_View, GObject::Name());
	return TRUE;
}

char *GView::Name()
{
	if (_View)
	{
		int Length = GetWindowTextLength(_View) + 1;
		char *Buf = NEW(char[Length]);
		if (Buf AND GetWindowText(_View, Buf, Length) > 0)
		{
			if (NOT GObject::Name() OR
				strcmp(GObject::Name(), Buf) != 0)
			{
				GObject::Name(Buf);
			}
		}
		DeleteArray(Buf);
	}
	return GObject::Name();
}

GView *GView::FindControl(OsView hCtrl)
{
	if (_View == hCtrl)
	{
		return this;
	}

	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(hCtrl);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

GView *GView::FindControl(int Id)
{
	if (GetId() == Id)
	{
		return this;
	}

	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(Id);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

void GView::MoveOnScreen()
{
	GRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);

	if (Pos.x2 >= Screen.X())
		Pos.Offset(Screen.X() - Pos.x2, 0);
	if (Pos.y2 >= Screen.Y())
		Pos.Offset(0, Screen.Y() - Pos.y2);
	if (Pos.x1 < 0)
		Pos.Offset(-Pos.x1, 0);
	if (Pos.y1 < 0)
		Pos.Offset(0, -Pos.y1);
}

void GView::MoveToCenter()
{
	GRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	Pos.Offset(-Pos.x1, -Pos.y1);
	Pos.Offset((Screen.X() - Pos.X()) / 2, (Screen.Y() - Pos.Y()) / 2);

	if (Handle())
	{
		SetPos(Pos, true);
	}
}

void GView::MoveToMouse()
{
	POINT Pt;
	if (GetCursorPos(&Pt))
	{
		Pos.Offset(-Pos.x1, -Pos.y1);
		Pos.Offset(Pt.x-(Pos.X()/2), Pt.y-(Pos.Y()/2));
		MoveOnScreen();
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////
GFlags::GFlags()
{
	WndFlags = GWF_VISIBLE | GWF_ENABLED;
}

bool GFlags::Enabled()
{
	return TestFlag(WndFlags, GWF_ENABLED);
}

void GFlags::Enabled(bool i)
{
	if (i) SetFlag(WndFlags, GWF_ENABLED);
	else ClearFlag(WndFlags, GWF_ENABLED);
}

bool GFlags::Visible()
{
	return TestFlag(WndFlags, GWF_VISIBLE);
}

void GFlags::Visible(bool i)
{
	if (i) SetFlag(WndFlags, GWF_VISIBLE);
	else ClearFlag(WndFlags, GWF_VISIBLE);
}

bool GFlags::Focus()
{
	return TestFlag(WndFlags, GWF_FOCUS);
}

void GFlags::Focus(bool i)
{
	if (i) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);
}

bool GFlags::DropTarget()
{
	return TestFlag(WndFlags, GWF_DROP_TARGET);
}

void GFlags::DropTarget(bool i)
{
	if (i) SetFlag(WndFlags, GWF_DROP_TARGET);
	else ClearFlag(WndFlags, GWF_DROP_TARGET);
}

#define GWF_ALL (GWF_SUNKEN | GWF_FLAT | GWF_RAISED)

bool GFlags::Sunken()
{
	return TestFlag(WndFlags, GWF_SUNKEN);
}

void GFlags::Sunken(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_SUNKEN);
}

bool GFlags::Flat()
{
	return TestFlag(WndFlags, GWF_FLAT);
}

void GFlags::Flat(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_FLAT);
}

bool GFlags::Raised()
{
	return TestFlag(WndFlags, GWF_RAISED);
}

void GFlags::Raised(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_RAISED);
}

