/*hdr
**      FILE:           GWidgets.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           8/9/1998
**      DESCRIPTION:    Dialog components
**
**      Copyright (C) 1998-2001 Matthew Allen
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include "Lgi.h"

///////////////////////////////////////////////////////////////////////////////////////////
short *DlgStrCopy(short *A, char *N)
{
	uchar *n = (uchar*) N;
	if (n)
	{
		while (*n)
		{
			*A++ = *n++;
		}
	}
	*A++ = 0;
	return A;
}

short *DlgPadToDWord(short *A, bool Seek = false)
{
	char *c = (char*) A;
	while (((int) c) & 3)
	{
		if (Seek)
		{
			c++;
		}
		else
		{
			*c++ = 0;
		}
	}
	return (short*) c;
}

///////////////////////////////////////////////////////////////////////////////////////////
GDialog::GDialog()
#ifdef LGI_RES
	: ResObject(Res_Dialog)
#endif
{
	Name("Dialog");
	IsModal = false;
	Mem = 0;
	WndFlags |= GWF_DIALOG;
	WndExStyle |= WS_EX_CONTROLPARENT;
	IsModal = true;
}

GDialog::~GDialog()
{
	DeleteObj(Mem);
}

bool GDialog::LoadFromResource(char *Resource)
{
	char n[256];
	bool Status = GLgiRes::LoadFromResource(Resource, Children, &Pos, n);
	if (Status)
	{
		Name(n);
	}
	return Status;
}

LRESULT CALLBACK DlgRedir(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (m == WM_INITDIALOG)
	{
		GDialog *NewWnd = (GDialog*) b;
		NewWnd->_View = hWnd;
		SetWindowLong(hWnd, GWL_USERDATA, (int) NewWnd);
		GApp::ActiveDialog = hWnd;
	}

	GDialog *Wnd = (GDialog*) GetWindowLong(hWnd, GWL_USERDATA);
	if (Wnd)
	{
		GMessage Msg(m, a, b);
		return Wnd->OnEvent(&Msg);
	}

	return 0;
}

int GDialog::DoModal()
{
	int Status = -1;

	IsModal = true;
	Mem = NEW(GMem(16<<10));
	if (Mem)
	{
		DLGTEMPLATE *Template = (DLGTEMPLATE*) Mem->Lock();
		if (Template)
		{
			GRegion r = Pos;
			r.x1 = (double)r.x1 / DIALOG_X;
			r.y1 = (double)r.y1 / DIALOG_Y;
			r.x2 = (double)r.x2 / DIALOG_X;
			r.y2 = (double)r.y2 / DIALOG_Y;

			Template->style =	// DS_ABSALIGN |
								DS_SETFONT |
								DS_NOFAILCREATE |
								DS_3DLOOK |
								WS_SYSMENU |
								WS_CAPTION |
								DS_SETFOREGROUND |
								DS_MODALFRAME;

			Template->dwExtendedStyle = WS_EX_DLGMODALFRAME;
			Template->cdit = 0;
			Template->x = r.x1;
			Template->y = r.y1;
			Template->cx = r.X();
			Template->cy = r.Y();

			IsModal = true;

			short *A = (short*) (Template+1);
			// menu
			*A++ = 0;
			// class
			*A++ = 0;
			// title
			A = DlgStrCopy(A, Name());
			// font
			*A++ = 8; // point size
			A = DlgStrCopy(A, "MS Sans Serif");
			A = DlgPadToDWord(A);

			GView *p = (Parent) ? ((Parent->Handle()) ? Parent : Parent->FindReal()) : 0;
			HWND hWindow = (p)?p->Handle():0;

			Status = DialogBoxIndirectParam(hAppInst(),
											Template,
											hWindow,
											(DLGPROC) DlgRedir, 
											(LPARAM) this);
		}

		DeleteObj(Mem);
	}

	return Status;
}

int GDialog::DoModeless()
{
	int Status = -1;
	Mem = NEW(GMem(16<<10));
	IsModal = false;
	if (Mem)
	{
		DLGTEMPLATE *Template = (DLGTEMPLATE*) Mem->Lock();
		if (Template)
		{
			GRegion r = Pos;
			r.x1 = (double)r.x1 / DIALOG_X;
			r.y1 = (double)r.y1 / DIALOG_Y;
			r.x2 = (double)r.x2 / DIALOG_X;
			r.y2 = (double)r.y2 / DIALOG_Y;

			Template->style =	WS_VISIBLE |
								DS_ABSALIGN |
								WS_SYSMENU |
								WS_CAPTION |
								DS_SETFONT |
								DS_NOFAILCREATE |
								DS_3DLOOK;

			Template->style |= DS_MODALFRAME;
			Template->dwExtendedStyle = WS_EX_DLGMODALFRAME;
			Template->cdit = 0;
			Template->x = r.x1;
			Template->y = r.y1;
			Template->cx = r.X();
			Template->cy = r.Y();

			IsModal = false;

			short *A = (short*) (Template+1);
			// menu
			*A++ = 0;
			// class
			*A++ = 0;
			// title
			A = DlgStrCopy(A, Name());
			// font
			*A++ = 8; // point size
			A = DlgStrCopy(A, "MS Sans Serif");
			A = DlgPadToDWord(A);

			GView *p = (Parent) ? ((Parent->Handle()) ? Parent : Parent->FindReal()) : 0;
			HWND hWindow = (p)?p->Handle():0;

			CreateDialogIndirectParam(	hAppInst(),
										Template,
										hWindow,
										(DLGPROC) DlgRedir, 
										(LPARAM) this);
		}
	}

	return Status;
}

extern GButton *FindDefault(GView *w);

int GDialog::OnEvent(GMessage *Msg)
{
	switch (Msg->Msg)
	{
		/*
		case DM_GETDEFID:
		{
			GView *Def = FindDefault(this);
			if (Def)
			{
				return MAKELONG(Def->GetId(), DC_HASDEFID);
			}
			break;
		}
		case WM_NCACTIVATE:
		{
			GApp::ActiveDialog = (Msg->a) ? hWnd : 0;
			break;
		}
		*/
		case WM_CREATE:
		{
			return 0;
		}
		case WM_INITDIALOG:
		{
			SetPos(Pos);	// resets the dialog to the correct
							// size when large fonts are used

			// GApp::ActiveDialog = hWnd;
			AttachChildren();
			OnCreate();

			return true;
			break;
		}
		case WM_CLOSE:
		{
			if (IsModal)
			{
				EndModal(0);
			}
			else
			{
				EndModeless(0);
			}
			return 0;
			break;
		}
	}

	return GView::OnEvent(Msg);
}

void GDialog::EndModal(int Code)
{
	EndDialog(Handle(), Code);
}

void GDialog::EndModeless(int Code)
{
	Quit(Code);
}

///////////////////////////////////////////////////////////////////////////////////////////
GControl::GControl(GWin32Class *Sub)
{
	SubClass = Sub;
	Pos.ZOff(10, 10);
}

GControl::~GControl()
{
}

int GControl::OnEvent(GMessage *Msg)
{
	int Status = 0;

	// Pre-OS event handler
	switch (Msg->Msg)
	{
		case WM_CREATE:
		{
			SetId(CtrlId);
			OnCreate();
			break;
		}
		case WM_SETTEXT:
		{
			GObject::Name((char*)Msg->b);
			break;
		}
		case WM_GETDLGCODE:
		{
			return GView::OnEvent(Msg);
		}
		case WM_NOTIFY:
		case WM_COMMAND:
		case WM_KEYDOWN:
		case WM_KEYUP:
		case WM_SYSKEYDOWN:
		case WM_SYSKEYUP:
		{
			GView::OnEvent(Msg);
			break;
		}
	}
	
	// OS event handler
	if (SubClass)
	{
		Status = SubClass->CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);
	}

	// Post-OS event handler
	switch (Msg->Msg)
	{
		case WM_CREATE:
		{
			SetFont(SysFont);
			break;
		}
	}

	return Status;
}

GdcPt2 GControl::SizeOfStr(char *Str)
{
	GdcPt2 Pt(0, 0);
	if (Str)
	{
		for (char *s=Str; s AND *s; )
		{
			char *e = strchr(s, '\n');
			if (NOT e) e = s + strlen(s);

			int x = 0, y = 0;
			SysFont->Size(&x, &y, s, (int)e-(int)s);

			Pt.y += y;
			Pt.x = max(Pt.x, x);

			s = (*e=='\n') ? e + 1 : 0;
		}
	}

	return Pt;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Button
GWin32Class GButton::WndClass;

GButton::GButton(int id, int x, int y, int cx, int cy, char *name) :
	GControl(&WndClass)
#ifdef LGI_RES
	, ResObject(Res_Button)
#endif
{
	Def = FALSE;
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);

	WndClass.SubClass(Class(), "BUTTON");
}

GButton::~GButton()
{
}

char *GButton::Class()
{
	return LGI_BUTTON;
}

uint32 GButton::Style()
{
	return ((Def) ? BS_DEFPUSHBUTTON : BS_PUSHBUTTON) |
			BS_TEXT |
			WS_CHILD |
			WS_TABSTOP |
			WS_VISIBLE |
			GView::Style();
}

bool GButton::Default()
{
	return Def;
}

void GButton::Default(bool b)
{
	Def = b;
}

int GButton::OnEvent(GMessage *Msg)
{
	if (MsgCode(Msg) == WM_GETDLGCODE)
	{
		return (Default() ?	DLGC_DEFPUSHBUTTON : DLGC_UNDEFPUSHBUTTON) |
				DLGC_WANTTAB;
	}

	return GControl::OnEvent(Msg);
}

///////////////////////////////////////////////////////////////////////////////////////////
// Edit
GWin32Class GEdit::WndClass;

GEdit::GEdit(int id, int x, int y, int cx, int cy, char *name) :
	GControl(&WndClass)
#ifdef LGI_RES
	, ResObject(Res_EditBox)
#endif
{
	GdcPt2 Size = SizeOfStr((name)?name:"A");
	if (cx < 0) cx = Size.x + 6;
	if (cy < 0) cy = Size.y + 4;

	MultiLineEdit = false;
	PasswordEdit = false;
	SetId(id);
	Sunken(true);
	if (name) Name(name);
	Pos.Set(x, y, x+max(cx, 10), y+max(cy, 10));

	WndClass.SubClass(Class(), "EDIT");
}

GEdit::~GEdit()
{
}

int GEdit::SysOnNotify(int Code)
{
	int Len = SendMessage(Handle(), WM_GETTEXTLENGTH, 0, 0);
	char *Str = NEW(char[Len+1]);
	if (Str)
	{
		SendMessage(Handle(), WM_GETTEXT, (WPARAM) (Len+1), (LPARAM) Str);
		GObject::Name(Str);
		DeleteArray(Str);
	}
	return 0;
}

void GEdit::MultiLine(bool m)
{
	MultiLineEdit = m;
}

void GEdit::Password(bool m)
{
	PasswordEdit = m;
}

void GEdit::Value(int i)
{
	char Str[32];
	sprintf(Str, "%i", i);
	Name(Str);
}

int GEdit::Value()
{
	char *n = Name();
	return (n) ? atoi(n) : 0;
}

int GEdit::OnEvent(GMessage *Msg)
{
	switch (Msg->Msg)
	{
		case WM_GETDLGCODE:
		{
			return WndClass.CallParent(Handle(), Msg->Msg, Msg->a, Msg->b) |
					DLGC_WANTTAB;
		}
		case WM_KEYDOWN:
		{
			if (Msg->a == VK_TAB AND MultiLine())
			{
				// this causes the multiline Edit control to
				// insert tabs into the text and not move
				// to the next field.
				return 0;
			}

			GView::OnEvent(Msg);
			break;
		}
		case WM_CHAR:
		{
			if (Msg->a == VK_TAB AND NOT MultiLine())
			{
				// single line edit controls make "click"
				// when they get a TAB char... this
				// avoids that.
				return 0;
			}
			break;
		}
	}

	return GControl::OnEvent(Msg);
}

void GEdit::OnKey(GKey &k)
{
	if (k.c == VK_RETURN)
	{
		GView *n = (GetNotify()) ? GetNotify() : GetParent();
		if (n) n->OnNotify(this, k.c);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////
// Check box
GWin32Class GCheckBox::WndClass;

GCheckBox::GCheckBox(int id, int x, int y, int cx, int cy, char *name, int InitState) :
	GControl(&WndClass)
#ifdef LGI_RES
	, ResObject(Res_CheckBox)
#endif
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	Val = InitState;
	Three = FALSE;
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	WndClass.SubClass(Class(), "BUTTON");
}

GCheckBox::~GCheckBox()
{
}

int GCheckBox::Value()
{
	if (Handle())
	{
		Val = SendMessage(Handle(), BM_GETCHECK, 0, 0);
	}
	return Val;
}

void GCheckBox::Value(int b)
{
	Val = b;
	if (Handle())
	{
		SendMessage(Handle(), BM_SETCHECK, Val, 0);
	}
}

int GCheckBox::OnEvent(GMessage *Msg)
{
	if (Msg->Msg == WM_CREATE)
	{
		Value(Val);
	}

	return GControl::OnEvent(Msg);
}

//////////////////////////////////////////////////////////////////////////////////////////////////
// Text
GWin32Class GText::WndClass;

GText::GText(int id, int x, int y, int cx, int cy, char *name) :
	GControl(&WndClass)
#ifdef LGI_RES
	, ResObject(Res_StaticText)
#endif
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x;
	if (cy < 0) cy = Size.y;

	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);

	WndClass.SubClass(Class(), "STATIC");
}

GText::~GText()
{
}

int GText::OnEvent(GMessage *Msg)
{
	return GControl::OnEvent(Msg);
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio group
GWin32Class GRadioGroup::WndClass;
int GRadioGroup::NextId = 20000;

GRadioGroup::GRadioGroup(int id, int x, int y, int cx, int cy, char *name, int Init) :
	GControl(&WndClass)
#ifdef LGI_RES
	, ResObject(Res_Group)
#endif
{
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	InitialValue = Init;

	GView::WndClass = LGI_RADIOGROUP;
	WndStyle |= BS_GROUPBOX;
	WndStyle &= ~WS_TABSTOP;
	// WndClass.Class.hbrBackground = (HBRUSH)(COLOR_BTNFACE+1);
	WndClass.SubClass(Class(), "BUTTON");
}

GRadioGroup::~GRadioGroup()
{
}

void GRadioGroup::OnCreate()
{
	AttachChildren();
	Value(InitialValue);
}

int GRadioGroup::Value()
{
	if (Handle())
	{
		int i=0;
		for (	GView *w = Children.First();
				w;
				w = Children.Next())
		{
			GRadioButton *But = dynamic_cast<GRadioButton*>(w);
			if (But)
			{
				if (But->Value())
				{
					return i;
				}

				i++;
			}
		}
	}
	else
	{
		return InitialValue;
	}

	return -1;
}

void GRadioGroup::Value(int Which)
{
	if (Handle())
	{
		int i=0;
		for (	GView *w = Children.First();
				w;
				w = Children.Next())
		{
			GRadioButton *But = dynamic_cast<GRadioButton*>(w);
			if (But)
			{
				But->Value(Which == i);
				i++;
			}
		}
	}
	else
	{
		InitialValue = Which;
	}
}

int GRadioGroup::OnNotify(GView *Ctrl, int Flags)
{
	GView *n = GetNotify() ? GetNotify() : GetParent();
	if (n)
	{
		return n->OnNotify(Ctrl, Flags);
	}
	return 0;
}

int GRadioGroup::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case WM_ERASEBKGND:
		{
			// bloody Win32 control doesn't paint it's background... so
			// I fix it here.
			GScreenDC Dc(GetDC(Handle()), Handle());
			Dc.Colour(LC_MED, 24);
			Dc.Rectangle();
			InvalidateRect(Handle(), NULL, false);
			return true;
		}
		case WM_COMMAND:
		{
			GView *Ctrl = FindControl((HWND) Msg->b);
			GView *n = GetNotify() ? GetNotify() : GetParent();
			if (Ctrl AND n)
			{
				short Code = HIWORD(Msg->a);
				switch (Code)
				{
					case BN_CLICKED: // BUTTON
					{
						n->OnNotify(this, 0);
						break;
					}
				}
			}
			break;
		}
	}

	return GControl::OnEvent(Msg);
}

GRadioButton *GRadioGroup::Append(int x, int y, char *name)
{
	GRadioButton *But = NEW(GRadioButton(NextId++, x, y, -1, -1, name, Children.GetItems() == 0));
	if (But)
	{
		Children.Insert(But);
	}
	return But;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio button
GWin32Class GRadioButton::WndClass;

GRadioButton::GRadioButton(int id, int x, int y, int cx, int cy, char *name, bool first) :
	GControl(&WndClass)
#ifdef LGI_RES
	, ResObject(Res_RadioBox)
#endif
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	First = first;
	InitVal = false;
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);

	WndClass.SubClass(Class(), "BUTTON");
}

GRadioButton::~GRadioButton()
{
}

int GRadioButton::Value()
{
	if (Handle())
	{
		return SendMessage(Handle(), BM_GETCHECK, 0, 0);
	}
	return 0;
}

void GRadioButton::Value(int i)
{
	InitVal = i;
	if (Handle())
	{
		SendMessage(Handle(), BM_SETCHECK, i, 0);
	}
}

int GRadioButton::OnEvent(GMessage *Msg)
{
	int Status = GControl::OnEvent(Msg);

	switch (MsgCode(Msg))
	{
		case WM_CREATE:
		{
			Value(InitVal);

			if (Parent AND
				this == Parent->Children.First())
			{
				int s = GetWindowLong(Handle(), GWL_STYLE) | WS_GROUP;
				SetWindowLong(Handle(), GWL_STYLE, s);
			}
			break;
		}
		case WM_GETDLGCODE:
		{
			return SubClass->CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);
			break;
		}
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////////////
// Slider control
GWin32Class GSlider::WndClass;

GSlider::GSlider(int id, int x, int y, int cx, int cy, char *name, bool vert) :
	GControl(&WndClass)
#ifdef LGI_RES
	, ResObject(Res_Slider)
#endif
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	Name(name);
	Vertical = vert;
	Min = Max = 0;
	Val = 0;

	WndClass.SubClass(Class(), TRACKBAR_CLASS);
}

GSlider::~GSlider()
{
}

int GSlider::SysOnNotify(int Code)
{
	return 0;
}

void GSlider::Value(int i)
{
	Val = i;
	if (Handle())
	{
		SendMessage(Handle(), TBM_SETPOS, TRUE, Val);
	}
}

int GSlider::Value()
{
	if (Handle())
	{
		Val = SendMessage(Handle(), TBM_GETPOS, 0, 0);
	}
	return Val;
}

void GSlider::GetLimits(int &min, int &max)
{
	if (Handle())
	{
		min = Min = SendMessage(Handle(), TBM_GETRANGEMIN, 0, 0);
		max = Max = SendMessage(Handle(), TBM_GETRANGEMAX, 0, 0);
	}
	else
	{
		min = Min;
		max = Max;
	}
}

void GSlider::SetLimits(int min, int max)
{
	Min = min;
	Max = max;
	if (Handle())
	{
		SendMessage(Handle(), TBM_SETRANGE, TRUE, (LPARAM) MAKELONG(min, max)); 
	}
}

int GSlider::OnEvent(GMessage *Msg)
{
	int Status = GControl::OnEvent(Msg);

	switch (MsgCode(Msg))
	{
		case WM_HSCROLL:
		case WM_VSCROLL:
		{
			GView *n = (GetNotify()) ? GetNotify() : GetParent();
			if (n)
			{
				n->OnNotify(this, 0);
			}
			break;
		}
		case WM_CREATE:
		{
			SetLimits(Min, Max);
			Value(Val);
			break;
		}
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////////////
GWin32Class GBitmap::WndClass;

#ifdef _MT
class GBitmapThread : public GThread
{
	GBitmap *Bmp;
	char *File;
	GThread **Owner;

public:
	GBitmapThread(GBitmap *bmp, char *file, GThread **owner)
	{
		Bmp = bmp;
		File = NewStr(file);
		Owner = owner;
		if (Owner)
		{
			*Owner = this;
		}
		Run(true);
	}

	~GBitmapThread()
	{
		if (Owner)
		{
			*Owner = 0;
		}
		DeleteArray(File);
	}

	int Main()
	{
		if (Bmp)
		{
			GSurface *pDC = LoadDC(File);
			if (pDC)
			{
				Bmp->SetDC(pDC);
				GRegion r = Bmp->GetPos();
				r.Dimension(pDC->X()+4, pDC->Y()+4);
				Bmp->SetPos(r);
				Bmp->Invalidate();
			}
		}

		return 0;
	}
};
#endif

GBitmap::GBitmap(int id, int x, int y, char *FileName, bool Async)
#ifdef LGI_RES
	: ResObject(Res_Bitmap)
#endif
{
	pDC = 0;
	pThread = 0;

	SetId(id);
	Pos.ZOff(4, 4);
	Pos.Offset(x, y);

	#ifdef _MT
	if (NOT Async)
	{
	#endif
		pDC = LoadDC(FileName);
		if (pDC)
		{
			Pos.Dimension(pDC->X()+4, pDC->Y()+4);
		}
	#ifdef _MT
	}
	else
	{
		NEW(GBitmapThread(this, FileName, &pThread));
	}
	#endif

	WndClass.Create(Class());
}

GBitmap::~GBitmap()
{
	#ifdef _MT
	if (pThread)
	{
		pThread->Terminate();
		DeleteObj(pThread);
	}
	#endif
	DeleteObj(pDC);
}

void GBitmap::SetDC(GSurface *pNewDC)
{
	DeleteObj(pDC);
	pDC = pNewDC;
}

GSurface *GBitmap::GetDC()
{
	return pDC;
}

int GBitmap::OnEvent(GMessage *Msg)
{
	return GView::OnEvent(Msg);
}

void GBitmap::OnPaint(GSurface *pScreen)
{
	GRegion a(0, 0, Pos.X()-1, Pos.Y()-1);
	LgiWideBorder(pScreen, a, SUNKEN);
	if (pDC)
	{
		pScreen->Blt(2, 2, pDC);
	}
	else
	{
		pScreen->Colour(CBit(pScreen->GetBits(), GetSysColor(COLOR_WINDOW)));
		pScreen->Rectangle(&a);
	}
}

void GBitmap::OnMouseClick(GMouse &m)
{
	if (NOT m.Down() AND Parent)
	{
		((GDialog*) Parent)->OnNotify(this, 0);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////
GItemContainer::GItemContainer()
{
	Flags = 0;
	ImageList = 0;
}

GItemContainer::~GItemContainer()
{
	if (OwnList())
	{
		DeleteObj(ImageList);
	}
	else
	{
		ImageList = 0;
	}
}

bool GItemContainer::SetImageList(GImageList *list, bool Own)
{
	ImageList = list;
	OwnList(Own);
	AskImage(true);
	return ImageList != NULL;
}

bool GItemContainer::LoadImageList(char *File, int x, int y)
{
	GSurface *pDC = LoadDC(File);
	if (pDC)
	{
		ImageList = NEW(GImageList(x, y, pDC));
		if (ImageList)
		{
			#ifdef WIN32
			ImageList->Create(pDC->X(), pDC->Y(), pDC->GetBits());
			#endif
		}
	}
	return pDC != 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
GWin32Class GProgress::WndClass;

GProgress::GProgress(int id, int x, int y, int cx, int cy, char *name) :
	GControl(&WndClass)
#ifdef LGI_RES
	, ResObject(Res_Progress)
#endif
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	if (name) GControl::Name(name);

	WndStyle |= PBS_SMOOTH;
	WndClass.SubClass(Class(), PROGRESS_CLASS);
}

GProgress::~GProgress()
{
}

void GProgress::SetLimits(int l, int h)
{
	Low = l;
	High = h;
	if (Handle())
	{
		SendMessage(Handle(), PBM_SETRANGE32, Low, High);
	}
}

void GProgress::Value(int v)
{
	Val = v;
	if (Handle())
	{
		SendMessage(Handle(), PBM_SETPOS, (WPARAM) Val, 0);
	}
}

int GProgress::OnEvent(GMessage *Msg)
{
	return GControl::OnEvent(Msg);
}

