#include "Lgi.h"

///////////////////////////////////////////////////////////////////////////////////////////////
GWindow::GWindow()
{
	Menu = 0;
	WndStyle |= WS_TILEDWINDOW | WS_CLIPCHILDREN; // | WS_CLIPSIBLINGS;
	WndStyle &= ~WS_CHILD;
	WndExStyle |= WS_EX_CONTROLPARENT | WS_EX_ACCEPTFILES;
	LastFocus = 0;
	_QuitOnClose = true;
}

GWindow::~GWindow()
{
	if (Menu)
	{
		Menu->Detach();
		DeleteObj(Menu);
	}
}

void GWindow::OnPaint(GSurface *pDC)
{
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle();
}

bool GWindow::Visible()
{
	return GView::Visible();
}

void GWindow::Visible(bool v)
{
	if (v) Pour();
	GView::Visible(v);
}

void GWindow::Pour()
{
	GComplexRegion Client(GetClient());
	GComplexRegion Update;

	for (GView *v = Children.First(); v; v = Children.Next())
	{
		GRegion OldPos = v->Pos;
		Update.Union(&OldPos);

		if (v->Pour(Client))
		{
			if (NOT v->Visible())
			{
				v->Visible(true);
			}

			if (OldPos != v->Pos)
			{
				// position has changed update...
				v->Invalidate();
			}

			Client.Subtract(&v->Pos);
			Update.Subtract(&v->Pos);
		}
		else
		{
			// make the view not visible
			// v->Visible(FALSE);
		}
	}

	for (int i=0; i<Update.GetSize(); i++)
	{
		Invalidate(Update[i]);
	}
}

int GWindow::OnEvent(GMessage *Msg)
{
	int Status = 0;

	switch (MsgCode(Msg))
	{
		case WM_TRANSLATE_ACCELERATOR:
		{
			if (Menu)
			{
				Menu->OnEvent(Msg);
			}
			break;
		}
		case WM_SYSCOLORCHANGE:
		{
			LgiInitColours();
			break;
		}
		case WM_WINDOWPOSCHANGING:
		{
			WINDOWPOS *Info = (LPWINDOWPOS) Msg->b;
			if (Info)
			{
				if (_MinimumSize.x AND
					_MinimumSize.x > Info->cx)
				{
					Info->cx = _MinimumSize.x;
				}

				if (_MinimumSize.y AND
					_MinimumSize.y > Info->cy)
				{
					Info->cy = _MinimumSize.y;
				}
			}
			break;
		}
		case WM_ACTIVATE:
		{
			// This is a hack to make Windows set the focus of a child
			// control when you Alt-Tab in and out of a top level window
			if (LOWORD(Msg->a) != WA_INACTIVE)
			{
				// gaining focus
				if (LastFocus)
				{
					SetFocus(LastFocus);
					LastFocus = 0;
				}
			}
			else
			{
				// losing focus
				LastFocus = 0;
				HWND f = GetFocus();
				for (HWND p = ::GetParent(f); p; p = ::GetParent(p))
				{
					if (p == Handle())
					{
						LastFocus = f;
					}
				}
			}
			break;
		}
		case WM_CREATE:
		{
			Pour();
			OnCreate();
			break;
		}
		case WM_QUERYENDSESSION:
	 	case WM_CLOSE:
		{
			bool QuitApp = false;
			if (OnRequestClose())
			{
				QuitApp = true;

				if (GetQuitOnClose())
				{
					LgiCloseApp();
				}
				else
				{
					Quit();
				}
			}

			if (MsgCode(Msg) == WM_CLOSE)
			{
				return 0;
			}
			else
			{
				return QuitApp;
			}
			break;
		}
		case WM_DROPFILES:
		{
			HDROP hDrop = (HDROP) Msg->a;
			if (hDrop)
			{
				List<char> FileNames;
				int Count = DragQueryFile(hDrop, -1, NULL, 0);;
				char FileName[256];

				for (int i=0; i<Count; i++)
				{
					if (DragQueryFile(hDrop, i, FileName, sizeof(FileName)-1) > 0)
					{
						FileNames.Insert(NewStr(FileName));
					}
				}

				OnReceiveFiles(FileNames);
				DeleteObjList(&FileNames, char);
			}
			break;
		}
		case WM_COMMAND:
		{
			HWND OurWnd = Handle(); // copy onto the stack, because
									// we might lose the 'this' object in the
									// OnCommand handler which would delete
									// the memory containing the handle.

			Status = OnCommand(LOWORD(Msg->a), HIWORD(Msg->a), (HWND) Msg->b);
			if (NOT IsWindow(OurWnd))
			{
				// The window was deleted so break out now
				break;
			}
			// otherwise fall thru to the GView handler
		}
		default:
		{
			Status = GView::OnEvent(Msg);
			break;
		}
	}

	return Status;
}

void GWindow::OnPosChange()
{
	Pour();
}

