/*hdr
**      FILE:           Ftp.h
**      AUTHOR:         Matthew Allen
**      DATE:           7/9/99
**      DESCRIPTION:    i.Ftp header file
**
**      Copyright (C) 1999 Matthew Allen
**              fret@ozemail.com.au
*/

// Includes
#include <stdio.h>
#include <stdlib.h>
#include "Lgi.h"
#include "DragAndDrop.h"
#include "GXml.h"
#include "INet.h"

// Resource ID's
#include "resource.h"
#include "resdefs.h"

// Gui controls
#include "GTextView.h"
#include "GPanel.h"
#include "GProgressStatusPane.h"

// Ftp protocol
#include "IFtp.h"

////////////////////////////////////////////////////////////////////////////////////////////
#define	VER							1.39

#define FTP_COL_WRITE				Rgb24(0,0,255)
#define FTP_COL_READ				Rgb24(0,128,0)
#define FTP_COL_ERROR				Rgb24(255,0,0)
#define FTP_COL_INFO				Rgb24(0,0,0)

////////////////////////////////////////////////////////////////////////////////////////////
// Defines

// Account props
#define OPT_AccountName				"AccName"	// (char*)
#define OPT_RemoteHost				"Host"		// (char*)
#define OPT_Port					"Port"		// (int)
#define OPT_UserName				"User"		// (char*)
#define OPT_Password				"Password"	// (char*)
#define OPT_Anonymous				"Anon"		// (bool)(int)
#define OPT_LocalDir				"Local"		// (char*)
#define OPT_RemoteDir				"Remote"	// (char*)
#define OPT_UseActive				"UseActive"	// (bool)

#define OPT_Socks5Server			"Socks5Server"
#define OPT_Socks5UserName			"Socks5UserName"
#define OPT_Socks5Password			"Socks5Password"

// window messages
#define	IDM_OPEN					111
#define	IDM_LOAD					112
#define	IDM_SAVE					113
#define	IDM_SAVEAS					114
#define IDM_STOP					115
#define	IDM_CLOSE					118
#define IDM_EXIT					119

#define IDM_CONNECT					120
#define IDM_DISCONNECT				121
#define IDM_UPLOAD					122
#define IDM_DOWNLOAD				123
#define IDM_OPTIONS					124
#define	IDM_LIST					125
#define IDM_UP						126
#define IDM_CREATE_DIR				127
#define IDM_DELETE_DIR				128
#define IDM_DELETE					129
#define IDM_EXECUTE					130
#define IDM_RENAME					131
#define IDM_ABORT					132
#define IDM_SETPERMS				132
#define IDM_CD						133
#define IDM_TRANSFER				134
#define IDM_TRANSFER_MASK			135
#define IDM_VIEW_TEXT				136

#define	IDM_UNDO					201
#define	IDM_REDO					202
#define	IDM_CUT						203
#define	IDM_COPY					204
#define	IDM_PASTE					205
#define	IDM_FIND					206

#define IDM_TO						311
#define IDM_CC						312
#define IDM_EMPTY					313
#define IDM_EDIT					314

#define IDM_PROPERTIES				402

#define IDM_ACCOUNT_BASE			500

#define IDM_ABOUT					998
#define IDM_HELP					999

// File utility macros
#define SizeStrField(Var)			(sizeof(short)+SizeofStr(Var))
#define SizeIntField(Var)			(sizeof(short)+sizeof(ulong)+sizeof(Var))
#define SizeObjField(Var)			(sizeof(short)+sizeof(ulong)+(Var).Sizeof())

#define WriteStrField(Id, Var)		f << ((short)Id); WriteStr(f, Var)
#define WriteIntField(Id, Var)		f << ((short)Id); f << ((ulong)sizeof(Var)); f << Var
#define WriteObjField(Id, Var)		f << ((short)Id); f << ((ulong)(Var).Sizeof()); (Var).Serialize(f, Write)

#define ReadStrField(Id, Var)		case Id: Var = ReadStr(f); break
#define ReadIntField(Id, Var)		case Id: f >> FieldSize; f >> Var; break
#define ReadObjField(Id, Var)		case Id: f >> FieldSize; (Var).Serialize(f, Write); break

// Misc
#define	MainWnd						((AppWnd*)GApp::ObjInstance()->AppWnd)

extern char AppName[];
extern char HelpFile[];
extern char OptionsFile[];

////////////////////////////////////////////////////////////////////////////////////////////
// Functions

// Misc
extern int SizeOfFile(char *FileName);
extern char *OsName();
extern void GetFullAppName(char *s);
extern void GetFolderId(char *s);
extern void Error(char *str, ...);

#include "FileTransferProgress.h"

////////////////////////////////////////////////////////////////////////////////////////////
// Classes
class AppWnd;

/////////////////////////////////////////////////////////////////////////////
#define IDC_OVERWRITE			100
#define IDC_DIR					101
#define IDC_MOUNT				102

class ResumeDlg : public GDialog
{
public:
	ResumeDlg(GView *parent, IFtpEntry *e, bool Resumable);
	int OnNotify(GView *Ctrl, int Flags);
};

/////////////////////////////////////////////////////////////////////////////
class AboutDlg : public GDialog {

	char Text[512];

public:
	AboutDlg(GView *parent);
	int OnNotify(GView *Ctrl, int Flags);
};

class App : public GApp {
public:
	#ifdef WIN32
	App(HINSTANCE inst, char *cmd, int show) : GApp(inst, cmd, show) {}
	#else if BEOS
	App(char *MimeSig, char *CmdLine) : GApp(MimeSig, CmdLine) {}
	#endif
};

#include "RLog.h"
class GLogPanel : public GPanel, public RLog
{
	RLogView *View;

public:
	GLogPanel();
	~GLogPanel();

	bool Pour(GComplexRegion &r);
};

#define TASK_ListElements			1000
#define TASK_Cd						1001
#define TASK_CreateDir				1002
#define TASK_Delete					1003
#define TASK_SetPerms				1004
#define TASK_ExecuteFile			1005
#define TASK_ViewTextFile			1006
#define TASK_ProcessClick			1007
#define TASK_Login					1008
#define TASK_Transfer				1009
#define TASK_Rename					1010
#define TASK_Noop					1011

class FtpCmd
{
public:
	int Cmd;
	char Arg[256]; // for buffering if you need it
	char Arg2[256];
	int Int;
	class GFileSystemView *View;

	char *Str;
	GMouse Click;
	IFtpEntry *Entry;
	GSocket *Connection;

	FtpCmd()
	{
		Cmd = 0;
		Str = 0;
		View = 0;
		Arg[0] = 0;
		Arg2[0] = 0;
		Entry = 0;
		Connection = 0;
		Int = 0;
	}

	FtpCmd(char *s)
	{
		Cmd = 0;
		View = 0;
		Entry = 0;
		Connection = 0;
		Int = 0;
		
		if (s)
		{
			Str = Arg;
			strcpy(Arg, s);
		}
		else
		{
			Str = 0;
			Arg[0] = 0;
		}
	}

	~FtpCmd()
	{
		DeleteObj(Entry);
	}
};

class FtpWorkerThread : public GThread, public GSemaphore
{
	List<FtpCmd> Cmds;
	bool Loop;

public:
	FtpWorkerThread();

	// This loops around looking for commands to send
	int Main();
	
	// This que's a command for processing by the worker thread
	// The command will get poped off the que and sent to the
	// calling object.
	void PostCmd(FtpCmd *Cmd);

	// Props
	int GetCmdsLeft();
	void SetLoop(bool i);
};

class FtpList : public GList
{
public:
	class GFileSystemView *View;

	FtpList(int id, int x, int y, int cx, int cy) :
		GList(id, x, y, cx, cy)
	{
		View = 0;
	}

	void OnColumnClick(int Col, GMouse &m);
};

class GFileSystemView : public GLayout, public List<IFtpEntry>
{
	friend int FtpItemCompare(GListItem *a, GListItem *b, int Data);

protected:
	GImageList *ImageList;
	GToolBar *Commands;
	AppWnd *Wnd;
	IFtp *Ftp;
	bool Listing;

	int SortCol;
	bool SortAscend;

	GCommand CmdCreateDir;
	GCommand CmdDelete;
	GCommand CmdRenameFile;
	GCommand CmdSetPerms;
	GCommand CmdExecute;
	GCommand CmdViewText;
	GCommand CmdTransfer;
	GCommand CmdRefresh;
	GCommand CmdCdup;

public:
	GEdit *Edit;
	FtpList *Lst;
	GButton *Dir;
	GCombo *Mount;

	GFileSystemView *Opposite;
	FtpWorkerThread *Worker;

	GFileSystemView(AppWnd *Wnd, GImageList *ImageList, IFtp *ftp);
	~GFileSystemView();

	// Data
	void GetDir(char *Path, char *Append = 0);
	int GetSelection(List<IFtpEntry> &Sel);
	void PostCmd(FtpCmd *Cmd, bool ForceRemote = false);
	virtual bool IsCmdLocal(int Cmd) { return false; }
	virtual void EnableCommands(bool Enabled) {}

	// Menu
	virtual void GetMenuCommands(IFtpEntry *e, GSubMenu *RClick);
	virtual void OnMenuCommand(IFtpEntry *e, int Cmd);

	// Window
	bool Attach(GView *p);
	bool Pour(GComplexRegion &r);
	void OnPaint(GSurface *pDC);
	int OnNotify(GView *Ctrl, int Flags);
	void OnColumnClick(int Col, GMouse &m);
	virtual void OnClick(IFtpEntry *e, GMouse *m);

	// Insert a entry
	bool Insert(IFtpEntry *p, int Index = -1) { return false; }
	void Empty() { if (Lst) Lst->Empty(); }

	// Methods
	//
	//		These functions are called from the worker thread
	//		not the main UI thread. This is acheived by posting
	//		a command to the worker thread which then calls us
	//		back when it's ready to process it

	// Process a general command
	void DoCmd(FtpCmd *Cmd);

	// Command handlers
	virtual void ListElements() {}
	virtual bool Cd(char *Dir = 0) { return false; } // 0 = up
	virtual bool CreateDir(char *Dir) { return false; }
	virtual bool Delete(IFtpEntry *e) { return false; }
	virtual void SetPerms(IFtpEntry *e, int Perms) {}
	virtual void ExecuteFile(char *File) {}
	virtual void ViewTextFile(char *File) {}
	virtual void ProcessClick(IFtpEntry *e, GMouse *m) {}
	virtual bool TransferFile(FtpCmd *c)
	{
		// Arguments:
		//		Cmd = TASK_Transfer
		//		Arg = <full path name>
		//		Entry = <entry to transfer>
		//		Str = <directory transfer mask>
		//
		// Either Arg or Entry must be valid
		return false;
	}
	virtual bool RenameFile(char *From, char *To) { return false; }

	// Worker
	virtual IFtpEntry *HasDir(char *Name);
	virtual bool TransferDir(IFtpEntry *e, char *Mask = 0);
	virtual bool DeleteDir(IFtpEntry *e);
	virtual char DirChar() { return DIR_CHAR; }
};

class GLocalFS : public GFileSystemView
{
public:
	GLocalFS(AppWnd *Wnd, GImageList *ImageList, IFtp *ftp);

	bool IsCmdLocal(int Cmd);
	void OnItemSelect(GListItem *Item);
	void EnableCommands(bool Enabled);

	bool Insert(IFtpEntry *p, int Index = -1);
	void ListElements();
	bool Cd(char *Dir = 0);
	bool CreateDir(char *Dir);
	bool Delete(IFtpEntry *e);
	void SetPerms(IFtpEntry *e, int Perms);
	void ExecuteFile(char *File);
	void ViewTextFile(char *File);
	void ProcessClick(IFtpEntry *e, GMouse *m);
	bool TransferFile(FtpCmd *c);
	bool RenameFile(char *From, char *To);
};

class GRemoteFS : public GFileSystemView
{
public:
	GRemoteFS(AppWnd *Wnd, GImageList *ImageList, IFtp *ftp);

	bool IsCmdLocal(int Cmd);
	void OnItemSelect(GListItem *Item);
	void EnableCommands(bool Enabled);

	bool Insert(IFtpEntry *p, int Index = -1);
	void ListElements();
	bool Cd(char *Dir = 0);
	bool CreateDir(char *Dir);
	bool Delete(IFtpEntry *e);
	void SetPerms(IFtpEntry *e, int Perms);
	void ProcessClick(IFtpEntry *e, GMouse *m);
	bool TransferFile(FtpCmd *c);
	bool RenameFile(char *From, char *To);

	char DirChar() { return '/'; }
};

class ILog {

	// Logging
	RLog *Logger;
	Progress *Info;
	bool Logging;
	
public:
	ILog(RLog *log);
	~ILog();

	// Logging
	void WriteLog(char *s, int Len, COLOUR c);

	void LogRead(char *Data, int Len);
	void LogWrite(char *Data, int Len);
	void LogError(int ErrorCode, char *ErrorDescription);
	void LogInformation(char *Str);
	void LogSetParameter(int Which, int What);
};

class ILogConnection : public ILog, public GSocket
{
	AppWnd *Wnd;

public:
	ILogConnection(RLog *log, GNetwork *parent, AppWnd *wnd);

	void OnDisconnect();
	void OnRead(char *Data, int Len);
	void OnWrite(char *Data, int Len);
	void OnError(int ErrorCode, char *ErrorDescription);
	void OnInformation(char *Str);
	void SetParameter(int Which, int What);
};

class ISocks5LogConnection : public ILog, public GSocks5Socket
{
	AppWnd *Wnd;

public:
	ISocks5LogConnection(	char *proxy, int port, char *username, char *password,
							RLog *log, GNetwork *parent, AppWnd *wnd);

	void OnDisconnect();
	void OnRead(char *Data, int Len);
	void OnWrite(char *Data, int Len);
	void OnError(int ErrorCode, char *ErrorDescription);
	void OnInformation(char *Str);
	void SetParameter(int Which, int What);
};

class DirListItem : public GListItem
{
	GFileSystemView *View;
	char SizeStr[32];

public:
	IFtpEntry *Entry;

	DirListItem(IFtpEntry *entry, GFileSystemView *View);
	~DirListItem();

	IFtpEntry *GetEntry() { return Entry; }
	char *GetText(int i);
	int GetImage(int Flags = 0);
	void OnMouseClick(GMouse &m);
};

////////////////////////////////////////////////////////////////////////
class AppWnd :
	public GWindow,
	public IFtpDataPipe,
	public FileTransferWindow,
	public GDragDropTarget
{
	friend class HttpThread;
	friend class GRemoteFS;
	friend class GFileSystemView;
	friend class HttpThread;

protected:
	#ifdef WIN32
	static GWin32Class Class;
	#endif

	char			*OptionsFile;
	ObjProperties	Options;
	List<ObjProperties> Accounts;
	bool			Connecting;

	// UI
	GStatusBar		*Status;
	GStatusPane		*StatusInfo[_STATUS_MAX];
	GSubMenu		*File;
	GSubMenu		*Edit;
	GSubMenu		*AccountLst;
	GSubMenu		*Help;
	GToolBar		*Commands;

	// Commands
	GCommand		CmdConnect;
	GCommand		CmdDisconnect;
	GCommand		CmdAbort;
	GCommand		CmdHttp;

	// Storage
	int MaxDataRate; // in bytes / second

	// Main view
	GImageList		*ImageList;
	GSplitter		*Splitter;
	GLogPanel		*LogPanel;

	GFileSystemView	*Local;
	GFileSystemView *Remote;
	FileTransferProgress *Progress;

	FtpWorkerThread	*Worker;
	GThread *Http;

	// Vars
	GNetwork		Sockets;
	GSocket		*Connection;
	IFtp			*Ftp;
	int				LastCmd;

	// Methods
	void			SetupUi();
	void			ResetThread();
	void			UpdateMenus();

public:
	AppWnd();
	~AppWnd();

	// ---------------------------------------------------------------------
	// Methods
	ObjProperties *GetOptions() { return &Options; }
	void EnableCommands(bool Enabled);
	void Serialize(bool Write);
	void ReadProperties(XmlTag *t, ObjProperties *a);
	void WriteProperties(GFile &F, ObjProperties *Props, char *KeyName);
	void OnDisconnect();
	void SetTransfer(bool On);
	void ConnectTo(ObjProperties *Account);
	void OnCmd(FtpCmd *Cmd);
	void DoCmdLine();

	// Data throttle
	void SetDataRate(int Percent = 100);
	int WritePipe(GSocket *c, char *Data, int Len, int Flags);
	int ReadPipe(GSocket *c, char *Data, int Len, int Flags);

	// ---------------------------------------------------------------------
	// Window
	int OnNotify(GView *Ctrl, int Flags);
	int OnEvent(GMessage *Msg);
	int OnCommand(int Cmd, int Event, OsView Handle);
	bool OnRequestClose();
	void OnPulse();

	int OnDrop(uchar *Data, int Length, GdcPt2 Pt, int KeyState);
	int WillAccept(List<int> &Formats, GdcPt2 Pt, int KeyState);
};

#include "HttpDlg.h"