/*
**	FILE:			FtpApp.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			7/9/99
**	DESCRIPTION:	i.Ftp application
**
**	Copyright (C) 1999, Matthew Allen
**		fret@ozemail.com.au
*/

// Debug defines

// Includes
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <stdarg.h>

#include "Ftp.h"
#include "GToken.h"
#include "IHttp.h"
#include "GAbout.h"

#define IDM_HTTP				10001

#define IDLE_NOOP_TIME			60

#ifdef BEOS
#define isdigit(c) ((c) >= '0' AND (c) <= '9')
#endif

#define GFSV_DIR				(WM_USER+1200)
#define GFSV_LIST				(WM_USER+1201)

//////////////////////////////////////////////////////////////////////////////
char AppName[] = "i.Ftp";
char HelpFile[] = "index.html";
char OptionsFileName[] = "options.sgml";

//////////////////////////////////////////////////////////////////////////////
int Sorter(GListItem *a, GListItem *b, int Data)
{
	DirListItem *A = dynamic_cast<DirListItem*>(a);
	DirListItem *B = dynamic_cast<DirListItem*>(b);
	if (A AND B)
	{
		if (A->Entry->IsDir() AND
			NOT B->Entry->IsDir())
		{
			return -1;
		}
		if (NOT A->Entry->IsDir() AND
			B->Entry->IsDir())
		{
			return 1;
		}

		return stricmp(A->Entry->Name, B->Entry->Name);
	}
	return 0;
}

//////////////////////////////////////////////////////////////////////////////
DirListItem::DirListItem(IFtpEntry *entry, GFileSystemView *view)
{
	_UserData = this;
	SizeStr[0] = 0;
	Entry = entry;
	View = view;
}

DirListItem::~DirListItem()
{
	DeleteObj(Entry);
}

void DirListItem::OnMouseClick(GMouse &m)
{
	if (View)
	{
		View->OnClick(Entry, &m);
	}
}

int DirListItem::GetImage(int Flags)
{
	if (Entry)
	{
		if (Entry->Attributes & IFTP_SYM_LINK)
		{
			return 2;
		}
		
		return (Entry->IsDir()) ? 0 : 1;
	}

	return 1;
}

char *DirListItem::GetText(int i)
{
	if (NOT Entry) return 0;

	switch (i)
	{
		case 0:
		{
			if (Entry->Name) return Entry->Name;
			break;
		}
		case 1:
		{
			if (Entry->IsDir())
			{
				return "<dir>";
			}
			else if (Entry->Name)
			{
				char *Dot = strrchr(Entry->Name, '.');
				return (Dot) ? Dot + 1 : 0;
			}
			break;
		}
		case 2:
		{
			FormatSize(SizeStr, Entry->Size);
			return SizeStr;
			break;
		}
		case 3:
		{
			if (Entry->Date) return Entry->Date;
			break;
		}
		case 4:
		{
			if (Entry->Time) return Entry->Time;
			break;
		}
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////////////
ILog::ILog(RLog *log)
{
	Info = 0;
	Logger = log;
	Logging = true;
}

ILog::~ILog()
{
}

void ILog::WriteLog(char *s, int Len, COLOUR c)
{
	if (Logging AND
		Logger AND
		s AND
		Len > 0)
	{
		if (Len > 5 AND
			strnicmp(s, "PASS ", 5) == 0)
		{
			Logger->Write(c, "PASS *******", 12);
			return;
		}
		
		char *Eol;
		while ((Len > 0) AND (Eol = strnstr(s, "\r\n", Len)))
		{
			int l = (int)Eol-(int)s;
			Logger->Write(c, s, l);
			s += l + 2;
			Len -= l + 2;
		}

		if (Len > 0)
		{
			Logger->Write(c, s, Len);
		}
	}
}

void ILog::LogRead(char *Data, int Len)
{
	if (Info)
	{
		Info->Value(Info->Value() + Len);
	}

	if (Data)
	{
		bool Error = Data[0] == '5' AND
						IsDigit(Data[1]) AND
						IsDigit(Data[2]) AND
						Data[3] == ' ';

		WriteLog(Data, Len, Error ? FTP_COL_ERROR : FTP_COL_READ);
	}
}

void ILog::LogWrite(char *Data, int Len)
{
	if (Info)
	{
		Info->Value(Info->Value() + Len);
	}

	if (Data)
	{
		WriteLog(Data, Len, FTP_COL_WRITE);
	}
}

void ILog::LogError(int ErrorCode, char *ErrorDescription)
{
	if (Logger)
	{
		if (ErrorCode >= 0)
		{
			Logger->Print(FTP_COL_ERROR, "Error(%i): %s", ErrorCode, ErrorDescription);
		}
		else
		{
			Logger->Print(FTP_COL_ERROR, "Error: %s", ErrorDescription);
		}
	}
}

void ILog::LogInformation(char *Str)
{
	if (Logger)
	{
		Logger->Print(FTP_COL_INFO, "Information: %s", Str);
	}
}

void ILog::LogSetParameter(int Which, int What)
{
	switch (Which)
	{
		case PARAMETER_LOG:
		{
			Logging = What != 0;
			break;
		}
		case PARAMETER_PROGRESS:
		{
			Info = (Progress*) What;
			break;
		}
	}
}

///////////////////////////////////////////////////////////////
ILogConnection::ILogConnection(RLog *log, GNetwork *parent, AppWnd *wnd) :
	ILog(log)
{
	Wnd = wnd;
	Parent = parent;
}

void ILogConnection::OnDisconnect()
{
	if (Wnd)
	{
		Wnd->OnDisconnect();
	}

	LogInformation("[INET] Disconnect");
}

void ILogConnection::OnRead(char *Data, int Len) { LogRead(Data, Len); }
void ILogConnection::OnWrite(char *Data, int Len) { LogWrite(Data, Len); }
void ILogConnection::OnError(int ErrorCode, char *ErrorDescription) { LogError(ErrorCode, ErrorDescription); }
void ILogConnection::OnInformation(char *Str) { LogInformation(Str); }
void ILogConnection::SetParameter(int Which, int What) { LogSetParameter(Which, What); }

//////////////////////////////////////////////////////////////////////////////
ISocks5LogConnection::ISocks5LogConnection(	char *proxy,
											int port,
											char *username,
											char *password,
											RLog *log,
											GNetwork *parent,
											AppWnd *wnd) :
	ILog(log),
	GSocks5Socket(proxy, port, username, password)
{
	Wnd = wnd;
	Parent = parent;
}

void ISocks5LogConnection::OnDisconnect()
{
	if (Wnd)
	{
		Wnd->OnDisconnect();
	}

	LogInformation("[INET] Disconnect");
}

void ISocks5LogConnection::OnRead(char *Data, int Len) { LogRead(Data, Len); }
void ISocks5LogConnection::OnWrite(char *Data, int Len) { LogWrite(Data, Len); }
void ISocks5LogConnection::OnError(int ErrorCode, char *ErrorDescription) { LogError(ErrorCode, ErrorDescription); }
void ISocks5LogConnection::OnInformation(char *Str) { LogInformation(Str); }
void ISocks5LogConnection::SetParameter(int Which, int What) { LogSetParameter(Which, What); }

//////////////////////////////////////////////////////////////////////////////
char *OsName()
{
	#ifdef WIN32
	return "Win32";
	#endif
	#ifdef BEOS
	return "BeOS";
	#endif
	#ifdef LINUX
	return "Linux";
	#endif
}

void GetFullAppName(char *s)
{
	if (s)
	{
		#ifndef _DEBUG
		char *Build = "Release";
		#else
		char *Build = "Debug";
		#endif

		sprintf(s, "%s v%02.2f (%s %s)", AppName, VER, OsName(), Build);
	}
}

int SizeOfFile(char *FileName)
{
	GFile F;
	if (F.Open(FileName, O_READ))
	{
		return F.GetSize();
	}
	return 0;
}

void Error(char *str, ...)
{
	if (str)
	{
		char buffer[256];
		va_list arg;
		va_start(arg ,str);
		vsprintf(buffer, str, arg);
		va_end(arg);

		LgiMsg(MainWnd, buffer, "Scribe: Error", MB_OK);
	}
}

//////////////////////////////////////////////////////////////////////////////
void ConvertUrlToAccount(char *Url, ObjProperties *Account)
{
	/*
	Fields.Insert(NEW(DataDlgField(DATA_STR, IDC_SERVER, OPT_RemoteHost)));
	Fields.Insert(NEW(DataDlgField(DATA_INT, IDC_PORT, OPT_Port)));
	Fields.Insert(NEW(DataDlgField(DATA_STR, IDC_USERNAME, OPT_UserName)));
	Fields.Insert(NEW(DataDlgField(DATA_STR, IDC_PASSWORD, OPT_Password)));
	Fields.Insert(NEW(DataDlgField(DATA_BOOL, IDC_ANONYMOUS, OPT_Anonymous)));
	Fields.Insert(NEW(DataDlgField(DATA_STR, IDC_REMOTE, OPT_RemoteDir)));
	Fields.Insert(NEW(DataDlgField(DATA_STR, IDC_LOCAL, OPT_LocalDir)));
	Fields.Insert(NEW(DataDlgField(DATA_BOOL, IDC_USE_ACTIVE, OPT_UseActive)));
	*/

	if (Url AND Account)
	{
		// check for URL and decode
		char *Protocol = "ftp://";

		if (strnicmp(Url, Protocol, strlen(Protocol)) == 0)
		{
			// ftp:// URL
			char *Buf = NewStr(Url + strlen(Protocol));
			char *s = Buf;
			
			// User/Password
			Account->Set(OPT_Anonymous, (int)1);
			char *At = strchr(s, '@');
			if (At)
			{
				*At = 0;
				char *Colon = strchr(s, ':');
				if (Colon)
				{
					*Colon = 0;
					Account->Set(OPT_UserName, s);
					Account->Set(OPT_Password, Colon+1);
					Account->Set(OPT_Anonymous, (int)0);
				}
				s = At + 1;
			}

			// Remote path
			char *Dir = strchr(s, '/');
			if (Dir)
			{
				// Chop off any file name
				char *LastDir = strrchr(Dir, '/');
				if (LastDir AND strchr(LastDir, '.'))
				{
					// er it's a file name... choppy chop
					*LastDir = 0;
				}

				if (strlen(Dir) > 0)
				{
					Account->Set(OPT_RemoteDir, Dir + 1);
				}
				*Dir++ = 0;
			}

			// Port?
			char *Colon = strchr(s, ':');
			if (Colon)
			{
				*Colon++ = 0;
				Account->Set(OPT_Port, atoi(Colon));
			}

			// All thats left is the remote host
			Account->Set(OPT_RemoteHost, s);

			DeleteArray(Buf);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
char *TempStr = "(temporary)";
#define	IDC_TAB							90

class ConnectDlg : public GDialog, public DataDlgTools
{
	List<ObjProperties> *Accounts;
	ObjProperties *Temp;
	int Index;
	GCombo *AccountCombo;
	DataDlgFieldList Fields;

	DataDlgFieldList OptFields;
	DataDlgTools OptTools;

	ObjProperties *GetAccount(char *Name)
	{
		if (Name)
		{
			for (ObjProperties *a = Accounts->First(); a; a = Accounts->Next())
			{
				char *n;
				if (a->Get(OPT_AccountName, n) AND
					stricmp(Name, n) == 0)
				{
					return a;
				}
			}
		}
		return 0;
	}

	void LoadAccountList()
	{
		if (AccountCombo AND Accounts)
		{
			while (AccountCombo->Delete(0));

			for (ObjProperties *a = Accounts->First(); a; a = Accounts->Next())
			{
				char *n;
				if (a->Get(OPT_AccountName, n))
				{
					AccountCombo->Insert(n);
				}
				else
				{
					Accounts->Delete(a);
					DeleteObj(a);
				}
			}

			AccountCombo->Index(Index);
		}
	}

public:
	ObjProperties *Account;

	ConnectDlg(AppWnd *parent, List<ObjProperties> *accounts)
	{
		SetParent(parent);
		Accounts = accounts;
		Dlg = this;
		Account = 0;
		Index = 0;

		if (parent)
		{
			OptTools.Set(this, parent->GetOptions());
			OptFields.Insert(NEW(DataDlgField(DATA_STR, IDC_SOCKS5_SERVER, OPT_Socks5Server)));
			OptFields.Insert(NEW(DataDlgField(DATA_STR, IDC_SOCKS5_USERNAME, OPT_Socks5UserName)));
			OptFields.Insert(NEW(DataDlgField(DATA_STR, IDC_SOCKS5_PASSWORD, OPT_Socks5Password)));
		}

		if (LoadFromResource(MAKEINTRESOURCE(IDD_CONNECT)))
		{
			GTabView *Tab = 0;
			#ifdef WIN32
			Children.Insert(Tab = NEW(GTabView(IDC_TAB, 5, 5, Pos.X()-24, Pos.Y()-68, "Tab")));
			#else if BEOS
			Children.Insert(Tab = NEW(GTabView(IDC_TAB, 5, 5, Pos.X()-24, Pos.Y()-58, "Tab")));
			#endif
			if (Tab AND Accounts)
			{
				GTabPage *Connect = Tab->Append("Remote Host");
				if (Connect)
				{
					Connect->AppendDialogFromResource(MAKEINTRESOURCE(IDD_REMOTE));
				}

				GTabPage *Proxy = Tab->Append("Proxy/Firewall");
				if (Proxy)
				{
					Proxy->AppendDialogFromResource(MAKEINTRESOURCE(IDD_SOCKS5));
				}

				MoveToCenter();

				Options = Temp = NEW(ObjProperties);
				if (Options)
				{
					Temp->Set(OPT_AccountName, TempStr);
					Temp->Set(OPT_Port, 21);

					Accounts->Insert(Temp, 0);
				}

				AccountCombo = dynamic_cast<GCombo*>(FindControl(IDC_ACCOUNT));
				LoadAccountList();

				Fields.Insert(NEW(DataDlgField(DATA_STR, IDC_SERVER, OPT_RemoteHost)));
				Fields.Insert(NEW(DataDlgField(DATA_INT, IDC_PORT, OPT_Port)));
				Fields.Insert(NEW(DataDlgField(DATA_STR, IDC_USERNAME, OPT_UserName)));
				Fields.Insert(NEW(DataDlgField(DATA_STR, IDC_PASSWORD, OPT_Password)));
				Fields.Insert(NEW(DataDlgField(DATA_BOOL, IDC_ANONYMOUS, OPT_Anonymous)));
				Fields.Insert(NEW(DataDlgField(DATA_STR, IDC_REMOTE, OPT_RemoteDir)));
				Fields.Insert(NEW(DataDlgField(DATA_STR, IDC_LOCAL, OPT_LocalDir)));
				Fields.Insert(NEW(DataDlgField(DATA_BOOL, IDC_USE_ACTIVE, OPT_UseActive)));

				// Data -> Controls
				Serialize(false);
				
				// Load options
				OptTools.ProcessFields(OptFields, false);

				DoModal();
			}
		}
	}

	~ConnectDlg()
	{
		Accounts->Delete(Temp);
		DeleteObj(Temp);
	}

	char *GetEditStr(int Which)
	{
		GEdit *e = dynamic_cast<GEdit*>(FindControl(Which));
		if (e)
		{
			return e->GEdit::Name();
		}
		return 0;
	}

	void SetEditStr(int Which, char *Str)
	{
		GEdit *e = dynamic_cast<GEdit*>(FindControl(Which));
		if (e)
		{
			e->GEdit::Name((Str)?Str:(char*)"");
		}
	}

	void Serialize(bool Write)
	{
		if (Write)
		{
			GCheckBox *Anon = dynamic_cast<GCheckBox*>(FindControl(IDC_ANONYMOUS));

			// check for URL and decode
			char *Protocol = "ftp://";

			char *Server = NewStr(GetEditStr(IDC_SERVER));
			if (Server AND strnicmp(Server, Protocol, strlen(Protocol)) == 0)
			{
				// ftp:// URL
				char *s = Server + strlen(Protocol);
				
				// User/Password
				char *At = strchr(s, '@');
				if (At)
				{
					*At = 0;
					char *Colon = strchr(s, ':');
					if (Colon)
					{
						*Colon = 0;
						SetEditStr(IDC_USERNAME, s);
						SetEditStr(IDC_PASSWORD, Colon+1);
					}
					s = At + 1;

					if (Anon)
					{
						Anon->Value(false);
					}
				}
				else if (Anon)
				{
					Anon->Value(true);
				}

				// Remote path
				char *Dir = strchr(s, '/');
				if (Dir)
				{
					// Chop off any file name
					char *LastDir = 0;
					for (char *End = Dir+strlen(Dir)-1; End>Dir; End--)
					{
						if (*End == '/')
						{
							LastDir = End;
							break;
						}
					}
					if (LastDir AND strchr(LastDir, '.'))
					{
						// er it's a file name... choppy chop
						*LastDir = 0;
					}

					SetEditStr(IDC_REMOTE, Dir);
					*Dir++ = 0;
				}

				// Port?
				char *Colon = strchr(s, ':');
				if (Colon)
				{
					*Colon++ = 0;
					SetEditStr(IDC_PORT, Colon);
				}

				// All thats left is the remote host
				SetEditStr(IDC_SERVER, s);
			}
			DeleteArray(Server);

			// Check for user / anon
			if (Anon AND NOT ValidStr(GetEditStr(IDC_USERNAME)))
			{
				Anon->Value(true);
			}

			// Data -> Controls
			SaveFields(Fields);
		}
		else
		{
			// Data -> Controls
			LoadFields(Fields);
		}
	}

	int OnNotify(GView *Ctrl, int Flags)
	{
		switch (Ctrl->GetId())
		{
			case IDC_ACCOUNT:
			{
				if (AccountCombo)
				{
					int i = AccountCombo->Index();
					if (i >= 0 AND Index != i)
					{
						Index = i;
						ObjProperties *Acc = Accounts->ItemAt(Index);
						if (Acc)
						{
							Serialize(true);
							Options = Acc;
							Serialize(false);
							AccountCombo->Index(Index);
						}
					}
					// else user is typing something
				}
				break;
			}
			case IDC_ADD:
			{
				if (AccountCombo)
				{
					// Controls -> Data
					Serialize(true);

					ObjProperties *Acc = 0;
					if (Options)
					{
						// save the temp into the list
						Acc = NEW(ObjProperties);
						if (Acc)
						{
							// copy to a new object
							*Acc = *Temp;

							// Work out the account name
							char AName[256] = "";
							GInput Dlg(this, AName, "Enter account name:", AppName);
							if (Dlg.DoModal() AND Dlg.Str)
							{
								strcpy(AName, Dlg.Str);
							}
							else return 0;
							
							Acc->Set(OPT_AccountName, AName);

							// new current record
							Options = Acc;
							Accounts->Insert(Acc);
							Index = Accounts->IndexOf(Acc);

							// delete keys out of temp... starts blank
							while (Temp->FirstKey())
							{
								Temp->DeleteKey();
							}
							Temp->Set(OPT_AccountName, TempStr);
							Temp->Set(OPT_Port, 21);

							// load the data back into the UI
							// Data -> Controls
							Serialize(false);
						}
					}

					LoadAccountList();
				}
				break;
			}
			case IDC_DELETE:
			{
				if (AccountCombo AND Index > 0)
				{
					Accounts->Delete(Options);
					Options = Accounts->ItemAt(--Index);
					Serialize(false); // Data -> Controls
					LoadAccountList();
				}
				break;
			}
			case IDC_BROWSE:
			{
				GEdit *LocalDir = dynamic_cast<GEdit*>(FindControl(IDC_LOCAL));
				if (LocalDir)
				{
					GFileSelect Select;
		
					Select.Name(LocalDir->GEdit::Name());
					Select.Parent(this);
					if (Select.OpenFolder())
					{
						LocalDir->GEdit::Name(Select.Name());
					}
				}
				break;
			}
			case IDC_CONNECT:
			{
				// Save connect fields
				Serialize(true);
				Account = Options;

				// Save options
				OptTools.ProcessFields(OptFields, true);

				EndModal(1);
				break;
			}
			case IDCANCEL:
			{
				SaveFields(Fields);

				// Save options
				OptTools.ProcessFields(OptFields, true);

				EndModal(0);
				break;
			}
		}
		return 0;
	}
};

ResumeDlg::ResumeDlg(GView *parent, IFtpEntry *e, bool Resumable)
{
	Pos.ZOff(340, 100);
	MoveToCenter();
	SetParent(parent);
	if (e)
	{
		char Str[256];
		sprintf(Str, "The file '%s' already exists, what would you like to do?", e->Name);
		Children.Insert(NEW(GText(-1, 10, 10, -1, -1, Str)));
		
		int Buttons = Resumable ? 3 : 2;
		int Width = (Buttons * 60) + ((Buttons - 1) * 10);
		int x = (Pos.X() - Width) / 2;
		int y = 40;
		Children.Insert(NEW(GButton(IDC_OVERWRITE, x, y, 60, 20, "OverWrite")));
		x += 70;
		if (Resumable)
		{
			Children.Insert(NEW(GButton(IDC_RESUME, x, y, 60, 20, "Resume")));
			x += 70;
		}
		Children.Insert(NEW(GButton(IDCANCEL, x, y, 60, 20, "Cancel")));
		x += 70;
	}
}

int ResumeDlg::OnNotify(GView *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
		case IDC_OVERWRITE:
		case IDC_RESUME:
		case IDCANCEL:
		{
			EndModal(Ctrl->GetId());
			break;
		}
	}

	return 0;
}

class PermsDlg : public GDialog
{
	bool Win32;
	GCheckBox *Chk[9];

public:
	int Perms;

	PermsDlg(GView *parent, bool win32, int Init = 0)
	{
		Win32 = win32;
		Perms = Init;
		SetParent(parent);
		GView::Name("Set Permissions");

		if (Win32)
		{
			// Read/System/Hidden/Archive
			Children.Insert(Chk[0] = NEW(GCheckBox(100, 40, 10, -1, -1, "Read Only")));
			Children.Insert(Chk[1] = NEW(GCheckBox(101, 40, 40, -1, -1, "Archive")));
			Children.Insert(Chk[2] = NEW(GCheckBox(102, 40, 70, -1, -1, "Hidden")));
			Children.Insert(Chk[3] = NEW(GCheckBox(103, 40, 100, -1, -1, "System")));

			if (Perms & IFTP_READ) Chk[0]->Value(true);
			if (Perms & IFTP_ARCHIVE) Chk[1]->Value(true);
			if (Perms & IFTP_HIDDEN) Chk[2]->Value(true);
			if (Perms & IFTP_SYSTEM) Chk[3]->Value(true);

			Children.Insert(NEW(GButton(IDOK,		30, 140, 60, 20, "Ok")));
			Children.Insert(NEW(GButton(IDCANCEL,	100, 140, 60, 20, "Cancel")));

			Pos.ZOff(200, 200);
		}
		else
		{
			// Read/Write/Execute Owner/Group/Global

			// Row headings
			Children.Insert(NEW(GText(-1, 10, 40, -1, -1, "Owner")));
			Children.Insert(NEW(GText(-1, 10, 40+30, -1, -1, "Group")));
			Children.Insert(NEW(GText(-1, 10, 40+60, -1, -1, "Global")));

			// Col headings
			Children.Insert(NEW(GText(-1, 70, 10, -1, -1, "Read")));
			Children.Insert(NEW(GText(-1, 70+50, 10, -1, -1, "Write")));
			Children.Insert(NEW(GText(-1, 70+100, 10, -1, -1, "Execute")));

			// Checkboxes
			int n=0;
			for (int y=40; y<=100; y+=30)
			{
				for (int x=70; x<=180; x+=50)
				{
					Children.Insert(Chk[n] = NEW(GCheckBox(100+n, x, y, 20, 20, "")));
					n++;
				}
			}

			if (Perms & IFTP_READ) Chk[0]->Value(true);
			if (Perms & IFTP_WRITE) Chk[1]->Value(true);
			if (Perms & IFTP_EXECUTE) Chk[2]->Value(true);
			if (Perms & IFTP_GRP_READ) Chk[3]->Value(true);
			if (Perms & IFTP_GRP_WRITE) Chk[4]->Value(true);
			if (Perms & IFTP_GRP_EXECUTE) Chk[5]->Value(true);
			if (Perms & IFTP_GLOB_READ) Chk[6]->Value(true);
			if (Perms & IFTP_GLOB_WRITE) Chk[7]->Value(true);
			if (Perms & IFTP_GLOB_EXECUTE) Chk[8]->Value(true);

			Children.Insert(NEW(GButton(IDOK, 70, 140, 60, 20, "Ok")));
			Children.Insert(NEW(GButton(IDCANCEL, 140, 140, 60, 20, "Cancel")));

			Pos.ZOff(260, 200);
		}

		MoveToCenter();
	}

	int OnNotify(GView *Ctrl, int Flags)
	{
		switch (Ctrl->GetId())
		{
			case IDOK:
			{
				Perms = 0;
				if (Win32)
				{
					if (Chk[0]->Value()) Perms |= IFTP_READ;
					if (Chk[1]->Value()) Perms |= IFTP_ARCHIVE;
					if (Chk[2]->Value()) Perms |= IFTP_HIDDEN;
					if (Chk[3]->Value()) Perms |= IFTP_SYSTEM;
				}
				else
				{
					if (Chk[0]->Value()) Perms |= IFTP_READ;
					if (Chk[1]->Value()) Perms |= IFTP_WRITE;
					if (Chk[2]->Value()) Perms |= IFTP_EXECUTE;
					
					if (Chk[3]->Value()) Perms |= IFTP_GRP_READ;
					if (Chk[4]->Value()) Perms |= IFTP_GRP_WRITE;
					if (Chk[5]->Value()) Perms |= IFTP_GRP_EXECUTE;
					
					if (Chk[6]->Value()) Perms |= IFTP_GLOB_READ;
					if (Chk[7]->Value()) Perms |= IFTP_GLOB_WRITE;
					if (Chk[8]->Value()) Perms |= IFTP_GLOB_EXECUTE;
				}
				// fall thru
			}
			case IDCANCEL:
			{
				EndModal(Ctrl->GetId());
				break;
			}
		}

		return 0;
	}
};

//////////////////////////////////////////////////////////////////////////////
#ifdef WIN32
GWin32Class AppWnd::Class;
#endif

AppWnd::AppWnd()
{
	// init some variables
	GApp::ObjInstance()->AppWnd = this;
	ImageList = 0;
	Splitter = 0;
	OptionsFile = 0;
	Connection = 0;
	Ftp = 0;
	ImageList = 0;
	MaxDataRate = -1;
	AccountLst = 0;
	Local = 0;
	LastCmd = 0;

	Http = 0;
	Worker = NEW(FtpWorkerThread);

	// Startup UI
	GRegion r(50, 50, 870, 550);
	SetPos(r); // default location
	Serialize(false);
	Name(AppName);
	DropTarget(true);

	#ifdef WIN32
	Class.Class.hIcon = LoadIcon(hAppInst(), MAKEINTRESOURCE(IDI_ICON1));
	Class.Create(AppName);
	if (NOT Create(Class.Name(), NULL))
	{
		LgiMsg(0, "Couldn't create main window.");
		LgiExitApp();
	}
	else
	#endif
	{
		SetWindow(this);
		SetupUi();
	}

	DoCmdLine();
}

AppWnd::~AppWnd()
{
	if (Splitter)
	{
		Options.Set("split-pos", Splitter->Split());
	}

	Serialize(true);
	DeleteObj(OptionsFile);
	DeleteObj(ImageList);
	DeleteObj(Worker);
	DeleteObj(Progress);
	DeleteObj(Ftp);

	for (ObjProperties *a = Accounts.First(); a; a = Accounts.Next())
	{
		DeleteObj(a);
	}

	#ifdef BEOS
	LgiExitApp();
	#endif
}

void AppWnd::DoCmdLine()
{
	char Src[256];
	char Dest[256];

	if (GApp::ObjInstance()->GetOption(Src, 's') AND
		GApp::ObjInstance()->GetOption(Dest, 'd'))
	{
		ObjProperties Account;

		if (stristr(Src, "ftp://"))
		{
			// Download
		}
		else if (stristr(Dest, "ftp://"))
		{
			// Upload
			ConvertUrlToAccount(Dest, &Account);
			ConnectTo(&Account);

			// Post Upload Cmd
			FtpCmd *Cmd = NEW(FtpCmd);
			if (Cmd)
			{
				Cmd->Cmd = TASK_Transfer;
				Cmd->View = Local;
				Cmd->Connection = Connection;
				strcpy(Cmd->Arg, Src);

				Local->PostCmd(Cmd);
			}

			// Post List Cmd
			Cmd = NEW(FtpCmd);
			if (Cmd)
			{
				Cmd->Cmd = TASK_ListElements;
				Cmd->View = Remote;
				Cmd->Connection = Connection;
				Remote->PostCmd(Cmd);
			}
		}
	}
}	

void AppWnd::OnPulse()
{
	int Now = LgiCurrentTime();
	int Diff = Now - LastCmd;
	bool Running = Worker ? Worker->IsRunning() : false;

	if (LastCmd AND
		Remote AND
		NOT Running AND
		Diff > (IDLE_NOOP_TIME * 1000))
	{
		FtpCmd *Cmd = NEW(FtpCmd);
		if (Cmd)
		{
			Cmd->Cmd = TASK_Noop;
			Cmd->View = Remote;
			Remote->PostCmd(Cmd);
		}
	}

	if (StatusInfo[0])
	{
		char Str[256] = "";
		if (Running)
		{
			sprintf(Str, "Busy (Queue: %i)", Worker?Worker->GetCmdsLeft()+1:0);
		}
		else
		{
			int Seconds = (LgiCurrentTime() - LastCmd) / 1000;
			sprintf(Str, "Idle: %i:%02.2i", Seconds/60, Seconds%60);
		}
		StatusInfo[0]->Name(Str);
	}
}

void AppWnd::SetDataRate(int Max)
{
	MaxDataRate = Max;
}

int AppWnd::WritePipe(GSocket *c, char *Data, int Len, int Flags)
{
	if (MaxDataRate >= 0)
	{
		int InitialLen = Len;
		while (Len > 0)
		{
			int Now = LgiCurrentTime();
			int SecondLen = min(Len, MaxDataRate);
			int Written = c->GSocket::Write(Data, SecondLen, Flags);
			if (Written > 0)
			{
				Data += Written;
				Len -= Written;
				while (LgiCurrentTime() - Now < 1000)
				{
					// wait till next second
				}
			}
			else
			{
				break;
			}
		}

		return InitialLen - Len;
	}
	else
	{
		// Push it through as fast as we can
		return c->GSocket::Write(Data, Len, Flags);
	}

	return 0;
}

int AppWnd::ReadPipe(GSocket *c, char *Data, int Len, int Flags)
{
	if (MaxDataRate >= 0)
	{
		int InitialLen = Len;
		while (Len > 0)
		{
			int Now = LgiCurrentTime();
			int SecondLen = min(Len, MaxDataRate);
			int Read = c->GSocket::Read(Data, SecondLen, Flags);
			if (Read > 0)
			{
				Data += Read;
				Len -= Read;
				while (LgiCurrentTime() - Now < 1000)
				{
					// wait till next second
				}
			}
			else
			{
				break;
			}
		}

		return InitialLen - Len;
	}
	else
	{
		// Read it in as fast as we can
		return c->GSocket::Read(Data, Len, Flags);
	}

	return 0;
}

bool AppWnd::OnRequestClose()
{
	if (Worker)
	{
		Worker->Terminate();
	}

	if (Http)
	{
		Http->Terminate();
	}

	return true;
}

void AppWnd::SetupUi()
{
	Menu = NEW(GMenu);
	if (Menu)
	{
		#ifdef BEOS
		Menu->Attach(this);
		#endif

		File = Menu->AppendSub("&File");
		if (File)
		{
			CmdConnect.MenuItem = File->AppendItem("&Connect", IDM_CONNECT, true);
			CmdDisconnect.MenuItem = File->AppendItem("Discon&nect", IDM_DISCONNECT, false);
			File->AppendSeparator();

			// broken
			CmdHttp.MenuItem = File->AppendItem("HTTP Download", IDM_HTTP, true);
			File->AppendSeparator();

			File->AppendItem("E&xit", IDM_EXIT, true);
		}

		AccountLst = Menu->AppendSub("&Accounts");

		Help = Menu->AppendSub("&Help");
		if (Help)
		{
			Help->AppendItem("&Help",		IDM_HELP, TRUE);
			Help->AppendSeparator();
			Help->AppendItem("&About",		IDM_ABOUT, TRUE);
		}

		#ifdef WIN32
		Menu->Attach(this);
		#endif
	}

	Commands = LgiLoadToolbar(this, "_commands.gif");
	if (Commands)
	{
		Commands->Attach(this);

		CmdConnect.ToolButton = Commands->AppendButton("Connect", IDM_CONNECT, TBT_PUSH);
		CmdDisconnect.ToolButton = Commands->AppendButton("Disconnect", IDM_DISCONNECT, TBT_PUSH, false);
		Commands->AppendSeparator();

		CmdAbort.ToolButton = Commands->AppendButton("Abort Transfer", IDM_ABORT, TBT_PUSH, false);
	}

	char *FileName = LgiFindFile("_icons.gif");
	if (FileName)
	{
		GSurface *pDC = LoadDC(FileName);
		if (pDC)
		{
			ImageList = NEW(GImageList(16, 16, pDC));
			#ifdef WIN32
			if (ImageList)
			{
				ImageList->Create(pDC->X(), pDC->Y(), GdcD->GetBits());
			}
			#endif
		}
		DeleteArray(FileName);
	}

	Status = NEW(GStatusBar);
	if (Status)
	{
		Status->Attach(this);
		StatusInfo[0] = Status->AppendPane("", 2);
		Progress = NEW(FileTransferProgress(this, Status, StatusInfo+1, true));
	}

	LogPanel = NEW(GLogPanel);
	if (LogPanel)
	{
		Pour();
		LogPanel->SetClosedSize((Commands) ? Commands->Y()-1 : 16);
		LogPanel->Attach(this);
	}

	// Worker thread doesn't need to suck too much CPU
	if (Worker)
	{
		Worker->Priority(PRIORITY_LOW);
	}

	// Ftp protocol class
	Ftp = NEW(IFtp(this));
	if (Ftp)
	{
		Ftp->Meter = Progress;
	}

	// Split the local / remote views..
	Splitter = NEW(GSplitter);
	if (Splitter)
	{
		int Pos = 400;
		Options.Get("split-pos", Pos);

		Splitter->Attach(this);
		Splitter->Split(Pos);

		Local = NEW(GLocalFS(this, ImageList, Ftp));
		Remote = NEW(GRemoteFS(this, ImageList, Ftp));

		if (Local)
		{
			Local->Worker = Worker;
			Local->Opposite = Remote;
			Splitter->SetViewA(Local, true);
		}

		if (Remote)
		{
			Remote->Worker = Worker;
			Remote->Opposite = Local;
			Splitter->SetViewB(Remote, true);
		}
	}

	if (Local)
	{
		Local->ListElements();
	}

	EnableCommands(false);
	UpdateMenus();

	#ifdef WIN32
	Pour();
	ShowWindow(Handle(), LgiApp->Show());
	UpdateWindow(Handle());
	#endif
	#ifdef BEOS
	Visible(true);
	#endif

}

void AppWnd::UpdateMenus()
{
	if (AccountLst)
	{
		// empty current menu
		while (AccountLst->RemoveItem(0));

		// refill with current accounts
		int i=0;
		for (ObjProperties *p=Accounts.First(); p; p=Accounts.Next(), i++)
		{
			char *Name;
			if (p->Get(OPT_AccountName, Name))
			{
				AccountLst->AppendItem(Name, IDM_ACCOUNT_BASE+i, true);
			}
		}
	}
}

void AppWnd::ResetThread()
{
	if (Worker->IsRunning() AND Ftp)
	{
		// close the files and sockets...
		Ftp->Close();
	}

	// terminate...
	if (Worker)
	{
		Worker->Terminate();
		DeleteObj(Worker);
	}

	// setup a new thread...
	Worker = NEW(FtpWorkerThread);

	if (Local)
	{
		Local->Worker = Worker;
	}

	if (Remote)
	{
		Remote->Worker = Worker;
	}
}

void AppWnd::EnableCommands(bool Enabled)
{
	CmdConnect.Enabled(NOT Enabled);
	CmdHttp.Enabled(NOT Enabled);
	CmdDisconnect.Enabled(Enabled);
	CmdAbort.Enabled(false);

	if (Local)
	{
		Local->EnableCommands(Enabled);
	}

	if (Remote)
	{
		Remote->EnableCommands(Enabled);
	}

	if (Enabled)
	{
		LastCmd = LgiCurrentTime();
		SetPulse(100);
	}
	else
	{
		LastCmd = 0;
		SetPulse();
	}
}

void AppWnd::OnCmd(FtpCmd *Cmd)
{
	if (NOT Cmd->View->IsCmdLocal(Cmd->Cmd))
	{
		LastCmd = LgiCurrentTime();
	}

	if (Connection AND
		NOT Connection->IsOpen())
	{
		EnableCommands(false);
	}
}

int AppWnd::OnDrop(uchar *Data, int Length, GdcPt2 Pt, int KeyState)
{
	if (Length > 5)
	{
		if ((Data[0] == 'f' OR Data[0] == 'h') AND Data[1] == 0)
		{
			// probably a UNICODE string... naughty, try and convert
			char *Out = (char*)Data;
			for (int16 *In = (int16*)Data; *In AND *In != '\n'; *In++)
			{
				*Out++ = *In;
			}
			*Out++ = 0;
		}

		if (strnicmp((char*)Data, "ftp://", 6) == 0)
		{
			// Ftp URL
			ObjProperties Account;
			ConvertUrlToAccount((char*)Data, &Account);
			ConnectTo(&Account);
		}
		else if (strnicmp((char*)Data, "http://", 7) == 0)
		{
			// Http URL
			Http = NEW(HttpThread(this, (char*)Data));
		}

		return DROPEFFECT_COPY;
	}

	return DROPEFFECT_NONE;
}

int AppWnd::WillAccept(List<int> &Formats, GdcPt2 Pt, int KeyState)
{
	int Status = DROPEFFECT_NONE;
	int Url = RegisterClipboardFormat(CFSTR_SHELLURL);

	for (int *i=Formats.First(); i; )
	{
		if (*i == Url)
		{
			// hello girls.... a URL
			Status = DROPEFFECT_COPY;
			i = Formats.Next();
		}
		else
		{
			// tell the d'n'd code that we don't support this
			Formats.Delete(i);
			i = Formats.Current();
		}
	}

	return DROPEFFECT_NONE;
}

int AppWnd::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case WM_CHANGE:
		{
			return OnNotify((GView*) MsgA(Msg), MsgB(Msg));
		}
		case WM_DESCRIBE:
		{
			char *Text = (char*) MsgB(Msg);
			if (Text)
			{
				// SetStatusText(Text, 0);
			}
			break;
		}
	}
	
	return GWindow::OnEvent(Msg);
}

void AppWnd::OnDisconnect()
{
	EnableCommands(false);

	if (Remote)
	{
		Remote->SetCtrlName(GFSV_DIR, "");
		Remote->Empty();
	}

	if (StatusInfo[0])
	{
		StatusInfo[0]->Name("");
	}
	
	if (Progress)
	{
		Progress->Value(0);
	}
}

void AppWnd::ConnectTo(ObjProperties *Account)
{
	if (Account AND Ftp)
	{
		char *Socks5Server = 0;
		char *Socks5UserName = 0;
		char *Socks5Password = 0;

		Options.Get(OPT_Socks5Server, Socks5Server);
		Options.Get(OPT_Socks5UserName, Socks5UserName);
		Options.Get(OPT_Socks5Password, Socks5Password);

		Ftp->Close();
		Connection = 0;
		if (ValidStr(Socks5Server))
		{
			char Buf[256];
			strcpy(Buf, Socks5Server);
			char *Colon = strchr(Buf, ':');
			int Port = 1080;
			if (Colon)
			{
				Port = atoi(Colon+1);
				*Colon = 0;
			}

			Connection = NEW(ISocks5LogConnection(	Buf,
													Port,
													Socks5UserName,
													Socks5Password,
													LogPanel,
													&Sockets,
													this));
		}
		else
		{
			Connection = NEW(ILogConnection(LogPanel, &Sockets, this));
		}

		char *RemoteHost = 0;
		int Port = 0;
		char *UserName = 0;
		char *Password = 0;
		char *RemoteDir = 0;
		char *LocalDir = 0;
		int Anonymous = 0;
		int UseActive = 0;
		char CmdStr[256];

		Account->Get(OPT_RemoteHost, RemoteHost);
		Account->Get(OPT_Port, Port);
		Account->Get(OPT_UserName, UserName);
		Account->Get(OPT_Password, Password);
		Account->Get(OPT_Anonymous, Anonymous);
		Account->Get(OPT_LocalDir, LocalDir);
		Account->Get(OPT_RemoteDir, RemoteDir);
		Account->Get(OPT_UseActive, UseActive);

		// Set local dir
		if (ValidStr(LocalDir) AND Local)
		{
			FtpCmd *Cmd = NEW(FtpCmd(LocalDir));
			if (Cmd)
			{
				Cmd->Cmd = TASK_Cd;
				Cmd->View = Local;
				Local->PostCmd(Cmd);
			}
		}

		// Loging to remote server
		LogPanel->Print(FTP_COL_INFO, "Connecting to '%s'...", RemoteHost);
		if (Anonymous)
		{
			sprintf(CmdStr, "%s,%i,%s,%s,%s,%i", RemoteHost, Port, "anonymous", "user@localhost.com", (ValidStr(RemoteDir))?RemoteDir:" ", UseActive);
		}
		else
		{
			sprintf(CmdStr, "%s,%i,%s,%s,%s,%i", RemoteHost, Port, UserName, Password, (ValidStr(RemoteDir))?RemoteDir:" ", UseActive);
		}

		// Enabled abort command
		Connecting = true;

		// Post login command
		FtpCmd *Cmd = NEW(FtpCmd(CmdStr));
		if (Cmd)
		{
			Cmd->Cmd = TASK_Login;
			Cmd->View = Remote;
			Cmd->Connection = Connection;
			Local->PostCmd(Cmd);
		}
	}
}

int AppWnd::OnCommand(int Cmd, int Event, OsView Handle)
{
	if (Cmd >= IDM_ACCOUNT_BASE AND Cmd < Accounts.GetItems()+IDM_ACCOUNT_BASE)
	{
		ConnectTo(Accounts.ItemAt(Cmd - IDM_ACCOUNT_BASE));
	}

	switch (Cmd)
	{
		case IDM_HTTP:
		{
			Http = NEW(HttpThread(this));
			break;
		}
		case IDM_CONNECT:
		{
			{
				ConnectDlg Dlg(this, &Accounts);
				ConnectTo(Dlg.Account);
			}
			UpdateMenus();
			break;
		}
		case IDM_DISCONNECT:
		{
			if (Ftp)
			{
				ResetThread();
				Remote->Empty();
				Ftp->Close();
				EnableCommands(false);
			}
			break;
		}

		case IDM_ABORT:
		{
			if (Connecting)
			{
				ResetThread();

				Remote->Empty();
				Ftp->Close();
				EnableCommands(false);
				Connecting = false;

				LogPanel->Print(FTP_COL_INFO, "Connect canceled.");
			}
			else if (Progress)
			{
				Progress->Cancel(true);
			}
			break;
		}
		case IDM_EXIT:
		{
			LgiCloseApp();
			break;
		}
		case IDM_HELP:
		{
			char Path[256];

			if (LgiGetExePath(Path))
			{
				#ifdef _DEBUG
				LgiTrimDir(Path);
				strcat(Path, DIR_STR);
				strcat(Path, "Code");
				strcat(Path, DIR_STR);
				#endif
				if (Path[strlen(Path)-1] != DIR_CHAR) strcat(Path, DIR_STR);
				sprintf(Path+strlen(Path), "help%sindex.html", DIR_STR);

				if (NOT LgiExecute(Path, NULL, NULL))
				{
					LgiMsg(this, "Couldn't find the help file '%s'", AppName, MB_OK, Path);
				}
			}
			break;
		}
		case IDM_ABOUT:
		{
			GAbout Dlg(this, AppName, VER, "Graphical FTP client.", "_about.gif");
			break;
		}
	}

	return 0;
}

int AppWnd::OnNotify(GView *Ctrl, int Flags)
{
	switch (Ctrl->GetId())
	{
	}
	return 0;
}

void AppWnd::WriteProperties(GFile &F, ObjProperties *Props, char *KeyName)
{
	if (Props AND KeyName)
	{
		char *Str = NEW(char[16<<10]);
		if (Str)
		{
			*Str = 0;
			sprintf(Str, "<%s", KeyName);
			for (bool b = Props->FirstKey(); b; b = Props->NextKey())
			{
				switch (Props->KeyType())
				{
					case OBJ_INT:
					{
						int n;
						if (Props->Get(Props->KeyName(), n))
						{
							sprintf(Str+strlen(Str), "\r\n\t%s=%i", Props->KeyName(), n);
						}
						break;
					}
					case OBJ_STRING:
					{
						char *s;
						if (Props->Get(Props->KeyName(), s) AND ValidStr(s))
						{
							sprintf(Str+strlen(Str), "\r\n\t%s='%s'", Props->KeyName(), s);
						}
						break;
					}
				}
			}

			sprintf(Str+strlen(Str), ">\r\n");
			F.Write(Str, strlen(Str));
			DeleteArray(Str);
		}
	}
}

void AppWnd::ReadProperties(XmlTag *t, ObjProperties *a)
{
	if (t AND a)
	{
		for (XmlValue *v = t->Values.First(); v; v = t->Values.Next())
		{
			bool IsNumber = true;
			for (char *s = v->Value; *s AND IsNumber; s++)
			{
				IsNumber &= (*s >= '0' AND *s <= '9');
			}

			if (IsNumber)
			{
				a->Set(v->Name, atoi(v->Value));
			}
			else
			{
				a->Set(v->Name, v->Value);
			}
		}
	}
}

void AppWnd::Serialize(bool Write)
{
	char *AccountsKeyName = "account";
	char *OptionsKeyName = "options";

	if (NOT OptionsFile)
	{
		char File[256];
		if (LgiApp->GetOption(File, 'o'))
		{
			if (FileExists(File))
			{
				OptionsFile = NewStr(File);
			}
			else
			{
				char *Temp = LgiFindFile(File);
				if (Temp)
				{
					OptionsFile = Temp;
				}
				else
				{
					OptionsFile = NewStr(File);
				}
			}
		}

		if (NOT OptionsFile)
		{
			OptionsFile = LgiFindFile(OptionsFileName);
		}

		if (NOT OptionsFile)
		{
			char Temp[256];
			if (LgiGetExePath(Temp))
			{
				#ifdef _DEBUG
				LgiTrimDir(Temp);
				#endif
				strcat(Temp, DIR_STR);
				strcat(Temp, OptionsFileName);
				OptionsFile = NewStr(Temp);
			}
		}
	}

	GFile F;
	if (OptionsFile AND
		F.Open(OptionsFile, (Write)?O_WRITE:O_READ))
	{
		if (Write)
		{
			F.SetSize(0);

			// Save window position
			#ifdef WIN32
			if (Handle() AND NOT IsIconic(Handle()) AND NOT IsZoomed(Handle()))
			#endif
			{
				GRegion r = GetPos();
				Options.Set("x", r.x1);
				Options.Set("y", r.y1);
				Options.Set("cx", r.X()-1);
				Options.Set("cy", r.Y()-1);
			}

			WriteProperties(F, &Options, OptionsKeyName);
			for (ObjProperties *a = Accounts.First(); a; a = Accounts.Next())
			{
				WriteProperties(F, a, AccountsKeyName);
			}
		}
		else
		{
			Xml X;
			if (X.ParseXmlFile(F) > 0)
			{
				for (XmlTag *t = X.Values.First(); t; t = X.Values.Next())
				{
					if (t->Name)
					{
						if (stricmp(t->Name, AccountsKeyName) == 0)
						{
							ObjProperties *a = NEW(ObjProperties);
							if (a)
							{
								ReadProperties(t, a);
								Accounts.Insert(a);
							}
						}
						else if (stricmp(t->Name, OptionsKeyName) == 0)
						{
							ReadProperties(t, &Options);
						}
					}
				}
			}

			GRegion r;
			if (Options.Get("x", r.x1) AND
				Options.Get("y", r.y1) AND
				Options.Get("cx", r.x2) AND
				Options.Get("cy", r.y2))
			{
				r.x2 += r.x1;
				r.y2 += r.y1;
				SetPos(r);
			}
		}
	}
}

void AppWnd::SetTransfer(bool On)
{
	if (Progress)
	{
		Progress->Cancel(false);
	}

	CmdAbort.Enabled(On);
}

/*
void AppWnd::SetStatusText(char *Text, int Pane)
{
	if (Pane >= 0 AND Pane < STATUS_MAX AND StatusInfo[Pane])
	{
		StatusInfo[Pane]->Name(Text);
	}
}

void AppWnd::OnReceiveFiles(List<char> &Files)
{
	if (Files.GetItems() > 0)
	{
		for (char *FileName = Files.First(); FileName; FileName = Files.Next())
		{
		}
	}
}

*/

//////////////////////////////////////////////////////////////////////
void FtpList::OnColumnClick(int Col, GMouse &m)
{
	if (View)
	{
		View->OnColumnClick(Col, m);
	}
}

GFileSystemView::GFileSystemView(AppWnd *wnd, GImageList *ImageList, IFtp *ftp)
{
	Wnd = wnd;
	Opposite = 0;
	Ftp = ftp;
	Worker = 0;
	Listing = true;
	GRegion r(0, 0, 100, 120);
	SetPos(r);
	Lst = 0;
	Dir = 0;
	SortCol = -1;
	SortAscend = true;
	Mount = 0;
	#ifdef WIN32
	WndFlags |= WS_CLIPCHILDREN;
	#endif

	Children.Insert(Edit = NEW(GEdit(GFSV_DIR, 0, 0, 100, 20, "")));
	Children.Insert(Lst = NEW(FtpList(GFSV_LIST, 0, 20, 100, 100)));
	if (Lst)
	{
		Lst->View = this;
		Lst->Sunken(false);
		Lst->AddColumn("Name", 140);
		Lst->AddColumn("Type", 40);
		Lst->AddColumn("Size", 70);
		Lst->AddColumn("Date", 70);
		Lst->AddColumn("Time", 70);
		Lst->SetImageList(ImageList, false);
	}

	Commands = LgiLoadToolbar(this, "_pane.gif");
	if (Commands)
	{
		Children.Insert(Commands);

		CmdCreateDir.ToolButton = Commands->AppendButton("Create Directory", IDM_CREATE_DIR, TBT_PUSH);
		CmdDelete.ToolButton =  Commands->AppendButton("Delete", IDM_DELETE, TBT_PUSH);
		CmdRenameFile.ToolButton =  Commands->AppendButton("Rename File", IDM_RENAME, TBT_PUSH);
		CmdSetPerms.ToolButton = Commands->AppendButton("Set Permissions", IDM_SETPERMS, TBT_PUSH);
		CmdExecute.ToolButton = Commands->AppendButton("Execute File", IDM_EXECUTE, TBT_PUSH);
		CmdViewText.ToolButton = Commands->AppendButton("View Text", IDM_VIEW_TEXT, TBT_PUSH);

		Commands->AppendSeparator();

		CmdTransfer.ToolButton = Commands->AppendButton("Transfer", IDM_UPLOAD, TBT_PUSH);
		Commands->AppendSeparator();

		CmdRefresh.ToolButton = Commands->AppendButton("List", IDM_LIST, TBT_PUSH);
		CmdCdup.ToolButton = Commands->AppendButton("Up to parent directory", IDM_UP, TBT_PUSH);
	}
}

GFileSystemView::~GFileSystemView()
{
}

int FtpItemCompare(GListItem *a, GListItem *b, int Data)
{
	DirListItem *A = (DirListItem*) a->_UserData;
	DirListItem *B = (DirListItem*) b->_UserData;
	GFileSystemView *View = (GFileSystemView*)Data;
	int Col = View->SortCol;
	int Status = 0;

	if (A AND B)
	{
		NameCompare:
		switch (Col)
		{
			case 0:
			case 1:
			{
				char *Sa = A->GetText(Col);
				char *Sb = B->GetText(Col);
				if (Sa AND Sb)
				{
					Status = stricmp(Sa, Sb);
				}
				break;
			}
			case 2:
			{
				Status = A->Entry->Size - B->Entry->Size;
				break;
			}
			case 3:
			{
				// return A->Entry->Date.Compare(&B->Entry->Date);
				break;
			}
		}

		if (Status == 0 AND Col > 0)
		{
			Col = 0;
			goto NameCompare;
		}
	}

	return View->SortAscend ? Status : -Status;
}

void GFileSystemView::OnColumnClick(int Col, GMouse &m)
{
	if (Col <= 2)
	{
		if (Col == SortCol)
		{
			SortAscend = NOT SortAscend;
		}
		else
		{
			SortCol = Col;
			SortAscend = true;
		}

		Lst->Sort(FtpItemCompare, (int)this);

		for (int i=0; i<Lst->GetColumns(); i++)
		{
			if (i == SortCol)
			{
				Lst->ColumnAt(i)->Mark(SortAscend ? GLI_MARK_DOWN_ARROW : GLI_MARK_UP_ARROW);
			}
			else
			{
				Lst->ColumnAt(i)->Mark(GLI_MARK_NONE);
			}
		}
	}
}

void GFileSystemView::PostCmd(FtpCmd *Cmd, bool ForceRemote)
{
	if (Cmd)
	{
		if (NOT ForceRemote AND
			Cmd->View->IsCmdLocal(Cmd->Cmd))
		{
			DoCmd(Cmd);
		}
		else
		{
			Worker->PostCmd(Cmd);
		}
	}
}

bool GFileSystemView::Attach(GView *p)
{
	bool Status = GView::Attach(p);
	if (Status)
	{
		AttachChildren();
	}
	return Status;
}

bool GFileSystemView::Pour(GComplexRegion &r)
{
	r.Offset(-Pos.x1, -Pos.y1);

	if (Commands)
	{
		Commands->Pour(r);
		r.Subtract(&Commands->GetPos());
	}

	GRegion *Rgn = FindLargest(r);
	if (Rgn AND Edit)
	{
		GRegion e(Rgn->x1, Rgn->y1, Rgn->x2, Rgn->y1 + 19);

		if (Mount)
		{
			GRegion m = e;
			e.x1 += Mount->X();
			m.x2 = e.x1 - 1;
			Mount->SetPos(m);
			r.Subtract(&m);
		}

		if (Dir)
		{
			GRegion d = e;
			e.x2 -= Dir->X();
			d.x1 = e.x2 + 1;
			Dir->SetPos(d);
			r.Subtract(&d);
		}

		Edit->SetPos(e);
		r.Subtract(&e);
	}

	Rgn = FindLargest(r);
	if (Rgn AND Lst)
	{
		Lst->SetPos(*Rgn);
		r.Subtract(Rgn);
	}

	return true;
}

void GFileSystemView::OnPaint(GSurface *pDC)
{
	GRegion a = GetPos();
	a.Offset(-a.x1, -a.y1);
	GComplexRegion r(a);

	for (GView *c = Children.First(); c; c = Children.Next())
	{
		r.Subtract(&c->GetPos());
	}

	pDC->Colour(LC_MED, 24);
	for (GRegion *b = r.First(); b; b = r.Next())
	{
		pDC->Rectangle(b);
	}
}

void GFileSystemView::GetMenuCommands(IFtpEntry *e, GSubMenu *RClick)
{
	if (RClick)
	{
		RClick->AppendItem("Change Dir", IDM_CD, e->IsDir() OR TestFlag(e->Attributes, IFTP_SYM_LINK));
		RClick->AppendItem("Transfer", IDM_TRANSFER, true);
		RClick->AppendItem("Transfer with mask", IDM_TRANSFER_MASK, true);
	}
}

void GFileSystemView::OnMenuCommand(IFtpEntry *e, int Cmd)
{
	char *Mask = 0;
	
	switch (Cmd)
	{
		case IDM_CD:
		{
			FtpCmd *Cmd = NEW(FtpCmd(e->Name));
			if (Cmd)
			{
				Cmd->Cmd = TASK_Cd;
				Cmd->Entry = NEW(IFtpEntry(e));
				Cmd->View = this;
				PostCmd(Cmd);
			}
			break;
		}
		case IDM_TRANSFER_MASK:
		{
			GInput Dlg(this, "", "Enter file masks (separate with space)", AppName);
			if (Dlg.DoModal() == IDOK)
			{
				Mask = NewStr(Dlg.Str);
			}
			else
			{
				break;
			}
			// fall thru
		}
		case IDM_TRANSFER:
		{
			List<IFtpEntry> Sel;
			if (GetSelection(Sel) > 0)
			{
				FtpCmd *Cmd = 0;

				for (IFtpEntry *e = Sel.First(); e; e = Sel.Next())
				{
					Cmd = NEW(FtpCmd(Mask));
					if (Cmd)
					{
						Cmd->Cmd = TASK_Transfer;
						Cmd->Entry = NEW(IFtpEntry(e));
						Cmd->View = this;
						PostCmd(Cmd);
					}
				}
				
				Cmd = NEW(FtpCmd);
				if (Cmd)
				{
					Cmd->Cmd = TASK_ListElements;
					Cmd->View = Opposite;
					PostCmd(Cmd, NOT IsCmdLocal(TASK_ListElements) );
				}
			}
			break;
		}
	}
}

void GFileSystemView::OnClick(IFtpEntry *e, GMouse *m)
{
	if (e AND m->Right())
	{
		GSubMenu *RClick = NEW(GSubMenu);
		if (RClick)
		{
			GetMenuCommands(e, RClick);

			if (GetMouse(*m, true))
			{
				int Cmd = RClick->Float(this, m->x, m->y, false);
				if (Cmd)
				{
					OnMenuCommand(e, Cmd);
				}
			}
		}
	}

	if (e AND m AND Worker)
	{
		FtpCmd *Cmd = NEW(FtpCmd);
		if (Cmd)
		{
			Cmd->Cmd = TASK_ProcessClick;
			Cmd->Entry = NEW(IFtpEntry(e));
			Cmd->Click = *m;
			Cmd->View = this;
			PostCmd(Cmd);
		}
	}
}

int GFileSystemView::OnNotify(GView *Ctrl, int Flags)
{
	if (Worker)
	{
		switch (Ctrl->GetId())
		{
			case IDC_DIR:
			{
				if (Edit)
				{
					LgiExecute(Edit->Name(), "", "");
				}
				break;
			}
			case IDC_MOUNT:
			{
				if (Mount AND Edit)
				{
					char *Drive = Mount->Name();
					if (Mount->Index() > 0)
					{
						FtpCmd *Cmd = NEW(FtpCmd(Drive));
						if (Cmd)
						{
							Cmd->Cmd = TASK_Cd;
							Cmd->View = this;
							PostCmd(Cmd);
						}
						Mount->Index(0);
					}
				}
				break;
			}
			case IDM_LIST:
			{
				FtpCmd *Cmd = NEW(FtpCmd);
				if (Cmd)
				{
					Cmd->Cmd = TASK_ListElements;
					Cmd->View = this;
					PostCmd(Cmd);
				}
				break;
			}
			case GFSV_DIR:
			{
				if (Edit AND Flags == 13)
				{
					FtpCmd *Cmd = NEW(FtpCmd(Edit->Name()));
					if (Cmd)
					{
						Cmd->Cmd = TASK_Cd;
						Cmd->View = this;
						PostCmd(Cmd);
					}
				}
				break;
			}
			case IDM_UPLOAD:
			{
				List<IFtpEntry> Sel;
				if (GetSelection(Sel) > 0)
				{
					for (IFtpEntry *e = Sel.First(); e; e = Sel.Next())
					{
						FtpCmd *Cmd = NEW(FtpCmd);
						if (Cmd)
						{
							Cmd->Cmd = TASK_Transfer;
							Cmd->Entry = NEW(IFtpEntry(e));
							Cmd->View = this;
							PostCmd(Cmd);
						}
					}

					FtpCmd *Cmd = NEW(FtpCmd);
					if (Cmd)
					{
						Cmd->Cmd = TASK_ListElements;
						Cmd->View = Opposite;
						PostCmd(Cmd, NOT IsCmdLocal(TASK_ListElements) );
					}
				}
				break;
			}
			case IDM_UP:
			{
				FtpCmd *Cmd = NEW(FtpCmd);
				if (Cmd)
				{
					Cmd->Cmd = TASK_Cd;
					Cmd->View = this;
					PostCmd(Cmd);
				}
				break;
			}
			case IDM_CREATE_DIR:
			{
				GInput Inp(this, "", "Enter the new directory name:", AppName);
				if (Inp.DoModal())
				{
					FtpCmd *Cmd = NEW(FtpCmd(Inp.Str));
					if (Cmd)
					{
						Cmd->Cmd = TASK_CreateDir;
						Cmd->View = this;
						PostCmd(Cmd);
					}
				}
				break;
			}
			case IDM_RENAME:
			{
				List<IFtpEntry> Sel;
				if (GetSelection(Sel) > 0 AND
					Sel.GetItems() == 1)
				{
					IFtpEntry *e = Sel.First();
					if (e)
					{
						GInput Inp(this, e->Name, "Enter the new file name:", AppName);
						if (Inp.DoModal())
						{
							FtpCmd *Cmd = NEW(FtpCmd);
							if (Cmd)
							{
								Cmd->Cmd = TASK_Rename;
								Cmd->View = this;
								strcpy(Cmd->Arg, e->Name); // from
								strcpy(Cmd->Arg2, Inp.Str); // to
								PostCmd(Cmd);
							}
						}
					}
				}
				break;
			}
			case IDM_DELETE:
			{
				List<IFtpEntry> Sel;
				if (GetSelection(Sel) > 0)
				{
					for (IFtpEntry *e = Sel.First(); e; e = Sel.Next())
					{
						FtpCmd *Cmd = NEW(FtpCmd);
						if (Cmd)
						{
							Cmd->Cmd = TASK_Delete;
							Cmd->View = this;
							Cmd->Entry = NEW(IFtpEntry(e));
							PostCmd(Cmd);
						}
					}

					FtpCmd *Cmd = NEW(FtpCmd);
					if (Cmd)
					{
						Cmd->Cmd = TASK_ListElements;
						Cmd->View = this;
						PostCmd(Cmd, NOT IsCmdLocal(TASK_ListElements));
					}
				}
				break;
			}
			case IDM_EXECUTE:
			{
				List<IFtpEntry> Sel;
				if (GetSelection(Sel) > 0)
				{
					for (IFtpEntry *e = Sel.First(); e; e = Sel.Next())
					{
						FtpCmd *Cmd = NEW(FtpCmd(e->Name));
						if (Cmd)
						{
							Cmd->Cmd = TASK_ExecuteFile;
							Cmd->View = this;
							PostCmd(Cmd);
						}
					}
				}
				break;
			}
			case IDM_VIEW_TEXT:
			{
				List<IFtpEntry> Sel;
				if (GetSelection(Sel) > 0)
				{
					for (IFtpEntry *e = Sel.First(); e; e = Sel.Next())
					{
						FtpCmd *Cmd = NEW(FtpCmd(e->Name));
						if (Cmd)
						{
							Cmd->Cmd = TASK_ViewTextFile;
							Cmd->View = this;
							PostCmd(Cmd);
						}
					}
				}
				break;
			}
			case IDM_SETPERMS:
			{
				List<IFtpEntry> Sel;
				if (GetSelection(Sel) > 0)
				{
					IFtpEntry *e = Sel.First();
					if (e)
					{
						#ifdef WIN32
						PermsDlg Dlg(this, IsCmdLocal(IDM_SETPERMS), e->Attributes);
						#else
						PermsDlg Dlg(this, false, e->Attributes);
						#endif
						if (Dlg.DoModal() == IDOK)
						{
							FtpCmd *Cmd = NEW(FtpCmd);
							if (Cmd)
							{
								Cmd->Cmd = TASK_SetPerms;
								Cmd->View = this;
								Cmd->Entry = NEW(IFtpEntry(e));
								Cmd->Int = Dlg.Perms;
								PostCmd(Cmd);
							}
						}
					}
				}
				break;
			}
		}
	}

	return 0;
}

int GFileSystemView::GetSelection(List<IFtpEntry> &Sel)
{
	int Status = 0;
	if (Lst)
	{
		List<GListItem> Ls;
		if (Lst AND Lst->GetSelection(Ls))
		{
			for (DirListItem *i = dynamic_cast<DirListItem*>(Ls.First());
				i;
				i = dynamic_cast<DirListItem*>(Ls.Next()))
			{
				if (i->GetEntry())
				{
					Sel.Insert(i->GetEntry());
					Status++;
				}
			}
		}
	}
	return Status;
}

void GFileSystemView::GetDir(char *Path, char *Append)
{
	char *Dir = (Edit) ? Edit->Name() : 0;
	if (Path AND Dir)
	{
		strcpy(Path, Dir);
		if (Append)
		{
			if (Path[strlen(Path)-1] != DirChar())
			{
				char *e = Path+strlen(Path);
				e[0] = DirChar();
				e[1] = 0;
			}
			strcat(Path, Append);
		}
	}
}

void GFileSystemView::DoCmd(FtpCmd *Cmd)
{
	if (Cmd AND Cmd->View)
	{
		switch (Cmd->Cmd)
		{
			case TASK_ListElements:
			{
				Cmd->View->ListElements();
				break;
			}
			case TASK_Cd:
			{
				Cmd->View->Cd(Cmd->Str);
				break;
			}
			case TASK_CreateDir:
			{
				Cmd->View->CreateDir(Cmd->Str);
				break;
			}
			case TASK_Delete:
			{
				if (Cmd->Entry->IsDir())
				{
					Cmd->View->DeleteDir(Cmd->Entry);
				}
				else
				{
					Cmd->View->Delete(Cmd->Entry);
				}
				break;
			}
			case TASK_ExecuteFile:
			{
				Cmd->View->ExecuteFile(Cmd->Str);
				break;
			}
			case TASK_ViewTextFile:
			{
				Cmd->View->ViewTextFile(Cmd->Str);
				break;
			}
			case TASK_ProcessClick:
			{
				Cmd->View->ProcessClick(Cmd->Entry, &Cmd->Click);
				break;
			}
			case TASK_Login:
			{
				TokenParser T(Cmd->Str, ",");
				if (Ftp AND
					T.Tokens >= 4)
				{
					if (T.Get(5))
					{
						Ftp->IsForceActive(atoi(T.Get(5)));
					}

					#define CONNECT_RETIES 6
					bool Open = false;
					for (int i=0; NOT Open AND i<CONNECT_RETIES AND Wnd->Connecting; i++)
					{
						if (i>0)
						{
							Wnd->LogPanel->Print(FTP_COL_INFO, "Retry(%i/%i): Connect...", i, CONNECT_RETIES-1);
						}

						Wnd->CmdAbort.Enabled(true);
						Wnd->CmdConnect.Enabled(false);
						Open = Ftp->Open(Cmd->Connection, T.Get(0), atoi(T.Get(1)), T.Get(2), T.Get(3));
					}

					Wnd->Connecting = false;

					if (Open)
					{
						char *RemoteDir = T.Get(4);
						if (ValidStr(RemoteDir))
						{
							Ftp->SetDir(RemoteDir);
						}

						ListElements();
						Wnd->EnableCommands(true);
					}
				}
				break;
			}
			case TASK_Transfer:
			{
				Cmd->View->TransferFile(Cmd);
				break;
			}
			case TASK_Rename:
			{
				Cmd->View->RenameFile(Cmd->Arg, Cmd->Arg2);
				break;
			}
			case TASK_SetPerms:
			{
				Cmd->View->SetPerms(Cmd->Entry, Cmd->Int);
				break;
			}
			case TASK_Noop:
			{
				Ftp->Noop();
				break;
			}
		}

		if (Wnd)
		{
			Wnd->OnCmd(Cmd);
		}
	}
}

IFtpEntry *GFileSystemView::HasDir(char *Name)
{
	if (Lst AND Name)
	{
		for (DirListItem *i = dynamic_cast<DirListItem*>(Lst->First());
			i; i = dynamic_cast<DirListItem*>(Lst->Next()))
		{
			if (i->GetEntry() AND
				i->GetEntry()->Name AND
				i->GetEntry()->IsDir() AND
				stricmp(i->GetEntry()->Name, Name) == 0)
			{
				return i->GetEntry();
			}
		}
	}

	return 0;
}

bool GFileSystemView::TransferDir(IFtpEntry *e, char *Mask)
{
	bool Status = false;
	if (Lst AND
		e AND
		e->IsDir())
	{
		IFtpEntry *d = Opposite->HasDir(e->Name);
		if (d)
		{
			// change to
			Status = Opposite->Cd(d->Name);
		}
		else
		{
			// create
			if (Status = Opposite->CreateDir(e->Name))
			{
				Status = Opposite->Cd(e->Name);
			}
		}

		if (Status = Cd(e->Name))
		{
			TokenParser T(Mask, " ");

			// For all items in directory
			List<IFtpEntry> Local;
			{
				// copy all the items in the current directory into a temp
				// list, so that as we traverse the tree we don't lose items
				// in the list.
				DirListItem *i;
				for (i = dynamic_cast<DirListItem*>(Lst->First()); i;
					 i = dynamic_cast<DirListItem*>(Lst->Next()))
				{
					if (i->GetEntry())
					{
						Local.Insert(NEW(IFtpEntry(i->GetEntry())));
					}
				}
			}

			{
				// iterate the saved dir listing and process each item in turn
				for (IFtpEntry *i = Local.First(); i; i = Local.First())
				{
					if (i->IsDir())
					{
						// Dir
						Status &= TransferDir(i, Mask);
					}
					else
					{
						// File
						bool Transfer = Mask == 0;
						if (Mask)
						{
							for (int n=0; n<T.Tokens; n++)
							{
								if (Match(i->Name, T.Get(n)))
								{
									Transfer = true;
									break;
								}
							}
						}

						if (Transfer)
						{
							char Path[256];
							if (IsCmdLocal(TASK_ListElements))
							{
								// upload
								GetDir(Path, i->Name);
								Status &= Ftp->UploadFile(Path, i->Name);

								// destination is remote, we can't waste time/bandwidth
								// updating the UI as we go...
							}
							else
							{
								// download
								Opposite->GetDir(Path, i->Name);
								Status &= Ftp->DownloadFile(Path, i);

								// we can list the local contents as we go as the command
								// is local
								Opposite->ListElements();
							}
						}
					}

					Local.Delete(i);
					DeleteObj(i);
				}
			}

			Status &= Cd();
			Status &= Opposite->Cd();
		}
	}
	return Status;
}

bool GFileSystemView::DeleteDir(IFtpEntry *e)
{
	bool Status = false;
	if (Lst AND
		e AND
		e->IsDir())
	{
		// Save all current items...
		List<DirListItem> Init;
		DirListItem *i;
		for (i = dynamic_cast<DirListItem*>(Lst->First());
			i; i = dynamic_cast<DirListItem*>(Lst->Next()))
		{
			Init.Insert(i);
		}
		Lst->RemoveAll();

		// Change to dir...
		if (Status = Cd(e->Name))
		{
			// For all items in directory
			List<DirListItem> Local;
			for (i = dynamic_cast<DirListItem*>(Lst->First());
				i; i = dynamic_cast<DirListItem*>(Lst->Next()))
			{
				Local.Insert(i);
			}
			Lst->RemoveAll();

			for (i = Local.First();
				i; i = Local.First())
			{
				if (i->GetEntry())
				{
					if (i->GetEntry()->IsDir())
					{
						// Dir
						Status &= DeleteDir(i->GetEntry());
					}
					else
					{
						// File
						Status &= Delete(i->GetEntry());
					}
				}

				Local.Delete(i);
				DeleteObj(i);
			}

			Status &= Cd();
			Status &= Delete(e);
		}

		// ReLoad all current items.
		for (i=Init.First(); i; i=Init.Next())
		{
			if (i->GetEntry() == e)
			{
				DeleteObj(i);
			}
			else
			{
				Lst->Insert(i);
			}
		}
	}
	return Status;
}

///////////////////////////////////////////////////////////////////////
GLocalFS::GLocalFS(AppWnd *Wnd, GImageList *ImageList, IFtp *ftp) :
	GFileSystemView(Wnd, ImageList, ftp)
{
	Children.Insert(Dir = NEW(GButton(IDC_DIR, 0, 0, 44, 20, "Explore")));
	Children.Insert(Mount = NEW(GCombo(IDC_MOUNT, 0, 0, 50, 20, "")));
	if (Mount)
	{
		Mount->Insert("Drive:");
		for (int i=0; i<FileDev->GetNumDrives(); i++)
		{
			DriveInfo *d = FileDev->GetDriveInfo(i);
			if (d)
			{
				Mount->Insert(d->Name);
			}
		}
	}
}

void GLocalFS::EnableCommands(bool Enabled)
{
	CmdTransfer.Enabled(Enabled);
}

bool GLocalFS::IsCmdLocal(int Cmd)
{
	return (Cmd != TASK_Transfer);
}

bool GLocalFS::Insert(IFtpEntry *p, int Index)
{
	if (Lst)
	{
		Lst->Insert(NEW(DirListItem(p, this)), (p->IsDir()) ? 0 : -1);
	}

	return true;
}

void GLocalFS::ListElements()
{
	if (Listing)
	{
		if (Lst)
		{
			Lst->Empty();
		}

		char DirName[256];
		if (FileDev->GetCurrentDirectory(DirName, sizeof(DirName)))
		{
			if (Edit)
			{
				Edit->Name(DirName);
			}

			if (DirName[strlen(DirName)-1] == DIR_CHAR)
			{
				DirName[strlen(DirName)-1] = 0;
			}
			sprintf(DirName+strlen(DirName), "%c*.*", DIR_CHAR);

			VDirectory *Dir = FileDev->GetDir(NULL);
			if (Dir)
			{
				for (int Status = Dir->First(DirName, 0, 0); Status; Status = Dir->Next())
				{
					if (stricmp(Dir->GetName(), ".") == 0 OR
						stricmp(Dir->GetName(), "..") == 0)
						continue;

					IFtpEntry *e = NEW(IFtpEntry);
					if (e)
					{
						e->Name = NewStr(Dir->GetName());
						e->Size = Dir->GetSize();
						
						if (Dir->IsDir()) e->Attributes |= IFTP_DIR;
						#ifdef WIN32
						if (Dir->GetAttributes() & FA_READONLY) e->Attributes |= IFTP_READ;
						if (Dir->GetAttributes() & FA_ARCHIVE) e->Attributes |= IFTP_ARCHIVE;
						if (Dir->GetAttributes() & FA_HIDDEN) e->Attributes |= IFTP_HIDDEN;
						if (Dir->GetAttributes() & FA_SYSTEM) e->Attributes |= IFTP_SYSTEM;
						#else
						if (Dir->GetAttributes() & S_IRUSR) e->Attributes |= IFTP_READ;
						if (Dir->GetAttributes() & S_IWUSR) e->Attributes |= IFTP_WRITE;
						if (Dir->GetAttributes() & S_IXUSR) e->Attributes |= IFTP_EXECUTE;

						if (Dir->GetAttributes() & S_IRGRP) e->Attributes |= IFTP_GRP_READ;
						if (Dir->GetAttributes() & S_IWGRP) e->Attributes |= IFTP_GRP_WRITE;
						if (Dir->GetAttributes() & S_IXGRP) e->Attributes |= IFTP_GRP_EXECUTE;

						if (Dir->GetAttributes() & S_IROTH) e->Attributes |= IFTP_GLOB_READ;
						if (Dir->GetAttributes() & S_IWOTH) e->Attributes |= IFTP_GLOB_WRITE;
						if (Dir->GetAttributes() & S_IXOTH) e->Attributes |= IFTP_GLOB_EXECUTE;
						#endif
						
						quad GDateTime = Dir->GetLastWriteTime();
						char Str[256];
						Dir->ConvertToDate(Str, GDateTime);
						e->Date = NewStr(Str);
						Dir->ConvertToTime(Str, GDateTime);
						e->Time = NewStr(Str);
						
						Insert(e);
					}
				}

				DeleteObj(Dir);
			}
		}

		if (Lst)
		{
			Lst->Sort(Sorter, 0);
		}
	}
}

bool GLocalFS::Cd(char *Dir)
{
	bool Status = false;
	char DirName[256];

	if (FileDev->GetCurrentDirectory(DirName, sizeof(DirName)))
	{
		if (Dir)
		{
			if (strchr(Dir, DIR_CHAR))
			{
				strcpy(DirName, Dir);
			}
			else
			{
				if (DirName[strlen(DirName)-1] != DIR_CHAR)
				{
					strcat(DirName, DIR_STR);
				}
				strcat(DirName, Dir);
			}
		}
		else
		{
			for (char *s = DirName + strlen(DirName) - 1;
				s >= DirName;
				s--)
			{
				if (*s == DIR_CHAR)
				{
					if ((int)s-(int)DirName <= 2)
					{
						s[1] = 0;
					}
					else
					{
						*s = 0;
					}
					break;
				}
			}
		}

		Status = FileDev->SetCurrentDirectory(DirName);
		ListElements();
	}

	return Status;
}

bool GLocalFS::CreateDir(char *Dir)
{
	if (Dir)
	{
		if (FileDev->CreateDirectory(Dir))
		{
			ListElements();
			return true;
		}
	}

	return false;
}

bool GLocalFS::Delete(IFtpEntry *e)
{
	if (e)
	{
		if (e->IsDir())
		{
			return FileDev->RemoveDirectory(e->Name);
		}
		else
		{
			char Path[256];
			GetDir(Path, e->Name);
			return FileDev->DeleteFile(Path);
		}
	}

	return false;
}

void GLocalFS::SetPerms(IFtpEntry *e, int Perms)
{
	if (e AND Perms)
	{
		char Path[256];
		GetDir(Path, e->Name);

		#ifdef WIN32
		if (SetFileAttributes(Path,
				((Perms & IFTP_READ) ? FILE_ATTRIBUTE_READONLY : 0) |
				((Perms & IFTP_ARCHIVE) ? FILE_ATTRIBUTE_ARCHIVE : 0) |
				((Perms & IFTP_HIDDEN) ? FILE_ATTRIBUTE_HIDDEN : 0) |
				((Perms & IFTP_SYSTEM) ? FILE_ATTRIBUTE_SYSTEM : 0)))
		{
			ListElements();
		}
		#else if BEOS
		BEntry Entry(Path);
		int p = 0;

		if (Perms & IFTP_READ) p |= S_IRUSR;
		if (Perms & IFTP_WRITE) p |= S_IWUSR;
		if (Perms & IFTP_EXECUTE) p |= S_IXUSR;

		if (Perms & IFTP_GRP_READ) p |= S_IRGRP;
		if (Perms & IFTP_GRP_WRITE) p |= S_IWGRP;
		if (Perms & IFTP_GRP_EXECUTE) p |= S_IXGRP;

		if (Perms & IFTP_GLOB_READ) p |= S_IROTH;
		if (Perms & IFTP_GLOB_WRITE) p |= S_IWOTH;
		if (Perms & IFTP_GLOB_EXECUTE) p |= S_IXOTH;

		if (Entry.SetPermissions(p) == B_OK)
		{
			ListElements();
		}
		#endif
	}
}

void GLocalFS::ViewTextFile(char *File)
{
	char App[256];
	if (File AND
		LgiGetAppForMimeType("text/plain", App))
	{
		bool ArgInserted = false;
		char Path[256];
		GetDir(Path, File);

		for (char *a=strchr(App, '%'); a AND *a; a=strchr(a, '%'))
		{
			if (isdigit(a[1]) OR a[1] == 'L')
			{
				// argument placeholder
				if (NOT ArgInserted)
				{
					int Len = strlen(Path);
					a += 2;
					memmove(a + Len - 2, a, strlen(a) + 1);
					memcpy(a - 2, Path, Len);
					ArgInserted = true;
				}
				break;
			}
			else
			{
				// environment variable
				char *End = strchr(a+1, '%');
				if (End)
				{
					*End++ = 0;
					if (stricmp(a+1, "systemroot") == 0)
					{
						char Temp[256];
						#ifdef WIN32
						GetWindowsDirectory(Temp, sizeof(Temp));
						#else
						strcpy(Temp, "/tmp");
						#endif
						int Len = strlen(Temp);

						if (Temp[strlen(Temp)-1] == DIR_CHAR) Temp[strlen(Temp)-1] = 0;
						strcat(Temp, End);
						strcpy(App, Temp);

						a = App + Len;
					}
				}
			} 
		}

		if (NOT ArgInserted)
		{
			strcat(App, " ");
			strcat(App, Path);
		}

		// find end of program
		char *Arg = strstr(App, ".exe");
		if (Arg)
		{
			Arg += 4;
			if (strchr("\'\"", *Arg))
			{
				Arg++;
			}

			// isolate the arguments...
			*Arg++ = 0;
			LgiExecute(App, Arg, ".");
		}
	}
}

void GLocalFS::ExecuteFile(char *File)
{
	if (File)
	{
		char Path[256];
		GetDir(Path, File);
		LgiExecute(Path, "", ".");
	}
}

bool GLocalFS::TransferFile(FtpCmd *c)
{
	bool Status = false;

	if (c AND Ftp)
	{
		char Path[256];
		if (c->Entry)
		{
			if (c->Entry->IsDir())
			{
				Status = TransferDir(c->Entry, c->Str);
				Path[0] = 0;
			}
			else
			{
				GetDir(Path, c->Entry->Name);
			}
		}
		else
		{
			strcpy(Path, c->Arg);
		}

		char *File = strrchr(Path, DIR_CHAR);
		if (File)
		{
			MainWnd->SetTransfer(true);
			Status = Ftp->UploadFile(Path, File + 1);
			MainWnd->SetTransfer(false);
		}
	}

	return Status;
}

void GLocalFS::ProcessClick(IFtpEntry *e, GMouse *m)
{
	if (e)
	{
		if (m->Left() AND m->Double())
		{
			if (e->IsDir())
			{
				// dir
				Cd(e->Name);
			}
			else
			{
				// file
				FtpCmd *Cmd = NEW(FtpCmd);
				if (Cmd)
				{
					Cmd->Cmd = TASK_Transfer;
					Cmd->Entry = NEW(IFtpEntry(e));
					Cmd->View = this;
					PostCmd(Cmd);
				}

				if (Opposite)
				{
					Cmd = NEW(FtpCmd);
					if (Cmd)
					{
						Cmd->Cmd = TASK_ListElements;
						Cmd->View = Opposite;
						PostCmd(Cmd, NOT IsCmdLocal(TASK_ListElements));
					}
				}
			}
		}
	}
}

bool GLocalFS::RenameFile(char *From, char *To)
{
	if (From AND To)
	{
		if (FileDev->MoveFile(From, To))
		{
			ListElements();
			return true;
		}
	}

	return false;
}

///////////////////////////////////////////////////////////////////////
GRemoteFS::GRemoteFS(AppWnd *Wnd, GImageList *ImageList, IFtp *ftp) :
	GFileSystemView(Wnd, ImageList, ftp)
{
}

void GRemoteFS::EnableCommands(bool Enabled)
{
	CmdCreateDir.Enabled(Enabled);
	CmdDelete.Enabled(Enabled);
	CmdRenameFile.Enabled(Enabled);
	CmdSetPerms.Enabled(Enabled);
	CmdExecute.Enabled(false);
	CmdViewText.Enabled(false);
	CmdTransfer.Enabled(Enabled);
	CmdRefresh.Enabled(Enabled);
	CmdCdup.Enabled(Enabled);
}

bool GRemoteFS::IsCmdLocal(int Cmd)
{
	return Cmd == TASK_ProcessClick;
}

void GRemoteFS::OnItemSelect(GListItem *Item)
{
	if (Ftp AND Ftp->IsOpen())
	{
		
	}
}

bool GRemoteFS::Insert(IFtpEntry *p, int Index)
{
	if (Lst)
	{
		Lst->Insert(NEW(DirListItem(p, this)), (p->IsDir()) ? 0 : -1);
	}

	return true;
}

void GRemoteFS::ListElements()
{
	if (Listing)
	{
		if (Lst)
		{
			Lst->Empty();
		}

		if (Ftp)
		{
			char DirName[256];
			if (Ftp->GetDir(DirName))
			{
				if (Edit)
				{
					Edit->Name(DirName);
				}

				Ftp->ListDir(this);
			}
		}

		if (Lst)
		{
			Lst->Sort(Sorter, 0);
		}
	}
}

bool GRemoteFS::Cd(char *Dir)
{
	bool Status = false;
	if (Ftp)
	{
		if (ValidStr(Dir))
		{
			Status = Ftp->SetDir(Dir);
		}
		else
		{
			Status = Ftp->UpDir();
		}

		ListElements();
	}

	return Status;
}

bool GRemoteFS::CreateDir(char *Dir)
{
	if (Ftp AND Dir)
	{
		if (Ftp->CreateDir(Dir))
		{
			ListElements();
			return true;
		}
	}

	return false;
}

bool GRemoteFS::Delete(IFtpEntry *e)
{
	if (Ftp AND e)
	{
		if (e->IsDir())
		{
			return Ftp->DeleteDir(e->Name);
		}
		else
		{
			return Ftp->DeleteFile(e->Name);
		}
	}

	return false;
}

void GRemoteFS::SetPerms(IFtpEntry *e, int Perms)
{
	if (Ftp AND e AND e->Name AND NOT e->IsDir())
	{
		if (Ftp->SetPerms(e->Name, Perms))
		{
			ListElements();
		}
	}
}

bool GRemoteFS::TransferFile(FtpCmd *c)
{
	bool Status = false;

	// download file
	if (c AND Opposite)
	{
		char Path[256];
		if (c->Entry)
		{
			if (c->Entry->IsDir())
			{
				Status = TransferDir(c->Entry, c->Str);
				Path[0] = 0;
			}
			else
			{
				Opposite->GetDir(Path, c->Entry->Name);
			}
		}
		else
		{
			strcpy(Path, c->Arg);
		}

		char *File = strrchr(Path, DIR_CHAR);
		if (File AND c->Entry)
		{
			if (FileExists(Path))
			{
				int LocalSize = FileSize(Path);
				bool Resumable = c->Entry->Size > LocalSize;
				
				ResumeDlg Dlg(this, c->Entry, Resumable);
				int Result = Dlg.DoModal();
				if (Result == IDC_OVERWRITE)
				{
					FileDev->DeleteFile(Path);
				}
				else if (Result == IDC_RESUME)
				{
					if (NOT Ftp->ResumeAt(LocalSize))
					{
						if (LgiMsg(this,"This server doesn't support resumable downloads\n"
										"Overwrite instead?", AppName, MB_YESNO) == IDNO)
						{
							return false;
						}
					}
				}
				else
				{
					return false;
				}
			}

			MainWnd->SetTransfer(true);
			Status = Ftp->DownloadFile(Path, c->Entry);
			MainWnd->SetTransfer(false);
		}
	}

	return Status;
}

void GRemoteFS::ProcessClick(IFtpEntry *e, GMouse *m)
{
	if (e AND
		e->Name AND
		Opposite AND
		Ftp)
	{
		if (m->Left())
		{
			if (m->Double())
			{
				// double click
				if (e->IsDir() AND (e->Attributes & IFTP_SYM_LINK) == 0)
				{
					// dir

					// change dir to e->Name
					FtpCmd *Cmd = NEW(FtpCmd(e->Name));
					if (Cmd)
					{
						Cmd->Cmd = TASK_Cd;
						Cmd->Entry = NEW(IFtpEntry(e));
						Cmd->View = this;
						PostCmd(Cmd);
					}

					/* old local code....
					if (Ftp->SetDir(e->Name))
					{
						ListElements();
					}
					*/
				}
				else
				{
					// file
					FtpCmd *Cmd = NEW(FtpCmd);
					if (Cmd)
					{
						Cmd->Cmd = TASK_Transfer;
						Cmd->Entry = e;
						Cmd->View = this;
						PostCmd(Cmd);

						// re-list contents after
						Cmd = NEW(FtpCmd);
						if (Cmd)
						{
							Cmd->Cmd = TASK_ListElements;
							Cmd->View = Opposite;
							PostCmd(Cmd, Opposite->IsCmdLocal(TASK_ListElements) );
						}
					}

					/* old code
					if (TransferFile(TransferCmd))
					{
						Opposite->ListElements();
					}
					*/
				}
			}
			else
			{
				// single click
			}
		}
	}
}

bool GRemoteFS::RenameFile(char *From, char *To)
{
	if (Ftp AND From AND To)
	{
		if (Ftp->RenameFile(From, To))
		{
			ListElements();
		}
	}
	return false;
}

////////////////////////////////////////////////////////////
FtpWorkerThread::FtpWorkerThread()
{
	Loop = true;
	Run(true);
}

int FtpWorkerThread::Main()
{
	while (Loop)
	{
		if (LockWithTimeout(CLOCKS_PER_SEC))
		{
			FtpCmd *Cmd = Cmds.First();
			if (Cmd)
			{
				Cmds.Delete(Cmd);
			}

			Unlock();

			if (Cmd)
			{
				if (Cmd->View)
				{
					Cmd->View->DoCmd(Cmd);
				}
				DeleteObj(Cmd);
			}
			else
			{
				Run(false);
			}
		}
	}

	return 0;
}

void FtpWorkerThread::PostCmd(FtpCmd *Cmd)
{
	if (Cmd AND Lock())
	{
		Cmds.Insert(Cmd);
		Unlock();

		Run(true);
	}
}

void FtpWorkerThread::SetLoop(bool i)
{
	if (Lock())
	{
		Loop = i;
		Unlock();
	}
}

int FtpWorkerThread::GetCmdsLeft()
{
	int Status = 0;
	if (Lock())
	{
		Status = Cmds.GetItems();
		Unlock();
	}
	return Status;
}

